Package: tune
Title: Tidy Tuning Tools
Version: 2.1.0
Authors@R: c(
    person("Max", "Kuhn", , "max@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2402-136X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: The ability to tune models is important. 'tune' contains
    functions and classes to be used in conjunction with other
    'tidymodels' packages for finding reasonable values of
    hyper-parameters in models, preprocessing methods, and post-processing
    steps.
License: MIT + file LICENSE
URL: https://tune.tidymodels.org/, https://github.com/tidymodels/tune
BugReports: https://github.com/tidymodels/tune/issues
Depends: R (>= 4.1)
Imports: cli (>= 3.3.0), dials (>= 1.4.3), dplyr (>= 1.1.0), GauPro (>=
        0.2.15), generics (>= 0.1.2), ggplot2, glue (>= 1.6.2), hardhat
        (>= 1.4.2), lifecycle, parallel, parsnip (>= 1.5.0), purrr (>=
        1.0.0), recipes (>= 1.3.2), rlang (>= 1.1.4), rsample (>=
        1.3.2), tailor (>= 0.1.0), tibble (>= 3.1.0), tidyr (>= 1.2.0),
        tidyselect (>= 1.1.2), vctrs (>= 0.6.1), withr, workflows (>=
        1.3.0), yardstick (>= 1.4.0)
Suggests: C50, censored (>= 0.3.0), covr, future (>= 1.33.0),
        future.apply, kernlab, kknn, knitr, mgcv, mirai (>= 2.4.0),
        modeldata, probably, scales, spelling, splines2, survival,
        testthat (>= 3.0.0), xgboost, xml2
Config/Needs/website: pkgdown, tidymodels, kknn, tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-17 04:15:55 UTC; max
Author: Max Kuhn [aut, cre] (ORCID: <https://orcid.org/0000-0003-2402-136X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Max Kuhn <max@posit.co>
Repository: CRAN
Date/Publication: 2026-04-17 14:30:02 UTC
Built: R 4.6.0; ; 2026-04-25 13:33:06 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",iteration_results)
S3method("[",resample_results)
S3method("[",tune_results)
S3method("names<-",iteration_results)
S3method("names<-",resample_results)
S3method("names<-",tune_results)
S3method(.get_fingerprint,tune_results)
S3method(.use_case_weights_with_yardstick,default)
S3method(.use_case_weights_with_yardstick,hardhat_frequency_weights)
S3method(.use_case_weights_with_yardstick,hardhat_importance_weights)
S3method(augment,last_fit)
S3method(augment,resample_results)
S3method(augment,tune_results)
S3method(autoplot,resample_results)
S3method(autoplot,tune_results)
S3method(collect_extracts,default)
S3method(collect_extracts,tune_results)
S3method(collect_metrics,default)
S3method(collect_metrics,tune_results)
S3method(collect_notes,default)
S3method(collect_notes,tune_results)
S3method(collect_predictions,default)
S3method(collect_predictions,tune_results)
S3method(compute_metrics,default)
S3method(compute_metrics,tune_results)
S3method(dplyr_reconstruct,iteration_results)
S3method(dplyr_reconstruct,resample_results)
S3method(dplyr_reconstruct,tune_results)
S3method(extract_fit_engine,tune_results)
S3method(extract_fit_parsnip,tune_results)
S3method(extract_mold,tune_results)
S3method(extract_preprocessor,tune_results)
S3method(extract_recipe,tune_results)
S3method(extract_spec_parsnip,tune_results)
S3method(extract_workflow,last_fit)
S3method(extract_workflow,tune_results)
S3method(fit_best,default)
S3method(fit_best,tune_results)
S3method(fit_resamples,default)
S3method(fit_resamples,model_spec)
S3method(fit_resamples,workflow)
S3method(int_pctl,tune_results)
S3method(last_fit,default)
S3method(last_fit,model_fit)
S3method(last_fit,model_spec)
S3method(last_fit,workflow)
S3method(load_pkgs,character)
S3method(load_pkgs,model_spec)
S3method(load_pkgs,workflow)
S3method(merge,model_spec)
S3method(merge,recipe)
S3method(min_grid,C5_rules)
S3method(min_grid,boost_tree)
S3method(min_grid,cubist_rules)
S3method(min_grid,linear_reg)
S3method(min_grid,logistic_reg)
S3method(min_grid,mars)
S3method(min_grid,model_spec)
S3method(min_grid,multinom_reg)
S3method(min_grid,nearest_neighbor)
S3method(min_grid,pls)
S3method(min_grid,poisson_reg)
S3method(min_grid,proportional_hazards)
S3method(min_grid,rule_fit)
S3method(obj_sum,resample_results)
S3method(obj_sum,tune_race)
S3method(obj_sum,tune_results)
S3method(outcome_names,formula)
S3method(outcome_names,recipe)
S3method(outcome_names,terms)
S3method(outcome_names,tune_results)
S3method(outcome_names,workflow)
S3method(outcome_names,workflow_variables)
S3method(predict,conf_bound)
S3method(predict,exp_improve)
S3method(predict,last_fit)
S3method(predict,prob_improve)
S3method(predict,tune_results)
S3method(pretty,tune_results)
S3method(print,control_bayes)
S3method(print,control_grid)
S3method(print,control_last_fit)
S3method(print,prob_improve)
S3method(print,tune_results)
S3method(select_best,default)
S3method(select_best,tune_results)
S3method(select_by_one_std_err,default)
S3method(select_by_one_std_err,tune_results)
S3method(select_by_pct_loss,default)
S3method(select_by_pct_loss,tune_results)
S3method(show_best,default)
S3method(show_best,tune_results)
S3method(size_sum,tune_results)
S3method(tune_bayes,default)
S3method(tune_bayes,model_spec)
S3method(tune_bayes,workflow)
S3method(tune_grid,default)
S3method(tune_grid,model_spec)
S3method(tune_grid,workflow)
S3method(vec_cast,data.frame.iteration_results)
S3method(vec_cast,data.frame.resample_results)
S3method(vec_cast,data.frame.tune_results)
S3method(vec_cast,iteration_results.data.frame)
S3method(vec_cast,iteration_results.iteration_results)
S3method(vec_cast,iteration_results.tbl_df)
S3method(vec_cast,resample_results.data.frame)
S3method(vec_cast,resample_results.resample_results)
S3method(vec_cast,resample_results.tbl_df)
S3method(vec_cast,tbl_df.iteration_results)
S3method(vec_cast,tbl_df.resample_results)
S3method(vec_cast,tbl_df.tune_results)
S3method(vec_cast,tune_results.data.frame)
S3method(vec_cast,tune_results.tbl_df)
S3method(vec_cast,tune_results.tune_results)
S3method(vec_ptype2,data.frame.iteration_results)
S3method(vec_ptype2,data.frame.resample_results)
S3method(vec_ptype2,data.frame.tune_results)
S3method(vec_ptype2,iteration_results.data.frame)
S3method(vec_ptype2,iteration_results.iteration_results)
S3method(vec_ptype2,iteration_results.tbl_df)
S3method(vec_ptype2,resample_results.data.frame)
S3method(vec_ptype2,resample_results.resample_results)
S3method(vec_ptype2,resample_results.tbl_df)
S3method(vec_ptype2,tbl_df.iteration_results)
S3method(vec_ptype2,tbl_df.resample_results)
S3method(vec_ptype2,tbl_df.tune_results)
S3method(vec_ptype2,tune_results.data.frame)
S3method(vec_ptype2,tune_results.tbl_df)
S3method(vec_ptype2,tune_results.tune_results)
S3method(vec_restore,iteration_results)
S3method(vec_restore,resample_results)
S3method(vec_restore,tune_results)
export(.catch_and_log)
export(.check_grid)
export(.check_param_objects)
export(.config_key_from_metrics)
export(.create_weight_mapping)
export(.determine_pred_types)
export(.effective_sample_size)
export(.estimate_metrics)
export(.filter_perf_metrics)
export(.get_config_key)
export(.get_data_subsets)
export(.get_extra_col_names)
export(.get_fingerprint)
export(.get_resample_weights)
export(.get_tune_eval_time_target)
export(.get_tune_eval_times)
export(.get_tune_metric_names)
export(.get_tune_metrics)
export(.get_tune_outcome_names)
export(.get_tune_parameter_names)
export(.get_tune_parameters)
export(.get_tune_workflow)
export(.has_preprocessor)
export(.has_preprocessor_formula)
export(.has_preprocessor_recipe)
export(.has_preprocessor_variables)
export(.has_spec)
export(.is_cataclysmic)
export(.load_namespace)
export(.loop_over_all_stages)
export(.loop_over_all_stages2)
export(.make_static)
export(.needs_finalization)
export(.par_fns)
export(.set_workflow)
export(.set_workflow_recipe)
export(.set_workflow_spec)
export(.stash_last_result)
export(.update_model)
export(.update_parallel_over)
export(.update_recipe)
export(.use_case_weights_with_yardstick)
export(.validate_resample_weights)
export(.weighted_sd)
export(add_resample_weights)
export(augment)
export(autoplot)
export(calculate_resample_weights)
export(check_eval_time_arg)
export(check_initial)
export(check_metric_in_tune_results)
export(check_metrics)
export(check_metrics_arg)
export(check_parameters)
export(check_rset)
export(check_time)
export(check_workflow)
export(choose_eval_time)
export(choose_framework)
export(choose_metric)
export(collect_extracts)
export(collect_metrics)
export(collect_notes)
export(collect_predictions)
export(compute_metrics)
export(conf_bound)
export(conf_mat_resampled)
export(control_bayes)
export(control_grid)
export(control_last_fit)
export(control_resamples)
export(coord_obs_pred)
export(empty_ellipses)
export(encode_set)
export(estimate_tune_results)
export(eval_mirai)
export(exp_improve)
export(expo_decay)
export(extract_fit_engine)
export(extract_fit_parsnip)
export(extract_mold)
export(extract_parameter_set_dials)
export(extract_preprocessor)
export(extract_recipe)
export(extract_resample_weights)
export(extract_spec_parsnip)
export(extract_workflow)
export(filter_parameters)
export(finalize_model)
export(finalize_recipe)
export(finalize_tailor)
export(finalize_workflow)
export(finalize_workflow_preprocessor)
export(first_eval_time)
export(first_metric)
export(fit_best)
export(fit_max_value)
export(fit_resamples)
export(forge_from_workflow)
export(future_installed)
export(get_future_workers)
export(get_metric_time)
export(get_mirai_workers)
export(get_parallel_seeds)
export(get_tune_colors)
export(has_non_par_pkgs)
export(initialize_catalog)
export(int_pctl)
export(is_preprocessor)
export(is_recipe)
export(is_workflow)
export(last_fit)
export(load_pkgs)
export(loop_call)
export(maybe_choose_eval_time)
export(message_wrap)
export(metrics_info)
export(min_grid)
export(min_grid.C5_rules)
export(min_grid.boost_tree)
export(min_grid.cubist_rules)
export(min_grid.linear_reg)
export(min_grid.logistic_reg)
export(min_grid.mars)
export(min_grid.multinom_reg)
export(min_grid.nearest_neighbor)
export(min_grid.pls)
export(min_grid.poisson_reg)
export(min_grid.proportional_hazards)
export(min_grid.rule_fit)
export(mirai_installed)
export(new_backend_options)
export(new_bare_tibble)
export(new_iteration_results)
export(outcome_names)
export(parameters)
export(prob_improve)
export(pull_rset_attributes)
export(required_pkgs)
export(schedule_grid)
export(select_best)
export(select_by_one_std_err)
export(select_by_pct_loss)
export(show_best)
export(show_notes)
export(tunable)
export(tune)
export(tune_args)
export(tune_bayes)
export(tune_grid)
export(val_class_and_single)
export(val_class_or_null)
import(rlang)
import(vctrs)
import(workflows)
importFrom(cli,cli_abort)
importFrom(cli,cli_alert)
importFrom(cli,cli_alert_danger)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_alert_warning)
importFrom(cli,cli_inform)
importFrom(cli,cli_warn)
importFrom(cli,qty)
importFrom(dials,encode_unit)
importFrom(dials,is_unknown)
importFrom(dials,parameters)
importFrom(dplyr,all_of)
importFrom(dplyr,anti_join)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_when)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,dplyr_reconstruct)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,inner_join)
importFrom(dplyr,last)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_if)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,sample_n)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,starts_with)
importFrom(dplyr,ungroup)
importFrom(dplyr,vars)
importFrom(generics,augment)
importFrom(generics,min_grid)
importFrom(generics,required_pkgs)
importFrom(generics,tunable)
importFrom(generics,tune_args)
importFrom(ggplot2,aes)
importFrom(ggplot2,aes_)
importFrom(ggplot2,aes_string)
importFrom(ggplot2,autoplot)
importFrom(ggplot2,facet_grid)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_errorbar)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,xlab)
importFrom(ggplot2,ylab)
importFrom(glue,glue)
importFrom(glue,glue_collapse)
importFrom(hardhat,extract_fit_engine)
importFrom(hardhat,extract_fit_parsnip)
importFrom(hardhat,extract_mold)
importFrom(hardhat,extract_parameter_set_dials)
importFrom(hardhat,extract_preprocessor)
importFrom(hardhat,extract_recipe)
importFrom(hardhat,extract_spec_parsnip)
importFrom(hardhat,extract_workflow)
importFrom(hardhat,tune)
importFrom(lifecycle,deprecated)
importFrom(parsnip,get_from_env)
importFrom(parsnip,required_pkgs)
importFrom(purrr,map_int)
importFrom(recipes,all_outcomes)
importFrom(recipes,all_predictors)
importFrom(rlang,"%||%")
importFrom(rlang,":=")
importFrom(rlang,call2)
importFrom(rlang,call_name)
importFrom(rlang,caller_env)
importFrom(rlang,env_get)
importFrom(rlang,eval_tidy)
importFrom(rlang,expr)
importFrom(rlang,is_false)
importFrom(rlang,is_function)
importFrom(rlang,is_missing)
importFrom(rlang,is_quosure)
importFrom(rlang,is_quosures)
importFrom(rlang,ns_env)
importFrom(rlang,quo_get_expr)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(rsample,.get_fingerprint)
importFrom(rsample,int_pctl)
importFrom(stats,dnorm)
importFrom(stats,median)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,qt)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,update)
importFrom(tibble,obj_sum)
importFrom(tibble,size_sum)
importFrom(tidyr,nest)
importFrom(tidyr,unnest)
importFrom(yardstick,accuracy)
importFrom(yardstick,brier_class)
importFrom(yardstick,brier_survival)
importFrom(yardstick,rmse)
importFrom(yardstick,roc_auc)
importFrom(yardstick,rsq)
importFrom(yardstick,weighted_interval_score)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tune <a href='https://tune.tidymodels.org'><img src='man/figures/logo.png' alt='Package hex logo. A black sticker with technicolor dials representing varying parameter values.' align="right" height="139" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/tidymodels/tune/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidymodels/tune/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/tidymodels/tune/branch/main/graph/badge.svg)](https://app.codecov.io/gh/tidymodels/tune)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![CRAN
status](https://www.r-pkg.org/badges/version/tune)](https://CRAN.R-project.org/package=tune)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
<!-- badges: end -->

## Overview

The goal of tune is to facilitate hyperparameter tuning for the
tidymodels packages. It relies heavily on
[recipes](https://recipes.tidymodels.org/),
[parsnip](https://parsnip.tidymodels.org/), and
[dials](https://dials.tidymodels.org/).

## Installation

Install from CRAN:

``` r
install.packages("tune", repos = "http://cran.r-project.org") #or your local mirror
```

or you can install the current development version using:

``` r
# install.packages("pak")
pak::pak("tidymodels/tune")
```

## Examples

There are several package vignettes, as well as articles available at
[tidymodels.org](https://www.tidymodels.org/), demonstrating how to use
tune.

Good places to begin include:

- [Getting started with cell segmentation
  data](https://www.tidymodels.org/start/tuning/)
- [Getting started with Ames housing
  data](https://tune.tidymodels.org/articles/getting_started.html)

More advanced resources available are:

- [Basic grid search for an SVM
  model](https://www.tidymodels.org/learn/work/tune-svm/)
- [Iterative Bayesian optimization of a classification
  model](https://www.tidymodels.org/learn/work/bayes-opt/)
- [Advanced text mining
  example](https://tune.tidymodels.org/articles/extras/text_analysis.html)
- [Notes on optimizations and parallel
  processing](https://tune.tidymodels.org/articles/extras/optimizations.html)
- [Details on acquisition function for scoring parameter
  combinations](https://tune.tidymodels.org/articles/acquisition_functions.html)

## Contributing

This project is released with a [Contributor Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

- For questions and discussions about tidymodels packages, modeling, and
  machine learning, please [post on Posit
  Community](https://forum.posit.co/new-topic?category_id=15&tags=tidymodels,question).

- If you think you have encountered a bug, please [submit an
  issue](https://github.com/tidymodels/tune/issues).

- Either way, learn how to create and share a
  [reprex](https://reprex.tidyverse.org/articles/articles/learn-reprex.html)
  (a minimal, reproducible example), to clearly communicate about your
  code.

- Check out further details on [contributing guidelines for tidymodels
  packages](https://www.tidymodels.org/contribute/) and [how to get
  help](https://www.tidymodels.org/help/).

library(broom)
library(dials)
library(dplyr)
library(ggplot2)
library(modeldata)
library(parsnip)
library(purrr)
library(recipes)
library(rsample)
library(tailor)
library(tidyr)
library(tune)
library(workflows)
library(yardstick)
library(scales)
library(censored)
library(sessioninfo)
library(testthat)
# also will require prodlim, mboost, kknn, and kernlab

# ------------------------------------------------------------------------------
# "mt_*" test objects used in test-predictions.R, test-extract.R, and test-autoplot.R

set.seed(455)
folds <- vfold_cv(mtcars, v = 5)

simple_rec <- recipe(mpg ~ ., data = mtcars)

form <- mpg ~ .

spline_rec <-
  recipe(mpg ~ ., data = mtcars) |>
  step_normalize(all_predictors()) |>
  step_bs(disp, deg_free = tune())

lm_mod <- linear_reg() |> set_engine("lm")

knn_mod <-
  nearest_neighbor(mode = "regression", neighbors = tune()) |>
  set_engine("kknn")

knn_mod_two <-
  nearest_neighbor(
    mode = "regression",
    neighbors = tune("K"),
    weight_func = tune()
  ) |>
  set_engine("kknn")

get_coefs <- function(x) {
  x |>
    extract_fit_parsnip() |>
    tidy()
}

verb <- FALSE
g_ctrl <- control_grid(verbose = verb, save_pred = TRUE, extract = get_coefs)
b_ctrl <- control_bayes(verbose = verb, save_pred = TRUE, extract = get_coefs)

# ------------------------------------------------------------------------------

mt_spln_lm <-
  workflow() |>
  add_recipe(spline_rec) |>
  add_model(lm_mod)

mt_spln_knn <-
  workflow() |>
  add_recipe(spline_rec) |>
  add_model(knn_mod)

mt_knn <-
  workflow() |>
  add_recipe(simple_rec) |>
  add_model(knn_mod)

# ------------------------------------------------------------------------------

set.seed(8825)
mt_spln_lm_grid <-
  tune_grid(mt_spln_lm, resamples = folds, control = g_ctrl)

set.seed(8825)
mt_spln_lm_bo <-
  tune_bayes(
    mt_spln_lm,
    resamples = folds,
    initial = 2,
    iter = 3,
    control = b_ctrl
  )

# ------------------------------------------------------------------------------

set.seed(8825)
mt_spln_knn_grid <-
  tune_grid(
    mt_spln_knn,
    resamples = folds,
    grid = grid_regular(extract_parameter_set_dials(mt_spln_knn)),
    control = g_ctrl
  )

set.seed(8825)
mt_spln_knn_bo <-
  tune_bayes(mt_spln_knn, resamples = folds, iter = 3, control = b_ctrl)

set.seed(8825)
mt_spln_knn_bo_sep <-
  tune_bayes(
    knn_mod_two,
    spline_rec,
    resamples = folds,
    iter = 3,
    control = b_ctrl
  )

# ------------------------------------------------------------------------------

set.seed(8825)
mt_knn_grid <- tune_grid(mt_knn, resamples = folds, control = g_ctrl)

set.seed(8825)
mt_knn_bo <-
  tune_bayes(mt_knn, resamples = folds, iter = 3, control = b_ctrl)

# ------------------------------------------------------------------------------

save(
  list = grep("^mt_", ls(), value = TRUE),
  file = test_path("data", "test_objects.RData"),
  version = 2,
  compress = "xz"
)

# ------------------------------------------------------------------------------
# "knn_*" test objects used in test-predictions.R, test-autoplot.R, test-GP.R
# and test-select_best.R

data(two_class_dat, package = "modeldata")
set.seed(7898)
data_folds <- vfold_cv(two_class_dat, repeats = 5)

two_class_rec <-
  recipe(Class ~ ., data = two_class_dat) |>
  step_normalize(A, B)

knn_model <-
  nearest_neighbor(
    mode = "classification",
    neighbors = tune("K"),
    weight_func = tune(),
    dist_power = tune("exponent")
  ) |>
  set_engine("kknn")

two_class_wflow <-
  workflow() |>
  add_recipe(two_class_rec) |>
  add_model(knn_model)

two_class_set <-
  extract_parameter_set_dials(two_class_wflow) |>
  update(K = neighbors(c(1, 50))) |>
  update(exponent = dist_power(c(1 / 10, 2)))

set.seed(2494)
two_class_grid <-
  two_class_set |>
  grid_max_entropy(size = 10)

class_metrics <- metric_set(roc_auc, accuracy, kap, mcc)

knn_results <-
  tune_grid(
    two_class_wflow,
    resamples = data_folds,
    grid = two_class_grid,
    metrics = class_metrics
  )


knn_set <- two_class_set

knn_gp <-
  tune:::fit_gp(
    dat = collect_metrics(knn_results),
    pset = knn_set,
    metric = "accuracy",
    control = control_bayes()
  )

saveRDS(
  knn_results,
  file = testthat::test_path("data", "knn_results.rds"),
  version = 2,
  compress = "xz"
)

saveRDS(
  two_class_set,
  file = testthat::test_path("data", "knn_set.rds"),
  version = 2,
  compress = "xz"
)

saveRDS(
  two_class_grid,
  file = testthat::test_path("data", "knn_grid.rds"),
  version = 2,
  compress = "xz"
)

saveRDS(
  knn_set,
  file = testthat::test_path("data", "knn_set.rds"),
  version = 2,
  compress = "xz"
)

saveRDS(
  knn_gp,
  file = testthat::test_path("data", "knn_gp.rds"),
  version = 2,
  compress = "xz"
)

# ------------------------------------------------------------------------------
# "svm_*" test objects used in numerous test files

svm_model <-
  svm_poly(
    mode = "classification",
    cost = tune(),
    degree = tune("%^*#"),
    scale_factor = tune()
  ) |>
  set_engine("kernlab")

two_class_wflow <-
  workflow() |>
  add_recipe(two_class_rec) |>
  add_model(svm_model)

two_class_set <-
  extract_parameter_set_dials(two_class_wflow) |>
  update(cost = cost(c(-10, 4)))

set.seed(2494)
two_class_grid <-
  two_class_set |>
  grid_max_entropy(size = 5)

class_only <- metric_set(accuracy, kap, mcc)

svm_results <-
  tune_grid(
    two_class_wflow,
    resamples = data_folds,
    grid = two_class_grid,
    metrics = class_only,
    control = control_grid(save_pred = TRUE)
  )

saveRDS(
  svm_results,
  file = testthat::test_path("data", "svm_results.rds"),
  version = 2,
  compress = "xz"
)

two_class_reg_grid <-
  two_class_set |>
  grid_regular(levels = c(5, 4, 2))

svm_reg_results <-
  tune_grid(
    two_class_wflow,
    resamples = data_folds,
    grid = two_class_reg_grid,
    metrics = class_only,
    control = control_grid(save_pred = TRUE)
  )

saveRDS(
  svm_reg_results,
  file = testthat::test_path("data", "svm_reg_results.rds"),
  version = 2,
  compress = "xz"
)

# ------------------------------------------------------------------------------

set.seed(7898)
data_folds <- vfold_cv(mtcars, repeats = 2)

# ------------------------------------------------------------------------------
# "rcv_results" used in test-autoplot.R, test-select_best.R, and test-estimate.R

base_rec <-
  recipe(mpg ~ ., data = mtcars) |>
  step_normalize(all_predictors())

disp_rec <-
  base_rec |>
  step_bs(disp, degree = tune(), deg_free = tune()) |>
  step_bs(wt, degree = tune("wt degree"), deg_free = tune("wt df"))

lm_model <-
  linear_reg(mode = "regression") |>
  set_engine("lm")

cars_wflow <-
  workflow() |>
  add_recipe(disp_rec) |>
  add_model(lm_model)

cars_set <-
  cars_wflow |>
  extract_parameter_set_dials() |>
  update(degree = degree_int(1:2)) |>
  update(deg_free = deg_free(c(2, 10))) |>
  update(`wt degree` = degree_int(1:2)) |>
  update(`wt df` = deg_free(c(2, 10)))

set.seed(255)
cars_grid <-
  cars_set |>
  grid_regular(levels = c(3, 2, 3, 2))

rcv_results <-
  tune_grid(
    cars_wflow,
    resamples = data_folds,
    grid = cars_grid,
    control = control_grid(verbose = FALSE, save_pred = TRUE)
  )

saveRDS(
  rcv_results,
  file = testthat::test_path("data", "rcv_results.rds"),
  version = 2,
  compress = "xz"
)


# ------------------------------------------------------------------------------
# Object classed with `resample_results` for use in vctrs/dplyr tests

set.seed(6735)

folds <- vfold_cv(mtcars, v = 3)

rec <- recipe(mpg ~ ., data = mtcars)

mod <- linear_reg() |>
  set_engine("lm")

lm_resamples <- fit_resamples(mod, rec, folds)

lm_resamples

saveRDS(
  lm_resamples,
  file = testthat::test_path("data", "lm_resamples.rds"),
  version = 2,
  compress = "xz"
)

# ------------------------------------------------------------------------------
# Object classed with `iteration_results` for use in vctrs/dplyr tests

set.seed(7898)
folds <- vfold_cv(mtcars, v = 2)

rec <- recipe(mpg ~ ., data = mtcars) |>
  step_normalize(all_predictors()) |>
  step_ns(disp, deg_free = tune())

mod <- linear_reg(mode = "regression") |>
  set_engine("lm")

wflow <- workflow() |>
  add_recipe(rec) |>
  add_model(mod)

set.seed(2934)
lm_bayes <- tune_bayes(wflow, folds, initial = 4, iter = 3)

saveRDS(
  lm_bayes,
  file = testthat::test_path("data", "lm_bayes.rds"),
  version = 2,
  compress = "xz"
)

# ------------------------------------------------------------------------------
# A single survival model

set.seed(1)
sim_dat <- prodlim::SimSurv(200) |>
  mutate(event_time = Surv(time, event)) |>
  select(event_time, X1, X2)

set.seed(2)
sim_rs <- vfold_cv(sim_dat)

time_points <- c(10, 1, 5, 15)

boost_spec <-
  boost_tree(trees = tune()) |>
  set_mode("censored regression") |>
  set_engine("mboost")

srv_mtr <-
  metric_set(
    brier_survival,
    roc_auc_survival,
    brier_survival_integrated,
    concordance_survival
  )

set.seed(2193)
surv_boost_tree_res <-
  boost_spec |>
  tune_grid(
    event_time ~ X1 + X2,
    resamples = sim_rs,
    grid = tibble(trees = c(1, 5, 10, 20, 100)),
    metrics = srv_mtr,
    eval_time = time_points
  )

saveRDS(
  surv_boost_tree_res,
  file = testthat::test_path("data", "surv_boost_tree_res.rds"),
  version = 2,
  compress = "xz"
)


# ------------------------------------------------------------------------------

sessioninfo::session_info()

if (!interactive()) {
  q("no")
}

# ------------------------------------------------------------------------------
# Expected results produced by tune 1.3.0 to compare to later implementations
# This example tunes a model over a single submodel parameter.

library(tidymodels)
library(sessioninfo)

# ------------------------------------------------------------------------------

tidymodels_prefer()
theme_set(theme_bw())
options(pillar.advice = FALSE, pillar.min_title_chars = Inf)

# ------------------------------------------------------------------------------

set.seed(1)
dat <- sim_classification(1000)
rs <- vfold_cv(dat)

# ------------------------------------------------------------------------------

mod <- nearest_neighbor(neighbors = tune(), weight_func = "triangular") |>
  set_mode("classification")

submodel_only_wflow <- workflow(class ~ ., mod)

submodel_only_grid <- tibble(neighbors = 3:10)

submodel_only_res <-
  submodel_only_wflow |>
  tune_grid(
    resamples = rs,
    grid = submodel_only_grid,
    control = control_grid(save_pred = TRUE)
  )

submodel_only_metrics <- collect_metrics(submodel_only_res, summarize = FALSE)
submodel_only_pred <- collect_predictions(submodel_only_res, summarize = FALSE)

# ------------------------------------------------------------------------------

save(
  submodel_only_metrics,
  submodel_only_pred,
  file = "submodel_only_example.RData"
)

# ------------------------------------------------------------------------------

sessioninfo::session_info()

if (!interactive()) {
  q("no")
}

# ------------------------------------------------------------------------------
# Expected results produced by tune 1.3.0 to compare to later implementations
# This example tunes a preprocess and a model. The model has two parameters, one
# of which is a submodel parameter. A regular design was used that is unbalanced
# by removing the first candidate.

library(tidymodels)
library(sessioninfo)

# ------------------------------------------------------------------------------

tidymodels_prefer()
theme_set(theme_bw())
options(pillar.advice = FALSE, pillar.min_title_chars = Inf)

# ------------------------------------------------------------------------------

set.seed(1)
dat <- sim_regression(1000)
rs <- vfold_cv(dat)

# ------------------------------------------------------------------------------

rec <- recipe(outcome ~ ., data = dat) |>
  step_pca(all_numeric_predictors(), num_comp = tune())

mod <- nearest_neighbor(neighbors = tune("k"), weight_func = tune()) |>
  set_mode("regression")

submodel_wflow <- workflow(rec, mod)

# submodel_grd <-
#   submodel_wflow |>
#   extract_parameter_set_dials() |>
#   update(
#     k = neighbors(c(4, 20)),
#     num_comp = num_comp(c(2, 10))
#   ) |>
#   grid_regular(levels = c(4, 3, 2)) |>
#   slice(-1)

# fmt: skip
submodel_grid <-
  tibble::tribble(
    ~k,   ~weight_func, ~num_comp,
     9L,  "rectangular",        2L,
    14L,  "rectangular",        2L,
    20L,  "rectangular",        2L,
     4L,   "triangular",        2L,
     9L,   "triangular",        2L,
    14L,   "triangular",        2L,
    20L,   "triangular",        2L,
     4L, "epanechnikov",        2L,
     9L, "epanechnikov",        2L,
    14L, "epanechnikov",        2L,
    20L, "epanechnikov",        2L,
     4L,  "rectangular",       10L,
     9L,  "rectangular",       10L,
    14L,  "rectangular",       10L,
    20L,  "rectangular",       10L,
     4L,   "triangular",       10L,
     9L,   "triangular",       10L,
    14L,   "triangular",       10L,
    20L,   "triangular",       10L,
     4L, "epanechnikov",       10L,
     9L, "epanechnikov",       10L,
    14L, "epanechnikov",       10L,
    20L, "epanechnikov",       10L
  )

submodel_res <-
  submodel_wflow |>
  tune_grid(
    resamples = rs,
    grid = submodel_grid,
    control = control_grid(save_pred = TRUE)
  )

submodel_metrics <- collect_metrics(submodel_res, summarize = FALSE)
submodel_pred <- collect_predictions(submodel_res, summarize = FALSE)

# ------------------------------------------------------------------------------

save(submodel_metrics, submodel_pred, file = "submodel_example.RData")

# ------------------------------------------------------------------------------

sessioninfo::session_info()

if (!interactive()) {
  q("no")
}

# ------------------------------------------------------------------------------
# Expected results produced by tune 1.3.0 to compare to later implementations
# This "simple" example is for model-only results with no submodels

library(tidymodels)
library(sessioninfo)

# ------------------------------------------------------------------------------

tidymodels_prefer()
theme_set(theme_bw())
options(pillar.advice = FALSE, pillar.min_title_chars = Inf)

# ------------------------------------------------------------------------------

set.seed(1)
dat <- sim_regression(1000)
rs <- vfold_cv(dat)

# ------------------------------------------------------------------------------

mod <- nearest_neighbor(neighbors = 11, weight_func = tune()) |>
  set_mode("regression")

simple_wflow <- workflow(outcome ~ ., mod)

simple_grid <-
  tibble::tribble(
    ~weight_func   ,
    "rectangular"  ,
    "triangular"   ,
    "epanechnikov"
  )

simple_res <-
  simple_wflow |>
  tune_grid(
    resamples = rs,
    grid = simple_grid,
    control = control_grid(save_pred = TRUE)
  )

simple_metrics <- collect_metrics(simple_res, summarize = FALSE)
simple_pred <- collect_predictions(simple_res, summarize = FALSE)

# ------------------------------------------------------------------------------

save(simple_metrics, simple_pred, file = "simple_example.RData")

# ------------------------------------------------------------------------------

sessioninfo::session_info()

if (!interactive()) {
  q("no")
}

#' Control aspects of the grid search process
#'
#' @inheritParams control_bayes
#'
#' @inheritSection collect_predictions Hyperparameters and extracted objects
#'
#' @details
#'
#' For `extract`, this function can be used to output the model object, the
#' recipe (if used), or some components of either or both. When evaluated, the
#' function's sole argument has a fitted workflow If the formula method is used,
#' the recipe element will be `NULL`.
#'
#' The results of the `extract` function are added to a list column in the
#' output called `.extracts`. Each element of this list is a tibble with tuning
#' parameter column and a list column (also called `.extracts`) that contains
#' the results of the function. If no extraction function is used, there is no
#' `.extracts` column in the resulting object. See [tune_bayes()] for more
#' specific details.
#'
#' Note that for [collect_predictions()], it is possible that each row of the
#'  original data point might be represented multiple times per tuning
#'  parameter. For example, if the bootstrap or repeated cross-validation are
#'  used, there will be multiple rows since the sample data point has been
#'  evaluated multiple times. This may cause issues when merging the predictions
#'  with the original data.
#'
#' [control_resamples()] is an alias for [control_grid()] and is meant to be
#' used with [fit_resamples()].
#' @export
control_grid <- function(
  verbose = FALSE,
  allow_par = TRUE,
  extract = NULL,
  save_pred = FALSE,
  pkgs = NULL,
  save_workflow = FALSE,
  event_level = "first",
  parallel_over = NULL,
  backend_options = NULL,
  workflow_size = 100.0
) {
  # Any added arguments should also be added in superset control functions
  # in other packages

  # add options for seeds per resample
  check_bool(verbose)
  check_bool(allow_par)
  check_bool(save_pred)
  check_bool(save_workflow)
  check_string(event_level)
  check_character(pkgs, allow_null = TRUE)
  check_function(extract, allow_null = TRUE)
  check_number_decimal(workflow_size)

  val_parallel_over(parallel_over, "control_grid()")

  res <- list(
    verbose = verbose,
    allow_par = allow_par,
    extract = extract,
    save_pred = save_pred,
    pkgs = pkgs,
    save_workflow = save_workflow,
    event_level = event_level,
    parallel_over = parallel_over,
    backend_options = backend_options,
    workflow_size = workflow_size
  )

  class(res) <- c("control_grid", "control_resamples")
  res
}

# Helper function to print control settings using cli
print_control_settings <- function(x, default = FALSE, defaults = NULL) {
  # Get the fields to print
  fields <- names(x)

  # Optionally reduce to only non-defaults
  if (default && !is.null(defaults)) {
    fields <- fields[
      !vapply(
        fields,
        function(field) {
          identical(x[[field]], defaults[[field]])
        },
        logical(1)
      )
    ]
  }

  # Build formatted lines for each field
  for (field in fields) {
    value <- x[[field]]

    if (is.function(value)) {
      cli::cli_bullets(c(" " = "{.arg {field}}: <function>"))
    } else if (inherits(value, "tune_backend_options")) {
      cli::cli_bullets(c(" " = "{.arg {field}}: <backend_options>"))
    } else if (is.null(value)) {
      cli::cli_bullets(c(" " = "{.arg {field}}: NULL"))
    } else {
      cli::cli_bullets(c(" " = "{.arg {field}}: {.val {value}}"))
    }
  }
}

#' @export
print.control_grid <- function(x, default = FALSE, ...) {
  cli::cli_text("{.emph Grid/resamples control object}")
  defaults <- control_grid()
  print_control_settings(x, default = default, defaults = defaults)
  invisible(x)
}

#' @rdname control_grid
#' @export
control_resamples <- control_grid

#' Control aspects of the last fit process
#'
#' @inheritParams control_grid
#'
#' @details
#'
#' [control_last_fit()] is a wrapper around [control_resamples()] and is meant
#'   to be used with [last_fit()].
#'
#' @export
control_last_fit <- function(
  verbose = FALSE,
  event_level = "first",
  allow_par = FALSE
) {
  # Any added arguments should also be added in superset control functions
  # in other packages

  extr <- function(x) x
  control <-
    control_resamples(
      verbose = verbose,
      allow_par = allow_par,
      event_level = event_level,
      extract = extr,
      save_pred = TRUE,
      save_workflow = FALSE
    )
  class(control) <- c("control_last_fit", class(control))
  control
}

#' @export
print.control_last_fit <- function(x, default = FALSE, ...) {
  cli::cli_text("{.emph Last fit control object}")
  defaults <- control_last_fit()
  print_control_settings(x, default = default, defaults = defaults)
  invisible(x)
}

# ------------------------------------------------------------------------------

#' Control aspects of the Bayesian search process
#'
#' @param verbose A logical for logging results (other than warnings and errors,
#'   which are always shown) as they are generated during training in a single
#'   R process. When using most parallel backends, this argument typically will
#'   not result in any logging. If using a dark IDE theme, some logging messages
#'   might be hard to see; try setting the `tidymodels.dark` option with
#'   `options(tidymodels.dark = TRUE)` to print lighter colors.
#' @param verbose_iter A logical for logging results of the Bayesian search
#'   process. Defaults to FALSE. If using a dark IDE theme, some logging
#'   messages might be hard to see; try setting the `tidymodels.dark` option
#'   with `options(tidymodels.dark = TRUE)` to print lighter colors.
#' @param no_improve The integer cutoff for the number of iterations without
#'   better results.
#' @param uncertain The number of iterations with no improvement before an
#'  uncertainty sample is created where a sample with high predicted variance is
#'  chosen (i.e., in a region that has not yet been explored). The iteration
#'  counter is reset after each uncertainty sample. For example, if `uncertain =
#'  10`, this condition is triggered every 10 samples with no improvement.
#' @param seed An integer for controlling the random number stream. Tuning
#' functions are sensitive to both the state of RNG set outside of tuning
#' functions with `set.seed()` as well as the value set here. The value of the
#' former determines RNG for the higher-level tuning process, like grid
#' generation and setting the value of this argument if left as default. The
#' value of this argument determines RNG state in workers for each iteration
#' of model fitting, determined by the value of `parallel_over`.
#' @param time_limit A number for the minimum number of _minutes_ (elapsed) that
#'   the function should execute. The elapsed time is evaluated at internal
#'   checkpoints and, if over time, the results at that time are returned (with
#'   a warning). This means that the `time_limit` is not an exact limit, but a
#'   minimum time limit.
#'
#'   Note that timing begins immediately on execution. Thus, if the
#'   `initial` argument to [tune_bayes()] is supplied as a number, the elapsed
#'   time will include the time needed to generate initialization results.
#' @param extract An optional function with at least one argument (or `NULL`)
#'   that can be used to retain arbitrary objects from the model fit object,
#'   recipe, or other elements of the workflow.
#' @param save_pred A logical for whether the out-of-sample predictions should
#'   be saved for each model _evaluated_.
#' @param pkgs An optional character string of R package names that should be
#'   loaded (by namespace) during parallel processing.
#' @param save_workflow A logical for whether the workflow should be appended
#'  to the output as an attribute.
#' @param save_gp_scoring A logical to save the intermediate Gaussian process
#'   models for each iteration of the search. These are saved to
#'  `tempdir()` with names `gp_candidates_{i}.RData` where `i` is the iteration.
#'  These results are deleted when the R session ends. This option is only
#'  useful for teaching purposes.
#' @param event_level A single string containing either `"first"` or `"second"`.
#'   This argument is passed on to yardstick metric functions when any type
#'   of class prediction is made, and specifies which level of the outcome
#'   is considered the "event".
#' @param parallel_over A single string containing either `"resamples"` or
#'   `"everything"` describing how to use parallel processing. Alternatively,
#'   `NULL` is allowed, which chooses between `"resamples"` and `"everything"`
#'   automatically.
#'
#'   If `"resamples"`, then tuning will be performed in parallel over resamples
#'   alone. Within each resample, the preprocessor (i.e. recipe or formula) is
#'   processed once, and is then reused across all models that need to be fit.
#'
#'   If `"everything"`, then tuning will be performed in parallel at two levels.
#'   An outer parallel loop will iterate over resamples. Additionally, an
#'   inner parallel loop will iterate over all unique combinations of
#'   preprocessor and model tuning parameters for that specific resample. This
#'   will result in the preprocessor being re-processed multiple times, but
#'   can be faster if that processing is extremely fast.
#'
#'   If `NULL`, chooses `"resamples"` if there are more than one resample,
#'   otherwise chooses `"everything"` to attempt to maximize core utilization.
#'
#'   Note that switching between `parallel_over` strategies is not guaranteed
#'   to use the same random number generation schemes. However, re-tuning a
#'   model using the same `parallel_over` strategy is guaranteed to be
#'   reproducible between runs.
#' @param backend_options An object of class `"tune_backend_options"` as created
#'   by `tune::new_backend_options()`, used to pass arguments to specific tuning
#'   backend. Defaults to `NULL` for default backend options.
#' @param allow_par A logical to allow parallel processing (if a parallel
#'   backend is registered).
#' @param workflow_size A non-negative number (in MB) that is used as a
#' threshold for a warning regarding the size of the workflow. Only used when
#' `save_workflow = TRUE`.
#'
#' @inheritSection collect_predictions Hyperparameters and extracted objects
#'
#' @details
#'
#' For `extract`, this function can be used to output the model object, the
#' recipe (if used), or some components of either or both. When evaluated, the
#' function's sole argument has a fitted workflow If the formula method is used,
#' the recipe element will be `NULL`.
#'
#' The results of the `extract` function are added to a list column in the
#' output called `.extracts`. Each element of this list is a tibble with tuning
#' parameter column and a list column (also called `.extracts`) that contains
#' the results of the function. If no extraction function is used, there is no
#' `.extracts` column in the resulting object. See [tune_bayes()] for more
#' specific details.
#'
#' Note that for [collect_predictions()], it is possible that each row of the
#'  original data point might be represented multiple times per tuning
#'  parameter. For example, if the bootstrap or repeated cross-validation are
#'  used, there will be multiple rows since the sample data point has been
#'  evaluated multiple times. This may cause issues when merging the predictions
#'  with the original data.
#' @export
control_bayes <-
  function(
    verbose = FALSE,
    verbose_iter = FALSE,
    no_improve = 10L,
    uncertain = Inf,
    seed = sample.int(10^5, 1),
    extract = NULL,
    save_pred = FALSE,
    time_limit = NA,
    pkgs = NULL,
    save_workflow = FALSE,
    save_gp_scoring = FALSE,
    event_level = "first",
    parallel_over = NULL,
    backend_options = NULL,
    allow_par = TRUE,
    workflow_size = 100.0
  ) {
    # Any added arguments should also be added in superset control functions
    # in other packages

    # add options for seeds per resample
    check_bool(verbose)
    check_bool(verbose_iter)
    check_bool(allow_par)
    check_bool(save_pred)
    check_bool(save_workflow)
    check_bool(save_gp_scoring)
    check_character(pkgs, allow_null = TRUE)
    check_function(extract, allow_null = TRUE)
    check_number_whole(no_improve, min = 0, allow_infinite = TRUE)
    check_number_whole(uncertain, min = 0, allow_infinite = TRUE)
    check_number_whole(seed)
    check_number_decimal(workflow_size)

    check_time_limit_arg(time_limit)

    val_parallel_over(parallel_over, "control_bayes()")

    if (!is.infinite(uncertain) && uncertain > no_improve) {
      cli::cli_warn(
        "Uncertainty sample scheduled after {uncertain} poor iterations but the
         search will stop after {no_improve}."
      )
    }

    res <-
      list(
        verbose = verbose,
        verbose_iter = verbose_iter,
        allow_par = allow_par,
        no_improve = no_improve,
        uncertain = uncertain,
        seed = seed,
        extract = extract,
        save_pred = save_pred,
        time_limit = time_limit,
        pkgs = pkgs,
        save_workflow = save_workflow,
        save_gp_scoring = save_gp_scoring,
        event_level = event_level,
        parallel_over = parallel_over,
        backend_options = backend_options,
        workflow_size = workflow_size
      )

    class(res) <- "control_bayes"
    res
  }

#' @export
print.control_bayes <- function(x, default = FALSE, ...) {
  cli::cli_text("{.emph Bayes control object}")
  defaults <- control_bayes()
  print_control_settings(x, default = default, defaults = defaults)
  invisible(x)
}

# ------------------------------------------------------------------------------

val_parallel_over <- function(parallel_over, where) {
  check_string(parallel_over, allow_null = TRUE)
  if (!is.null(parallel_over)) {
    rlang::arg_match0(
      parallel_over,
      c("resamples", "everything"),
      "parallel_over"
    )
  }

  invisible(NULL)
}

#' @export
#' @keywords internal
#' @rdname control_grid
new_backend_options <- function(..., class = character()) {
  out <- rlang::list2(...)

  if (any(rlang::names2(out) == "")) {
    cli::cli_abort("All backend options must be named.")
  }

  structure(out, class = c(class, "tune_backend_options"))
}

test_that("`event_level` is passed through in tune_grid()", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kernlab")

  set.seed(123)

  data("two_class_dat", package = "modeldata")
  dat <- two_class_dat[1:50, ]

  spec <- parsnip::svm_linear(mode = "classification", cost = tune()) |>
    parsnip::set_engine("kernlab")

  control <- control_grid(event_level = "second", save_pred = TRUE)
  resamples <- rsample::bootstraps(dat, times = 1)
  set <- yardstick::metric_set(yardstick::roc_auc, yardstick::sens)
  grid <- tibble::tibble(cost = 2L)

  result <- tune_grid(
    spec,
    Class ~ .,
    resamples = resamples,
    metrics = set,
    grid = grid,
    control = control
  )

  metrics <- result$.metrics[[1]]
  estimates <- metrics$.estimate
  predictions <- result$.predictions[[1]]

  expected_sens <- yardstick::sens_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_class,
    event_level = "second"
  )

  expected_roc_auc <- yardstick::roc_auc_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_Class2,
    event_level = "second"
  )

  expect_identical(estimates[metrics$.metric == "sens"], expected_sens)
  expect_identical(estimates[metrics$.metric == "roc_auc"], expected_roc_auc)
})

test_that("`event_level` is passed through in fit_resamples()", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kernlab")

  set.seed(123)

  data("two_class_dat", package = "modeldata")
  dat <- two_class_dat[1:50, ]

  spec <- parsnip::svm_linear(mode = "classification", cost = 2) |>
    parsnip::set_engine("kernlab")

  control <- control_resamples(event_level = "second", save_pred = TRUE)
  resamples <- rsample::bootstraps(dat, times = 1)
  set <- yardstick::metric_set(yardstick::roc_auc, yardstick::sens)

  result <- fit_resamples(
    spec,
    Class ~ .,
    resamples = resamples,
    metrics = set,
    control = control
  )

  metrics <- result$.metrics[[1]]
  estimates <- metrics$.estimate
  predictions <- result$.predictions[[1]]

  expected_sens <- yardstick::sens_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_class,
    event_level = "second"
  )

  expected_roc_auc <- yardstick::roc_auc_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_Class2,
    event_level = "second"
  )

  expect_identical(estimates[metrics$.metric == "sens"], expected_sens)
  expect_identical(estimates[metrics$.metric == "roc_auc"], expected_roc_auc)
})

test_that("`event_level` is passed through in tune_bayes()", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kernlab")

  set.seed(123)

  data("two_class_dat", package = "modeldata")
  dat <- two_class_dat[1:50, ]

  spec <- parsnip::svm_linear(mode = "classification", cost = tune()) |>
    parsnip::set_engine("kernlab")

  control_grid <- control_grid(save_pred = TRUE)
  control <- control_bayes(event_level = "second", save_pred = TRUE)
  resamples <- rsample::bootstraps(dat, times = 1)
  set <- yardstick::metric_set(yardstick::roc_auc, yardstick::sens)

  result <- tune_bayes(
    spec,
    Class ~ .,
    resamples = resamples,
    metrics = set,
    iter = 1,
    control = control
  )

  result <- result[result$.iter == 1, ]

  metrics <- result$.metrics[[1]]
  estimates <- metrics$.estimate
  predictions <- result$.predictions[[1]]

  expected_sens <- yardstick::sens_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_class,
    event_level = "second"
  )

  expected_roc_auc <- yardstick::roc_auc_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_Class2,
    event_level = "second"
  )

  expect_identical(estimates[metrics$.metric == "sens"], expected_sens)
  expect_identical(estimates[metrics$.metric == "roc_auc"], expected_roc_auc)
})

test_that("`event_level` is passed through in last_fit()", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kernlab")

  set.seed(123)

  data("two_class_dat", package = "modeldata")
  dat <- two_class_dat[1:10, ]

  spec <- parsnip::svm_linear(mode = "classification", cost = 0.5) |>
    parsnip::set_engine("kernlab")

  control <- control_last_fit(event_level = "second")
  split <- rsample::initial_split(dat)
  set <- yardstick::metric_set(yardstick::roc_auc, yardstick::sens)

  result <- last_fit(
    spec,
    Class ~ .,
    split = split,
    metrics = set,
    control = control
  )

  metrics <- result$.metrics[[1]]
  estimates <- metrics$.estimate
  predictions <- result$.predictions[[1]]

  expected_sens <- yardstick::sens_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_class,
    event_level = "second"
  )

  expected_roc_auc <- yardstick::roc_auc_vec(
    truth = predictions$Class,
    estimate = predictions$.pred_Class2,
    event_level = "second"
  )

  expect_identical(estimates[metrics$.metric == "sens"], expected_sens)
  expect_identical(estimates[metrics$.metric == "roc_auc"], expected_roc_auc)
})

test_that("basic functionality", {
  svm_reg_results <- readRDS(test_path("data", "svm_reg_results.rds"))
  load(test_path("data", "test_objects.RData"))
  options(width = 200, pillar.advice = FALSE, pillar.min_title_chars = Inf)

  filtered_grid <- filter_parameters(
    svm_reg_results,
    parameters = tibble::tibble(`%^*#` = 1)
  )
  expect_true(all(purrr::map_lgl(
    filtered_grid$.metrics,
    \(.x) all(.x$`%^*#` == 1)
  )))
  expect_true(all(purrr::map_lgl(
    filtered_grid$.predictions,
    \(.x) all(.x$`%^*#` == 1)
  )))

  best_param <- select_best(svm_reg_results, metric = "accuracy")
  best_grid <- filter_parameters(svm_reg_results, parameters = best_param)
  expect_true(all(purrr::map_lgl(
    best_grid$.metrics,
    \(.x) all(.x$cost == best_param$cost)
  )))
  expect_true(all(purrr::map_lgl(
    best_grid$.metrics,
    \(.x) all(.x$`%^*#` == best_param$`%^*#`)
  )))
  expect_true(all(purrr::map_lgl(
    best_grid$.metrics,
    \(.x) all(.x$scale_factor == best_param$scale_factor)
  )))
  expect_true(all(purrr::map_lgl(
    best_grid$.predictions,
    \(.x) all(.x$cost == best_param$cost)
  )))
  expect_true(all(purrr::map_lgl(
    best_grid$.predictions,
    \(.x) all(.x$`%^*#` == best_param$`%^*#`)
  )))
  expect_true(all(purrr::map_lgl(
    best_grid$.predictions,
    \(.x) all(.x$scale_factor == best_param$scale_factor)
  )))

  filtered_grid <- filter_parameters(
    mt_knn_bo,
    parameters = tibble::tibble(neighbors = 1)
  )
  expect_true(all(purrr::map_lgl(
    filtered_grid$.metrics,
    \(.x) all(.x$neighbors == 1)
  )))
  expect_true(all(purrr::map_lgl(
    filtered_grid$.predictions,
    \(.x) all(.x$neighbors == 1)
  )))

  best_param <- select_best(mt_knn_bo, metric = "rmse")
  best_grid <- filter_parameters(mt_knn_bo, parameters = best_param)
  expect_true(all(purrr::map_lgl(
    best_grid$.metrics,
    \(.x) all(.x$neighbors == best_param$neighbors)
  )))
  expect_true(all(purrr::map_lgl(
    best_grid$.predictions,
    \(.x) all(.x$neighbors == best_param$neighbors)
  )))

  rec_tune_1 <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_normalize(recipes::all_predictors()) |>
    recipes::step_pca(recipes::all_predictors(), num_comp = tune())

  lm_mod <- parsnip::linear_reg() |> parsnip::set_engine("lm")

  set.seed(363)
  mt_folds <- rsample::vfold_cv(mtcars, v = 3)

  extr_rec <- function(x) {
    extract_recipe(x) |> tidy(number = 2)
  }
  lm_res <-
    workflow() |>
    add_recipe(rec_tune_1) |>
    add_model(lm_mod) |>
    tune_grid(
      resamples = mt_folds,
      control = control_grid(extract = extr_rec),
      grid = tibble::tibble(num_comp = 1:2)
    )

  filtered_grid <- filter_parameters(
    lm_res,
    parameters = tibble::tibble(num_comp = 1)
  )
  expect_true(all(purrr::map_lgl(
    filtered_grid$.metrics,
    \(.x) all(.x$num_comp == 1)
  )))
  expect_true(all(purrr::map_lgl(
    filtered_grid$.extracts,
    \(.x) all(.x$num_comp == 1)
  )))

  best_param <- select_best(lm_res, metric = "rmse")
  best_grid <- filter_parameters(lm_res, parameters = best_param)
  expect_true(all(purrr::map_lgl(
    best_grid$.metrics,
    \(.x) all(.x$num_comp == best_param$num_comp)
  )))
  expect_true(all(purrr::map_lgl(
    best_grid$.extracts,
    \(.x) all(.x$num_comp == best_param$num_comp)
  )))

  filtered_grid <- filter_parameters(svm_reg_results, `%^*#` == 1)
  expect_true(all(purrr::map_lgl(
    filtered_grid$.metrics,
    \(.x) all(.x$`%^*#` == 1)
  )))
  expect_true(all(purrr::map_lgl(
    filtered_grid$.predictions,
    \(.x) all(.x$`%^*#` == 1)
  )))
})

# ------------------------------------------------------------------------------

test_that("bad inputs", {
  svm_reg_results <- readRDS(test_path("data", "svm_reg_results.rds"))

  expect_snapshot(error = TRUE, {
    filter_parameters(
      collect_metrics(svm_reg_results),
      parameters = tibble::tibble(`%^*#` = 1)
    )
  })
  expect_snapshot(error = TRUE, {
    filter_parameters(svm_reg_results, parameters = tibble::tibble(soup = 1))
  })
  expect_snapshot(error = TRUE, {
    filter_parameters(svm_reg_results, tibble::tibble(soup = 1))
  })
  expect_snapshot(error = TRUE, {
    filter_parameters(svm_reg_results, cost < 1, tibble::tibble(soup = 1))
  })
  expect_snapshot(error = TRUE, {
    filter_parameters(
      svm_reg_results,
      parameters = tibble::tibble(`%^*#` = 1 / 3)
    )
  })
  expect_snapshot(
    filter_parameters(
      svm_reg_results,
      parameters = tibble::tibble(`%^*#` = 1, soup = 2)
    )
  )
  expect_snapshot(
    res <- filter_parameters(
      svm_reg_results,
      parameters = tibble::tibble(`%^*#` = 1, soup = 2, boop = 3)
    )
  )
  expect_snapshot(
    res <- filter_parameters(
      svm_reg_results,
      parameters = tibble::tibble(`%^*#` = 1, soup = 2, boop = 3, loop = 4)
    )
  )
})

test_that("tune recipe only", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()
  set.seed(363)
  mt_folds <- rsample::vfold_cv(mtcars, v = 5)

  extr_1_1 <- function(x) {
    extract_recipe(x) |> tidy(number = 2)
  }
  before_kind <- RNGkind()[[1]]
  expect_no_error(
    res_1_1 <-
      workflow() |>
      add_recipe(helper_objects$rec_tune_1) |>
      add_model(helper_objects$lm_mod) |>
      tune_grid(
        resamples = mt_folds,
        control = control_grid(extract = extr_1_1)
      )
  )
  after_kind <- RNGkind()[[1]]
  expect_equal(before_kind, after_kind)
  expect_no_error(extract_1_1 <- dplyr::bind_rows(res_1_1$.extracts))

  expect_true(all(names(extract_1_1) == c("num_comp", ".extracts", ".config")))
  expect_true(
    all(purrr::map_lgl(extract_1_1$.extracts, tibble::is_tibble)),
  )
})

# ------------------------------------------------------------------------------

test_that("tune model only", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()
  set.seed(363)
  mt_folds <- rsample::vfold_cv(mtcars, v = 5)

  extr_2_1 <- function(x) {
    mod <- extract_fit_engine(x)
    tibble(index = mod@alphaindex[[1]], estimate = mod@coef[[1]])
  }

  expect_no_error(
    res_2_1 <-
      workflow() |>
      add_recipe(helper_objects$rec_no_tune_1) |>
      add_model(helper_objects$svm_mod) |>
      tune_grid(
        resamples = mt_folds,
        grid = 2,
        control = control_grid(extract = extr_2_1)
      )
  )
  expect_no_error(extract_2_1 <- dplyr::bind_rows(res_2_1$.extracts))

  expect_true(all(names(extract_2_1) == c("cost", ".extracts", ".config")))
  expect_true(
    all(purrr::map_lgl(extract_2_1$.extracts, tibble::is_tibble)),
  )
  expect_true(
    all(purrr::map_lgl(
      extract_2_1$.extracts,
      \(.x) all(names(.x) == c("index", "estimate"))
    )),
  )

  extr_2_2 <- function(x) {
    tibble(is_null_rec = is.null(extract_recipe(x)))
  }

  # should not fail:
  expect_no_error(
    res_2_2 <-
      workflow() |>
      add_recipe(helper_objects$rec_tune_1) |>
      add_model(helper_objects$lm_mod) |>
      tune_grid(
        resamples = mt_folds,
        grid = 2,
        control = control_grid(extract = extr_2_2)
      )
  )

  expect_no_error(
    extract_2_2 <-
      dplyr::bind_rows(res_2_2$.extracts) |>
      tidyr::unnest(cols = c(.extracts))
  )
  expect_true(all(!extract_2_2$is_null_rec))
})

test_that("mis-specified extract function", {
  wf <-
    workflows::workflow(
      preprocessor = mpg ~ .,
      spec = parsnip::linear_reg()
    )

  set.seed(1)
  boots <- rsample::bootstraps(mtcars, 3)

  raise_warning <- function(x) {
    warning("AHHH")
    TRUE
  }
  raise_error <- function(x) {
    stop("AHHH")
    TRUE
  }
  raise_both <- function(x) {
    warning("AH")
    stop("AHHH")
    TRUE
  }
  raise_error_once <- local({
    first <- TRUE

    function(x) {
      if (first) {
        first <<- FALSE
        stop("oh no")
      }

      "hi"
    }
  })

  expect_snapshot(
    res_extract_warning <-
      fit_resamples(
        wf,
        boots,
        control = control_resamples(extract = raise_warning)
      ),
    transform = catalog_lines
  )

  expect_snapshot(
    res_extract_error <-
      fit_resamples(
        wf,
        boots,
        control = control_resamples(extract = raise_error)
      ),
    transform = catalog_lines
  )

  expect_snapshot(
    res_extract_both <-
      fit_resamples(
        wf,
        boots,
        control = control_resamples(extract = raise_both)
      ),
    transform = catalog_lines
  )

  expect_snapshot(
    res_extract_error_once <-
      fit_resamples(
        wf,
        boots,
        control = control_resamples(extract = raise_error_once)
      ),
    transform = catalog_lines
  )

  expect_snapshot(res_extract_warning)
  expect_snapshot(res_extract_error)
  expect_snapshot(res_extract_both)
  expect_snapshot(res_extract_error_once)

  expect_true(res_extract_warning$.extracts[[1]]$.extracts[[1]])
  expect_snapshot(res_extract_warning$.notes[[1]])

  expect_snapshot(res_extract_error$.extracts[[1]]$.extracts[[1]])
  expect_snapshot(res_extract_error$.notes[[1]])

  expect_snapshot(res_extract_both$.extracts[[1]]$.extracts[[1]])

  expect_snapshot(res_extract_error_once$.extracts[[1]]$.extracts[[1]])
  expect_equal(res_extract_error_once$.extracts[[2]]$.extracts[[1]], "hi")
  expect_equal(res_extract_error_once$.extracts[[3]]$.extracts[[1]], "hi")
})

# ------------------------------------------------------------------------------

test_that("tune model and recipe", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()
  set.seed(363)
  mt_folds <- rsample::vfold_cv(mtcars, v = 5)

  extr_3_1 <- function(x) {
    x
  }

  wflow_3 <-
    workflow() |>
    add_recipe(helper_objects$rec_tune_1) |>
    add_model(helper_objects$svm_mod)
  set.seed(35)
  grid_3 <-
    extract_parameter_set_dials(wflow_3) |>
    update(num_comp = dials::num_comp(c(2, 5))) |>
    dials::grid_space_filling(size = 4)

  expect_no_error(
    res_3_1 <- tune_grid(
      wflow_3,
      resamples = mt_folds,
      grid = grid_3,
      control = control_grid(extract = extr_3_1)
    )
  )
  expect_no_error(extract_3_1 <- dplyr::bind_rows(res_3_1$.extracts))

  expect_named(extract_3_1, c("cost", "num_comp", ".extracts", ".config"))
  expect_true(
    all(purrr::map_lgl(extract_3_1$.extracts, inherits, "workflow")),
  )
})

# ------------------------------------------------------------------------------

test_that("check .config in extracts", {
  load(test_path("data", "test_objects.RData"))

  # recipe only
  for (i in 1:nrow(mt_spln_lm_grid)) {
    expect_true(any(names(mt_spln_lm_grid$.extracts[[i]]) == ".config"))
  }

  for (i in 1:nrow(mt_spln_lm_bo)) {
    expect_true(any(names(mt_spln_lm_bo$.extracts[[i]]) == ".config"))
  }

  recipe_only_configs <-
    full_join(
      mt_spln_lm_bo |>
        filter(id == first(id)) |>
        select(.iter, .metrics) |>
        unnest(cols = .metrics) |>
        filter(.metric == first(.metric)),
      mt_spln_lm_bo |>
        filter(id == first(id)) |>
        select(.iter, .extracts) |>
        unnest(cols = .extracts),
      by = c(".iter", "deg_free")
    )

  expect_equal(
    recipe_only_configs$.config.x,
    recipe_only_configs$.config.y
  )

  # recipe and model
  for (i in 1:nrow(mt_spln_knn_grid)) {
    expect_true(any(names(mt_spln_knn_grid$.extracts[[i]]) == ".config"))
  }

  for (i in 1:nrow(mt_spln_knn_bo)) {
    expect_true(any(names(mt_spln_knn_bo$.extracts[[i]]) == ".config"))
  }

  # model only
  for (i in 1:nrow(mt_knn_grid)) {
    expect_true(any(names(mt_knn_grid$.extracts[[i]]) == ".config"))
  }

  for (i in 1:nrow(mt_knn_bo)) {
    expect_true(any(names(mt_knn_bo$.extracts[[i]]) == ".config"))
  }
})

test_that("fit_best", {
  skip_if_not_installed("kknn")
  skip_if_not_installed("modeldata")

  library(recipes)
  library(rsample)
  library(dplyr)
  library(parsnip)

  data(meats, package = "modeldata")
  meats <- meats |> select(-water, -fat)

  set.seed(1)
  meat_split <- initial_split(meats)
  meat_train <- training(meat_split)
  meat_test <- testing(meat_split)

  set.seed(2)
  meat_rs <- vfold_cv(meat_train, v = 3)

  pca_rec <-
    recipe(protein ~ ., data = meat_train) |>
    step_pca(all_predictors(), num_comp = tune())

  knn_mod <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")

  ctrl <- control_grid(save_workflow = TRUE)

  set.seed(3)
  knn_pca_res <-
    tune_grid(knn_mod, pca_rec, resamples = meat_rs, grid = 3, control = ctrl)

  expect_silent(knn_fit <- fit_best(knn_pca_res))
  expect_true(knn_fit$trained)
  expect_silent(fit_best(knn_pca_res, metric = "rsq"))
  expect_snapshot(fit_best(knn_pca_res, verbose = TRUE))
  expect_snapshot(
    tmp <-
      fit_best(
        knn_pca_res,
        verbose = TRUE,
        parameters = tibble(neighbors = 1, num_comp = 1)
      )
  )

  expect_snapshot_error(
    fit_best(1L)
  )
  expect_snapshot_error(
    fit_best(tibble())
  )
  expect_snapshot_error(
    fit_best(knn_pca_res, metric = "WAT")
  )
  expect_snapshot_error(
    fit_best(knn_pca_res, parameters = tibble())
  )
  expect_snapshot_error(
    fit_best(knn_pca_res, parameters = tibble(neighbors = 2))
  )
  expect_snapshot_error(
    fit_best(knn_pca_res, chickens = 2)
  )
  data(example_ames_knn)
  expect_snapshot_error(
    fit_best(ames_iter_search)
  )
})

test_that("fit_best() works with validation split: 3-way split", {
  skip_if_not_installed("kknn")
  skip_if_not_installed("modeldata")
  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(23598723)
  initial_val_split <- rsample::initial_validation_split(ames)
  val_set <- validation_set(initial_val_split)

  f <- Sale_Price ~ Gr_Liv_Area + Year_Built
  knn_mod <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")
  wflow <- workflow(f, knn_mod)

  tune_res <- tune_grid(
    wflow,
    grid = tibble(neighbors = c(1, 5)),
    resamples = val_set,
    control = control_grid(save_workflow = TRUE)
  ) |>
    suppressWarnings()
  set.seed(3)
  fit_on_train <- fit_best(tune_res)
  pred <- predict(fit_on_train, testing(initial_val_split))

  set.seed(3)
  exp_fit_on_train <- nearest_neighbor(neighbors = 5) |>
    set_mode("regression") |>
    fit(f, training(initial_val_split))
  exp_pred <- predict(exp_fit_on_train, testing(initial_val_split))

  expect_equal(pred, exp_pred)
})

test_that("fit_best() works with validation split: 2x 2-way splits", {
  skip_if_not_installed("kknn")
  skip_if_not_installed("modeldata")
  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(23598723)
  split <- initial_validation_split(ames)
  train <- training(split)
  val_set <- validation_set(split)

  f <- Sale_Price ~ Gr_Liv_Area + Year_Built
  knn_mod <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")
  wflow <- workflow(f, knn_mod)

  tune_res <- tune_grid(
    wflow,
    grid = tibble(neighbors = c(1, 5)),
    resamples = val_set,
    control = control_grid(save_workflow = TRUE)
  )
  set.seed(3)
  fit_on_train_and_val <- fit_best(tune_res)
  pred <- predict(fit_on_train_and_val, testing(split))

  set.seed(3)
  exp_fit_on_train_and_val <- nearest_neighbor(neighbors = 5) |>
    set_mode("regression") |>
    fit(f, train)
  exp_pred <- predict(exp_fit_on_train_and_val, testing(split))

  expect_equal(pred, exp_pred)
})

test_that("fit_best() warns when metric or eval_time are specified in addition to parameters", {
  skip_if_not_installed("kknn")

  knn_mod <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")
  res <-
    tune_grid(
      workflow(mpg ~ ., knn_mod),
      bootstraps(mtcars),
      control = control_grid(save_workflow = TRUE)
    )

  tune_params <- select_best(res, metric = "rmse")
  expect_snapshot(
    manual_wf <- fit_best(res, metric = "rmse", parameters = tune_params)
  )
  expect_snapshot(
    manual_wf <- fit_best(
      res,
      metric = "rmse",
      eval_time = 10,
      parameters = tune_params
    )
  )
})

test_that("percentile intervals - resamples only", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("rsample", minimum_version = "1.2.1.9000")
  library(rsample)
  library(parsnip)

  data(Sacramento, package = "modeldata")
  set.seed(13)
  sac_rs <- vfold_cv(Sacramento)
  lm_res <-
    linear_reg() |>
    fit_resamples(
      log10(price) ~ beds + baths + sqft + type + latitude + longitude,
      resamples = sac_rs,
      control = control_resamples(save_pred = TRUE)
    )
  template <- dplyr::tibble(
    .metric = character(0),
    .estimator = character(0),
    .lower = numeric(0),
    .estimate = numeric(0),
    .upper = numeric(0),
    .config = character(0)
  )
  set.seed(1)
  expect_snapshot(int_res_1 <- int_pctl(lm_res, times = 500))
  expect_equal(int_res_1[0, ], template)
  expect_equal(nrow(int_res_1), 2)

  expect_snapshot(
    int_pctl(lm_res, times = 2000, metrics = "rmse"),
    error = TRUE
  )

  # check to make sure that alpha works
  set.seed(1)
  expect_snapshot(int_res_2 <- int_pctl(lm_res, times = 500, alpha = .25))
  expect_true(int_res_2$.lower[1] > int_res_1$.lower[1])
  expect_true(int_res_2$.upper[1] < int_res_1$.upper[1])
})


test_that("percentile intervals - last fit", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("rsample", minimum_version = "1.1.1.9000")
  library(rsample)
  library(parsnip)
  library(yardstick)

  data(Sacramento, package = "modeldata")
  set.seed(1)
  sac_split <- initial_split(Sacramento)

  lm_res <-
    linear_reg() |>
    last_fit(
      log10(price) ~ beds + baths + sqft + type + latitude + longitude,
      metrics = metric_set(mae),
      split = sac_split
    )
  template <- dplyr::tibble(
    .metric = character(0),
    .estimator = character(0),
    .lower = numeric(0),
    .estimate = numeric(0),
    .upper = numeric(0),
    .config = character(0)
  )
  set.seed(1)
  expect_snapshot(int_res_1 <- int_pctl(lm_res, times = 200))
  expect_equal(int_res_1[0, ], template)
  expect_equal(nrow(int_res_1), 1)
  set.seed(1)
  expect_snapshot(int_res_2 <- int_pctl(lm_res, times = 200))
  expect_equal(int_res_1, int_res_2)
})


test_that("percentile intervals - grid + bayes tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("C50")
  skip_if_not_installed("rsample", minimum_version = "1.1.1.9000")
  library(rsample)
  library(parsnip)
  library(yardstick)

  data("two_class_dat", package = "modeldata")
  set.seed(1)
  cls_rs <- vfold_cv(two_class_dat)

  c5_res <-
    decision_tree(min_n = tune()) |>
    set_engine("C5.0") |>
    set_mode("classification") |>
    tune_grid(
      Class ~ .,
      resamples = cls_rs,
      grid = dplyr::tibble(min_n = c(5, 20, 40)),
      metrics = metric_set(sens),
      control = control_grid(save_pred = TRUE)
    )
  template <- dplyr::tibble(
    min_n = numeric(0),
    .metric = character(0),
    .estimator = character(0),
    .lower = numeric(0),
    .estimate = numeric(0),
    .upper = numeric(0),
    .config = character(0)
  )

  expect_snapshot(int_res_1 <- int_pctl(c5_res))
  expect_equal(int_res_1[0, ], template)
  expect_equal(nrow(int_res_1), 3)

  # ------------------------------------------------------------------------------

  expect_snapshot({
    set.seed(92)
    c5_bo_res <-
      decision_tree(min_n = tune()) |>
      set_engine("C5.0") |>
      set_mode("classification") |>
      tune_bayes(
        Class ~ .,
        resamples = cls_rs,
        initial = c5_res,
        iter = 1,
        metrics = metric_set(sens),
        control = control_bayes(save_pred = TRUE)
      )
  })
  template <- dplyr::tibble(
    min_n = integer(0),
    .metric = character(0),
    .estimator = character(0),
    .lower = numeric(0),
    .estimate = numeric(0),
    .upper = numeric(0),
    .config = character(0),
    .iter = integer(0)
  )
  set.seed(1)
  int_res_2 <- int_pctl(c5_bo_res)
  expect_equal(int_res_2[0, ], template)
  expect_equal(nrow(int_res_2), 4)
  set.seed(1)
  int_res_3 <- int_pctl(c5_bo_res, event_level = "second")
  expect_true(all(int_res_3$.estimate < int_res_2$.estimate))

  ###

  set.seed(1)
  int_res_2_reps <- int_pctl(c5_bo_res, keep_replicates = TRUE)
  expect_named(
    int_res_2_reps,
    # fmt: skip
    c("min_n", ".metric", ".estimator", ".lower", ".estimate", ".upper",
      ".config", ".iter", ".values")
  )

  qnts <- quantile(
    int_res_2_reps$.values[[1]]$estimate,
    probs = c(0.025, 0.5, 0.975)
  )
  expect_equal(
    c(
      int_res_2_reps$.lower[1],
      int_res_2_reps$.estimate[1],
      int_res_2_reps$.upper[1]
    ),
    as.vector(qnts),
    tolerance = 0.001
  )

  # ------------------------------------------------------------------------------

  c5_mixed_res <-
    decision_tree(min_n = tune()) |>
    set_engine("C5.0") |>
    set_mode("classification") |>
    tune_grid(
      Class ~ .,
      resamples = cls_rs,
      grid = dplyr::tibble(min_n = c(20, 40)),
      metrics = metric_set(roc_auc, sens),
      control = control_grid(save_pred = TRUE)
    )
  template <- dplyr::tibble(
    min_n = numeric(0),
    .metric = character(0),
    .estimator = character(0),
    .lower = numeric(0),
    .estimate = numeric(0),
    .upper = numeric(0),
    .config = character(0)
  )
  set.seed(2093)
  int_res_4 <- int_pctl(c5_mixed_res)
  expect_equal(int_res_4[0, ], template)
  expect_equal(nrow(int_res_4), 4)
})


test_that("percentile intervals - grid tuning with validation set", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("C50")
  skip_if_not_installed("rsample", minimum_version = "1.1.1.9000")
  library(rsample)
  library(parsnip)
  library(yardstick)

  data("two_class_dat", package = "modeldata")
  set.seed(1)
  cls_split <- initial_validation_split(two_class_dat, prop = c(.8, .15))
  cls_rs <- validation_set(cls_split)

  c5_res <-
    decision_tree(min_n = tune()) |>
    set_engine("C5.0") |>
    set_mode("classification") |>
    tune_grid(
      Class ~ .,
      resamples = cls_rs,
      grid = dplyr::tibble(min_n = c(5, 20, 40)),
      metrics = metric_set(sens),
      control = control_grid(save_pred = TRUE)
    )
  template <- dplyr::tibble(
    min_n = numeric(0),
    .metric = character(0),
    .estimator = character(0),
    .lower = numeric(0),
    .estimate = numeric(0),
    .upper = numeric(0),
    .config = character(0)
  )

  expect_snapshot(int_res_1 <- int_pctl(c5_res))
  expect_equal(int_res_1[0, ], template)
  expect_equal(nrow(int_res_1), 3)
})

rcv_results <- readRDS(test_path("data", "rcv_results.rds"))

opt <- getOption("dplyr.summarise.inform", default = "FALSE")
options(dplyr.summarise.inform = FALSE)

compl <-
  unnest(rcv_results, .metrics) |>
  group_by(
    deg_free,
    degree,
    `wt df`,
    `wt degree`,
    .config,
    .metric,
    .estimator
  ) |>
  summarize(
    mean = mean(.estimate, na.rm = TRUE),
    n = sum(!is.na(.estimator)),
    std_err = sd(.estimate, na.rm = TRUE) / sqrt(n)
  ) |>
  ungroup() |>
  arrange(.config)

options(dplyr.summarise.inform = opt)

test_that("estimate method", {
  expect_equal(
    collect_metrics(rcv_results)[, names(compl)] |> arrange(.config),
    compl
  )
})

test_that("estimate method (with apparent resample)", {
  skip_on_cran()
  skip_if_not_installed("kknn")

  library(parsnip)
  library(rsample)
  library(yardstick)

  m_set <- metric_set(rmse)

  res <-
    fit_resamples(
      nearest_neighbor("regression"),
      mpg ~ cyl + hp,
      bootstraps(mtcars, 5, apparent = TRUE),
      metrics = m_set,
      control = control_grid(save_pred = TRUE)
    )

  collected_sum <-
    collect_metrics(res) |>
    select(mean, n, std_err)

  collected_manual <-
    res |>
    dplyr::filter(id != "Apparent") |>
    tidyr::unnest(.metrics) |>
    summarize(
      mean = mean(.estimate),
      n = sum(!is.na(.estimator)),
      std_err = sd(.estimate) / sqrt(n)
    )

  expect_equal(collected_sum, collected_manual)
})

test_that("control functions respect hierarchy with condense_control", {
  ctrl <- parsnip::condense_control(
    control_grid(),
    control_resamples()
  )

  expect_equal(
    ctrl,
    control_resamples()
  )

  ctrl <- parsnip::condense_control(
    control_last_fit(),
    control_grid()
  )

  expect_equal(
    ctrl,
    control_grid(
      extract = control_last_fit()$extract,
      save_pred = TRUE,
      allow_par = FALSE
    )
  )

  ctrl <- parsnip::condense_control(
    control_last_fit(),
    control_resamples()
  )

  expect_equal(
    ctrl,
    control_resamples(
      extract = control_last_fit()$extract,
      save_pred = TRUE,
      allow_par = FALSE
    )
  )

  ctrl <- parsnip::condense_control(
    control_bayes(),
    control_grid()
  )

  expect_equal(
    ctrl,
    control_grid()
  )

  ctrl <- parsnip::condense_control(
    control_bayes(),
    control_resamples()
  )

  expect_equal(
    ctrl,
    control_resamples()
  )
})

test_that("workflow size warning", {
  withr::local_options(width = 500)

  # A larger data set to trip the warning
  MTCARS <- mtcars[rep(1:32, each = 1000), ]

  lm_rec <- recipe(mpg ~ ., data = MTCARS)
  lm_wflow <- workflow(lm_rec, linear_reg() |> set_engine("lm", x = TRUE))
  # About 2.7MB when fit

  expect_silent({
    set.seed(1)
    no_warning <-
      lm_wflow |>
      fit_resamples(
        resamples = vfold_cv(MTCARS),
        control = control_resamples(save_workflow = TRUE, workflow_size = Inf)
      )
  })

  expect_snapshot({
    set.seed(1)
    warns <-
      lm_wflow |>
      fit_resamples(
        resamples = vfold_cv(MTCARS),
        control = control_resamples(save_workflow = TRUE, workflow_size = 2)
      )
  })

  expect_silent({
    set.seed(1)
    no_save <-
      lm_wflow |>
      fit_resamples(
        resamples = vfold_cv(MTCARS),
        control = control_resamples(save_workflow = FALSE, workflow_size = 2)
      )
  })
})

test_that("control object print methods", {
  expect_snapshot(control_grid())
  expect_snapshot(control_grid(verbose = TRUE, save_pred = TRUE))
  expect_snapshot(control_grid(pkgs = c("pkg1", "pkg2"), extract = I))

  set.seed(123)
  expect_snapshot(control_bayes())
  expect_snapshot(control_bayes(
    verbose_iter = TRUE,
    no_improve = 5,
    save_gp_scoring = TRUE
  ))

  expect_snapshot(control_last_fit())
  expect_snapshot(control_last_fit(verbose = TRUE))
})

test_that("control object print methods with default = TRUE", {
  expect_snapshot(print(
    control_grid(verbose = TRUE, pkgs = c("pkg1", "pkg2")),
    default = TRUE
  ))
  set.seed(456)
  expect_snapshot(print(
    control_bayes(verbose_iter = TRUE, no_improve = 5),
    default = TRUE
  ))
  expect_snapshot(print(control_last_fit(verbose = TRUE), default = TRUE))
})

test_that("augment fit_resamples", {
  skip_if_not_installed("modeldata")
  data(two_class_dat, package = "modeldata")

  # ------------------------------------------------------------------------------

  lr_spec <- parsnip::logistic_reg() |> parsnip::set_engine("glm")

  set.seed(1)
  two_class_dat <- as.data.frame(two_class_dat)
  bt1 <- rsample::bootstraps(two_class_dat, times = 30)

  set.seed(1)
  fit_1 <- fit_resamples(
    lr_spec,
    Class ~ .,
    bt1,
    control = control_resamples(save_pred = TRUE)
  )
  aug_1 <- augment(fit_1)
  expect_true(nrow(aug_1) == nrow(two_class_dat))
  expect_equal(aug_1[["A"]], two_class_dat[["A"]])
  expect_true(sum(!is.na(aug_1$.pred_class)) == nrow(two_class_dat))
  expect_snapshot(names(aug_1))

  expect_s3_class_bare_tibble(aug_1)

  expect_snapshot(error = TRUE, augment(fit_1, hey = "you"))
})


test_that("augment fit_resamples", {
  skip_if_not_installed("modeldata")
  data(two_class_dat, package = "modeldata")

  # ------------------------------------------------------------------------------

  skip_if(new_rng_snapshots)
  lr_spec <- parsnip::logistic_reg() |> parsnip::set_engine("glm")

  set.seed(1)
  two_class_dat <- as.data.frame(two_class_dat)
  bt2 <- rsample::bootstraps(two_class_dat, times = 3)

  set.seed(1)
  fit_2 <- fit_resamples(
    lr_spec,
    Class ~ .,
    bt2,
    control = control_resamples(save_pred = TRUE)
  )
  expect_snapshot(aug_2 <- augment(fit_2))

  expect_true(nrow(aug_2) == nrow(two_class_dat))
  expect_equal(aug_2[["A"]], two_class_dat[["A"]])
  expect_true(sum(!is.na(aug_2$.pred_class)) < nrow(two_class_dat))
  expect_snapshot(names(aug_2))
  expect_s3_class_bare_tibble(aug_2)
})

# ------------------------------------------------------------------------------

test_that("augment tune_grid", {
  skip_if_not_installed("kernlab")

  svm_spec <- parsnip::svm_linear(cost = tune(), margin = 0.1) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("regression")
  set.seed(1)
  cv1 <- rsample::vfold_cv(mtcars)

  set.seed(1)
  fit_1 <- tune_grid(
    svm_spec,
    mpg ~ .,
    cv1,
    grid = data.frame(cost = 1:3),
    control = control_grid(save_pred = TRUE)
  )
  aug_1 <- augment(fit_1)
  expect_true(nrow(aug_1) == nrow(mtcars))
  expect_equal(aug_1[["wt"]], mtcars[["wt"]])
  expect_true(sum(!is.na(aug_1$.pred)) == nrow(mtcars))

  expect_snapshot(names(aug_1))
  expect_s3_class_bare_tibble(aug_1)

  aug_2 <- augment(fit_1, parameters = data.frame(cost = 3))
  expect_true(any(abs(aug_1$.pred - aug_2$.pred) > 1))

  expect_snapshot(error = TRUE, {
    augment(fit_1, parameters = list(cost = 3))
  })

  expect_snapshot(error = TRUE, {
    augment(fit_1, parameters = data.frame(cost = 3:4))
  })

  expect_snapshot(error = TRUE, {
    augment(fit_1, cost = 4)
  })

  # ------------------------------------------------------------------------------

  suppressMessages({
    set.seed(1)
    fit_2 <- tune_bayes(
      svm_spec,
      mpg ~ .,
      cv1,
      initial = fit_1,
      iter = 2,
      param_info = parameters(dials::cost(c(-10, 5))),
      control = control_bayes(save_pred = TRUE)
    )
  })
  aug_3 <- augment(fit_2)
  expect_true(nrow(aug_3) == nrow(mtcars))
  expect_equal(aug_3[["wt"]], mtcars[["wt"]])
  expect_true(sum(!is.na(aug_3$.pred)) == nrow(mtcars))
  expect_snapshot(names(aug_3))
  expect_s3_class_bare_tibble(aug_3)
})


# ------------------------------------------------------------------------------

test_that("augment last_fit", {
  skip_if_not_installed("modeldata")
  data(two_class_dat, package = "modeldata")

  # ------------------------------------------------------------------------------

  lr_spec <- parsnip::logistic_reg() |> parsnip::set_engine("glm")
  set.seed(1)
  split <- rsample::initial_split(two_class_dat)
  fit_1 <- last_fit(lr_spec, Class ~ ., split = split)

  aug_1 <- augment(fit_1)
  expect_true(nrow(aug_1) == nrow(rsample::assessment(split)))
  expect_equal(aug_1[["A"]], rsample::assessment(split)[["A"]])
  expect_true(
    sum(!is.na(aug_1$.pred_class)) == nrow(rsample::assessment(split))
  )
  expect_snapshot(names(aug_1))
  expect_s3_class_bare_tibble(aug_1)

  test_data <-
    split |>
    assessment() |>
    mutate(.row = as.integer(split, data = "assessment"))

  exp_augment <-
    fit_1 |>
    collect_predictions(summarize = TRUE) |>
    arrange(.row, .config) |>
    inner_join(test_data |> select(-Class), by = ".row") |>
    dplyr::select(!!!names(aug_1))

  expect_equal(aug_1, exp_augment)

  expect_snapshot(error = TRUE, augment(fit_1, potato = TRUE))
})

test_that("gives same output as collect_metrics() when metrics match", {
  skip_on_cran()
  skip_if_not_installed("kknn")

  library(parsnip)
  library(rsample)
  library(yardstick)

  m_set <- metric_set(rmse)

  res <-
    tune_grid(
      nearest_neighbor("regression", neighbors = tune()),
      mpg ~ cyl + hp,
      bootstraps(mtcars, 5),
      grid = 3,
      metrics = m_set,
      control = control_grid(save_pred = TRUE)
    )

  collected_sum <- collect_metrics(res)
  computed_sum <- compute_metrics(res, m_set)

  expect_equal(collected_sum, computed_sum)

  collected_unsum <- collect_metrics(res, summarize = FALSE)
  computed_unsum <- compute_metrics(res, m_set, summarize = FALSE)

  expect_equal(collected_unsum, computed_unsum)
})

test_that("gives same output as collect_metrics() when metrics match (apparent)", {
  skip_on_cran()
  skip_if_not_installed("kknn")

  library(parsnip)
  library(rsample)
  library(yardstick)

  m_set <- metric_set(rmse)

  res <-
    fit_resamples(
      nearest_neighbor("regression"),
      mpg ~ cyl + hp,
      bootstraps(mtcars, 5, apparent = TRUE),
      metrics = m_set,
      control = control_grid(save_pred = TRUE, allow_par = FALSE)
    )

  collected_sum <- collect_metrics(res)
  computed_sum <- compute_metrics(res, m_set)

  expect_equal(collected_sum, computed_sum)

  collected_unsum <- collect_metrics(res, summarize = FALSE)
  computed_unsum <- compute_metrics(res, m_set, summarize = FALSE)

  expect_equal(collected_unsum, computed_unsum)
})

test_that("`metrics` argument works (numeric metrics)", {
  skip_on_cran()
  skip_if_not_installed("kknn")

  library(parsnip)
  library(rsample)
  library(yardstick)

  m_set_rmse <- metric_set(rmse)
  m_set_rsq <- metric_set(rsq)

  set.seed(1)

  res_rmse <-
    tune_grid(
      nearest_neighbor("regression", neighbors = tune()),
      mpg ~ cyl + hp,
      bootstraps(mtcars, 5),
      grid = 3,
      metrics = m_set_rmse,
      control = control_grid(save_pred = TRUE)
    )

  set.seed(1)

  res_rsq <-
    tune_grid(
      nearest_neighbor("regression", neighbors = tune()),
      mpg ~ cyl + hp,
      bootstraps(mtcars, 5),
      grid = 3,
      metrics = m_set_rsq,
      control = control_grid(save_pred = TRUE)
    )

  collected_sum_rsq <- collect_metrics(res_rsq)
  computed_sum_rsq <- compute_metrics(res_rmse, m_set_rsq)

  expect_equal(collected_sum_rsq, computed_sum_rsq)

  collected_sum_rmse <- collect_metrics(res_rmse)
  computed_sum_rmse <- compute_metrics(res_rsq, m_set_rmse)

  expect_equal(collected_sum_rsq, computed_sum_rsq)

  collected_unsum_rsq <- collect_metrics(res_rsq, summarize = FALSE)
  computed_unsum_rsq <- compute_metrics(res_rmse, m_set_rsq, summarize = FALSE)

  expect_equal(collected_unsum_rsq, computed_unsum_rsq)

  collected_unsum_rmse <- collect_metrics(res_rmse, summarize = FALSE)
  computed_unsum_rmse <- compute_metrics(res_rsq, m_set_rmse, summarize = FALSE)

  expect_equal(collected_unsum_rmse, computed_unsum_rmse)
})

test_that("`metrics` argument works (compatible class metric types)", {
  skip_on_cran()
  skip_if_not_installed("kknn")
  skip_if_not_installed("modeldata")

  library(parsnip)
  library(rsample)
  library(yardstick)
  library(modeldata)

  set.seed(1)

  res_acc_auc <-
    tune_grid(
      nearest_neighbor("classification", neighbors = tune()),
      Class ~ .,
      vfold_cv(two_class_dat, 5),
      metrics = metric_set(accuracy, roc_auc),
      grid = 3,
      control = control_grid(save_pred = TRUE)
    )

  set.seed(1)

  res_class <-
    tune_grid(
      nearest_neighbor("classification", neighbors = tune()),
      Class ~ .,
      vfold_cv(two_class_dat, 5),
      metrics = metric_set(precision),
      grid = 3,
      control = control_grid(save_pred = TRUE)
    )

  set.seed(1)

  res_prob <-
    tune_grid(
      nearest_neighbor("classification", neighbors = tune()),
      Class ~ .,
      vfold_cv(two_class_dat, 5),
      metrics = metric_set(mn_log_loss),
      grid = 3,
      control = control_grid(save_pred = TRUE)
    )

  collected_sum_class <- collect_metrics(res_class)
  computed_sum_class <- compute_metrics(res_acc_auc, metric_set(precision))

  expect_equal(collected_sum_class, computed_sum_class)

  collected_unsum_class <- collect_metrics(res_class, summarize = FALSE)
  computed_unsum_class <- compute_metrics(
    res_acc_auc,
    metric_set(precision),
    summarize = FALSE
  )

  expect_equal(collected_unsum_class, computed_unsum_class)

  collected_sum_prob <- collect_metrics(res_prob)
  computed_sum_prob <- compute_metrics(res_acc_auc, metric_set(mn_log_loss))

  expect_equal(collected_sum_prob, computed_sum_prob)

  collected_unsum_prob <- collect_metrics(res_prob, summarize = FALSE)
  computed_unsum_prob <- compute_metrics(
    res_acc_auc,
    metric_set(mn_log_loss),
    summarize = FALSE
  )

  expect_equal(collected_unsum_prob, computed_unsum_prob)
})

test_that("`metrics` argument works (differing class metric types)", {
  skip_on_cran()
  skip_if_not_installed("kknn")
  skip_if_not_installed("modeldata")

  library(parsnip)
  library(rsample)
  library(yardstick)
  library(modeldata)

  # class metric
  m_set_acc <- metric_set(accuracy)
  # prob metric
  m_set_auc <- metric_set(roc_auc)

  set.seed(1)

  res_acc <-
    tune_grid(
      nearest_neighbor("classification", neighbors = tune()),
      Class ~ .,
      vfold_cv(two_class_dat, 5),
      metrics = m_set_acc,
      grid = 3,
      control = control_grid(save_pred = TRUE)
    )

  set.seed(1)

  res_auc <-
    tune_grid(
      nearest_neighbor("classification", neighbors = tune()),
      Class ~ .,
      vfold_cv(two_class_dat, 5),
      metrics = m_set_auc,
      grid = 3,
      control = control_grid(save_pred = TRUE)
    )

  expect_snapshot(
    compute_metrics(res_auc, m_set_acc),
    error = TRUE
  )

  expect_snapshot(
    compute_metrics(res_acc, m_set_auc),
    error = TRUE
  )
})

test_that("`metrics` argument works (iterative tuning)", {
  # the pattern of tuning first with one metric, then with another,
  # and then testing equality of collected vs computed metrics is a bit
  # more complicated here, as a different metric will lead to different
  # proposed hyperparameters in the gaussian process.

  skip_on_cran()
  skip_if_not_installed("kknn")

  library(parsnip)
  library(rsample)
  library(yardstick)

  m_set_rmse <- metric_set(rmse)
  m_set_both <- metric_set(rmse, rsq)

  expect_snapshot({
    set.seed(1)
    res_rmse <-
      tune_bayes(
        nearest_neighbor("regression", neighbors = tune()),
        mpg ~ .,
        vfold_cv(mtcars, v = 3),
        metrics = m_set_rmse,
        control = tune::control_bayes(save_pred = TRUE),
        iter = 2,
        initial = 3
      )
  })

  expect_snapshot({
    set.seed(1)
    res_both <-
      tune_bayes(
        nearest_neighbor("regression", neighbors = tune()),
        mpg ~ .,
        vfold_cv(mtcars, v = 3),
        metrics = m_set_both,
        control = tune::control_bayes(save_pred = TRUE),
        iter = 2,
        initial = 3
      )
  })

  collected_sum_rmse <- collect_metrics(res_rmse)
  computed_sum_rmse <- compute_metrics(res_both, m_set_rmse)

  expect_equal(collected_sum_rmse, computed_sum_rmse)

  collected_unsum_rmse <- collect_metrics(res_rmse, summarize = FALSE)
  computed_unsum_rmse <- compute_metrics(
    res_both,
    m_set_rmse,
    summarize = FALSE
  )

  expect_equal(collected_unsum_rmse, computed_unsum_rmse)
})

test_that("errors informatively with bad input", {
  skip_on_cran()
  skip_if_not_installed("kknn")

  library(parsnip)
  library(rsample)
  library(yardstick)

  set.seed(1)

  res_rmse <-
    tune_grid(
      nearest_neighbor("regression", neighbors = tune()),
      mpg ~ cyl + hp,
      bootstraps(mtcars, 5),
      grid = 3,
      metrics = metric_set(rmse)
    )

  res_rmse_save_pred <-
    tune_grid(
      nearest_neighbor("regression", neighbors = tune()),
      mpg ~ cyl + hp,
      bootstraps(mtcars, 5),
      grid = 3,
      metrics = metric_set(rmse),
      control = control_grid(save_pred = TRUE)
    )

  expect_snapshot(
    compute_metrics(res_rmse, metric_set(rsq)),
    error = TRUE
  )

  expect_snapshot(
    compute_metrics("boop", metric_set(rsq)),
    error = TRUE
  )

  expect_snapshot(
    compute_metrics(res_rmse_save_pred, "wheee"),
    error = TRUE
  )
})

# ------------------------------------------------------------------------------

#' @export
`[.iteration_results` <- function(x, i, j, ...) {
  out <- NextMethod()
  iteration_results_reconstruct(out, x)
}

#' @export
`names<-.iteration_results` <- function(x, value) {
  out <- NextMethod()
  iteration_results_reconstruct(out, x)
}

# ------------------------------------------------------------------------------

# Same as `tune_results`, except that the vctrs invariants also require that
# the `.iter` column sticks around.

#' @export
#' @keywords internal
#' @rdname empty_ellipses
#' @param parameters A `parameters` object.
#' @param metrics A metric set.
#' @param eval_time A numeric vector of time points where dynamic event time
#' metrics should be computed (e.g. the time-dependent ROC curve, etc).
#' @param outcomes A character vector of outcome names.
#' @param rset_info Attributes from an `rset` object.
#' @param workflow The workflow used to fit the iteration results.
new_iteration_results <-
  function(
    x,
    parameters,
    metrics,
    eval_time,
    eval_time_target,
    outcomes = character(0),
    rset_info,
    workflow
  ) {
    new_tune_results(
      x = x,
      parameters = parameters,
      metrics = metrics,
      eval_time = eval_time,
      eval_time_target = eval_time_target,
      outcomes = outcomes,
      rset_info = rset_info,
      workflow = workflow,
      class = "iteration_results"
    )
  }

#' Calculate and format metrics from tuning functions
#'
#' @description
#'
#' This function computes metrics from tuning results. The arguments and
#' output formats are closely related to those from [collect_metrics()], but
#' this function additionally takes a `metrics` argument with a
#' [metric set][yardstick::metric_set()] for new metrics to compute. This
#' allows for computing new performance metrics without requiring users to
#' re-evaluate models against resamples.
#'
#' Note that the [control option][control_grid()] `save_pred = TRUE` must
#' have been supplied when generating `x`.
#'
#' @param x The results of a tuning function like [tune_grid()] or
#' [fit_resamples()], generated with the control option `save_pred = TRUE`.
#' @param metrics A [metric set][yardstick::metric_set()] of new metrics
#' to compute. See the "Details" section below for more information.
#' @param summarize A single logical value indicating whether metrics should
#' be summarized over resamples (`TRUE`) or return the values for each
#' individual resample. See [collect_metrics()] for more details on how
#' metrics are summarized.
#' @inheritParams control_grid
#' @inheritParams collect_metrics
#'
#' @return A tibble. See [collect_metrics()] for more details on the return value.
#'
#' @details
#'
#' Each metric in the set supplied to the `metrics` argument must have a metric
#' type (usually `"numeric"`, `"class"`, or `"prob"`) that matches some metric
#' evaluated when generating `x`. e.g. For example, if `x` was generated with
#' only hard `"class"` metrics, this function can't compute metrics that take in
#' class probabilities (`"prob"`.) By default, the tuning functions used to
#' generate `x` compute metrics of all needed types.
#'
#' @examplesIf tune:::should_run_examples()
#' # load needed packages:
#' library(parsnip)
#' library(rsample)
#' library(yardstick)
#'
#' # evaluate a linear regression against resamples.
#' # note that we pass `save_pred = TRUE`:
#' res <-
#'   fit_resamples(
#'     linear_reg(),
#'     mpg ~ cyl + hp,
#'     bootstraps(mtcars, 5),
#'     control = control_grid(save_pred = TRUE)
#'   )
#'
#' # to return the metrics supplied to `fit_resamples()`:
#' collect_metrics(res)
#'
#' # to compute new metrics:
#' compute_metrics(res, metric_set(mae))
#'
#' # if `metrics` is the same as that passed to `fit_resamples()`,
#' # then `collect_metrics()` and `compute_metrics()` give the same
#' # output, though `compute_metrics()` is quite a bit slower:
#' all.equal(
#'   collect_metrics(res),
#'   compute_metrics(res, metric_set(rmse, rsq))
#' )
#' @name compute_metrics
#' @export
compute_metrics <- function(x, metrics, summarize, event_level, ...) {
  UseMethod("compute_metrics")
}

#' @export
#' @rdname compute_metrics
compute_metrics.default <- function(
  x,
  metrics,
  summarize = TRUE,
  event_level = "first",
  ...
) {
  cli::cli_abort(
    "No {.fn compute_metrics} method exists for
                 {.obj_type_friendly {x}}."
  )
}

#' @export
#' @rdname compute_metrics
compute_metrics.tune_results <- function(
  x,
  metrics,
  ...,
  summarize = TRUE,
  event_level = "first"
) {
  rlang::check_dots_empty()
  check_bool(summarize)
  if (!".predictions" %in% names(x)) {
    cli::cli_abort(
      "{.arg x} must have been generated with the control argument
      {.code save_pred = TRUE}."
    )
  }

  if (!inherits(metrics, "metric_set")) {
    cli::cli_abort("{.arg metrics} must be a metric set.")
  }

  new_metrics_info <- metrics_info(metrics)
  old_metrics_info <- metrics_info(attr(x, "metrics"))

  # if only a class metric is supplied to a tuning function, then
  # `save_pred = TRUE` will only return the class predictions from
  # each model. same story goes for prob metrics.
  if (!all(new_metrics_info$type %in% old_metrics_info$type)) {
    cli::cli_abort(c(
      "The supplied `metrics` argument has metrics of type \\
       {.val {new_metrics_info$type}}, while the metrics used to generate \\
       predictions only used {.val {old_metrics_info$type}} metrics.",
      "i" = "To save predictions for {new_metrics_info$type} metrics, \\
             generate {.arg x} with metrics of that type."
    ))
  }

  param_names <- .get_tune_parameter_names(x)
  outcome_name <- .get_tune_outcome_names(x)

  preds <- x$.predictions
  preds <- purrr::map2(preds, x$id, mutate_id)
  preds <- purrr::list_rbind(preds)
  preds <- dplyr::group_by(preds, id, .config)

  # call `.estimate_metrics` with additional groupings
  mtrcs <-
    .estimate_metrics(
      preds,
      metric = metrics,
      param_names = c(param_names, "id", ".config"),
      outcome_name = outcome_name,
      event_level = event_level,
      metrics_info = new_metrics_info
    )

  # re-order so `.config` comes last
  mtrcs <- mtrcs[c(setdiff(names(mtrcs), ".config"), ".config")]

  # nest by resample id
  nest_cols <- "id"

  # Convert the iterative .configs into numbers
  if (any(grepl("^[iI]ter", mtrcs$.config))) {
    mtrcs$.iter <- .config_to_.iter(mtrcs$.config)
    nest_cols <- c(nest_cols, ".iter")
  }

  mtrcs <- nest(mtrcs, .by = all_of(nest_cols), .key = ".metrics")
  match_locations <- vec_locate_matches(x[nest_cols], mtrcs[nest_cols])
  x$.metrics <- vec_slice(mtrcs$.metrics, match_locations$haystack)

  attr(x, "metrics") <- metrics

  collect_metrics(x = x, summarize = summarize, ...)
}

mutate_id <- function(tbl, id) {
  tbl$id <- id
  tbl
}

#' @export
obj_sum.tune_results <- function(x) {
  null_metrics <- purrr::map_lgl(x$.metrics, is.null)
  if (!all(null_metrics)) {
    res <- "tune[+]"
  } else {
    res <- "tune[x]"
  }
  res
}

# also used for resample_results
#' @export
size_sum.tune_results <- function(x) {
  ""
}

#' @export
obj_sum.resample_results <- function(x) {
  null_metrics <- purrr::map_lgl(x$.metrics, is.null)
  if (!all(null_metrics)) {
    res <- "rsmp[+]"
  } else {
    res <- "rsmp[x]"
  }
  res
}

#' @export
obj_sum.tune_race <- function(x) {
  null_metrics <- purrr::map_lgl(x$.metrics, is.null)
  if (!all(null_metrics)) {
    res <- "race[+]"
  } else {
    res <- "race[x]"
  }
  res
}

# Functions related to computing predicted performance

# will add further options later for surv_prob etc.
pred_type <- function(x) {
  cls <- class(x)[class(x) != "function"][1]
  res <- dplyr::case_when(
    cls == "class_metric" ~ "class",
    cls == "prob_metric" ~ "prob",
    cls == "ordered_prob_metric" ~ "prob",
    cls == "numeric_metric" ~ "numeric",
    cls == "dynamic_survival_metric" ~ "survival",
    cls == "integrated_survival_metric" ~ "survival",
    cls == "static_survival_metric" ~ "time",
    cls == "linear_pred_survival_metric" ~ "linear_pred",
    cls == "quantile_metric" ~ "quantile",
    TRUE ~ "unknown"
  )
  res
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
metrics_info <- function(x) {
  metric_list <- rlang::env_get(environment(x), "fns")
  metric_dir <- purrr::map_chr(metric_list, attr, "direction")
  res <- tibble::new_tibble(
    list(
      .metric = names(metric_dir),
      direction = unname(metric_dir),
      type = unname(purrr::map_chr(metric_list, pred_type))
    ),
    nrow = length(metric_dir)
  )
  res
}

#' Internal functions used by other tidymodels packages
#'
#' These are not to be meant to be invoked directly by users.
#' @param dat A data set.
#' @param metric A metric set.
#' @param param_names A character vector of tuning parameter names.
#' @param outcome_name A character string for the column of `dat` that is the
#' outcome.
#' @param event_level A logical passed from the control function.
#' @param x A character vector of package names.
#' @param .expr Code to execute.
#' @param ... Object to pass to the internal `update_printer()` function.
#' @param bad_only A logical for whether warnings and errors should be caught.
#' @param notes Character data to add to the logging.
#' @param workflow A workflow.
#' @param grid_preprocessor A tibble with parameter information.
#' @param new_data A data frame or matrix of predictors to process.
#' @param metrics_info The output of `tune:::metrics_info(metrics)`---only
#' included as an argument to allow for pre-computing.
#' @param catalog A logical passed to `update_printer()` giving whether the message
#' is compatible with the issue cataloger. Defaults to `TRUE`. Updates that are
#' always unique and do not represent a tuning "issue" can bypass the cataloger
#' by setting `catalog = FALSE`.
#' @keywords internal
#' @name tune-internal-functions
#' @export
.estimate_metrics <- function(
  dat,
  metric,
  param_names,
  outcome_name,
  event_level,
  metrics_info = metrics_info(metrics)
) {
  # The call stack is:
  #
  # tune_grid_loop()
  #   .loop_over_all_stages() <many times>
  #    .estimate_metrics()

  # predictions made in:
  #
  # predict_all_types()
  #   predict_wrapper()
  #     parsnip::predict() (or multi_predict())

  if (inherits(dat, "try-error")) {
    return(NULL)
  }

  # Determine the type of prediction that is required
  types <- unique(metrics_info$type)

  if (length(outcome_name) > 1L) {
    cli::cli_abort(
      "Multiple outcomes ({.val {outcome_name}}) are not supported in {.fn .estimate_metrics}.",
      .internal = TRUE
    )
  }

  if (case_weights_column_name() %in% names(dat)) {
    case_weights <- sym(case_weights_column_name())
  } else {
    case_weights <- NULL
  }

  if (all(types == "numeric")) {
    estimate_reg(dat, metric, param_names, outcome_name, case_weights)
  } else if (all(types == "class" | types == "prob")) {
    estimate_class_prob(
      dat,
      metric,
      param_names,
      outcome_name,
      case_weights,
      types,
      event_level
    )
  } else if (
    all(types == "time" | types == "survival" | types == "linear_pred")
  ) {
    estimate_surv(dat, metric, param_names, outcome_name, case_weights, types)
  } else if (all(types == "quantile")) {
    estimate_quantile_reg(
      dat,
      metric,
      param_names,
      outcome_name,
      case_weights,
      types
    )
  } else {
    cli::cli_abort("Metric type not yet supported by {.pkg tune}.")
  }
}

estimate_reg <- function(dat, metric, param_names, outcome_name, case_weights) {
  dat |>
    dplyr::group_by(!!!rlang::syms(param_names)) |>
    metric(
      estimate = .pred,
      truth = !!sym(outcome_name),
      case_weights = !!case_weights
    )
}

estimate_quantile_reg <- function(
  dat,
  metric,
  param_names,
  outcome_name,
  case_weights,
  types # not currently used but perhaps in future
) {
  dat |>
    dplyr::group_by(!!!rlang::syms(param_names)) |>
    metric(
      estimate = .pred_quantile,
      truth = !!sym(outcome_name),
      case_weights = !!case_weights
    )
}

estimate_class_prob <- function(
  dat,
  metric,
  param_names,
  outcome_name,
  case_weights,
  types,
  event_level
) {
  truth <- sym(outcome_name)

  estimate <- NULL
  if (any(types == "class")) {
    estimate <- sym(".pred_class")
  }

  probs <- NULL
  if (any(types == "prob")) {
    levels <- levels(dat[[outcome_name]])
    probs <- paste0(".pred_", levels)

    # Special case binary class prob metrics,
    # as yardstick requires only 1 column passed through
    if (length(probs) == 2) {
      if (identical(event_level, "first")) {
        probs <- probs[[1]]
      } else if (identical(event_level, "second")) {
        probs <- probs[[2]]
      } else {
        cli::cli_abort(
          "{.arg event_level} must be either {.val first} or
                        {.val second}."
        )
      }
    }
  }

  dat |>
    dplyr::group_by(!!!rlang::syms(param_names)) |>
    metric(
      truth = !!truth,
      estimate = !!estimate,
      !!!probs,
      case_weights = !!case_weights,
      event_level = event_level
    )
}

# ------------------------------------------------------------------------------

estimate_surv <- function(
  dat,
  metric,
  param_names,
  outcome_name,
  case_weights,
  types
) {
  #  potentially need to work around submodel parameters since those are within .pred
  dat <- unnest_parameters(dat, param_names)
  dat |>
    dplyr::group_by(!!!rlang::syms(param_names)) |>
    metric(
      truth = !!rlang::sym(outcome_name),
      estimate = !!maybe_estimate(metric),
      case_weights = !!case_weights,
      !!maybe_surv_prob(metric)
    )
}

unnest_parameters <- function(x, params = NULL) {
  if (is.null(params)) {
    return(x)
  }

  # When multi_predict() is used, .pred will have the tuning parameter values.
  # Other (non-submodel) parameters will be outside of 'x'.
  # If this happens, pull the submodel parameters out of .pred and put them at
  # the out level ('x') with the rest (if any)

  outer_nms <- names(x)
  has_pred <- any(outer_nms == ".pred")
  if (!has_pred) {
    return(x)
  }

  inner_nms <- names(x$.pred[[1]])
  has_inner_params <- any(params %in% inner_nms)
  if (!has_inner_params) {
    return(x)
  }

  x <- x |> parsnip::add_rowindex()

  others <- x |> dplyr::select(-.pred)
  rm_cols <- c(".pred_censored", ".weight_time")
  nest_cols <- c(".eval_time", ".pred_survival", ".weight_censored")
  x <-
    x |>
    dplyr::select(.pred, .row) |>
    tidyr::unnest(cols = c(.pred)) |>
    dplyr::select(-dplyr::any_of(rm_cols)) |>
    tidyr::nest(
      .pred = c(all_of(nest_cols)),
      .by = c(.row, dplyr::all_of(params))
    ) |>
    dplyr::full_join(others, by = ".row") |>
    dplyr::select(-.row)
  x
}

maybe_estimate <- function(x) {
  info <- tibble::as_tibble(x)

  has_static <- any(info$class == "static_survival_metric")
  has_linear_pred <- any(info$class == "linear_pred_survival_metric")
  has_static_and_linear <- has_static && has_linear_pred

  if (has_static_and_linear) {
    res <- rlang::expr(c(
      static = .pred_time,
      linear_pred = .pred_linear_pred
    ))
  } else if (has_static) {
    res <- rlang::sym(".pred_time")
  } else if (has_linear_pred) {
    res <- rlang::sym(".pred_linear_pred")
  } else {
    res <- NULL
  }
  res
}

dyn_inputs <- c("integrated_survival_metric", "dynamic_survival_metric")

maybe_surv_prob <- function(x) {
  info <- tibble::as_tibble(x)
  if (any(info$class %in% dyn_inputs)) {
    res <- rlang::sym(".pred")
  } else {
    res <- NULL
  }
  res
}

test_that("prediction types", {
  expect_equal(tune:::pred_type(accuracy), "class")
  expect_equal(tune:::pred_type(roc_auc), "prob")
  expect_equal(tune:::pred_type(rsq), "numeric")
  expect_equal(
    tune:::pred_type(yardstick::metric_set(
      yardstick::accuracy,
      yardstick::roc_auc
    )),
    "unknown"
  )
})

test_that("metric information", {
  set_1 <- tune:::metrics_info(
    yardstick::metric_set(
      yardstick::accuracy,
      yardstick::kap,
      yardstick::roc_auc
    )
  )
  set_2 <- tune:::metrics_info(
    yardstick::metric_set(yardstick::rmse, yardstick::rsq)
  )
  expect_equal(
    set_1,
    tibble::tibble(
      .metric = c("accuracy", "kap", "roc_auc"),
      direction = rep("maximize", 3),
      type = c("class", "class", "prob")
    )
  )
  expect_equal(
    set_2,
    tibble::tibble(
      .metric = c("rmse", "rsq"),
      direction = c("minimize", "maximize"),
      type = c("numeric", "numeric")
    )
  )
})

tune_grid_loop <- function(
  resamples,
  grid,
  workflow,
  param_info,
  metrics,
  eval_time,
  control
) {
  if (is.null(grid)) {
    grid <- tibble::tibble()
  }

  mtr_info <- tibble::as_tibble(metrics)

  control <- .update_parallel_over(control, resamples, grid)

  # Determine how to process the tasks (i.e. parallel engines or sequential)
  strategy <- choose_framework(workflow, control)

  # For last_fit, we want to keep the RNG stream as-is to maintain reproducibility
  # with manual execution. Otherwise, generate parallel seeds even if work is
  # being executed sequentially. `resamples$id` is set in `last_fit_workflow`.
  if (all(resamples$id == "train/test split")) {
    resamples$.seeds <- purrr::map(resamples$id, \(x) integer(0))
  } else {
    # Make and set the worker/process seeds if workers get resamples
    resamples$.seeds <- get_parallel_seeds(nrow(resamples))
  }

  # We'll significantly alter the rsample object but will need it intact later;
  # Make a copy and save some information before proceeding
  rset_info <- pull_rset_attributes(resamples)
  split_args <- rsample::.get_split_args(resamples)

  resamples <- new_bare_tibble(resamples)
  # Break it up row-wise to facilitate mapping/parallelism
  resamples <- vec_list_rowwise(resamples)

  # add an on.exit(); esc puts us in code

  # Notes on debugging:
  # 1. You can set `options(future.debug = TRUE)` to help
  # 2. If you are debugging .loop_over_all_stages, use the control option
  #    `allow_par = FALSE`; that will use `lapply()` so that you can see output.

  # ------------------------------------------------------------------------------

  # fmt: skip
  tm_pkgs <- c("rsample", "workflows", "hardhat", "tune", "parsnip", "tailor",
               "yardstick")
  load_pkgs <- c(required_pkgs(workflow), control$pkgs, tm_pkgs)
  load_pkgs <- unique(load_pkgs)

  is_inst <- purrr::map_lgl(load_pkgs, rlang::is_installed)
  if (any(!is_inst)) {
    nms <- load_pkgs[!is_inst]
    cli::cli_abort(
      "Some package installs are needed: {.pkg {nms}}",
      call = NULL
    )
  }

  par_opt <- list()

  # ----------------------------------------------------------------------------
  # Collect "static" data into a single object for a cleaner interface

  static <- .make_static(
    workflow,
    param_info = param_info,
    grid = grid,
    metrics = metrics,
    eval_time = eval_time,
    split_args = split_args,
    control = control,
    pkgs = load_pkgs,
    strategy = strategy
  )

  # ----------------------------------------------------------------------------
  # Control execution

  # We'll make a call that defines how we iterate over resamples and grid points.
  # That call changes depending on the value of `control$parallel_over`. There
  # are two possible options:
  # - parallel_over == "resamples" means that we lapply() over the existing
  #   `resamples` object.
  # - parallel_over == "everything" means that we lapply() over the every
  #   combination of resamples x grid candidates. That index (called `inds`) has
  #   not been created at this point.
  #
  # We'll create `inds` only if needed and then use the same code to make the
  # call objects (`cl`) that will be executed in the current environment. `cl`
  # will contain a reference to either `resamples` or `inds` depending on the
  # `parallel_over` value in `control`.

  if (control$parallel_over == "everything") {
    # If multiple resamples but preprocessing is cheap (or just a validation set).
    # Loop over grid rows and splits
    candidates <- get_row_wise_grid(workflow, grid)

    # Break all combinations of resamples and candidates into a list of integers
    # for each combination.
    inds <- tidyr::crossing(s = seq_along(candidates), b = seq_along(resamples))

    # Split up for better processing
    inds <- vec_list_rowwise(inds)
  }

  cl <- loop_call(control$parallel_over, strategy, par_opt)
  res <- rlang::eval_bare(cl)

  # ----------------------------------------------------------------------------
  # Separate results into different components

  res <- dplyr::bind_rows(res)

  outcome_names <- unique(unlist(res$outcome_names))
  res$outcome_names <- NULL

  resamples <- dplyr::bind_rows(resamples)
  id_cols <- grep("^id", names(resamples), value = TRUE)

  if (control$parallel_over == "resamples") {
    res <- dplyr::full_join(resamples, res, by = id_cols)
  } else {
    # Clean up for variable columns and make into a function
    pool_cols <- grep("^\\.", names(res), value = TRUE)
    res <- res |>
      dplyr::summarize(
        dplyr::across(dplyr::matches("^\\."), ~ list(purrr::list_rbind(.x))),
        .by = c(!!!id_cols)
      ) |>
      dplyr::full_join(resamples, by = id_cols)
  }

  res <- res |>
    dplyr::relocate(
      splits,
      dplyr::starts_with("id"),
      .metrics,
      .notes,
      dplyr::any_of(".predictions"),
      dplyr::any_of(".extracts")
    )

  res <-
    new_tune_results(
      x = res,
      parameters = param_info,
      metrics = metrics,
      eval_time = eval_time,
      eval_time_target = NULL,
      outcomes = outcome_names,
      rset_info = rset_info,
      workflow = workflow
    )

  res
}

vec_list_rowwise <- function(x) {
  vctrs::vec_split(x, by = 1:nrow(x))$val
}

# nocov start

.onLoad <- function(libname, pkgname) {
  ns <- rlang::ns_env("tune")

  ns$manange_global_limit()

  # Modified version of the cli .onLoad()
  # We can't use cli::symbol$tick because the width of the character
  # looks awful when you output it alongside info / warning characters
  makeActiveBinding(
    "tune_symbol",
    function() {
      # If `cli.unicode` is set we use that
      opt <- getOption("cli.unicode", NULL)

      if (!is.null(opt)) {
        if (isTRUE(opt)) {
          return(tune_symbol_utf8)
        } else {
          return(tune_symbol_ascii)
        }
      }

      # Otherwise we try to auto-detect
      if (cli::is_utf8_output()) {
        tune_symbol_utf8
      } else if (is_latex_output()) {
        tune_symbol_ascii
      } else if (is_windows()) {
        tune_symbol_windows
      } else {
        tune_symbol_ascii
      }
    },
    ns
  )

  makeActiveBinding(
    "tune_color",
    function() {
      opt <- getOption("tidymodels.dark", NULL)

      if (!is.null(opt)) {
        if (isTRUE(opt)) {
          return(tune_color_dark)
        } else {
          return(tune_color_light)
        }
      }

      tune_color_light
    },
    ns
  )
}

# nocov end

#' @import vctrs
NULL

# ------------------------------------------------------------------------------
# tune_results

#' @export
vec_restore.tune_results <- function(x, to, ...) {
  tune_results_reconstruct(x, to)
}


#' @export
vec_ptype2.tune_results.tune_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.tune_results.tune_results")
}
#' @export
vec_ptype2.tune_results.tbl_df <- function(x, y, ...) {
  stop_never_called("vec_ptype2.tune_results.tbl_df")
}
#' @export
vec_ptype2.tbl_df.tune_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.tbl_df.tune_results")
}
#' @export
vec_ptype2.tune_results.data.frame <- function(x, y, ...) {
  stop_never_called("vec_ptype2.tune_results.data.frame")
}
#' @export
vec_ptype2.data.frame.tune_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.data.frame.tune_results")
}


#' @export
vec_cast.tune_results.tune_results <- function(x, to, ...) {
  stop_incompatible_cast_tune_results(x, to, ...)
}
#' @export
vec_cast.tune_results.tbl_df <- function(x, to, ...) {
  stop_incompatible_cast_tune_results(x, to, ...)
}
#' @export
vec_cast.tbl_df.tune_results <- function(x, to, ...) {
  tib_cast(x, to, ...)
}
#' @export
vec_cast.tune_results.data.frame <- function(x, to, ...) {
  stop_incompatible_cast_tune_results(x, to, ...)
}
#' @export
vec_cast.data.frame.tune_results <- function(x, to, ...) {
  df_cast(x, to, ...)
}

# ------------------------------------------------------------------------------
# resample_results

#' @export
vec_restore.resample_results <- function(x, to, ...) {
  resample_results_reconstruct(x, to)
}


#' @export
vec_ptype2.resample_results.resample_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.resample_results.resample_results")
}
#' @export
vec_ptype2.resample_results.tbl_df <- function(x, y, ...) {
  stop_never_called("vec_ptype2.resample_results.tbl_df")
}
#' @export
vec_ptype2.tbl_df.resample_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.tbl_df.resample_results")
}
#' @export
vec_ptype2.resample_results.data.frame <- function(x, y, ...) {
  stop_never_called("vec_ptype2.resample_results.data.frame")
}
#' @export
vec_ptype2.data.frame.resample_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.data.frame.resample_results")
}


#' @export
vec_cast.resample_results.resample_results <- function(x, to, ...) {
  stop_incompatible_cast_resample_results(x, to, ...)
}
#' @export
vec_cast.resample_results.tbl_df <- function(x, to, ...) {
  stop_incompatible_cast_resample_results(x, to, ...)
}
#' @export
vec_cast.tbl_df.resample_results <- function(x, to, ...) {
  tib_cast(x, to, ...)
}
#' @export
vec_cast.resample_results.data.frame <- function(x, to, ...) {
  stop_incompatible_cast_resample_results(x, to, ...)
}
#' @export
vec_cast.data.frame.resample_results <- function(x, to, ...) {
  df_cast(x, to, ...)
}

# ------------------------------------------------------------------------------
# iteration_results

#' @export
vec_restore.iteration_results <- function(x, to, ...) {
  iteration_results_reconstruct(x, to)
}


#' @export
vec_ptype2.iteration_results.iteration_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.iteration_results.iteration_results")
}
#' @export
vec_ptype2.iteration_results.tbl_df <- function(x, y, ...) {
  stop_never_called("vec_ptype2.iteration_results.tbl_df")
}
#' @export
vec_ptype2.tbl_df.iteration_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.tbl_df.iteration_results")
}
#' @export
vec_ptype2.iteration_results.data.frame <- function(x, y, ...) {
  stop_never_called("vec_ptype2.iteration_results.data.frame")
}
#' @export
vec_ptype2.data.frame.iteration_results <- function(x, y, ...) {
  stop_never_called("vec_ptype2.data.frame.iteration_results")
}


#' @export
vec_cast.iteration_results.iteration_results <- function(x, to, ...) {
  stop_incompatible_cast_iteration_results(x, to, ...)
}
#' @export
vec_cast.iteration_results.tbl_df <- function(x, to, ...) {
  stop_incompatible_cast_iteration_results(x, to, ...)
}
#' @export
vec_cast.tbl_df.iteration_results <- function(x, to, ...) {
  tib_cast(x, to, ...)
}
#' @export
vec_cast.iteration_results.data.frame <- function(x, to, ...) {
  stop_incompatible_cast_iteration_results(x, to, ...)
}
#' @export
vec_cast.data.frame.iteration_results <- function(x, to, ...) {
  df_cast(x, to, ...)
}

# ------------------------------------------------------------------------------

stop_incompatible_cast_tune_results <- function(
  x,
  to,
  ...,
  x_arg = "",
  to_arg = ""
) {
  details <- "Can't cast to a <tune_results> because attributes are likely incompatible."
  stop_incompatible_cast(
    x,
    to,
    x_arg = x_arg,
    to_arg = to_arg,
    details = details
  )
}

stop_incompatible_cast_resample_results <- function(
  x,
  to,
  ...,
  x_arg = "",
  to_arg = ""
) {
  details <- "Can't cast to a <resample_results> because attributes are likely incompatible."
  stop_incompatible_cast(
    x,
    to,
    x_arg = x_arg,
    to_arg = to_arg,
    details = details
  )
}

stop_incompatible_cast_iteration_results <- function(
  x,
  to,
  ...,
  x_arg = "",
  to_arg = ""
) {
  details <- "Can't cast to a <iteration_results> because attributes are likely incompatible."
  stop_incompatible_cast(
    x,
    to,
    x_arg = x_arg,
    to_arg = to_arg,
    details = details
  )
}

stop_never_called <- function(fn) {
  cli::cli_abort("Internal error: {.fn {fn}} should never be called.")
}

#' Bootstrap confidence intervals for performance metrics
#'
#' Using out-of-sample predictions, the bootstrap is used to create percentile
#' confidence intervals.
#' @inheritParams collect_predictions
#' @inheritParams rlang::args_dots_empty
#' @inheritParams rsample::int_pctl
#' @inheritParams rsample::bootstraps
#' @param .data A object with class `tune_results` where the `save_pred = TRUE`
#' option was used in the control function.
#' @param metrics A [yardstick::metric_set()]. By default, it uses the same
#' metrics as the original object.
#' @param allow_par A logical to allow parallel processing (if a parallel
#' backend is registered).
#' @param event_level A single string. Either `"first"` or `"second"` to specify
#' which level of truth to consider as the "event".
#' @param eval_time A vector of evaluation times for censored regression models.
#' `NULL` is appropriate otherwise. If `NULL` is used with censored models, a
#' evaluation time is selected, and a warning is issued.
#' @param keep_replicates A logic for saving the individual estimates from each
#' bootstrap sample (as a list column called `.values`).
#' @return A tibble of metrics with additional columns for `.lower` and
#' `.upper` (and potentially, `.values`).
#' @details
#' For each model configuration (if any), this function takes bootstrap samples
#' of the out-of-sample predicted values. For each bootstrap sample, the metrics
#' are computed and these are used to compute confidence intervals.
#' See [rsample::int_pctl()] and the references therein for more details.
#'
#' Note that the `.estimate` column is likely to be different from the results
#' given by [collect_metrics()] since a different estimator is used. Since
#' random numbers are used in sampling, set the random number seed prior to
#' running this function.
#'
#' The number of bootstrap samples should be large to have reliable intervals.
#' The defaults reflect the fewest samples that should be used.
#'
#' The computations for each configuration can be extensive. To increase
#' computational efficiency parallel processing can be used. The \pkg{future}
#' package is used here. To execute the resampling iterations in parallel,
#' specify a [plan][future::plan] with future first. The `allow_par` argument
#' can be used to avoid parallelism.
#'
#' Also, if a censored regression model used numerous evaluation times, the
#' computations can take a long time unless the times are filtered with the
#' `eval_time` argument.
#' @seealso [rsample::int_pctl()]
#' @references Davison, A., & Hinkley, D. (1997). _Bootstrap Methods and their
#'  Application_. Cambridge: Cambridge University Press.
#'  doi:10.1017/CBO9780511802843
#' @examplesIf !tune:::is_cran_check()
#' if (rlang::is_installed("modeldata")) {
#'   data(Sacramento, package = "modeldata")
#'   library(rsample)
#'   library(parsnip)
#'
#'   set.seed(13)
#'   sac_rs <- vfold_cv(Sacramento)
#'
#'   lm_res <-
#'     linear_reg() |>
#'     fit_resamples(
#'       log10(price) ~ beds + baths + sqft + type + latitude + longitude,
#'       resamples = sac_rs,
#'       control = control_resamples(save_pred = TRUE)
#'     )
#'
#'   set.seed(31)
#'   int_pctl(lm_res)
#' }
#' @export
int_pctl.tune_results <- function(
  .data,
  metrics = NULL,
  eval_time = NULL,
  times = 1001,
  parameters = NULL,
  alpha = 0.05,
  allow_par = TRUE,
  event_level = "first",
  keep_replicates = FALSE,
  ...
) {
  rlang::check_dots_empty()

  if (is.null(metrics)) {
    metrics <- .get_tune_metrics(.data)
  } else {
    if (!inherits(metrics, "metric_set")) {
      cli::cli_abort(
        "{.arg metrics} should be a metric set as generated by {.fun yardstick::metric_set}."
      )
    }
  }

  if (is.null(eval_time)) {
    eval_time <- .get_tune_eval_times(.data)
    eval_time <- maybe_choose_eval_time(.data, metrics, eval_time)
  } else {
    eval_time <- unique(eval_time)
    check_eval_time_in_tune_results(.data, eval_time)
    # Are there at least a minimal number of evaluation times?
    check_enough_eval_times(eval_time, metrics)
  }

  .data$.predictions <- filter_predictions_by_eval_time(
    .data$.predictions,
    eval_time
  )

  y_nm <- outcome_names(.data)

  config_keys <- get_configs(
    .data,
    parameters = parameters,
    as_list = FALSE
  )

  metrics_info <- metrics_info(metrics)
  res <-
    int_pctl_iter(
      config_keys,
      .data,
      metrics = metrics,
      times = times,
      allow_par = allow_par,
      event_level = event_level,
      alpha = alpha,
      metrics_info = metrics_info,
      keep_replicates = keep_replicates
    )

  res
}

# ------------------------------------------------------------------------------

int_pctl_iter <- function(
  config,
  x,
  metrics,
  times,
  allow_par,
  event_level,
  alpha,
  metrics_info,
  keep_replicates
) {
  y_nm <- outcome_names(x)
  param_names <- .get_tune_parameter_names(x)

  preds <- collect_predictions(x, summarize = TRUE, parameters = config)

  rs <- rsample::bootstraps(preds, times = times, strata = ".config")

  strategy <-
    choose_framework(
      object = NULL,
      control = list(allow_par = allow_par),
      verbose = FALSE,
      default = "mirai"
    )

  cl <- pctl_call(strategy)
  rs$.metrics <- rlang::eval_bare(cl)

  res <- rsample::int_pctl(rs, .metrics, alpha = alpha)
  has_iter <- any(names(res) == ".iter")

  nest_cols <- c(".config", "term")
  if (has_iter) {
    nest_cols <- c(".iter", nest_cols)
  }

  if (keep_replicates) {
    rs_est <- purrr::list_rbind(rs$.metrics)

    rs_est <-
      vctrs::vec_split(
        x = rs_est[setdiff(colnames(rs_est), nest_cols)],
        by = rs_est[nest_cols]
      )

    rs_est <- vctrs::vec_cbind(
      rs_est$key,
      tibble::new_tibble(list(.values = rs_est$val))
    )

    res <- dplyr::full_join(res, rs_est, by = nest_cols)
  }

  res$.metric <- res$term
  res$term <- NULL
  res$.estimator <- "bootstrap"
  res$.alpha <- NULL
  res$.method <- NULL

  by_cols <- nest_cols[nest_cols != "term"]
  res <- dplyr::inner_join(res, config, by = by_cols)

  sort_cols <- rlang::syms(c(".metric", by_cols))
  res <- res |> dplyr::arrange(!!!sort_cols)
  first_cols <- c(
    param_names,
    ".metric",
    ".estimator",
    ".lower",
    ".estimate",
    ".upper",
    ".config"
  )
  if (has_iter) {
    first_cols <- c(first_cols, ".iter")
  }

  res[, c(first_cols, setdiff(names(res), first_cols))]
}

boot_metrics <- function(
  split,
  y,
  metrics,
  param_names = NULL,
  event_level,
  metrics_info,
  configs
) {
  dat <- rsample::analysis(split)

  res <-
    .estimate_metrics(
      dat,
      metric = metrics,
      param_names = param_names,
      outcome_name = y,
      event_level = event_level,
      metrics_info = metrics_info
    )

  if (length(param_names) > 0) {
    res <- dplyr::inner_join(res, configs, by = param_names)
  } else {
    res <- vctrs::vec_cbind(res, configs)
  }

  res$term <- res$.metric
  res$.metric <- NULL
  res$estimate <- res$.estimate
  res$.estimate <- NULL
  res$.estimator <- NULL
  res
}

# ------------------------------------------------------------------------------

get_configs <- function(x, parameters = NULL, as_list = TRUE) {
  param <- .get_tune_parameter_names(x)
  config_cols <- c(".config", ".iter", param)
  config_keys <-
    collect_metrics(x, summarize = FALSE) |>
    dplyr::distinct(dplyr::pick(dplyr::any_of(config_cols)))
  if (!is.null(parameters)) {
    merge_cols <- intersect(names(config_keys), names(parameters))
    config_keys <- dplyr::inner_join(config_keys, parameters, by = merge_cols)
  }
  if (as_list) {
    config_keys <- vctrs::vec_chop(config_keys, as.list(1:nrow(config_keys)))
  }
  config_keys
}

filter_predictions_by_eval_time <- function(x, eval_time = NULL) {
  if (is.null(eval_time)) {
    return(x)
  }
  purrr::map(x, thin_time, times = eval_time)
}

thin_time <- function(x, times) {
  subset_time <- function(x, times) {
    x[x$.eval_time %in% times, ]
  }

  x$.pred <- purrr::map(x$.pred, subset_time, times = times)
  x
}

# This are new changes for parallel process via future _or_ mirai. These
# functions should be able to replace what is contained in `parallel.R`

# ------------------------------------------------------------------------------

#' Internal functions to help use parallel processing
#' @export
#' @keywords internal
#' @rdname internal-parallel
#' @param object A workflow.
#' @param control A control object
#' @param verbose A logical for printing
has_non_par_pkgs <- function(object, control, verbose = FALSE) {
  pkgs <- character(0)

  if (!is.null(object)) {
    pkgs <- required_pkgs(object)
  }
  if (!is.null(control)) {
    pkgs <- c(pkgs, control$pkgs)
  }
  pkgs <- unique(pkgs)
  if (length(pkgs) == 0) {
    return(FALSE)
  }
  naughty_list <- c("keras", "rJava")
  has_pkg <- pkgs %in% naughty_list
  if (any(has_pkg)) {
    pkgs <- pkgs[has_pkg]
    if (verbose) {
      cli::cli_inform(
        "These packages cannot be used with explicit parallel processing: {.pkg {pkgs}}."
      )
    }
  }
  any(has_pkg)
}

# ------------------------------------------------------------------------------

#' @export
#' @keywords internal
#' @rdname internal-parallel
future_installed <- function() {
  rlang::is_installed("future")
}

#' @export
#' @keywords internal
#' @rdname internal-parallel
mirai_installed <- function() {
  rlang::is_installed("mirai")
}

#' @export
#' @keywords internal
#' @rdname internal-parallel
get_future_workers <- function(verbose) {
  has_future <- future_installed()

  if (has_future) {
    future_workers <- future::nbrOfWorkers()
    if (verbose) {
      if (future_workers == 0) {
        cli::cli_inform(
          "{.pkg future} is not active."
        )
      } else {
        cli::cli_inform(
          "{.pkg future} is active with {future_workers} worker{?s}."
        )
      }
    }
  } else {
    if (verbose) {
      cli::cli_inform("{.pkg future} is not installed.")
    }

    future_workers <- 0L
  }
  future_workers
}

#' @export
#' @keywords internal
#' @rdname internal-parallel
get_mirai_workers <- function(verbose) {
  if (!mirai_installed()) {
    if (verbose) {
      cli::cli_inform("{.pkg mirai} is not installed.")
    }
    return(0L)
  }

  # note connections will be 0 if `!daemons_set()`
  mirai_workers <- mirai::status()$connections

  if (verbose) {
    if (mirai_workers == 0) {
      cli::cli_inform(
        "{.pkg mirai} is not active."
      )
    } else {
      cli::cli_inform(
        "{.pkg mirai} is active with {mirai_workers} worker{?s}."
      )
    }
  }

  mirai_workers
}

#' @export
#' @keywords internal
#' @rdname internal-parallel
#' @param default The default parallel processor.
choose_framework <- function(
  object = NULL,
  control = NULL,
  verbose = FALSE,
  default = "mirai"
) {
  if (!is.null(control)) {
    if (!control$allow_par) {
      return("sequential")
    }
  }

  if (has_non_par_pkgs(object, control, verbose)) {
    return("sequential")
  }

  has_future <- future_installed()
  has_mirai <- mirai_installed()

  if (!has_future & !has_mirai) {
    if (verbose) {
      cli::cli_inform("Neither {.pkg mirai} or {.pkg future} are installed.")
    }
    return("sequential")
  }

  mirai_workers <- get_mirai_workers(verbose)
  future_workers <- get_future_workers(verbose)

  neither <- future_workers < 2 & mirai_workers < 2
  both <- future_workers >= 2 & mirai_workers >= 2

  if (neither) {
    if (verbose) {
      cli::cli_inform("Too few workers for parallel processing.")
    }
    return("sequential")
  }

  if (both) {
    if (verbose) {
      cli::cli_inform(
        "Multiple workers exist for both {.pkg mirai} and {.pkg future};
        falling back to the default of {.pkg {default}}."
      )
    }
    return(default)
  }

  if (future_workers >= 2) {
    res <- "future"
  } else {
    res <- "mirai"
  }

  if (verbose) {
    cli::cli_inform("{.pkg {res}} will be used for parallel processing}.")
  }

  res
}

#' @export
#' @keywords internal
#' @rdname internal-parallel
#' @param workers The number of existing workers
get_parallel_seeds <- function(workers) {
  # Get current rng info and save
  orig_state <- .Random.seed
  orig_kind <- RNGkind()[1]
  # Reset the stream to get new rng's
  on.exit({
    RNGkind(orig_kind)
    assign(".Random.seed", orig_state, globalenv())
  })

  # Set to type used for multiple streams
  RNGkind("L'Ecuyer-CMRG")

  # Capture the seed to make more seeds.
  .seed <- .Random.seed

  res <- vector(mode = "list", length = workers)
  for (i in seq_along(res)) {
    res[[i]] <- parallel::nextRNGSubStream(.seed)
    .seed <- parallel::nextRNGStream(.seed)
  }

  res
}

# ------------------------------------------------------------------------------

#' Support for parallel processing in tune
#'
#' @description
#'
#' \pkg{tune} can enable simultaneous parallel computations. Tierney (2008)
#'  defined different classes of parallel processing techniques:
#'
#'  - _Implicit_ is when a function uses low-level tools to perform a
#'  calculation that is small in scope in parallel. Examples are using
#' multithreaded linear algebra libraries (e.g., BLAS) or basic R vectorization
#' functions.
#'  - _Explicit_ parallelization occurs when the user requests that some
#' calculations should be run by generating multiple new R (sub)processes. These
#' calculations can be more complex than those for implicit parallel
#'  processing.
#'
#' For example, some decision tree libraries can implicitly parallelize their
#' search for the optimal splitting routine using multiple threads.
#'
#' Alternatively, if you are resampling a model _B_ times, you can explicitly
#' create _B_ new R jobs to train _B_ boosted trees in parallel and return their
#' resampling results to the main R process (e.g., [fit_resamples()]).
#'
#' There are two frameworks that can be used to explicitly parallel process
#' your work in \pkg{tune}: the [future][future::future] package and the
#' [mirai][mirai:: mirai] package. Previously, you could use the
#' [foreach][foreach::foreach] package, but this has been deprecated as of
#' version 1.2.1 of tune.
#'
#' By default, no parallelism is used to process models in \pkg{tune}; you have
#' to opt-in.
#'
#' ## Using future
#'
#' You should install the package and choose your flavor of parallelism using
#' the [plan][future::plan] function. This allows you to specify the number of
#' worker processes and the specific technology to use.
#'
#' For example, you can use:
#'
#' ```r
#'    library(future)
#'    plan(multisession, workers = 4)
#' ```
#' and work will be conducted simultaneously (unless there is an exception; see
#' the section below).
#'
#' If you had previously used \pkg{foreach}, this would replace your existing
#' code that probably looked like:
#'
#' ```r
#'    library(doBackend)
#'    registerDoBackend(cores = 4)
#' ```
#'
#' See [future::plan()] for possible options other than `multisession`.
#'
#' Note that \pkg{tune} resets the _maximum_ limit of memory of global variables
#' (e.g., attached packages) to be greater than the default when the package is
#' loaded. This value can be altered using `options(future.globals.maxSize)`.
#'
#' If you want \pkg{future} to use \pkg{mirai} parallel workers, you can
#' install and load the \pkg{future.mirai} package.
#'
#' ## Using mirai
#'
#' To set the specific for parallel processing with \pkg{mirai}, use the
#' [mirai::daemons()] function. The first argument, `n`, determines the number
#' of parallel workers. Using `daemons(0)` reverts to sequential processing.
#'
#' The arguments `url` and `remote` are used to set up and launch parallel
#' processes over the network for distributed computing. See [mirai::daemons()]
#' documentation for more details.
#'
#' ## Reverting to sequential processing
#'
#' There are a few times when you might specify that you wish to use parallel
#' processing, but it will revert to sequential execution:
#'
#' - Many of the control functions (e.g. [control_grid()]) have an argument
#'   called `allow_par`. If this is set to `FALSE`, parallel backends will
#'   always be ignored.
#' - Some packages, such as \pkg{rJava} and \pkg{keras} are not compatible with
#'   explicit parallelization. If any of these packages are used, sequential
#'   processing occurs.
#' - If you specify fewer than two workers, or if there is only a single task,
#'   the computations will occur sequentially.
#'
#' ## Expectations for reproducibility
#'
#' We advise that you _always_ run [set.seed()] with a seed value just prior to
#' using a function that uses (or might use) random numbers. Given this:
#'
#' - You should expect to get the same results if you run that section of code
#'   repeatedly, conditional on using version 1.4.0 of tune.
#' - You should expect differences in results between version 1.4.0 of tune and
#'   previous versions.
#' - When using [last_fit()], you should be able to get the same results as
#'   manually using [generics::fit()] and [predict()] to do the same work.
#' - When running with or without parallel processing (using any backend
#'   package), you should be able to achieve the same results from
#'   [fit_resamples()] and the various tuning functions.
#'
#' Specific exceptions:
#'
#' - For SVM classification models using the \pkg{kernlab} package, the random
#'   number generator is independent of R, and there is no argument to control
#'   it. Unfortunately, it is likely to give you different results from
#'   run-to-run.
#' - For some deep learning packages (e.g., \pkg{tensorflow}, \pkg{keras}, and
#'   \pkg{torch}), it is very difficult to achieve reproducible results. This
#'   is especially true when using GPUs for computations. Additionally, we have
#'   seen differences in computations (stochastic or non-random) between
#'   platforms due to the packages' use of different numerical tolerance
#'   constants across operating systems.
#'
#' ## Handling package dependencies
#'
#' \pkg{tune} knows what packages are required to fit a workflow object.
#'
#' When computations are run sequentially, an initial check is made to see if
#' they are installed. This triggers the packages to be loaded but not visible
#' in the search path.
#'
#' In parallel, the required packages are fully loaded (i.e., loaded and seen
#' in the search path), as they were previously with \pkg{foreach}, in the
#' worker processes (but not the main R session).
#'
#' @references
#' \url{https://www.tmwr.org/grid-search#parallel-processing}
#'
#' Tierney, Luke. "Implicit and explicit parallel computing in R." COMPSTAT
#' 2008: Proceedings in Computational Statistics. Physica-Verlag HD, 2008.
#' @name parallelism
NULL

# ------------------------------------------------------------------------------
# Choosing how to execute the looping structure

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.update_parallel_over <- function(control, resamples, grid) {
  num_candidates <- nrow(grid)

  if (is.null(control$parallel_over) | num_candidates == 0) {
    control$parallel_over <- "resamples"
  }
  if (length(resamples$splits) == 1 & num_candidates > 0) {
    control$parallel_over <- "everything"
  }
  control
}

# mirai_map() acts a little different form map(), lapply(), etc. It requires
# that the elements in .args be the args (not symbols). It also requires an
# extra step to collect the results and coerce them into a list.

#' @export
#' @keywords internal
#' @rdname internal-parallel
#' @param .x A list.
#' @param .f A function
#' @param ...,.args Options to pass to other functions.
#'
eval_mirai <- function(.x, .f, ..., .args) {
  .args <- lapply(.args, get, envir = parent.frame())
  res <- mirai::mirai_map(.x, .f, ..., .args = .args)
  mirai::collect_mirai(res)
}

#' @export
#' @keywords internal
#' @rdname internal-parallel
.par_fns <- function(framework) {
  if (framework == "sequential") {
    res <- list(fn = "lapply", ns = NULL)
  } else if (framework == "future") {
    res <- list(fn = "future_lapply", ns = "future.apply")
  } else if (framework == "mirai") {
    res <- list(fn = "eval_mirai", ns = NULL)
  } else {
    cli::cli_abort("Frmework {.val framework} is unknown.")
  }
  res
}

#'@export
#'@keywords internal
#'@rdname empty_ellipses
loop_call <-
  function(strategy, framework, opts) {
    if (strategy == "resamples") {
      base_cl <- rlang::call2(
        .par_fns(framework)[[1]],
        .ns = .par_fns(framework)[[2]],
        quote(resamples),
        quote(tune::.loop_over_all_stages)
      )
      base_args <- list(grid = quote(grid), static = quote(static))
    } else {
      base_cl <- rlang::call2(
        .par_fns(framework)[[1]],
        .ns = .par_fns(framework)[[2]],
        quote(inds),
        quote(tune::.loop_over_all_stages2)
      )
      base_args <- list(
        resamples = quote(resamples),
        grid = quote(candidates),
        static = quote(static)
      )
    }

    # Configure arguments
    base_args <- c(base_args, opts)

    if (framework == "future") {
      rlang::check_installed("future")

      future_opts <- list(
        future.label = "tune-grid-%d",
        future.stdout = TRUE,
        future.seed = TRUE,
        future.globals = names(base_args)
      )
      base_args <- c(base_args, future_opts)
    }

    if (framework == "mirai") {
      rlang::check_installed("mirai")
      cl <- rlang::call_modify(base_cl, .args = base_args)
    } else {
      cl <- rlang::call_modify(base_cl, !!!base_args)
    }
    cl
  }

# ------------------------------------------------------------------------------
# for int_pctl

pctl_call <- function(framework, args = list()) {
  if (framework == "future") {
    rlang::check_installed("future")

    future_opts <- list(
      future.label = "int-pctl-%d",
      future.stdout = TRUE,
      future.seed = TRUE
    )
    if (!is.null(names(args))) {
      future_opts$future.globals <- names(args)
    }
    args <- c(args, future_opts)
  }

  main_args <- list(
    y = quote(y_nm),
    metrics = quote(metrics),
    event_level = quote(event_level),
    metrics_info = quote(metrics_info),
    param_names = quote(param_names),
    configs = quote(config)
  )
  args <- c(main_args, args)

  base_cl <- rlang::call2(
    .par_fns(framework)[[1]],
    .ns = .par_fns(framework)[[2]],
    quote(rs$splits),
    quote(boot_metrics)
  )
  if (framework == "mirai") {
    rlang::check_installed("mirai")
    cl <- rlang::call_modify(base_cl, .args = args)
  } else {
    cl <- rlang::call_modify(base_cl, !!!args)
  }

  cl
}

# ------------------------------------------------------------------------------

manange_global_limit <- function(min = 1e9) {
  currrent_value <- getOption("future.globals.maxSize")
  if (is.null(currrent_value)) {
    options(future.globals.maxSize = min)
  } else {
    if (currrent_value < min) {
      options(future.globals.maxSize = min)
    }
  }
  invisible(NULL)
}

test_that("no parallelism", {
  skip_if_not_installed("mirai")

  svm_spec <- svm_rbf(mode = "classification")
  ctrl_no <- control_grid(allow_par = FALSE)
  ctrl_java <- control_grid(pkgs = "rJava")

  # ---------------------------------------------------------------------------
  # default

  expect_equal(tune:::choose_framework(), "sequential")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))

  # ----------------------------------------------------------------------------
  # Due to

  expect_equal(tune:::choose_framework(control = ctrl_no), "sequential")
  expect_equal(tune:::choose_framework(svm_spec), "sequential")
})

test_that("enable future parallelism", {
  skip_if_not_installed("future")
  skip_if_not_installed("mirai")

  # temporary issue with GHA
  skip_if(grepl("^darwin", R.version$os) & Sys.Date() < as.Date("2025-08-01"))

  svm_spec <- svm_rbf(mode = "classification")
  ctrl_no <- control_grid(allow_par = FALSE)
  ctrl_java <- control_grid(pkgs = "rJava")

  library(future)

  # ----------------------------------------------------------------------------
  # sequential; not enough workers

  expect_equal(tune:::choose_framework(), "sequential")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))

  # ----------------------------------------------------------------------------
  # parallel

  plan(multisession(workers = 2))
  expect_equal(tune:::choose_framework(), "future")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))
  expect_equal(tune:::choose_framework(svm_spec), "future")

  # ----------------------------------------------------------------------------
  # sequential due to restrictions

  expect_equal(tune:::choose_framework(control = ctrl_no), "sequential")
  expect_snapshot(tune:::choose_framework(control = ctrl_no, verbose = TRUE))

  expect_equal(tune:::choose_framework(control = ctrl_java), "sequential")
  expect_snapshot(tune:::choose_framework(control = ctrl_java, verbose = TRUE))

  plan(multisession(workers = 1))

  expect_equal(tune:::choose_framework(), "sequential")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))

  plan(sequential)
  detach("package:future", character.only = TRUE)

  expect_equal(tune:::choose_framework(), "sequential")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))
})

test_that("enable mirai parallelism", {
  skip_if_not_installed("mirai", minimum_version = "2.4.0")

  svm_spec <- svm_rbf(mode = "classification")
  ctrl_no <- control_grid(allow_par = FALSE)
  ctrl_java <- control_grid(pkgs = "rJava")

  library(mirai)

  # ----------------------------------------------------------------------------
  # sequential; not enough workers

  expect_equal(tune:::choose_framework(), "sequential")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))

  # ----------------------------------------------------------------------------
  # parallel

  daemons(2)

  expect_equal(tune:::choose_framework(), "mirai")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))
  expect_equal(tune:::choose_framework(svm_spec), "mirai")

  # ----------------------------------------------------------------------------
  # sequential due to restrictions

  expect_equal(tune:::choose_framework(control = ctrl_no), "sequential")
  expect_snapshot(tune:::choose_framework(control = ctrl_no, verbose = TRUE))

  expect_equal(tune:::choose_framework(control = ctrl_java), "sequential")
  expect_snapshot(tune:::choose_framework(control = ctrl_java, verbose = TRUE))

  daemons(0)
  daemons(1)

  expect_equal(tune:::choose_framework(), "sequential")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))

  # teardown
  daemons(0)
  ## Give daemons a chance to shutdown
  Sys.sleep(1)

  expect_equal(tune:::choose_framework(), "sequential")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))
})

test_that("break parallelism tie", {
  skip_if_not_installed("mirai", minimum_version = "2.4.0")
  skip_if_not_installed("future")
  skip_on_cran()

  # temporary issue with GHA
  skip_if(grepl("^darwin", R.version$os) & Sys.Date() < as.Date("2025-08-01"))

  library(mirai)
  daemons(2)

  library(future)
  plan(multisession(workers = 2))

  expect_equal(tune:::choose_framework(), "mirai")
  expect_snapshot(tune:::choose_framework(verbose = TRUE))

  # teardown
  daemons(0)
  ## Give daemons a chance to shutdown
  Sys.sleep(1)

  plan(sequential)
})

test_that("loop execution code", {
  skip_if_not_installed("mirai")

  # sequential
  expect_snapshot(tune:::loop_call("resamples", "sequential", list()))
  expect_snapshot(
    tune:::loop_call(
      "resamples",
      "sequential",
      list(a = quote(a))
    )
  )
  expect_snapshot(tune:::loop_call("everything", "sequential", list()))
  expect_snapshot(
    tune:::loop_call(
      "everything",
      "sequential",
      list(a = quote(a))
    )
  )

  # future
  expect_snapshot(tune:::loop_call("resamples", "future", list()))
  expect_snapshot(
    tune:::loop_call(
      "resamples",
      "future",
      list(a = quote(a))
    )
  )
  expect_snapshot(tune:::loop_call("everything", "future", list()))
  expect_snapshot(
    tune:::loop_call(
      "everything",
      "future",
      list(a = quote(a))
    )
  )

  # mirai
  expect_snapshot(tune:::loop_call("resamples", "mirai", list()))
  expect_snapshot(
    tune:::loop_call(
      "resamples",
      "mirai",
      list(a = quote(a))
    )
  )
  expect_snapshot(tune:::loop_call("everything", "mirai", list()))
  expect_snapshot(
    tune:::loop_call(
      "everything",
      "mirai",
      list(a = quote(a))
    )
  )
})


test_that("same results using mirai", {
  skip_if_not_installed("mirai")
  skip_if_not_installed("xgboost")
  skip_if_not_installed("modeldata")
  skip_on_cran()

  set.seed(1)
  dat <- modeldata::sim_regression(500)
  rs <- vfold_cv(dat)

  # ------------------------------------------------------------------------------

  mod <- boost_tree(min_n = tune(), trees = 20, learn_rate = tune()) |>
    set_mode("regression")

  simple_wflow <- workflow(outcome ~ ., mod)

  set.seed(1)
  seq_res <-
    simple_wflow |>
    tune_grid(
      resamples = rs,
      grid = 4
    )

  seq_mtr <- collect_metrics(seq_res)

  tmp <- mirai::daemons(2)

  set.seed(1)
  mirai_res <-
    simple_wflow |>
    tune_grid(
      resamples = rs,
      grid = 4
    )

  mirai_mtr <- collect_metrics(mirai_res)

  expect_equal(seq_mtr, mirai_mtr)

  tmp <- mirai::daemons(0)
})


test_that("same results using future", {
  skip_if_not_installed("future")
  skip_if_not_installed("future.apply")
  skip_if_not_installed("xgboost")
  skip_if_not_installed("modeldata")
  skip_on_cran()

  # temporary issue with GHA
  skip_if(grepl("^darwin", R.version$os) & Sys.Date() < as.Date("2025-08-01"))

  set.seed(1)
  dat <- modeldata::sim_regression(500)
  rs <- vfold_cv(dat)

  # ------------------------------------------------------------------------------

  mod <- boost_tree(min_n = tune(), trees = 20, learn_rate = tune()) |>
    set_mode("regression")

  simple_wflow <- workflow(outcome ~ ., mod)

  set.seed(1)
  seq_res <-
    simple_wflow |>
    tune_grid(
      resamples = rs,
      grid = 4
    )

  seq_mtr <- collect_metrics(seq_res)

  future::plan(future::multisession(workers = 2))

  set.seed(1)
  future_res <-
    simple_wflow |>
    tune_grid(
      resamples = rs,
      grid = 4
    )

  future_mtr <- collect_metrics(future_res)

  expect_equal(seq_mtr, future_mtr)

  future::plan("sequential")
})


test_that("generating parallel seeds does not affect the space-time continuum", {
  og_seed <- .Random.seed

  before_kind <- RNGkind()
  set.seed(1)
  before_rnd <- runif(1)

  set.seed(1)
  tmp <- tune:::get_parallel_seeds(3)
  after_rng <- runif(1)

  expect_equal(3, length(tmp))
  expect_equal(before_rnd, after_rng)
  expect_equal(before_kind, RNGkind())

  # The seeds work
  assign(".Random.seed", tmp[[1]], envir = .GlobalEnv)
  unif_1 <- runif(1)

  assign(".Random.seed", tmp[[1]], envir = .GlobalEnv)
  unif_2 <- runif(1)

  expect_equal(unif_1, unif_2)
  assign(".Random.seed", og_seed, envir = .GlobalEnv)
})

#' Determine names of the outcome data in a workflow
#'
#' @param x An object.
#' @param ... Not used.
#' @return A character string of variable names
#' @keywords internal
#' @examples
#' library(dplyr)
#' lm(cbind(mpg, wt) ~ ., data = mtcars) |>
#'   purrr::pluck(terms) |>
#'   outcome_names()
#' @export
outcome_names <- function(x, ...) {
  UseMethod("outcome_names")
}

#' @export
#' @rdname outcome_names
outcome_names.terms <- function(x, ...) {
  if (length(x) == 2) {
    res <- character(0)
  } else {
    res <- all.vars(x[[2]])
  }
  res
}

#' @export
#' @rdname outcome_names
outcome_names.formula <- outcome_names.terms


#' @export
#' @rdname outcome_names
outcome_names.recipe <- function(x, ...) {
  y <- summary(x)
  y$variable[y$role == "outcome" & !is.na(y$role)]
}

#' @export
#' @rdname outcome_names
outcome_names.workflow <- function(x, ...) {
  if (inherits(extract_spec_parsnip(x), "cluster_spec")) {
    return(NULL)
  }

  if (!is.null(x$pre$mold)) {
    y_vals <- extract_mold(x)$outcomes
    res <- colnames(y_vals)
  } else {
    preprocessor <- extract_preprocessor(x)
    res <- outcome_names(preprocessor, ...)
  }
  res
}

#' @export
#' @rdname outcome_names
outcome_names.tune_results <- function(x, ...) {
  att <- attributes(x)
  if (any(names(att) == "outcomes")) {
    res <- att$outcomes
  } else {
    res <- NA_character_
  }
  res
}

#' @export
#' @rdname outcome_names
outcome_names.workflow_variables <- function(x, data = NULL, ...) {
  if (!inherits(data, "data.frame")) {
    cli::cli_abort(
      "When using {.fn outcome_names} on an object for class {.cls {class(x)}}, the
      argument {.arg data} should be a data frame, not {.obj_type_friendly {data}}."
    )
  }
  names(dplyr::select(data, !!x$outcomes))
}

test_that("base R objects", {
  lm_1 <- lm(mpg ~ ., data = mtcars)
  expect_equal(outcome_names(lm_1$terms), "mpg")

  lm_2 <- lm(cbind(mpg, wt) ~ ., data = mtcars)
  expect_equal(outcome_names(lm_2$terms), c("mpg", "wt"))

  f_1 <- mpg ~ .
  expect_equal(outcome_names(f_1), "mpg")

  f_2 <- cbind(mpg, wt) ~ .
  expect_equal(outcome_names(f_2), c("mpg", "wt"))

  f_3 <- ~.
  expect_equal(outcome_names(f_3), character(0))
})

## -----------------------------------------------------------------------------

test_that("recipes", {
  rec_1 <- recipes::recipe(mpg ~ ., data = mtcars)
  expect_equal(outcome_names(rec_1), "mpg")
  expect_equal(outcome_names(recipes::prep(rec_1)), "mpg")

  rec_2 <- recipes::recipe(mpg + wt ~ ., data = mtcars)
  expect_equal(outcome_names(rec_2), c("mpg", "wt"))
  expect_equal(outcome_names(recipes::prep(rec_2)), c("mpg", "wt"))

  rec_3 <- recipes::recipe(~., data = mtcars)
  expect_equal(outcome_names(rec_3), character(0))
  expect_equal(outcome_names(recipes::prep(rec_3)), character(0))

  rec_4 <- recipes::recipe(mpg ~ ., data = mtcars) |> recipes::step_rm(mpg)
  expect_equal(outcome_names(rec_4), "mpg")
  expect_equal(outcome_names(recipes::prep(rec_4)), character(0))

  rec_5 <- recipes::recipe(mtcars)
  expect_equal(outcome_names(rec_5), character())
  expect_equal(outcome_names(recipes::prep(rec_5)), character(0))

  rec_6 <- recipes::recipe(mtcars) |>
    recipes::update_role(mpg, new_role = "outcome")
  expect_equal(outcome_names(rec_6), "mpg")
  expect_equal(outcome_names(recipes::prep(rec_6)), "mpg")

  rec_7 <- recipes::recipe(mtcars) |>
    recipes::update_role(mpg, disp, new_role = "outcome")
  expect_equal(outcome_names(rec_7), c("mpg", "disp"))
  expect_equal(outcome_names(recipes::prep(rec_7)), c("mpg", "disp"))

  rec_8 <- recipes::recipe(mtcars) |>
    recipes::update_role(mpg, disp, new_role = "outcome") |>
    recipes::step_rm(mpg)
  expect_equal(outcome_names(rec_8), c("mpg", "disp"))
  expect_equal(outcome_names(recipes::prep(rec_8)), "disp")
})


## -----------------------------------------------------------------------------

test_that("workflows + recipes", {
  rec_1 <- recipes::recipe(mpg ~ ., data = mtcars)
  rec_2 <- recipes::recipe(mpg + wt ~ ., data = mtcars)
  rec_3 <- recipes::recipe(mtcars) |>
    recipes::update_role(mpg, new_role = "outcome")
  lm_mod <- parsnip::linear_reg() |> parsnip::set_engine("lm")
  wflow <- workflow() |> add_model(lm_mod)

  wflow_1 <- wflow |> add_recipe(rec_1)
  expect_equal(outcome_names(wflow_1), "mpg")
  expect_equal(outcome_names(parsnip::fit(wflow_1, mtcars)), "mpg")

  wflow_2 <- wflow |> add_recipe(rec_2)
  expect_equal(outcome_names(wflow_2), c("mpg", "wt"))
  expect_equal(outcome_names(parsnip::fit(wflow_2, mtcars)), c("mpg", "wt"))

  wflow_3 <- wflow |> add_recipe(rec_3)
  expect_equal(outcome_names(wflow_3), "mpg")
  expect_equal(outcome_names(parsnip::fit(wflow_3, mtcars)), "mpg")
})


## -----------------------------------------------------------------------------

test_that("workflows + formulas", {
  lm_mod <- parsnip::linear_reg() |> parsnip::set_engine("lm")
  wflow <- workflow() |> add_model(lm_mod)

  wflow_1 <- wflow |> add_formula(mpg ~ .)
  expect_equal(outcome_names(wflow_1), "mpg")
  expect_equal(outcome_names(parsnip::fit(wflow_1, mtcars)), "mpg")

  wflow_2 <- wflow |> add_formula(mpg + wt ~ .)
  expect_equal(outcome_names(wflow_2), c("mpg", "wt"))
  expect_equal(outcome_names(parsnip::fit(wflow_2, mtcars)), c("mpg", "wt"))
})


## -----------------------------------------------------------------------------

test_that("tune_results objects", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  res <- lin_mod |>
    fit_resamples(mpg ~ ., folds) |>
    outcome_names()
  expect_equal(res, "mpg")
})

#' Remove some tuning parameter results
#'
#' For objects produced by the `tune_*()` functions, there may only be a subset
#' of tuning parameter combinations of interest. For large data sets, it might be
#' helpful to be able to remove some results. This function trims the `.metrics`
#' column of unwanted results as well as columns `.predictions` and `.extracts`
#' (if they were requested).
#'
#' @param x An object of class `tune_results` that has multiple tuning parameters.
#' @param ... Expressions that return a logical value, and are defined in terms
#'  of the tuning parameter values. If multiple expressions are included, they
#'  are combined with the `&` operator. Only rows for which all conditions
#'  evaluate to `TRUE` are kept.
#' @param parameters A tibble of tuning parameter values that can be used to
#'  filter the predicted values before processing. This tibble should only have
#'  columns for tuning parameter identifiers (e.g. `"my_param"` if
#'  `tune("my_param")` was used). There can be multiple rows and one or more
#'  columns. **If used, this parameter must be named.**
#' @return A version of `x` where the lists columns only retain the parameter
#' combinations in `parameters` or satisfied by the filtering logic.
#' @examples
#' library(dplyr)
#' library(tibble)
#'
#' # For grid search:
#' data("example_ames_knn")
#'
#' ## -----------------------------------------------------------------------------
#' # select all combinations using the 'rank' weighting scheme
#'
#' ames_grid_search |>
#'   collect_metrics()
#'
#' filter_parameters(ames_grid_search, weight_func == "rank") |>
#'   collect_metrics()
#'
#' rank_only <- tibble::tibble(weight_func = "rank")
#' filter_parameters(ames_grid_search, parameters = rank_only) |>
#'   collect_metrics()
#'
#' ## -----------------------------------------------------------------------------
#' # Keep only the results from the numerically best combination
#'
#' ames_iter_search |>
#'   collect_metrics()
#'
#' best_param <- select_best(ames_iter_search, metric = "rmse")
#' ames_iter_search |>
#'   filter_parameters(parameters = best_param) |>
#'   collect_metrics()
#' @details
#' Removing some parameter combinations might affect the results of `autoplot()`
#' for the object.
#' @export
filter_parameters <- function(x, ..., parameters = NULL) {
  cl_x <- as.character(match.call()$x)
  check_filter_dots(rlang::enquos(...))
  # check for type
  if (!inherits(x, "tune_results")) {
    cli::cli_abort(
      "{.arg {caller_arg(x)}} should have class {.cls tune_results};
                    {.obj_type_friendly {x}} was passed."
    )
  }
  x <- filter_by_join(x, parameters, nm = cl_x)
  x <- filter_by_filter(x, ...)
  x
}

filter_by_join <- function(x, parameters = NULL, nm = "") {
  if (is.null(parameters)) {
    return(x)
  }
  # check parameters vs x
  param_names <- .get_tune_parameter_names(x)
  filter_names <- names(parameters)
  filter_names <- filter_names[filter_names != ".config"]
  if (length(intersect(filter_names, param_names)) == 0) {
    cli::cli_abort(
      "There are no columns in {.arg parameters} that match with {.val {nm}}."
    )
  }
  extra_names <- setdiff(filter_names, param_names)
  if (length(extra_names) > 0) {
    cli_warn(
      "{qty(extra_names)} The column{?s} {.var {extra_names}} passed in \\
       {.arg parameters} {?is/are} not needed and will be ignored."
    )

    parameters <- parameters[, filter_names %in% param_names]
  }

  # run a test to make sure that there are no issues in filtering
  tst_orig <- bind_rows(x$.metrics)
  tst_filtered <- filter_join_iter(tst_orig, parameters)
  if (nrow(tst_filtered) == 0) {
    cli::cli_abort("No parameter combinations were selected using your subset.")
  }

  x$.metrics <- purrr::map(x$.metrics, filter_join_iter, .subset = parameters)
  if (any(names(x) == ".predictions")) {
    x$.predictions <- purrr::map(
      x$.predictions,
      filter_join_iter,
      .subset = parameters
    )
  }
  if (any(names(x) == ".extracts")) {
    x$.extracts <- purrr::map(
      x$.extracts,
      filter_join_iter,
      .subset = parameters
    )
  }

  x
}

filter_join_iter <- function(x, .subset) {
  if (isTRUE(nrow(x) == 0) || all(is.null(x))) {
    return(x)
  }
  dplyr::inner_join(x, .subset, by = names(.subset))
}

filter_by_filter <- function(x, ...) {
  dots <- rlang::enquos(...)
  if (rlang::is_empty(dots)) {
    return(x)
  }
  x$.metrics <- purrr::map(x$.metrics, ~ dplyr::filter(.x, !!!dots))
  if (any(names(x) == ".predictions")) {
    x$.predictions <- purrr::map(x$.predictions, \(.x) {
      dplyr::filter(.x, !!!dots)
    })
  }
  if (any(names(x) == ".extracts")) {
    x$.extracts <- purrr::map(x$.extracts, \(.x) dplyr::filter(.x, !!!dots))
  }
  x
}

check_filter_dots <- function(dots, call = rlang::caller_env()) {
  res <- purrr::map(dots, \(.x) try(rlang::eval_tidy(.x), silent = TRUE))

  if (any(purrr::map_lgl(res, inherits, "data.frame"))) {
    cli::cli_abort(
      c(
        "An element passed to {.arg ...} is a data frame rather than a filter
         expression.",
        "i" = "Did you forget to name the {.arg parameters} argument?"
      ),
      call = call
    )
  }
}

# As an example, if we fit a boosted tree  model and tune over
# trees = 1:20 and min_n = c(20, 30)
# we should only have to fit two models:
#
#   trees = 20 & min_n = 20
#   trees = 20 & min_n = 30
#
# The logic related to how this "mini grid" gets made is model-specific.
#
# To get the full set of predictions, we need to know, for each of these two
# models, what values of num_terms to give to the multi_predict() function.
#
# The current idea is to have a list column of the extra models for prediction.
# For the example above:
#
#   # A tibble: 2 x 3
#     trees min_n .submodels
#     <dbl> <dbl> <list>
#   1    20    20 <named list [1]>
#   2    20    30 <named list [1]>
#
# and the .submodels would both be
#
#  list(trees = 1:19)
#
# There are a lot of other things to consider in future versions like grids
# where there are multiple columns with the same name (maybe the results of
# a recipe) and so on.

#' Determine the minimum set of model fits
#'
#' `min_grid()` determines exactly what models should be fit in order to
#'  evaluate the entire set of tuning parameter combinations. This is for
#'  internal use only and the API may change in the near future.
#'
#' `fit_max_value()` can be used in other packages to implement a `min_grid()`
#' method.
#' @param x A model specification.
#' @param grid A tibble with tuning parameter combinations.
#' @param ... Not currently used.
#' @return A tibble with the minimum tuning parameters to fit and an additional
#' list column with the parameter combinations used for prediction.
#' @keywords internal
#' @examplesIf tune:::should_run_examples(suggests = "xgboost")
#' library(dplyr)
#' library(dials)
#' library(parsnip)
#'
#' ## -----------------------------------------------------------------------------
#' ## No ability to exploit submodels:
#'
#' svm_spec <-
#'   svm_poly(cost = tune(), degree = tune()) |>
#'   set_engine("kernlab") |>
#'   set_mode("regression")
#'
#' svm_grid <-
#'   svm_spec |>
#'   extract_parameter_set_dials() |>
#'   grid_regular(levels = 3)
#'
#' min_grid(svm_spec, svm_grid)
#'
#' ## -----------------------------------------------------------------------------
#' ## Can use submodels
#'
#' xgb_spec <-
#'   boost_tree(trees = tune(), min_n = tune()) |>
#'   set_engine("xgboost") |>
#'   set_mode("regression")
#'
#' xgb_grid <-
#'   xgb_spec |>
#'   extract_parameter_set_dials() |>
#'   grid_regular(levels = 3)
#'
#' min_grid(xgb_spec, xgb_grid)
#' @export
#' @rdname min_grid
min_grid.model_spec <- function(x, grid, ...) {
  # x is a `model_spec` object from parsnip
  # grid is a tibble of tuning parameter values with names
  #  matching the parameter names.
  blank_submodels(grid)
}

# ------------------------------------------------------------------------------
# helper functions

# Template for model results that do no have the sub-model feature
blank_submodels <- function(grid) {
  grid |>
    dplyr::mutate(.submodels = purrr::map(1:nrow(grid), \(x) list())) |>
    dplyr::mutate_if(is.factor, as.character)
}

get_submodel_info <- function(spec) {
  if (is.null(spec$engine)) {
    cli::cli_abort("Please set the model's engine.")
  }
  param_info <-
    get_from_env(paste0(class(spec)[1], "_args")) |>
    dplyr::filter(engine == spec$engine) |>
    dplyr::select(name = parsnip, has_submodel) |>
    dplyr::full_join(
      hardhat::extract_parameter_set_dials(spec) |> dplyr::select(name, id),
      by = "name"
    ) |>
    dplyr::mutate(id = ifelse(is.na(id), name, id)) |>
    # In case the parameter is an engine parameter
    dplyr::mutate(
      has_submodel = ifelse(is.na(has_submodel), FALSE, has_submodel)
    )

  param_info
}

# Assumes only one sub-model parameter and that the fitted one is the
# maximum value
submod_only <- function(grid) {
  if (nrow(grid) == 1) {
    grid$.submodels <- list(list())
    return(grid)
  }
  nm <- colnames(grid)[1]
  fit_only <- tibble::new_tibble(
    list(nm = max(grid[[nm]], na.rm = TRUE)),
    nrow = 1
  )
  names(fit_only) <- nm
  sub_mods <- list(grid[[nm]][-which.max(grid[[nm]])])
  names(sub_mods) <- nm
  fit_only$.submodels <- list(sub_mods)
  dplyr::select(fit_only, dplyr::all_of(names(grid)), .submodels)
}

# Assumes only one sub-model parameter and that the fitted one is the
# maximum value
submod_and_others <- function(grid, fixed_args) {
  orig_names <- names(grid)
  subm_nm <- orig_names[!(orig_names %in% fixed_args)]

  # avoid more rlangedness related to names until end:
  grid <- grid |> dplyr::rename(..val = !!subm_nm)

  fit_only <-
    grid |>
    dplyr::group_by(!!!rlang::syms(fixed_args)) |>
    dplyr::summarize(max_val = max(..val, na.rm = TRUE)) |>
    dplyr::ungroup()

  if (utils::packageVersion("dplyr") >= "1.0.99.9000") {
    min_grid_df <-
      dplyr::full_join(fit_only, grid, by = fixed_args, multiple = "all")
  } else {
    min_grid_df <-
      dplyr::full_join(fit_only, grid, by = fixed_args)
  }

  min_grid_df <- min_grid_df |>
    dplyr::filter(..val != max_val) |>
    dplyr::group_by(!!!rlang::syms(fixed_args)) |>
    dplyr::summarize(.submodels = list(tibble::lst(!!subm_nm := ..val))) |>
    dplyr::ungroup()

  if (utils::packageVersion("dplyr") >= "1.0.99.9000") {
    min_grid_df <- min_grid_df |>
      dplyr::full_join(fit_only, by = fixed_args, multiple = "all")
  } else {
    min_grid_df <- min_grid_df |>
      dplyr::full_join(fit_only, by = fixed_args)
  }

  min_grid_df <- min_grid_df |>
    dplyr::rename(!!subm_nm := max_val)

  min_grid_df$.submodels <-
    dplyr::if_else(
      !purrr::map_lgl(min_grid_df$.submodels, rlang::is_null),
      min_grid_df$.submodels,
      purrr::map(1:nrow(min_grid_df), \(x) list())
    )

  dplyr::select(min_grid_df, dplyr::all_of(orig_names), .submodels) |>
    dplyr::mutate_if(is.factor, as.character)
}

# Determine the correct sub-model structure when the sub-model parameter's
# fit value should be the maximum (e.g. fit the largest number of boosting
# iterations and use muti_predict() for the others)
# Assumes a single sub-model parameter
#' @export
#' @rdname min_grid
fit_max_value <- function(x, grid, ...) {
  gr_nms <- names(grid)
  param_info <- get_submodel_info(x)
  sub_nm <- param_info$id[param_info$has_submodel]

  if (length(sub_nm) == 0 | !any(names(grid) %in% sub_nm)) {
    return(blank_submodels(grid))
  }

  fixed_args <- gr_nms[gr_nms != sub_nm]

  if (length(fixed_args) == 0) {
    res <- submod_only(grid)
  } else {
    res <- submod_and_others(grid, fixed_args)
  }
  res
}


# ------------------------------------------------------------------------------
# specific methods

# ------------------------------------------------------------------------------
# Boosted trees

#' @export
#' @export min_grid.boost_tree
#' @rdname min_grid
min_grid.boost_tree <- fit_max_value

# ------------------------------------------------------------------------------
# linear regression

#' @export
#' @export min_grid.linear_reg
#' @rdname min_grid
min_grid.linear_reg <- function(x, grid, ...) {
  # This is basically `fit_max_value()` with an extra error trap
  gr_nms <- names(grid)
  param_info <- get_submodel_info(x)
  sub_nm <- param_info$id[param_info$has_submodel]

  if (x$engine == "glmnet") {
    no_penalty(grid, sub_nm)
  }

  if (length(sub_nm) == 0) {
    return(blank_submodels(grid))
  }

  fixed_args <- gr_nms[gr_nms != sub_nm]

  if (length(fixed_args) == 0) {
    res <- submod_only(grid)
  } else {
    res <- submod_and_others(grid, fixed_args)
  }
  res
}

no_penalty <- function(x, nm) {
  if (length(nm) == 0 || all(colnames(x) != nm)) {
    cli::cli_abort("At least one penalty value is required for {.pkg glmnet}.")
  }
  invisible(NULL)
}

# ------------------------------------------------------------------------------
# logistic regression

#' @export
#' @export min_grid.logistic_reg
#' @rdname min_grid
min_grid.logistic_reg <- min_grid.linear_reg

# ------------------------------------------------------------------------------
# mars

#' @export
#' @export min_grid.mars
#' @rdname min_grid
min_grid.mars <- fit_max_value

# ------------------------------------------------------------------------------
# multinomial regression

#' @export
#' @export min_grid.multinom_reg
#' @rdname min_grid
min_grid.multinom_reg <- min_grid.linear_reg

# ------------------------------------------------------------------------------
# Cox PH regression

#' @export
#' @export min_grid.proportional_hazards
#' @rdname min_grid
min_grid.proportional_hazards <- min_grid.linear_reg

# ------------------------------------------------------------------------------
# Knn

#' @export
#' @export min_grid.nearest_neighbor
#' @rdname min_grid
min_grid.nearest_neighbor <- fit_max_value

## -----------------------------------------------------------------------------
## rules package

#' @export
#' @export min_grid.cubist_rules
#' @rdname min_grid
min_grid.cubist_rules <- fit_max_value

#' @export
#' @export min_grid.C5_rules
#' @rdname min_grid
min_grid.C5_rules <- fit_max_value

#' @export
#' @export min_grid.rule_fit
#' @rdname min_grid
min_grid.rule_fit <- fit_max_value

## -----------------------------------------------------------------------------
## plsmod package

#' @export
#' @export min_grid.pls
#' @rdname min_grid
min_grid.pls <- fit_max_value

## -----------------------------------------------------------------------------
## poissonreg package

#' @export
#' @export min_grid.poisson_reg
#' @rdname min_grid
min_grid.poisson_reg <- fit_max_value

trim_ipcw <- function(x) {
  x$.weight_time <- NULL
  x$.pred_censored <- NULL
  x
}

maybe_add_ipcw <- function(.data, model, types) {
  if (!any(types == "survival")) {
    return(.data)
  }
  res <- parsnip::.censoring_weights_graf(model, .data)
  res$.pred <- purrr::map(res$.pred, trim_ipcw)
  res
}

#' Get time for analysis of dynamic survival metrics
#' @param metrics A metric set.
#' @param eval_time A vector of evaluation times.
#' @export
#' @keywords internal
get_metric_time <- function(metrics, eval_time) {
  info <- tibble::as_tibble(metrics)
  if (any(info$class == "dynamic_survival_metric")) {
    eval_time <- eval_time[1]
  } else {
    eval_time <- NULL
  }
  eval_time
}

predict_wrapper <- function(model, new_data, type, eval_time, subgrid = NULL) {
  if (is.null(subgrid)) {
    fn <- "predict.model_fit"
  } else {
    fn <- "multi_predict"
  }

  cl <-
    rlang::call2(
      fn,
      .ns = "parsnip",
      object = rlang::expr(model),
      new_data = rlang::expr(new_data),
      type = type
    )

  # Add in censored regression evaluation times (if needed)
  has_type <- type %in% dyn_surv_types
  if (
    model$spec$mode == "censored regression" & !is.null(eval_time) & has_type
  ) {
    cl <- rlang::call_modify(cl, eval_time = eval_time)
  }

  # When there are sub-models:
  if (!is.null(subgrid)) {
    cl <- rlang::call_modify(cl, !!!subgrid)
  }
  res <- rlang::eval_tidy(cl)

  res
}

# ------------------------------------------------------------------------------

#' @export
#' @rdname tune-internal-functions
forge_from_workflow <- function(new_data, workflow) {
  blueprint <- workflow$pre$mold$blueprint
  forged <- hardhat::forge(new_data, blueprint, outcomes = TRUE)

  forged
}

get_metrics_by <- function(metric_set) {
  metrics <- attr(metric_set, "metrics")
  metrics_by <- purrr::map(metrics, attr, "by")
  unique(unlist(metrics_by, use.names = FALSE))
}

# metrics_by is the output of `get_metrics_by()`---it's assumed that wherever
# `has_metrics_by()` is needed, `get_metrics_by()` output will be needed too.
has_metrics_by <- function(metrics_by) {
  length(metrics_by) > 0
}

# ------------------------------------------------------------------------------

#' @export
#' @rdname tune-internal-functions
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `finalize_workflow_preprocessor()` is deprecated.
finalize_workflow_preprocessor <- function(workflow, grid_preprocessor) {
  lifecycle::deprecate_warn("2.1.0", "finalize_workflow_preprocessor()")

  # Already finalized, nothing to tune
  if (ncol(grid_preprocessor) == 0L) {
    return(workflow)
  }

  recipe <- extract_preprocessor(workflow)
  recipe <- merge(recipe, grid_preprocessor)$x[[1]]

  workflow <- .set_workflow_recipe(workflow, recipe)

  workflow
}

# ------------------------------------------------------------------------------

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.has_preprocessor <- function(workflow) {
  .has_preprocessor_recipe(workflow) ||
    .has_preprocessor_formula(workflow) ||
    .has_preprocessor_variables(workflow)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.has_preprocessor_recipe <- function(workflow) {
  "recipe" %in% names(workflow$pre$actions)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.has_preprocessor_formula <- function(workflow) {
  "formula" %in% names(workflow$pre$actions)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.has_preprocessor_variables <- function(workflow) {
  "variables" %in% names(workflow$pre$actions)
}

has_postprocessor <- function(workflow) {
  "tailor" %in% names(workflow$post$actions)
}

has_case_weights <- function(workflow) {
  "case_weights" %in% names(workflow$pre$actions)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.has_spec <- function(workflow) {
  "model" %in% names(workflow$fit$actions)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.set_workflow_spec <- function(workflow, spec) {
  workflow$fit$actions$model$spec <- spec
  workflow
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.set_workflow_recipe <- function(workflow, recipe) {
  workflow$pre$actions$recipe$recipe <- recipe
  workflow
}

set_workflow_tailor <- function(workflow, tailor) {
  workflow$post$actions$tailor$tailor <- tailor
  workflow
}

test_that("evaluation time", {
  skip_if_not_installed("censored")
  skip_if_not_installed("yardstick", minimum_version = "1.1.0.9000")

  suppressPackageStartupMessages(library(tune))
  suppressPackageStartupMessages(library(censored))
  suppressPackageStartupMessages(library(yardstick))
  suppressPackageStartupMessages(library(rsample))

  spec <- survival_reg()
  set.seed(1)
  rs <- vfold_cv(stanford2, strata = status)
  .time <- seq(1, 1000, length = 5)
  mtr <- metric_set(brier_survival)
  reg_mtr <- metric_set(rmse)

  expect_snapshot(
    error = TRUE,
    spec |> tune_grid(Surv(time, status) ~ ., resamples = rs, metrics = mtr)
  )
  expect_snapshot(
    error = TRUE,
    spec |>
      tune_grid(Surv(time, status) ~ ., resamples = rs, metrics = reg_mtr)
  )
  expect_snapshot(
    linear_reg() |>
      tune_grid(age ~ ., resamples = rs, metrics = reg_mtr, eval_time = 1)
  )

  expect_snapshot(
    error = TRUE,
    no_usable_times <-
      spec |>
      tune_grid(
        Surv(time, status) ~ .,
        resamples = rs,
        metrics = mtr,
        eval_time = c(-1, Inf)
      )
  )

  times <- 4:1
  expect_equal(get_metric_time(metric_set(brier_survival), times), 4)
  expect_equal(get_metric_time(metric_set(concordance_survival), times), NULL)
  expect_equal(
    get_metric_time(metric_set(brier_survival_integrated), times),
    NULL
  )
  expect_equal(
    get_metric_time(
      metric_set(
        brier_survival,
        brier_survival_integrated,
        concordance_survival
      ),
      times
    ),
    4
  )
})

library(testthat)
library(tune)

# The submission to CRAN has been rejected several times because the _total_
# time to check the package is too long (>10 min). Most of this is attributable
# to the install time for dependencies. However, CRAN has been very inconsistent
# about this with one submission stating that the rule is that the check time
# should be less than 10 min while in another that the total check time,
# _aggregated over all architectures_, should be under 10 min (e.g. adding the
# windows check times for 32 and 64 bit to the linux test times and so on).
# Since they have been undependable on this, we'll avoid running all tests on
# CRAN.

if (identical(Sys.getenv("NOT_CRAN"), "true")) {
  # emulates `testthat:::on_cran()`
  if (requireNamespace("xml2")) {
    test_check(
      "tune",
      reporter = MultiReporter$new(
        reporters = list(
          JunitReporter$new(file = "test-results.xml"),
          CheckReporter$new()
        )
      )
    )
  } else {
    test_check("tune")
  }
}

#' @importFrom stats predict

#' @noRd
#' @method predict tune_results
#' @export
predict.tune_results <- function(object, ...) {
  cli::cli_abort(c(
    "`predict()` is not well-defined for tuning results.",
    "i" = "To predict with the optimal model configuration from tuning \\
           results, ensure that the tuning result was generated with the \\
           {.help [control option](tune::control_grid)} \\
           {.code save_workflow = TRUE}, run \\
           {.help [{.fun fit_best}](tune::fit_best)}, and then predict using \\
           {.help [{.fun predict}](workflows::predict.workflow)} on its output.",
    "i" = "To collect predictions from tuning results, ensure that \\
           the tuning result was generated with the \\
           {.help [control option](tune::control_grid)} \\
           {.code save_pred = TRUE} and run \\
           {.help [{.fun collect_predictions}](tune::collect_predictions)}."
  ))
}

#' @noRd
#' @method predict last_fit
#' @export
predict.last_fit <- function(object, ...) {
  cli::cli_abort(c(
    "`predict()` is not well-defined for {.code last_fit()} results.",
    "i" = "To predict with the model fitted in {.code last_fit()}, first run \\
           {.help [{.fun extract_workflow}](tune::extract_workflow.last_fit)} \\
           and then predict using \\
           {.help [{.fun predict}](workflows::predict.workflow)} on its output."
  ))
}

#' Schedule a grid
#'
#' @param grid A tibble containing the parameter grid.
#' @param wflow The workflow object for which we schedule the grid.
#'
#' @returns A schedule object, inheriting from either 'single_schedule',
#' 'grid_schedule', or 'resample_schedule'.
#'
#' @keywords internal
#' @export
schedule_grid <- function(grid, wflow) {
  if (!tibble::is_tibble(grid)) {
    cli::cli_abort(
      "Argument {.arg grid} must be a tibble, not {.obj_type_friendly {grid}}."
    )
  }
  if (!inherits(wflow, "workflow")) {
    cli::cli_abort(
      "Argument {.arg wflow} must be a workflow object, not {.obj_type_friendly {wflow}}."
    )
  }

  schedule <- schedule_stages(grid, wflow)

  param_info <- tune_args(wflow)

  og_cls <- class(schedule)
  if (nrow(param_info) == 0) {
    cls <- "resample_schedule"
  } else {
    cls <- "grid_schedule"
  }

  if (nrow(grid) == 1) {
    cls <- c("single_schedule", cls)
  }

  class(schedule) <- c(cls, "schedule", og_cls)

  schedule
}

schedule_stages <- function(grid, wflow) {
  # Which parameter belongs to which stage and which is a submodel parameter?
  param_info <- get_param_info(wflow)

  model_source <- get_model_source(wflow)

  # schedule preprocessing stage and push the rest into a nested tibble
  param_pre_stage <- param_info |>
    dplyr::filter(source == "recipe") |>
    dplyr::pull(id)
  schedule <- grid |>
    tidyr::nest(.by = dplyr::all_of(param_pre_stage), .key = "model_stage")

  # schedule next stages nested within `schedule_model_stage_i()`
  schedule |>
    dplyr::mutate(
      model_stage = purrr::map(
        model_stage,
        schedule_model_stage_i,
        param_info = param_info,
        wflow = wflow,
        model_source = model_source
      )
    )
}

schedule_model_stage_i <- function(
  model_stage,
  param_info,
  wflow,
  model_source = "model_spec"
) {
  model_param <- param_info |>
    dplyr::filter(source == model_source) |>
    dplyr::pull(id)
  non_submodel_param <- param_info |>
    dplyr::filter(source == model_source & !has_submodel) |>
    dplyr::pull(id)

  any_non_submodel_param <- length(non_submodel_param) > 0

  # schedule model parameters
  schedule <- min_model_grid(model_stage, model_param, wflow)

  # push remaining parameters into the next stage
  next_stage <- model_stage |>
    tidyr::nest(
      .by = dplyr::all_of(non_submodel_param),
      .key = "predict_stage"
    )

  if (any_non_submodel_param) {
    # min_model_grid() may change the row order, thus use next_stage as the
    # "left" data frame here to preserve the original row order
    schedule <- next_stage |>
      dplyr::left_join(schedule, by = non_submodel_param) |>
      dplyr::relocate(dplyr::all_of(model_param))
  } else {
    schedule <- dplyr::bind_cols(schedule, next_stage)
  }

  # schedule next stages nested within `schedule_predict_stage_i()`
  schedule |>
    dplyr::mutate(
      predict_stage = purrr::map(
        predict_stage,
        schedule_predict_stage_i,
        param_info = param_info
      )
    )
}

min_model_grid <- function(grid, model_param, wflow) {
  # work on only the model parameters
  model_grid <- grid |>
    dplyr::select(dplyr::all_of(model_param)) |>
    dplyr::distinct()

  if (nrow(model_grid) < 1) {
    return(model_grid)
  }

  # But we check to see if there are _any_ model parameters. If not,
  # act like `min_grid.model_spec()` and just add a placeholder for submodels.
  # We want to avoid method dispatch to a more specific method (e.g.
  # `min_grid.logistic_reg()`) because it assumes that _there is_ a model
  # parameter and, at least for glmnet models, it fails. See issue #1119
  if (ncol(model_grid) == 0) {
    res <-
      min_grid.model_spec(extract_spec_parsnip(wflow), model_grid) |>
      dplyr::select(dplyr::all_of(model_param))
  } else {
    res <-
      min_grid(extract_spec_parsnip(wflow), model_grid) |>
      dplyr::select(dplyr::all_of(model_param))
  }
  res
}

schedule_predict_stage_i <- function(predict_stage, param_info) {
  submodel_param <- param_info |>
    dplyr::filter(source == "model_spec" & has_submodel) |>
    dplyr::pull(id)

  predict_stage |>
    tidyr::nest(
      .by = dplyr::all_of(submodel_param),
      .key = "post_stage"
    )
}

get_param_info <- function(wflow) {
  param_info <- tune_args(wflow) |>
    dplyr::select(name, id, source)

  model_spec <- extract_spec_parsnip(wflow)
  model_type <- class(model_spec)[1]
  model_eng <- model_spec$engine

  # NULL for tidyclust workflows since tidyclust models are not in parsnip's
  # registry
  model_param <- parsnip::get_from_env(paste0(model_type, "_args"))

  if (!is.null(model_param)) {
    model_param <- model_param |>
      dplyr::filter(engine == model_spec$engine) |>
      dplyr::select(name = parsnip, has_submodel)
    param_info <- dplyr::left_join(param_info, model_param, by = "name")
  } else {
    # tidyclust models never use submodels
    param_info$has_submodel <- FALSE
  }

  if (!"has_submodel" %in% names(param_info)) {
    param_info$has_submodel <- FALSE
  } else {
    param_info$has_submodel[is.na(param_info$has_submodel)] <- FALSE
  }

  param_info
}

# Determine the source type for model parameters
# Returns "model_spec" for parsnip models, "cluster_spec" for tidyclust models
get_model_source <- function(wflow) {
  spec <- extract_spec_parsnip(wflow)
  if (inherits(spec, "cluster_spec")) {
    "cluster_spec"
  } else {
    "model_spec"
  }
}

# `get_param_info()` -----------------------------------------------------

test_that("`get_param_info()` works for a workflow without tags for tuning", {
  wflow <- workflow(mpg ~ ., parsnip::linear_reg())
  param_info <- get_param_info(wflow)

  expect_named(param_info, c("name", "id", "source", "has_submodel"))
  expect_identical(nrow(param_info), 0L)
})

test_that("`get_param_info()` works for a workflow with tags for tuning", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("probably")

  # tuning tags in all components
  wflow <- workflow(rec_tune, mod_tune_no_submodel, tlr_tune)

  param_info <- get_param_info(wflow)

  expect_named(param_info, c("name", "id", "source", "has_submodel"))
  expect_identical(
    param_info$id,
    c("min_n", "threshold", "disp_df", "lower_limit")
  )
  expect_identical(param_info$has_submodel, c(FALSE, FALSE, FALSE, FALSE))
})

test_that("`get_param_info()` works when there are submodel parameters", {
  skip_if_not_installed("probably")

  # tuning tags only in model spec
  rec_no_steps <- recipes::recipe(mpg ~ ., data = mtcars)

  wflow <- workflow(rec_no_steps, mod_tune, tlr_no_tune)

  param_info <- get_param_info(wflow)

  expect_named(param_info, c("name", "id", "source", "has_submodel"))
  expect_identical(param_info$name, c("trees", "min_n"))
  expect_identical(param_info$has_submodel, c(TRUE, FALSE))
})


# `schedule_predict_stage_i()` -------------------------------------------

test_that("`schedule_predict_stage_i()` works with: no submodel, no post-processing", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_no_tune, tlr_no_tune)
  param_info <- get_param_info(wflow)
  grid_predict_stage <- tibble::tibble()

  schedule <- schedule_predict_stage_i(grid_predict_stage, param_info)
  expect_named(schedule, c("post_stage"))
  expect_identical(nrow(schedule), 0L)
})

test_that("`schedule_predict_stage_i()` works with: no submodel, with post-processing", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_no_tune, tlr_tune)
  param_info <- get_param_info(wflow)
  grid_predict_stage <- tibble::tibble(lower_limit = 1:2)

  schedule <- schedule_predict_stage_i(grid_predict_stage, param_info)
  expect_named(schedule, c("post_stage"))
  expect_identical(nrow(schedule), 1L)
  expect_identical(schedule$post_stage[[1]], grid_predict_stage)
})

test_that("`schedule_predict_stage_i()` works with: with submodel, no post-processing", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_tune_submodel, tlr_no_tune)
  param_info <- get_param_info(wflow)
  grid_predict_stage <- tibble::tibble(trees = 1:2)

  schedule <- schedule_predict_stage_i(grid_predict_stage, param_info)
  expect_named(schedule, c("trees", "post_stage"))
  expect_identical(schedule$trees, grid_predict_stage$trees)
  expect_identical(
    purrr::map_int(schedule$post_stage, ncol),
    c(0L, 0L)
  )
})

test_that("`schedule_predict_stage_i()` works with: with submodel, with post-processing", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_tune_submodel, tlr_tune)
  param_info <- get_param_info(wflow)
  # semi-regular grid
  grid_predict_stage <- list(
    tibble::tibble(trees = 1L, lower_limit = 1L),
    tibble::tibble(trees = c(2L, 2L), lower_limit = 1:2)
  ) |>
    purrr::list_rbind()

  schedule <- schedule_predict_stage_i(grid_predict_stage, param_info)
  expect_named(schedule, c("trees", "post_stage"))
  expect_identical(schedule$trees, 1:2)
  expect_identical(
    schedule$post_stage[[1]],
    tibble::tibble(lower_limit = 1L)
  )
  expect_identical(
    schedule$post_stage[[2]],
    tibble::tibble(lower_limit = 1:2)
  )
})


# `schedule_model_stage_i()` ---------------------------------------------

test_that("`schedule_model_stage_i()` works with: no tuning at all", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_no_tune, tlr_no_tune)
  param_info <- get_param_info(wflow)
  grid_model_stage <- tibble::tibble()

  schedule <- schedule_model_stage_i(grid_model_stage, param_info, wflow)
  expect_named(schedule, c("predict_stage"))
  expect_identical(nrow(schedule), 0L)
})

test_that("`schedule_model_stage_i()` works with only non-submodel: with non-submodel, no submodel, no post", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_tune_no_submodel, tlr_no_tune)
  param_info <- get_param_info(wflow)
  grid_model_stage <- tibble::tibble(min_n = 1:2)

  schedule <- schedule_model_stage_i(grid_model_stage, param_info, wflow)
  expect_named(schedule, c("min_n", "predict_stage"))
  expect_identical(nrow(schedule), 2L)
  expect_identical(schedule$min_n, 1:2)
  expect_identical(
    purrr::map_chr(schedule$predict_stage, \(x) class(x)[1]),
    c("tbl_df", "tbl_df")
  )
  expect_identical(
    purrr::map_chr(schedule$predict_stage, names),
    c("post_stage", "post_stage")
  )
})

test_that("`schedule_model_stage_i()` works with only submodel: no non-submodel, with submodel, no post", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_tune_submodel, tlr_no_tune)
  param_info <- get_param_info(wflow)
  grid_model_stage <- tibble::tibble(trees = 1:2)

  schedule <- schedule_model_stage_i(grid_model_stage, param_info, wflow)
  expect_named(schedule, c("trees", "predict_stage"))
  expect_identical(nrow(schedule), 1L)
  expect_identical(schedule$trees, 2L)
  expect_identical(
    purrr::map_chr(schedule$predict_stage, \(x) class(x)[1]),
    c("tbl_df")
  )
  expect_identical(
    purrr::map(schedule$predict_stage, names),
    list(c("trees", "post_stage"))
  )
  expect_identical(
    schedule$predict_stage[[1]] |> pull("trees"),
    1:2
  )
})

test_that("`schedule_model_stage_i()` works with only post: no non-submodel, no submodel, with post", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_no_tune, tlr_tune)
  param_info <- get_param_info(wflow)
  grid_model_stage <- tibble::tibble(lower_limit = 1:2)

  schedule <- schedule_model_stage_i(grid_model_stage, param_info, wflow)
  expect_named(schedule, c("predict_stage"))
  expect_identical(nrow(schedule), 1L)
  expect_identical(
    purrr::map_chr(schedule$predict_stage, \(x) class(x)[1]),
    c("tbl_df")
  )
  expect_identical(
    purrr::map(schedule$predict_stage, names),
    list(c("post_stage"))
  )
  expect_identical(
    schedule$predict_stage[[1]] |> pull("post_stage"),
    list(tibble::tibble(lower_limit = 1:2))
  )
})

test_that("`schedule_model_stage_i()` works with both model types only: with non-submodel, with submodel, no post", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_tune, tlr_no_tune)
  param_info <- get_param_info(wflow)
  # irregular grid
  grid_model_stage <- list(
    tibble::tibble(trees = 1:2, min_n = 1L),
    tibble::tibble(trees = 2L, min_n = 2L),
    tibble::tibble(trees = 2:3, min_n = 3L)
  ) |>
    purrr::list_rbind()

  schedule <- schedule_model_stage_i(grid_model_stage, param_info, wflow)

  expect_named(schedule, c("trees", "min_n", "predict_stage"))
  expect_identical(nrow(schedule), 3L)
  expect_identical(schedule$trees, c(2L, 2L, 3L))
  expect_identical(schedule$min_n, 1:3)
  expect_identical(
    purrr::map_chr(schedule$predict_stage, \(x) class(x)[1]) |> unique(),
    "tbl_df"
  )
  expect_identical(
    purrr::map(schedule$predict_stage, names),
    list(
      c("trees", "post_stage"),
      c("trees", "post_stage"),
      c("trees", "post_stage")
    )
  )
  expect_identical(
    schedule |>
      filter(min_n == 1) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("trees"),
    1:2
  )
  expect_identical(
    schedule |>
      filter(min_n == 2) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("trees"),
    2L
  )
  expect_identical(
    schedule |>
      filter(min_n == 3) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("trees"),
    2:3
  )
})

test_that("`schedule_model_stage_i()` works without submodel: with non-submodel, no submodel, with post", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_tune_no_submodel, tlr_tune)
  param_info <- get_param_info(wflow)
  # semi-regular grid
  grid_model_stage <- list(
    tibble::tibble(min_n = 1L, lower_limit = 1:2),
    tibble::tibble(min_n = 2L, lower_limit = 1:3),
    tibble::tibble(min_n = 3L, lower_limit = 3L)
  ) |>
    purrr::list_rbind()

  schedule <- schedule_model_stage_i(grid_model_stage, param_info, wflow)
  expect_named(schedule, c("min_n", "predict_stage"))
  expect_identical(nrow(schedule), 3L)
  expect_identical(schedule$min_n, 1:3)
  expect_identical(
    purrr::map_chr(schedule$predict_stage, \(x) class(x)[1]) |> unique(),
    "tbl_df"
  )
  expect_identical(
    purrr::map(schedule$predict_stage, names) |>
      purrr::list_c() |>
      unique(),
    "post_stage"
  )
  expect_identical(
    schedule |>
      filter(min_n == 1) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("post_stage") |>
      purrr::pluck(1) |>
      pull("lower_limit"),
    1:2
  )
  expect_identical(
    schedule |>
      filter(min_n == 2) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("post_stage") |>
      purrr::pluck(1) |>
      pull("lower_limit"),
    1:3
  )
  expect_identical(
    schedule |>
      filter(min_n == 3) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("post_stage") |>
      purrr::pluck(1) |>
      pull("lower_limit"),
    3L
  )
})

test_that("`schedule_model_stage_i()` works everything: with non-submodel, with submodel, with post", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_tune, tlr_tune)
  param_info <- get_param_info(wflow)
  # semi-regular grid
  grid_model_stage <- list(
    tibble::tibble(
      min_n = 1L,
      trees = c(1L, 1L, 2L, 2L),
      lower_limit = c(1:2, 1:2)
    ),
    tibble::tibble(
      min_n = 2L,
      trees = c(1L, rep(2L, 3)),
      lower_limit = c(1L, 1:3)
    ),
    # another row to be combined under min_n = 1
    tibble::tibble(min_n = 1L, trees = 3L, lower_limit = 4L)
  ) |>
    purrr::list_rbind()

  schedule <- schedule_model_stage_i(grid_model_stage, param_info, wflow)
  expect_named(schedule, c("trees", "min_n", "predict_stage"))
  expect_identical(nrow(schedule), 2L)
  expect_identical(schedule$min_n, 1:2)
  expect_identical(
    purrr::map_chr(schedule$predict_stage, \(x) class(x)[1]) |> unique(),
    "tbl_df"
  )
  expect_identical(
    purrr::map(schedule$predict_stage, names) |>
      purrr::list_c() |>
      unique(),
    c("trees", "post_stage")
  )
  expect_identical(
    schedule |>
      filter(min_n == 1) |>
      select(-predict_stage),
    tibble::tibble(trees = 3L, min_n = 1L)
  )
  expect_identical(
    schedule |>
      filter(min_n == 1) |>
      pull(predict_stage) |>
      purrr::pluck(1) |>
      pull(trees),
    1:3
  )
  expect_identical(
    schedule |>
      filter(min_n == 1) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("post_stage") |>
      purrr::list_rbind() |>
      pull("lower_limit"),
    c(1:2, 1:2, 4L)
  )
  expect_identical(
    schedule |>
      filter(min_n == 2) |>
      select(-predict_stage),
    tibble::tibble(trees = 2L, min_n = 2L)
  )
  expect_identical(
    schedule |>
      filter(min_n == 2) |>
      pull(predict_stage) |>
      purrr::pluck(1) |>
      pull(trees),
    1:2
  )
  expect_identical(
    schedule |>
      filter(min_n == 2) |>
      pull("predict_stage") |>
      purrr::pluck(1) |>
      pull("post_stage") |>
      purrr::list_rbind() |>
      pull("lower_limit"),
    c(1L, 1:3)
  )
})


# `schedule_stages()` ----------------------------------------------------

test_that("`schedule_stages()` works without preprocessing", {
  skip_if_not_installed("probably")

  wflow <- workflow(mpg ~ ., mod_no_tune, tlr_no_tune)
  grid <- tibble::tibble()

  schedule <- schedule_stages(grid, wflow)
  expect_named(schedule, c("model_stage"))
  expect_identical(nrow(schedule), 0L)
})

test_that("`schedule_stages()` works with preprocessing", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("probably")

  wflow <- workflow(rec_tune, mod_no_tune, tlr_no_tune)
  grid <- tibble::tibble(
    threshold = rep(1:2, each = 2),
    disp_df = c(1:2, 1:2)
  )

  schedule <- schedule_stages(grid, wflow)
  expect_named(schedule, c("threshold", "disp_df", "model_stage"))
  expect_identical(nrow(schedule), 4L)
  expect_identical(
    schedule |> dplyr::select(-model_stage),
    grid
  )
})

# `schedule_grid()` ------------------------------------------------------

# No tuning or postprocesing estimation

test_that("grid processing schedule - no parameters", {
  wflow_nada <- workflow(outcome ~ ., parsnip::logistic_reg())
  grid_nada <- tibble::tibble()

  sched_nada <- schedule_grid(grid_nada, wflow_nada)

  expect_named(sched_nada, "model_stage")
  expect_equal(nrow(sched_nada), 0)

  expect_s3_class(
    sched_nada,
    c("resample_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - recipe and model", {
  skip_if_not_installed("splines2")

  wflow_pre_only <- workflow(rec_no_tune, parsnip::logistic_reg())
  grid_pre_only <- tibble::tibble()
  sched_pre_only <- schedule_grid(grid_pre_only, wflow_pre_only)

  expect_named(sched_pre_only, c("model_stage"))
  expect_equal(nrow(sched_pre_only), 0)

  expect_s3_class(
    sched_pre_only,
    c("resample_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - recipe, model, and post", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("probably")

  wflow_three <- workflow(rec_no_tune, parsnip::linear_reg(), tlr_no_tune)
  grid_three <- tibble::tibble()
  sched_three <- schedule_grid(grid_three, wflow_three)

  expect_named(sched_three, c("model_stage"))
  expect_equal(nrow(sched_three), 0)

  expect_s3_class(
    sched_three,
    c("resample_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

# Tuning, no postprocesing estimation

test_that("grid processing schedule - recipe only", {
  skip_if_not_installed("splines2")

  wflow_pre_only <- workflow(rec_tune, parsnip::logistic_reg())
  grid_pre_only <-
    extract_parameter_set_dials(wflow_pre_only) |>
    dials::grid_regular(levels = 3) |>
    arrange(threshold, disp_df)
  sched_pre_only <-
    schedule_grid(grid_pre_only, wflow_pre_only)

  expect_named(sched_pre_only, c("threshold", "disp_df", "model_stage"))
  expect_equal(nrow(sched_pre_only), nrow(grid_pre_only))

  # All of the other nested tibbles should be empty
  expect_equal(
    sched_pre_only |>
      tidyr::unnest(model_stage) |>
      tidyr::unnest(predict_stage) |>
      tidyr::unnest(post_stage),
    grid_pre_only
  )

  expect_s3_class(
    sched_pre_only,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - model only, no submodels", {
  wflow_rf_only <- workflow(outcome ~ ., mod_tune_no_submodel)
  grid_rf_only <-
    extract_parameter_set_dials(wflow_rf_only) |>
    dials::grid_regular(levels = 3)
  sched_rf_only <-
    schedule_grid(grid_rf_only, wflow_rf_only)

  expect_named(sched_rf_only, c("model_stage"))
  expect_equal(nrow(sched_rf_only), 1L)

  rf_n <- length(sched_rf_only$model_stage)
  for (i in 1:rf_n) {
    # No real need for the loop here
    expect_named(sched_rf_only$model_stage[[i]], c("min_n", "predict_stage"))
    expect_equal(
      sched_rf_only$model_stage[[i]] |>
        tidyr::unnest(predict_stage) |>
        tidyr::unnest(post_stage),
      grid_rf_only
    )
  }

  expect_s3_class(
    sched_rf_only,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - model only, submodels, regular grid", {
  wflow_bst <- workflow(outcome ~ ., mod_tune)
  grid_bst <- extract_parameter_set_dials(wflow_bst) |>
    dials::grid_regular(levels = 3)

  min_n_only <- grid_bst |> dplyr::distinct(min_n) |> dplyr::arrange(min_n)
  trees_only <- grid_bst |> dplyr::distinct(trees) |> dplyr::arrange(trees)

  # regular grid
  sched_bst <- schedule_grid(grid_bst, wflow_bst)

  expect_named(sched_bst, c("model_stage"))
  expect_equal(nrow(sched_bst), 1L)

  reg_n <- length(sched_bst$model_stage)
  for (i in 1:reg_n) {
    expect_named(
      sched_bst$model_stage[[i]],
      c("trees", "min_n", "predict_stage")
    )

    expect_equal(
      sched_bst$model_stage[[i]] |>
        dplyr::select(-trees, -predict_stage),
      min_n_only
    )

    for (j in seq_along(sched_bst$model_stage[[i]]$predict_stage)) {
      expect_named(
        sched_bst$model_stage[[i]]$predict_stage[[j]],
        c("trees", "post_stage")
      )
      expect_equal(
        sched_bst$model_stage[[i]]$predict_stage[[j]] |>
          dplyr::select(trees),
        trees_only
      )
    }

    expect_equal(
      sched_bst$model_stage[[i]] |>
        dplyr::select(-trees) |>
        tidyr::unnest(predict_stage) |>
        tidyr::unnest(post_stage) |>
        dplyr::select(trees, min_n),
      grid_bst
    )
  }

  expect_s3_class(
    sched_bst,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - model only, submodels, SFD grid", {
  wflow_bst <- workflow(outcome ~ ., mod_tune)
  grid_sfd_bst <- extract_parameter_set_dials(wflow_bst) |>
    dials::grid_space_filling(size = 5, type = "uniform")
  sched_sfd_bst <- schedule_grid(grid_sfd_bst, wflow_bst)

  expect_named(sched_sfd_bst, c("model_stage"))
  expect_equal(nrow(sched_sfd_bst), 1L)

  irreg_n <- length(sched_sfd_bst$model_stage)
  expect_equal(irreg_n, 1L)

  expect_named(
    sched_sfd_bst$model_stage[[1]],
    c("trees", "min_n", "predict_stage")
  )
  expect_equal(
    sched_sfd_bst$model_stage[[1]] |>
      dplyr::select(-predict_stage) |>
      dplyr::select(trees, min_n) |>
      dplyr::arrange(trees, min_n),
    grid_sfd_bst |>
      dplyr::select(trees, min_n) |>
      dplyr::arrange(trees, min_n)
  )

  expect_equal(
    sched_sfd_bst$model_stage[[1]] |>
      dplyr::select(-trees) |>
      tidyr::unnest(predict_stage) |>
      tidyr::unnest(post_stage) |>
      dplyr::select(trees, min_n) |>
      dplyr::arrange(trees, min_n),
    grid_sfd_bst |>
      dplyr::select(trees, min_n) |>
      dplyr::arrange(trees, min_n)
  )

  expect_s3_class(
    sched_sfd_bst,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - model only, submodels, irregular design", {
  wflow_bst <- workflow(outcome ~ ., mod_tune)
  grid_odd_bst <- tibble::tibble(
    min_n = c(1, 1, 2, 3, 4, 5),
    trees = rep(1:2, 3)
  )
  sched_odd_bst <- schedule_grid(grid_odd_bst, wflow_bst)

  expect_named(sched_odd_bst, c("model_stage"))
  expect_equal(nrow(sched_odd_bst), 1L)

  odd_n <- length(sched_odd_bst$model_stage)
  expect_equal(odd_n, 1L)

  expect_named(
    sched_odd_bst$model_stage[[1]],
    c("trees", "min_n", "predict_stage")
  )
  expect_equal(
    sched_odd_bst$model_stage[[1]] |>
      dplyr::select(-predict_stage) |>
      dplyr::select(trees, min_n),
    tibble::tibble(trees = c(2, 1, 2, 1, 2), min_n = c(1, 2, 3, 4, 5))
  )

  for (i in 1:nrow(sched_odd_bst$model_stage[[1]])) {
    prd <- sched_odd_bst$model_stage[[1]]$predict_stage[[i]]
    if (sched_odd_bst$model_stage[[1]]$min_n[i] == 1) {
      expect_equal(nrow(prd), 2L)
    } else {
      expect_equal(nrow(prd), 1L)
    }
    expect_true(
      all(purrr::map_int(prd$post_stage, nrow) == 1)
    )
  }

  expect_s3_class(
    sched_odd_bst,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - model only, submodels, 1 point design", {
  wflow_bst <- workflow(outcome ~ ., mod_tune)

  set.seed(1)
  grid_1_pt <- extract_parameter_set_dials(wflow_bst) |>
    dials::grid_random(size = 1)
  sched_1_pt <- schedule_grid(grid_1_pt, wflow_bst)

  expect_named(sched_1_pt, c("model_stage"))
  expect_equal(nrow(sched_1_pt), 1L)
  expect_equal(length(sched_1_pt$model_stage), 1L)
  expect_named(
    sched_1_pt$model_stage[[1]],
    c("trees", "min_n", "predict_stage")
  )

  expect_equal(
    length(sched_1_pt$model_stage[[1]]$predict_stage),
    1L
  )
  expect_named(
    sched_1_pt$model_stage[[1]]$predict_stage[[1]],
    c("trees", "post_stage")
  )

  expect_equal(
    length(sched_1_pt$model_stage[[1]]$predict_stage[[1]]$post_stage),
    1L
  )
  expect_equal(
    dim(sched_1_pt$model_stage[[1]]$predict_stage[[1]]$post_stage[[1]]),
    1:0
  )

  expect_s3_class(
    sched_1_pt,
    c(
      "single_schedule",
      "grid_schedule",
      "schedule",
      "tbl_df",
      "tbl",
      "data.frame"
    )
  )
})

test_that("grid processing schedule - postprocessing only", {
  skip_if_not_installed("probably")

  wflow_thrsh <- workflow(outcome ~ ., parsnip::linear_reg(), tlr_tune)
  grid_thrsh <- extract_parameter_set_dials(wflow_thrsh) |>
    update(lower_limit = dials::lower_limit(c(0, 1))) |>
    dials::grid_regular(levels = 3)

  sched_thrsh <- schedule_grid(grid_thrsh, wflow_thrsh)

  expect_named(sched_thrsh, c("model_stage"))
  expect_equal(nrow(sched_thrsh), 1L)

  expect_named(sched_thrsh$model_stage[[1]], c("predict_stage"))
  expect_equal(nrow(sched_thrsh$model_stage[[1]]), 1L)

  expect_named(
    sched_thrsh$model_stage[[1]]$predict_stage[[1]],
    c("post_stage")
  )
  expect_equal(nrow(sched_thrsh$model_stage[[1]]), 1L)

  expect_equal(
    sched_thrsh$model_stage[[1]]$predict_stage[[1]]$post_stage[[1]],
    grid_thrsh
  )

  expect_s3_class(
    sched_thrsh,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - recipe + postprocessing, regular grid", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("probably")

  wflow_pre_post <- workflow(rec_tune, parsnip::linear_reg(), tlr_tune)
  grid_pre_post <-
    extract_parameter_set_dials(wflow_pre_post) |>
    update(lower_limit = dials::lower_limit(c(0, 1))) |>
    dials::grid_regular(levels = 3)

  grid_pre <-
    grid_pre_post |>
    distinct(threshold, disp_df)
  grid_post <-
    grid_pre_post |>
    distinct(lower_limit) |>
    arrange(lower_limit)

  sched_pre_post <- schedule_grid(grid_pre_post, wflow_pre_post)

  expect_named(sched_pre_post, c("threshold", "disp_df", "model_stage"))
  expect_equal(
    sched_pre_post |> select(-model_stage) |> tibble::as_tibble(),
    grid_pre
  )

  for (i in seq_along(sched_pre_post$model_stage)) {
    expect_named(sched_pre_post$model_stage[[i]], c("predict_stage"))
    expect_equal(nrow(sched_pre_post$model_stage[[i]]), 1L)
  }

  for (i in seq_along(sched_pre_post$model_stage)) {
    expect_named(
      sched_pre_post$model_stage[[i]]$predict_stage[[1]],
      c("post_stage")
    )
    expect_identical(
      sched_pre_post$model_stage[[i]]$predict_stage[[1]]$post_stage[[1]] |>
        arrange(lower_limit),
      grid_post
    )
  }

  expect_s3_class(
    sched_pre_post,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - recipe + postprocessing, irregular grid", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("probably")

  wflow_pre_post <- workflow(rec_tune, parsnip::linear_reg(), tlr_tune)
  grid_pre_post <-
    extract_parameter_set_dials(wflow_pre_post) |>
    update(lower_limit = dials::lower_limit(c(0, 1))) |>
    dials::grid_regular() |>
    dplyr::slice(-c(1, 14))

  grid_pre <-
    grid_pre_post |>
    distinct(threshold, disp_df)

  grids_post <-
    grid_pre_post |>
    dplyr::group_nest(threshold, disp_df) |>
    mutate(data = purrr::map(data, \(.x) arrange(.x, lower_limit)))

  sched_pre_post <- schedule_grid(grid_pre_post, wflow_pre_post)

  expect_named(sched_pre_post, c("threshold", "disp_df", "model_stage"))
  expect_equal(
    sched_pre_post |> select(-model_stage) |> tibble::as_tibble(),
    grid_pre
  )

  for (i in seq_along(sched_pre_post$model_stage)) {
    expect_named(sched_pre_post$model_stage[[i]], c("predict_stage"))
    expect_equal(nrow(sched_pre_post$model_stage[[i]]), 1L)
  }

  for (i in seq_along(sched_pre_post$model_stage)) {
    expect_named(
      sched_pre_post$model_stage[[i]]$predict_stage[[1]],
      c("post_stage")
    )

    pre_grid_i <-
      sched_pre_post |>
      slice(i) |>
      select(threshold, disp_df)

    post_grid_i <-
      pre_grid_i |>
      inner_join(grids_post, by = dplyr::join_by(threshold, disp_df)) |>
      purrr::pluck("data") |>
      purrr::pluck(1) |>
      arrange(lower_limit)

    expect_identical(
      sched_pre_post$model_stage[[i]]$predict_stage[[1]]$post_stage[[1]] |>
        arrange(lower_limit),
      post_grid_i
    )
  }

  expect_s3_class(
    sched_pre_post,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - recipe + model, no submodels, regular grid", {
  skip_if_not_installed("splines2")

  wflow_pre_model <- workflow(rec_tune, mod_tune_no_submodel)
  grid_pre_model <-
    extract_parameter_set_dials(wflow_pre_model) |>
    dials::grid_regular()

  grid_pre <-
    grid_pre_model |>
    distinct(threshold, disp_df)

  grid_model <-
    grid_pre_model |>
    distinct(min_n) |>
    arrange(min_n)

  sched_pre_model <- schedule_grid(grid_pre_model, wflow_pre_model)

  expect_named(sched_pre_model, c("threshold", "disp_df", "model_stage"))
  expect_equal(
    sched_pre_model |> select(-model_stage) |> tibble::as_tibble(),
    grid_pre
  )

  for (i in seq_along(sched_pre_model$model_stage)) {
    expect_named(sched_pre_model$model_stage[[i]], c("min_n", "predict_stage"))
    expect_equal(
      sched_pre_model$model_stage[[i]] |> select(min_n) |> arrange(min_n),
      grid_model
    )
  }

  for (i in seq_along(sched_pre_model$model_stage)) {
    expect_named(
      sched_pre_model$model_stage[[i]]$predict_stage[[1]],
      c("post_stage")
    )

    expect_equal(
      nrow(sched_pre_model$model_stage[[i]]$predict_stage[[1]]),
      1L
    )
    expect_equal(
      nrow(sched_pre_model$model_stage[[i]]$predict_stage[[1]]$post_stage[[1]]),
      1L
    )
  }

  expect_s3_class(
    sched_pre_model,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - recipe + model, submodels, irregular grid", {
  skip_if_not_installed("splines2")

  wflow_pre_model <- workflow(rec_tune, mod_tune)
  grid_pre_model <-
    extract_parameter_set_dials(wflow_pre_model) |>
    dials::grid_regular() |>
    # This will make the submodel parameter (trees) unbalanced for some
    # combination of parameters of the other parameters.
    slice(-c(1, 2, 11))

  grid_pre <-
    grid_pre_model |>
    distinct(threshold, disp_df)

  grid_model <-
    grid_pre_model |>
    dplyr::group_nest(threshold, disp_df) |>
    mutate(
      data = purrr::map(
        data,
        \(.x) .x |> dplyr::summarize(trees = max(trees), .by = c(min_n))
      ),
      data = purrr::map(data, \(.x) .x |> arrange(min_n))
    )

  sched_pre_model <- schedule_grid(grid_pre_model, wflow_pre_model)

  expect_named(sched_pre_model, c("threshold", "disp_df", "model_stage"))
  expect_equal(
    sched_pre_model |> select(-model_stage) |> tibble::as_tibble(),
    grid_pre
  )

  for (i in seq_along(sched_pre_model$model_stage)) {
    model_i <- sched_pre_model$model_stage[[i]]
    expect_named(model_i, c("trees", "min_n", "predict_stage"))
    expect_equal(
      model_i |> select(min_n, trees) |> arrange(min_n),
      grid_model$data[[i]]
    )

    for (j in seq_along(sched_pre_model$model_stage[[i]]$predict_stage)) {
      predict_j <- model_i$predict_stage[[j]]

      # We need to figure out the trees that need predicting for the current
      # set of other parameters.

      # Get the settings that have already be resolved:
      other_ij <-
        model_i |>
        select(-predict_stage, -trees) |>
        slice(j) |>
        vctrs::vec_cbind(
          sched_pre_model |>
            select(threshold, disp_df) |>
            slice(i)
        )
      # What are the matching values from the grid?
      trees_ij <-
        grid_pre_model |>
        inner_join(other_ij, by = c("min_n", "threshold", "disp_df")) |>
        select(trees)

      expect_equal(
        predict_j |> select(trees) |> arrange(trees),
        trees_ij |> arrange(trees)
      )
    }
  }

  expect_s3_class(
    sched_pre_model,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("grid processing schedule - recipe + model + tailor, submodels, irregular grid", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("probably")

  wflow_pre_model_post <- workflow(rec_tune, mod_tune, tlr_tune)
  grid_pre_model_post <-
    extract_parameter_set_dials(wflow_pre_model_post) |>
    update(lower_limit = dials::lower_limit(c(0, 1))) |>
    dials::grid_regular() |>
    # This will make the submodel parameter (trees) unbalanced for some
    # combination of parameters of the other parameters.
    slice(seq(1, 240, by = 7))

  grid_pre <-
    grid_pre_model_post |>
    distinct(threshold, disp_df)

  grid_model <-
    grid_pre_model_post |>
    select(-lower_limit) |>
    dplyr::group_nest(threshold, disp_df) |>
    mutate(
      data = purrr::map(
        data,
        \(.x) .x |> dplyr::summarize(trees = max(trees), .by = c(min_n))
      ),
      data = purrr::map(data, \(.x) .x |> arrange(min_n))
    )

  sched_pre_model_post <- schedule_grid(
    grid_pre_model_post,
    wflow_pre_model_post
  )

  expect_named(sched_pre_model_post, c("threshold", "disp_df", "model_stage"))
  expect_equal(
    sched_pre_model_post |> select(-model_stage) |> tibble::as_tibble(),
    grid_pre
  )

  for (i in seq_along(sched_pre_model_post$model_stage)) {
    model_i <- sched_pre_model_post$model_stage[[i]]

    # Get the current set of preproc parameters to remove
    other_i <-
      sched_pre_model_post[i, ] |>
      dplyr::select(-model_stage)

    # We expect to evaluate these specific models for this set of preprocessors
    exp_i <-
      grid_pre_model_post |>
      inner_join(other_i, by = c("threshold", "disp_df")) |>
      arrange(trees, min_n, lower_limit) |>
      select(trees, min_n, lower_limit)

    # What we will evaluate:
    subgrid_i <-
      model_i |>
      select(-trees) |>
      unnest(predict_stage) |>
      unnest(post_stage) |>
      arrange(trees, min_n, lower_limit) |>
      select(trees, min_n, lower_limit)

    expect_equal(subgrid_i, exp_i)

    for (j in seq_along(sched_pre_model_post$model_stage[[i]]$predict_stage)) {
      model_ij <- model_i[j, ]
      expect_named(model_ij, c("trees", "min_n", "predict_stage"))

      predict_j <- model_ij$predict_stage[[1]]
      expect_named(predict_j, c("trees", "post_stage"))

      exp_post_grid <-
        # Condition on the current set of non-submodel or post param to see
        # what we should be evaluating:
        model_ij |>
        dplyr::select(-trees) |>
        vctrs::vec_cbind(other_i) |>
        dplyr::inner_join(
          grid_pre_model_post,
          by = c("threshold", "disp_df", "min_n")
        ) |>
        dplyr::select(trees, lower_limit) |>
        dplyr::arrange(trees, lower_limit)

      # Which as scheduled to be evaluated:
      subgrid_ij <-
        predict_j |>
        unnest(post_stage) |>
        dplyr::arrange(trees, lower_limit)

      expect_equal(subgrid_ij, exp_post_grid)
    }
  }

  expect_s3_class(
    sched_pre_model_post,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )
})

test_that("parameter information with engine, recipe, and tailor parameters", {
  mlp_spec <-
    mlp(
      hidden_units = tune(),
      penalty = tune(),
      learn_rate = tune(),
      epochs = 500,
      activation = tune()
    ) |>
    set_engine(
      "brulee",
      stop_iter = tune(),
      class_weights = tune()
    ) |>
    set_mode("classification")

  rec <- recipe(Class ~ ., data = tibble(Class = "a", x = 1)) |>
    step_pca(all_numeric_predictors(), num_comp = tune())

  tlr <- tailor() |> adjust_probability_threshold(threshold = tune())

  mlp_wflow <- workflow(rec, mlp_spec, tlr)

  mlp_grid <-
    mlp_wflow |>
    extract_parameter_set_dials() |>
    grid_space_filling(size = 4)

  mlp_info <- tune:::get_param_info(mlp_wflow)
  expect_true(all(!is.na(mlp_info$has_submodel)))
})

#' Plot tuning search results
#'
#' @param object A tibble of results from [tune_grid()] or [tune_bayes()].
#' @param type A single character value. Choices are `"marginals"` (for a plot
#'  of each predictor versus performance; see Details below), `"parameters"`
#'  (each parameter versus search iteration), or `"performance"` (performance
#'  versus iteration). The latter two choices are only used for [tune_bayes()].
#' @param metric A character vector or `NULL` for which metric to plot. By
#' default, all metrics will be shown via facets. Possible options are
#' the entries in `.metric` column of `collect_metrics(object)`.
#' @param width A number for the width of the confidence interval bars when
#' `type = "performance"`. A value of zero prevents them from being shown.
#' @param eval_time A numeric vector of time points where dynamic event time
#' metrics should be chosen (e.g. the time-dependent ROC curve, etc). The
#' values should be consistent with the values used to create `object`.
#' @param call The call to be displayed in warnings or errors.
#' @param ... For plots with a regular grid, this is passed to `format()` and is
#' applied to a parameter used to color points. Otherwise, it is not used.
#' @return A `ggplot2` object.
#' @details
#'
#' When the results of `tune_grid()` are used with `autoplot()`, it tries to
#'  determine whether a _regular grid_ was used.
#'
#' ## Regular grids
#'
#'   For regular grids with one or more numeric tuning parameters, the parameter
#'  with the most unique values is used on the x-axis. If there are categorical
#'  parameters, the first is used to color the geometries. All other parameters
#'  are used in column faceting.
#'
#'   The plot has the performance metric(s) on the y-axis. If there are multiple
#'  metrics, these are row-faceted.
#'
#'   If there are more than five tuning parameters, the "marginal effects" plots
#'  are used instead.
#'
#' ## Irregular grids
#'
#' For space-filling or random grids, a _marginal_ effect plot is created. A
#'  panel is made for each numeric parameter so that each parameter is on the
#'  x-axis and performance is on the y-xis. If there are multiple metrics, these
#'  are row-faceted.
#'
#' A single categorical parameter is shown as colors. If there are two or more
#'  non-numeric parameters, an error is given. A similar result occurs is only
#'  non-numeric parameters are in the grid. In these cases, we suggest using
#'  `collect_metrics()` and `ggplot()` to create a plot that is appropriate for
#'  the data.
#'
#' If a parameter has an associated transformation associated with it (as
#' determined by the parameter object used to create it), the plot shows the
#' values in the transformed units (and is labeled with the transformation type).
#'
#' Parameters are labeled using the labels found in the parameter object
#' _except_ when an identifier was used (e.g. `neighbors = tune("K")`).
#'
#' @seealso [tune_grid()], [tune_bayes()]
#' @examplesIf tune:::should_run_examples()
#' # For grid search:
#' data("example_ames_knn")
#'
#' # Plot the tuning parameter values versus performance
#' autoplot(ames_grid_search, metric = "rmse")
#'
#'
#' # For iterative search:
#' # Plot the tuning parameter values versus performance
#' autoplot(ames_iter_search, metric = "rmse", type = "marginals")
#'
#' # Plot tuning parameters versus iterations
#' autoplot(ames_iter_search, metric = "rmse", type = "parameters")
#'
#' # Plot performance over iterations
#' autoplot(ames_iter_search, metric = "rmse", type = "performance")
#' @export
autoplot.tune_results <-
  function(
    object,
    type = c("marginals", "parameters", "performance"),
    metric = NULL,
    eval_time = NULL,
    width = NULL,
    call = rlang::current_env(),
    ...
  ) {
    type <- match.arg(type)
    has_iter <- any(names(object) == ".iter")
    if (!has_iter && type != "marginals") {
      cli::cli_abort(
        "{.code type = {glue::double_quote(type)}} is only used with
                      iterative search results."
      )
    }
    pset <- .get_tune_parameters(object)
    if (any(is.na(pset$object))) {
      p_names <- pset$id[is.na(pset$object)]
      cli::cli_abort(
        "Some parameters do not have corresponding parameter objects and
         cannot be used with {.fn autoplot}: {.arg {p_names}}."
      )
    }

    if (type == "parameters") {
      if (!is.null(eval_time)) {
        cli::cli_warn(
          "{.arg eval_time} is not used with {.code autoplot(..., type = 'parameters')}."
        )
      }
      p <- plot_param_vs_iter(object, call)
    } else {
      if (type == "performance") {
        p <- plot_perf_vs_iter(
          object,
          metric,
          eval_time = eval_time,
          width,
          call
        )
      } else {
        if (use_regular_grid_plot(object)) {
          p <- plot_regular_grid(
            object,
            metric = metric,
            eval_time = eval_time,
            call,
            ...
          )
        } else {
          p <- plot_marginals(
            object,
            metric = metric,
            eval_time = eval_time,
            call
          )
        }
      }
    }
    p
  }

#' @export
autoplot.resample_results <- function(object, ...) {
  cli::cli_abort(
    "There is no {.fn autoplot} implementation for {.cls resample_results}."
  )
}

# ------------------------------------------------------------------------------

get_param_object <- function(x) {
  att <- attributes(x)
  if (any(names(att) == "parameters")) {
    res <- att$parameters
  } else {
    res <- NULL
  }
  res
}

get_param_columns <- function(x) {
  prm <- get_param_object(x)
  if (!is.null(prm)) {
    res <- prm$id
  } else {
    dat <- collect_metrics(x)
    other_names <- c(
      ".metric",
      ".estimator",
      "mean",
      "n",
      "std_err",
      ".iter",
      ".config"
    )
    res <- names(dat)[!(names(dat) %in% other_names)]
  }
  res
}

# Use the user-given id for the parameter or the parameter label?
get_param_label <- function(x, id_val) {
  x <- tibble::as_tibble(x)
  y <- dplyr::filter(x, id == id_val) |> dplyr::slice(1)
  num_param <- sum(x$name == y$name)
  no_special_id <- y$name == y$id
  if (no_special_id && num_param == 1) {
    res <- y$object[[1]]$label
  } else {
    res <- id_val
  }
  res
}

# When visualizing fairness metrics, collapse the
# `.metric` and `.by` columns into one string, e.g.
# `demographic_parity(gender)`. Otherwise, `.by` would
# be ignored, so the same fairness metric applied to
# different data-columns (e.g. `demographic_parity(gender)` and
# `demographic_parity(race)`) would be obscured in output.
#
# For non-fairness metrics (i.e. observations with `is.na(.by)`),
# the `.metric` column is unaffected.
paste_param_by <- function(x) {
  if (".by" %in% colnames(x)) {
    x <-
      x |>
      dplyr::mutate(
        .metric = case_when(
          !is.na(.by) ~ paste0(.metric, "(", .by, ")"),
          .default = .metric
        )
      )
  }

  x
}

# ------------------------------------------------------------------------------

is_factorial <- function(x, cutoff = 0.95) {
  n <- nrow(x)
  p <- ncol(x)
  vals <- purrr::map(x, unique)
  full_fact <-
    tidyr::crossing(!!!vals) |>
    dplyr::full_join(x |> dplyr::mutate(..obs = 1), by = names(x))
  mean(!is.na(full_fact$..obs)) >= cutoff
}


is_regular_grid <- function(grid) {
  num_points <- nrow(grid)
  p <- ncol(grid)

  if (p == 1) {
    return(TRUE)
  }

  if (p <= 5) {
    ff <- is_factorial(grid)
    if (ff) {
      return(TRUE)
    }
  }

  pct_unique <- purrr::map_int(grid, \(.x) length(unique(.x))) / num_points
  max_pct_unique <- max(pct_unique, na.rm = TRUE)
  np_ratio <- p / num_points

  # Derived from simulation data and C5.0 tree
  if (max_pct_unique > 1 / 2) {
    res <- FALSE
  }
  if (max_pct_unique <= 1 / 2 & max_pct_unique <= 1 / 6) {
    res <- TRUE
  }
  if (max_pct_unique <= 1 / 2 & max_pct_unique > 1 / 6 & np_ratio > 0.05) {
    res <- TRUE
  }
  if (max_pct_unique <= 1 / 2 & max_pct_unique > 1 / 6 & np_ratio <= 0.05) {
    res <- FALSE
  }
  res
}

# This will eventually change and use a parameters object.
use_regular_grid_plot <- function(x) {
  dat <- collect_metrics(x)
  param_cols <- get_param_columns(x)
  grd <- dat |>
    dplyr::select(all_of(param_cols)) |>
    distinct()
  is_regular_grid(grd)
}

# ------------------------------------------------------------------------------

process_autoplot_metrics <- function(x, metric, eval_time) {
  met_set <- .get_tune_metrics(x)
  any_dyn <- any(purrr::map_lgl(metric, \(.x) is_dyn(met_set, .x)))

  x <- estimate_tune_results(x)

  # This next line updates the .metric columns to be consistent with what the
  # metric set produces. For example, in the data, there might be a value of
  # "demographic_parity" but the metric set knows about
  # "demographic_parity(cyl)". `paste_param_by()` makes sure that they both
  # have the same value (if any).
  x <- paste_param_by(x)

  x <- x |>
    dplyr::filter(.metric %in% metric) |>
    dplyr::filter(!is.na(mean))

  num_eval_times <- length(eval_time[!is.na(eval_time)])

  if (any_dyn & num_eval_times > 0) {
    x <- x |>
      dplyr::filter(.eval_time %in% eval_time) |>
      dplyr::mutate(
        .metric = dplyr::if_else(
          condition = !is.na(.eval_time),
          true = paste0(.metric, " @", format(.eval_time, digits = 5)),
          false = .metric
        )
      )
  }
  x
}

# ------------------------------------------------------------------------------

plot_perf_vs_iter <- function(
  x,
  metric = NULL,
  eval_time = NULL,
  width = NULL,
  call = rlang::caller_env()
) {
  if (is.null(width)) {
    width <- max(x$.iter) / 75
  }

  metric <- check_autoplot_metrics(x, metric, call)
  eval_time <- check_autoplot_eval_times(x, metric, eval_time, call)

  x <- process_autoplot_metrics(x, metric, eval_time)

  search_iter <-
    x |>
    dplyr::filter(.iter > 0 & std_err > 0) |>
    dplyr::mutate(const = ifelse(n > 0, qt(0.975, n), 0))

  p <-
    ggplot(x, aes(x = .iter, y = mean)) +
    geom_point() +
    xlab("Iteration") +
    scale_x_continuous(breaks = iter_breaks(x$.iter))

  if (nrow(search_iter) > 0 & width > 0) {
    p <-
      p +
      geom_errorbar(
        data = search_iter,
        aes(ymin = mean - const * std_err, ymax = mean + const * std_err),
        width = width
      )
  }

  if (length(unique(x$.metric)) > 1) {
    p <- p + facet_wrap(~.metric, scales = "free_y")
  } else {
    p <- p + ylab(unique(x$.metric))
  }
  p
}

plot_param_vs_iter <- function(x, call = rlang::caller_env()) {
  param_cols <- get_param_columns(x)
  pset <- get_param_object(x)
  if (is.null(pset)) {
    cli::cli_abort(
      "{.fn autoplot} requires objects made with {.pkg tune}
                    version 0.1.0 or later."
    )
  }

  # ----------------------------------------------------------------------------
  # Collect and filter resampling results

  x <- estimate_tune_results(x)
  is_num <- purrr::map_lgl(
    x |> dplyr::select(dplyr::all_of(param_cols)),
    is.numeric
  )
  num_param_cols <- param_cols[is_num]

  # ----------------------------------------------------------------------------
  # Transform and re-label when needed. Previous vectors of names are updated.

  for (prm in param_cols) {
    pobj <- pset$object[[which(pset$id == prm)]]
    lab <- get_param_label(pset, prm)
    if (!is.null(pobj$trans)) {
      x[[prm]] <- pobj$trans$transform(x[[prm]])
      new_name <- paste0(lab, " (", pobj$trans$name, ")")
    } else {
      new_name <- lab
    }
    names(x)[names(x) == prm] <- new_name
    num_param_cols[num_param_cols == prm] <- new_name
    param_cols[param_cols == prm] <- new_name
  }

  # ----------------------------------------------------------------------------
  # Stack numeric columns for filtering

  x <-
    x |>
    dplyr::select(.iter, dplyr::all_of(num_param_cols)) |>
    tidyr::pivot_longer(cols = dplyr::all_of(num_param_cols))

  # ------------------------------------------------------------------------------

  p <-
    ggplot(x, aes(x = .iter, y = value)) +
    geom_point() +
    xlab("Iteration") +
    scale_x_continuous(breaks = iter_breaks(x$.iter))

  if (length(param_cols) == 1) {
    p <- p + ylab(param_cols)
  } else {
    p <- p + ylab("") + facet_wrap(~name, scales = "free_y")
  }

  p
}

plot_marginals <- function(
  x,
  metric = NULL,
  eval_time = NULL,
  call = rlang::caller_env()
) {
  param_cols <- get_param_columns(x)
  pset <- get_param_object(x)
  if (is.null(pset)) {
    cli::cli_abort(
      "{.fn autoplot} requires objects made with {.pkg tune}
                   version 0.1.0 or later."
    )
  }

  # ----------------------------------------------------------------------------
  # Collect and filter resampling results

  is_race <- inherits(x, "tune_race")

  metric <- check_autoplot_metrics(x, metric, call)
  eval_time <- check_autoplot_eval_times(x, metric, eval_time, call)

  x <- process_autoplot_metrics(x, metric, eval_time)

  # ----------------------------------------------------------------------------
  # Check types of parameters then sort by unique values

  is_num <- purrr::map_lgl(
    x |> dplyr::select(dplyr::all_of(param_cols)),
    is.numeric
  )
  num_val <- purrr::map_int(
    x |> dplyr::select(dplyr::all_of(param_cols)),
    \(.x) length(unique(.x))
  )

  if (any(num_val < 2)) {
    rm_param <- param_cols[num_val < 2]
    param_cols <- param_cols[num_val >= 2]
    is_num <- is_num[num_val >= 2]
    x <- x |> dplyr::select(-dplyr::all_of(rm_param))
  }

  if (any(!is_num)) {
    num_param_cols <- param_cols[is_num]
    chr_param_cols <- param_cols[!is_num]
    if (length(chr_param_cols) > 1) {
      cli::cli_abort(
        "Currently cannot autoplot grids with 2+ non-numeric parameters."
      )
    }
    if (length(num_param_cols) == 0) {
      cli::cli_abort(
        "Currently cannot autoplot grids with only non-numeric parameters."
      )
    }
    num_val <- num_val[param_cols %in% chr_param_cols]
    names(num_val) <- chr_param_cols
    num_val <- sort(num_val, decreasing = TRUE)
  } else {
    num_param_cols <- param_cols
    chr_param_cols <- character(0)
  }

  # ----------------------------------------------------------------------------
  # Transform and re-label when needed. Previous vectors of names are updated.

  for (prm in param_cols) {
    pobj <- pset$object[[which(pset$id == prm)]]
    lab <- get_param_label(pset, prm)
    if (!is.null(pobj$trans)) {
      x[[prm]] <- pobj$trans$transform(x[[prm]])
      new_name <- paste0(lab, " (", pobj$trans$name, ")")
    } else {
      new_name <- lab
    }
    names(x)[names(x) == prm] <- new_name
    num_param_cols[num_param_cols == prm] <- new_name
    chr_param_cols[chr_param_cols == prm] <- new_name
    param_cols[param_cols == prm] <- new_name
  }

  # ----------------------------------------------------------------------------
  # Stack numeric parameters for faceting.

  x <-
    x |>
    dplyr::rename(`# resamples` = n) |>
    dplyr::select(dplyr::all_of(param_cols), mean, `# resamples`, .metric) |>
    tidyr::pivot_longer(cols = dplyr::all_of(num_param_cols))

  # ----------------------------------------------------------------------------

  p <- ggplot(x, aes(x = value, y = mean))

  if (length(chr_param_cols) > 0) {
    if (is_race) {
      p <- p +
        geom_point(aes(
          col = !!sym(chr_param_cols),
          alpha = `# resamples`,
          size = resamples
        ))
      p <- p + ggplot2::labs(color = chr_param_cols)
    } else {
      p <- p + geom_point(aes(col = !!sym(chr_param_cols)), alpha = .7)
      p <- p + ggplot2::labs(color = chr_param_cols)
    }
  } else {
    if (is_race) {
      p <- p + geom_point(aes(alpha = `# resamples`, size = `# resamples`))
    } else {
      p <- p + geom_point(alpha = .7)
    }
  }

  if (length(unique(x$.metric)) > 1) {
    if (length(num_param_cols) == 1) {
      p <-
        p +
        facet_wrap(~.metric, scales = "free_y") +
        xlab(num_param_cols) +
        ylab("")
    } else {
      p <-
        p +
        ggplot2::facet_grid(.metric ~ name, scales = "free") +
        xlab("") +
        ylab("")
    }
  } else {
    if (length(num_param_cols) == 1) {
      p <- p + xlab(num_param_cols) + ylab(unique(x$.metric))
    } else {
      p <-
        p +
        facet_wrap(~name, scales = "free_x") +
        xlab("") +
        ylab(unique(x$.metric))
    }
  }

  p
}


plot_regular_grid <- function(
  x,
  metric = NULL,
  eval_time = NULL,
  call = rlang::caller_env(),
  ...
) {
  # Collect and filter resampling results

  is_race <- inherits(x, "tune_race")

  metric <- check_autoplot_metrics(x, metric, call)
  eval_time <- check_autoplot_eval_times(x, metric, eval_time, call)

  dat <- process_autoplot_metrics(x, metric, eval_time)

  check_singular_metric(dat, call)

  multi_metrics <- length(unique(dat$.metric)) > 1

  # ----------------------------------------------------------------------------
  # Get information about parameters

  param_cols <- get_param_columns(x)
  pset <- get_param_object(x)
  if (is.null(pset)) {
    cli::cli_abort(
      "The {.fn autoplot} function requires objects made with
                    {.pkg tune} version 0.1.0 or later."
    )
  }

  grd <- dat |> dplyr::select(all_of(param_cols))

  # ----------------------------------------------------------------------------
  # Determine which parameter goes on the x-axis and their types

  is_num <- purrr::map_lgl(grd, is.numeric)
  num_param_cols <- param_cols[is_num]
  chr_param_cols <- param_cols[!is_num]

  num_values <- purrr::map_int(grd[, num_param_cols], \(.x) length(unique(.x)))
  num_values <- sort(num_values, decreasing = TRUE)

  if (!any(is_num)) {
    x_col <- chr_param_cols[1]
    grp_cols <- chr_param_cols[-1]
  } else {
    x_col <- names(num_values)[1]
    grp_cols <- c(chr_param_cols, names(num_values)[-1])
  }

  g <- length(grp_cols)

  # ----------------------------------------------------------------------------

  if (g > 5) {
    return(plot_marginals(x, metric, eval_time, call))
  }

  # ----------------------------------------------------------------------------
  # Transform and re-label when needed. Previous vectors of names are updated.

  x_col_prm <- pset$object[[which(pset$id == x_col)]]
  if (inherits(x_col_prm, "quant_param") && !is.null(x_col_prm$trans)) {
    trans <- x_col_prm$trans
  } else {
    trans <- NULL
  }

  for (prm in param_cols) {
    pobj <- pset$object[[which(pset$id == prm)]]
    new_name <- get_param_label(pset, prm)
    names(dat)[names(dat) == prm] <- new_name
    grp_cols[grp_cols == prm] <- new_name
    x_col[x_col == prm] <- new_name
    param_cols[param_cols == prm] <- new_name
  }

  # ----------------------------------------------------------------------------

  dat <-
    dat |>
    dplyr::rename(`# resamples` = n) |>
    dplyr::select(dplyr::all_of(param_cols), mean, `# resamples`, .metric) |>
    tidyr::pivot_longer(cols = dplyr::all_of(x_col))

  # ------------------------------------------------------------------------------

  if (g >= 1) {
    # Here we know that there is at least one grouping parameter. For the first,
    # we assign it to the color aesthetic. If it is numeric, it is converted to
    # character using format().
    col_col <- grp_cols[1]
    if (is.numeric(dat[[col_col]])) {
      dat[[col_col]] <- format(dat[[col_col]], ...)
    }
    col_col <- rlang::ensym(col_col)
    p <- ggplot(
      dat,
      aes(value, y = mean, col = {{ col_col }}, group = {{ col_col }})
    )
    # Since `col_col` has either the parameter id or the parameter label, use
    # is in the key:

    p <- p + ggplot2::labs(color = col_col, x = x_col)

    if (g >= 2) {
      # Since there at 2 - 5 grouping parameters, the others are assigned to
      # column facets. Row facets will be for performance metrics.
      facets <- grp_cols[-1]
      facets <- purrr::map(facets, sym)
      facets <- rlang::quos(!!!facets)
      # faceting variables
      if (multi_metrics) {
        p <- p +
          facet_grid(
            rows = vars(.metric),
            vars(!!!facets),
            labeller = ggplot2::labeller(.cols = ggplot2::label_both),
            scales = "free_y"
          )
      } else {
        p <-
          p +
          facet_wrap(
            vars(!!!facets),
            labeller = ggplot2::labeller(.cols = ggplot2::label_both)
          )
      }
    } else if (multi_metrics) {
      p <- p + facet_grid(rows = vars(.metric), scales = "free_y")
    }
  } else {
    # Only a single parameter and potentially multiple metrics.
    p <- ggplot(dat, aes(x = value, y = mean))
    if (multi_metrics) {
      p <- p + facet_wrap(~.metric, scales = "free_y", ncol = 1)
    }
  }

  if (is_race) {
    p <- p + geom_point(aes(alpha = `# resamples`, size = `# resamples`))
  } else {
    p <- p + geom_point(size = 1)
  }

  if (multi_metrics) {
    p <- p + ylab("")
  } else {
    dat$.metric[1]
    p <- p + ylab(dat$.metric[1])
  }
  if (nrow(pset) == 1) {
    x_lab <- pset$object[[1]]$label
    p <- p + xlab(x_lab)
  }

  if (any(is_num)) {
    p <- p + geom_line()
  }

  if (!is.null(trans)) {
    p <- p + ggplot2::scale_x_continuous(trans = trans)
  }

  p
}


# ------------------------------------------------------------------------------

iter_breaks <- function(x, num = 10) {
  un_z <- sort(unique(x))
  n <- length(un_z)
  if (n < num) {
    num <- n
  }
  brks <- pretty(un_z, n = num)
  brks <- unique(floor(brks))
  names(brks) <- attr(brks, "labels")
  brks
}

# Overview ---------------------------------------------------------------------
#
# General logging to the screen is used in various places. The code in this file
# is split into 4 sections.
#
# The first section contains functions that tune uses to let the user know what
# is happening. The main function is `update_printer()`, which is used in
# `logging.R` and throughout the package. Some examples of these are
# `Fold01: preprocessor 1/1, model 1/1` from `tune_grid()` and
# `Generating 3 candidates` from `tune_bayes()`. This function should be fully
# controlled by `control$verbose`.
#
# The second section contains everything related to the catalog object. This
# object contains all the errors and warnings that are captured when handling
# workflows inside the training loop. The catalog is a tibble with the columns
# `type`, `note`, `n`, and `id`.
#
# - `type` being whether it is a warning or an error,
# - `note` is what the condition was as a string,
# - `n` is how many times it was seen
# - `id` is a unique identifier (created using `lbls`)
#
# All of this infrastructure is done using the `tune_env` environment that
# contains the catalog itself, and other useful information that toggles whether
# we are using the catalog and ensures it is used correctly across the fits.
#
# The third section, `catching and logging`, is what is being used in tune or
# extension packages. `.catch_and_log()` is wrapping some code that we want to
# collect on. If an error or warning is seen, it will be captured and handled
# using the code in this and the previous section.
#
# The forth section contains functions that are exclusively used in
# `tune_bayes()`.

# Tune printing ----------------------------------------------------------------

update_printer <- function(
  control,
  split_labels = NULL,
  task,
  type = "success",
  catalog = TRUE,
  ...
) {
  if (!any(control$verbose, control$verbose_iter)) {
    return(invisible(NULL))
  }

  if (task == "internal") {
    return(NULL)
  }

  if (!is.null(split_labels)) {
    labs <- rev(unlist(split_labels))
    labs <- paste0(labs, collapse = ", ")
    labs <- paste0(labs, ": ")
  } else {
    labs <- ""
  }

  # see https://github.com/r-lib/cli/issues/92
  task <- gsub("\\{", "", task)

  task <- paste0(labs, task)
  siren(task, type = type)
  NULL
}

siren <- function(x, type = "info") {
  tune_color <- get_tune_colors()
  types <- names(tune_color$message)
  type <- match.arg(type, types)

  msg <- glue::glue(x, .trim = FALSE)

  symb <- dplyr::case_when(
    type == "warning" ~ tune_color$symbol$warning("!"),
    type == "go" ~ tune_color$symbol$go(cli::symbol$pointer),
    type == "danger" ~ tune_color$symbol$danger("x"),
    type == "success" ~ tune_color$symbol$success(tune_symbol$success),
    type == "info" ~ tune_color$symbol$info("i")
  )

  msg <- dplyr::case_when(
    type == "warning" ~ tune_color$message$warning(msg),
    type == "go" ~ tune_color$message$go(msg),
    type == "danger" ~ tune_color$message$danger(msg),
    type == "success" ~ tune_color$message$success(msg),
    type == "info" ~ tune_color$message$info(msg)
  )

  if (inherits(msg, "character")) {
    msg <- as.character(msg)
  }

  msg <- paste(symb, msg)

  message(msg)
}

# Catalog ----------------------------------------------------------------------

tune_env <- rlang::new_environment(
  data = list(
    progress_env = NULL,
    progress_active = FALSE,
    progress_catalog = NULL,
    progress_started = FALSE
  )
)

lbls <- c(LETTERS, letters, 1:1e3)

# determines whether a currently running tuning process uses the cataloger.
uses_catalog <- function() {
  isTRUE(tune_env$progress_active && !is_testing())
}

# copied from testthat::is_testing
is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

# determines whether a tuning process is currently active.
# this function is intended to guard calls to `initialize_catalog()`
# inside of machinery that is wrapped by other tuning
# methods. e.g. `tune_bayes()` has an issue cataloger that ought not to be
# overwritten when fitting over an initial grid.
catalog_is_active <- function() {
  tune_env$progress_active
}

# initializes machinery for the tune catalog inside of an environment.
# the `env` should be an execution environment that persists throughout the
# tuning process for a given tuning approach and exits once tuning is completed.
#
# this function attaches an exit handler (see `on.exit()`) to the execution
# environment of the function it's called within, by default. pay close
# attention when other exit handlers are attached to the same environment;
# a call to `on.exit()` in `env` after this function is called will cause
# issues with the catalog. (see #845.)
#
#' @rdname tune-internal-functions
#' @export
initialize_catalog <- function(
  control,
  env = rlang::caller_env(),
  workflow = NULL
) {
  catalog <-
    tibble::new_tibble(
      list(
        type = character(0),
        note = character(0),
        n = numeric(0),
        id = numeric(0)
      ),
      nrow = 0
    )

  if (is.null(workflow)) {
    workflow <- workflow()
  }

  if (
    choose_framework(workflow, control) == "sequential" &&
      !is_testing() &&
      !control$verbose
  ) {
    progress_active <- TRUE
  } else {
    progress_active <- FALSE
  }

  rlang::env_bind(tune_env, progress_env = env)

  rlang::env_bind(tune_env, progress_catalog = catalog)
  withr::defer(
    rlang::env_bind(tune_env, progress_catalog = NULL),
    envir = env
  )

  rlang::env_bind(tune_env, progress_active = progress_active)
  withr::defer(
    rlang::env_bind(tune_env, progress_active = FALSE),
    envir = env
  )
  withr::defer(
    rlang::env_bind(tune_env, progress_started = FALSE),
    envir = env
  )

  invisible(NULL)
}

# given a catalog, summarize errors and warnings in a 1-length glue vector.
# for use by the progress bar inside of `tune_catalog()`.
summarize_catalog <- function(catalog, sep = "   ") {
  if (nrow(catalog) == 0) {
    return("")
  }

  res <- dplyr::arrange(catalog, id)
  res <- dplyr::mutate(
    res,
    color = dplyr::if_else(
      type == "warning",
      list(cli::col_yellow),
      list(cli::col_red)
    )
  )
  res <- dplyr::rowwise(res)
  res <- dplyr::mutate(
    res,
    msg = glue::glue("{color(cli::style_bold(lbls[id]))}: x{n}")
  )
  res <- dplyr::ungroup(res)
  res <- dplyr::pull(res, msg)
  res <- glue::glue_collapse(res, sep = sep)

  res
}

# catching and logging ---------------------------------------------------------

#' @export
#' @rdname tune-internal-functions
.catch_and_log <- function(
  .expr,
  ...,
  bad_only = FALSE,
  notes,
  catalog = TRUE
) {
  dots <- list(...)
  update_printer(..., type = "info", catalog = catalog, task = dots$location)
  tmp <- catcher(.expr)

  if (has_log_notes(tmp)) {
    notes <- append_log_notes(notes, tmp, dots$location)
    catalog_log(notes)
  }
  tmp <- remove_log_notes(tmp)
  assign("notes", notes, envir = parent.frame())

  tmp
}

catcher <- function(expr) {
  signals <- list()
  add_cond <- function(cnd) {
    signals <<- append(signals, list(rlang::cnd_entrace(cnd)))
    rlang::cnd_muffle(cnd)
  }

  res <- rlang::try_fetch(
    expr,
    warning = add_cond,
    error = function(e) {
      structure(
        catch_message(e),
        class = "try-error",
        # if a simple error, add a traceback.
        # otherwise, pass the condition right along.
        condition = rlang::`%||%`(rlang::cnd_entrace(e), e)
      )
    }
  )

  attr(res, "notes") <- signals
  res
}

has_log_notes <- function(x) {
  is_failure(x) || NROW(attr(x, "notes")) > 0
}

is_failure <- function(x) {
  inherits(x, "try-error")
}

append_log_notes <- function(notes, x, location) {
  if (is.null(notes)) {
    notes <- new_note()
  }

  wrns <- attr(x, "notes")
  if (length(wrns) > 0) {
    for (wrn in wrns) {
      type <- "warning"
      note <- wrn$message

      notes <- tibble::add_row(
        notes,
        location = unclass(location),
        type = type,
        note = note,
        trace = list(wrn$trace)
      )
    }
  }

  if (is_failure(x)) {
    type <- "error"
    x <- attr(x, "condition")
    note <- conditionMessage(x)

    notes <- tibble::add_row(
      notes,
      location = unclass(location),
      type = type,
      note = note,
      trace = list(x$trace)
    )
  }

  notes
}

new_note <- function(
  location = character(0),
  type = character(0),
  note = character(0),
  trace = list()
) {
  tibble::new_tibble(
    list(
      location = location,
      type = type,
      note = note,
      trace = trace
    )
  )
}


catalog_log <- function(x) {
  catalog <- rlang::env_get(tune_env, "progress_catalog")

  if (is.null(catalog)) {
    catalog <- tibble::new_tibble(
      list(
        type = character(0),
        note = character(0),
        n = numeric(0),
        id = numeric(0)
      ),
      nrow = 0
    )
  }

  for (i in seq_along(x$note)) {
    x_note <- x$note[i]
    x_type <- x$type[i]

    if (x_note %in% catalog$note) {
      idx <- match(x_note, catalog$note)
      catalog$n[idx] <- catalog$n[idx] + 1
    } else {
      new_id <- nrow(catalog) + 1
      catalog <- tibble::add_row(
        catalog,
        tibble::tibble(
          type = x_type,
          note = x_note,
          n = 1,
          id = new_id
        )
      )

      # construct issue summary
      color <- if (x_type == "warning") cli::col_yellow else cli::col_red
      # pad by nchar(label) + nchar("warning") + additional spaces and symbols
      pad <- nchar(new_id) + 14L
      justify <- paste0("\n", strrep("\u00a0", pad))
      note <- gsub("\n", justify, x_note)
      # pad `nchar("warning") - nchar("error")` spaces to the right of the `:`
      if (x_type == "error") {
        note <- paste0("\u00a0\u00a0", note)
      }
      msg <- glue::glue(
        "{color(cli::style_bold(lbls[new_id]))} | {color(x_type)}: {note}"
      )
      cli::cli_alert(msg)
    }
  }

  rlang::env_bind(tune_env, progress_catalog = catalog)
  if (uses_catalog()) {
    rlang::env_bind(
      tune_env$progress_env,
      catalog_summary = summarize_catalog(catalog)
    )

    if (!tune_env$progress_started) {
      rlang::with_options(
        cli::cli_progress_bar(
          type = "custom",
          format = "There were issues with some computations   {catalog_summary}",
          clear = FALSE,
          .envir = tune_env$progress_env
        ),
        cli.progress_show_after = 0
      )
      rlang::env_bind(tune_env, progress_started = TRUE)
    }

    cli::cli_progress_update(.envir = tune_env$progress_env)
  }

  return(NULL)
}

remove_log_notes <- function(x) {
  attr(x, "notes") <- NULL
  x
}

# Bayes specific printing ------------------------------------------------------

log_best <- function(control, iter, info, digits = 4) {
  if (!isTRUE(control$verbose_iter)) {
    return(invisible(NULL))
  }

  message("")
  message(cli::rule(left = cli::style_bold(paste("Iteration", iter))))
  message("")

  msg <-
    paste0(
      "Current best:\t\t",
      info$metrics,
      "=",
      signif(info$best_val, digits = digits),
      " (@iter ",
      info$best_iter,
      ")"
    )
  update_printer(
    control,
    split_labels = NULL,
    task = msg,
    type = "info",
    catalog = FALSE
  )
}

check_and_log_flow <- function(control, results) {
  if (!isTRUE(control$verbose_iter)) {
    return(invisible(NULL))
  }

  if (all(is.na(results$.mean))) {
    if (nrow(results) < 2) {
      update_printer(
        control,
        split_labels = NULL,
        task = "Halting search",
        type = "danger",
        catalog = FALSE
      )
      eval.parent(parse(text = "break"))
    } else {
      update_printer(
        control,
        split_labels = NULL,
        task = "Skipping to next iteration",
        type = "danger",
        catalog = FALSE
      )
      eval.parent(parse(text = "next"))
    }
  }
  invisible(NULL)
}

log_progress <- function(
  control,
  x,
  maximize = TRUE,
  objective = NULL,
  eval_time = NULL,
  digits = 4
) {
  if (!isTRUE(control$verbose_iter)) {
    return(invisible(NULL))
  }

  x <- dplyr::filter(x, .metric == objective)
  if (!is.null(eval_time)) {
    x <- dplyr::filter(x, .eval_time == eval_time)
  }

  if (maximize) {
    bst <- which.max(x$mean)
  } else {
    bst <- which.min(x$mean)
  }
  bst_iter <- x$.iter[bst]
  max_iter <- max(x$.iter)
  bst_val <- x$mean[x$.iter == max_iter]
  bst_se <- x$std_err[x$.iter == max_iter]
  msg <-
    paste0(
      " Newest results:\t",
      objective,
      "=",
      signif(bst_val, digits = digits)
    )
  if (!is.na(bst_se) && bst_se > 0) {
    msg <- paste0(msg, " (+/-", signif(bst_se, digits = digits - 1), ")")
  }

  if (bst_iter == max_iter) {
    msg <- paste0(cli::col_red(cli::symbol$heart), msg)
  } else {
    msg <- paste0(cli::col_silver(cli::symbol$circle_cross), msg)
  }
  message(msg)
}

param_msg <- function(control, candidate) {
  if (!isTRUE(control$verbose_iter)) {
    return(invisible(NULL))
  }
  candidate <- candidate[,
    !(names(candidate) %in% c(".mean", ".sd", "objective"))
  ]
  p_chr <- paste0(
    names(candidate),
    "=",
    format(as.data.frame(candidate), digits = 3)
  )
  p_chr <- paste0(p_chr, collapse = ", ")
  message_wrap(
    p_chr,
    prefix = "i",
    color_text = get_tune_colors()$message$info,
    color_prefix = get_tune_colors()$symbol$info
  )
  invisible(NULL)
}

#' Write a message that respects the line width
#'
#' @param x A character string of the message text.
#' @param width An integer for the width.
#' @param prefix An optional string to go on the first line of the message.
#' @param color_text,color_prefix A function (or `NULL`) that is used to color
#'  the text and/or prefix.
#' @return The processed text is returned (invisibly) but a message is written.
#' @examples
#' library(cli)
#' Gaiman <-
#'   paste(
#'     '"Good point." Bod was pleased with himself, and glad he had thought of',
#'     "asking the poet for advice. Really, he thought, if you couldn't trust a",
#'     "poet to offer sensible advice, who could you trust?",
#'     collapse = ""
#'   )
#' message_wrap(Gaiman)
#' message_wrap(Gaiman, width = 20, prefix = "-")
#' message_wrap(Gaiman,
#'   width = 30, prefix = "-",
#'   color_text = cli::col_silver
#' )
#' message_wrap(Gaiman,
#'   width = 30, prefix = "-",
#'   color_text = cli::style_underline,
#'   color_prefix = cli::col_green
#' )
#' @export
message_wrap <-
  function(
    x,
    width = options()$width - 2,
    prefix = "",
    color_text = NULL,
    color_prefix = color_text
  ) {
    check_string(x)
    check_function(color_text, allow_null = TRUE)
    check_function(color_prefix, allow_null = TRUE)
    n <- nchar(prefix)
    if (n > 0) {
      buffer <- paste0(rep(" ", n + 1), collapse = "")
    } else {
      buffer <- ""
    }
    msg <- strwrap(x, width = width - n - 1)
    if (!is.null(color_text)) {
      msg <- purrr::map_chr(msg, color_text)
    }
    if (!is.null(color_prefix)) {
      prefix <- color_prefix(prefix)
    }
    msg[-length(msg)] <- paste0(msg[-length(msg)], "\n")
    msg[-1] <- paste0(buffer, msg[-1])
    if (n > 0) {
      msg[1] <- paste(prefix, msg[1])
    }
    message(msg)
    invisible(msg)
  }


acq_summarizer <- function(control, iter, objective = NULL, digits = 4) {
  if (!isTRUE(control$verbose_iter)) {
    return(invisible(NULL))
  }
  if (inherits(objective, "conf_bound") && is.function(objective$kappa)) {
    val <- paste0(
      "Kappa value: ",
      signif(objective$kappa(iter), digits = digits)
    )
  } else {
    if (
      inherits(objective, c("exp_improve", "prob_improve")) &&
        is.function(objective$trade_off)
    ) {
      val <- paste0(
        "Trade-off value: ",
        signif(objective$trade_off(iter), digits = digits)
      )
    } else {
      val <- NULL
    }
  }
  if (!is.null(val)) {
    update_printer(
      control,
      split_labels = NULL,
      task = val,
      type = "info",
      catalog = FALSE
    )
  }
  invisible(NULL)
}

# End --------------------------------------------------------------------------

test_that("catch errors", {
  res_1 <- tune:::catcher(log("a"))
  expect_true(class(res_1) == "try-error")
  expect_true(length(attr(res_1, "notes")) == 0)

  res_2 <- tune:::catcher(log(1))
  expect_true(res_2 == log(1))
  expect_true(length(attr(res_2, "notes")) == 0)

  res_3 <- tune:::catcher(log(-1))
  expect_true(is.nan(res_3))
  expect_true(class(attr(res_3, "notes")) == "list")
  expect_true(inherits(attr(res_3, "notes")[[1]], "simpleWarning"))
  expect_true(inherits(attr(res_3, "notes")[[1]], "warning"))
  expect_true(inherits(attr(res_3, "notes")[[1]], "condition"))
})

test_that("low-level messages", {
  expect_snapshot(
    error = TRUE,
    siren("a", "werewolf")
  )
  expect_snapshot(siren("bat", "info"))
  expect_snapshot(siren("bat", "go"))
  expect_snapshot(siren("bat", "danger"))
  expect_snapshot(siren("bat", "warning"))

  skip_on_os("windows")
  expect_snapshot(siren("bat", "success"))
})

test_that("update_printer", {
  ctrl_t <- control_grid(verbose = TRUE)
  ctrl_f <- control_grid(verbose = FALSE)
  rs <- labels(rsample::vfold_cv(mtcars)$splits[[1]])

  expect_snapshot(update_printer(ctrl_t, rs, task = "cube", type = "go"))
  expect_snapshot(update_printer(ctrl_t, NULL, task = "cube", type = "go"))
  expect_silent(update_printer(ctrl_f, NULL, task = "cube", type = "go"))

  skip_on_os("windows")
  expect_snapshot(update_printer(ctrl_t, rs, task = "cube", type = "success"))
})

test_that("logging iterations", {
  ctrl_t <- control_bayes(verbose_iter = TRUE)
  ctrl_f <- control_bayes(verbose_iter = FALSE)
  sc_1 <- list(
    best_val = 7,
    best_iter = 2,
    last_impr = 3,
    uncertainty = 0,
    overall_iter = 1,
    metrics = .8,
    max = FALSE
  )

  expect_snapshot(log_best(ctrl_t, 10, sc_1))
  expect_silent(log_best(ctrl_f, 10, sc_1))
})

test_that("logging search info", {
  ctrl_t <- control_bayes(verbose_iter = TRUE)
  tb_1 <- tibble::tibble(.mean = 1:3)

  expect_silent(check_and_log_flow(ctrl_t, tb_1))
  expect_snapshot(
    error = TRUE,
    check_and_log_flow(ctrl_t, tb_1 |> mutate(.mean = .mean * NA))
  )
  expect_snapshot(
    error = TRUE,
    check_and_log_flow(ctrl_t, tb_1 |> mutate(.mean = .mean * NA) |> slice(1))
  )
})

test_that("current results", {
  ctrl_t <- control_bayes(verbose_iter = TRUE)
  ctrl_f <- control_bayes(verbose_iter = FALSE)
  tb_2 <-
    tibble::tibble(
      .metric = rep(letters[1:2], each = 4),
      mean = 1:8,
      .iter = 1:8,
      std_err = (1:8) / 10
    )

  expect_snapshot(
    log_progress(ctrl_t, tb_2, maximize = FALSE, objective = "a")
  )
  expect_snapshot(
    log_progress(ctrl_t, tb_2, maximize = TRUE, objective = "b")
  )
  expect_silent(log_progress(ctrl_f, tb_2, maximize = TRUE, objective = "a"))

  skip_on_os("windows")

  expect_snapshot(
    log_progress(ctrl_t, tb_2, maximize = TRUE, objective = "a")
  )
})


test_that("show parameters", {
  ctrl_t <- control_bayes(verbose_iter = TRUE)
  ctrl_f <- control_bayes(verbose_iter = FALSE)

  expect_snapshot(param_msg(ctrl_t, iris[1, 4:5]))
  expect_silent(param_msg(ctrl_f, iris[1, 4:5]))
})


test_that("acquisition functions", {
  ctrl_t <- control_bayes(verbose_iter = TRUE)
  ctrl_f <- control_bayes(verbose_iter = FALSE)

  expect_silent(acq_summarizer(ctrl_t, 1))
  expect_silent(acq_summarizer(ctrl_t, 1, conf_bound()))
  expect_silent(acq_summarizer(ctrl_f, 1, conf_bound()))
  expect_snapshot(acq_summarizer(ctrl_t, 1, conf_bound(I)))
  expect_snapshot(acq_summarizer(ctrl_t, 1, exp_improve(I)))
  expect_snapshot(acq_summarizer(ctrl_t, 1, prob_improve(I)))
})

## -----------------------------------------------------------------------------

test_that("message_wrap", {
  text <-
    paste(
      "A data frame of tuning combinations or a positive integer. The data",
      "frame should have columns for each parameter being tuned and rows for",
      "tuning parameter candidates. An integer denotes the number of candidate",
      "parameter sets to be created automatically.",
      collapse = ""
    )
  verify_output(
    test_path("message_wrap/width_30.txt"),
    message_wrap(text, width = 30),
    crayon = TRUE
  )
  verify_output(
    test_path("message_wrap/width_50_prefix.txt"),
    message_wrap(text, width = 50, prefix = "'grid':"),
    crayon = TRUE
  )
  verify_output(
    test_path("message_wrap/width_50_prefix_red_text.txt"),
    message_wrap(
      text,
      width = 50,
      prefix = "'grid':",
      color_text = cli::col_red
    ),
    crayon = TRUE
  )
  verify_output(
    test_path("message_wrap/width_50_prefix_red_text_blue_prefix.txt"),
    message_wrap(
      text,
      width = 50,
      prefix = "'grid':",
      color_text = cli::col_red,
      color_prefix = cli::col_blue
    ),
    crayon = TRUE
  )
})

test_that("interactive logger works (fit_resamples, warning + error)", {
  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )

  raise_warning <- function(x) {
    warning("ope! yikes.")
  }
  raise_error <- function(x) {
    stop("AHHhH")
  }

  set.seed(1)
  expect_snapshot(
    {
      res_fit <-
        fit_resamples(
          parsnip::nearest_neighbor("regression", "kknn"),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5),
          control = control_resamples(
            extract = function(x) {
              raise_warning()
              raise_error()
            }
          )
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})

test_that("interactive logger works (fit_resamples, rlang warning + error)", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )

  raise_warning_rl <- function(x) {
    rlang::warn("ope! yikes. (but rlang)")
  }
  raise_error_rl <- function(x) {
    rlang::abort("AHHhH (but rlang)")
  }

  set.seed(1)
  expect_snapshot(
    {
      res_fit <-
        fit_resamples(
          parsnip::nearest_neighbor("regression", "kknn"),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5),
          control = control_resamples(
            extract = function(x) {
              raise_warning_rl()
              raise_error_rl()
            }
          )
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})


test_that("interactive logger works (fit_resamples, multiline)", {
  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )
  skip_on_cran()

  raise_multiline_conditions <- function(x) {
    cli::cli_warn(c("hmmm what's happening", "uuuhhHhH"))
    cli::cli_abort(c("aHHHksdjvndiuf", "!" = "),:"))
    x
  }

  set.seed(1)
  expect_snapshot(
    {
      res_fit <-
        fit_resamples(
          parsnip::nearest_neighbor("regression", "kknn"),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5),
          control = control_resamples(
            extract = raise_multiline_conditions
          )
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})

test_that("interactive logger works (fit_resamples, occasional error)", {
  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )
  skip_on_cran()

  raise_error_later <- function() {
    local({
      count <- 0
      function(x) {
        count <<- count + 1
        if (count > 3) {
          stop("this errors now! ha!")
        }
        "hi"
      }
    })
  }
  later <- raise_error_later()

  set.seed(1)
  expect_snapshot(
    {
      res_fit <-
        fit_resamples(
          parsnip::nearest_neighbor("regression", "kknn"),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5),
          control = control_resamples(extract = later)
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})

test_that("interactive logger works (fit_resamples, occasional errors)", {
  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )

  skip_on_cran()

  raise_error_once <- function() {
    local({
      first <- TRUE
      function(x) {
        if (first) {
          first <<- FALSE
          stop("oh no")
        }

        "hi"
      }
    })
  }

  raise_error_later <- function() {
    local({
      count <- 0
      function(x) {
        count <<- count + 1
        if (count > 3) {
          stop("this errors now! ha!")
        }
        "hi"
      }
    })
  }

  once <- raise_error_once()
  later <- raise_error_later()

  set.seed(1)
  expect_snapshot(
    {
      res_fit <-
        fit_resamples(
          parsnip::nearest_neighbor("regression", "kknn"),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 10),
          control = control_resamples(
            extract = function(x) {
              once()
              later()
            }
          )
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})


test_that("interactive logger works (fit_resamples, many distinct errors)", {
  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )

  skip_on_cran()
  # Enough characters to see 'E: x1'
  rlang::local_options(cli.width = 84)

  raise_error_numbered <- function() {
    local({
      count <- 0
      function(x) {
        count <<- count + 1
        stop(paste0("error number ", count))
        "hi"
      }
    })
  }
  numbered <- raise_error_numbered()

  set.seed(1)
  expect_snapshot(
    {
      res_fit <-
        fit_resamples(
          parsnip::nearest_neighbor("regression", "kknn"),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5),
          control = control_resamples(extract = numbered)
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})

test_that("interactive logger works (tune grid, error)", {
  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )
  skip_on_cran()

  raise_error <- function(x) {
    stop("AHHhH")
  }

  set.seed(1)
  expect_snapshot(
    {
      res_fit <-
        tune_grid(
          parsnip::nearest_neighbor("regression", "kknn", dist_power = tune()),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5),
          grid = 5,
          control = control_grid(extract = raise_error)
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})

test_that("interactive logger works (bayesian, error)", {
  skip_if(
    choose_framework(workflow(), control_grid(allow_par = FALSE)) !=
      "sequential",
    "Will not catalog: parallelism is enabled"
  )
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  local_mocked_bindings(
    is_testing = function() {
      FALSE
    },
    initialize_catalog = redefer_initialize_catalog(rlang::current_env())
  )

  skip_on_cran()

  raise_error <- function(x) {
    stop("AHHhH")
  }

  set.seed(1)
  expect_snapshot(
    {
      res_grid <-
        tune_bayes(
          parsnip::nearest_neighbor("regression", "kknn", dist_power = tune()),
          Sale_Price ~ .,
          rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5),
          initial = 5,
          iter = 5,
          control = control_bayes(extract = raise_error)
        )
    },
    transform = catalog_lines
  )

  # `catalog_summary_test` written to this env via `redefer_initialize_catalog()`
  expect_snapshot(catalog_summary_test)
})

# Notes for easier reading are in `inst/loop.qmd`.
# Notes on debugging:
# 1. You can set `options(future.debug = TRUE)` to help
# 2. If you are debugging .loop_over_all_stages, use the control option
#    `allow_par = FALSE`; that will use `lapply()` so that you can see output.

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.loop_over_all_stages <- function(resamples, grid, static) {
  # Some packages may use random numbers so attach them prior to initializing
  # the RNG seed
  attach_pkgs(static$pkgs, strategy = static$strategy)

  # Initialize some objects
  seed_length <- length(resamples$.seeds[[1]])

  # If we are using last_fit() (= zero seed length), don't mess with the RNG
  # stream; otherwise set everything up.
  if (seed_length > 0) {
    orig_seed <- .Random.seed
    # Set seed within the worker process
    assign(".Random.seed", resamples$.seeds[[1]], envir = .GlobalEnv)
    resamples$.seeds <- NULL
    withr::defer(assign(".Random.seed", orig_seed, envir = .GlobalEnv))
  }

  split <- resamples$splits[[1]]
  split_labs <- resamples |>
    dplyr::select(dplyr::starts_with("id"))
  # Determine if we're using prediction-based or model-based metrics
  is_model_metrics <- identical(static$metric_type, "model")

  pred_reserve <- NULL
  metrics_reserve <- NULL
  notes <- new_note()
  extracts <- NULL

  sched <- schedule_grid(grid, static$wflow)

  config_tbl <- static$configs

  # Append data partitions here; these are the same for the duration of this function
  data_splits <- .get_data_subsets(static$wflow, split, static$split_args)
  static <- update_static(static, data_splits)

  # Now that we have data, determine the names of the outcome data. NOTE that
  # if an inline function is used (e.g. add_formula(log(mpg) ~ .)), We will
  # potentially change it later. See #1024
  static$y_name <- outcome_names(static$wflow, data = split$data)

  # ----------------------------------------------------------------------------
  # Iterate over preprocessors

  num_iterations_pre <- max(nrow(sched), 1)

  for (iter_pre in seq_len(num_iterations_pre)) {
    current_sched_pre <- sched[iter_pre, ]
    current_grid <- remove_stage(current_sched_pre)

    location <- glue::glue("preprocessor {iter_pre}/{num_iterations_pre}")

    # Note: finalize_fit_pre() will process the data used for modeling. We'll
    # also need to process the data used for prediction in the same way. That
    # will happen below (via process_prediction_data()).

    current_wflow <- .catch_and_log(
      finalize_fit_pre(static$wflow, current_sched_pre, static),
      control = static$control,
      split_labels = split_labs,
      location = location,
      notes = notes
    )

    if (is_failure(current_wflow)) {
      next
    }

    # Now we can process the data being predicted. These do not change
    # over the next two loops, so compute them here, once.
    location <- glue::glue(
      "preprocessor {iter_pre}/{num_iterations_pre} (prediction data)"
    )
    pred_data <-
      .catch_and_log(
        process_prediction_data(current_wflow, static),
        control = static$control,
        split_labels = split_labs,
        location = location,
        notes = notes
      )
    if (is_failure(pred_data)) {
      next
    }

    # Also process the calibration data (if needed for postprocessor fitting).
    has_post_estimation <- static$post_estimation
    if (has_post_estimation) {
      location <- glue::glue(
        "preprocessor {iter_pre}/{num_iterations_pre} (calibration data)"
      )
      cal_pred_data <- .catch_and_log(
        process_prediction_data(current_wflow, static, source = "cal"),
        control = static$control,
        split_labels = split_labs,
        location = location,
        notes = notes
      )
      if (is_failure(cal_pred_data)) {
        next
      }
    }

    # Update y_name in case the workflow had an inline function like `log(mpg) ~ .`
    static$y_name <- outcome_names(current_wflow)

    num_iterations_model <- max(nrow(current_sched_pre$model_stage[[1]]), 1)

    # --------------------------------------------------------------------------
    # Iterate over model parameters

    # Make a copy of the current workflow so that we can finalize it multiple
    # times, since finalize_*() functions will only update parameters whose
    # values currently are tune()
    wflow_with_fitted_pre <- current_wflow

    grid_with_pre <- current_grid

    for (iter_model in seq_len(num_iterations_model)) {
      current_sched_model <- current_sched_pre$model_stage[[1]][iter_model, ]
      current_grid <- extend_grid(grid_with_pre, current_sched_model)

      # Splice in any parameters marked for tuning and fit the model
      location <- glue::glue(
        "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model}"
      )

      current_wflow <- .catch_and_log(
        finalize_fit_model(wflow_with_fitted_pre, current_sched_model),
        control = static$control,
        split_labels = split_labs,
        location = location,
        notes = notes
      )

      if (is_failure(current_wflow)) {
        next
      }

      # ------------------------------------------------------------------------
      # For model-based metrics: compute metrics directly and handle extractions
      if (is_model_metrics) {
        current_wflow <- workflows::.fit_finalize(current_wflow)

        # Compute model-based metrics directly from workflow + data
        location <- glue::glue(
          "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model} (model metrics)"
        )

        current_metrics <- .catch_and_log(
          static$metrics(current_wflow, new_data = static$data$pred$data),
          control = static$control,
          split_labels = split_labs,
          location = location,
          notes = notes
        )

        if (!is_failure(current_metrics)) {
          current_metrics <- dplyr::bind_cols(current_grid, current_metrics)
          metrics_reserve <- dplyr::bind_rows(metrics_reserve, current_metrics)
        }

        # Extractions for model-based metrics
        if (!is.null(static$control$extract)) {
          location <- glue::glue(
            "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model} (extracts)"
          )
          elt_extract <- .catch_and_log(
            extract_details(current_wflow, static$control$extract),
            control = static$control,
            split_labels = split_labs,
            location = location,
            notes = notes
          )

          extracts <- append_extracts(
            extracts,
            elt_extract,
            current_grid,
            static
          )
        }

        # Skip prediction/post loops for model-based metrics
        next
      }

      # ------------------------------------------------------------------------
      # For prediction-based metrics: continue with predict/post loops

      has_submodel <- has_sub_param(current_sched_model$predict_stage[[1]])
      num_iterations_pred <- max(
        nrow(current_sched_model$predict_stage[[1]]),
        1
      )

      # --------------------------------------------------------------------------
      # Predict all submodels at once (if applicable), then iterate over them
      # for postprocessing

      if (has_submodel) {
        # Collect all submodel values and predict once
        sched_pred_all_submodels <- current_sched_model$predict_stage[[1]]
        sub_nm <- get_sub_param(sched_pred_all_submodels)
        grid_pred_all_submodels <- sched_pred_all_submodels[,
          sub_nm,
          drop = FALSE
        ]

        # Submodel parameters will be added in the predict stage
        grid_with_pre_model <- current_grid |>
          dplyr::select(-dplyr::all_of(sub_nm))

        location <- glue::glue(
          "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model} (predictions)"
        )
        pred_all_submodels <- .catch_and_log(
          predict_all_types(
            current_wflow,
            pred_data,
            static,
            grid_pred_all_submodels
          ),
          control = static$control,
          split_labels = split_labs,
          location = location,
          notes = notes
        )

        if (is_failure(pred_all_submodels)) {
          next
        }
        pred_all_submodels <- remove_log_notes(pred_all_submodels)

        # Also predict on calibration data for all submodels at once
        if (has_post_estimation) {
          location <- glue::glue(
            "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model} (calibration predictions)"
          )
          cal_pred_all_submodels <- .catch_and_log(
            predict_all_types(
              current_wflow,
              cal_pred_data,
              static,
              grid_pred_all_submodels,
              source = "cal"
            ),
            control = static$control,
            split_labels = split_labs,
            location = location,
            notes = notes
          )

          if (is_failure(cal_pred_all_submodels)) {
            next
          }
          cal_pred_all_submodels <- remove_log_notes(cal_pred_all_submodels)
        }
      }

      for (iter_pred in seq_len(num_iterations_pred)) {
        current_sched_pred <- current_sched_model$predict_stage[[1]][
          iter_pred,
        ]

        if (has_submodel) {
          sub_nm <- get_sub_param(current_sched_pred)
          sub_val <- current_sched_pred[[sub_nm]]

          # Add submodel param to grid
          current_grid <- extend_grid(grid_with_pre_model, current_sched_pred)

          # Filter to this submodel's predictions (already computed above)
          current_pred <- pred_all_submodels |>
            dplyr::filter(.data[[sub_nm]] == sub_val) |>
            dplyr::select(-dplyr::all_of(sub_nm))
        } else {
          location <- glue::glue(
            "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model} (predictions)"
          )
          current_pred <- .catch_and_log(
            predict_all_types(current_wflow, pred_data, static),
            control = static$control,
            split_labels = split_labs,
            location = location,
            notes = notes
          )

          if (is_failure(current_pred)) {
            next
          }
          current_pred <- remove_log_notes(current_pred)
        }

        has_post <- has_tailor(current_wflow)
        num_iterations_post <- max(nrow(current_sched_pred$post_stage[[1]]), 1)

        # if the postprocessor does not require a fit,
        # this does not cause data leakage
        current_cal_pred <- current_pred

        if (has_post_estimation) {
          if (has_submodel) {
            current_cal_pred <- cal_pred_all_submodels |>
              dplyr::filter(.data[[sub_nm]] == sub_val) |>
              dplyr::select(-dplyr::all_of(sub_nm))
          } else {
            location <- glue::glue(
              "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model} (calibration predictions)"
            )
            current_cal_pred <- .catch_and_log(
              predict_all_types(
                current_wflow,
                cal_pred_data,
                static,
                source = "cal"
              ),
              control = static$control,
              split_labels = split_labs,
              location = location,
              notes = notes
            )

            if (is_failure(current_cal_pred)) {
              next
            }
            current_cal_pred <- remove_log_notes(current_cal_pred)
          }
        }

        # ----------------------------------------------------------------------
        # Iterate over postprocessors

        # Make a copy of the current workflow so that we can finalize it multiple
        # times, since finalize_*() functions will only update parameters whose
        # values currently are tune()
        wflow_with_fitted_pre_and_model <- current_wflow

        grid_with_pre_model_pred <- current_grid

        for (iter_post in seq_len(num_iterations_post)) {
          if (has_post) {
            current_sched_post <-
              current_sched_pred$post_stage[[1]][iter_post, ]

            current_grid <- extend_grid(
              grid_with_pre_model_pred,
              current_sched_post
            )

            location <- glue::glue(
              "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model}, postprocessing {iter_pred}/{num_iterations_pred}"
            )
            current_wflow <- .catch_and_log(
              finalize_fit_post(
                wflow_with_fitted_pre_and_model,
                current_cal_pred,
                grid = current_sched_post
              ),
              control = static$control,
              split_labels = split_labs,
              location = location,
              notes = notes
            )
            if (is_failure(current_wflow)) {
              next
            }

            # to predict, use the post-processor directly rather than the
            # workflow so that we don't have to generate the model predictions
            # a second time
            post_fit <- extract_postprocessor(current_wflow, estimated = TRUE)
            post_pred <- .catch_and_log(
              predict(post_fit, current_pred),
              control = static$control,
              split_labels = split_labs,
              location = location,
              notes = notes
            )
            if (is_failure(post_pred)) {
              next
            }

            final_pred <- dplyr::bind_cols(post_pred, current_grid)
          } else {
            # No postprocessor so just use what we have
            final_pred <- dplyr::bind_cols(current_pred, current_grid)
          }

          current_wflow <- workflows::.fit_finalize(current_wflow)

          # --------------------------------------------------------------------
          # Allocate predictions to an overall object

          pred_reserve <- dplyr::bind_rows(pred_reserve, final_pred)

          # --------------------------------------------------------------------
          # Extractions

          if (!is.null(static$control$extract)) {
            location <- glue::glue(
              "preprocessor {iter_pre}/{num_iterations_pre}, model {iter_model}/{num_iterations_model} (extracts)"
            )
            elt_extract <- .catch_and_log(
              extract_details(current_wflow, static$control$extract),
              control = static$control,
              split_labels = split_labs,
              location = location,
              notes = notes
            )

            extracts <- append_extracts(
              extracts,
              elt_extract,
              current_grid,
              static
            )

            if (is_failure(elt_extract)) {
              next
            }
          }

          # Output for these loops:
          # - pred_reserve (probably not null)
          # - extracts (may be null)
          # - notes
        } # post loop
      } # predict loop
    } # model loop
  } # pre loop

  # ----------------------------------------------------------------------------
  # Compute metrics on each config and eval_time

  if (is_model_metrics) {
    # For model-based metrics, metrics were computed inline
    if (is.null(metrics_reserve)) {
      all_metrics <- NULL
    } else {
      all_metrics <- add_configs(metrics_reserve, static)
      # Reorder columns to match original grid order
      grid_cols <- names(grid)
      other_cols <- setdiff(names(all_metrics), grid_cols)
      all_metrics <- all_metrics[c(grid_cols, other_cols)]
    }
  } else {
    # For prediction-based metrics, compute from collected predictions
    if (is.null(pred_reserve)) {
      all_metrics <- NULL
    } else {
      location <- glue::glue("internal")
      all_metrics <- .catch_and_log(
        pred_reserve |>
          dplyr::group_by(!!!rlang::syms(static$param_info$id)) |>
          .estimate_metrics(
            metric = static$metrics,
            param_names = static$param_info$id,
            outcome_name = static$y_name,
            event_level = static$control$event_level,
            metrics_info = metrics_info(static$metrics)
          ) |>
          add_configs(static),
        control = static$control,
        split_labels = split_labs,
        location = location,
        notes = notes
      )
    }
  }

  if (!is.null(extracts)) {
    extracts <- add_configs(extracts, static) |>
      dplyr::relocate(.config, .after = .extracts) |>
      dplyr::relocate(names(grid))

    # Failing rows are not in the output:
    empty_extract <- purrr::map_lgl(extracts$.extracts, is.null)
    extracts <- extracts[!empty_extract, ]
  }

  # ----------------------------------------------------------------------------
  # Return the results

  return_tbl <- tibble::tibble(
    .metrics = list(all_metrics),
    .notes = list(notes)
  )

  return_tbl$outcome_names <- static$y_name

  if (!is.null(static$control$extract)) {
    if (is.null(extracts)) {
      # Everything failed; return NULL for each row
      return_tbl$.extracts <- purrr::map(1:nrow(return_tbl), \(x) NULL)
    } else {
      return_tbl <- dplyr::mutate(return_tbl, .extracts = list(extracts))
    }
  }

  return_tbl <- vctrs::vec_cbind(return_tbl, split_labs)

  if (!is_model_metrics && static$control$save_pred) {
    if (is.null(pred_reserve)) {
      # Everything failed; return NULL for each row
      return_tbl$.predictions <- purrr::map(1:nrow(return_tbl), \(x) NULL)
    } else {
      return_tbl$.predictions <-
        list(
          add_configs(pred_reserve, static) |>
            # Filter out joined rows that corresponded to a config that failed
            dplyr::filter(!is.na(.row)) |>
            reorder_pred_cols(static$y_name, static$param_info$id)
        )
    }
  }

  return_tbl
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.loop_over_all_stages2 <- function(index, resamples, grid, static) {
  .loop_over_all_stages(resamples[[index$b]], grid[[index$s]], static)
}

# ------------------------------------------------------------------------------

# This will take a grid and make a list of subgrids that should be used when
# we parallel process over grid candidates. The function will make 1-row grids
# except when there is a submodel parameter. In that case, it will create a
# subgrid that has fixed values for non-submodel parameters and the associated
# values of the submodel.
get_row_wise_grid <- function(wflow, grid) {
  param_tuned <- tune_args(wflow)$id
  submodel <- wflow |>
    hardhat::extract_spec_parsnip() |>
    get_submodel_info() |>
    dplyr::filter(has_submodel) |>
    purrr::pluck("id")

  const_param <- setdiff(param_tuned, submodel)
  const_param <- rlang::syms(const_param)

  if (length(submodel) == 0) {
    inds <- seq_len(nrow(grid))
  } else {
    grid_inds <- grid |>
      parsnip::add_rowindex() |>
      dplyr::group_nest(!!!const_param) |>
      dplyr::mutate(inds = dplyr::row_number()) |>
      tidyr::unnest(c(data)) |>
      dplyr::select(-.row)
    grid <- grid_inds[, param_tuned]
    inds <- grid_inds$inds
  }
  vctrs::vec_split(grid, inds)$val
}

# ------------------------------------------------------------------------------

add_configs <- function(x, static) {
  config_tbl <- static$configs
  if (length(static$param_info$id) > 0) {
    x <- dplyr::left_join(x, config_tbl, by = static$param_info$id)
  } else {
    x <- dplyr::bind_cols(x, config_tbl)
  }

  dplyr::arrange(x, .config)
}

test_that("convert grid to rowwise - no submodels", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")

  # Model only
  wflow_1 <- workflow(Class ~ ., dt_spec, cls_est_post)
  param_1 <- wflow_1 |> extract_parameter_set_dials()
  grid_reg_1 <- grid_regular(param_1) |> arrange(min_n)

  res_reg_1 <- tune:::get_row_wise_grid(wflow_1, grid_reg_1)
  for (i in 1:nrow(grid_reg_1)) {
    sub <- grid_reg_1[i, ]
    expect_equal(res_reg_1[[i]], sub)
  }

  # recipe only
  wflow_2 <- workflow(puromycin_tune_rec, logistic_reg(), cls_est_post)
  param_2 <- wflow_2 |> extract_parameter_set_dials()
  grid_reg_2 <- grid_regular(param_2) |> arrange(degree)

  res_reg_2 <- tune:::get_row_wise_grid(wflow_2, grid_reg_2)
  for (i in 1:nrow(grid_reg_2)) {
    sub <- grid_reg_2[i, ]
    expect_equal(res_reg_2[[i]], sub)
  }

  # post only
  wflow_3 <- workflow(Class ~ ., logistic_reg(), cls_post)
  param_3 <- wflow_3 |> extract_parameter_set_dials()
  grid_reg_3 <- grid_regular(param_3) |> arrange(cut)

  res_reg_3 <- tune:::get_row_wise_grid(wflow_3, grid_reg_3)
  for (i in 1:nrow(grid_reg_3)) {
    sub <- grid_reg_3[i, ]
    expect_equal(res_reg_3[[i]], sub)
  }

  # Recipe + Model
  wflow_4 <- workflow(puromycin_tune_rec, dt_spec, cls_est_post)
  param_4 <- wflow_4 |> extract_parameter_set_dials()
  grid_reg_4 <- grid_regular(param_4) |> arrange(degree, min_n)

  res_reg_4 <- tune:::get_row_wise_grid(wflow_4, grid_reg_4)
  for (i in 1:nrow(grid_reg_4)) {
    sub <- grid_reg_4[i, ]
    expect_equal(res_reg_4[[i]], sub)
  }

  # Recipe + Post
  wflow_5 <- workflow(puromycin_tune_rec, logistic_reg(), cls_post)
  param_5 <- wflow_5 |> extract_parameter_set_dials()
  grid_reg_5 <- grid_regular(param_5) |> arrange(degree, cut)

  res_reg_5 <- tune:::get_row_wise_grid(wflow_5, grid_reg_5)
  for (i in 1:nrow(grid_reg_5)) {
    sub <- grid_reg_5[i, ]
    expect_equal(res_reg_5[[i]], sub)
  }

  # Model + Post
  wflow_5 <- workflow(Class ~ ., dt_spec, cls_post)
  param_5 <- wflow_5 |> extract_parameter_set_dials()
  grid_reg_5 <- grid_regular(param_5) |> arrange(cut, min_n)

  res_reg_5 <- tune:::get_row_wise_grid(wflow_5, grid_reg_5)
  for (i in 1:nrow(grid_reg_5)) {
    sub <- grid_reg_5[i, ]
    expect_equal(res_reg_5[[i]], sub)
  }

  # Recipe + Model + Post
  wflow_6 <- workflow(puromycin_tune_rec, dt_spec, cls_post)
  param_6 <- wflow_6 |> extract_parameter_set_dials()
  grid_reg_6 <- grid_regular(param_6) |> arrange(cut, degree, min_n)

  res_reg_6 <- tune:::get_row_wise_grid(wflow_6, grid_reg_6)
  for (i in 1:nrow(grid_reg_6)) {
    sub <- grid_reg_6[i, ]
    expect_equal(res_reg_6[[i]], sub)
  }
})

test_that("convert grid to rowwise - submodels", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")

  # Model only
  wflow_1 <- workflow(Class ~ ., glmn_spec, reg_cal)
  param_1 <- wflow_1 |> extract_parameter_set_dials()

  grid_reg_1 <- grid_regular(param_1) |> arrange(mixture, penalty)
  res_reg_1 <- tune:::get_row_wise_grid(wflow_1, grid_reg_1)
  distinct_reg_1 <- distinct(grid_reg_1, mixture)
  for (i in 1:nrow(distinct_reg_1)) {
    sub <- grid_reg_1 |>
      dplyr::filter(mixture == distinct_reg_1$mixture[i])
    expect_equal(res_reg_1[[i]], sub)
  }

  grid_sfd_1 <- grid_space_filling(param_1) |> arrange(mixture, penalty)
  res_sfd_1 <- tune:::get_row_wise_grid(wflow_1, grid_sfd_1)
  distinct_sfd_1 <- distinct(grid_sfd_1, mixture)
  for (i in 1:nrow(distinct_sfd_1)) {
    sub <- grid_sfd_1 |>
      dplyr::filter(mixture == distinct_sfd_1$mixture[i])
    expect_equal(res_sfd_1[[i]], sub)
  }

  # Recipe + Model
  wflow_2 <- workflow(puromycin_tune_rec, glmn_spec, reg_cal)
  param_2 <- wflow_2 |> extract_parameter_set_dials()

  grid_reg_2 <- grid_regular(param_2)
  distinct_reg_2 <- distinct(grid_reg_2, mixture, degree) |>
    arrange(mixture, degree)

  res_reg_2 <- tune:::get_row_wise_grid(wflow_2, grid_reg_2)
  for (i in 1:nrow(distinct_reg_2)) {
    sub <- grid_reg_2 |>
      inner_join(distinct_reg_2[i, ], by = c("mixture", "degree"))
    expect_equal(res_reg_2[[i]], sub)
  }

  grid_sfd_2 <- grid_space_filling(param_2)
  distinct_sfd_2 <- distinct(grid_sfd_2, mixture, degree) |>
    arrange(mixture, degree)

  res_sfd_2 <- tune:::get_row_wise_grid(wflow_2, grid_sfd_2)
  for (i in 1:nrow(distinct_sfd_2)) {
    sub <- grid_sfd_2 |>
      inner_join(distinct_sfd_2[i, ], by = c("mixture", "degree"))
    expect_equal(res_sfd_2[[i]], sub)
  }

  # Model + Post
  wflow_3 <- workflow(Class ~ ., glmn_spec, reg_cal_tune)
  param_3 <- wflow_3 |> extract_parameter_set_dials()

  grid_reg_3 <- grid_regular(param_3)
  distinct_reg_3 <- distinct(grid_reg_3, method, mixture) |>
    arrange(mixture, method)

  res_reg_3 <- tune:::get_row_wise_grid(wflow_3, grid_reg_3)
  for (i in 1:nrow(distinct_reg_3)) {
    sub <- grid_reg_3 |>
      inner_join(distinct_reg_3[i, ], by = c("mixture", "method"))
    expect_equal(res_reg_3[[i]], sub)
  }

  grid_sfd_3 <- grid_space_filling(param_3)
  distinct_sfd_3 <- distinct(grid_sfd_3, method, mixture) |>
    arrange(mixture, method)

  res_sfd_3 <- tune:::get_row_wise_grid(wflow_3, grid_sfd_3)
  for (i in 1:nrow(distinct_sfd_3)) {
    sub <- grid_sfd_3 |>
      inner_join(distinct_sfd_3[i, ], by = c("mixture", "method"))
    expect_equal(res_sfd_3[[i]], sub)
  }

  # Recipe + Model + Post
  wflow_4 <- workflow(puromycin_tune_rec, glmn_spec, reg_cal_tune)
  param_4 <- wflow_4 |> extract_parameter_set_dials()

  grid_reg_4 <- grid_regular(param_4)
  distinct_reg_4 <- distinct(grid_reg_4, degree, method, mixture) |>
    arrange(mixture, degree, method)

  res_reg_4 <- tune:::get_row_wise_grid(wflow_4, grid_reg_4)
  for (i in 1:nrow(distinct_reg_4)) {
    sub <- grid_reg_4 |>
      inner_join(distinct_reg_4[i, ], by = c("mixture", "degree", "method"))
    expect_equal(res_reg_4[[i]], sub)
  }

  grid_sfd_4 <- grid_space_filling(param_4)
  distinct_sfd_4 <- distinct(grid_sfd_4, degree, method, mixture) |>
    arrange(mixture, degree, method)

  res_sfd_4 <- tune:::get_row_wise_grid(wflow_4, grid_sfd_4)
  for (i in 1:nrow(distinct_sfd_4)) {
    sub <- grid_sfd_4 |>
      inner_join(distinct_sfd_4[i, ], by = c("mixture", "degree", "method"))
    expect_equal(res_sfd_4[[i]], sub)
  }
})

#' @importFrom dplyr filter select full_join mutate bind_rows case_when vars
#' @importFrom dplyr all_of ungroup slice bind_cols pull sample_n desc anti_join
#' @importFrom dplyr distinct arrange rename mutate_if starts_with inner_join
#' @importFrom dplyr last
#' @importFrom purrr map_int
#' @importFrom rlang call2 ns_env is_quosure is_quosures quo_get_expr call_name
#' @importFrom rlang is_false eval_tidy expr sym syms env_get is_function :=
#' @importFrom rlang is_missing %||% caller_env
#' @importFrom glue glue glue_collapse
#' @importFrom dials is_unknown encode_unit
#' @importFrom stats sd qt qnorm dnorm pnorm predict model.matrix setNames
#' @importFrom stats model.matrix model.response model.frame update median
#' @importFrom yardstick rsq rmse accuracy roc_auc brier_survival brier_class
#' @importFrom yardstick weighted_interval_score
#' @importFrom tidyr unnest nest
#' @importFrom parsnip get_from_env required_pkgs
#' @importFrom recipes all_predictors all_outcomes
#' @importFrom ggplot2 ggplot aes xlab geom_point geom_errorbar facet_wrap ylab
#' @importFrom ggplot2 facet_grid geom_line aes_string aes_ scale_x_continuous
#' @importFrom cli cli_alert_danger cli_alert_info cli_alert_warning
#' @importFrom cli cli_alert_success cli_alert
#' @importFrom cli cli_inform cli_warn cli_abort qty
#' @importFrom tibble obj_sum size_sum
#' @import rlang

# ------------------------------------------------------------------------------
# Only a small number of functions in workflows.
# It is worth just importing everything.

#' @import workflows

# ------------------------------------------------------------------------------

utils::globalVariables(
  c(
    ".",
    "engine",
    "name",
    "func",
    "parsnip",
    "call_name",
    ".step",
    "call_info",
    "component",
    "component_id",
    "id",
    "control",
    ".pred",
    ".metric",
    ".estimator",
    ".estimate",
    "n",
    "note",
    "object",
    "splits",
    "grid",
    "resamples",
    ".iter",
    "mean",
    ".submodels",
    "metrics",
    "data",
    ".mean",
    ".sd",
    "iteration",
    "pkg",
    ".pred_class",
    "std_err",
    "const",
    "objective",
    "delta",
    "sd_trunc",
    "snr",
    "z",
    "..val",
    "max_val",
    "has_submodel",
    "res",
    ".extracts",
    ".metrics",
    "value",
    ".notes",
    ".loss",
    ".bound",
    ".column",
    ".totals",
    ".value",
    "direction",
    ".config",
    "Freq",
    "Prediction",
    "Truth",
    ".seed",
    ".order",
    ".iter_model",
    ".iter_preprocessor",
    ".iter_config",
    ".msg_model",
    "# resamples",
    "seed",
    "pre",
    "type",
    "rowwise",
    ".best",
    "location",
    "msg",
    "..object",
    ".eval_time",
    ".pred_linear_pred",
    ".pred_survival",
    ".pred_time",
    ".weight_censored",
    "nice_time",
    "time_metric",
    ".lower",
    ".upper",
    "i",
    "results",
    "term",
    ".alpha",
    ".method",
    "old_term",
    ".lab_pre",
    ".model",
    ".num_models",
    "model_stage",
    "predict_stage",
    "se",
    "num",
    "user",
    ".resample_weight",
    "effective_n",
    ".index",
    ".quantile_levels",
    ".pred_quantile",
    ".row_orig"
  )
)

# ------------------------------------------------------------------------------

release_bullets <- function() {
  c(
    "**Do this before checks!**. Update dependencies with `devtools::install_dev_deps()` and update the test objects via `R CMD BATCH --vanilla inst/test_objects.R`."
  )
}

# ------------------------------------------------------------------------------
# data on model prediction types

# Will predictions have one value per row and should be in a list column?
dyn_surv_types <- c("survival", "hazard")
dyn_quant_types <- "quantile"
dyn_types <- c(dyn_surv_types, dyn_quant_types)

#' @importFrom dials parameters
#' @export
dials::parameters

#' @importFrom ggplot2 autoplot
#' @export
ggplot2::autoplot

#' @importFrom generics required_pkgs
#' @export
generics::required_pkgs

#' @importFrom hardhat tune
#' @export
hardhat::tune

#' @importFrom generics tunable
#' @export
generics::tunable

#' @importFrom generics tune_args
#' @export
generics::tune_args

#' @importFrom generics min_grid
#' @export
generics::min_grid

#' @importFrom generics augment
#' @export
generics::augment

#' @importFrom rsample .get_fingerprint
#' @export
rsample::.get_fingerprint

#' @importFrom rsample int_pctl
#' @export
rsample::int_pctl

#' @importFrom hardhat extract_spec_parsnip
#' @export
hardhat::extract_spec_parsnip
#'
#' @importFrom hardhat extract_recipe
#' @export
hardhat::extract_recipe
#'
#' @importFrom hardhat extract_fit_parsnip
#' @export
hardhat::extract_fit_parsnip
#'
#' @importFrom hardhat extract_fit_engine
#' @export
hardhat::extract_fit_engine
#'
#' @importFrom hardhat extract_mold
#' @export
hardhat::extract_mold
#'
#' @importFrom hardhat extract_preprocessor
#' @export
hardhat::extract_preprocessor
#'
#' @importFrom hardhat extract_workflow
#' @export
hardhat::extract_workflow
#'
#' @importFrom hardhat extract_parameter_set_dials
#' @export
hardhat::extract_parameter_set_dials

#' Merge parameter grid values into objects
#'
#' @description
#'
#' `merge()` can be used to easily update any of the arguments in a
#'  \pkg{parsnip} model or recipe.
#'
#' @param x A recipe or model specification object.
#' @param y A data frame or a parameter grid resulting from one of the
#'  `grid_*` functions. The column names should correspond to the parameter
#'  names (or their annotations) in the object.
#' @param ... Not used but required for S3 completeness.
#'
#' @return A tibble with a column `x` that has as many rows as were in `y`.
#' @keywords internal
#' @examplesIf tune:::should_run_examples(suggests = c("xgboost", "modeldata", "splines2"))
#' library(tibble)
#' library(recipes)
#' library(parsnip)
#' library(dials)
#'
#' pca_rec <-
#'   recipe(mpg ~ ., data = mtcars) |>
#'   step_impute_knn(all_predictors(), neighbors = tune()) |>
#'   step_pca(all_predictors(), num_comp = tune())
#'
#' pca_grid <-
#'   tribble(
#'     ~neighbors, ~num_comp,
#'              1,         1,
#'              5,         1,
#'              1,         2,
#'              5,         2
#'   )
#'
#' merge(pca_rec, pca_grid)
#'
#' spline_rec <-
#'   recipe(mpg ~ ., data = mtcars) |>
#'   step_spline_natural(disp, deg_free = tune("disp df")) |>
#'   step_spline_natural(wt, deg_free = tune("wt df"))
#'
#' spline_grid <-
#'   tribble(
#'     ~"disp df", ~ "wt df",
#'     3,         3,
#'     5,         3,
#'     3,         5,
#'     5,         5
#'   )
#'
#' merge(pca_rec, pca_grid)
#'
#' data(hpc_data, package = "modeldata")
#'
#' xgb_mod <-
#'   boost_tree(trees = tune(), min_n = tune()) |>
#'   set_engine("xgboost")
#'
#' set.seed(254)
#' xgb_grid <-
#'   extract_parameter_set_dials(xgb_mod) |>
#'   finalize(hpc_data) |>
#'   grid_max_entropy(size = 3)
#'
#' merge(xgb_mod, xgb_grid)
#' @export
merge.recipe <- function(x, y, ...) {
  merger(x, y, ...)
}

#' @export
#' @rdname merge.recipe
merge.model_spec <- function(x, y, ...) {
  merger(x, y, ...)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.update_model <- function(
  grid,
  object,
  pset,
  step_id,
  nms,
  ...,
  source = "model_spec"
) {
  for (i in nms) {
    param_info <- pset |> dplyr::filter(id == i & .data$source == .env$source)
    if (nrow(param_info) > 1) {
      cli::cli_abort("Cannot update; there are too many parameters.")
    }
    if (nrow(param_info) == 1) {
      if (param_info$component_id == "main") {
        object$args[[param_info$name]] <-
          rlang::as_quosure(grid[[i]][[1]], env = rlang::empty_env())
      } else {
        object$eng_args[[param_info$name]] <-
          rlang::as_quosure(grid[[i]][[1]], env = rlang::empty_env())
      }
    }
  }
  object
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.update_recipe <- function(grid, object, pset, step_id, nms, ...) {
  for (i in nms) {
    param_info <- pset |> dplyr::filter(id == i & source == "recipe")
    if (nrow(param_info) == 1) {
      idx <- which(step_id == param_info$component_id)
      # check index
      # should use the contructor but maybe dangerous/difficult
      object$steps[[idx]][[param_info$name]] <- grid[[i]]
    }
  }
  object
}


merger <- function(x, y, ...) {
  if (!is.data.frame(y)) {
    cli::cli_abort("The second argument should be a data frame.")
  }
  pset <- hardhat::extract_parameter_set_dials(x)

  if (nrow(pset) == 0) {
    res <- purrr::map(seq_len(nrow(y)), \(.x) x)
    res <- tibble::new_tibble(list(x = res), nrow = length(res))
    return(res)
  }
  grid_name <- colnames(y)
  # We will deliberately allow `y` to lack some tunable parameters in `x`

  if (inherits(x, "recipe")) {
    updater <- .update_recipe
    step_ids <- purrr::map_chr(x$steps, "id")
  } else {
    updater <- .update_model
    step_ids <- NULL
  }

  if (!any(grid_name %in% pset$id)) {
    res <- purrr::map(seq_len(nrow(y)), \(.x) x)
    res <- tibble::new_tibble(list(x = res), nrow = length(res))
    return(res)
  }

  y |>
    dplyr::mutate(
      ..object = purrr::map(
        1:nrow(y),
        \(.x) updater(y[.x, ], x, pset, step_ids, grid_name)
      )
    ) |>
    dplyr::select(x = ..object)
}

# ------------------------------------------------------------------------------
# Helpers for .loop_over_all_stages()

# Note: in loop(), we add more elements for the outcome name(s), and the
# data partitions
#' @export
#' @keywords internal
#' @rdname empty_ellipses
.make_static <- function(
  workflow,
  param_info,
  grid,
  metrics,
  eval_time,
  split_args,
  control,
  pkgs = "tune",
  strategy = "sequential",
  data = list(fit = NULL, pred = NULL, cal = NULL)
) {
  # check inputs
  if (!inherits(workflow, "workflow")) {
    cli::cli_abort("{.arg workflow} should be a {.cls workflow} object.")
  }
  if (!inherits(param_info, "parameters")) {
    cli::cli_abort("{.arg param_info} should be a {.cls parameters} object.")
  }
  if (!inherits(metrics, "metric_set")) {
    cli::cli_abort("{.arg metrics} should be a {.cls metric_set} object.")
  }
  if (!(inherits(eval_time, "numeric") || is.null(eval_time))) {
    cli::cli_abort("{.arg eval_time} should be a numeric vector.")
  }

  configs <- .get_config_key(grid, workflow)

  metric_type <- get_metric_type(metrics)

  if (metric_type == "model") {
    # tidyclust model based metrics
    metric_info <- tibble::tibble(
      metric = character(0),
      class = character(0),
      direction = character(0)
    )
    pred_types <- character(0)
  } else {
    metric_info <- tibble::as_tibble(metrics)
    pred_types <- .determine_pred_types(workflow, metrics)
  }

  list(
    wflow = workflow,
    param_info = param_info,
    configs = configs,
    post_estimation = workflows::.workflow_postprocessor_requires_fit(
      workflow
    ),
    metrics = metrics,
    metric_info = metric_info,
    metric_type = metric_type,
    pred_types = pred_types,
    eval_time = eval_time,
    split_args = split_args,
    control = control,
    pkgs = pkgs,
    strategy = strategy,
    data = data
  )
}

get_metric_type <- function(metrics) {
  if (inherits(metrics, "cluster_metric_set")) {
    "model"
  } else {
    "prediction"
  }
}

check_static_data <- function(x, elem = "fit") {
  if (is.null(x)) {
    return(x)
  }

  nms <- sort(names(x))
  if (!identical(nms, c("data", "ind"))) {
    cli::cli_abort(
      "{.arg data_*} arguments should have names {.val data} and
                   {.val ind}, not {.val {nms}} in the {.field {elem}} slot."
    )
  }

  if (!is.integer(x[["ind"]])) {
    cli::cli_abort(
      "Element {.arg ind} should be an integer in the
                   {.field {elem}} slot."
    )
  }

  if (!is.data.frame(x[["data"]])) {
    cli::cli_abort(
      "Element {.arg data} should be a tibble in the {.field {elem}} slot."
    )
    if (!tibble::is_tibble(x[["data"]])) {
      x[["data"]] <- tibble::as_tibble(x[["data"]])
    }
  }

  x
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.get_data_subsets <- function(wflow, split, split_args = NULL) {
  fit_lst <- pred_lst <- cal_lst <- list(data = NULL, ind = NULL)
  pred_lst$data <- rsample::assessment(split)
  pred_lst$ind <- as.integer(split, data = "assessment")
  if (workflows::.workflow_postprocessor_requires_fit(wflow)) {
    # if the workflow has a postprocessor that needs training (i.e. calibration),
    # further split the analysis data into an "inner" analysis and
    # assessment set.
    # * the preprocessor and model (excluding the post-processor) are fitted
    #   on `analysis(internal_calibration_split(split))`, the inner analysis set (just
    #   referred to as analysis)
    # * that model generates predictions on `calibration(internal_calibration_split(split))`,
    #   the calibration set
    # * the post-processor is trained on the predictions generated from the
    #   calibration set
    # * the model (including the post-processor) generates predictions on the
    #   assessment set and those predictions are assessed with performance metrics
    split <- rsample::internal_calibration_split(split, split_args = split_args)

    cal_lst$ind <- as.integer(split, data = "assessment")
    cal_lst$data <- vctrs::vec_slice(split$data, cal_lst$ind)
  } else {
    cal_lst <- NULL
  }

  fit_lst$data <- rsample::analysis(split)
  fit_lst$ind <- as.integer(split, data = "analysis")
  list(fit = fit_lst, pred = pred_lst, cal = cal_lst)
}

update_static <- function(x, data) {
  check_static_data(data$fit)
  check_static_data(data$pred)
  check_static_data(data$cal)
  x$data <- data
  x
}

# ------------------------------------------------------------------------------
# For stages

remove_stage <- function(x) {
  stages <- c("model_stage", "predict_stage", "post_stage")
  x[, !(names(x) %in% stages)]
}

has_pre_param <- function(x) {
  any(names(x) != "model_stage")
}

has_mod_param <- function(x) {
  any(names(x) != "predict_stage")
}

# ------------------------------------------------------------------------------

# This is run on a `predict_stage` column:
has_sub_param <- function(x) {
  not_post_list <- names(x) != "post_stage"
  has_param_col <- any(not_post_list)
  if (!has_param_col) {
    return(FALSE)
  }
  param_col_nm <- names(x)[not_post_list]
  param_col <- x[[param_col_nm]]
  two_plus_vals <- length(param_col) > 1
  two_plus_vals
}

# This is run on a `predict_stage` column:
get_sub_param <- function(x) {
  not_post_list <- names(x) != "post_stage"
  names(x)[not_post_list]
}

# ------------------------------------------------------------------------------
# from workflows
# nocov start
has_tailor <- function(x) {
  "tailor" %in% names(x$post$actions)
}
# nocov end

# ------------------------------------------------------------------------------
# Prediction and postprocessing

finalize_fit_post <- function(
  wflow_current,
  predictions_calibration,
  grid = NULL
) {
  if (is.null(grid)) {
    grid <- dplyr::tibble()
  }

  post_obj <- hardhat::extract_postprocessor(wflow_current) |>
    finalize_tailor(grid)
  wflow_current <- set_workflow_tailor(wflow_current, post_obj)

  fit_post_from_predictions(wflow_current, predictions_calibration)
}

# This mimics `workflows::.fit_post()`, except it takes predictions instead of the
# unprocessed calibration data.
# We do that because we want to predict for all submodels at once
# (as we do for the assessment set).
fit_post_from_predictions <- function(wflow, predictions_calibration) {
  tailor_obj <- hardhat::extract_postprocessor(wflow, estimated = FALSE)
  outcome_names <- names(hardhat::extract_mold(wflow)$outcomes)

  post_fit <- tailor::fit(
    object = tailor_obj,
    .data = predictions_calibration,
    outcome = outcome_names,
    estimate = tidyselect::any_of(c(".pred", ".pred_class")),
    probabilities = c(
      tidyselect::contains(".pred_"),
      -tidyselect::matches("^\\.pred$|^\\.pred_class$")
    )
  )

  wflow$post$fit <- post_fit
  wflow
}

# ------------------------------------------------------------------------------

predict_all_types <- function(
  wflow_fit,
  processed_data,
  static,
  submodel_grid = NULL,
  source = c("pred", "cal")
) {
  source <- rlang::arg_match(source)
  .data <- static$data[[source]]$data
  .ind <- static$data[[source]]$ind

  model_fit <- wflow_fit |> hardhat::extract_fit_parsnip()

  sub_param <- names(submodel_grid)

  # Convert argument names to parsnip format see #1011
  submodel_grid <- engine_to_parsnip(static$wflow, submodel_grid)

  pred <- NULL
  for (type_iter in static$pred_types) {
    tmp_res <- predict_wrapper(
      model = model_fit,
      new_data = processed_data$predictors,
      type = type_iter,
      eval_time = static$eval_time,
      subgrid = submodel_grid
    )
    tmp_res$.row <- .ind

    # predict_wrapper() is designed to predict all submodels at once; we get a
    # list column back called .pred with a single row. Collapse that and move
    # the submodel column.
    if (length(sub_param) > 0) {
      tmp_res <- tidyr::unnest(tmp_res, cols = c(.pred))

      # For censored regression (and eventually quantile regression),
      # dynamic predictions will generate a list column that can have multiple
      # predictions per training set row so those end up in a list column.
      # Since we unnested, we need to re-nest to get back to the original
      # format.
      if (type_iter %in% dyn_types) {
        tmp_res <-
          tmp_res |>
          tidyr::nest(.pred = c(-dplyr::all_of(sub_param), -.row))
      }
    }

    # Now go back to engine names
    tmp_res <- parsnip_to_engine(static$wflow, tmp_res)

    if (is.null(pred)) {
      pred <- tmp_res
    } else {
      pred <- dplyr::full_join(pred, tmp_res, by = c(sub_param, ".row"))
    }
  }

  pred <- pred |>
    dplyr::full_join(processed_data$outcomes, by = ".row")

  # Add implicitly grouped metric data, if applicable
  metrics_by <- get_metrics_by(static$metrics)
  if (has_metrics_by(metrics_by)) {
    .data$.row <- .ind
    pred <- dplyr::full_join(pred, .data[c(metrics_by, ".row")], by = ".row")
  }

  # Add case weights (if needed)
  if (has_case_weights(wflow_fit)) {
    case_weights <- extract_case_weights(.data, wflow_fit)
    if (.use_case_weights_with_yardstick(case_weights[[1]])) {
      case_weights <- dplyr::mutate(case_weights, .row = .ind)
      pred <- dplyr::full_join(pred, case_weights, by = ".row")
    }
  }

  pred <- maybe_add_ipcw(pred, model_fit, static$pred_types)

  pred
}

# ------------------------------------------------------------------------------
# Fitting/training functions

finalize_fit_pre <- function(wflow_current, grid, static) {
  pre_proc <- hardhat::extract_preprocessor(wflow_current)

  if (inherits(pre_proc, "recipe")) {
    grid <- remove_stage(grid)
    pre_proc_param <- hardhat::extract_parameter_set_dials(pre_proc)
    pre_proc_id <- pre_proc_param$id

    if (length(pre_proc_id) > 0) {
      grid <- grid[, pre_proc_id]
      pre_proc <- finalize_recipe(pre_proc, grid)
      wflow_current <- .set_workflow_recipe(wflow_current, pre_proc)
    }
  }
  workflows::.fit_pre(wflow_current, static$data$fit$data)
}

finalize_fit_model <- function(wflow_current, grid) {
  mod_spec <- hardhat::extract_spec_parsnip(wflow_current)

  grid <- remove_stage(grid)
  mod_param <- hardhat::extract_parameter_set_dials(mod_spec)
  mod_id <- mod_param$id

  if (length(mod_id) > 0) {
    grid <- grid[, mod_id]
    mod_spec <- finalize_model(mod_spec, grid)
    wflow_current <- .set_workflow_spec(wflow_current, mod_spec)
  }

  # .catch_and_log_fit()
  .fit_model(wflow_current, workflows::control_workflow())
}

# ------------------------------------------------------------------------------
# To call after the model is set and we loop over predict and/or post parameters
# See #1128
process_prediction_data <- function(
  wflow_fit,
  static,
  source = c("pred", "cal")
) {
  source <- rlang::arg_match(source)
  .data <- static$data[[source]]$data
  .ind <- static$data[[source]]$ind

  processed_data <- forge_from_workflow(.data, wflow_fit)

  if (nrow(processed_data$predictors) != nrow(.data)) {
    if (source == "pred") {
      set_name <- "assessment"
    } else {
      set_name <- "calibration"
    }
    msg <- "Some {set_name} set rows are not available at prediction time."
    if (.has_preprocessor_recipe(wflow_fit)) {
      msg <- c(
        msg,
        i = "Consider using {.code skip = TRUE} on any recipe steps that
             remove rows to avoid calling them on the {set_name} set."
      )
    } else {
      msg <- c(msg, i = "Did your preprocessing steps filter or remove rows?")
    }
    cli::cli_abort(msg)
  }

  processed_data$outcomes <- processed_data$outcomes |>
    dplyr::mutate(.row = .ind)
  processed_data
}


# ------------------------------------------------------------------------------
# Misc functions

extend_grid <- function(...) {
  list(...) |> purrr::map(remove_stage) |> purrr::list_cbind()
}

# ------------------------------------------------------------------------------
# Add .config to grid

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.get_config_key <- function(grid, wflow) {
  info <- tune_args(wflow)
  key <- grid

  only_param <- setdiff(info$id, names(grid))
  if (length(only_param) > 0) {
    cli::cli_abort(
      "Some parameters are tagged for tuning but are not in the grid:
      {.arg {only_param}}",
      call = NULL
    )
  }

  only_grid <- setdiff(names(grid), info$id)
  if (length(only_grid) > 0) {
    cli::cli_abort(
      "Some parameters are in the grid but are not tagged for tuning:
      {.arg {only_grid}}",
      call = NULL
    )
  }

  pre_param <- info$id[info$source == "recipe"]
  if (length(pre_param) > 0) {
    key <- make_config_labs(grid, pre_param) |>
      dplyr::full_join(key, by = pre_param)
  } else {
    key <- key |>
      dplyr::mutate(pre = "pre0")
  }

  model_source <- get_model_source(wflow)
  mod_param <- info$id[info$source == model_source]
  if (length(mod_param) > 0) {
    key <- make_config_labs(grid, mod_param, "mod") |>
      dplyr::full_join(key, by = mod_param)
  } else {
    key <- key |>
      dplyr::mutate(mod = "mod0")
  }

  post_param <- info$id[info$source == "tailor"]
  if (length(post_param) > 0) {
    key <- make_config_labs(grid, post_param, "post") |>
      dplyr::full_join(key, by = post_param)
  } else {
    key <- key |>
      dplyr::mutate(post = "post0")
  }

  # in the case of resampling without tuning, grid and thus key are 0-row tibbles
  if (nrow(key) < 1) {
    key <- dplyr::tibble(
      pre = "pre0",
      mod = "mod0",
      post = "post0"
    )
  }

  key$.config <- paste(key$pre, key$mod, key$post, sep = "_")
  key$.config <- gsub("_$", "", key$.config)
  key |>
    dplyr::arrange(.config) |>
    dplyr::select(dplyr::all_of(info$id), .config)
}

make_config_labs <- function(grid, param, val = "pre") {
  res <- grid |>
    dplyr::select(dplyr::all_of(param)) |>
    dplyr::distinct() |>
    dplyr::arrange(!!!rlang::syms(param)) |>
    dplyr::mutate(
      num = format(dplyr::row_number()),
      num = gsub(" ", "0", num),
      {{ val }} := paste0(val, num)
    ) |>
    dplyr::select(-num)

  res
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.determine_pred_types <- function(wflow, metrics) {
  model_mode <- extract_spec_parsnip(wflow)$mode

  # For clustering (partition mode), no predictions needed
  if (model_mode == "partition") {
    return(character(0))
  }

  pred_types <- unique(metrics_info(metrics)$type)
  if (has_tailor(wflow)) {
    post <- extract_postprocessor(wflow)
    post_out <- purrr::map(post$adjustments, "outputs")
    post_in <- purrr::map(post$adjustments, "inputs")
    post_types <- unlist(c(post_out, post_in))
    post_types[grepl("probability", post_types)] <- "prob"
    post_cls <- purrr::map(post$adjustments, class)
    post_cls <- unlist(post_cls)
    if (any(post_cls == "probability_calibration")) {
      post_types <- c(post_types, "class", "prob")
    }
    post_cls <- unique(post_cls)
    pred_types <- unique(c(pred_types, post_types))
  }

  if (any(pred_types == "everything")) {
    if (model_mode == "regression") {
      pred_types <- c(pred_types, "numeric")
    } else if (model_mode == "classification") {
      pred_types <- c(pred_types, "class", "prob")
    } else if (model_mode == "censored regression") {
      pred_types <- c(
        pred_types,
        "static_survival_metric",
        "dynamic_survival_metric"
      )
    } else {
      cli::cli_abort(
        "No prediction types are known for mode {.val model_mode}."
      )
    }

    pred_types <- pred_types[pred_types != "everything"]
  }

  sort(unique(pred_types))
}

reorder_pred_cols <- function(x, outcome = character(0), param = character(0)) {
  x |>
    dplyr::relocate(
      # Outcome first
      dplyr::all_of(outcome),
      # Dynamic columns next
      dplyr::any_of("^\\.eval_time"),
      # dplyr::any_of(".quantile_level"),     # placeholder for future value
      # Prediction columns
      dplyr::matches("^\\.pred_time$"),
      dplyr::matches("^\\.pred$"),
      dplyr::matches("^\\.pred_class$"),
      dplyr::matches("^\\.pred_."),
      dplyr::matches("^\\.resid"),
      # Row indicator
      dplyr::any_of("^\\.row"),
      # Resample indicator(s)
      dplyr::matches("^id$"),
      dplyr::matches("^id[1-9]$"),
      .before = dplyr::everything()
    ) |>
    # Put any tuning parameters at the end, .config at the very end
    dplyr::relocate(
      dplyr::all_of(param),
      dplyr::any_of(".config"),
      .after = dplyr::everything()
    )
}


engine_to_parsnip <- function(wflow, grid) {
  if (is.null(grid)) {
    return(grid)
  }
  grid_nm <- names(grid)
  key <- parsnip::.model_param_name_key(wflow) |>
    dplyr::filter(user != parsnip & user %in% grid_nm) |>
    dplyr::select(-engine)

  if (nrow(key) == 0) {
    return(grid)
  }
  nm_lst <- key$user
  names(nm_lst) <- key$parsnip
  dplyr::rename(grid, dplyr::all_of(nm_lst))
}

parsnip_to_engine <- function(wflow, grid) {
  grid_nm <- names(grid)
  key <- parsnip::.model_param_name_key(wflow) |>
    dplyr::filter(user != parsnip & parsnip %in% grid_nm) |>
    dplyr::select(-engine)

  if (nrow(key) == 0) {
    return(grid)
  }
  nm_lst <- key$parsnip
  names(nm_lst) <- key$user
  dplyr::rename(grid, dplyr::all_of(nm_lst))
}

# ------------------------------------------------------------------------------

attach_pkgs <- function(pkgs, strategy = "sequential") {
  sshh_load <- purrr::quietly(library)

  if (length(pkgs) > 0 & strategy != "sequential") {
    # In parallel, load it all
    pkgs_res <- purrr::map(pkgs, ~ sshh_load(.x, character.only = TRUE))
  }

  invisible(pkgs)
}

extract_details <- function(object, extractor) {
  if (is.null(extractor)) {
    return(list())
  }
  extractor(object)
}

append_extracts <- function(extracts, elt_extract, current_grid, static) {
  if (is.null(extracts)) {
    extracts <- tibble::tibble(.extracts = list(1))
    if (nrow(static$param_info) > 0) {
      extracts <- tibble::add_column(current_grid, .extracts = list(1))
    }
    extracts <- extracts[integer(), ]
  }

  if (nrow(static$param_info) > 0) {
    extracts <- tibble::add_row(
      extracts,
      tibble::add_column(current_grid, .extracts = list(elt_extract))
    )
  } else {
    extracts <- tibble::add_row(
      extracts,
      tibble::tibble(.extracts = list(elt_extract))
    )
  }

  extracts
}

# These tests reference results generated by tune 1.3.0. The code to generate
# them (and the results) are found in the `inst` directory.

test_that("verifying .loop_over_all_stages, no submodels, no post estimation or tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  load(system.file(
    "regression_tests",
    "simple_example.RData",
    package = "tune"
  ))

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_regression(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  mod <- nearest_neighbor(neighbors = 11, weight_func = tune()) |>
    set_mode("regression")
  wflow <- workflow(outcome ~ ., mod, reg_post)
  grd <- wflow |> extract_parameter_set_dials() |> grid_regular()

  static_1 <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = metric_set(rmse, rsq),
    eval_time = NULL,
    split_args = rs_args,
    control = control_grid()
  )

  data_1 <- tune:::.get_data_subsets(wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "outcome"

  simple_res <- tune:::.loop_over_all_stages(rs_iter, grd, static_1)
  expect_named(simple_res, c(".metrics", ".notes", "outcome_names", "id"))
  expect_true(nrow(simple_res) == 1)

  # A linear transformation so R^2 should be the same
  exp_rsq_mtr <-
    simple_metrics |>
    dplyr::filter(id == "Fold01" & .metric == "rsq") |>
    dplyr::select(all_of(names(simple_res$.metrics[[1]]))) |>
    arrange(weight_func) |>
    dplyr::select(-.config)

  obs_rsq_simple_mtr <-
    simple_res$.metrics[[1]] |>
    dplyr::filter(.metric == "rsq") |>
    arrange(weight_func) |>
    dplyr::select(-.config)

  expect_equal(obs_rsq_simple_mtr, exp_rsq_mtr)

  # rmse should be bad
  exp_rmse_mtr <-
    simple_metrics |>
    dplyr::filter(id == "Fold01" & .metric == "rmse") |>
    dplyr::select(all_of(names(simple_res$.metrics[[1]]))) |>
    arrange(weight_func) |>
    dplyr::select(expected = .estimate, weight_func)

  obs_rmse_simple_mtr <-
    simple_res$.metrics[[1]] |>
    dplyr::filter(.metric == "rmse") |>
    arrange(weight_func) |>
    dplyr::select(obs = .estimate, weight_func)

  rmse_diff <-
    full_join(exp_rmse_mtr, obs_rmse_simple_mtr, by = "weight_func") |>
    mutate(diffs = obs - expected)

  expect_true(all(rmse_diff$diffs > 0))
})

test_that("verifying .loop_over_all_stages, submodels, no post estimation or tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  load(system.file(
    "regression_tests",
    "submodel_example.RData",
    package = "tune"
  ))

  ctrl <- tune::control_grid()

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_regression(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  rec <- recipe(outcome ~ ., data = dat) |>
    step_pca(all_numeric_predictors(), num_comp = tune())

  mod <- nearest_neighbor(neighbors = tune("k"), weight_func = tune()) |>
    set_mode("regression")

  submodel_wflow <- workflow(rec, mod, reg_post)

  # fmt: skip
  submodel_grid <-
    tibble::tribble(
      ~k,   ~weight_func, ~num_comp,
      9L,  "rectangular",        2L,
      14L,  "rectangular",        2L,
      20L,  "rectangular",        2L,
      4L,   "triangular",        2L,
      9L,   "triangular",        2L,
      14L,   "triangular",        2L,
      20L,   "triangular",        2L,
      4L, "epanechnikov",        2L,
      9L, "epanechnikov",        2L,
      14L, "epanechnikov",        2L,
      20L, "epanechnikov",        2L,
      4L,  "rectangular",       10L,
      9L,  "rectangular",       10L,
      14L,  "rectangular",       10L,
      20L,  "rectangular",       10L,
      4L,   "triangular",       10L,
      9L,   "triangular",       10L,
      14L,   "triangular",       10L,
      20L,   "triangular",       10L,
      4L, "epanechnikov",       10L,
      9L, "epanechnikov",       10L,
      14L, "epanechnikov",       10L,
      20L, "epanechnikov",       10L
    )

  # ------------------------------------------------------------------------------

  static_1 <- tune:::.make_static(
    submodel_wflow,
    param_info = submodel_wflow |> extract_parameter_set_dials(),
    grid = submodel_grid,
    metrics = metric_set(rmse, rsq),
    eval_time = NULL,
    split_args = rs_args,
    control = ctrl
  )

  data_1 <- tune:::.get_data_subsets(submodel_wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "outcome"

  submodel_res <- tune:::.loop_over_all_stages(rs_iter, submodel_grid, static_1)
  expect_named(submodel_res, c(".metrics", ".notes", "outcome_names", "id"))
  expect_true(nrow(submodel_res) == 1)

  # A linear transformation so R^2 should be the same
  exp_rsq_mtr <-
    submodel_metrics |>
    dplyr::filter(id == "Fold01" & .metric == "rsq") |>
    dplyr::select(all_of(names(submodel_res$.metrics[[1]]))) |>
    arrange(weight_func, k, num_comp) |>
    dplyr::select(-.config)

  obs_rsq_submodel_mtr <-
    submodel_res$.metrics[[1]] |>
    dplyr::filter(.metric == "rsq") |>
    arrange(weight_func, k, num_comp) |>
    dplyr::select(-.config)

  expect_equal(obs_rsq_submodel_mtr, exp_rsq_mtr)

  # rmse should be bad
  exp_rmse_mtr <-
    submodel_metrics |>
    dplyr::filter(id == "Fold01" & .metric == "rmse") |>
    dplyr::select(all_of(names(submodel_res$.metrics[[1]]))) |>
    arrange(weight_func, k, num_comp) |>
    dplyr::select(expected = .estimate, weight_func, k, num_comp)

  obs_rmse_submodel_mtr <-
    submodel_res$.metrics[[1]] |>
    dplyr::filter(.metric == "rmse") |>
    arrange(weight_func, k, num_comp) |>
    dplyr::select(obs = .estimate, weight_func, k, num_comp)

  rmse_diff <-
    full_join(
      exp_rmse_mtr,
      obs_rmse_submodel_mtr,
      by = c("weight_func", "k", "num_comp")
    ) |>
    mutate(diffs = obs - expected)

  expect_true(all(rmse_diff$diffs > 0))
})

test_that("verifying .loop_over_all_stages, submodels only, no post estimation or tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")

  load(system.file(
    "regression_tests",
    "submodel_only_example.RData",
    package = "tune"
  ))

  ctrl <- tune::control_grid()

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_classification(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  mod <- nearest_neighbor(neighbors = tune(), weight_func = "triangular") |>
    set_mode("classification")

  submodel_only_wflow <- workflow(class ~ ., mod, cls_tenth)

  submodel_only_grid <- tibble(neighbors = 3:10)

  # ------------------------------------------------------------------------------

  static_1 <- tune:::.make_static(
    submodel_only_wflow,
    param_info = submodel_only_wflow |> extract_parameter_set_dials(),
    grid = submodel_only_grid,
    metrics = metric_set(accuracy, roc_auc, brier_class),
    eval_time = NULL,
    split_args = rs_args,
    control = ctrl
  )

  data_1 <- tune:::.get_data_subsets(submodel_only_wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "class"

  submodel_only_res <- tune:::.loop_over_all_stages(
    rs_iter,
    submodel_only_grid,
    static_1
  )
  expect_named(
    submodel_only_res,
    c(".metrics", ".notes", "outcome_names", "id")
  )
  expect_true(nrow(submodel_only_res) == 1)

  # Thresholding so accuracy should be different and Brier and ROC should be
  # unchanged
  exp_prob_mtr <-
    submodel_only_metrics |>
    dplyr::filter(id == "Fold01" & .metric != "accuracy") |>
    dplyr::select(all_of(names(submodel_only_res$.metrics[[1]]))) |>
    arrange(neighbors, .metric) |>
    dplyr::select(-.config)

  obs_prob_submodel_only_mtr <-
    submodel_only_res$.metrics[[1]] |>
    dplyr::filter(.metric != "accuracy") |>
    arrange(neighbors, .metric) |>
    dplyr::select(-.config)

  expect_equal(obs_prob_submodel_only_mtr, exp_prob_mtr)

  # accuracy should be worse
  exp_acc_mtr <-
    submodel_only_metrics |>
    dplyr::filter(id == "Fold01" & .metric == "accuracy") |>
    dplyr::select(all_of(names(submodel_only_res$.metrics[[1]]))) |>
    arrange(neighbors, .metric) |>
    dplyr::select(expected = .estimate, neighbors, .metric)

  obs_acc_submodel_only_mtr <-
    submodel_only_res$.metrics[[1]] |>
    dplyr::filter(.metric == "accuracy") |>
    arrange(neighbors, .metric) |>
    dplyr::select(obs = .estimate, neighbors, .metric)

  rmse_diff <-
    full_join(
      exp_acc_mtr,
      obs_acc_submodel_only_mtr,
      by = c("neighbors", ".metric")
    ) |>
    mutate(diffs = obs - expected)

  expect_true(all(rmse_diff$diffs < 0))
})

# These tests reference results generated by tune 1.3.0. The code to generate
# them (and the results) are found in the `inst` directory.

test_that("verifying .loop_over_all_stages, no submodels, post estimation without tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  load(system.file(
    "regression_tests",
    "simple_example.RData",
    package = "tune"
  ))

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_regression(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  mod <- nearest_neighbor(neighbors = 11, weight_func = tune()) |>
    set_mode("regression")
  wflow <- workflow(outcome ~ ., mod, reg_cal)
  grd <- wflow |> extract_parameter_set_dials() |> grid_regular()

  static_1 <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = metric_set(rmse, rsq),
    eval_time = NULL,
    split_args = rs_args,
    control = control_grid(save_pred = TRUE)
  )

  data_1 <- tune:::.get_data_subsets(wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "outcome"

  simple_res <- tune:::.loop_over_all_stages(rs_iter, grd, static_1)
  expect_true(!is.null(simple_res$.metrics[[1]]))
  expect_named(
    simple_res,
    c(".metrics", ".notes", "outcome_names", "id", ".predictions")
  )
  expect_true(nrow(simple_res) == 1)
  expect_equal(
    nrow(simple_res$.predictions[[1]]),
    nrow(data_1$pred$data) * nrow(grd)
  )

  # TODO more tests can be added when calibration method = "none" is implemented
})

test_that("verifying .loop_over_all_stages, submodels, post estimation without tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  load(system.file(
    "regression_tests",
    "submodel_example.RData",
    package = "tune"
  ))

  ctrl <- tune::control_grid(save_pred = TRUE)

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_regression(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  rec <- recipe(outcome ~ ., data = dat) |>
    step_pca(all_numeric_predictors(), num_comp = tune())

  mod <- nearest_neighbor(neighbors = tune("k"), weight_func = tune()) |>
    set_mode("regression")

  submodel_wflow <- workflow(rec, mod, reg_cal)

  # fmt: skip
  submodel_grid <-
    tibble::tribble(
      ~k,   ~weight_func, ~num_comp,
      9L,  "rectangular",        2L,
      14L,  "rectangular",        2L,
      20L,  "rectangular",        2L,
      4L,   "triangular",        2L,
      9L,   "triangular",        2L,
      14L,   "triangular",        2L,
      20L,   "triangular",        2L,
      4L, "epanechnikov",        2L,
      9L, "epanechnikov",        2L,
      14L, "epanechnikov",        2L,
      20L, "epanechnikov",        2L,
      4L,  "rectangular",       10L,
      9L,  "rectangular",       10L,
      14L,  "rectangular",       10L,
      20L,  "rectangular",       10L,
      4L,   "triangular",       10L,
      9L,   "triangular",       10L,
      14L,   "triangular",       10L,
      20L,   "triangular",       10L,
      4L, "epanechnikov",       10L,
      9L, "epanechnikov",       10L,
      14L, "epanechnikov",       10L,
      20L, "epanechnikov",       10L
    )

  # ------------------------------------------------------------------------------

  static_1 <- tune:::.make_static(
    submodel_wflow,
    param_info = submodel_wflow |> extract_parameter_set_dials(),
    grid = submodel_grid,
    metrics = metric_set(rmse, rsq),
    eval_time = NULL,
    split_args = rs_args,
    control = ctrl
  )

  data_1 <- tune:::.get_data_subsets(submodel_wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "outcome"

  submodel_res <- tune:::.loop_over_all_stages(rs_iter, submodel_grid, static_1)
  expect_true(!is.null(submodel_res$.metrics[[1]]))
  expect_named(
    submodel_res,
    c(".metrics", ".notes", "outcome_names", "id", ".predictions")
  )
  expect_true(nrow(submodel_res) == 1)
  expect_equal(
    nrow(submodel_res$.predictions[[1]]),
    nrow(data_1$pred$data) * nrow(submodel_grid)
  )
})

test_that("verifying .loop_over_all_stages, submodels only, post estimation without tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  load(system.file(
    "regression_tests",
    "submodel_only_example.RData",
    package = "tune"
  ))

  ctrl <- tune::control_grid(save_pred = TRUE)

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_classification(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  mod <- nearest_neighbor(neighbors = tune(), weight_func = "triangular") |>
    set_mode("classification")

  submodel_only_wflow <- workflow(class ~ ., mod, cls_est_post)

  submodel_only_grid <- tibble(neighbors = 3:10)

  # ------------------------------------------------------------------------------

  static_1 <- tune:::.make_static(
    submodel_only_wflow,
    param_info = submodel_only_wflow |> extract_parameter_set_dials(),
    grid = submodel_only_grid,
    metrics = metric_set(accuracy, roc_auc, brier_class),
    eval_time = NULL,
    split_args = rs_args,
    control = ctrl
  )

  data_1 <- tune:::.get_data_subsets(submodel_only_wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "class"

  submodel_only_res <- tune:::.loop_over_all_stages(
    rs_iter,
    submodel_only_grid,
    static_1
  )
  expect_true(!is.null(submodel_only_res$.metrics[[1]]))
  expect_named(
    submodel_only_res,
    c(".metrics", ".notes", "outcome_names", "id", ".predictions")
  )
  expect_true(nrow(submodel_only_res) == 1)
  expect_equal(
    nrow(submodel_only_res$.predictions[[1]]),
    nrow(data_1$pred$data) * nrow(submodel_only_grid)
  )

  # TODO more tests can be added when calibration method = "none" is implemented
})

test_that("submodel calibration varies per submodel value (#1144)", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("glmnet")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  set.seed(1)
  dat <- modeldata::sim_regression(500)
  rs <- vfold_cv(dat, v = 3)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(.seeds = get_parallel_seeds(1))

  wflow <- workflow(outcome ~ ., glmn_spec, reg_cal)
  grd <- tibble(penalty = c(0.001, 0.1), mixture = c(1, 1))

  extract_cal_coef <- function(wf) {
    postprocessor <- extract_postprocessor(wf, estimated = TRUE)
    postprocessor$adjustments[[1]]$results$fit$estimates[[1]]$estimate |>
      coefficients()
  }

  static <- .make_static(
    wflow,
    param_info = extract_parameter_set_dials(wflow),
    grid = grd,
    metrics = metric_set(rmse),
    eval_time = NULL,
    split_args = rs_args,
    control = control_grid(
      save_pred = TRUE,
      extract = extract_cal_coef
    )
  )

  data_splits <- .get_data_subsets(wflow, rs_split, rs_args)
  static <- update_static(static, data_splits)
  static$y_name <- "outcome"

  res <- .loop_over_all_stages(rs_iter, grd, static)
  extracts <- res$.extracts[[1]]

  loop_cal_p1 <- extracts$.extracts[[1]]
  loop_cal_p2 <- extracts$.extracts[[2]]

  # The two calibration models should differ because different penalty values
  # produce different predictions on the calibration set.
  expect_false(identical(loop_cal_p1, loop_cal_p2))

  withr::local_preserve_seed()
  assign(".Random.seed", rs_iter$.seeds[[1]], envir = .GlobalEnv)
  inner_split <- rsample::internal_calibration_split(
    rs_split,
    split_args = rs_args
  )

  ref_wf <- workflow(
    outcome ~ .,
    linear_reg(penalty = 0.001, mixture = 1, engine = "glmnet"),
    reg_cal
  ) |>
    fit(
      data = analysis(inner_split),
      data_calibration = calibration(inner_split)
    )

  ref_cal <- extract_cal_coef(ref_wf)

  expect_equal(loop_cal_p1, ref_cal)
})

test_that("set configurations - single element", {
  skip_if_not_installed("probably")

  wflow_pre <- workflow(puromycin_tune_rec, linear_reg())
  wflow_mod <- workflow(Class ~ ., dt_spec)
  wflow_post <- workflow(Class ~ ., logistic_reg(), cls_post)

  # ------------------------------------------------------------------------------

  expect_snapshot(
    tune:::.get_config_key(
      tibble::tibble(polynimial_degree = 1:2),
      wflow_pre
    ),
    error = TRUE
  )

  expect_snapshot(
    tune:::.get_config_key(
      tibble::tibble(degree = 1:3),
      wflow_pre
    )
  )

  expect_snapshot(
    tune:::.get_config_key(
      tibble::tibble(degree = 1),
      wflow_pre
    )
  )

  expect_snapshot(
    tune:::.get_config_key(
      tibble::tibble(min_n = 10:12),
      wflow_mod
    )
  )

  expect_snapshot(
    tune:::.get_config_key(
      tibble::tibble(cut = seq(0, 1, length.out = 5)),
      wflow_post
    )
  )
})

test_that("set configurations - two elements", {
  skip_if_not_installed("probably")

  wflow_1 <- workflow(puromycin_tune_rec, knn_reg_spec)
  wflow_2 <- workflow(Class ~ ., dt_spec, cls_post)
  wflow_3 <- workflow(puromycin_tune_rec, logistic_reg(), cls_post)

  # ------------------------------------------------------------------------------

  expect_snapshot(
    tune:::.get_config_key(
      tidyr::crossing(degree = 1:3, neighbors = 1:2),
      wflow_1
    )
  )

  expect_snapshot(
    tune:::.get_config_key(
      tidyr::crossing(min_n = 1:3, cut = (1:5) / 5) |> slice(-1),
      wflow_2
    )
  )

  grid <- grid <- dials:::grid_space_filling(
    parameters(degree_int(), threshold()),
    size = 10
  ) |>
    setNames(c("degree", "cut"))
  expect_snapshot(
    tune:::.get_config_key(
      grid,
      wflow_3
    )
  )
})

test_that("set configurations - all or none", {
  skip_if_not_installed("probably")

  wflow_none <- workflow(Class ~ ., logistic_reg())
  wflow_all <- workflow(puromycin_tune_rec, dt_spec, cls_post)

  # ------------------------------------------------------------------------------

  expect_snapshot(
    tune:::.get_config_key(
      tibble(),
      wflow_none
    )
  )

  grid <- grid <- dials:::grid_space_filling(
    parameters(degree_int(), min_n(), threshold())
  ) |>
    setNames(c("degree", "min_n", "cut"))

  expect_snapshot(
    tune:::.get_config_key(
      grid,
      wflow_all
    )
  )
})

test_that("predict classification - no submodels - no calibration", {
  skip_if_not_installed("modeldata")

  cls <- make_post_data()

  pca_rec <- recipe(class ~ ., data = cls$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  wflow <- workflow(pca_rec, logistic_reg())
  grd <- tibble()

  wflow_fit <- fit(wflow, cls$data)

  class_only <- metric_set(accuracy)
  prob_only <- metric_set(brier_class)
  both_types <- metric_set(brier_class, accuracy)

  data_1 <- tune:::.get_data_subsets(wflow, cls$rs$splits[[1]], cls$args)

  fac_0 <- factor(levels = levels(cls$data$class))

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------
  # Only predict classes

  static_class <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = class_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_class <- tune:::update_static(static_class, data_1)
  static_class$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_class)

  class_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_class,
    submodel_grid = NULL
  )

  expect_equal(
    class_res[0, ],
    tibble(.pred_class = fac_0, .row = integer(0), class = fac_0)
  )
  expect_equal(nrow(class_res), nrow(assessment(cls$rs$splits[[1]])))

  # ----------------------------------------------------------------------------
  # Only predict probabilities

  static_prob <- tune:::.make_static(
    wflow,
    pred_df_1,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = prob_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_prob <- tune:::update_static(static_prob, data_1)
  static_prob$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_prob)

  prob_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_prob,
    submodel_grid = NULL
  )

  expect_equal(
    prob_res[0, ],
    tibble(
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      .row = integer(0),
      class = fac_0
    )
  )
  expect_equal(nrow(prob_res), nrow(assessment(cls$rs$splits[[1]])))

  # ----------------------------------------------------------------------------
  # Both prediction types

  static_both <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = both_types,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_both <- tune:::update_static(static_both, data_1)
  static_both$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_both)

  both_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_both,
    submodel_grid = NULL
  )

  expect_equal(
    both_res[0, ],
    tibble(
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(both_res), nrow(assessment(cls$rs$splits[[1]])))
})

test_that("predict classification - no submodels - with calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  cls <- make_post_data()

  pca_rec <- recipe(class ~ ., data = cls$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  cal_pst <- tailor() |> adjust_probability_calibration()

  wflow <- workflow(pca_rec, logistic_reg(), cal_pst)
  wflow_fit <- fit(wflow, cls$data, data_calibration = cls$data)
  grd <- tibble()

  class_only <- metric_set(accuracy)
  prob_only <- metric_set(brier_class)
  both_types <- metric_set(brier_class, accuracy)

  data_1 <- tune:::.get_data_subsets(wflow, cls$rs$splits[[1]], cls$args)

  fac_0 <- factor(levels = levels(cls$data$class))

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------
  # Only predict classes

  static_class <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = class_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_class <- tune:::update_static(static_class, data_1)
  static_class$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_class)

  class_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_class,
    submodel_grid = NULL
  )

  expect_equal(
    class_res_prd[0, ],
    tibble(
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(class_res_prd), nrow(data_1$pred$data))

  # ----------------------------------------------------------------------------
  # Only predict probabilities

  static_prob <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = prob_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_prob <- tune:::update_static(static_prob, data_1)
  static_prob$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_prob)

  prob_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_prob,
    submodel_grid = NULL
  )

  expect_equal(
    prob_res_prd[0, ],
    tibble(
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(prob_res_prd), nrow(data_1$pred$data))

  # ----------------------------------------------------------------------------
  # Both prediction types

  static_both <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = both_types,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_both <- tune:::update_static(static_both, data_1)
  static_both$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_both)

  both_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_both,
    submodel_grid = NULL
  )

  expect_equal(
    both_res_prd[0, ],
    tibble(
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(prob_res_prd), nrow(data_1$pred$data))
})

test_that("predict classification - with submodels - no calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  cls <- make_post_data()

  pca_rec <- recipe(class ~ ., data = cls$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  knn_cls_spec <- parsnip::nearest_neighbor(
    mode = "classification",
    neighbors = 12
  )

  five_neighbors <- tibble(neighbors = 5)

  wflow <- workflow(pca_rec, knn_cls_spec)
  wflow_fit <- fit(wflow, cls$data)
  grd <- tibble()

  class_only <- metric_set(accuracy)
  prob_only <- metric_set(brier_class)
  both_types <- metric_set(brier_class, accuracy)

  data_1 <- tune:::.get_data_subsets(wflow, cls$rs$splits[[1]], cls$args)

  fac_0 <- factor(levels = levels(cls$data$class))

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------
  # Only predict classes

  static_class <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = class_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_class <- tune:::update_static(static_class, data_1)
  static_class$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_class)

  class_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_class,
    submodel_grid = five_neighbors
  )

  expect_equal(
    class_res[0, ],
    tibble(
      neighbors = double(0),
      .pred_class = fac_0,
      .row = integer(0),
      class = fac_0
    )
  )
  expect_equal(nrow(class_res), nrow(assessment(cls$rs$splits[[1]])))

  # ----------------------------------------------------------------------------
  # Only predict probabilities

  static_prob <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = prob_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_prob <- tune:::update_static(static_prob, data_1)
  static_prob$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_prob)

  prob_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_prob,
    submodel_grid = five_neighbors
  )

  expect_equal(
    prob_res[0, ],
    tibble(
      neighbors = double(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      .row = integer(0),
      class = fac_0
    )
  )
  expect_equal(nrow(prob_res), nrow(assessment(cls$rs$splits[[1]])))

  # ----------------------------------------------------------------------------
  # Both prediction types

  static_both <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = both_types,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_both <- tune:::update_static(static_both, data_1)
  static_both$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_both)

  both_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_both,
    submodel_grid = five_neighbors
  )

  expect_equal(
    both_res[0, ],
    tibble(
      neighbors = double(0),
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(both_res), nrow(assessment(cls$rs$splits[[1]])))
})

test_that("predict classification - with submodels - with calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  cls <- make_post_data()

  pca_rec <- recipe(class ~ ., data = cls$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  knn_cls_spec <- parsnip::nearest_neighbor(
    mode = "classification",
    neighbors = 12
  )

  five_neighbors <- tibble(neighbors = 5)

  cal_pst <- tailor() |> adjust_probability_calibration()

  wflow <- workflow(pca_rec, knn_cls_spec, cal_pst)
  wflow_fit <- fit(wflow, cls$data, data_calibration = cls$data)
  grd <- tibble()

  class_only <- metric_set(accuracy)
  prob_only <- metric_set(brier_class)
  both_types <- metric_set(brier_class, accuracy)

  data_1 <- tune:::.get_data_subsets(wflow, cls$rs$splits[[1]], cls$args)

  fac_0 <- factor(levels = levels(cls$data$class))

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------
  # Only predict classes

  static_class <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = class_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_class <- tune:::update_static(static_class, data_1)
  static_class$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_class)

  class_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_class,
    submodel_grid = five_neighbors
  )

  expect_equal(
    class_res_prd[0, ],
    tibble(
      neighbors = double(0),
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(class_res_prd), nrow(data_1$pred$data))

  # ----------------------------------------------------------------------------
  # Only predict probabilities

  static_prob <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = prob_only,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_prob <- tune:::update_static(static_prob, data_1)
  static_prob$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_prob)

  prob_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_prob,
    submodel_grid = five_neighbors
  )

  expect_equal(
    prob_res_prd[0, ],
    tibble(
      neighbors = double(0),
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(prob_res_prd), nrow(data_1$pred$data))

  # ----------------------------------------------------------------------------
  # Both prediction types

  static_both <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = both_types,
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )

  static_both <- tune:::update_static(static_both, data_1)
  static_both$y_name <- "class"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_both)

  both_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_both,
    submodel_grid = five_neighbors
  )

  expect_equal(
    both_res_prd[0, ],
    tibble(
      neighbors = double(0),
      .pred_class = fac_0,
      .row = integer(0),
      .pred_class_1 = double(0),
      .pred_class_2 = double(0),
      class = fac_0
    )
  )
  expect_equal(nrow(prob_res_prd), nrow(data_1$pred$data))
})

test_that("predict regression - no submodels - no calibration", {
  skip_if_not_installed("modeldata")

  reg <- make_post_data(mode = "regression")

  pca_rec <- recipe(outcome ~ ., data = reg$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  wflow <- workflow(pca_rec, linear_reg())
  wflow_fit <- fit(wflow, reg$data)
  grd <- tibble()

  reg_mtr <- metric_set(rmse)

  data_1 <- tune:::.get_data_subsets(wflow, reg$rs$splits[[1]], reg$args)

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------

  static <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = reg_mtr,
    eval_time = NULL,
    split_args = reg$args,
    control = ctrl
  )

  static <- tune:::update_static(static, data_1)
  static$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static)

  class_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static,
    submodel_grid = NULL
  )

  expect_equal(
    class_res[0, ],
    tibble(.pred = double(0), .row = integer(0), outcome = double(0))
  )
  expect_equal(nrow(class_res), nrow(assessment(reg$rs$splits[[1]])))
})

test_that("predict regression - no submodels - with calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  reg <- make_post_data(mode = "regression")

  pca_rec <- recipe(outcome ~ ., data = reg$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  reg_pst <- tailor() |> adjust_numeric_calibration()

  wflow <- workflow(pca_rec, linear_reg(), reg_pst)
  wflow_fit <- fit(wflow, reg$data, data_calibration = reg$data)
  grd <- tibble()

  reg_mtr <- metric_set(rmse)

  data_1 <- tune:::.get_data_subsets(wflow, reg$rs$splits[[1]], reg$args)

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------

  static <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = reg_mtr,
    eval_time = NULL,
    split_args = reg$args,
    control = ctrl
  )

  static <- tune:::update_static(static, data_1)
  static$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static)

  class_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static,
    submodel_grid = NULL
  )

  expect_equal(
    class_res_prd[0, ],
    tibble(
      .pred = double(0),
      .row = integer(0),
      outcome = double(0)
    )
  )
  expect_equal(nrow(class_res_prd), nrow(data_1$pred$data))
})

test_that("predict regression - with submodels - no calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  reg <- make_post_data(mode = "regression")

  pca_rec <- recipe(outcome ~ ., data = reg$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  knn_reg_spec <- parsnip::nearest_neighbor(
    mode = "regression",
    neighbors = 12
  )

  five_neighbors <- tibble(neighbors = 5)

  wflow <- workflow(pca_rec, knn_reg_spec)
  wflow_fit <- fit(wflow, reg$data)

  reg_mtr <- metric_set(rmse)

  data_1 <- tune:::.get_data_subsets(wflow, reg$rs$splits[[1]], reg$args)

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------

  static <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = tibble(),
    metrics = reg_mtr,
    eval_time = NULL,
    split_args = reg$args,
    control = ctrl
  )

  static <- tune:::update_static(static, data_1)
  static$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static)

  class_res <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static,
    submodel_grid = five_neighbors
  )

  expect_equal(
    class_res[0, ],
    tibble(
      neighbors = double(0),
      .pred = double(0),
      .row = integer(0),
      outcome = double(0)
    )
  )
  expect_equal(nrow(class_res), nrow(assessment(reg$rs$splits[[1]])))
})

test_that("predict regression - with submodels - with calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  reg <- make_post_data(mode = "regression")

  pca_rec <- recipe(outcome ~ ., data = reg$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  knn_reg_spec <- parsnip::nearest_neighbor(
    mode = "regression",
    neighbors = 12
  )

  five_neighbors <- tibble(neighbors = 5)

  reg_pst <- tailor() |> adjust_numeric_calibration()

  wflow <- workflow(pca_rec, knn_reg_spec, reg_pst)
  wflow_fit <- fit(wflow, reg$data, data_calibration = reg$data)

  reg_mtr <- metric_set(rmse)

  data_1 <- tune:::.get_data_subsets(wflow, reg$rs$splits[[1]], reg$args)

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------

  static <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = tibble(),
    metrics = reg_mtr,
    eval_time = NULL,
    split_args = reg$args,
    control = ctrl
  )

  static <- tune:::update_static(static, data_1)
  static$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static)

  class_res_prd <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static,
    submodel_grid = five_neighbors
  )

  expect_equal(
    class_res_prd[0, ],
    tibble(
      neighbors = double(0),
      .pred = double(0),
      .row = integer(0),
      outcome = double(0)
    )
  )
  expect_equal(nrow(class_res_prd), nrow(data_1$pred$data))
})

test_that("predict censored regression - no submodels - no calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("censored")
  skip_if_not_installed("survival")

  library(censored)

  cens <- make_post_data(mode = "censored")

  pca_rec <- recipe(outcome ~ ., data = cens$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  wflow <- workflow(pca_rec, survival_reg())
  wflow_fit <- fit(wflow, cens$data)
  grd <- tibble()

  mtr_stc <- metric_set(concordance_survival)
  mtr_dyn <- metric_set(brier_survival)
  mtr_int <- metric_set(brier_survival_integrated)
  mtr_all <- metric_set(
    concordance_survival,
    brier_survival,
    brier_survival_integrated
  )

  .times <- c(15, 25)

  data_1 <- tune:::.get_data_subsets(wflow, cens$rs$splits[[1]], cens$args)

  ctrl <- tune::control_grid()

  # ----------------------------------------------------------------------------
  # static metrics

  static_stc <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = mtr_stc,
    eval_time = .times,
    split_args = cens$args,
    control = ctrl
  )

  static_stc <- tune:::update_static(static_stc, data_1)
  static_stc$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_stc)

  res_stc <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_stc,
    submodel_grid = NULL
  )

  expect_equal(
    res_stc[0, ],
    tibble(.pred_time = numeric(0), .row = integer(0), outcome = surv_0)
  )
  expect_equal(nrow(res_stc), nrow(assessment(cens$rs$splits[[1]])))

  # ----------------------------------------------------------------------------
  # dynamic metrics

  static_dyn <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = mtr_dyn,
    eval_time = .times,
    split_args = cens$args,
    control = ctrl
  )

  static_dyn <- tune:::update_static(static_dyn, data_1)
  static_dyn$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_dyn)

  res_dyn <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_dyn,
    submodel_grid = NULL
  )

  expect_equal(
    res_dyn[0, ],
    tibble(.pred = list(), .row = integer(0), outcome = surv_0)
  )

  expect_equal(
    res_dyn$.pred[[1]][0, ],
    pred_dyn_0
  )
  expect_equal(nrow(res_dyn), nrow(assessment(cens$rs$splits[[1]])))
})

# TODO put this in extratests
test_that("predict censored regression - submodels - no calibration", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("censored")
  skip_if_not_installed("survival")
  skip_if_not_installed("glmnet")

  library(censored)

  cens <- make_post_data(mode = "censored")

  pca_rec <- recipe(outcome ~ ., data = cens$data) |>
    step_pca(all_numeric_predictors(), num_comp = 2)

  glmn_cens <- proportional_hazards(penalty = tune()) |> set_engine("glmnet")

  wflow <- workflow(pca_rec, glmn_cens)

  grd <-
    wflow |>
    extract_parameter_set_dials() |>
    grid_space_filling(size = 3)

  mtr_stc <- metric_set(concordance_survival)
  mtr_dyn <- metric_set(brier_survival)
  mtr_int <- metric_set(brier_survival_integrated)
  mtr_all <- metric_set(
    concordance_survival,
    brier_survival,
    brier_survival_integrated
  )

  .times <- c(15, 25)

  data_1 <- tune:::.get_data_subsets(wflow, cens$rs$splits[[1]], cens$args)

  ctrl <- tune::control_grid()

  wflow_fit <-
    wflow |>
    finalize_workflow(grd[1, ]) |>
    fit(cens$data)

  # ----------------------------------------------------------------------------
  # static metrics

  static_stc <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = mtr_stc,
    eval_time = .times,
    split_args = cens$args,
    control = ctrl
  )

  static_stc <- tune:::update_static(static_stc, data_1)
  static_stc$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_stc)

  res_stc <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_stc,
    submodel_grid = NULL
  )

  expect_equal(
    res_stc[0, ],
    tibble(.pred_time = numeric(0), .row = integer(0), outcome = surv_0)
  )
  expect_equal(nrow(res_stc), nrow(assessment(cens$rs$splits[[1]])))

  # ----------------------------------------------------------------------------
  # dynamic metrics

  static_dyn <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = mtr_dyn,
    eval_time = .times,
    split_args = cens$args,
    control = ctrl
  )

  static_dyn <- tune:::update_static(static_dyn, data_1)
  static_dyn$y_name <- "outcome"
  pred_df_1 <- tune:::process_prediction_data(wflow_fit, static_dyn)

  res_dyn <- tune:::predict_all_types(
    wflow_fit,
    pred_df_1,
    static_dyn,
    submodel_grid = NULL
  )

  expect_equal(
    res_dyn[0, ],
    tibble(.pred = list(), .row = integer(0), outcome = surv_0)
  )

  expect_equal(
    res_dyn$.pred[[1]][0, ],
    pred_dyn_0
  )
  expect_equal(nrow(res_dyn), nrow(assessment(cens$rs$splits[[1]])))
})

# ------------------------------------------------------------------------------
# Row availability checks

test_that("process_prediction_data errors when recipe drops rows", {
  skip_if_not_installed("modeldata")

  cls <- make_post_data()

  filter_rec <- recipe(class ~ ., data = cls$data) |>
    step_filter(non_linear_1 > 0, skip = FALSE)

  wflow <- workflow(filter_rec, logistic_reg())
  wflow_fit <- fit(wflow, cls$data)

  ctrl <- tune::control_grid()
  data_1 <- tune:::.get_data_subsets(wflow, cls$rs$splits[[1]], cls$args)

  static <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = tibble(),
    metrics = metric_set(accuracy),
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )
  static <- tune:::update_static(static, data_1)

  expect_snapshot(
    tune:::process_prediction_data(wflow_fit, static),
    error = TRUE
  )
})

test_that("process_prediction_data errors when recipe drops calibration rows", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")

  cls <- make_post_data()

  filter_rec <- recipe(class ~ ., data = cls$data) |>
    step_filter(non_linear_1 > 0, skip = FALSE)

  wflow <- workflow(filter_rec, logistic_reg(), cls_est_post)
  # TODO just use cls$data
  # after https://github.com/tidymodels/workflows/issues/315 is fixed
  wflow_fit <- fit(
    wflow,
    data = cls$data,
    data_calibration = cls$data |> filter(non_linear_1 > 0)
  )

  ctrl <- tune::control_grid()
  data_1 <- tune:::.get_data_subsets(wflow, cls$rs$splits[[1]], cls$args)

  static <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = tibble(),
    metrics = metric_set(accuracy),
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )
  static <- tune:::update_static(static, data_1)

  expect_snapshot(
    tune:::process_prediction_data(wflow_fit, static, source = "cal"),
    error = TRUE
  )
})

test_that("process_prediction_data errors when non-recipe preprocessor drops rows", {
  skip_if_not_installed("modeldata")

  cls <- make_post_data()

  wflow <- workflow(class ~ ., logistic_reg())
  wflow_fit <- fit(wflow, cls$data)

  ctrl <- tune::control_grid()
  data_1 <- tune:::.get_data_subsets(wflow, cls$rs$splits[[1]], cls$args)

  static <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = tibble(),
    metrics = metric_set(accuracy),
    eval_time = NULL,
    split_args = cls$args,
    control = ctrl
  )
  static <- tune:::update_static(static, data_1)

  local_mocked_bindings(
    forge_from_workflow = function(new_data, workflow) {
      res <- hardhat::forge(
        new_data,
        workflow$pre$mold$blueprint,
        outcomes = TRUE
      )
      res$predictors <- res$predictors[1:5, ]
      res
    }
  )

  expect_snapshot(
    tune:::process_prediction_data(wflow_fit, static),
    error = TRUE
  )
})

test_that("stage helpers - recipe only", {
  wflow_1 <- workflows::workflow(puromycin_tune_rec, linear_reg())
  grid_1 <- wflow_1 |>
    extract_parameter_set_dials() |>
    grid_regular()
  sched_1 <- schedule_grid(grid_1, wflow_1)

  expect_equal(dim(tune:::remove_stage(sched_1)), c(3L, 1L))
  expect_s3_class(
    sched_1,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )

  expect_true(tune:::has_pre_param(sched_1))
  expect_false(tune:::has_sub_param(sched_1$model_stage[[1]]$predict_stage))
})

test_that("stage helpers - model only", {
  # Tunable model with submodels

  wflow_1 <- workflows::workflow(class ~ ., glmn_spec)
  grid_1 <- wflow_1 |>
    extract_parameter_set_dials() |>
    grid_regular() |>
    slice(-1)
  sched_1 <- schedule_grid(grid_1, wflow_1)

  expect_equal(dim(tune:::remove_stage(sched_1)), c(1L, 0L))

  res_1 <- tune:::remove_stage(sched_1 |> mutate(col = 1))
  exp_1 <- tibble::tibble(col = 1)
  class(exp_1) <- c("grid_schedule", "schedule", class(exp_1))
  expect_equal(res_1, exp_1)

  expect_false(tune:::has_pre_param(sched_1))
  expect_true(tune:::has_mod_param(sched_1))
  expect_equal(
    purrr:::map_lgl(
      sched_1$model_stage[[1]]$predict_stage,
      tune:::has_sub_param
    ),
    rep(TRUE, 3)
  )
  expect_equal(
    purrr:::map_chr(
      sched_1$model_stage[[1]]$predict_stage,
      tune:::get_sub_param
    ),
    rep("penalty", 3)
  )

  # ----------------------------------------------------------------------------
  # Tunable model without submodels

  wflow_2 <- workflows::workflow(class ~ ., dt_spec)
  grid_2 <- wflow_2 |>
    extract_parameter_set_dials() |>
    grid_regular()
  sched_2 <- schedule_grid(grid_2, wflow_2)

  expect_equal(dim(tune:::remove_stage(sched_2)), c(1L, 0L))

  res_2 <- tune:::remove_stage(sched_2 |> mutate(col = 1))
  exp_2 <- tibble::tibble(col = 1)
  class(exp_2) <- c("grid_schedule", "schedule", class(exp_2))
  expect_equal(res_2, exp_2)

  expect_false(tune:::has_pre_param(sched_2))
  expect_true(tune:::has_mod_param(sched_2))
  expect_equal(
    purrr:::map_lgl(
      sched_2$model_stage[[1]]$predict_stage,
      tune:::has_sub_param
    ),
    rep(FALSE, 3)
  )
  expect_equal(
    purrr:::map(
      sched_2$model_stage[[1]]$predict_stage,
      tune:::get_sub_param
    ),
    list(character(0), character(0), character(0))
  )
})

test_that("stage helpers - post only", {
  skip_if_not_installed("probably")

  wflow_1 <- workflows::workflow(class ~ ., logistic_reg(), cls_post)
  grid_1 <- wflow_1 |>
    extract_parameter_set_dials() |>
    grid_regular()
  sched_1 <- schedule_grid(grid_1, wflow_1)

  expect_equal(dim(tune:::remove_stage(sched_1)), c(1L, 0L))
  expect_s3_class(
    sched_1,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )

  expect_false(tune:::has_pre_param(sched_1))
  expect_false(tune:::has_mod_param(sched_1$model_stage[[1]]))
})

test_that("stage helpers - recipe and model", {
  # No submodels

  wflow_1 <- workflows::workflow(puromycin_tune_rec, dt_spec)
  grid_1 <- wflow_1 |>
    extract_parameter_set_dials() |>
    grid_regular() |>
    slice(-1)
  sched_1 <- schedule_grid(grid_1, wflow_1)

  expect_equal(dim(tune:::remove_stage(sched_1)), c(3L, 1L))
  expect_s3_class(
    sched_1,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )

  expect_true(tune:::has_pre_param(sched_1))

  expect_equal(
    purrr::map_lgl(sched_1$model_stage, tune:::has_mod_param),
    rep(TRUE, 3)
  )

  expect_equal(
    purrr::map_lgl(
      sched_1$model_stage,
      \(.x) tune:::has_sub_param(.x$predict_stage)
    ),
    rep(FALSE, 3)
  )

  # With submodels

  wflow_2 <- workflows::workflow(puromycin_tune_rec, glmn_spec)
  grid_2 <- wflow_2 |>
    extract_parameter_set_dials() |>
    grid_regular() |>
    slice(-1)
  sched_2 <- schedule_grid(grid_2, wflow_2)

  expect_equal(dim(tune:::remove_stage(sched_2)), c(3L, 1L))
  expect_s3_class(
    sched_2,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )

  expect_true(tune:::has_pre_param(sched_2))

  expect_equal(
    purrr::map_lgl(sched_2$model_stage, tune:::has_mod_param),
    rep(TRUE, 3)
  )

  expect_equal(
    purrr::map_lgl(
      sched_2$model_stage,
      \(.x) tune:::has_sub_param(.x$predict_stage[[1]])
    ),
    rep(TRUE, 3)
  )
})

test_that("stage helpers - recipe and post", {
  skip_if_not_installed("probably")

  wflow_1 <- workflows::workflow(puromycin_tune_rec, logistic_reg(), cls_post)
  grid_1 <- wflow_1 |>
    extract_parameter_set_dials() |>
    grid_regular()
  sched_1 <- schedule_grid(grid_1, wflow_1)

  expect_equal(dim(tune:::remove_stage(sched_1)), c(3L, 1L))
  expect_s3_class(
    sched_1,
    c("grid_schedule", "schedule", "tbl_df", "tbl", "data.frame")
  )

  expect_true(tune:::has_pre_param(sched_1))

  expect_false(tune:::has_mod_param(sched_1$model_stage[[1]]))
})

test_that("maker static object", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  mc_cv_args <- rsample::.get_split_args(two_class_rs)

  wflow <- workflow(Class ~ ., dt_spec, cls_est_post)

  res <- tune:::.make_static(
    wflow,
    param_info = wflow |> extract_parameter_set_dials(),
    grid = dt_grid,
    metrics = metric_set(accuracy),
    eval_time = NULL,
    split_args = mc_cv_args,
    control = control_resamples()
  )

  expect_true(is.list(res))
  expect_named(
    res,
    c(
      "wflow",
      "param_info",
      "configs",
      "post_estimation",
      "metrics",
      "metric_info",
      "metric_type",
      "pred_types",
      "eval_time",
      "split_args",
      "control",
      "pkgs",
      "strategy",
      "data"
    )
  )

  expect_snapshot(
    tune:::.make_static(
      1,
      param_info = wflow |> extract_parameter_set_dials(),
      grid = dt_grid,
      metrics = metric_set(accuracy),
      eval_time = NULL,
      split_args = mc_cv_args,
      control = control_resamples()
    ),
    error = TRUE
  )

  expect_snapshot(
    tune:::.make_static(
      wflow,
      param_info = 2,
      grid = dt_grid,
      metrics = metric_set(accuracy),
      eval_time = NULL,
      split_args = mc_cv_args,
      control = control_resamples()
    ),
    error = TRUE
  )

  expect_snapshot(
    tune:::.make_static(
      wflow,
      param_info = wflow |> extract_parameter_set_dials(),
      grid = dt_grid,
      metrics = 3,
      eval_time = NULL,
      split_args = mc_cv_args,
      control = control_resamples()
    ),
    error = TRUE
  )

  expect_snapshot(
    tune:::.make_static(
      wflow,
      param_info = wflow |> extract_parameter_set_dials(),
      grid = dt_grid,
      metrics = metric_set(accuracy),
      eval_time = "four",
      split_args = mc_cv_args,
      control = control_resamples()
    ),
    error = TRUE
  )

  partitions <- tune:::.get_data_subsets(wflow, two_class_rs$splits[[1]])

  res_with_data <- tune:::update_static(res, partitions)
  expect_true(is.list(res_with_data))
  expect_named(
    res_with_data,
    c(
      "wflow",
      "param_info",
      "configs",
      "post_estimation",
      "metrics",
      "metric_info",
      "metric_type",
      "pred_types",
      "eval_time",
      "split_args",
      "control",
      "pkgs",
      "strategy",
      "data"
    )
  )

  bad_part <- partitions
  bad_part$cal$data <- 1
  expect_snapshot(tune:::update_static(res, bad_part), error = TRUE)

  bad_part <- partitions
  bad_part$cal$ind <- "two"
  expect_snapshot(tune:::update_static(res, bad_part), error = TRUE)
})

test_that("finalize_fit_pre() with formulas", {
  skip_if_not_installed("modeldata")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  rs_split <- two_class_rs$splits[[1]]

  # ----------------------------------------------------------------------------

  form_wflow <- workflow(Class ~ ., dt_spec)
  grd <-
    form_wflow |>
    extract_parameter_set_dials() |>
    grid_space_filling(size = 3)

  form_stc <- tune:::.make_static(
    form_wflow,
    param_info = form_wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = metric_set(brier_class, spec),
    eval_time = NULL,
    split_args = rsample::.get_split_args(two_class_rs),
    control = control_grid()
  )
  form_stc <- tune:::update_static(
    form_stc,
    tune:::.get_data_subsets(form_stc$wflow, rs_split, form_stc$split_args)
  )
  form_stc$y_name <- "Class"

  form_res <- tune:::finalize_fit_pre(form_wflow, dt_grid, form_stc)
  expect_s3_class(form_res, "workflow")
  form_pre_res <- form_res |> extract_mold()
  expect_named(form_pre_res, c("predictors", "outcomes", "blueprint", "extras"))
})

# ------------------------------------------------------------------------------

test_that("finalize_fit_pre() with recipes", {
  skip_if_not_installed("modeldata")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  rs_split <- two_class_rs$splits[[1]]

  # ----------------------------------------------------------------------------

  rec <- recipe(Class ~ ., data = two_class_dat) |>
    step_normalize(A, B)

  rec_wflow <- workflow(rec, dt_spec)
  grd <-
    rec_wflow |>
    extract_parameter_set_dials() |>
    grid_space_filling(size = 3)

  rec_stc <- tune:::.make_static(
    rec_wflow,
    param_info = rec_wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = metric_set(brier_class, spec),
    eval_time = NULL,
    split_args = rsample::.get_split_args(two_class_rs),
    control = control_grid()
  )
  rec_stc <- tune:::update_static(
    rec_stc,
    tune:::.get_data_subsets(rec_stc$wflow, rs_split, rec_stc$split_args)
  )

  rec_res <- tune:::finalize_fit_pre(rec_wflow, dt_grid, rec_stc)
  expect_s3_class(rec_res, "workflow")
  rec_pre_res <- rec_res |> extract_mold()
  expect_named(rec_pre_res, c("predictors", "outcomes", "blueprint", "extras"))
  rec_pre_res <- rec_res |> extract_recipe()
  expect_s3_class(rec_pre_res, "recipe")
})

test_that("finalize_fit_pre() with tuned recipes", {
  skip_if_not_installed("modeldata")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  rs_split <- two_class_rs$splits[[1]]

  # ----------------------------------------------------------------------------

  rec <- recipe(Class ~ ., data = two_class_dat) |>
    step_pca(A, B, num_comp = tune("comps"))

  rec_grid <- tibble(comps = 1)

  rec_wflow <- workflow(rec, dt_spec)
  grd <-
    rec_wflow |>
    extract_parameter_set_dials() |>
    grid_space_filling(size = 3)

  rec_stc <- tune:::.make_static(
    rec_wflow,
    param_info = rec_wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = metric_set(brier_class, spec),
    eval_time = NULL,
    split_args = rsample::.get_split_args(two_class_rs),
    control = control_grid()
  )
  rec_stc <- tune:::update_static(
    rec_stc,
    tune:::.get_data_subsets(rec_stc$wflow, rs_split, rec_stc$split_args)
  )

  rec_res <- tune:::finalize_fit_pre(rec_wflow, rec_grid, rec_stc)
  expect_s3_class(rec_res, "workflow")
  rec_pre_res <- rec_res |> extract_mold()
  expect_named(rec_pre_res, c("predictors", "outcomes", "blueprint", "extras"))
  rec_pre_res <- rec_res |> extract_recipe()
  expect_s3_class(rec_pre_res, "recipe")
  rec_pre_data <- bake(rec_pre_res, new_data = two_class_dat[1:2, ])
  expect_named(rec_pre_data, c("Class", "PC1"))
})

# ------------------------------------------------------------------------------

test_that("finalize_fit_pre() with selectors", {
  skip_if_not_installed("modeldata")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  rs_split <- two_class_rs$splits[[1]]

  # ----------------------------------------------------------------------------

  vars_wflow <- workflow(spec = dt_spec) |>
    add_variables(outcomes = c(Class), predictors = c(A, B))
  grd <-
    vars_wflow |>
    extract_parameter_set_dials() |>
    grid_space_filling(size = 3)

  vars_stc <- tune:::.make_static(
    vars_wflow,
    param_info = vars_wflow |> extract_parameter_set_dials(),
    grid = grd,
    metrics = metric_set(brier_class, spec),
    eval_time = NULL,
    split_args = rsample::.get_split_args(two_class_rs),
    control = control_grid()
  )
  vars_stc <- tune:::update_static(
    vars_stc,
    tune:::.get_data_subsets(vars_stc$wflow, rs_split, vars_stc$split_args)
  )

  vars_res <- tune:::finalize_fit_pre(vars_wflow, dt_grid, vars_stc)
  expect_s3_class(vars_res, "workflow")
  vars_pre_res <- vars_res |> extract_mold()
  expect_named(vars_pre_res, c("predictors", "outcomes", "blueprint", "extras"))
})

test_that("finalize_fit_model() for classification", {
  skip_if_not_installed("modeldata")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  rs_split <- two_class_rs$splits[[1]]

  # ----------------------------------------------------------------------------

  dt_wflow <- workflow(Class ~ ., dt_spec)
  dt_grid <- tibble(min_n = 5)

  dt_stc <- tune:::.make_static(
    dt_wflow,
    param_info = dt_wflow |> extract_parameter_set_dials(),
    grid = dt_grid,
    metrics = metric_set(brier_class, spec),
    eval_time = NULL,
    split_args = rsample::.get_split_args(two_class_rs),
    control = control_grid()
  )
  dt_stc <- tune:::update_static(
    dt_stc,
    tune:::.get_data_subsets(dt_stc$wflow, rs_split, dt_stc$split_args)
  )

  dt_0_res <- tune:::finalize_fit_pre(dt_wflow, dt_grid, dt_stc)
  dt_res <- tune:::finalize_fit_model(dt_0_res, dt_grid)
  expect_s3_class(dt_res, "workflow")
  dt_res <- dt_res |> extract_fit_parsnip()
  expect_s3_class(dt_res, c("_C5.0", "model_fit"))
})

test_that("extract data subsets - no postprocessing", {
  skip_if_not_installed("modeldata")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  rs_split <- two_class_rs$splits[[1]]

  wflow_1 <- workflow(Class ~ ., dt_spec)
  data_1 <- tune:::.get_data_subsets(wflow_1, rs_split)
  expect_named(data_1, c("fit", "pred", "cal"))
  expect_equal(data_1$fit$data, analysis(rs_split))
  expect_equal(data_1$pred$data, assessment(rs_split))
  expect_equal(data_1$fit$ind, as.integer(rs_split))
  expect_equal(data_1$pred$ind, as.integer(rs_split, data = "assessment"))
})

test_that("extract data subsets - no estimated postprocessing", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  rs_split <- two_class_rs$splits[[1]]

  wflow_1 <- workflow(Class ~ ., dt_spec, cls_post)
  data_1 <- tune:::.get_data_subsets(wflow_1, rs_split)
  expect_named(data_1, c("fit", "pred", "cal"))
  expect_equal(data_1$fit$data, analysis(rs_split))
  expect_equal(data_1$pred$data, assessment(rs_split))
  expect_equal(data_1$fit$ind, as.integer(rs_split))
  expect_equal(data_1$pred$ind, as.integer(rs_split, data = "assessment"))
})

test_that("extract data subsets - estimated postprocessing", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")

  data("two_class_dat", package = "modeldata")
  two_class_rs <- mc_cv(two_class_dat, times = 2)
  mc_cv_args <- rsample::.get_split_args(two_class_rs)

  rs_split <- two_class_rs$splits[[1]]

  set.seed(1)
  rs_sub_split <- rsample::internal_calibration_split(rs_split, mc_cv_args)

  wflow_1 <- workflow(Class ~ ., dt_spec, cls_est_post)

  set.seed(1)
  data_1 <- tune:::.get_data_subsets(wflow_1, rs_split, mc_cv_args)

  expect_named(
    data_1,
    c("fit", "pred", "cal")
  )
  expect_equal(data_1$fit$data, analysis(rs_sub_split))
  expect_equal(data_1$cal$data, calibration(rs_sub_split))
  expect_equal(data_1$pred$data, assessment(rs_split))

  expect_equal(data_1$fit$ind, as.integer(rs_sub_split))
  expect_equal(data_1$cal$ind, as.integer(rs_sub_split, data = "assessment"))
  expect_equal(data_1$pred$ind, as.integer(rs_split, data = "assessment"))
})

test_that("tailor trains calibrator", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")

  cls <- make_post_data()

  wflow_cal <- workflow(class ~ ., logistic_reg(), cls_est_post)
  wflow_fit <- .fit_pre(wflow_cal, cls$data) |>
    .fit_model(control = control_workflow())

  cal_preds <- augment(extract_fit_parsnip(wflow_fit), cls$data)
  wflow_fit <- finalize_fit_post(wflow_fit, cal_preds, grid = tibble())
  res <- extract_postprocessor(wflow_fit, estimated = TRUE)
  expect_s3_class(res, "tailor")
  expect_true(res$adjustments[[1]]$trained)
  expect_equal(
    res$adjustments[[1]]$results$fit$estimates[[1]]$estimate$df.null,
    nrow(cls$data) - 1
  )
})

test_that("tailor updated with grid and fit", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  cls <- make_post_data()

  wflow <- workflow(class ~ ., logistic_reg())

  cal_and_cut <- tailor::tailor() |>
    tailor::adjust_probability_calibration() |>
    tailor::adjust_probability_threshold(threshold = tune("cut"))

  wflow_cal <- workflow(class ~ ., logistic_reg(), cal_and_cut)
  wflow_fit <- .fit_pre(wflow_cal, cls$data) |>
    .fit_model(control = control_workflow())

  cal_preds <- augment(extract_fit_parsnip(wflow_fit), cls$data)
  wflow_fit <- finalize_fit_post(wflow_fit, cal_preds, grid = tibble(cut = 0))
  wflow_fit <- .fit_finalize(wflow_fit)

  re_predicted <- predict(wflow_fit, cls$data)
  res <- extract_postprocessor(wflow_fit, estimated = TRUE)

  expect_s3_class(res, "tailor")
  expect_true(res$adjustments[[1]]$trained)
  expect_equal(
    res$adjustments[[1]]$results$fit$estimates[[1]]$estimate$df.null,
    nrow(cls$data) - 1
  )
  expect_true(all(re_predicted$.pred_class == "class_1"))
})

# These tests reference results generated by tune 1.3.0. The code to generate
# them (and the results) are found in the `inst` directory.

test_that("verifying .loop_over_all_stages, no submodels, post estimation without tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  load(system.file(
    "regression_tests",
    "simple_example.RData",
    package = "tune"
  ))

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_regression(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  mod <- nearest_neighbor(neighbors = 11, weight_func = tune()) |>
    set_mode("regression")
  wflow <- workflow(outcome ~ ., mod, reg_cal_max)
  max_param <-
    wflow |>
    extract_parameter_set_dials() |>
    update(upper_limit = upper_limit(c(0, 1)))

  grd <- max_param |> grid_regular(levels = c(3, 2))
  upper_vals <- sort(unique(grd$upper_limit))

  static_1 <- tune:::.make_static(
    wflow,
    param_info = max_param,
    grid = grd,
    metrics = metric_set(rmse, rsq),
    eval_time = NULL,
    split_args = rs_args,
    control = control_grid(save_pred = TRUE)
  )

  data_1 <- tune:::.get_data_subsets(wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "outcome"

  simple_res <- tune:::.loop_over_all_stages(rs_iter, grd, static_1)
  expect_true(!is.null(simple_res$.metrics[[1]]))
  expect_named(
    simple_res,
    c(".metrics", ".notes", "outcome_names", "id", ".predictions")
  )
  expect_true(nrow(simple_res) == 1)
  expect_equal(
    nrow(simple_res$.predictions[[1]]),
    nrow(data_1$pred$data) * nrow(grd)
  )

  # TODO more tests can be added when calibration method = "none" is implemented
})

test_that("verifying .loop_over_all_stages, submodels, post estimation without tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  load(system.file(
    "regression_tests",
    "submodel_example.RData",
    package = "tune"
  ))

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_regression(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  rec <- recipe(outcome ~ ., data = dat) |>
    step_pca(all_numeric_predictors(), num_comp = tune())

  mod <- nearest_neighbor(neighbors = tune("k"), weight_func = tune()) |>
    set_mode("regression")

  submodel_wflow <- workflow(rec, mod, reg_cal_max)
  max_param <-
    submodel_wflow |>
    extract_parameter_set_dials() |>
    update(upper_limit = upper_limit(c(0, 1)))

  upper_vals <- c(0, 1)

  # fmt: skip
  submodel_grid <-
    tibble::tribble(
      ~k,   ~weight_func, ~num_comp,
      9L,  "rectangular",        2L,
      14L,  "rectangular",        2L,
      20L,  "rectangular",        2L,
      4L,   "triangular",        2L,
      9L,   "triangular",        2L,
      14L,   "triangular",        2L,
      20L,   "triangular",        2L,
      4L, "epanechnikov",        2L,
      9L, "epanechnikov",        2L,
      14L, "epanechnikov",        2L,
      20L, "epanechnikov",        2L,
      4L,  "rectangular",       10L,
      9L,  "rectangular",       10L,
      14L,  "rectangular",       10L,
      20L,  "rectangular",       10L,
      4L,   "triangular",       10L,
      9L,   "triangular",       10L,
      14L,   "triangular",       10L,
      20L,   "triangular",       10L,
      4L, "epanechnikov",       10L,
      9L, "epanechnikov",       10L,
      14L, "epanechnikov",       10L,
      20L, "epanechnikov",       10L
    ) |>
    tidyr::crossing(upper_limit = upper_vals)

  # ------------------------------------------------------------------------------

  static_1 <- tune:::.make_static(
    submodel_wflow,
    param_info = max_param,
    grid = submodel_grid,
    metrics = metric_set(rmse),
    eval_time = NULL,
    split_args = rs_args,
    control = tune::control_grid(save_pred = TRUE)
  )

  data_1 <- tune:::.get_data_subsets(submodel_wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "outcome"

  submodel_res <- tune:::.loop_over_all_stages(rs_iter, submodel_grid, static_1)
  expect_true(!is.null(submodel_res$.metrics[[1]]))
  expect_named(
    submodel_res,
    c(".metrics", ".notes", "outcome_names", "id", ".predictions")
  )
  expect_true(nrow(submodel_res) == 1)
  expect_equal(
    nrow(submodel_res$.predictions[[1]]),
    nrow(data_1$pred$data) * nrow(submodel_grid)
  )
})

test_that("verifying .loop_over_all_stages, submodels only, post estimation without tuning", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")
  skip_if_not_installed("probably")
  skip_if_not_installed("mgcv")

  load(system.file(
    "regression_tests",
    "submodel_only_example.RData",
    package = "tune"
  ))

  # ------------------------------------------------------------------------------

  set.seed(1)
  dat <- modeldata::sim_classification(1000)
  rs <- vfold_cv(dat)

  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)

  rs_iter <- tune:::vec_list_rowwise(rs) |>
    purrr::pluck(1) |>
    mutate(
      .seeds = tune:::get_parallel_seeds(1)
    )

  # ------------------------------------------------------------------------------

  mod <- nearest_neighbor(neighbors = tune(), weight_func = "triangular") |>
    set_mode("classification")

  submodel_only_wflow <- workflow(class ~ ., mod, cls_cal_tune_post)
  cut_vals <- c(.1, .9)
  submodel_only_grid <- tidyr::crossing(neighbors = 3:10, cut = cut_vals)

  # ------------------------------------------------------------------------------

  static_1 <- tune:::.make_static(
    submodel_only_wflow,
    param_info = submodel_only_wflow |> extract_parameter_set_dials(),
    grid = submodel_only_grid,
    metrics = metric_set(accuracy, roc_auc, brier_class),
    eval_time = NULL,
    split_args = rs_args,
    control = tune::control_grid(save_pred = TRUE)
  )

  data_1 <- tune:::.get_data_subsets(submodel_only_wflow, rs_split, rs_args)
  static_1 <- tune:::update_static(static_1, data_1)
  static_1$y_name <- "class"

  submodel_only_res <- tune:::.loop_over_all_stages(
    rs_iter,
    submodel_only_grid,
    static_1
  )
  expect_true(!is.null(submodel_only_res$.metrics[[1]]))
  expect_named(
    submodel_only_res,
    c(".metrics", ".notes", "outcome_names", "id", ".predictions")
  )
  expect_true(nrow(submodel_only_res) == 1)
  expect_equal(
    nrow(submodel_only_res$.predictions[[1]]),
    nrow(data_1$pred$data) * nrow(submodel_only_grid)
  )

  # TODO more tests can be added when calibration method = "none" is implemented
})

#' Internal functions for developers
#'
#' These are not intended for use by the general public.
#' @param x An object.
#' @param ... Other options
#' @keywords internal
#' @export
empty_ellipses <- function(...) {
  dots <- rlang::enquos(...)
  if (length(dots) > 0) {
    nms <- names(dots)
    no_name <- nms == ""
    if (!any(no_name)) {
      cli::cli_warn(
        "The {.code ...} are not used in this function but {length(dots)}
         object{?s} {?was/were} passed: {.val {names(dots)}}"
      )
    } else if (all(no_name)) {
      cli::cli_warn(
        "The {.code ...} are not used in this function but {length(dots)}
         unnamed object{?s} {?was/were} passed."
      )
    } else {
      cli::cli_warn(
        "The {.code ...} are not used in this function but {length(dots)}
         object{?s} {?was/were} passed."
      )
    }
  }
  invisible(NULL)
}


#' @export
#' @keywords internal
#' @rdname empty_ellipses
is_recipe <- function(x) {
  inherits(x, "recipe")
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
is_preprocessor <- function(x) {
  is_recipe(x) || rlang::is_formula(x)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
is_workflow <- function(x) {
  inherits(x, "workflow")
}

# adapted from ps:::is_cran_check()
is_cran_check <- function() {
  if (identical(Sys.getenv("NOT_CRAN"), "true")) {
    FALSE
  } else {
    Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") != ""
  }
}

# suggests: a character vector of package names, giving packages
#           listed in Suggests that are needed for the example.
# for use a la `@examplesIf tune:::should_run_examples()`
should_run_examples <- function(suggests = NULL) {
  has_needed_installs <- TRUE

  if (!is.null(suggests)) {
    has_needed_installs <- rlang::is_installed(suggests)
  }

  has_needed_installs && !is_cran_check()
}

# new_tibble() currently doesn't strip attributes
# https://github.com/tidyverse/tibble/pull/769
#' @export
#' @keywords internal
#' @rdname empty_ellipses
new_bare_tibble <- function(x, ..., class = character()) {
  x <- vctrs::new_data_frame(x)
  tibble::new_tibble(x, nrow = nrow(x), ..., class = class)
}

# a helper that takes in a .config vector and returns the corresponding `.iter`.
# entries from initial results, e.g. `pre2_mod1_post0`, are assigned
# `.iter = 0`.
.config_to_.iter <- function(.config) {
  .iter <- .config
  nonzero <- grepl("^[iI]ter", .iter)
  .iter <- ifelse(nonzero, gsub("^[iI]ter", "", .iter), "0")
  .iter <- as.numeric(.iter)
  .iter
}

`%||%` <- function(x, y) {
  if (rlang::is_null(x)) y else x
}

## -----------------------------------------------------------------------------

#' Various accessor functions
#'
#' These functions return different attributes from objects with class
#' `tune_result`.
#'
#' @param x An object of class `tune_result`.
#' @return
#' \itemize{
#'   \item `.get_tune_parameters()` returns a `dials` `parameter` object or a tibble.
#'   \item `.get_tune_parameter_names()`, `.get_tune_metric_names()`, and
#'    `.get_tune_outcome_names()` return a character string.
#'   \item `.get_tune_metrics()` returns a metric set or NULL.
#'   \item `.get_tune_workflow()` returns the workflow used to fit the
#'   resamples (if `save_workflow` was set to `TRUE` during fitting) or NULL.
#' }
#' @keywords internal
#' @export
#' @rdname tune_accessor
.get_tune_parameters <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "parameters")) {
    res <- x$parameters
  } else {
    res <- tibble::new_tibble(list())
  }
  res
}

#' @export
#' @rdname tune_accessor
.get_tune_parameter_names <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "parameters")) {
    res <- x$parameters$id
  } else {
    res <- character(0)
  }
  res
}


#' @export
#' @rdname tune_accessor
# This will return any other columns that should be added to the group_by()
# when computing the final (averaged) resampling estimate
.get_extra_col_names <- function(x) {
  res <- character(0)
  mtrcs <- x$.metrics[[1]]
  if (any(names(mtrcs) == ".eval_time")) {
    res <- c(res, ".eval_time")
  }
  if (any(names(mtrcs) == ".by")) {
    res <- c(res, ".by")
  }
  res
}


#' @export
#' @rdname tune_accessor
.get_tune_metrics <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "metrics")) {
    res <- x$metrics
  } else {
    res <- NULL
  }
  res
}

#' @export
#' @rdname tune_accessor
.get_tune_metric_names <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "metrics")) {
    res <- names(attributes(x$metrics)$metrics)
  } else {
    res <- character(0)
  }
  res
}


#' @export
#' @rdname tune_accessor
.get_tune_eval_times <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "eval_time")) {
    res <- x$eval_time
  } else {
    res <- NULL
  }
  res
}

#' @export
#' @rdname tune_accessor
.get_tune_eval_time_target <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "eval_time_target")) {
    res <- x$eval_time_target
  } else {
    res <- NULL
  }
  res
}

#' @export
#' @rdname tune_accessor
.get_tune_outcome_names <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "outcomes")) {
    res <- x$outcomes
  } else {
    res <- character(0)
  }
  res
}

#' @export
#' @rdname tune_accessor
.get_tune_workflow <- function(x) {
  x <- attributes(x)
  if (any(names(x) == "workflow")) {
    res <- x$workflow
  } else {
    res <- NULL
  }
  res
}

#' @export
#' @rdname tune_accessor
.get_fingerprint.tune_results <- function(x, ...) {
  att <- attributes(x)$rset_info$att
  if (any(names(att) == "fingerprint")) {
    res <- att$fingerprint
  } else {
    res <- NA_character_
  }
  res
}

# Get a textual summary of the type of resampling
#' @export
pretty.tune_results <- function(x, ...) {
  attr(x, "rset_info")$label
}

#' Resampling weights utility functions
#'
#' These are internal functions for handling variable resampling weights in
#' hyperparameter tuning.
#'
#' @param x A tune_results object.
#' @param weights Numeric vector of weights.
#' @param id_names Character vector of ID column names.
#' @param metrics_data The metrics data frame.
#' @param w Numeric vector of weights.
#' @param num_resamples Integer number of resamples.
#'
#' @return Various return values depending on the function.
#' @keywords internal
#' @name resample_weights_utils
#' @aliases .create_weight_mapping .weighted_sd .effective_sample_size .validate_resample_weights
#' @export
#' @rdname resample_weights_utils
.get_resample_weights <- function(x) {
  rset_info <- attr(x, "rset_info")
  if (is.null(rset_info)) {
    return(NULL)
  }

  # Access weights from rset_info attributes using correct path
  weights <- rset_info$att[[".resample_weights"]]

  weights
}

#' @export
#' @rdname resample_weights_utils
.create_weight_mapping <- function(weights, id_names, metrics_data) {
  # Get unique combinations of ID columns from the metrics data
  unique_ids <- dplyr::distinct(metrics_data, !!!rlang::syms(id_names))

  if (nrow(unique_ids) != length(weights)) {
    cli::cli_warn(
      c(
        "Number of weights ({length(weights)}) does not match number of resamples ({nrow(unique_ids)}).",
        "Weights will be ignored."
      )
    )
    return(NULL)
  }

  # Add weights to the unique ID combinations
  unique_ids$.resample_weight <- weights
  unique_ids
}

#' @export
#' @rdname resample_weights_utils
.weighted_sd <- function(x, w) {
  if (all(is.na(x))) {
    return(NA_real_)
  }

  # Remove NA values and corresponding weights
  valid <- !is.na(x)
  x_valid <- x[valid]
  w_valid <- w[valid]

  if (length(x_valid) <= 1) {
    return(NA_real_)
  }

  # Calculate weighted variance
  weighted_var <-
    tibble::as_tibble_col(x) |>
    stats::cov.wt(wt = w, cor = FALSE)

  weighted_var <- weighted_var$cov[1, 1]

  sqrt(weighted_var)
}

#' @export
#' @rdname resample_weights_utils
.effective_sample_size <- function(w) {
  # Remove NA weights
  w <- w[!is.na(w)]

  if (length(w) == 0) {
    return(0)
  }

  # Calculate effective sample size: (sum of weights)^2 / sum of squared weights
  sum_w <- sum(w)
  sum_w_sq <- sum(w^2)

  if (sum_w_sq == 0) {
    return(0)
  }

  sum_w^2 / sum_w_sq
}

#' @export
#' @rdname resample_weights_utils
.validate_resample_weights <- function(weights, num_resamples) {
  if (is.null(weights)) {
    return(NULL)
  }

  if (!is.numeric(weights)) {
    cli::cli_abort("{.arg weights} must be numeric.")
  }

  if (length(weights) != num_resamples) {
    cli::cli_abort(
      "Length of {.arg weights} ({length(weights)}) must equal number of resamples ({num_resamples})."
    )
  }

  if (any(weights < 0)) {
    cli::cli_abort("{.arg weights} must be non-negative.")
  }

  if (all(weights == 0)) {
    cli::cli_abort("At least one weight must be positive.")
  }

  # Return normalized weights
  normalized_weights <- weights / sum(weights)

  # If equal, equivalent to not weighting
  expected_equal <- 1 / num_resamples
  if (
    isTRUE(all.equal(normalized_weights, rep(expected_equal, num_resamples)))
  ) {
    return(NULL)
  }

  return(normalized_weights)
}

#' Add resample weights to an rset object
#'
#' This function allows you to specify custom weights for resamples. Weights
#' are automatically normalized to sum to 1.
#'
#' @param rset An rset object from \pkg{rsample}.
#' @param weights A numeric vector of weights, one per resample. Weights will be
#' normalized.
#' @return The rset object with weights added as an attribute.
#' @details
#' Resampling weights are useful when assessment sets (i.e., held out data) have
#' different sizes or when you want to upweight certain resamples in the evaluation.
#' The weights are stored as an attribute and used automatically during
#' metric aggregation.
#' @seealso [calculate_resample_weights()], [extract_resample_weights()]
#' @examples
#' library(rsample)
#' folds <- vfold_cv(mtcars, v = 3)
#' # Give equal weight to all folds
#' weighted_folds <- add_resample_weights(folds, c(1, 1, 1))
#' # Emphasize the first fold
#' weighted_folds <- add_resample_weights(folds, c(0.5, 0.25, 0.25))
#' @export
add_resample_weights <- function(rset, weights) {
  if (!inherits(rset, "rset")) {
    cli::cli_abort("{.arg rset} must be an rset object.")
  }

  # Validate weights
  weights <- .validate_resample_weights(weights, nrow(rset))

  # Add weights as an attribute
  attr(rset, ".resample_weights") <- weights

  rset
}

#' Calculate resample weights from resample sizes
#'
#' This convenience function calculates weights proportional to the number of
#' observations in each resample's analysis set. Larger resamples get higher weights.
#' This ensures that resamples with more data have proportionally more influence
#' on the final aggregated metrics.
#'
#' @param rset An rset object from \pkg{rsample}.
#' @return A numeric vector of weights proportional to resample sizes, normalized
#'   to sum to 1.
#' @details
#' This is particularly useful for time-based resamples (e.g., expanding window CV)
#' or stratified sampling  where resamples might have slightly different sizes, in
#' which resamples are imbalanced.
#' @seealso [add_resample_weights()], [extract_resample_weights()]
#' @examples
#' library(rsample)
#' folds <- vfold_cv(mtcars, v = 3)
#' weights <- calculate_resample_weights(folds)
#' weighted_folds <- add_resample_weights(folds, weights)
#' @export
calculate_resample_weights <- function(rset) {
  if (!inherits(rset, "rset")) {
    cli::cli_abort("{.arg rset} must be an rset object.")
  }

  # Calculate the size of each analysis set
  resample_sizes <- purrr::map_int(rset$splits, ~ nrow(rsample::analysis(.x)))

  # Return weights proportional to resample sizes
  resample_sizes / sum(resample_sizes)
}

#' Extract resample weights from rset or tuning objects
#'
#' This function provides a consistent interface to access resample weights
#' regardless of whether they were added to an rset object or are stored
#' in `tune_results` after tuning.
#'
#' @param x An rset object with resample weights, or a `tune_results` object.
#' @return A numeric vector of resample weights, or NULL if no weights are present.
#' @export
#' @examples
#' \dontrun{
#' library(rsample)
#' folds <- vfold_cv(mtcars, v = 3)
#' weighted_folds <- add_resample_weights(folds, c(0.2, 0.3, 0.5))
#' extract_resample_weights(weighted_folds)
#' }
extract_resample_weights <- function(x) {
  if (inherits(x, "rset")) {
    # For rset objects, weights are stored as an attribute
    res <- attr(x, ".resample_weights")
  } else if (inherits(x, c("tune_results", "resample_results"))) {
    # For tune results, use the internal function
    res <- .get_resample_weights(x)
  } else {
    cli::cli_abort("{.arg x} must be an rset or tune_results object.")
  }
  res
}

# ------------------------------------------------------------------------------

#' Save most recent results to search path
#' @param x An object.
#' @return NULL, invisibly.
#' @details The function will assign `x` to `.Last.tune.result` and put it in
#' the search path.
#' @export
.stash_last_result <- function(x) {
  if (!"org:r-lib" %in% search()) {
    do.call(
      "attach",
      list(new.env(), pos = length(search()), name = "org:r-lib")
    )
  }
  env <- as.environment("org:r-lib")
  env$.Last.tune.result <- x
  invisible(NULL)
}

test_that("check for no arguments pass to ...", {
  expect_snapshot(
    empty_ellipses(a = 5)
  )
  expect_snapshot(
    empty_ellipses(a = 5, b = 1)
  )
  expect_snapshot(
    empty_ellipses(5)
  )
  expect_snapshot(
    empty_ellipses(1, 5)
  )
  expect_snapshot(
    empty_ellipses(1, second = 2)
  )
})

rec_tune_1 <-
  recipes::recipe(mpg ~ ., data = mtcars) |>
  recipes::step_normalize(recipes::all_predictors()) |>
  recipes::step_pca(recipes::all_predictors(), num_comp = tune())

rec_no_tune_1 <-
  recipes::recipe(mpg ~ ., data = mtcars) |>
  recipes::step_normalize(recipes::all_predictors())

lm_mod <- parsnip::linear_reg() |> parsnip::set_engine("lm")

svm_mod <- parsnip::svm_rbf(mode = "regression", cost = tune()) |>
  parsnip::set_engine("kernlab")

iter1 <- 2
iter2 <- 2
iterT <- iter1 + iter2

# ------------------------------------------------------------------------------

test_that("tune recipe only", {
  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(rec_tune_1) |>
    add_model(lm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(1, 15)))
  folds <- rsample::vfold_cv(mtcars)
  control <- control_bayes(extract = identity)

  expect_snapshot({
    set.seed(2)
    res <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2,
      control = control
    )
  })

  res_est <- collect_metrics(res)
  res_workflow <- res$.extracts[[1]]$.extracts[[1]]

  expect_equal(res, .Last.tune.result)

  # Ensure tunable parameters in recipe are finalized
  num_comp <- res_workflow$pre$actions$recipe$recipe$steps[[2]]$num_comp

  expect_equal(unique(res$id), folds$id)
  expect_equal(nrow(res_est), iterT * 2)
  expect_equal(sum(res_est$.metric == "rmse"), iterT)
  expect_equal(sum(res_est$.metric == "rsq"), iterT)
  expect_equal(dplyr::n_distinct(res_est$.config), iterT)
  expect_equal(res_est$n, rep(10, iterT * 2))
  expect_false(identical(num_comp, expr(tune())))
  expect_true(res_workflow$trained)
  expect_null(.get_tune_eval_times(res))
  expect_null(.get_tune_eval_time_target(res))

  # test verbose options
  set.seed(1)
  expect_snapshot(
    tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2,
      control = control_bayes(verbose = TRUE)
    )
  )

  set.seed(1)
  expect_snapshot(
    tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2,
      control = control_bayes(verbose_iter = TRUE)
    )
  )

  set.seed(1)
  expect_snapshot(
    tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2,
      control = control_bayes(verbose_iter = TRUE, verbose = TRUE)
    )
  )
})

# ------------------------------------------------------------------------------

test_that("tune model only (with recipe)", {
  skip_if_not_installed("kernlab")

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(rec_no_tune_1) |>
    add_model(svm_mod)
  pset <- extract_parameter_set_dials(wflow)
  folds <- rsample::vfold_cv(mtcars)
  suppressMessages({
    res <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2
    )
  })

  expect_equal(res, .Last.tune.result)
  expect_equal(unique(res$id), folds$id)
  res_est <- collect_metrics(res)
  expect_equal(nrow(res_est), iterT * 2)
  expect_equal(sum(res_est$.metric == "rmse"), iterT)
  expect_equal(sum(res_est$.metric == "rsq"), iterT)
  expect_equal(dplyr::n_distinct(res_est$.config), iterT)
  expect_equal(res_est$n, rep(10, iterT * 2))
})

# ------------------------------------------------------------------------------

test_that("tune model only (with variables)", {
  skip_if_not_installed("kernlab")

  set.seed(4400)

  wflow <- workflow() |>
    add_variables(mpg, everything()) |>
    add_model(svm_mod)

  pset <- extract_parameter_set_dials(wflow)

  folds <- rsample::vfold_cv(mtcars)

  suppressMessages({
    res <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2
    )
  })

  expect_equal(unique(res$id), folds$id)

  res_est <- collect_metrics(res)

  expect_equal(nrow(res_est), iterT * 2)
  expect_equal(sum(res_est$.metric == "rmse"), iterT)
  expect_equal(sum(res_est$.metric == "rsq"), iterT)
  expect_equal(dplyr::n_distinct(res_est$.config), iterT)
  expect_equal(res_est$n, rep(10, iterT * 2))
})

# ------------------------------------------------------------------------------

test_that("tune model only (with recipe, multi-predict)", {
  skip_if_not_installed("kernlab")

  skip_on_cran()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(rec_no_tune_1) |>
    add_model(svm_mod)
  pset <- extract_parameter_set_dials(wflow)
  folds <- rsample::vfold_cv(mtcars)
  suppressMessages({
    res <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2
    )
  })

  expect_equal(unique(res$id), folds$id)
  expect_equal(
    colnames(res$.metrics[[1]]),
    c("cost", ".metric", ".estimator", ".estimate", ".config")
  )
  res_est <- collect_metrics(res)
  expect_equal(nrow(res_est), iterT * 2)
  expect_equal(sum(res_est$.metric == "rmse"), iterT)
  expect_equal(sum(res_est$.metric == "rsq"), iterT)
  expect_equal(dplyr::n_distinct(res_est$.config), iterT)
  expect_equal(res_est$n, rep(10, iterT * 2))
})

# ------------------------------------------------------------------------------

test_that("tune model and recipe", {
  skip_if_not_installed("kernlab")

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(rec_tune_1) |>
    add_model(svm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(1, 3)))
  folds <- rsample::vfold_cv(mtcars)
  suppressMessages({
    res <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2
    )
  })

  expect_equal(unique(res$id), folds$id)
  expect_equal(
    colnames(res$.metrics[[1]]),
    c("cost", "num_comp", ".metric", ".estimator", ".estimate", ".config")
  )
  res_est <- collect_metrics(res)
  expect_equal(nrow(res_est), iterT * 2)
  expect_equal(sum(res_est$.metric == "rmse"), iterT)
  expect_equal(sum(res_est$.metric == "rsq"), iterT)
  expect_equal(dplyr::n_distinct(res_est$.config), iterT)
  expect_equal(res_est$n, rep(10, iterT * 2))
})

# ------------------------------------------------------------------------------

test_that("tune model and recipe (multi-predict)", {
  skip_if_not_installed("kernlab")
  skip_on_cran()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(rec_tune_1) |>
    add_model(svm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(2, 3)))
  grid <- dials::grid_regular(pset, levels = c(3, 2))
  folds <- rsample::vfold_cv(mtcars)
  suppressMessages({
    res <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2
    )
  })

  expect_equal(unique(res$id), folds$id)
  res_est <- collect_metrics(res)
  expect_equal(nrow(res_est), iterT * 2)
  expect_equal(sum(res_est$.metric == "rmse"), iterT)
  expect_equal(sum(res_est$.metric == "rsq"), iterT)
  expect_equal(dplyr::n_distinct(res_est$.config), iterT)
  expect_equal(res_est$n, rep(10, iterT * 2))
})

# ------------------------------------------------------------------------------

test_that("tune recipe only - failure in recipe is caught elegantly", {
  skip_if_not_installed("splines2")

  # With tune_grid() this tests for NA values in the grid.
  # This is not applicable for tune_bayes().

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_b(disp, deg_free = tune())

  model <- parsnip::linear_reg(mode = "regression") |>
    parsnip::set_engine("lm")

  # NA values not allowed in recipe
  cars_grid <- tibble(deg_free = c(3, NA_real_, 4))

  expect_snapshot({
    cars_init_res <- tune_grid(
      model,
      preprocessor = rec,
      resamples = data_folds,
      grid = cars_grid
    )
  })

  expect_snapshot({
    set.seed(283)
    cars_bayes_res <- tune_bayes(
      model,
      preprocessor = rec,
      resamples = data_folds,
      initial = cars_init_res,
      iter = 2
    )
  })

  exp_failures <- nrow(data_folds) * sum(!complete.cases(cars_grid))
  obs_init_failures <- collect_notes(cars_init_res) |>
    filter(type == "error") |>
    nrow()
  obs_failures <- collect_notes(cars_bayes_res) |>
    filter(type == "error") |>
    nrow()

  exp_init_grid_res <-
    cars_grid |> tidyr::drop_na() |> distinct(deg_free) |> nrow()

  expect_equal(obs_init_failures, obs_failures)
  expect_equal(obs_failures, exp_failures)

  all_notes <- collect_notes(cars_bayes_res)
  expect_equal(nrow(all_notes), 5L)

  expect_equal(
    collect_metrics(cars_bayes_res) |> distinct(deg_free) |> nrow(),
    3
  )
})

test_that("tune model only - failure in recipe is caught elegantly", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("kernlab")

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  # NA values not allowed in recipe
  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_b(disp, deg_free = NA_real_)

  expect_snapshot({
    cars_res <- tune_bayes(
      svm_mod,
      preprocessor = rec,
      resamples = data_folds
    )
  })

  expect_s3_class(cars_res, "iteration_results")
})

test_that("tune model only - failure in formula is caught elegantly", {
  skip_if_not_installed("kernlab")

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  # these terms don't exist!
  wflow <- workflow() |>
    add_formula(y ~ z) |>
    add_model(svm_mod)

  expect_snapshot({
    cars_res <- tune_bayes(
      wflow,
      resamples = data_folds,
      control = control_bayes(
        extract = function(x) {
          1
        },
        save_pred = TRUE
      )
    )
  })

  expect_s3_class(cars_res, "iteration_results")
})

test_that("tune model and recipe - failure in recipe is caught elegantly", {
  skip("test is not implemented for tune_bayes()")
  skip_if_not_installed("splines2")
  skip_if_not_installed("kernlab")

  # With tune_grid() this tests for NA values in the grid.
  # This is not applicable for tune_bayes().

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_b(disp, deg_free = tune())

  # NA values not allowed in recipe
  cars_grid <- tibble(
    deg_free = c(3L, NA_real_, 10L),
    cost = c(0.1, 0.01, 0.001)
  )

  suppressMessages({
    cars_init_res <- tune_grid(
      svm_mod,
      preprocessor = rec,
      resamples = data_folds,
      grid = cars_grid
    )
  })

  suppressMessages({
    set.seed(283)
    cars_bayes_res <- tune_bayes(
      svm_mod,
      preprocessor = rec,
      resamples = data_folds,
      initial = cars_init_res,
      iter = 2
    )
  })

  exp_failures <- nrow(data_folds) * sum(!complete.cases(cars_grid))
  obs_init_failures <- collect_notes(cars_init_res) |>
    filter(type == "error") |>
    nrow()
  obs_failures <- collect_notes(cars_bayes_res) |>
    filter(type == "error") |>
    nrow()

  exp_init_grid_res <-
    cars_grid |> tidyr::drop_na() |> distinct(deg_free, cost) |> nrow()

  expect_equal(obs_init_failures, obs_failures)
  expect_equal(obs_failures, exp_failures)

  all_notes <- collect_notes(cars_bayes_res)
  expect_equal(nrow(all_notes), 7L)

  expect_equal(
    collect_metrics(cars_bayes_res) |> distinct(deg_free, cost) |> nrow(),
    exp_init_grid_res + 2
  )
})

test_that("argument order gives an error for recipes", {
  expect_snapshot(error = TRUE, {
    tune_bayes(
      rec_tune_1,
      model = lm_mod,
      resamples = rsample::vfold_cv(mtcars, v = 2),
      param_info = extract_parameter_set_dials(rec_tune_1),
      iter = iter1,
      initial = iter2
    )
  })
})

test_that("argument order gives an error for formula", {
  skip_if_not_installed("kernlab")

  expect_snapshot(error = TRUE, {
    tune_bayes(
      mpg ~ .,
      svm_mod,
      resamples = rsample::vfold_cv(mtcars, v = 2),
      param_info = extract_parameter_set_dials(svm_mod),
      initial = iter1,
      iter = iter2
    )
  })
})

test_that("retain extra attributes and saved GP candidates", {
  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(rec_tune_1) |>
    add_model(lm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(1, 5)))
  folds <- rsample::vfold_cv(mtcars)
  ctrl <- control_bayes(save_gp_scoring = TRUE)
  suppressMessages({
    res <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2,
      control = ctrl
    )
  })
  att <- attributes(res)
  att_names <- names(att)
  expect_true(any(att_names == "metrics"))
  expect_true(any(att_names == "outcomes"))
  expect_true(any(att_names == "parameters"))

  expect_true(is.character(att$outcomes))
  expect_true(att$outcomes == "mpg")
  expect_true(inherits(att$parameters, "parameters"))
  expect_true(inherits(att$metrics, "metric_set"))

  files <- list.files(path = tempdir(), pattern = "^gp_candidates")
  expect_true(length(files) == iter2)

  current_objs <- c(ls(), "current_objs")
  load(file.path(tempdir(), "gp_candidates_1.RData"))
  new_obj <- ls()
  expect_snapshot(setdiff(new_obj, current_objs))

  expect_snapshot(
    res2 <- tune_bayes(
      wflow,
      resamples = folds,
      param_info = pset,
      initial = iter1,
      iter = iter2,
      control = control_bayes(save_workflow = TRUE)
    )
  )
  expect_null(attr(res, "workflow"))
  expect_true(inherits(attr(res2, "workflow"), "workflow"))
})

# ------------------------------------------------------------------------------

test_that("too few starting values", {
  options(width = 120)
  # TODO Add specific checks with racing objects once finetune is released
  expect_silent(tune:::check_bayes_initial_size(5, 30, FALSE))

  expect_snapshot(tune:::check_bayes_initial_size(5, 3, FALSE))
  expect_snapshot(tune:::check_bayes_initial_size(5, 3, TRUE))
  expect_snapshot(tune:::check_bayes_initial_size(2, 2, FALSE))

  expect_snapshot(error = TRUE, tune:::check_bayes_initial_size(5, 1, FALSE))
  expect_snapshot(error = TRUE, tune:::check_bayes_initial_size(5, 1, TRUE))

  expect_snapshot(error = TRUE, tune:::check_bayes_initial_size(1, 1, FALSE))
})

# ------------------------------------------------------------------------------

test_that("missing performance values", {
  skip_if(new_rng_snapshots)
  skip_if(packageVersion("dplyr") < "1.1.1")
  skip_if_not_installed("modeldata")
  skip_if_not_installed("pec")

  data(ames, package = "modeldata")

  mod <- parsnip::decision_tree(cost_complexity = tune()) |>
    parsnip::set_mode("regression")

  set.seed(1)
  split <- rsample::initial_validation_split(ames)
  folds <- rsample::validation_set(split)

  expect_snapshot({
    set.seed(1)
    res <-
      mod |>
      tune_bayes(
        Sale_Price ~
          Neighborhood +
          Gr_Liv_Area +
          Year_Built +
          Bldg_Type +
          Latitude +
          Longitude,
        resamples = folds,
        initial = 3,
        metrics = yardstick::metric_set(rsq),
        param_info = parameters(dials::cost_complexity(c(-2, 0)))
      )
  })

  expect_snapshot(error = TRUE, {
    set.seed(2)
    res_fail <-
      mod |>
      tune_bayes(
        Sale_Price ~
          Neighborhood +
          Gr_Liv_Area +
          Year_Built +
          Bldg_Type +
          Latitude +
          Longitude,
        resamples = folds,
        initial = 5,
        metrics = yardstick::metric_set(rsq),
        param_info = parameters(dials::cost_complexity(c(0, 0.5)))
      )
  })
})

# ------------------------------------------------------------------------------
test_that("tune_bayes() output for `iter` edge cases (#721)", {
  skip_if_not_installed("kknn")

  # for `iter = 0`, ought to match `tune_grid()`
  boots <- rsample::bootstraps(mtcars)
  wf <-
    workflows::workflow(
      mpg ~ .,
      parsnip::nearest_neighbor("regression", "kknn", neighbors = tune())
    )

  ctrl_bayes <- control_bayes(seed = 1)

  set.seed(1)
  res_bayes <- tune_bayes(
    wf,
    boots,
    iter = 0,
    initial = 10,
    control = ctrl_bayes
  )

  set.seed(1)
  res_grid <- tune_grid(wf, boots)

  expect_equal(
    collect_metrics(res_bayes) |> dplyr::select(-.iter),
    collect_metrics(res_grid)
  )

  # for `iter < 0`, ought to error
  expect_snapshot(
    error = TRUE,
    tune_bayes(wf, boots, iter = -1)
  )

  expect_snapshot(
    error = TRUE,
    tune_bayes(wf, boots, iter = c(-1, 0, 1))
  )

  expect_snapshot(
    error = TRUE,
    tune_bayes(wf, boots, iter = c(0, 1, 2))
  )

  expect_snapshot(
    error = TRUE,
    tune_bayes(wf, boots, iter = NA)
  )

  expect_snapshot(
    error = TRUE,
    tune_bayes(wf, boots, iter = NULL)
  )
})

# ------------------------------------------------------------------------------

test_that("tune_bayes loggining doesn't error with failed model", {
  # no failed results:
  res_1 <- purrr::map_dfr(
    ames_iter_search$.metrics,
    tune:::set_config,
    config = "beratna"
  )
  expect_true(all(res_1$.config == "beratna"))

  has_failure <- tune:::vec_list_rowwise(ames_iter_search$.metrics[[1]])[1:3]
  has_failure[2] <- list(NULL)
  res_2 <- purrr::map(
    has_failure,
    tune:::set_config,
    config = "sasa ke?"
  )
  expect_null(res_2[[2]])
  expect_equal(res_2[[1]]$.config, "sasa ke?")
  expect_equal(res_2[[3]]$.config, "sasa ke?")
})

test_that("exponential decay", {
  expect_equal(
    expo_decay(1, start_val = 0, limit_val = 1, slope = 1 / 5),
    0
  )
  expect_equal(
    expo_decay(1000, start_val = 0, limit_val = 1, slope = 1 / 5),
    1
  )
  expect_equal(
    expo_decay(10, start_val = 0, limit_val = 50, slope = 1),
    (1 - exp(-9)) * 50
  )
})

# ------------------------------------------------------------------------------

test_that("in-line formulas on outcome", {
  # see issues 121
  w1 <-
    workflow() |>
    add_formula(log(mpg) ~ .) |>
    add_model(parsnip::linear_reg() |> parsnip::set_engine("lm"))

  expect_no_error(
    f1 <- fit_resamples(w1, resamples = rsample::vfold_cv(mtcars))
  )
  expect_true(inherits(f1, "resample_results"))

  w2 <-
    workflow() |>
    add_recipe(
      recipes::recipe(mpg ~ ., data = mtcars) |> recipes::step_log(mpg)
    ) |>
    add_model(parsnip::linear_reg() |> parsnip::set_engine("lm"))

  expect_no_error(
    f2 <- fit_resamples(w2, resamples = rsample::vfold_cv(mtcars))
  )
  expect_true(inherits(f2, "resample_results"))
})

# ------------------------------------------------------------------------------

test_that("empty ellipses", {
  expect_no_error(tune:::empty_ellipses())
  expect_snapshot(tune:::empty_ellipses(a = 1))
})

# ------------------------------------------------------------------------------

test_that("accessor functions", {
  load(test_path("data", "test_objects.RData"))

  expect_equal(
    .get_tune_parameter_names(mt_spln_knn_bo),
    attributes(mt_spln_knn_bo)$parameters$id
  )
  attr(mt_spln_knn_bo, "parameters") <- NULL
  expect_equal(.get_tune_parameter_names(mt_spln_knn_bo), character(0))

  expect_equal(.get_tune_metrics(mt_knn_bo), attributes(mt_knn_bo)$metrics)
  attr(mt_knn_bo, "metrics") <- NULL
  expect_null(.get_tune_metrics(mt_knn_bo))

  expect_equal(
    .get_tune_metric_names(mt_spln_knn_bo),
    names(attributes(attributes(mt_spln_knn_bo)$metrics)$metrics)
  )
  attr(mt_spln_knn_bo, "metrics") <- NULL
  expect_equal(.get_tune_metric_names(mt_spln_knn_bo), character(0))

  expect_equal(
    .get_tune_outcome_names(mt_spln_knn_bo),
    attributes(mt_spln_knn_bo)$outcomes
  )
  attr(mt_spln_knn_bo, "outcomes") <- NULL
  expect_equal(.get_tune_outcome_names(mt_spln_knn_bo), character(0))
})

test_that("accessor functions", {
  skip_if(utils::packageVersion("dials") <= "0.0.7")

  load(test_path("data", "test_objects.RData"))

  expect_equal(
    tibble::as_tibble(.get_tune_parameters(mt_knn_bo)),
    tibble::as_tibble(attributes(mt_knn_bo)$parameters)
  )
  attr(mt_knn_bo, "parameters") <- NULL
  expect_equal(.get_tune_parameters(mt_knn_bo), tibble::tibble())
})

# ------------------------------------------------------------------------------

test_that("rsample fingerprinting", {
  expect_equal(
    .get_fingerprint(ames_grid_search),
    "7a961db8a275a9f0d9b6733a967cea03"
  )
  expect_equal(
    .get_fingerprint(ames_iter_search),
    "7a961db8a275a9f0d9b6733a967cea03"
  )
})

test_that("formula method", {
  set.seed(23598723)
  split <- rsample::initial_split(mtcars)

  f <- mpg ~ cyl + poly(disp, 2) + hp + drat + wt + qsec + vs + am + gear + carb
  lm_fit <- lm(f, data = rsample::training(split))
  test_pred <- predict(lm_fit, rsample::testing(split))
  rmse_test <- yardstick::rsq_vec(
    rsample::testing(split) |> pull(mpg),
    test_pred
  )

  res <- parsnip::linear_reg() |>
    parsnip::set_engine("lm") |>
    last_fit(f, split)

  expect_equal(res, .Last.tune.result)

  expect_equal(
    coef(extract_fit_engine(res$.workflow[[1]])),
    coef(lm_fit),
    ignore_attr = TRUE
  )
  expect_equal(res$.metrics[[1]]$.estimate[[2]], rmse_test)
  expect_equal(res$.predictions[[1]]$.pred, unname(test_pred))
  expect_true(res$.workflow[[1]]$trained)
  expect_equal(
    nrow(predict(res$.workflow[[1]], rsample::testing(split))),
    nrow(rsample::testing(split))
  )
  expect_null(.get_tune_eval_times(res))
  expect_null(.get_tune_eval_time_target(res))
})

test_that("recipe method", {
  set.seed(23598723)
  split <- rsample::initial_split(mtcars)

  f <- mpg ~ cyl + poly(disp, 2) + hp + drat + wt + qsec + vs + am + gear + carb
  lm_fit <- lm(f, data = rsample::training(split))
  test_pred <- predict(lm_fit, rsample::testing(split))
  rmse_test <- yardstick::rsq_vec(
    rsample::testing(split) |> pull(mpg),
    test_pred
  )

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_poly(disp)
  res <- parsnip::linear_reg() |>
    parsnip::set_engine("lm") |>
    last_fit(rec, split)

  expect_equal(
    sort(coef(extract_fit_engine(res$.workflow[[1]]))),
    sort(coef(lm_fit)),
    ignore_attr = TRUE
  )
  expect_equal(res$.metrics[[1]]$.estimate[[2]], rmse_test)
  expect_equal(res$.predictions[[1]]$.pred, unname(test_pred))
  expect_true(res$.workflow[[1]]$trained)
  expect_equal(
    nrow(predict(res$.workflow[[1]], rsample::testing(split))),
    nrow(rsample::testing(split))
  )
})

test_that("model_fit method", {
  library(parsnip)

  lm_fit <- linear_reg() |> fit(mpg ~ ., data = mtcars)

  expect_snapshot(last_fit(lm_fit), error = TRUE)
})

test_that("workflow method", {
  library(parsnip)

  lm_fit <- workflows::workflow(mpg ~ ., linear_reg()) |> fit(data = mtcars)

  expect_snapshot(last_fit(lm_fit), error = TRUE)
})

test_that("collect metrics of last fit", {
  set.seed(23598723)
  split <- rsample::initial_split(mtcars)
  f <- mpg ~ cyl + poly(disp, 2) + hp + drat + wt + qsec + vs + am + gear + carb
  res <- parsnip::linear_reg() |>
    parsnip::set_engine("lm") |>
    last_fit(f, split)
  met <- collect_metrics(res)
  expect_true(inherits(met, "tbl_df"))
  expect_equal(names(met), c(".metric", ".estimator", ".estimate", ".config"))
})


test_that("ellipses with last_fit", {
  options(width = 200, pillar.advice = FALSE, pillar.min_title_chars = Inf)

  set.seed(23598723)
  split <- rsample::initial_split(mtcars)
  f <- mpg ~ cyl + poly(disp, 2) + hp + drat + wt + qsec + vs + am + gear + carb

  expect_snapshot(
    linear_reg() |>
      set_engine("lm") |>
      last_fit(f, split, something = "wrong")
  )
})

test_that("argument order gives errors for recipe/formula", {
  options(width = 200, pillar.advice = FALSE, pillar.min_title_chars = Inf)

  set.seed(23598723)
  split <- rsample::initial_split(mtcars)

  f <- mpg ~ cyl + poly(disp, 2) + hp + drat + wt + qsec + vs + am + gear + carb

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |> recipes::step_poly(disp)
  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect_snapshot(error = TRUE, {
    last_fit(rec, lin_mod, split)
  })
  expect_snapshot(error = TRUE, {
    last_fit(f, lin_mod, split)
  })
})

test_that("same results of last_fit() and fit() (#300)", {
  skip_if_not_installed("randomForest")

  rf <- parsnip::rand_forest(mtry = 2, trees = 5) |>
    parsnip::set_engine("randomForest") |>
    parsnip::set_mode("regression")

  wflow <- workflows::workflow() |>
    workflows::add_model(rf) |>
    workflows::add_formula(mpg ~ .)

  set.seed(23598723)
  split <- rsample::initial_split(mtcars)

  set.seed(1)
  lf_obj <- last_fit(wflow, split = split)

  set.seed(1)
  r_obj <- fit(wflow, data = rsample::analysis(split))
  r_pred <- predict(r_obj, rsample::assessment(split))

  expect_equal(
    lf_obj$.predictions[[1]]$.pred,
    r_pred$.pred
  )
})


test_that("`last_fit()` when objects need tuning", {
  skip_if_not_installed("splines2")

  options(width = 200, pillar.advice = FALSE, pillar.min_title_chars = Inf)

  rec <- recipe(mpg ~ ., data = mtcars) |>
    step_spline_natural(disp, deg_free = tune())
  spec_1 <- linear_reg(penalty = tune()) |> set_engine("glmnet")
  spec_2 <- linear_reg()
  wflow_1 <- workflow(rec, spec_1)
  wflow_2 <- workflow(mpg ~ ., spec_1)
  wflow_3 <- workflow(rec, spec_2)
  split <- rsample::initial_split(mtcars)

  expect_snapshot_error(last_fit(wflow_1, split))
  expect_snapshot_error(last_fit(wflow_2, split))
  expect_snapshot_error(last_fit(wflow_3, split))
})

test_that("last_fit() excludes validation set for initial_validation_split objects", {
  skip_if_not_installed("modeldata")
  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(23598723)
  split <- rsample::initial_validation_split(ames)

  f <- Sale_Price ~ Gr_Liv_Area + Year_Built
  lm_fit <- lm(f, data = rsample::training(split))
  test_pred <- predict(lm_fit, rsample::testing(split))
  rmse_test <- yardstick::rsq_vec(
    rsample::testing(split) |> pull(Sale_Price),
    test_pred
  )

  res <- parsnip::linear_reg() |>
    parsnip::set_engine("lm") |>
    last_fit(f, split)

  expect_equal(res, .Last.tune.result)

  expect_equal(
    coef(extract_fit_engine(res$.workflow[[1]])),
    coef(lm_fit),
    ignore_attr = TRUE
  )
  expect_equal(res$.metrics[[1]]$.estimate[[2]], rmse_test)
  expect_equal(res$.predictions[[1]]$.pred, unname(test_pred))
  expect_true(res$.workflow[[1]]$trained)
  expect_equal(
    nrow(predict(res$.workflow[[1]], rsample::testing(split))),
    nrow(rsample::testing(split))
  )
})

test_that("last_fit() can include validation set for initial_validation_split objects", {
  skip_if_not_installed("modeldata")
  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(23598723)
  split <- rsample::initial_validation_split(ames)

  f <- Sale_Price ~ Gr_Liv_Area + Year_Built
  train_val <- rbind(rsample::training(split), rsample::validation(split))
  lm_fit <- lm(f, data = train_val)
  test_pred <- predict(lm_fit, rsample::testing(split))
  rmse_test <- yardstick::rsq_vec(
    rsample::testing(split) |> pull(Sale_Price),
    test_pred
  )

  res <- parsnip::linear_reg() |>
    parsnip::set_engine("lm") |>
    last_fit(f, split, add_validation_set = TRUE)

  expect_equal(res, .Last.tune.result)

  expect_equal(
    coef(extract_fit_engine(res$.workflow[[1]])),
    coef(lm_fit),
    ignore_attr = TRUE
  )
  expect_equal(res$.metrics[[1]]$.estimate[[2]], rmse_test)
  expect_equal(res$.predictions[[1]]$.pred, unname(test_pred))
  expect_true(res$.workflow[[1]]$trained)
  expect_equal(
    nrow(predict(res$.workflow[[1]], rsample::testing(split))),
    nrow(rsample::testing(split))
  )
})

test_that("can use `last_fit()` with a workflow - postprocessor (requires training)", {
  skip_if_not_installed("mgcv")
  skip_if_not_installed("tailor", minimum_version = "0.0.0.9002")
  skip_if_not_installed("probably")

  y <- seq(0, 7, .001)
  dat <- data.frame(y = y, x = y + (y - 3)^2)

  dat

  set.seed(1)
  split <- rsample::initial_split(dat)

  set.seed(1)
  internal_calibration_split <- rsample::internal_calibration_split(
    split,
    split_args = list()
  )

  # ----------------------------------------------------------------------------

  wflow <-
    workflows::workflow(
      y ~ x,
      parsnip::linear_reg()
    ) |>
    workflows::add_tailor(
      tailor::tailor() |> tailor::adjust_numeric_calibration("linear")
    )

  # ----------------------------------------------------------------------------

  set.seed(1)
  last_fit_res <-
    last_fit(
      wflow,
      split
    )

  last_fit_preds <- collect_predictions(last_fit_res)
  last_fit_wflow <- extract_workflow(last_fit_res)
  expect_true(is_trained_workflow(last_fit_wflow))
  last_fit_cal <-
    last_fit_wflow |>
    extract_postprocessor() |>
    purrr::pluck("adjustments") |>
    purrr::pluck(1) |>
    purrr::pluck("results") |>
    purrr::pluck("fit") |>
    purrr::pluck("estimates") |>
    purrr::pluck(1) |>
    purrr::pluck("estimate")

  # ----------------------------------------------------------------------------

  # TODO inner split is getting different random numbers; check against no cal instead
  set.seed(1)
  wflow_res <-
    generics::fit(
      wflow,
      rsample::analysis(internal_calibration_split),
      data_calibration = rsample::calibration(internal_calibration_split)
    )

  wflow_cal <-
    wflow_res |>
    extract_postprocessor() |>
    purrr::pluck("adjustments") |>
    purrr::pluck(1) |>
    purrr::pluck("results") |>
    purrr::pluck("fit") |>
    purrr::pluck("estimates") |>
    purrr::pluck(1) |>
    purrr::pluck("estimate")

  wflow_preds <- predict(wflow_res, rsample::assessment(split))

  # ----------------------------------------------------------------------------

  expect_equal(last_fit_cal, wflow_cal)

  expect_equal(last_fit_preds[".pred"], wflow_preds)
})

test_that("can use `last_fit()` with a workflow - postprocessor (does not require training)", {
  skip_if_not_installed("tailor", minimum_version = "0.0.0.9002")
  skip_if_not_installed("probably")

  y <- seq(0, 7, .001)
  dat <- data.frame(y = y, x = y + (y - 3)^2)

  dat

  set.seed(1)
  split <- rsample::initial_split(dat)

  wflow <-
    workflows::workflow(
      y ~ x,
      parsnip::linear_reg()
    ) |>
    workflows::add_tailor(
      tailor::tailor() |> tailor::adjust_numeric_range(lower_limit = 1)
    )

  set.seed(1)
  last_fit_res <-
    last_fit(
      wflow,
      split
    )

  last_fit_preds <- collect_predictions(last_fit_res)

  set.seed(1)
  wflow_res <- generics::fit(wflow, rsample::analysis(split))
  wflow_preds <- predict(wflow_res, rsample::assessment(split))

  expect_equal(last_fit_preds[".pred"], wflow_preds)
})

# fit_resamples() --------------------------------------------------------------

test_that("`fit_resamples()` returns a `resample_result` object", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  result <- lin_mod |>
    fit_resamples(mpg ~ ., folds)

  expect_s3_class(result, "resample_results")

  expect_equal(result, .Last.tune.result)

  expect_null(.get_tune_eval_times(result))
  expect_null(.get_tune_eval_time_target(result))
})

test_that("can use `fit_resamples()` with a formula", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect_warning(
    result <- lin_mod |>
      fit_resamples(mpg ~ ., folds),
    NA
  )

  expect_equal(nrow(result$.metrics[[1]]), 2L)
})

test_that("can use `fit_resamples()` with a recipe", {
  skip_if_not_installed("splines2")

  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp) |>
    recipes::step_spline_natural(wt)

  lin_mod <- linear_reg() |>
    set_engine("lm")

  # Ensure the recipes are prepped and returned
  control <- control_resamples(extract = function(x) x)

  result <- fit_resamples(lin_mod, rec, folds, control = control)

  prepped_rec <- extract_recipe(result$.extracts[[1]][[1]][[1]])

  expect_true(prepped_rec$steps[[1]]$trained)
  expect_true(prepped_rec$steps[[2]]$trained)

  expect_equal(nrow(result$.metrics[[1]]), 2L)
})

test_that("can use `fit_resamples()` with a workflow - recipe", {
  skip_if_not_installed("splines2")

  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp) |>
    recipes::step_spline_natural(wt)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  workflow <- workflow() |>
    add_recipe(rec) |>
    add_model(lin_mod)

  expect <- fit_resamples(lin_mod, rec, folds)

  result <- fit_resamples(workflow, folds)

  expect_equal(collect_metrics(expect), collect_metrics(result))
})

test_that("can use `fit_resamples()` with a workflow - variables", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  workflow <- workflow() |>
    add_variables(mpg, c(cyl, disp)) |>
    add_model(lin_mod)

  expect <- fit_resamples(lin_mod, mpg ~ cyl + disp, folds)

  result <- fit_resamples(workflow, folds)

  expect_equal(collect_metrics(expect), collect_metrics(result))
})

test_that("can use `fit_resamples()` with a workflow - formula", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  workflow <- workflow() |>
    add_formula(mpg ~ cyl + disp) |>
    add_model(lin_mod)

  expect <- fit_resamples(lin_mod, mpg ~ cyl + disp, folds)

  result <- fit_resamples(workflow, folds)

  expect_equal(collect_metrics(expect), collect_metrics(result))
})

test_that("extracted workflow is finalized", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  workflow <- workflow() |>
    add_variables(mpg, c(cyl, disp)) |>
    add_model(lin_mod)

  control <- control_resamples(extract = identity)

  result <- fit_resamples(workflow, folds, control = control)
  result_workflow <- result$.extracts[[1]]$.extracts[[1]]

  expect_true(result_workflow$trained)
})

test_that("can use `fit_resamples()` with a workflow - postprocessor (requires training)", {
  skip_if_not_installed("tailor", minimum_version = "0.0.0.9002")
  skip_if_not_installed("probably")

  y <- seq(0, 7, .001)
  dat <- data.frame(y = y, x = y + (y - 3)^2)

  folds <- rsample::vfold_cv(dat, v = 2)

  wflow_no_post <-
    workflows::workflow(
      y ~ x,
      parsnip::linear_reg()
    )

  wflow_post <-
    wflow_no_post |>
    workflows::add_tailor(
      tailor::tailor() |> tailor::adjust_numeric_calibration("linear")
    )

  set.seed(1)
  no_post_res <-
    fit_resamples(
      wflow_no_post,
      folds,
      control = control_resamples(save_pred = TRUE)
    )

  set.seed(1)
  post_res <-
    fit_resamples(
      wflow_post,
      folds,
      control = control_resamples(save_pred = TRUE, extract = identity)
    )

  no_post_preds <-
    collect_predictions(no_post_res) |>
    dplyr::rename(no_post = .pred)
  post_preds <-
    collect_predictions(post_res) |>
    dplyr::rename(post = .pred) |>
    dplyr::select(.row, post, id, .config)

  both_preds <-
    dplyr::full_join(no_post_preds, post_preds, by = c(".row", "id", ".config"))

  expect_true(all(both_preds$post != both_preds$no_post))
  expect_true(
    post_res$.extracts[[1]]$.extracts[[1]] |> tune:::has_postprocessor()
  )
})

test_that("can use `fit_resamples()` with a workflow - postprocessor (no training)", {
  skip_if_not_installed("tailor", minimum_version = "0.0.0.9002")
  skip_if_not_installed("probably")

  y <- seq(0, 7, .001)
  dat <- data.frame(y = y, x = y + (y - 3)^2)

  folds <- rsample::vfold_cv(dat, v = 2)

  wflow_no_post <-
    workflows::workflow(
      y ~ x,
      parsnip::linear_reg()
    )

  wflow_post <-
    wflow_no_post |>
    workflows::add_tailor(
      tailor::tailor() |> tailor::adjust_numeric_range(lower_limit = 3)
    )

  set.seed(1)
  no_post_res <-
    fit_resamples(
      wflow_no_post,
      folds,
      control = control_resamples(save_pred = TRUE)
    )

  set.seed(1)
  post_res <-
    fit_resamples(
      wflow_post,
      folds,
      control = control_resamples(save_pred = TRUE, extract = identity)
    )

  no_post_preds <-
    collect_predictions(no_post_res) |>
    dplyr::rename(no_post = .pred)
  post_preds <-
    collect_predictions(post_res) |>
    dplyr::rename(post = .pred) |>
    dplyr::select(.row, post, id, .config)

  both_preds <-
    dplyr::full_join(
      no_post_preds,
      post_preds,
      by = c(".row", "id", ".config")
    ) |>
    filter(no_post < 3)

  expect_true(all(both_preds$post == 3))
  expect_true(
    post_res$.extracts[[1]]$.extracts[[1]] |> tune:::has_postprocessor()
  )
})

# Error capture ----------------------------------------------------------------

test_that("failure in recipe is caught elegantly", {
  skip_if_not_installed("splines2")

  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = NA_real_)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  control <- control_resamples(extract = function(x) x, save_pred = TRUE)

  expect_snapshot(
    result <- fit_resamples(lin_mod, rec, folds, control = control)
  )

  notes <- result$.notes
  note <- notes[[1]]$note

  extract <- result$.extracts
  predictions <- result$.predictions

  expect_length(notes, 2L)

  expect_snapshot(note)

  expect_equal(extract, list(NULL, NULL))
  expect_equal(predictions, list(NULL, NULL))
})

test_that("failure in variables tidyselect specification is caught elegantly", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  workflow <- workflow() |>
    add_model(lin_mod) |>
    add_variables(mpg, foobar)

  control <- control_resamples(extract = function(x) x, save_pred = TRUE)

  expect_snapshot(
    result <- fit_resamples(workflow, folds, control = control)
  )

  notes <- result$.notes
  note <- notes[[1]]$note

  extract <- result$.extracts
  predictions <- result$.predictions

  expect_length(notes, 2L)

  expect_snapshot(note)

  expect_equal(extract, list(NULL, NULL))
  expect_equal(predictions, list(NULL, NULL))
})

test_that("classification models generate correct error message", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipes::recipe(vs ~ ., data = mtcars)

  log_mod <- parsnip::logistic_reg() |>
    parsnip::set_engine("glm")

  control <- control_resamples(extract = function(x) x, save_pred = TRUE)

  expect_snapshot(
    result <- fit_resamples(log_mod, rec, folds, control = control)
  )

  notes <- result$.notes
  note <- notes[[1]]$note

  extract <- result$.extracts
  predictions <- result$.predictions

  expect_length(notes, 2L)

  expect_snapshot(note)

  expect_equal(extract, list(NULL, NULL))
  expect_equal(predictions, list(NULL, NULL))
})


# tune_grid() fallback ---------------------------------------------------------

test_that("`tune_grid()` falls back to `fit_resamples()` - formula", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect <- fit_resamples(lin_mod, mpg ~ ., folds)

  expect_snapshot(
    result <- tune_grid(lin_mod, mpg ~ ., folds)
  )

  expect_equal(collect_metrics(expect), collect_metrics(result))
})

test_that("`tune_grid()` falls back to `fit_resamples()` - workflow variables", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  wf <- workflow() |>
    add_model(lin_mod) |>
    add_variables(mpg, c(cyl, disp))

  expect <- fit_resamples(wf, folds)

  expect_snapshot(
    result <- tune_grid(wf, folds)
  )

  expect_equal(collect_metrics(expect), collect_metrics(result))
})

test_that("`tune_grid()` ignores `grid` if there are no tuning parameters", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect <- lin_mod |>
    fit_resamples(mpg ~ ., folds)

  expect_snapshot(
    result <- lin_mod |> tune_grid(mpg ~ ., grid = data.frame(x = 1), folds)
  )

  expect_equal(collect_metrics(expect), collect_metrics(result))
})


# autoplot() -------------------------------------------------------------------

test_that("cannot autoplot `fit_resamples()` results", {
  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  result <- lin_mod |>
    fit_resamples(mpg ~ ., folds)

  expect_snapshot(error = TRUE, {
    autoplot(result)
  })
})

test_that("ellipses with fit_resamples", {
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect_snapshot(
    lin_mod |> fit_resamples(mpg ~ ., folds, something = "wrong")
  )
})

test_that("argument order gives errors for recipe/formula", {
  skip_if_not_installed("splines2")

  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp) |>
    recipes::step_spline_natural(wt)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect_snapshot(error = TRUE, {
    fit_resamples(rec, lin_mod, folds)
  })
  expect_snapshot(error = TRUE, {
    fit_resamples(mpg ~ ., lin_mod, folds)
  })
})


test_that("retain extra attributes", {
  skip_if_not_installed("splines2")

  set.seed(6735)
  folds <- rsample::vfold_cv(mtcars, v = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  res <- lin_mod |>
    fit_resamples(mpg ~ ., folds)

  att <- attributes(res)
  att_names <- names(att)
  expect_true(any(att_names == "metrics"))
  expect_true(any(att_names == "outcomes"))
  expect_true(any(att_names == "parameters"))

  expect_true(is.character(att$outcomes))
  expect_true(att$outcomes == "mpg")
  expect_true(inherits(att$parameters, "parameters"))
  expect_true(inherits(att$metrics, "metric_set"))

  res2 <- lin_mod |>
    fit_resamples(
      mpg ~ .,
      folds,
      control = control_resamples(save_workflow = TRUE)
    )
  expect_null(attr(res, "workflow"))
  expect_true(inherits(attr(res2, "workflow"), "workflow"))
})


test_that("`fit_resamples()` when objects need tuning", {
  rec <- recipe(mpg ~ ., data = mtcars) |>
    step_spline_natural(disp, deg_free = tune())
  spec_1 <- linear_reg(penalty = tune()) |> set_engine("glmnet")
  spec_2 <- linear_reg()
  wflow_1 <- workflow(rec, spec_1)
  wflow_2 <- workflow(mpg ~ ., spec_1)
  wflow_3 <- workflow(rec, spec_2)
  rs <- rsample::vfold_cv(mtcars)

  expect_snapshot_error(fit_resamples(wflow_1, rs))
  expect_snapshot_error(fit_resamples(wflow_2, rs))
  expect_snapshot_error(fit_resamples(wflow_3, rs))
})

check_predictions <- function(split, pred, tune_df, obj) {
  assess <- rsample::assessment(split)
  n_te <- nrow(assess)
  n_pm <- nrow(tune_df)
  ind_te <- as.integer(split, data = "assessment")
  expect_true(tibble::is_tibble(pred))
  expect_equal(nrow(pred), n_te * n_pm)
  exp_nms <- c(
    "mpg",
    ".pred",
    ".row",
    .get_tune_parameter_names(obj),
    ".config"
  )
  expect_equal(names(pred), exp_nms)
  expect_equal(sort(unique(ind_te)), sort(unique(pred$.row)))
  TRUE
}

# Test file for variable fold weights functionality

# Setup test data
set.seed(42)
test_data <- data.frame(
  x1 = rnorm(50),
  x2 = rnorm(50),
  x3 = rnorm(50)
)
test_data$y <- 2 * test_data$x1 + 3 * test_data$x2 + rnorm(50, sd = 0.5)

set.seed(123)
folds <- rsample::vfold_cv(mtcars, v = 3)

# Helper function to create a simple model
create_test_model <- function() {
  parsnip::linear_reg() |> parsnip::set_engine("lm")
}


test_that("add_resample_weights() validates inputs correctly", {
  expect_snapshot(
    add_resample_weights("not_an_rset", c(0.5, 0.3, 0.2)),
    error = TRUE
  )

  expect_snapshot(
    add_resample_weights(folds, c("a", "b", "c")),
    error = TRUE
  )

  expect_snapshot(
    add_resample_weights(folds, c(0.5, 0.3)),
    error = TRUE
  )

  expect_snapshot(
    add_resample_weights(folds, c(-0.1, 0.5, 0.6)),
    error = TRUE
  )

  expect_snapshot(
    add_resample_weights(folds, c(0, 0, 0)),
    error = TRUE
  )
})

test_that("add_resample_weights() adds weights correctly", {
  weights <- c(0.1, 0.5, 0.4)
  weighted_folds <- add_resample_weights(folds, weights)

  # Weights get normalized to sum to 1
  expected_weights <- weights / sum(weights)

  expect_s3_class(weighted_folds, "rset")
  expect_equal(attr(weighted_folds, ".resample_weights"), expected_weights)
  expect_equal(nrow(weighted_folds), nrow(folds))
})

test_that("calculate_resample_weights() works correctly", {
  auto_weights <- calculate_resample_weights(folds)

  expect_type(auto_weights, "double")
  expect_length(auto_weights, nrow(folds))
  expect_true(all(auto_weights > 0))
  expect_true(abs(sum(auto_weights) - 1) < 1e-10)
})

test_that("weights are preserved through tuning pipeline", {
  weights <- c(0.1, 0.5, 0.4)
  weighted_folds <- add_resample_weights(folds, weights)

  mod <- create_test_model()

  suppressWarnings({
    res <- tune_grid(
      mod,
      mpg ~ .,
      resamples = weighted_folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  metrics <- collect_metrics(res)
  expect_equal(nrow(metrics), 1)
  expect_true("mean" %in% names(metrics))
  expect_true(is.numeric(metrics$mean))
})

test_that("weights affect metric aggregation", {
  weights <- c(0.1, 0.5, 0.4)
  weighted_folds <- add_resample_weights(folds, weights)

  mod <- create_test_model()

  suppressWarnings({
    # Unweighted results
    res_unweighted <- tune_grid(
      mod,
      mpg ~ .,
      resamples = folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )

    # Weighted results
    res_weighted <- tune_grid(
      mod,
      mpg ~ .,
      resamples = weighted_folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  unweighted_rmse <- collect_metrics(res_unweighted)$mean[1]
  weighted_rmse <- collect_metrics(res_weighted)$mean[1]

  expect_true(is.numeric(unweighted_rmse))
  expect_true(is.numeric(weighted_rmse))
  expect_false(is.na(unweighted_rmse))
  expect_false(is.na(weighted_rmse))
})

test_that("extreme weights show larger effect", {
  skip_if_not_installed("kknn")

  # Create folds for this specific test
  set.seed(42)
  test_folds <- rsample::vfold_cv(test_data, v = 3)

  # Regular weights
  weights <- c(0.6, 0.2, 0.2)
  weighted_folds <- add_resample_weights(test_folds, weights)

  # Extreme weights
  extreme_weights <- c(0.95, 0.025, 0.025)
  extreme_weighted_folds <- add_resample_weights(test_folds, extreme_weights)

  # Create a model with tuning parameter
  knn_spec <- parsnip::nearest_neighbor(neighbors = tune()) |>
    parsnip::set_engine("kknn") |>
    parsnip::set_mode("regression")

  param_grid <- data.frame(neighbors = c(3, 5))

  suppressWarnings({
    # Unweighted
    res_unweighted <- tune_grid(
      knn_spec,
      y ~ .,
      resamples = test_folds,
      grid = param_grid,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )

    # Regular weights
    res_weighted <- tune_grid(
      knn_spec,
      y ~ .,
      resamples = weighted_folds,
      grid = param_grid,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )

    # Extreme weights
    res_extreme <- tune_grid(
      knn_spec,
      y ~ .,
      resamples = extreme_weighted_folds,
      grid = param_grid,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  unweighted_metrics <- collect_metrics(res_unweighted)
  weighted_metrics <- collect_metrics(res_weighted)
  extreme_metrics <- collect_metrics(res_extreme)

  # Check that results exist and are sensible
  expect_equal(nrow(unweighted_metrics), 2)
  expect_equal(nrow(weighted_metrics), 2)
  expect_equal(nrow(extreme_metrics), 2)

  # Calculate differences
  regular_diff <- max(abs(unweighted_metrics$mean - weighted_metrics$mean))
  extreme_diff <- max(abs(unweighted_metrics$mean - extreme_metrics$mean))

  expect_true(regular_diff >= 0)
  expect_true(extreme_diff >= 0)
  expect_true(all(is.finite(c(regular_diff, extreme_diff))))
})

test_that("weight normalization works correctly", {
  expect_equal(
    tune:::.validate_resample_weights(c(3, 6, 9), 3),
    c(1 / 6, 1 / 3, 1 / 2) # normalized to sum to 1
  )

  expect_equal(
    tune:::.validate_resample_weights(c(0.2, 0.3, 0.5), 3),
    c(0.2, 0.3, 0.5) # already normalized to sum to 1
  )
})

test_that("equal weights return NULL", {
  # Simplest integer match
  expect_null(tune:::.validate_resample_weights(c(2, 2, 2), 3))

  # Fractional match
  expect_null(tune:::.validate_resample_weights(c(1 / 3, 1 / 3, 1 / 3), 3))

  # Check more reseampless
  expect_null(tune:::.validate_resample_weights(c(1, 1, 1, 1, 1), 5))
})

test_that("unequal weights do not return NULL", {
  # Check non-null decimal values
  result <- tune:::.validate_resample_weights(c(0.1, 0.5, 0.4), 3)
  expect_false(is.null(result))
  expect_equal(result, c(0.1, 0.5, 0.4))

  # Non-null fractional values
  result2 <- tune:::.validate_resample_weights(c(1, 2, 3), 3)
  expect_false(is.null(result2))
  expect_equal(result2, c(1 / 6, 2 / 6, 3 / 6))
})

test_that("add_resample_weights with equal weights returns NULL attribute", {
  # Adding equal weights should trigger NULL assignment
  equal_weighted_folds <- add_resample_weights(folds, c(1, 1, 1))
  expect_null(attr(equal_weighted_folds, ".resample_weights"))

  # Verify it's still an rset object
  expect_s3_class(equal_weighted_folds, "rset")
})

test_that("equal weights produce same results as no weights", {
  mod <- create_test_model()

  suppressWarnings({
    # Results with no weights
    res_no_weights <- tune_grid(
      mod,
      mpg ~ .,
      resamples = folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )

    # Results with equal weights
    equal_weighted_folds <- add_resample_weights(folds, c(1, 1, 1))
    res_equal_weights <- tune_grid(
      mod,
      mpg ~ .,
      resamples = equal_weighted_folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  metrics_no_weights <- collect_metrics(res_no_weights)
  metrics_equal_weights <- collect_metrics(res_equal_weights)

  # Results should match
  expect_equal(metrics_no_weights$mean, metrics_equal_weights$mean)
  expect_equal(metrics_no_weights$std_err, metrics_equal_weights$std_err)
})

test_that("weighted statistics functions work correctly", {
  x <- c(1, 2, 3, 4, 5)
  w <- c(0.1, 0.2, 0.3, 0.2, 0.2)

  weighted_sd <- tune:::.weighted_sd(x, w)

  expect_true(is.numeric(weighted_sd))
  expect_false(is.na(weighted_sd))
  expect_true(weighted_sd >= 0)

  # Test with NA values
  x_na <- c(1, 2, NA, 4, 5)
  weighted_sd_na <- tune:::.weighted_sd(x_na[!is.na(x_na)], w[!is.na(x_na)])

  expect_true(is.numeric(weighted_sd_na))

  # Test edge cases
  expect_true(is.na(tune:::.weighted_sd(c(1), c(1)))) # single value
})

test_that("fold weight extraction works", {
  weights <- c(0.1, 0.5, 0.4)
  weighted_folds <- add_resample_weights(folds, weights)

  # Weights get normalized to sum to 1
  expected_weights <- weights / sum(weights)

  mod <- create_test_model()

  suppressWarnings({
    res <- tune_grid(
      mod,
      mpg ~ .,
      resamples = weighted_folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  extracted_weights <- tune:::.get_resample_weights(res)
  expect_equal(extracted_weights, expected_weights)
})

test_that("individual fold metrics can be collected", {
  weights <- c(0.1, 0.5, 0.4)
  weighted_folds <- add_resample_weights(folds, weights)

  mod <- create_test_model()

  suppressWarnings({
    res <- tune_grid(
      mod,
      mpg ~ .,
      resamples = weighted_folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  # Collect individual fold metrics
  individual_metrics <- collect_metrics(res, summarize = FALSE)

  expect_true(nrow(individual_metrics) >= 3) # At least one metric per fold
  expect_true("id" %in% names(individual_metrics))
  expect_true(".estimate" %in% names(individual_metrics))
  expect_true(all(is.finite(individual_metrics$.estimate)))
})

test_that("backwards compatibility - no weights", {
  mod <- create_test_model()

  suppressWarnings({
    res <- tune_grid(
      mod,
      mpg ~ .,
      resamples = folds, # No weights
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  metrics <- collect_metrics(res)
  expect_equal(nrow(metrics), 1)
  expect_true("mean" %in% names(metrics))
  expect_true(is.numeric(metrics$mean))
  expect_false(is.na(metrics$mean))
})

test_that("rset tibble conversion includes fold weights", {
  weights <- c(0.1, 0.4, 0.5)
  weighted_folds <- add_resample_weights(folds, weights)

  # Convert to tibble manually (this is what our print method does)
  x_tbl <- tibble::as_tibble(weighted_folds)
  x_tbl$resample_weight <- weights

  # Verify the structure
  expect_true("resample_weight" %in% names(x_tbl))
  expect_equal(x_tbl$resample_weight, weights)
  expect_equal(nrow(x_tbl), 3)
})

test_that("extract_resample_weights() works with rset objects", {
  weights <- c(0.2, 0.3, 0.5)
  weighted_folds <- add_resample_weights(folds, weights)

  # Should return the weights
  extracted_weights <- extract_resample_weights(weighted_folds)
  expect_equal(extracted_weights, weights)

  # Should return NULL for unweighted rsets
  unweighted_result <- extract_resample_weights(folds)
  expect_null(unweighted_result)
})

test_that("extract_resample_weights() works with tune_results objects", {
  weights <- c(0.1, 0.5, 0.4)
  weighted_folds <- add_resample_weights(folds, weights)

  mod <- create_test_model()

  suppressWarnings({
    res <- tune_grid(
      mod,
      mpg ~ .,
      resamples = weighted_folds,
      grid = 1,
      metrics = yardstick::metric_set(yardstick::rmse),
      control = control_grid(verbose = FALSE)
    )
  })

  # Should extract weights from tune results
  extracted_weights <- extract_resample_weights(res)
  expected_weights <- weights / sum(weights) # normalized
  expect_equal(extracted_weights, expected_weights)
})

test_that("extract_resample_weights() validates input types", {
  expect_snapshot(
    extract_resample_weights("not_valid_input"),
    error = TRUE
  )

  expect_snapshot(
    extract_resample_weights(data.frame(x = 1:3)),
    error = TRUE
  )
})

test_that("tune recipe only", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_tune_1) |>
    add_model(helper_objects$lm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(1, 3)))
  grid <- dials::grid_regular(pset, levels = 3)
  folds <- rsample::vfold_cv(mtcars)
  control <- control_grid(extract = identity)

  res <- tune_grid(wflow, resamples = folds, grid = grid, control = control)
  res_est <- collect_metrics(res)
  res_workflow <- res$.extracts[[1]]$.extracts[[1]]

  expect_equal(res, .Last.tune.result)

  # Ensure tunable parameters in recipe are finalized
  num_comp <- res_workflow$pre$actions$recipe$recipe$steps[[2]]$num_comp

  expect_equal(res$id, folds$id)
  expect_equal(nrow(res_est), nrow(grid) * 2)
  expect_equal(sum(res_est$.metric == "rmse"), nrow(grid))
  expect_equal(sum(res_est$.metric == "rsq"), nrow(grid))
  expect_equal(res_est$n, rep(10, nrow(grid) * 2))
  expect_false(identical(num_comp, expr(tune())))
  expect_true(res_workflow$trained)
  expect_null(.get_tune_eval_times(res))
  expect_null(.get_tune_eval_time_target(res))
})

# ------------------------------------------------------------------------------

test_that("tune model only (with recipe)", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_no_tune_1) |>
    add_model(helper_objects$svm_mod)
  pset <- extract_parameter_set_dials(wflow)
  grid <- dials::grid_regular(pset, levels = 3)
  folds <- rsample::vfold_cv(mtcars)
  control <- control_grid(extract = identity)

  res <- tune_grid(wflow, resamples = folds, grid = grid, control = control)
  res_est <- collect_metrics(res)
  res_workflow <- res$.extracts[[1]]$.extracts[[1]]

  # Ensure tunable parameters in spec are finalized
  cost_quo <- res_workflow$fit$fit$spec$args$cost
  cost <- quo_get_expr(cost_quo)

  expect_equal(res$id, folds$id)
  expect_equal(nrow(res_est), nrow(grid) * 2)
  expect_equal(sum(res_est$.metric == "rmse"), nrow(grid))
  expect_equal(sum(res_est$.metric == "rsq"), nrow(grid))
  expect_equal(res_est$n, rep(10, nrow(grid) * 2))
  expect_false(identical(cost, expr(tune())))
  expect_true(res_workflow$trained)
})

# ------------------------------------------------------------------------------

test_that("tune model only (with variables)", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)

  wflow <- workflow() |>
    add_variables(mpg, everything()) |>
    add_model(helper_objects$svm_mod)

  pset <- extract_parameter_set_dials(wflow)
  grid <- dials::grid_regular(pset, levels = 3)

  folds <- rsample::vfold_cv(mtcars)

  res <- tune_grid(wflow, resamples = folds, grid = grid)

  expect_equal(res$id, folds$id)

  res_est <- collect_metrics(res)

  expect_equal(nrow(res_est), nrow(grid) * 2)
  expect_equal(sum(res_est$.metric == "rmse"), nrow(grid))
  expect_equal(sum(res_est$.metric == "rsq"), nrow(grid))
  expect_equal(res_est$n, rep(10, nrow(grid) * 2))
})

# ------------------------------------------------------------------------------

test_that("tune model only (with recipe, multi-predict)", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_no_tune_1) |>
    add_model(helper_objects$svm_mod)
  pset <- extract_parameter_set_dials(wflow)
  grid <- dials::grid_regular(pset, levels = 3)
  folds <- rsample::vfold_cv(mtcars)
  res <- tune_grid(wflow, resamples = folds, grid = grid)
  expect_equal(res$id, folds$id)
  expect_equal(
    colnames(res$.metrics[[1]]),
    c("cost", ".metric", ".estimator", ".estimate", ".config")
  )
  res_est <- collect_metrics(res)
  expect_equal(nrow(res_est), nrow(grid) * 2)
  expect_equal(sum(res_est$.metric == "rmse"), nrow(grid))
  expect_equal(sum(res_est$.metric == "rsq"), nrow(grid))
  expect_equal(res_est$n, rep(10, nrow(grid) * 2))
})

test_that("tune model only (without recipe, multi-predict. #695)", {
  skip_on_cran()
  skip_if_not_installed("kknn")

  knn_res <-
    tune_grid(
      parsnip::nearest_neighbor(mode = "regression", neighbors = tune("k")),
      mpg ~ .,
      resamples = rsample::bootstraps(mtcars, 5),
      grid = 4
    )

  expect_equal(nrow(knn_res$.notes[[1]]), 0)
})

# ------------------------------------------------------------------------------

test_that("tune model only (fairness - include `by` variable as predictor)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor("classification", "kknn", neighbors = tune())
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  res <- tune_grid(
    knn,
    vs ~ mpg + hp + cyl,
    resamples = boots,
    grid = n_grid,
    metrics = yardstick::metric_set(
      yardstick::roc_auc,
      yardstick::demographic_parity(cyl)
    )
  )

  expect_equal(
    colnames(res$.metrics[[1]]),
    c("neighbors", ".metric", ".by", ".estimator", ".estimate", ".config")
  )
  res_est <- collect_metrics(res)
  expect_equal(
    colnames(res_est),
    c(
      "neighbors",
      ".metric",
      ".estimator",
      ".by",
      "mean",
      "n",
      "std_err",
      ".config"
    )
  )
  expect_equal(nrow(res_est), n_grid * 2)
  expect_equal(sum(res_est$.by == "cyl", na.rm = TRUE), n_grid)
  expect_equal(sum(is.na(res_est$.by)), n_grid)
  expect_equal(sum(res_est$.metric == "roc_auc"), n_grid)
  expect_equal(sum(res_est$.metric == "demographic_parity"), n_grid)
  expect_equal(res_est$n, rep(3, n_grid * 2))
})

test_that("tune model only (fairness - don't include `by` variable as predictor)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor("classification", "kknn", neighbors = tune())
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  res2 <- tune_grid(
    knn,
    vs ~ mpg + hp,
    resamples = boots,
    grid = n_grid,
    metrics = yardstick::metric_set(
      yardstick::roc_auc,
      yardstick::demographic_parity(cyl)
    )
  )

  expect_equal(
    colnames(res2$.metrics[[1]]),
    c("neighbors", ".metric", ".by", ".estimator", ".estimate", ".config")
  )
  res_est2 <- collect_metrics(res2)
  expect_equal(
    colnames(res_est2),
    c(
      "neighbors",
      ".metric",
      ".estimator",
      ".by",
      "mean",
      "n",
      "std_err",
      ".config"
    )
  )
  expect_equal(nrow(res_est2), n_grid * 2)
  expect_equal(sum(res_est2$.by == "cyl", na.rm = TRUE), n_grid)
  expect_equal(sum(is.na(res_est2$.by)), n_grid)
  expect_equal(sum(res_est2$.metric == "roc_auc"), n_grid)
  expect_equal(sum(res_est2$.metric == "demographic_parity"), n_grid)
  expect_equal(res_est2$n, rep(3, n_grid * 2))
})

test_that("tune model only (fairness metrics - evaluate across multiple `by`)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor("classification", "kknn", neighbors = tune())
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  res3 <- tune_grid(
    knn,
    vs ~ mpg + hp,
    resamples = boots,
    grid = n_grid,
    metrics = yardstick::metric_set(
      yardstick::roc_auc,
      yardstick::demographic_parity(cyl),
      yardstick::equal_opportunity(am)
    )
  )

  expect_equal(
    colnames(res3$.metrics[[1]]),
    c("neighbors", ".metric", ".by", ".estimator", ".estimate", ".config")
  )
  res_est3 <- collect_metrics(res3)
  expect_equal(
    colnames(res_est3),
    c(
      "neighbors",
      ".metric",
      ".estimator",
      ".by",
      "mean",
      "n",
      "std_err",
      ".config"
    )
  )
  expect_equal(nrow(res_est3), n_grid * 3)
  expect_equal(sum(res_est3$.by == "cyl", na.rm = TRUE), n_grid)
  expect_equal(sum(is.na(res_est3$.by)), n_grid)
  expect_equal(sum(res_est3$.metric == "roc_auc"), n_grid)
  expect_equal(sum(res_est3$.metric == "demographic_parity"), n_grid)
  expect_equal(res_est3$n, rep(3, n_grid * 3))
})

test_that("tune model only (fairness - evaluate across multiple `by`, same metric)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor("classification", "kknn", neighbors = tune())
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  res4 <- tune_grid(
    knn,
    vs ~ mpg + hp,
    resamples = boots,
    grid = n_grid,
    metrics = yardstick::metric_set(
      yardstick::roc_auc,
      yardstick::demographic_parity(cyl),
      yardstick::demographic_parity(am)
    )
  )

  expect_equal(
    colnames(res4$.metrics[[1]]),
    c("neighbors", ".metric", ".by", ".estimator", ".estimate", ".config")
  )
  res_est4 <- collect_metrics(res4)
  expect_equal(
    colnames(res_est4),
    c(
      "neighbors",
      ".metric",
      ".estimator",
      ".by",
      "mean",
      "n",
      "std_err",
      ".config"
    )
  )
  expect_equal(nrow(res_est4), n_grid * 3)
  expect_equal(sum(res_est4$.by == "cyl", na.rm = TRUE), n_grid)
  expect_equal(sum(res_est4$.by == "am", na.rm = TRUE), n_grid)
  expect_equal(sum(is.na(res_est4$.by)), n_grid)
  expect_equal(sum(res_est4$.metric == "roc_auc"), n_grid)
  expect_equal(sum(res_est4$.metric == "demographic_parity"), n_grid * 2)
  expect_equal(res_est4$n, rep(3, n_grid * 3))
})

test_that("tune model only (fairness - evaluate only fairness metrics)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor("classification", "kknn", neighbors = tune())
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  res5 <- tune_grid(
    knn,
    vs ~ mpg + hp,
    resamples = boots,
    grid = n_grid,
    metrics = yardstick::metric_set(
      yardstick::demographic_parity(cyl)
    )
  )

  expect_equal(
    colnames(res5$.metrics[[1]]),
    c("neighbors", ".metric", ".by", ".estimator", ".estimate", ".config")
  )
  res_est5 <- collect_metrics(res5)
  expect_equal(
    colnames(res_est5),
    c(
      "neighbors",
      ".metric",
      ".estimator",
      ".by",
      "mean",
      "n",
      "std_err",
      ".config"
    )
  )
  expect_equal(nrow(res_est5), n_grid)
  expect_equal(sum(res_est5$.by == "cyl", na.rm = TRUE), n_grid)
  expect_equal(sum(is.na(res_est5$.by)), 0)
  expect_equal(sum(res_est5$.metric == "demographic_parity"), n_grid)
  expect_equal(res_est5$n, rep(3, n_grid))
})

# ------------------------------------------------------------------------------

test_that("tune model and recipe", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_tune_1) |>
    add_model(helper_objects$svm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(1, 3)))
  grid <- dials::grid_regular(pset, levels = 3)
  folds <- rsample::vfold_cv(mtcars)
  control <- control_grid(extract = identity)

  res <- tune_grid(wflow, resamples = folds, grid = grid, control = control)
  res_est <- collect_metrics(res)
  res_workflow <- res$.extracts[[1]]$.extracts[[1]]

  # Ensure tunable parameters in spec are finalized
  cost_quo <- res_workflow$fit$fit$spec$args$cost
  cost <- quo_get_expr(cost_quo)

  # Ensure tunable parameters in recipe are finalized
  num_comp <- res_workflow$pre$actions$recipe$recipe$steps[[2]]$num_comp

  expect_equal(res$id, folds$id)
  expect_equal(
    colnames(res$.metrics[[1]]),
    c("cost", "num_comp", ".metric", ".estimator", ".estimate", ".config")
  )
  expect_equal(nrow(res_est), nrow(grid) * 2)
  expect_equal(sum(res_est$.metric == "rmse"), nrow(grid))
  expect_equal(sum(res_est$.metric == "rsq"), nrow(grid))
  expect_equal(res_est$n, rep(10, nrow(grid) * 2))
  expect_false(identical(cost, expr(tune())))
  expect_false(identical(num_comp, expr(tune())))
  expect_true(res_workflow$trained)
})

# ------------------------------------------------------------------------------

test_that("tune model and recipe (multi-predict)", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_tune_1) |>
    add_model(helper_objects$svm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(2, 3)))
  grid <- dials::grid_regular(pset, levels = c(3, 2))
  folds <- rsample::vfold_cv(mtcars)
  res <- tune_grid(wflow, resamples = folds, grid = grid)
  expect_equal(res$id, folds$id)
  res_est <- collect_metrics(res)
  expect_equal(nrow(res_est), nrow(grid) * 2)
  expect_equal(sum(res_est$.metric == "rmse"), nrow(grid))
  expect_equal(sum(res_est$.metric == "rsq"), nrow(grid))
  expect_equal(res_est$n, rep(10, nrow(grid) * 2))
})

# ------------------------------------------------------------------------------

test_that('tune model and recipe (parallel_over = "everything")', {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_tune_1) |>
    add_model(helper_objects$svm_mod)
  pset <- extract_parameter_set_dials(wflow) |>
    update(num_comp = dials::num_comp(c(1, 3)))
  grid <- dials::grid_regular(pset, levels = 3)
  folds <- rsample::vfold_cv(mtcars)
  control <- control_grid(extract = identity, parallel_over = "everything")

  res <- tune_grid(wflow, resamples = folds, grid = grid, control = control)
  res_est <- collect_metrics(res)

  expect_equal(res$id, folds$id)
  expect_equal(
    colnames(res$.metrics[[1]]),
    c("cost", "num_comp", ".metric", ".estimator", ".estimate", ".config")
  )
  expect_equal(nrow(res_est), nrow(grid) * 2)
  expect_equal(sum(res_est$.metric == "rmse"), nrow(grid))
  expect_equal(sum(res_est$.metric == "rsq"), nrow(grid))
  expect_equal(res_est$n, rep(10, nrow(grid) * 2))
})

# ------------------------------------------------------------------------------

test_that("tune recipe only - failure in recipe is caught elegantly", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipe(mpg ~ ., data = mtcars) |>
    step_spline_b(disp, deg_free = tune())

  model <- linear_reg(mode = "regression") |>
    set_engine("lm")

  # NA values not allowed in recipe
  cars_grid <- tibble(deg_free = c(3, NA_real_, 4))

  # ask for predictions and extractions
  control <- control_grid(
    save_pred = TRUE,
    extract = function(x) 1L
  )

  suppressMessages({
    cars_res <- tune_grid(
      model,
      preprocessor = rec,
      resamples = data_folds,
      grid = cars_grid,
      control = control
    )
  })

  notes <- cars_res$.notes
  note <- notes[[1]]$note

  extract <- cars_res$.extracts[[1]]

  predictions <- cars_res$.predictions[[1]]
  used_deg_free <- sort(unique(predictions$deg_free))

  expect_length(notes, 2L)

  # failing rows are not in the output
  expect_equal(nrow(extract), 2L)
  expect_equal(extract$deg_free, c(3, 4))

  expect_equal(used_deg_free, c(3, 4))
})

test_that("tune model only - failure in recipe is caught elegantly", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  # NA values not allowed in recipe
  rec <- recipe(mpg ~ ., data = mtcars) |>
    step_spline_b(disp, deg_free = NA_real_)

  cars_grid <- tibble(cost = c(0.01, 0.02))

  expect_snapshot(
    cars_res <- tune_grid(
      helper_objects$svm_mod,
      preprocessor = rec,
      resamples = data_folds,
      grid = cars_grid,
      control = control_grid(
        extract = function(x) {
          1
        },
        save_pred = TRUE
      )
    )
  )

  notes <- cars_res$.notes
  note <- notes[[1]]$note

  extracts <- cars_res$.extracts
  predictions <- cars_res$.predictions

  expect_length(notes, 2L)

  # recipe failed - no models run
  expect_equal(extracts, list(NULL, NULL))
  expect_equal(predictions, list(NULL, NULL))
})

test_that("tune model only - failure in formula is caught elegantly", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  cars_grid <- tibble(cost = 0.01)

  # these terms don't exist!
  expect_snapshot(
    cars_res <- tune_grid(
      helper_objects$svm_mod,
      y ~ z,
      resamples = data_folds,
      grid = cars_grid,
      control = control_grid(
        extract = function(x) {
          1
        },
        save_pred = TRUE
      )
    ),
    transform = catalog_lines
  )

  notes <- cars_res$.notes
  note <- notes[[1]]$note

  extracts <- cars_res$.extracts
  predictions <- cars_res$.predictions

  expect_length(notes, 2L)

  # formula failed - no models run
  expect_equal(extracts, list(NULL, NULL))
  expect_equal(predictions, list(NULL, NULL))
})

test_that("tune model and recipe - failure in recipe is caught elegantly", {
  skip_if_not_installed("splines2")
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(7898)
  data_folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <- recipe(mpg ~ ., data = mtcars) |>
    step_spline_b(disp, deg_free = tune())

  # NA values not allowed in recipe
  cars_grid <- tibble(deg_free = c(NA_real_, 10L), cost = 0.01)

  suppressMessages({
    cars_res <- tune_grid(
      helper_objects$svm_mod,
      preprocessor = rec,
      resamples = data_folds,
      grid = cars_grid,
      control = control_grid(
        extract = function(x) {
          1
        },
        save_pred = TRUE
      )
    )
  })

  notes <- cars_res$.notes
  note <- notes[[1]]$note

  extract <- cars_res$.extracts[[1]]
  prediction <- cars_res$.predictions[[1]]

  expect_length(notes, 2L)

  # recipe failed half of the time, only 1 model passed
  expect_equal(nrow(extract), 1L)
  expect_equal(extract$deg_free, 10L)
  expect_equal(extract$cost, 0.01)

  expect_equal(
    unique(prediction[, c("deg_free", "cost")]),
    tibble(deg_free = 10, cost = 0.01)
  )
})

test_that("argument order gives errors for recipes", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  expect_snapshot(error = TRUE, {
    tune_grid(
      helper_objects$rec_tune_1,
      helper_objects$lm_mod,
      rsample::vfold_cv(mtcars, v = 2)
    )
  })
})

test_that("argument order gives errors for formula", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  expect_snapshot(error = TRUE, {
    tune_grid(mpg ~ ., helper_objects$lm_mod, rsample::vfold_cv(mtcars, v = 2))
  })
})

test_that("ellipses with tune_grid", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  wflow <- workflow() |>
    add_recipe(helper_objects$rec_tune_1) |>
    add_model(helper_objects$lm_mod)
  folds <- rsample::vfold_cv(mtcars)
  expect_snapshot(
    tune_grid(wflow, resamples = folds, grid = 3, something = "wrong")
  )
})


test_that("determining the grid type", {
  grid_1 <- expand.grid(a = 1:100, b = letters[1:2])
  expect_true(tune:::is_regular_grid(grid_1))
  expect_true(tune:::is_regular_grid(grid_1[-(1:10), ]))
  expect_false(tune:::is_regular_grid(grid_1[-(1:100), ]))
  set.seed(1932)
  grid_2 <- data.frame(a = runif(length(letters)), b = letters)
  expect_false(tune:::is_regular_grid(grid_2))
})


test_that("retain extra attributes", {
  skip_if_not_installed("kernlab")

  helper_objects <- helper_objects_tune()

  set.seed(4400)
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_no_tune_1) |>
    add_model(helper_objects$svm_mod)
  pset <- extract_parameter_set_dials(wflow)
  grid <- dials::grid_regular(pset, levels = 3)
  folds <- rsample::vfold_cv(mtcars)
  res <- tune_grid(wflow, resamples = folds, grid = grid)

  att <- attributes(res)
  att_names <- names(att)
  expect_true(any(att_names == "metrics"))
  expect_true(any(att_names == "outcomes"))
  expect_true(any(att_names == "parameters"))

  expect_true(is.character(att$outcomes))
  expect_true(att$outcomes == "mpg")
  expect_true(inherits(att$parameters, "parameters"))
  expect_true(inherits(att$metrics, "metric_set"))

  set.seed(4400)
  wflow <- workflow() |>
    add_formula(mpg ~ .) |>
    add_model(helper_objects$svm_mod)
  pset <- extract_parameter_set_dials(wflow)
  grid <- dials::grid_regular(pset, levels = 3)
  folds <- rsample::vfold_cv(mtcars)
  res <- tune_grid(wflow, resamples = folds, grid = grid)

  att <- attributes(res)
  att_names <- names(att)
  expect_true(any(att_names == "metrics"))
  expect_true(any(att_names == "outcomes"))
  expect_true(any(att_names == "parameters"))

  expect_true(is.character(att$outcomes))
  expect_true(att$outcomes == "mpg")
  expect_true(inherits(att$parameters, "parameters"))
  expect_true(inherits(att$metrics, "metric_set"))

  res2 <- tune_grid(
    wflow,
    resamples = folds,
    grid = grid,
    control = control_grid(save_workflow = TRUE)
  )
  expect_null(attr(res, "workflow"))
  expect_true(inherits(attr(res2, "workflow"), "workflow"))
})

test_that("finalize_workflow_preprocessor() is deprecated", {
  helper_objects <- helper_objects_tune()
  wflow <- workflow() |>
    add_recipe(helper_objects$rec_tune_1) |>
    add_model(helper_objects$lm_mod)
  grid <- tibble::tibble(num_comp = 1L)
  expect_snapshot(wflow <- finalize_workflow_preprocessor(wflow, grid))
})

#' Bayesian optimization of model parameters.
#'
#' [tune_bayes()] uses models to generate new candidate tuning parameter
#'  combinations based on previous results.
#'
#' @inheritParams tune_grid
#' @param metrics A [yardstick::metric_set()], or `NULL` to compute a standard
#' set of metrics. The first metric in `metrics` is the one that will be optimized.
#' @param iter The maximum number of search iterations.
#' @param objective A character string for what metric should be optimized or
#' an acquisition function object.
#' @param initial An initial set of results in a tidy format (as would result
#' from [tune_grid()]) or a positive integer. It is suggested that the number of
#' initial results be greater than the number of parameters being optimized.
#' @param control A control object created by [control_bayes()].
#' @param ... Not currently used.
#' @return A tibble of results that mirror those generated by [tune_grid()].
#' However, these results contain an `.iter` column and replicate the `rset`
#' object multiple times over iterations (at limited additional memory costs).
#' @seealso [control_bayes()], [tune()], [autoplot.tune_results()],
#'  [show_best()], [select_best()], [collect_predictions()],
#'  [collect_metrics()], [prob_improve()], [exp_improve()], [conf_bound()],
#'  [fit_resamples()]
#' @details
#'
#' The optimization starts with a set of initial results, such as those
#'  generated by [tune_grid()]. If none exist, the function will create several
#'  combinations and obtain their performance estimates.
#'
#' Using one of the performance estimates as the _model outcome_, a Gaussian
#'  process (GP) model is created where the previous tuning parameter combinations
#'  are used as the predictors.
#'
#' A large grid of potential hyperparameter combinations is predicted using
#'  the model and scored using an _acquisition function_. These functions
#'  usually combine the predicted mean and variance of the GP to decide the best
#'  parameter combination to try next. For more information, see the
#'  documentation for [exp_improve()] and the corresponding package vignette.
#'
#' The best combination is evaluated using resampling and the process continues.
#'
#' @section Parallel Processing:
#'
#' tune supports parallel processing with the \pkg{future} package. To execute
#' the resampling iterations in parallel, specify a [plan][future::plan] with
#' future first. The `allow_par` argument can be used to avoid parallelism.
#'
#' For the most part, warnings generated during training are shown as they occur
#' and are associated with a specific resample when
#' `control_bayes(verbose = TRUE)`. They are (usually) not aggregated until the
#' end of processing.
#'
#' For Bayesian optimization, parallel processing is used to estimate the
#' resampled performance values once a new candidate set of values are estimated.
#'
#' @section Initial Values:
#'
#' The results of [tune_grid()], or a previous run of [tune_bayes()] can be used
#' in the `initial` argument. `initial` can also be a positive integer. In this
#' case, a space-filling design will be used to populate a preliminary set of
#' results. For good results, the number of initial values should be more than
#' the number of parameters being optimized.
#'
#' The tuning parameter combinations that were tested are called _candidates_.
#'  Each candidate has a unique `.config` value that, for the initial grid search,
#'  has the pattern `pre{num}_mod{num}_post{num}`. The numbers include a zero
#'  when that element was static. For example, a value of `pre0_mod3_post4` means
#'  no preprocessors were tuned and the model and postprocessor(s) had at least
#'  three and four candidates, respectively. The iterative part of the
#'  search uses the pattern `iter{num}`. In each case, the numbers are
#'  zero-padded to enable proper sorting.
#'
#' @section Parameter Ranges and Values:
#'
#' In some cases, the tuning parameter values depend on the dimensions of the
#' data (they are said to contain [unknown][dials::unknown] values). For
#' example, `mtry` in random forest models depends on the number of predictors.
#' In such cases, the unknowns in the tuning parameter object must be determined
#' beforehand and passed to the function via the `param_info` argument.
#' [dials::finalize()] can be used to derive the data-dependent parameters.
#' Otherwise, a parameter set can be created via [dials::parameters()], and the
#' `dials` `update()` function can be used to specify the ranges or values.
#'
#' @section Performance Metrics:
#'
#' To use your own performance metrics, the [yardstick::metric_set()] function
#'  can be used to pick what should be measured for each model. If multiple
#'  metrics are desired, they can be bundled. For example, to estimate the area
#'  under the ROC curve as well as the sensitivity and specificity (under the
#'  typical probability cutoff of 0.50), the `metrics` argument could be given:
#'
#' \preformatted{
#'   metrics = metric_set(roc_auc, sens, spec)
#' }
#'
#' Each metric is calculated for each candidate model.
#'
#' If no metric set is provided, one is created:
#' \itemize{
#'   \item For regression models, the root mean squared error and coefficient
#'         of determination are computed.
#'   \item For classification, the area under the ROC curve and overall accuracy
#'         are computed.
#'   \item For censored regression, the dynamic Brier score
#'         ([yardstick::brier_survival()]) is used.
#'   \item For quantile regression, the weighted interval score
#'         ([yardstick::weighted_interval_score()]) is used.
#' }
#'
#' Note that the metrics also determine what type of predictions are estimated
#' during tuning. For example, in a classification problem, if metrics are used
#' that are all associated with hard class predictions, the classification
#' probabilities are not created.
#'
#' The out-of-sample estimates of these metrics are contained in a list column
#' called `.metrics`. This tibble contains a row for each metric and columns
#' for the value, the estimator type, and so on.
#'
#' [collect_metrics()] can be used for these objects to collapse the results
#' over the resampled (to obtain the final resampling estimates per tuning
#' parameter combination).
#'
#' @section Obtaining Predictions:
#'
#' When `control_bayes(save_pred = TRUE)`, the output tibble contains a list
#' column called `.predictions` that has the out-of-sample predictions for each
#' parameter combination in the grid and each fold (which can be very large).
#'
#' The elements of the tibble are tibbles with columns for the tuning
#' parameters, the row number from the original data object (`.row`), the
#' outcome data (with the same name(s) of the original data), and any columns
#' created by the predictions. For example, for simple regression problems, this
#' function generates a column called `.pred` and so on. As noted above, the
#' prediction columns that are returned are determined by the type of metric(s)
#' requested.
#'
#' This list column can be `unnested` using [tidyr::unnest()] or using the
#'  convenience function [collect_predictions()].
#'
#' @inheritSection tune_grid Extracting Information
#'
#' @template case-weights
#' @template censored-regression
#'
#' @section Optimizing Censored Regression Models:
#'
#' With dynamic performance metrics (e.g. Brier or ROC curves), performance is
#' calculated for every value of `eval_time` but the _first_ evaluation time
#' given by the user (e.g., `eval_time[1]`) is used to guide the optimization.
#'
#' @examplesIf tune:::should_run_examples(suggests = "kernlab") && !tune:::is_cran_check()
#' library(recipes)
#' library(rsample)
#' library(parsnip)
#'
#' # define resamples and minimal recipe on mtcars
#' set.seed(6735)
#' folds <- vfold_cv(mtcars, v = 5)
#'
#' car_rec <-
#'   recipe(mpg ~ ., data = mtcars) |>
#'   step_normalize(all_predictors())
#'
#' # define an svm with parameters to tune
#' svm_mod <-
#'   svm_rbf(cost = tune(), rbf_sigma = tune()) |>
#'   set_engine("kernlab") |>
#'   set_mode("regression")
#'
#' # use a space-filling design with 6 points
#' set.seed(3254)
#' svm_grid <- tune_grid(svm_mod, car_rec, folds, grid = 6)
#'
#' show_best(svm_grid, metric = "rmse")
#'
#' # use bayesian optimization to evaluate at 6 more points
#' set.seed(8241)
#' svm_bayes <- tune_bayes(svm_mod, car_rec, folds, initial = svm_grid, iter = 6)
#'
#' # note that bayesian optimization evaluated parameterizations
#' # similar to those that previously decreased rmse in svm_grid
#' show_best(svm_bayes, metric = "rmse")
#'
#' # specifying `initial` as a numeric rather than previous tuning results
#' # will result in `tune_bayes` initially evaluating an space-filling
#' # grid using `tune_grid` with `grid = initial`
#' set.seed(0239)
#' svm_init <- tune_bayes(svm_mod, car_rec, folds, initial = 6, iter = 6)
#'
#' show_best(svm_init, metric = "rmse")
#' @export
tune_bayes <- function(object, ...) {
  UseMethod("tune_bayes")
}

#' @export
tune_bayes.default <- function(object, ...) {
  cli::cli_abort(
    "The first argument to {.fn tune_bayes} should be either a model or workflow,
    not {.obj_type_friendly {object}}."
  )
}

#' @export
#' @rdname tune_bayes
tune_bayes.model_spec <- function(
  object,
  preprocessor,
  resamples,
  ...,
  iter = 10,
  param_info = NULL,
  metrics = NULL,
  eval_time = NULL,
  objective = exp_improve(),
  initial = 5,
  control = control_bayes()
) {
  if (rlang::is_missing(preprocessor) || !is_preprocessor(preprocessor)) {
    cli::cli_abort(tune_pp_msg)
  }

  # set `seed` so that calling `control_bayes()` doesn't alter RNG state (#721)
  control <- parsnip::condense_control(control, control_bayes(seed = 1))

  wflow <- add_model(workflow(), object)

  if (is_recipe(preprocessor)) {
    wflow <- add_recipe(wflow, preprocessor)
  } else if (rlang::is_formula(preprocessor)) {
    wflow <- add_formula(wflow, preprocessor)
  }

  tune_bayes_workflow(
    wflow,
    resamples = resamples,
    iter = iter,
    param_info = param_info,
    metrics = metrics,
    eval_time = eval_time,
    objective = objective,
    initial = initial,
    control = control,
    ...
  )
}


#' @export
#' @rdname tune_bayes
tune_bayes.workflow <-
  function(
    object,
    resamples,
    ...,
    iter = 10,
    param_info = NULL,
    metrics = NULL,
    eval_time = NULL,
    objective = exp_improve(),
    initial = 5,
    control = control_bayes()
  ) {
    # set `seed` so that calling `control_bayes()` doesn't alter RNG state (#721)
    control <- parsnip::condense_control(control, control_bayes(seed = 1))

    res <-
      tune_bayes_workflow(
        object,
        resamples = resamples,
        iter = iter,
        param_info = param_info,
        metrics = metrics,
        eval_time = eval_time,
        objective = objective,
        initial = initial,
        control = control,
        ...
      )
    .stash_last_result(res)
    res
  }

## TODOs from testing

# These lines are missing:
# -   i Gaussian process model
# -   ! The Gaussian process model is being fit using 1 features but only has 2
# -     data points to do so. This may cause errors or a poor model fit.
# -   v Gaussian process model
# -   i Generating 3 candidates
# -   i Predicted candidates

# Some failures are not being caught (or ignored); see "missing performance
# values" test
# -   ! Gaussian process model: no non-missing arguments to min; returning Inf, no non-missing arguments...
# -   x Gaussian process model: Error in seq_len(n - 1L): argument must be coercible to non-negative int...

# -   Error in `check_gp_failure()`:
# +   Error in `apply()`:
# -   ! Gaussian process model was not fit.
# +   ! dim(X) must have a positive length

tune_bayes_workflow <- function(
  object,
  resamples,
  iter = 10,
  param_info = NULL,
  metrics = NULL,
  eval_time = NULL,
  objective = exp_improve(),
  initial = 5,
  control,
  ...,
  call = caller_env()
) {
  clear_gp_results()
  start_time <- proc.time()[3]

  initialize_catalog(control = control, workflow = object)

  check_rset(resamples)
  rset_info <- pull_rset_attributes(resamples)

  check_iter(iter, call = call)
  if (iter > 0) {
    iter_chr <- recipes::names0(iter, "iter")
  } else {
    iter_chr <- "iter0"
  }

  metrics <- check_metrics_arg(metrics, object, call = call)
  opt_metric <- first_metric(metrics)
  opt_metric_name <- opt_metric$metric
  maximize <- opt_metric$direction == "maximize"

  eval_time <- check_eval_time_arg(eval_time, metrics, call = call)
  opt_metric_time <- first_eval_time(
    metrics,
    metric = opt_metric_name,
    eval_time = eval_time,
    call = call
  )

  if (is.null(param_info)) {
    param_info <- hardhat::extract_parameter_set_dials(object)
  }
  check_workflow(
    object,
    check_dials = is.null(param_info),
    pset = param_info,
    call = call
  )
  check_backend_options(control$backend_options)

  unsummarized <- check_initial(
    initial,
    pset = param_info,
    wflow = object,
    resamples = resamples,
    metrics = metrics,
    eval_time = eval_time,
    ctrl = control,
    checks = "bayes"
  )

  # Pull outcome names from initialization run
  outcomes <- peek_tune_results_outcomes(unsummarized)

  # Evaluate this portion of the function in a local environment using
  # `(function() expr)()` so that `on.exit()` doesn't touch the exit
  # handlers attached to the execution environment of this function
  # by `initialize_catalog()` (#828, #845).
  (function() {
    # Return whatever we have if there is a error (or execution is stopped)
    on.exit({
      cli::cli_alert_danger(
        "Optimization stopped prematurely; returning current results."
      )

      out <- new_iteration_results(
        x = unsummarized,
        parameters = param_info,
        metrics = metrics,
        eval_time = eval_time,
        eval_time_target = opt_metric_time,
        outcomes = outcomes,
        rset_info = rset_info,
        workflow = NULL
      )

      .stash_last_result(out)

      return(out)
    })

    # Preempt `estimate_tune_results()` error and rely
    # on `on.exit()` condition to return preliminary results
    if (.is_cataclysmic(unsummarized)) {
      return()
    }

    # Get the averaged resampling stats before stripping attributes
    mean_stats <- estimate_tune_results(unsummarized)

    # Strip off `tune_results` class and drop all attributes since
    # we add on an `iteration_results` class later.
    unsummarized <- new_bare_tibble(unsummarized)

    check_time(start_time, control$time_limit)

    score_card <- initial_info(
      mean_stats,
      opt_metric_name,
      maximize,
      opt_metric_time
    )

    if (control$verbose_iter) {
      msg <- paste("Optimizing", opt_metric_name, "using", objective$label)
      if (!is.null(opt_metric_time)) {
        msg <- paste(
          msg,
          "at evaluation time",
          format(opt_metric_time, digits = 3)
        )
      }
      message_wrap(msg)
    }

    prev_gp_mod <- NULL

    for (i in seq_len(iter) + score_card$overall_iter) {
      .notes <- new_note()

      log_best(control, i, score_card)

      check_time(start_time, control$time_limit)

      # Maybe remove .catch_and_log() here and do catching inside function
      set.seed(control$seed[1] + i)

      gp_mod <- fit_gp(
        mean_stats |> dplyr::select(-.iter),
        pset = param_info,
        metric = opt_metric_name,
        eval_time = opt_metric_time,
        control = control,
        ...
      )

      check_time(start_time, control$time_limit)

      set.seed(control$seed[1] + i + 1)
      candidates <-
        pred_gp(
          gp_mod,
          param_info,
          control = control,
          current = mean_stats |> dplyr::select(dplyr::all_of(param_info$id))
        )

      check_time(start_time, control$time_limit)

      acq_summarizer(control, iter = i, objective = objective)

      candidates <-
        dplyr::bind_cols(
          candidates,
          stats::predict(
            objective,
            candidates,
            iter = i,
            maximize = maximize,
            score_card$best_val
          )
        )

      check_time(start_time, control$time_limit)

      # TODO move some messages here
      # check_and_log_flow(control, candidates)

      save_gp_results(
        gp_mod,
        param_info,
        control,
        i,
        iter,
        candidates,
        score_card
      )

      candidates <- pick_candidate(candidates, score_card, control)

      if (score_card$uncertainty >= control$uncertain) {
        score_card$uncertainty <- -1 # is updated in update_score_card() below
      }

      check_time(start_time, control$time_limit)

      param_msg(control, candidates)
      set.seed(control$seed[1] + i + 2)
      tmp_res <-
        more_results(
          object,
          resamples = resamples,
          candidates = candidates,
          metrics = metrics,
          eval_time = eval_time,
          control = control,
          param_info = param_info
        )

      check_time(start_time, control$time_limit)

      all_bad <- .is_cataclysmic(tmp_res)
      if (!inherits(tmp_res, "try-error") & !all_bad) {
        tmp_res[[".metrics"]] <- purrr::map(
          tmp_res[[".metrics"]],
          set_config,
          config = iter_chr[i]
        )
        if (control$save_pred) {
          tmp_res[[".predictions"]] <- purrr::map(
            tmp_res[[".predictions"]],
            set_config,
            config = iter_chr[i]
          )
        }
        if (".extracts" %in% names(tmp_res)) {
          tmp_res[[".extracts"]] <- purrr::map(
            tmp_res[[".extracts"]],
            set_config,
            config = iter_chr[i]
          )
        }
        unsummarized <- dplyr::bind_rows(
          unsummarized,
          tmp_res |> mutate(.iter = i)
        )
        rs_estimate <- estimate_tune_results(tmp_res)
        mean_stats <- dplyr::bind_rows(
          mean_stats,
          rs_estimate |> dplyr::mutate(.iter = i)
        )
        score_card <- update_score_card(score_card, i, tmp_res)
        log_progress(
          control,
          x = mean_stats,
          maximize = maximize,
          objective = opt_metric_name,
          eval_time = opt_metric_time
        )
      } else {
        if (all_bad) {
          update_printer(
            control,
            split_labels = NULL,
            task = "All models failed",
            type = "danger"
          )
        }
        score_card$last_impr <- score_card$last_impr + 1
      }

      if (score_card$last_impr + 1 > control$no_improve) {
        cli::cli_alert_warning(
          "No improvement for {control$no_improve} iterations; returning current results."
        )
        break
      }
      prev_gp_mod <- gp_mod
      check_time(start_time, control$time_limit)
    }

    workflow_output <- .set_workflow(object, control)

    # Reset `on.exit()` hook
    on.exit()

    res <-
      new_iteration_results(
        x = unsummarized,
        parameters = param_info,
        metrics = metrics,
        eval_time = eval_time,
        eval_time_target = opt_metric_time,
        outcomes = outcomes,
        rset_info = rset_info,
        workflow = workflow_output
      )

    .stash_last_result(res)
    res
  })() # end of self calling lambda
}

create_initial_set <- function(param, n = NULL, checks) {
  .check_param_objects(param)
  if (is.null(n)) {
    n <- nrow(param) + 1
  }
  if (any(checks == "bayes")) {
    check_bayes_initial_size(nrow(param), n)
  }
  dials::grid_space_filling(param, size = n)
}

check_iter <- function(iter, call) {
  if (!(is.numeric(iter) && length(iter) == 1L && !is.na(iter) && iter >= 0)) {
    cli::cli_abort(
      "The {.arg iter} argument must be a non-negative integer.",
      call = call
    )
  }

  iter
}

set_config <- function(x, config = NULL, prefix = NULL) {
  if (!is.null(x)) {
    if (!is.null(prefix)) {
      x <- dplyr::mutate(x, .config = paste0(prefix, "_", .config))
    } else {
      x <- dplyr::mutate(x, .config = config)
    }
  }
  x
}

# ------------------------------------------------------------------------------

#' @export
#' @keywords internal
#' @rdname empty_ellipses
#' @param pset A `parameters` object.
#' @param as_matrix A logical for the return type.
encode_set <- function(x, pset, ..., as_matrix = FALSE) {
  rlang::check_dots_empty()
  # change the numeric variables to the transformed scale (if any)
  has_trans <- purrr::map_lgl(pset$object, \(.x) !is.null(.x$trans))
  if (any(has_trans)) {
    idx <- which(has_trans)
    for (i in idx) {
      x[[pset$id[i]]] <-
        dials::value_transform(pset$object[[i]], x[[pset$id[i]]])
    }
  }

  is_quant <- purrr::map_lgl(pset$object, inherits, "quant_param")
  # Convert all data to the [0, 1] scale based on their possible range (not on
  # their observed range)
  if (any(is_quant)) {
    new_vals <- purrr::map2(
      pset$object[is_quant],
      x[, is_quant],
      encode_unit,
      direction = "forward"
    )
    names(new_vals) <- names(x)[is_quant]
    new_vals <- tibble::as_tibble(new_vals)
    x[, is_quant] <- new_vals
  }

  # Ensure that the right levels are used to create dummy variables
  if (any(!is_quant)) {
    for (i in which(!is_quant)) {
      x[[i]] <- factor(x[[i]], levels = pset$object[[i]]$values)
    }
  }

  if (as_matrix) {
    x <- stats::model.matrix(~ . + 0, data = x)
  }
  x
}

update_score_card <- function(info, iter, results, control) {
  current_val <-
    results |>
    estimate_tune_results() |>
    dplyr::filter(.metric == info$metrics)

  if (!is.null(info$eval_time)) {
    current_val <-
      current_val |>
      dplyr::filter(.eval_time == info$eval_time)
  }

  current_val <- current_val$mean

  if (info$max) {
    is_better <- current_val > info$best_val
  } else {
    is_better <- current_val < info$best_val
  }

  if (!is.na(is_better) & is_better) {
    info$last_impr <- 0
    info$best_val <- current_val
    info$best_iter <- iter
    info$uncertainty <- 0
  } else {
    info$last_impr <- info$last_impr + 1
    info$uncertainty <- info$uncertainty + 1
  }
  info
}


# ------------------------------------------------------------------------------

# save opt_metric_name and maximize to simplify!!!!!!!!!!!!!!!
initial_info <- function(stats, metrics, maximize, eval_time) {
  best_res <-
    stats |>
    dplyr::filter(.metric == metrics) |>
    dplyr::filter(!is.na(mean))

  # TODO a lot of slice_min/slice_max can be used now
  if (maximize) {
    best_res <-
      best_res |>
      dplyr::arrange(desc(mean)) |>
      slice(1)
  } else {
    best_res <-
      best_res |>
      dplyr::arrange(mean) |>
      slice(1)
  }
  best_val <- best_res$mean[1]
  best_iter <- best_res$.iter[1]
  last_impr <- 0
  overall_iter <- max(stats$.iter)

  # outputs:

  list(
    best_val = best_val,
    best_iter = best_iter,
    last_impr = last_impr,
    uncertainty = 0,
    overall_iter = overall_iter,
    metrics = metrics,
    max = maximize,
    eval_time = eval_time
  )
}


# ------------------------------------------------------------------------------

more_results <- function(
  object,
  resamples,
  candidates,
  metrics,
  eval_time = NULL,
  control,
  param_info
) {
  update_printer(
    control,
    split_labels = NULL,
    task = "Estimating performance",
    type = "info"
  )

  candidates <- candidates[,
    !(names(candidates) %in% c(".mean", ".sd", "objective"))
  ]
  p_chr <- paste0(
    names(candidates),
    "=",
    format(as.data.frame(candidates), digits = 3)
  )

  tmp_res <-
    try(
      tune_grid(
        object,
        resamples = resamples,
        param_info = param_info,
        grid = candidates,
        metrics = metrics,
        eval_time = eval_time,
        control = control
      ),
      silent = TRUE
    )

  if (inherits(tmp_res, "try-error")) {
    update_printer(
      control,
      split_labels = NULL,
      task = "Couldn't estimate performance",
      type = "danger"
    )
  } else {
    all_bad <- .is_cataclysmic(tmp_res)
    if (all_bad) {
      p_chr <- glue::glue_collapse(
        p_chr,
        width = options()$width - 28,
        sep = ", "
      )
      msg <- paste("All models failed for:", p_chr)
      update_printer(control, split_labels = NULL, task = msg, type = "danger")
      tmp_res <- simpleError(msg)
    } else {
      update_printer(
        control,
        split_labels = NULL,
        task = "Estimating performance",
        type = "success"
      )
    }
  }
  tmp_res
}


#' @export
#' @keywords internal
#' @rdname empty_ellipses
.is_cataclysmic <- function(x) {
  is_err <- purrr::map_lgl(x$.metrics, inherits, c("simpleError", "error"))
  if (any(!is_err)) {
    is_good <- purrr::map_lgl(
      x$.metrics[!is_err],
      \(.x) tibble::is_tibble(.x) && nrow(.x) > 0
    )
    is_err[!is_err] <- !is_good
  }
  all(is_err)
}


#' @export
#' @keywords internal
#' @rdname empty_ellipses
#' @param origin The calculation start time.
#' @param limit The allowable time (in minutes).
check_time <- function(origin, limit) {
  if (is.na(limit)) {
    return(invisible(NULL))
  }
  now_time <- proc.time()[3]
  if (now_time - origin >= limit * 60) {
    cli::cli_abort("The time limit of {limit} minute{?s} has been reached.")
  }
  invisible(NULL)
}

# May be better to completely refactor things to a high-level call then use
# base's setTimeLimit().

## -----------------------------------------------------------------------------

save_gp_results <- function(x, pset, ctrl, i, iter, candidates, score_card) {
  if (!ctrl$save_gp_scoring) {
    return(invisible(NULL))
  }

  nm <- recipes::names0(iter, "gp_candidates_")[i]
  file_name <- glue::glue("{tempdir()}/{nm}.RData")
  gp_fit <- x
  res <- try(
    save(gp_fit, pset, i, candidates, score_card, file = file_name),
    silent = TRUE
  )
  if (inherits(res, "try-error")) {
    cli::cli_warn(
      "Could not save GP results at iteration {i}: {as.character(res))}"
    )
  }
  invisible(res)
}


clear_gp_results <- function() {
  gp_files <-
    list.files(tempdir(), pattern = "gp_candidates_", full.names = TRUE)
  res <- try(unlink(gp_files), silent = TRUE)
  invisible(res)
}

test_that("eval time inputs are checked for regression models", {
  skip_if_not_installed("kknn")

  library(parsnip)
  library(workflows)
  library(yardstick)
  library(rsample)

  # ------------------------------------------------------------------------------

  wflow <- workflow(mpg ~ ., linear_reg())
  knn_spec <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")
  wflow_tune <- workflow(mpg ~ ., knn_spec)

  set.seed(1)
  split <- initial_split(mtcars)
  rs <- vfold_cv(mtcars)

  times <- c(1, 1:3)

  # ------------------------------------------------------------------------------
  # setup metric sets

  met_reg <- metric_set(rmse)

  # ------------------------------------------------------------------------------
  # check inputs

  expect_snapshot(check_eval_time_arg(NULL, met_reg))
  expect_snapshot(check_eval_time_arg(times, met_reg))

  # ------------------------------------------------------------------------------
  # resampling

  expect_snapshot(
    res <- fit_resamples(wflow, rs, eval_time = times)
  )

  # ------------------------------------------------------------------------------
  # tuning

  expect_snapshot({
    set.seed(1)
    res <- tune_grid(wflow_tune, rs, eval_time = times)
  })

  expect_snapshot({
    set.seed(1)
    res <- tune_bayes(wflow_tune, rs, iter = 1, eval_time = times)
  })

  # ------------------------------------------------------------------------------
  # final fit

  expect_snapshot(
    res <- last_fit(wflow, split, eval_time = times)
  )
})

test_that("eval time are checked for classification models", {
  library(parsnip)
  library(workflows)
  library(yardstick)
  library(rsample)
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  data(two_class_dat, package = "modeldata")
  wflow <- workflow(Class ~ A + B, logistic_reg())
  knn_spec <- nearest_neighbor(neighbors = tune()) |>
    set_mode("classification")
  wflow_tune <- workflow(Class ~ A + B, knn_spec)

  set.seed(1)
  split <- initial_split(two_class_dat)
  rs <- vfold_cv(two_class_dat)

  times <- c(1, 1:3)

  # ------------------------------------------------------------------------------
  # setup metric sets

  met_cls <- metric_set(brier_class)

  # ------------------------------------------------------------------------------
  # check inputs

  expect_snapshot(check_eval_time_arg(NULL, met_cls))
  expect_snapshot(check_eval_time_arg(times, met_cls))

  # ------------------------------------------------------------------------------
  # resampling

  expect_snapshot(
    res <- fit_resamples(wflow, rs, eval_time = times)
  )

  # ------------------------------------------------------------------------------
  # tuning

  expect_snapshot({
    set.seed(1)
    res <- tune_grid(wflow_tune, rs, eval_time = times)
  })

  expect_snapshot({
    set.seed(1)
    res <- tune_bayes(wflow_tune, rs, iter = 1, eval_time = times)
  })

  # ------------------------------------------------------------------------------
  # final fit

  expect_snapshot(
    res <- last_fit(wflow, split, eval_time = times)
  )
})

test_that("eval time inputs are checked for censored regression models", {
  skip_if_not_installed("censored")

  library(parsnip)
  library(workflows)
  library(yardstick)
  library(rsample)
  suppressPackageStartupMessages(library(censored))

  stanford2$event_time <- Surv(stanford2$time, stanford2$status)
  stanford2 <- stanford2[, c("event_time", "age")]

  wflow <- workflow(event_time ~ age, survival_reg())
  sr_spec <- survival_reg(dist = tune())
  wflow_tune <- workflow(event_time ~ age, sr_spec)

  set.seed(1)
  split <- initial_split(stanford2)
  rs <- vfold_cv(stanford2)

  # ------------------------------------------------------------------------------
  # setup metric sets

  met_stc <- metric_set(concordance_survival)
  met_dyn <- metric_set(brier_survival)
  met_int <- metric_set(brier_survival_integrated)
  met_stc_dyn <- metric_set(concordance_survival, brier_survival)
  met_stc_int <- metric_set(concordance_survival, brier_survival_integrated)
  met_dyn_stc <- metric_set(brier_survival, concordance_survival)
  met_dyn_int <- metric_set(brier_survival, brier_survival_integrated)
  met_int_stc <- metric_set(brier_survival_integrated, concordance_survival)
  met_int_dyn <- metric_set(brier_survival_integrated, brier_survival)

  # ------------------------------------------------------------------------------
  # check inputs when eval_time left out

  expect_snapshot(check_eval_time_arg(NULL, met_stc))
  expect_snapshot(check_eval_time_arg(NULL, met_dyn), error = TRUE)
  expect_snapshot(check_eval_time_arg(NULL, met_int), error = TRUE)

  expect_snapshot(check_eval_time_arg(NULL, met_stc_dyn), error = TRUE)
  expect_snapshot(check_eval_time_arg(NULL, met_stc_int), error = TRUE)
  expect_snapshot(check_eval_time_arg(NULL, met_dyn_stc), error = TRUE)

  expect_snapshot(check_eval_time_arg(NULL, met_dyn_int), error = TRUE)
  expect_snapshot(check_eval_time_arg(NULL, met_int_stc), error = TRUE)
  expect_snapshot(check_eval_time_arg(NULL, met_int_dyn), error = TRUE)

  # ------------------------------------------------------------------------------
  # check inputs with single eval times

  expect_snapshot(check_eval_time_arg(2, met_stc))
  expect_snapshot(check_eval_time_arg(2, met_dyn))
  expect_snapshot(check_eval_time_arg(2, met_int), error = TRUE)

  expect_snapshot(check_eval_time_arg(2, met_stc_dyn))
  expect_snapshot(check_eval_time_arg(2, met_stc_int), error = TRUE)

  expect_snapshot(check_eval_time_arg(2, met_dyn_stc))
  expect_snapshot(check_eval_time_arg(2, met_dyn_int), error = TRUE)

  expect_snapshot(check_eval_time_arg(2, met_int_stc), error = TRUE)
  expect_snapshot(check_eval_time_arg(2, met_int_dyn), error = TRUE)

  # ------------------------------------------------------------------------------
  # check inputs with multiple eval times

  expect_snapshot(check_eval_time_arg(1:3, met_stc))
  expect_snapshot(check_eval_time_arg(1:3, met_dyn))
  expect_snapshot(check_eval_time_arg(1:3, met_int))

  expect_snapshot(check_eval_time_arg(1:3, met_stc_dyn))
  expect_snapshot(check_eval_time_arg(1:3, met_stc_int))
  expect_snapshot(check_eval_time_arg(1:3, met_dyn_stc))

  expect_snapshot(check_eval_time_arg(1:3, met_dyn_int))
  expect_snapshot(check_eval_time_arg(1:3, met_int_stc))
  expect_snapshot(check_eval_time_arg(1:3, met_int_dyn))

  # ------------------------------------------------------------------------------
  # resampling

  # no eval time
  expect_snapshot(res <- fit_resamples(wflow, rs, metrics = met_stc))
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_dyn), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_int), error = TRUE)

  expect_snapshot(fit_resamples(wflow, rs, metrics = met_stc_dyn), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_stc_int), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_dyn_stc), error = TRUE)

  expect_snapshot(fit_resamples(wflow, rs, metrics = met_dyn_int), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_int_stc), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_int_dyn), error = TRUE)

  # one eval time
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_stc, eval_time = 2)
  )
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_dyn, eval_time = 2)
  )
  expect_snapshot(
    fit_resamples(wflow, rs, metrics = met_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_stc_dyn, eval_time = 2)
  )
  expect_snapshot(
    fit_resamples(wflow, rs, metrics = met_stc_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_dyn_stc, eval_time = 2)
  )
  expect_snapshot(
    fit_resamples(wflow, rs, metrics = met_dyn_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    fit_resamples(wflow, rs, metrics = met_int_stc, eval_time = 2),
    error = TRUE
  )
  expect_snapshot(
    fit_resamples(wflow, rs, metrics = met_int_dyn, eval_time = 2),
    error = TRUE
  )

  # multiple eval times
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_dyn, eval_time = 1:3)
  )
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_stc_dyn, eval_time = 1:3)
  )
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_stc_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_dyn_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_dyn_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_int_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- fit_resamples(wflow, rs, metrics = met_int_dyn, eval_time = 1:3)
  )

  # ------------------------------------------------------------------------------
  # grid tuning (tune bayes tests in extratests repo)

  # no eval time
  expect_snapshot(res <- tune_grid(wflow_tune, rs, metrics = met_stc))
  expect_snapshot(tune_grid(wflow_tune, rs, metrics = met_dyn), error = TRUE)
  expect_snapshot(tune_grid(wflow_tune, rs, metrics = met_int), error = TRUE)

  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_stc_dyn),
    error = TRUE
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_stc_int),
    error = TRUE
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_dyn_stc),
    error = TRUE
  )

  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_dyn_int),
    error = TRUE
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_int_stc),
    error = TRUE
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_int_dyn),
    error = TRUE
  )

  # one eval time
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_stc, eval_time = 2)
  )
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_dyn, eval_time = 2)
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_stc_dyn, eval_time = 2)
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_stc_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_dyn_stc, eval_time = 2)
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_dyn_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_int_stc, eval_time = 2),
    error = TRUE
  )
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_int_dyn, eval_time = 2),
    error = TRUE
  )

  # multiple eval times
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_dyn, eval_time = 1:3)
  )
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_stc_dyn, eval_time = 1:3)
  )
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_stc_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_dyn_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_dyn_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_int_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- tune_grid(wflow_tune, rs, metrics = met_int_dyn, eval_time = 1:3)
  )

  # ------------------------------------------------------------------------------
  # last fit

  # no eval time
  expect_silent(res <- last_fit(wflow, split, metrics = met_stc))
  expect_snapshot(last_fit(wflow, split, metrics = met_dyn), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_int), error = TRUE)

  expect_snapshot(last_fit(wflow, split, metrics = met_stc_dyn), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_stc_int), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_dyn_stc), error = TRUE)

  expect_snapshot(last_fit(wflow, split, metrics = met_dyn_int), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_int_stc), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_int_dyn), error = TRUE)

  # one eval time
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_stc, eval_time = 2)
  )
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_dyn, eval_time = 2)
  )
  expect_snapshot(
    last_fit(wflow, split, metrics = met_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_stc_dyn, eval_time = 2)
  )
  expect_snapshot(
    last_fit(wflow, split, metrics = met_stc_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_dyn_stc, eval_time = 2)
  )
  expect_snapshot(
    last_fit(wflow, split, metrics = met_dyn_int, eval_time = 2),
    error = TRUE
  )

  expect_snapshot(
    last_fit(wflow, split, metrics = met_int_stc, eval_time = 2),
    error = TRUE
  )
  expect_snapshot(
    last_fit(wflow, split, metrics = met_int_dyn, eval_time = 2),
    error = TRUE
  )

  # multiple eval times
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_dyn, eval_time = 1:3)
  )
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_stc_dyn, eval_time = 1:3)
  )
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_stc_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_dyn_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_dyn_int, eval_time = 1:3)
  )

  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_int_stc, eval_time = 1:3)
  )
  expect_snapshot(
    res <- last_fit(wflow, split, metrics = met_int_dyn, eval_time = 1:3)
  )
})

test_that("two quantitative predictor marginal plot", {
  svm_results <- readRDS(test_path("data", "svm_results.rds"))

  p <- autoplot(svm_results)
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c("mean", "# resamples", ".metric", "name", "value")
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(value))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(p$labels$y, "")
  expect_equal(p$labels$x, "")

  p <- autoplot(svm_results, metric = "accuracy")
  expect_true(isTRUE(all.equal(unique(p$data$.metric), "accuracy")))
})


test_that("two quantitative predictor and one qualitative marginal plot", {
  knn_results <- readRDS(test_path("data", "knn_results.rds"))

  p <- autoplot(knn_results)
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c(
      "Distance Weighting Function",
      "mean",
      "# resamples",
      ".metric",
      "name",
      "value"
    )
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(value))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(p$labels$y, "")
  expect_equal(p$labels$x, "")
  expect_equal(p$labels$colour, "Distance Weighting Function")
})

test_that("not marginal plot with grid search", {
  knn_results <- readRDS(test_path("data", "knn_results.rds"))

  expect_snapshot(error = TRUE, autoplot(knn_results, type = "performance"))
  expect_snapshot(error = TRUE, autoplot(knn_results, type = "parameters"))
})


test_that("marginal plot labels and transformations - irregular grid", {
  svm_results <- readRDS(test_path("data", "svm_results.rds"))

  p <- autoplot(svm_results)
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c("mean", "# resamples", ".metric", "name", "value")
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(value))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(p$labels$y, "")
  expect_equal(p$labels$x, "")
  expect_equal(
    sort(unique(p$data$name)),
    c("%^*#", "Cost (log-2)", "Scale Factor (log-10)")
  )
  expect_equal(
    sort(unique(collect_metrics(svm_results)$cost)),
    sort(unique(2^(p$data$value[p$data$name == "Cost (log-2)"]))),
    tolerance = 0.001
  )
  expect_equal(
    sort(unique(collect_metrics(svm_results)$scale_factor)),
    sort(unique(10^(p$data$value[p$data$name == "Scale Factor (log-10)"]))),
    tolerance = 0.001
  )

  p <- autoplot(svm_results, metric = "accuracy")
  expect_true(isTRUE(all.equal(unique(p$data$.metric), "accuracy")))
})

# ------------------------------------------------------------------------------

test_that("marginal plot for iterative search", {
  load(test_path("data", "test_objects.RData"))

  p <- autoplot(mt_spln_knn_bo_sep)
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c(
      "Distance Weighting Function",
      "mean",
      "# resamples",
      ".metric",
      "name",
      "value"
    )
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(value))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(p$labels$y, "")
  expect_equal(p$labels$x, "")
  expect_equal(p$labels$colour, "Distance Weighting Function")

  p <- autoplot(mt_spln_knn_bo_sep, metric = "rmse")
  expect_true(isTRUE(all.equal(unique(p$data$.metric), "rmse")))
})


test_that("performance plot for iterative search", {
  load(test_path("data", "test_objects.RData"))

  p <- autoplot(mt_spln_knn_bo_sep, type = "performance")
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c(
      "K",
      "weight_func",
      "deg_free",
      ".metric",
      ".estimator",
      "mean",
      "n",
      "std_err",
      ".config",
      ".iter"
    )
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(.iter))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(p$labels$x, "Iteration")

  p <- autoplot(mt_spln_knn_bo_sep, type = "performance", metric = "rmse")
  expect_true(isTRUE(all.equal(unique(p$data$.metric), "rmse")))

  p <- autoplot(mt_spln_knn_bo_sep, type = "performance", width = 0)
  expect_true(isTRUE(all.equal(names(p$mapping), c("x", "y"))))
})


test_that("parameter plot for iterative search", {
  load(test_path("data", "test_objects.RData"))

  p <- autoplot(mt_spln_knn_bo_sep, type = "parameters")
  expect_s3_class(p, "ggplot")
  expect_equal(names(p$data), c(".iter", "name", "value"))

  name_vals <- sort(unique(p$data$name))
  expect_equal(name_vals, c("K", "Spline Degrees of Freedom"))
  expect_equal(rlang::get_expr(p$mapping$x), expr(.iter))
  expect_equal(rlang::get_expr(p$mapping$y), expr(value))
  expect_equal(p$labels$y, "")
  expect_equal(p$labels$x, "Iteration")
})


test_that("regular grid plot", {
  skip_if_not_installed("scales", "1.3.0")
  rcv_results <- readRDS(test_path("data", "rcv_results.rds"))
  svm_reg_results <- readRDS(test_path("data", "svm_reg_results.rds"))

  p <- autoplot(rcv_results)
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c(
      "degree",
      "wt df",
      "wt degree",
      "mean",
      "# resamples",
      ".metric",
      "name",
      "value"
    )
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(value))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(rlang::get_expr(p$mapping$col), sym("wt df"))
  expect_equal(rlang::get_expr(p$mapping$group), sym("wt df"))
  expect_equal(p$facet$vars(), c(".metric", "degree", "wt degree"))

  expect_equal(p$labels$y, "")
  expect_equal(p$labels$x, "deg_free")

  p <- autoplot(rcv_results, metric = "rmse")
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c(
      "degree",
      "wt df",
      "wt degree",
      "mean",
      "# resamples",
      ".metric",
      "name",
      "value"
    )
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(value))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(rlang::get_expr(p$mapping$col), sym("wt df"))
  expect_equal(rlang::get_expr(p$mapping$group), sym("wt df"))
  expect_equal(p$facet$vars(), c("degree", "wt degree"))

  expect_equal(p$labels$y, "rmse")
  expect_equal(p$labels$x, "deg_free")

  p <- autoplot(svm_reg_results)
  expect_s3_class(p, "ggplot")
  expect_equal(
    names(p$data),
    c("%^*#", "Scale Factor", "mean", "# resamples", ".metric", "name", "value")
  )
  expect_equal(rlang::get_expr(p$mapping$x), expr(value))
  expect_equal(rlang::get_expr(p$mapping$y), expr(mean))
  expect_equal(rlang::get_expr(p$mapping$col), sym("%^*#"))
  expect_equal(rlang::get_expr(p$mapping$group), sym("%^*#"))
  expect_equal(p$facet$vars(), c(".metric", "Scale Factor"))

  expect_equal(p$labels$y, "")
  expect_equal(p$labels$colour, as.name("%^*#"))
  expect_equal(p$labels$x, "Cost")

  expect_true(grepl("^trans", class(p$scales$scales[[1]]$trans)))
  expect_equal(p$scales$scales[[1]]$trans$name, "log-2")
  expect_equal(unique(p$data$name), "Cost")
})


test_that("coord_obs_pred", {
  skip_if_not_installed("modeldata")

  data(solubility_test, package = "modeldata")

  p <-
    ggplot2::ggplot(
      solubility_test,
      ggplot2::aes(x = solubility, y = prediction)
    ) +
    ggplot2::geom_abline(lty = 2) +
    ggplot2::geom_point(alpha = 0.5)

  rng <- range(solubility_test[[1]], solubility_test[[2]])

  p2 <- p + coord_obs_pred()

  expect_no_error(print(p2))

  expect_true(inherits(p2$coordinates, "CoordObsPred"))
  expect_equal(p2$coordinates$limits$x, rng)
  expect_equal(p2$coordinates$limits$y, rng)
  expect_equal(p2$coordinates$ratio, 1)

  solubility_test$solubility[1] <- NA
  p3 <-
    ggplot2::ggplot(
      solubility_test,
      ggplot2::aes(x = solubility, y = prediction)
    ) +
    ggplot2::geom_abline(lty = 2) +
    ggplot2::geom_point(alpha = 0.5)
  expect_snapshot_warning(print(p3 + coord_obs_pred()))
})

test_that("1D regular grid x labels", {
  skip_if_not_installed("kernlab")

  set.seed(1)
  res <-
    parsnip::svm_rbf(cost = tune()) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("regression") |>
    tune_grid(mpg ~ ., resamples = rsample::vfold_cv(mtcars, v = 5), grid = 3)
  expect_equal(autoplot(res)$labels$x, c(cost = "Cost"))
})

test_that("plot_regular_grid with fairness metrics (#773)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor("classification", "kknn", neighbors = tune())
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  res <- tune_grid(
    knn,
    vs ~ mpg + hp + cyl,
    resamples = boots,
    grid = n_grid,
    metrics = yardstick::metric_set(
      yardstick::roc_auc,
      yardstick::demographic_parity(cyl),
      yardstick::demographic_parity(am)
    )
  )

  res_plot <- autoplot(res)

  expect_contains(
    res_plot$data$.metric,
    c("demographic_parity(am)", "demographic_parity(cyl)", "roc_auc")
  )
})

test_that("plot_marginals with fairness metrics (#773)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor(
    "classification",
    "kknn",
    neighbors = tune(),
    weight_func = tune()
  )
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  res <- tune_grid(
    knn,
    vs ~ mpg + hp + cyl,
    resamples = boots,
    grid = n_grid,
    metrics = yardstick::metric_set(
      yardstick::roc_auc,
      yardstick::demographic_parity(cyl),
      yardstick::demographic_parity(am)
    )
  )

  res_plot <- autoplot(res)

  expect_contains(
    res_plot$data$.metric,
    c("demographic_parity(am)", "demographic_parity(cyl)", "roc_auc")
  )
})

test_that("plot_perf_vs_iter with fairness metrics (#773)", {
  skip_on_cran()
  skip_if_not_installed("yardstick", minimum_version = "1.2.0.9001")
  skip_if_not_installed("kknn")

  knn <- parsnip::nearest_neighbor("classification", "kknn", neighbors = tune())
  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)
  boots <- rsample::bootstraps(mtcars_fair, 3)
  n_grid <- 3

  set.seed(1)
  suppressMessages(
    res <- tune_bayes(
      knn,
      vs ~ mpg + hp + cyl,
      resamples = boots,
      metrics = yardstick::metric_set(
        yardstick::roc_auc,
        yardstick::demographic_parity(cyl),
        yardstick::demographic_parity(am)
      )
    )
  )

  res_plot <- autoplot(res, type = "performance")

  expect_contains(
    res_plot$data$.metric,
    c("demographic_parity(am)", "demographic_parity(cyl)", "roc_auc")
  )
})

test_that("regular grid plot", {
  skip_if_not_installed("ggplot2", minimum_version = "3.5.0")
  skip_if_not_installed("kernlab")

  svm_spec <-
    parsnip::svm_rbf(cost = tune()) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("regression")

  svm_grid <-
    svm_spec |>
    extract_parameter_set_dials() |>
    dials::grid_regular(levels = 1)

  set.seed(1)
  res <-
    svm_spec |>
    tune_grid(
      mpg ~ .,
      resamples = rsample::vfold_cv(mtcars, v = 5),
      grid = svm_grid
    )

  expect_snapshot(
    error = TRUE,
    autoplot(res)
  )
})

test_that("evaluation time warning for non-survival model", {
  skip_if_not_installed("kernlab")

  set.seed(1)
  res <-
    parsnip::svm_rbf(cost = tune()) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("regression") |>
    tune_grid(mpg ~ ., resamples = rsample::vfold_cv(mtcars, v = 5), grid = 2)

  expect_snapshot(foo <- autoplot(res, metric = "rmse", eval_time = 10))
})

test_that("metric inputs are checked for regression models", {
  skip_if_not_installed("kknn")

  library(parsnip)
  library(workflows)
  library(yardstick)
  library(rsample)

  wflow <- workflow(y ~ X1 + X2, linear_reg())
  knn_spec <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")
  wflow_tune <- workflow(y ~ X1 + X2, knn_spec)

  set.seed(1)
  split <- initial_split(mtcars)
  rs <- vfold_cv(mtcars)

  # ------------------------------------------------------------------------------
  # setup metric sets

  met_mix_int <-
    metric_set(brier_survival_integrated, brier_survival, concordance_survival)
  met_reg <- metric_set(rmse)
  met_cls <- metric_set(brier_class)

  # ------------------------------------------------------------------------------
  # check inputs

  expect_snapshot(check_metrics_arg(NULL, wflow))

  expect_snapshot(check_metrics_arg(met_reg, wflow))
  expect_snapshot(check_metrics_arg(met_cls, wflow), error = TRUE)
  expect_snapshot(check_metrics_arg(met_mix_int, wflow), error = TRUE)

  # ------------------------------------------------------------------------------
  # resampling

  expect_snapshot(fit_resamples(wflow, rs, metrics = met_cls), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_mix_int), error = TRUE)

  # ------------------------------------------------------------------------------
  # tuning

  expect_snapshot(tune_grid(wflow_tune, rs, metrics = met_cls), error = TRUE)
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_mix_int),
    error = TRUE
  )

  expect_snapshot(tune_bayes(wflow_tune, rs, metrics = met_cls), error = TRUE)
  expect_snapshot(
    tune_bayes(wflow_tune, rs, metrics = met_mix_int),
    error = TRUE
  )

  # ------------------------------------------------------------------------------
  # final fit

  expect_snapshot(last_fit(wflow, split, metrics = met_cls), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_mix_int), error = TRUE)
})

test_that("metric inputs are checked for classification models", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kknn")

  library(parsnip)
  library(workflows)
  library(yardstick)
  library(rsample)

  data(two_class_dat, package = "modeldata")
  wflow <- workflow(Class ~ A + B, logistic_reg())
  knn_spec <- nearest_neighbor(neighbors = tune()) |>
    set_mode("classification")
  wflow_tune <- workflow(Class ~ A + B, knn_spec)

  set.seed(1)
  split <- initial_split(two_class_dat)
  rs <- vfold_cv(two_class_dat)

  # ------------------------------------------------------------------------------
  # setup metric sets

  met_mix_int <-
    metric_set(brier_survival_integrated, brier_survival, concordance_survival)
  met_reg <- metric_set(rmse)
  met_cls <- metric_set(brier_class)

  # ------------------------------------------------------------------------------
  # check inputs

  expect_snapshot(check_metrics_arg(NULL, wflow))

  expect_snapshot(check_metrics_arg(met_reg, wflow), error = TRUE)
  expect_snapshot(check_metrics_arg(met_cls, wflow))
  expect_snapshot(check_metrics_arg(met_mix_int, wflow), error = TRUE)

  # ------------------------------------------------------------------------------
  # resampling

  expect_snapshot(fit_resamples(wflow, rs, metrics = met_reg), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_mix_int), error = TRUE)

  # ------------------------------------------------------------------------------
  # tuning

  expect_snapshot(tune_grid(wflow_tune, rs, metrics = met_reg), error = TRUE)
  expect_snapshot(
    tune_grid(wflow_tune, rs, metrics = met_mix_int),
    error = TRUE
  )

  expect_snapshot(tune_bayes(wflow_tune, rs, metrics = met_reg), error = TRUE)
  expect_snapshot(
    tune_bayes(wflow_tune, rs, metrics = met_mix_int),
    error = TRUE
  )

  # ------------------------------------------------------------------------------
  # final fit

  expect_snapshot(last_fit(wflow, split, metrics = met_reg), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_mix_int), error = TRUE)
})


test_that("metric inputs are checked for censored regression models", {
  skip_if_not_installed("censored")
  library(parsnip)
  library(workflows)
  library(yardstick)
  library(rsample)
  library(censored)

  stanford2$event_time <- Surv(stanford2$time, stanford2$status)
  stanford2 <- stanford2[, c("event_time", "age")]

  wflow <- workflow(event_time ~ age, survival_reg())
  sr_spec <- survival_reg(dist = tune())
  wflow_tune <- workflow(event_time ~ age, sr_spec)

  set.seed(1)
  split <- initial_split(stanford2)
  rs <- vfold_cv(stanford2)

  # ------------------------------------------------------------------------------
  # setup metric sets

  met_srv <- metric_set(concordance_survival)
  met_reg <- metric_set(rmse)
  met_cls <- metric_set(brier_class)

  # ------------------------------------------------------------------------------
  # check inputs

  expect_snapshot(check_metrics_arg(NULL, wflow))

  expect_snapshot(check_metrics_arg(met_reg, wflow), error = TRUE)
  expect_snapshot(check_metrics_arg(met_cls, wflow), error = TRUE)
  expect_snapshot(check_metrics_arg(met_srv, wflow))

  # ------------------------------------------------------------------------------
  # resampling

  expect_snapshot(fit_resamples(wflow, rs, metrics = met_cls), error = TRUE)
  expect_snapshot(fit_resamples(wflow, rs, metrics = met_reg), error = TRUE)

  # ------------------------------------------------------------------------------
  # tuning

  expect_snapshot(tune_grid(wflow_tune, rs, metrics = met_cls), error = TRUE)
  expect_snapshot(tune_grid(wflow_tune, rs, metrics = met_reg), error = TRUE)

  expect_snapshot(tune_bayes(wflow_tune, rs, metrics = met_cls), error = TRUE)
  expect_snapshot(tune_bayes(wflow_tune, rs, metrics = met_reg), error = TRUE)

  # ------------------------------------------------------------------------------
  # final fit

  expect_snapshot(last_fit(wflow, split, metrics = met_cls), error = TRUE)
  expect_snapshot(last_fit(wflow, split, metrics = met_reg), error = TRUE)
})

test_that("check for finalization with engine parameters", {
  pset_1 <- parameters(dials::mtry(), dials::penalty(), dials::mixture())
  pset_2 <- pset_1
  pset_2$object[[3]] <- NA

  pset_3 <- parameters(dials::mtry(1:2), dials::penalty(), dials::mixture())
  pset_4 <- pset_3
  pset_4$object[[3]] <- NA

  expect_true(.needs_finalization(pset_1))
  expect_true(.needs_finalization(pset_2))
  expect_true(.needs_finalization(pset_1, "potato"))
  expect_true(.needs_finalization(pset_2, "potato"))

  expect_false(.needs_finalization(pset_1, "mtry"))
  expect_false(.needs_finalization(pset_2, "mtry"))
  expect_false(.needs_finalization(pset_3, "mtry"))
  expect_false(.needs_finalization(pset_4, "mtry"))
  expect_false(.needs_finalization(pset_3))
  expect_false(.needs_finalization(pset_4))
})

## -----------------------------------------------------------------------------

test_that("tuning with engine parameters with dials objects", {
  skip_if_not_installed("randomForest")
  skip_if(utils::packageVersion("dials") <= "0.0.7")

  rf_mod <-
    parsnip::rand_forest(min_n = tune()) |>
    parsnip::set_engine("randomForest", maxnodes = tune()) |>
    parsnip::set_mode("regression")

  set.seed(192)
  rs <- rsample::vfold_cv(mtcars)

  set.seed(19828)
  expect_no_error(
    suppressMessages(
      rf_tune <- rf_mod |> tune_grid(mpg ~ ., resamples = rs, grid = 3)
    )
  )
  expect_no_error(
    p <- autoplot(rf_tune)
  )

  set.seed(283)
  expect_no_error(
    suppressMessages(
      rf_search <- rf_mod |>
        tune_bayes(mpg ~ ., resamples = rs, initial = 3, iter = 2)
    )
  )
  expect_no_error(
    p <- autoplot(rf_search)
  )
})

## -----------------------------------------------------------------------------

test_that("tuning with engine parameters without dials objects", {
  skip_if_not_installed("randomForest")
  skip_if(utils::packageVersion("dials") <= "0.0.7")

  ## ---------------------------------------------------------------------------

  rf_mod <-
    parsnip::rand_forest(min_n = tune()) |>
    parsnip::set_engine("randomForest", corr.bias = tune()) |>
    parsnip::set_mode("regression")

  grid <-
    data.frame(
      min_n = c(5, 10, 5, 10),
      corr.bias = c(TRUE, TRUE, FALSE, FALSE)
    )

  set.seed(192)
  rs <- rsample::vfold_cv(mtcars)

  ## ---------------------------------------------------------------------------

  expect_snapshot(error = TRUE, {
    rf_tune <- rf_mod |> tune_grid(mpg ~ ., resamples = rs, grid = 3)
  })

  ## ---------------------------------------------------------------------------

  expect_no_error(
    suppressMessages(
      rf_tune <- rf_mod |> tune_grid(mpg ~ ., resamples = rs, grid = grid)
    )
  )
  expect_snapshot(error = TRUE, {
    p <- autoplot(rf_tune)
  })

  ## ---------------------------------------------------------------------------

  set.seed(283)
  expect_snapshot(error = TRUE, {
    rf_search <- rf_mod |> tune_bayes(mpg ~ ., resamples = rs)
  })
})

#' Augment data with holdout predictions
#'
#' For `tune` objects that use resampling, these `augment()` methods will add
#' one or more columns for the hold-out predictions (i.e. from the assessment
#' set(s)).
#'
#' @param x An object resulting from one of the `tune_*()` functions,
#'  `fit_resamples()`, or `last_fit()`. The control specifications for these
#'  objects should have used the option `save_pred = TRUE`.
#' @param ... Not currently used.
#' @param parameters A data frame with a single row that indicates what
#' tuning parameters should be used to generate the predictions (for `tune_*()`
#' objects only). If `NULL`, `select_best(x)` will be used with the first
#' metric and, if applicable, the first evaluation time point, used to
#' create `x`.
#' @return A data frame with one or more additional columns for model
#' predictions.
#'
#' @details
#' For some resampling methods where rows may be replicated in multiple
#' assessment sets, the prediction columns will be averages of the holdout
#' results. Also, for these methods, it is possible that all rows of the
#' original data do not have holdout predictions (like a single bootstrap
#' resample). In this case, all rows are return and a warning is issued.
#'
#' For objects created by `last_fit()`, the test set data and predictions are
#' returned.
#'
#' Unlike other `augment()` methods, the predicted values for regression models
#' are in a column called `.pred` instead of `.fitted` (to be consistent with
#' other tidymodels conventions).
#'
#' For regression problems, an additional `.resid` column is added to the
#' results.
#'
#' @export
augment.tune_results <- function(x, ..., parameters = NULL) {
  rlang::check_dots_empty()

  # check/determine best settings
  if (is.null(parameters)) {
    obj_fun <- .get_tune_metric_names(x)[1]
    obj_eval_time <- choose_eval_time(
      x,
      metric = obj_fun,
      eval_time = NULL,
      quietly = TRUE
    )
    parameters <- select_best(x, metric = obj_fun, eval_time = obj_eval_time)
  } else {
    if (!is.data.frame(parameters) || nrow(parameters) > 1) {
      cli::cli_abort("{.arg parameters} should be a single row data frame.")
    }
  }

  pred <- collect_predictions(x, summarize = TRUE, parameters = parameters)
  y_nm <- .get_tune_outcome_names(x)
  merge_pred(x$splits[[1]]$data, pred, y_nm)
}

#' @rdname augment.tune_results
#' @export
augment.resample_results <- function(x, ...) {
  rlang::check_dots_empty()

  pred <- collect_predictions(x, summarize = TRUE)
  y_nm <- .get_tune_outcome_names(x)
  merge_pred(x$splits[[1]]$data, pred, y_nm)
}


#' @rdname augment.tune_results
#' @export
augment.last_fit <- function(x, ...) {
  rlang::check_dots_empty()

  pred <-
    x |>
    collect_predictions(summarize = TRUE) |>
    dplyr::arrange(.row)
  y_nm <- .get_tune_outcome_names(x)
  dat <-
    rsample::assessment(x$splits[[1]]) |>
    dplyr::mutate(
      .row = as.integer(x$splits[[1]], data = "assessment")
    ) |>
    dplyr::arrange(.row)

  merge_pred(dat, pred, y_nm)
}

merge_pred <- function(dat, pred, y) {
  pred_cols <- grep("^\\.pred", names(pred), value = TRUE)
  pred <- pred[order(pred$.row), ]
  pred <- pred[, c(".row", pred_cols)]
  if (nrow(pred) != nrow(dat)) {
    cli::cli_warn(
      "The original data had {nrow(dat)} rows but there were {nrow(pred)}
       hold-out predictions."
    )
  }

  if (!tibble::is_tibble(dat)) {
    dat <- tibble::as_tibble(dat)
  }

  if (!any(names(dat) == ".row")) {
    dat$.row <- seq_len(nrow(dat))
  }

  dat <- dplyr::left_join(dat, pred, by = ".row")
  dat$.row <- NULL
  if (all(pred_cols == ".pred")) {
    dat$.resid <- dat[[y]] - dat$.pred
  }
  reorder_pred_cols(dat, outcome = y, param = character(0))
}

step_logging_helper <-
  function(
    recipe,
    ...,
    type = NULL,
    role = "predictor",
    trained = FALSE,
    skip = FALSE,
    id = recipes::rand_id("logging_helper")
  ) {
    recipes::add_step(
      recipe,
      step_logging_helper_new(
        terms = rlang::enquos(...),
        type = type,
        role = role,
        trained = trained,
        skip = skip,
        id = id
      )
    )
  }

step_logging_helper_new <-
  function(terms, type, role, trained, skip, id) {
    recipes::step(
      subclass = "logging_helper",
      terms = terms,
      type = type,
      role = role,
      trained = trained,
      skip = skip,
      id = id
    )
  }

prep.step_logging_helper <- function(x, training, info = NULL, ...) {
  if (identical(x$type, "error")) {
    stop("testing error")
  }
  if (identical(x$type, "warning")) {
    warning("testing warning")
  }
  if (identical(x$type, "message")) {
    message("testing message")
  }

  step_logging_helper_new(
    terms = x$terms,
    type = x$type,
    role = x$role,
    trained = TRUE,
    skip = x$skip,
    id = x$id
  )
}

bake.step_logging_helper <- function(object, new_data, ...) {
  new_data
}

catalog_lines <- function(lines) {
  lines[grepl("^>", lines)]
}

test_that("preprocessor error doesn't stop grid", {
  skip_if_not_installed("modeldata")

  # Errors in preprocessing since the formula method will cry about list columns
  ames <- modeldata::ames[, c(72, 40:45)]
  ames$First_Flr_SF <- as.list(ames$First_Flr_SF)

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  exp <- bind_cols(
    folds,
    tibble(
      .metrics = list(NULL)
    )
  )

  expect_snapshot(
    res_fit <- tune_grid(
      parsnip::nearest_neighbor("regression", "kknn", dist_power = tune()),
      Sale_Price ~ .,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  expect_identical(
    dplyr::select(res_fit, -.notes),
    exp,
    ignore_attr = TRUE
  )

  expect_identical(nrow(res_fit$.notes[[1]]), 1L)
  expect_identical(ncol(res_fit$.notes[[1]]), 4L)
  expect_identical(
    vapply(res_fit$.notes[[1]], class, character(1)),
    c(
      location = "character",
      type = "character",
      note = "character",
      trace = "list"
    )
  )
})

test_that("model error doesn't stop grid", {
  skip_if_not_installed("modeldata")

  # Errors in modeling since the recipe passes the list columns through
  ames <- modeldata::ames[, c(72, 40:45)]
  ames$First_Flr_SF <- as.list(ames$First_Flr_SF)

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  exp <- bind_cols(
    folds,
    tibble(
      .metrics = list(NULL)
    )
  )

  rec_spec <- recipe(Sale_Price ~ ., ames)
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    dist_power = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  expect_identical(
    dplyr::select(res_fit, -.notes),
    exp,
    ignore_attr = TRUE
  )

  expect_identical(nrow(res_fit$.notes[[1]]), 2L)
  expect_identical(ncol(res_fit$.notes[[1]]), 4L)
  expect_identical(
    vapply(res_fit$.notes[[1]], class, character(1)),
    c(
      location = "character",
      type = "character",
      note = "character",
      trace = "list"
    )
  )
})

test_that("prediction error doesn't stop grid", {
  skip_if_not_installed("modeldata")

  # Errors in predictions as we have injected NA values into the testing splits
  ames <- modeldata::ames[, c(72, 40:45)]

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)
  for (i in seq_len(2)) {
    train_ids <- setdiff(seq_len(nrow(ames)), folds$splits[[i]]$in_id)
    folds$splits[[i]]$data$First_Flr_SF[train_ids] <- NA
  }

  exp <- bind_cols(
    folds,
    tibble(
      .metrics = list(NULL)
    )
  )

  rec_spec <- recipe(Sale_Price ~ ., ames)
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    dist_power = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  expect_identical(
    dplyr::select(res_fit, -.notes),
    exp,
    ignore_attr = TRUE
  )

  expect_identical(nrow(res_fit$.notes[[1]]), 2L)
  expect_identical(ncol(res_fit$.notes[[1]]), 4L)
  expect_identical(
    vapply(res_fit$.notes[[1]], class, character(1)),
    c(
      location = "character",
      type = "character",
      note = "character",
      trace = "list"
    )
  )
})

test_that("capturing error correctly in notes", {
  skip_if_not_installed("modeldata")

  assign(
    "prep.step_logging_helper",
    prep.step_logging_helper,
    envir = .GlobalEnv
  )
  assign(
    "bake.step_logging_helper",
    bake.step_logging_helper,
    envir = .GlobalEnv
  )
  ames <- modeldata::ames[, c(72, 40:45)]

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames) |>
    step_logging_helper(type = "error")
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    dist_power = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  exp <- tibble::tibble(
    location = "preprocessor 1/1",
    type = "error",
    note = "testing error"
  )

  expect_true(
    all(vapply(res_fit$.metrics, NROW, integer(1)) == 0)
  )
  rm(list = "prep.step_logging_helper", envir = .GlobalEnv)
  rm(list = "bake.step_logging_helper", envir = .GlobalEnv)
})

test_that("capturing warning correctly in notes", {
  skip_if_not_installed("modeldata")

  assign(
    "prep.step_logging_helper",
    prep.step_logging_helper,
    envir = .GlobalEnv
  )
  assign(
    "bake.step_logging_helper",
    bake.step_logging_helper,
    envir = .GlobalEnv
  )

  ames <- modeldata::ames[, c(72, 40:45)]

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames) |>
    step_logging_helper(type = "warning")
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    dist_power = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)
  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )
  exp <- tibble::tibble(
    location = "preprocessor 1/1",
    type = "warning",
    note = "testing warning",
    trace = res_fit$.notes[[1]]$trace
  )
  expect_identical(res_fit$.notes[[1]], exp)
  expect_identical(res_fit$.notes[[2]], exp)

  expect_true(
    all(vapply(res_fit$.metrics, NROW, integer(1)) > 0)
  )
  rm(list = "prep.step_logging_helper", envir = .GlobalEnv)
  rm(list = "bake.step_logging_helper", envir = .GlobalEnv)
})

test_that("doesn't capturing message in notes", {
  skip_if_not_installed("modeldata")

  assign(
    "prep.step_logging_helper",
    prep.step_logging_helper,
    envir = .GlobalEnv
  )
  assign(
    "bake.step_logging_helper",
    bake.step_logging_helper,
    envir = .GlobalEnv
  )

  ames <- modeldata::ames[, c(72, 40:45)]

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames) |>
    step_logging_helper(type = "message")
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    dist_power = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  exp <- tibble::tibble(
    location = character(0),
    type = character(0),
    note = character(0),
    trace = list()
  )

  expect_identical(res_fit$.notes[[1]], exp)
  expect_identical(res_fit$.notes[[2]], exp)

  expect_true(
    all(vapply(res_fit$.metrics, NROW, integer(1)) > 0)
  )
  rm(list = "prep.step_logging_helper", envir = .GlobalEnv)
  rm(list = "bake.step_logging_helper", envir = .GlobalEnv)
})

test_that("captures extract errors", {
  skip_if_not_installed("modeldata")

  set.seed(1234)
  ames <- modeldata::ames[, c(72, 40:45)]
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames)
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    neighbors = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  extract_error <- function(x) {
    stop("extract error")
  }

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE, extract = extract_error)
    ),
    transform = catalog_lines
  )

  expect_identical(
    nrow(collect_notes(res_fit)),
    4L
  )
})

test_that("captures kknn R errors", {
  skip_if_not_installed("modeldata")

  # kknn is gonna complain because one of the predictors is Inf
  ames <- modeldata::ames[, c(72, 40:45)]
  ames$Second_Flr_SF <- Inf

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames)
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    neighbors = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  expect_identical(
    nrow(collect_notes(res_fit)),
    2L
  )

  exp <- try(
    fit(finalize_workflow(wf_spec, tibble(neighbors = 10)), ames),
    silent = TRUE
  )
  exp <- attr(exp, "condition")
  exp <- conditionMessage(exp)

  expect_identical(
    rep(exp, 2),
    collect_notes(res_fit)$note
  )
})

test_that("captures xgboost C errors", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("modeldata")

  # xgboost is gonna complain because one of the predictors is Inf
  ames <- modeldata::ames[, c(72, 40:45)]
  ames$Second_Flr_SF <- Inf

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames)
  mod_spec <- parsnip::boost_tree(
    "regression",
    "xgboost",
    trees = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  trim_timestamp <- function(lines) {
    lines <- catalog_lines(lines)
    lines[!grepl("\\[[0-9:]+\\]", lines)]
  }

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = trim_timestamp
  )

  expect_identical(
    nrow(collect_notes(res_fit)),
    2L
  )

  exp <- try(
    fit(finalize_workflow(wf_spec, tibble(trees = 10)), ames),
    silent = TRUE
  )
  exp <- attr(exp, "condition")
  exp <- conditionMessage(exp)

  trim_error <- function(x) {
    gsub("\\[[0-9:]+\\]", "", x)
  }

  expect_identical(
    trim_error(
      rep(exp, 2)
    ),
    trim_error(
      collect_notes(res_fit)$note
    )
  )
})

test_that("captures cli styled errors", {
  skip_if_not_installed("modeldata")

  assign(
    "prep.step_logging_helper",
    prep.step_logging_helper,
    envir = .GlobalEnv
  )
  assign(
    "bake.step_logging_helper",
    bake.step_logging_helper,
    envir = .GlobalEnv
  )
  ames <- modeldata::ames[, c(72, 40:45)]

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames) |>
    step_logging_helper(type = "error")
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    dist_power = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  expect_identical(
    nrow(collect_notes(res_fit)),
    2L
  )

  exp <- try(
    fit(finalize_workflow(wf_spec, tibble(dist_power = 10)), ames),
    silent = TRUE
  )
  exp <- attr(exp, "condition")
  exp <- conditionMessage(exp)

  expect_identical(
    rep(exp, 2),
    collect_notes(res_fit)$note
  )
})


test_that("emitter works with errors", {
  skip_if_not_installed("modeldata")

  assign(
    "prep.step_logging_helper",
    prep.step_logging_helper,
    envir = .GlobalEnv
  )
  assign(
    "bake.step_logging_helper",
    bake.step_logging_helper,
    envir = .GlobalEnv
  )
  ames <- modeldata::ames[, c(72, 40:45)]

  set.seed(1234)
  folds <- rsample::vfold_cv(ames, 2)

  rec_spec <- recipe(Sale_Price ~ ., ames) |>
    step_logging_helper(type = "error")
  mod_spec <- parsnip::nearest_neighbor(
    "regression",
    "kknn",
    dist_power = tune()
  )

  wf_spec <- workflow(rec_spec, mod_spec)

  expect_snapshot(
    res_fit <- tune_grid(
      wf_spec,
      folds,
      grid = 2,
      control = control_grid(allow_par = FALSE)
    ),
    transform = catalog_lines
  )

  exp <- tibble::tibble(
    location = "preprocessor 1/1",
    type = "error",
    note = "testing error"
  )

  expect_true(
    all(vapply(res_fit$.metrics, NROW, integer(1)) == 0)
  )
  rm(list = "prep.step_logging_helper", envir = .GlobalEnv)
  rm(list = "bake.step_logging_helper", envir = .GlobalEnv)
})

#' Extract elements of `tune` objects
#'
#' @description
#' These functions extract various elements from a tune object. If they do
#' not exist yet, an error is thrown.
#'
#' - [`extract_preprocessor()`][extract_preprocessor.tune_results()] returns
#'   the formula, recipe, or variable
#'   expressions used for preprocessing.
#'
#' - [`extract_spec_parsnip()`][extract_spec_parsnip.tune_results()] returns
#'   the parsnip model specification.
#'
#' - [`extract_fit_parsnip()`][extract_fit_parsnip.tune_results()] returns the
#'   parsnip model fit object.
#'
#' - [`extract_fit_engine()`][extract_fit_engine.tune_results()] returns the
#'   engine specific fit embedded within
#'   a parsnip model fit. For example, when using [parsnip::linear_reg()]
#'   with the `"lm"` engine, this returns the underlying `lm` object.
#'
#' - [`extract_mold()`][extract_mold.tune_results()] returns the preprocessed
#'   "mold" object returned
#'   from [hardhat::mold()]. It contains information about the preprocessing,
#'   including either the prepped recipe, the formula terms object, or
#'   variable selectors.
#'
#' - [`extract_recipe()`][extract_recipe.tune_results()] returns the recipe.
#'   The `estimated` argument specifies
#'   whether the fitted or original recipe is returned.
#'
#' - [`extract_workflow()`][extract_workflow.tune_results()] returns the
#'   workflow object if the control option
#'   `save_workflow = TRUE` was used. The workflow will only have been
#'   estimated for objects produced by [last_fit()].
#'
#' @param x A `tune_results` object.
#' @param estimated A logical for whether the original (unfit) recipe or the
#' fitted recipe should be returned.
#' @param ... Not currently used.
#' @details
#' These functions supersede `extract_model()`.
#' @return
#' The extracted value from the `tune` tune_results, `x`, as described in the
#' description section.
#'
#' @name extract-tune
#' @examplesIf rlang::is_installed("splines2")
#' # example code
#'
#' library(recipes)
#' library(rsample)
#' library(parsnip)
#'
#' set.seed(6735)
#' tr_te_split <- initial_split(mtcars)
#'
#' spline_rec <- recipe(mpg ~ ., data = mtcars) |>
#'   step_spline_natural(disp)
#'
#' lin_mod <- linear_reg() |>
#'   set_engine("lm")
#'
#' spline_res <- last_fit(lin_mod, spline_rec, split = tr_te_split)
#'
#' extract_preprocessor(spline_res)
#'
#' # The `spec` is the parsnip spec before it has been fit.
#' # The `fit` is the fitted parsnip model.
#' extract_spec_parsnip(spline_res)
#' extract_fit_parsnip(spline_res)
#' extract_fit_engine(spline_res)
#'
#' # The mold is returned from `hardhat::mold()`, and contains the
#' # predictors, outcomes, and information about the preprocessing
#' # for use on new data at `predict()` time.
#' extract_mold(spline_res)
#'
#' # A useful shortcut is to extract the fitted recipe from the workflow
#' extract_recipe(spline_res)
#'
#' # That is identical to
#' identical(
#'   extract_mold(spline_res)$blueprint$recipe,
#'   extract_recipe(spline_res)
#' )
NULL

#' @export
#' @rdname extract-tune
extract_workflow.last_fit <- function(x, ...) {
  x$.workflow[[1]]
}


#' @export
#' @rdname extract-tune
extract_workflow.tune_results <- function(x, ...) {
  .get_tune_workflow(x)
}

#' @export
#' @rdname extract-tune
extract_spec_parsnip.tune_results <- function(x, ...) {
  extract_spec_parsnip(extract_workflow(x))
}

#' @export
#' @rdname extract-tune
extract_recipe.tune_results <- function(x, ..., estimated = TRUE) {
  check_empty_dots(...)
  check_bool(estimated)
  extract_recipe(extract_workflow(x), estimated = estimated)
}
check_empty_dots <- function(...) {
  opts <- list(...)
  if (any(names(opts) == "estimated")) {
    cli::cli_abort("'{.arg estimated}' should be a named argument.")
  }
  if (length(opts) > 0) {
    cli::cli_warn(
      "The {.code ...} are not used in this function but {length(opts)}
       object{?s} {?was/were} passed: {.val {names(opts)}}"
    )
  }
  invisible(NULL)
}


#' @export
#' @rdname extract-tune
extract_fit_parsnip.tune_results <- function(x, ...) {
  extract_fit_parsnip(extract_workflow(x))
}

#' @export
#' @rdname extract-tune
extract_fit_engine.tune_results <- function(x, ...) {
  extract_fit_engine(extract_workflow(x))
}

#' @export
#' @rdname extract-tune
extract_mold.tune_results <- function(x, ...) {
  extract_mold(extract_workflow(x))
}

#' @export
#' @rdname extract-tune
extract_preprocessor.tune_results <- function(x, ...) {
  extract_preprocessor(extract_workflow(x))
}

#' Fit the final best model to the training set and evaluate the test set
#'
#' [last_fit()] emulates the process where, after determining the best model,
#' the final fit on the entire training set is needed and is then evaluated on
#' the test set.
#'
#' @param object A `parsnip` model specification or an unfitted
#'   [workflow()][workflows::workflow]. No tuning parameters are allowed; if arguments
#'   have been marked with [tune()][hardhat::tune], their values must be
#'   [finalized][finalize_model].
#'
#' @param preprocessor A traditional model formula or a recipe created using
#'   [recipes::recipe()].
#'
#' @param split An `rsplit` object created from [rsample::initial_split()] or
#' [rsample::initial_validation_split()].
#'
#' @param metrics A [yardstick::metric_set()], or `NULL` to compute a standard
#'   set of metrics.
#'
#' @param control A [control_last_fit()] object used to fine tune the last fit
#'   process.
#'
#' @inheritParams tune_grid
#'
#' @param add_validation_set For 3-way splits into training, validation, and test
#' set via [rsample::initial_validation_split()], should the validation set be
#' included in the data set used to train the model. If not, only the training
#' set is used.
#'
#' @param ... Currently unused.
#'
#' @details
#' This function is intended to be used after fitting a _variety of models_
#'  and the final tuning parameters (if any) have been finalized. The next step
#'  would be to fit using the entire training set and verify performance using
#'  the test data.
#'
#' @template case-weights
#' @template censored-regression
#'
#' @section See also:
#'
#' [last_fit()] is closely related to [fit_best()]. They both
#' give you access to a workflow fitted on the training data but are situated
#' somewhat differently in the modeling workflow. [fit_best()] picks up
#' after a tuning function like [tune_grid()] to take you from tuning results
#' to fitted workflow, ready for you to predict and assess further. [last_fit()]
#' assumes you have made your choice of hyperparameters and finalized your
#' workflow to then take you from finalized workflow to fitted workflow and
#' further to performance assessment on the test data. While [fit_best()] gives
#' a fitted workflow, [last_fit()] gives you the performance results. If you
#' want the fitted workflow, you can extract it from the result of [last_fit()]
#' via [extract_workflow()][extract_workflow.tune_results].
#'
#'
#' @return A single row tibble that emulates the structure of `fit_resamples()`.
#' However, a list column called `.workflow` is also attached with the fitted
#' model (and recipe, if any) that used the training set. Helper functions
#' for formatting tuning results like [collect_metrics()] and
#' [collect_predictions()] can be used with `last_fit()` output.
#' @examplesIf tune:::should_run_examples() & rlang::is_installed("splines2")
#' library(recipes)
#' library(rsample)
#' library(parsnip)
#'
#' set.seed(6735)
#' tr_te_split <- initial_split(mtcars)
#'
#' spline_rec <- recipe(mpg ~ ., data = mtcars) |>
#'   step_spline_natural(disp)
#'
#' lin_mod <- linear_reg() |>
#'   set_engine("lm")
#'
#' spline_res <- last_fit(lin_mod, spline_rec, split = tr_te_split)
#' spline_res
#'
#' # test set metrics
#' collect_metrics(spline_res)
#'
#' # test set predictions
#' collect_predictions(spline_res)
#'
#' # or use a workflow
#'
#' library(workflows)
#' spline_wfl <-
#'   workflow() |>
#'   add_recipe(spline_rec) |>
#'   add_model(lin_mod)
#'
#' last_fit(spline_wfl, split = tr_te_split)
#' @export
last_fit <- function(object, ...) {
  UseMethod("last_fit")
}

#' @export
last_fit.default <- function(object, ...) {
  cli::cli_abort(
    "The first argument to {.fn last_fit} should be either a model or workflow,
    not {.obj_type_friendly {object}}."
  )
}

#' @export
last_fit.model_fit <- function(object, ...) {
  cli::cli_abort(c(
    "{.help [{.fun last_fit}](tune::last_fit)} is not well-defined for \\
     fitted model objects.",
    "i" = "{.help [{.fun last_fit}](tune::last_fit)} takes \\
           a {.help [model specification](parsnip::model_spec)} or \\
           {.help [unfitted workflow](workflows::workflow)} as its first \\
           argument."
  ))
}

tune_pp_msg <- "To tune a model specification, you must preprocess with a
                formula, recipe, or {.pkg dplyr} selectors, not
                {.obj_type_friendly {preprocessor}}."

#' @export
#' @rdname last_fit
last_fit.model_spec <- function(
  object,
  preprocessor,
  split,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_last_fit(),
  add_validation_set = FALSE
) {
  if (rlang::is_missing(preprocessor) || !is_preprocessor(preprocessor)) {
    cli::cli_abort(tune_pp_msg)
  }

  control <- parsnip::condense_control(control, control_last_fit())

  empty_ellipses(...)

  wflow <- add_model(workflow(), object)

  if (is_recipe(preprocessor)) {
    wflow <- add_recipe(wflow, preprocessor)
  } else if (rlang::is_formula(preprocessor)) {
    wflow <- add_formula(wflow, preprocessor)
  }

  last_fit_workflow(
    wflow,
    split = split,
    metrics = metrics,
    eval_time = eval_time,
    control = control,
    add_validation_set = add_validation_set
  )
}


#' @rdname last_fit
#' @export
last_fit.workflow <- function(
  object,
  split,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_last_fit(),
  add_validation_set = FALSE
) {
  empty_ellipses(...)

  control <- parsnip::condense_control(control, control_last_fit())

  last_fit_workflow(
    object,
    split = split,
    metrics = metrics,
    eval_time = eval_time,
    control = control,
    add_validation_set = add_validation_set
  )
}


last_fit_workflow <- function(
  object,
  split,
  metrics,
  eval_time = NULL,
  control,
  add_validation_set = FALSE,
  ...,
  call = rlang::caller_env()
) {
  rlang::check_dots_empty()
  check_no_tuning(object)

  if (workflows::is_trained_workflow(object)) {
    cli::cli_abort(
      "`last_fit()` is not well-defined for a fitted workflow.",
      call = call
    )
  }

  if (inherits(split, "initial_validation_split")) {
    split <- prepare_validation_split(split, add_validation_set)
  }
  splits <- list(split)
  resamples <- rsample::manual_rset(splits, ids = "train/test split")

  res <- resample_workflow(
    workflow = object,
    resamples = resamples,
    metrics = metrics,
    eval_time = eval_time,
    control = control,
    call = call
  )

  res$.workflow <- res$.extracts[[1]][[1]]
  res$.extracts <- NULL
  class(res) <- c("last_fit", class(res))
  class(res) <- unique(class(res))

  .stash_last_result(res)
  res
}


prepare_validation_split <- function(split, add_validation_set) {
  if (add_validation_set) {
    # equivalent to (unexported) rsample:::rsplit() without checks
    split <- structure(
      list(
        data = split$data,
        in_id = c(split$train_id, split$val_id),
        out_id = NA
      ),
      class = "rsplit"
    )
  } else {
    id_train_test <- seq_len(nrow(split$data))[-sort(split$val_id)]
    id_train <- match(split$train_id, id_train_test)

    split <- structure(
      list(
        data = split$data[-sort(split$val_id), , drop = FALSE],
        in_id = id_train,
        out_id = NA
      ),
      class = "rsplit"
    )
  }

  split
}

test_that("extract methods for last_fit objects", {
  lm_spec <- parsnip::linear_reg() |> parsnip::set_engine("lm")
  lm_prn_fit <- parsnip::fit(lm_spec, mpg ~ ., data = mtcars)
  lm_wflow <-
    workflow() |>
    add_model(lm_spec) |>
    add_formula(mpg ~ .)
  lm_res <- last_fit(lm_wflow, split = rsample::initial_split(mtcars))

  expect_true(inherits(extract_fit_engine(lm_res), "lm"))
  expect_true(inherits(extract_fit_parsnip(lm_res), "model_fit"))
  expect_true(inherits(extract_mold(lm_res), "list"))
  expect_true(inherits(extract_preprocessor(lm_res), "formula"))
  expect_true(inherits(extract_spec_parsnip(lm_res), "model_spec"))
  expect_true(inherits(extract_workflow(lm_res), "workflow"))
})

test_that("extract methods for resample_results objects", {
  lm_spec <- parsnip::linear_reg() |> parsnip::set_engine("lm")
  lm_rec_wflow <-
    workflow() |>
    add_model(lm_spec) |>
    add_recipe(
      recipes::recipe(mpg ~ ., data = mtcars) |>
        recipes::step_normalize(recipes::all_numeric_predictors())
    )
  lm_rec_res <- fit_resamples(
    lm_rec_wflow,
    resamples = rsample::vfold_cv(mtcars, v = 2),
    control = control_resamples(save_workflow = TRUE)
  )

  expect_true(inherits(extract_recipe(lm_rec_res, estimated = FALSE), "recipe"))
  expect_true(!extract_recipe(lm_rec_res, estimated = FALSE)$steps[[1]]$trained)
  expect_true(inherits(extract_workflow(lm_rec_res), "workflow"))
})

test_that("predict errors informatively when supplied tuning results", {
  skip_on_cran()

  data("example_ames_knn")

  expect_snapshot(predict(ames_grid_search), error = TRUE)
})

test_that("predict errors informatively when supplied last_fit() results", {
  skip_on_cran()

  library(rsample)
  library(parsnip)

  set.seed(6735)
  tr_te_split <- initial_split(mtcars)

  spline_res <- last_fit(linear_reg(), mpg ~ ., split = tr_te_split)

  expect_snapshot(predict(spline_res), error = TRUE)
})

new_rng_snapshots <- utils::compareVersion(
  "3.6.0",
  as.character(getRversion())
) >
  0

# New (as of 4.3.0) a new option generates different snapshots
rankdeficient_version <- any(names(formals("predict.lm")) == "rankdeficient")


helper_objects_tune <- function() {
  rec_tune_1 <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_normalize(recipes::all_predictors()) |>
    recipes::step_pca(recipes::all_predictors(), num_comp = tune())

  rec_no_tune_1 <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_normalize(recipes::all_predictors())

  lm_mod <- parsnip::linear_reg() |> parsnip::set_engine("lm")

  svm_mod <- parsnip::svm_rbf(mode = "regression", cost = tune()) |>
    parsnip::set_engine("kernlab")

  list(
    rec_tune_1 = rec_tune_1,
    rec_no_tune_1 = rec_no_tune_1,
    lm_mod = lm_mod,
    svm_mod = svm_mod
  )
}

# When not interactive, the cli progress bar is occasionally "reset" at random
# intervals throughout tuning, which means that intermediate counts may appear
# in snapshots. In an output like:
#
# > A | error:   AHHhH
# There were issues with some computations   A: x2
# There were issues with some computations   A: x4
# There were issues with some computations   A: x5
# There were issues with some computations   A: x5
#
# ...we just want to see the unique issues, e.g.
#
# > A | error:   AHHhH
catalog_lines <- function(lines) {
  lines[grepl("^>", lines)]
}

# Make a new binding to prevent infinite recursion when the original is mocked.
initialize_catalog_ <- tune:::initialize_catalog

# Sets a new exit handler on `initialize_catalog()` that stores the summary
# of issues before it's cleared along with the progress bar. Together with
# the above, we can test the full catalog output.
redefer_initialize_catalog <- function(test_env) {
  local({
    function(control, env = rlang::caller_env(), workflow = NULL) {
      initialize_catalog_(control, env, workflow)

      withr::defer(
        assign(
          "catalog_summary_test",
          tune:::tune_env$progress_env$catalog_summary,
          test_env
        ),
        envir = env,
        priority = "first"
      )

      NULL
    }
  })
}

# ------------------------------------------------------------------------------
# Objects to test grid processing

if (rlang::is_installed("splines2")) {
  rec_no_tune <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_corr(all_predictors(), threshold = .1) |>
    recipes::step_spline_natural(disp, deg_free = 5)

  rec_tune <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_corr(all_predictors(), threshold = tune()) |>
    recipes::step_spline_natural(disp, deg_free = tune("disp_df"))
}

mod_no_tune <-
  parsnip::rand_forest(mode = "regression")

mod_tune <-
  parsnip::boost_tree(
    trees = tune(),
    min_n = tune(),
    mode = "regression"
  )

mod_tune_submodel <-
  parsnip::boost_tree(
    trees = tune(),
    mode = "regression"
  )

mod_tune_no_submodel <-
  parsnip::rand_forest(
    min_n = tune(),
    mode = "regression"
  )

if (rlang::is_installed("probably")) {
  tlr_tune <-
    tailor::tailor() |>
    tailor::adjust_numeric_range(lower_limit = tune())

  tlr_tune_cal <-
    tailor::tailor() |>
    tailor::adjust_numeric_calibration(method = "linear") |>
    tailor::adjust_numeric_range(lower_limit = tune())

  tlr_no_tune <-
    tailor::tailor() |>
    tailor::adjust_numeric_range(lower_limit = 0)
}

#' Tools for selecting metrics and evaluation times
#'
#' @param mtr_set A [yardstick::metric_set()].
#' @param metric A character value for which metric is being used.
#' @param eval_time An optional vector of times to compute dynamic and/or
#' integrated metrics.
#' @param wflow A [workflows::workflow()].
#' @param x An object with class `tune_results`.
#' @param quietly Logical. Should warnings be muffled?
#' @param call The call to be displayed in warnings or errors.
#' @inheritParams rlang::args_dots_empty
#' @details
#' These are developer-facing functions used to compute and validate choices
#' for performance metrics. For survival analysis models, there are similar
#' functions for the evaluation time(s) required for dynamic and/or integrated
#' metrics.
#'
#' `choose_metric()` is used with functions such as [show_best()] or
#' [select_best()] where a single valid metric is required to rank models. If
#' no value is given by the user, the first metric value is used (with a
#' warning).
#'
#' For evaluation times, one is only required when the metric type is dynamic
#' (e.g. [yardstick::brier_survival()] or [yardstick::roc_auc_survival()]). For
#' these metrics, we require a single numeric value that was originally given
#' to the function used to produce `x` (such as [tune_grid()]).
#'
#' If a time is required and none is given, the first value in the vector
#' originally given in the `eval_time` argument is used (with a warning).
#'
#' `maybe_choose_eval_time()` is for cases where multiple evaluation times are
#' acceptable but you need to choose a good default. The "maybe" is because
#' the function that would use `maybe_choose_eval_time()` can accept multiple
#' metrics (like [autoplot()]).
#' @keywords internal
#' @export
choose_metric <- function(x, metric, ..., call = rlang::caller_env()) {
  rlang::check_dots_empty()

  mtr_set <- .get_tune_metrics(x)
  mtr_info <- tibble::as_tibble(mtr_set)

  if (is.null(metric)) {
    metric <- mtr_info$metric[1]
    cli::cli_warn(
      "No value of {.arg metric} was given; {.val {metric}}
                   will be used.",
      call = call
    )
  } else {
    metric <- check_mult_metrics(metric, call = call)
    check_metric_in_tune_results(mtr_info, metric, call = call)
  }

  mtr_info[mtr_info$metric == metric, ]
}

check_mult_metrics <- function(metric, ..., call = rlang::caller_env()) {
  rlang::check_dots_empty()

  num_metrics <- length(metric)
  metric <- metric[1]
  if (num_metrics > 1) {
    cli::cli_warn(
      "{num_metrics} metric{?s} were given; {.val {metric}} will
                   be used.",
      call = call
    )
  }
  metric
}

#' @rdname choose_metric
#' @export
check_metric_in_tune_results <- function(
  mtr_info,
  metric,
  ...,
  call = rlang::caller_env()
) {
  rlang::check_dots_empty(call = call)
  if (!any(mtr_info$metric == metric)) {
    cli::cli_abort(
      "{.val {metric}} was not in the metric set. Please choose
                    from: {.val {mtr_info$metric}}.",
      call = call
    )
  }
  invisible(NULL)
}

check_enough_eval_times <- function(eval_time, metrics) {
  num_times <- length(eval_time)
  max_times_req <- req_eval_times(metrics)
  cls <- tibble::as_tibble(metrics)$class
  uni_cls <- sort(unique(cls))
  if (max_times_req > num_times) {
    cli::cli_abort(
      "At least {max_times_req} evaluation time{?s} {?is/are}
                   required for the metric type(s) requested: {.val {uni_cls}}.
                   Only {num_times} unique time{?s} {?was/were} given."
    )
  }

  invisible(NULL)
}

contains_survival_metric <- function(mtr_info) {
  any(grepl("_survival", mtr_info$class))
}

# choose_eval_time() is called by show_best(), select_best(), and augment()
#' @rdname choose_metric
#' @export
choose_eval_time <- function(
  x,
  metric,
  ...,
  eval_time = NULL,
  quietly = FALSE,
  call = rlang::caller_env()
) {
  rlang::check_dots_empty(call = call)
  mtr_set <- .get_tune_metrics(x)
  mtr_info <- tibble::as_tibble(mtr_set)

  if (!contains_survival_metric(mtr_info)) {
    if (!is.null(eval_time) & !quietly) {
      cli::cli_warn(
        "{.arg eval_time} is only used for models with mode {.val censored regression}.",
        call = call
      )
    }
    return(NULL)
  }

  dyn_metric <- is_dyn(mtr_set, metric)

  # If we don't need an eval time but one is passed:
  if (!dyn_metric & !is.null(eval_time) & !quietly) {
    cli::cli_warn(
      "{.arg eval_time} is only used for dynamic survival metrics.",
      call = call
    )
  }

  # If we need an eval time, set it to the possible values so that
  # we can choose the first value
  if (dyn_metric && is.null(eval_time)) {
    eval_time <- .get_tune_eval_times(x)
  }

  eval_time <- first_eval_time(
    mtr_set,
    metric = metric,
    eval_time = eval_time,
    quietly = quietly,
    call = call
  )

  check_eval_time_in_tune_results(x, eval_time, call = call)

  eval_time
}

is_dyn <- function(mtr_set, metric) {
  mtr_info <- tibble::as_tibble(mtr_set)
  mtr_cls <- mtr_info$class[mtr_info$metric == metric]
  mtr_cls == "dynamic_survival_metric"
}

check_eval_time_in_tune_results <- function(
  x,
  eval_time,
  call = rlang::caller_env()
) {
  given_times <- .get_tune_eval_times(x)
  if (!is.null(eval_time)) {
    if (!any(eval_time %in% given_times)) {
      print_time <- format(eval_time, digits = 3)
      cli::cli_abort(
        "Evaluation time {print_time} is not in the results.",
        call = call
      )
    }
  }
  invisible(NULL)
}

#' @rdname choose_metric
#' @export
maybe_choose_eval_time <- function(x, mtr_set, eval_time) {
  mtr_info <- tibble::as_tibble(mtr_set)
  if (any(grepl("integrated", mtr_info$metric))) {
    return(.get_tune_eval_times(x))
  }
  eval_time <- purrr::map(
    mtr_info$metric,
    \(.x) choose_eval_time(x, .x, eval_time = eval_time, quietly = TRUE)
  )
  no_eval_time <- purrr::map_lgl(eval_time, is.null)
  if (all(no_eval_time)) {
    eval_time <- NULL
  } else {
    eval_time <- sort(unique(unlist(eval_time)))
  }
  eval_time
}


# ------------------------------------------------------------------------------

#' @rdname choose_metric
#' @export
first_metric <- function(mtr_set) {
  tibble::as_tibble(mtr_set)[1, ]
}

# first_eval_time() is called by show_best(), select_best(), and augment() (by way of
# choose_eval_time()) and directly by functions that need an objective function
# such as tune_bayes().
#' @rdname choose_metric
#' @export
first_eval_time <- function(
  mtr_set,
  ...,
  metric = NULL,
  eval_time = NULL,
  quietly = FALSE,
  call = rlang::caller_env()
) {
  rlang::check_dots_empty()

  num_times <- length(eval_time)

  if (is.null(metric)) {
    mtr_info <- first_metric(mtr_set)
    metric <- mtr_info$metric
  } else {
    mtr_info <- tibble::as_tibble(mtr_set)
    mtr_info <- mtr_info[mtr_info$metric == metric, ]
  }

  # Not a survival metric
  if (!contains_survival_metric(mtr_info)) {
    return(NULL)
  }

  # Not a metric that requires an eval_time
  no_time_req <- c(
    "static_survival_metric",
    "integrated_survival_metric",
    "linear_pred_survival_metric"
  )
  if (mtr_info$class %in% no_time_req) {
    return(NULL)
  }

  # checks for dynamic metrics
  if (num_times == 0) {
    cli::cli_abort(
      "A single evaluation time is required to use this metric.",
      call = call
    )
  } else if (num_times > 1) {
    eval_time <- eval_time[1]
    print_time <- format(eval_time, digits = 5)
    if (!quietly) {
      cli::cli_warn(
        "{.val {num_times}} evaluation times are available; the first
        will be used (i.e. {.code eval_time = {print_time}}).",
        call = call
      )
    }
  }

  eval_time
}

# ------------------------------------------------------------------------------

#' @rdname choose_metric
#' @export
.filter_perf_metrics <- function(x, metric, eval_time) {
  summary_res <- estimate_tune_results(x)
  summary_res <- summary_res[summary_res$.metric == metric, ]
  is_missing_mean <- is.na(summary_res$mean)
  summary_res <- summary_res[!is_missing_mean, ]

  if (!is.null(eval_time) && any(colnames(summary_res) == ".eval_time")) {
    summary_res <- summary_res[summary_res$.eval_time == eval_time, ]
  }
  if (nrow(summary_res) == 0) {
    cli::cli_abort(
      "No results are available. Please use {.fun collect_metrics}
                    to see if there were any issues."
    )
  }

  summary_res
}

# ------------------------------------------------------------------------------

#' @rdname choose_metric
#' @export
check_metrics_arg <- function(mtr_set, wflow, ..., call = rlang::caller_env()) {
  rlang::check_dots_empty(call = call)
  mode <- extract_spec_parsnip(wflow)$mode

  if (is.null(mtr_set)) {
    switch(
      mode,
      regression = {
        mtr_set <- yardstick::metric_set(rmse, rsq)
      },
      classification = {
        mtr_set <- yardstick::metric_set(roc_auc, accuracy, brier_class)
      },
      'censored regression' = {
        mtr_set <- yardstick::metric_set(brier_survival)
      },
      'quantile regression' = {
        mtr_set <- yardstick::metric_set(weighted_interval_score)
      },
      # workflows cannot be set with an unknown mode
      cli::cli_abort("Model value {.val {mode}} can't be used.", call = call)
    )

    return(mtr_set)
  }

  is_numeric_metric_set <- inherits(mtr_set, "numeric_metric_set")
  is_class_prob_metric_set <- inherits(mtr_set, "class_prob_metric_set")
  is_surv_metric_set <- inherits(mtr_set, c("survival_metric_set"))

  if (
    !is_numeric_metric_set && !is_class_prob_metric_set && !is_surv_metric_set
  ) {
    cli::cli_abort(
      "The {.arg metrics} argument should be the results of
                   {.fn yardstick::metric_set}.",
      call = call
    )
  }

  if (mode == "regression" && !is_numeric_metric_set) {
    cli::cli_abort(
      "The parsnip model has {.code mode} value of {.val {mode}},
                   but the {.arg metrics} is a metric set for a
                   different model mode.",
      call = call
    )
  }

  if (mode == "classification" && !is_class_prob_metric_set) {
    cli::cli_abort(
      "The parsnip model has {.code mode} value of {.val {mode}},
                   but the {.arg metrics} is a metric set for a
                   different model mode.",
      call = call
    )
  }

  if (mode == "censored regression" && !is_surv_metric_set) {
    cli::cli_abort(
      "The parsnip model has {.code mode} value of {.val {mode}},
                   but the {.arg metrics} is a metric set for a
                   different model mode.",
      call = call
    )
  }

  mtr_set
}

# ------------------------------------------------------------------------------

#' @rdname choose_metric
#' @export
check_eval_time_arg <- function(
  eval_time,
  mtr_set,
  ...,
  call = rlang::caller_env()
) {
  rlang::check_dots_empty(call = call)
  mtr_info <- tibble::as_tibble(mtr_set)

  # Not a survival metric
  if (!contains_survival_metric(mtr_info)) {
    if (!is.null(eval_time)) {
      cli::cli_warn(
        "{.arg eval_time} is only used for models with mode {.val censored regression}.",
        call = call
      )
    }
    return(NULL)
  }

  cls <- mtr_info$class
  uni_cls <- sort(unique(cls))
  eval_time <- .filter_eval_time(eval_time)

  num_times <- length(eval_time)
  max_times_req <- req_eval_times(mtr_set)

  # Are there at least a minimal number of evaluation times?
  check_enough_eval_times(eval_time, mtr_set)

  if (max_times_req == 0 & num_times > 0) {
    cli::cli_warn(
      "{.arg eval_time} is only used for dynamic or integrated survival metrics.",
      call = call
    )
    eval_time <- NULL
  }

  eval_time
}

req_eval_times <- function(mtr_set) {
  mtr_info <- tibble::as_tibble(mtr_set)
  cls <- mtr_info$class

  # Default for non-survival and static metrics
  max_req_times <- 0

  if (any(cls == "dynamic_survival_metric")) {
    max_req_times <- max(max_req_times, 1)
  }

  if (any(cls == "integrated_survival_metric")) {
    max_req_times <- max(max_req_times, 2)
  }

  max_req_times
}

# ------------------------------------------------------------------------------
# functions for autoplot

check_autoplot_metrics <- function(x, metric, call) {
  all_met <- .get_tune_metrics(x)
  all_info <- tibble::as_tibble(all_met)
  if (is.null(metric)) {
    metric <- all_info$metric
  }
  check_metric_in_tune_results(all_info, metric, call = call)
  metric
}

check_autoplot_eval_times <- function(x, metric, eval_time, call) {
  if (is.null(eval_time)) {
    if (!any(grepl("survival", metric))) {
      return(eval_time)
    }
    eval_time <- .get_tune_eval_times(x)[1]
  } else {
    any_dyn <- any(purrr::map_lgl(metric, \(.x) {
      is_dyn(.get_tune_metrics(x), .x)
    }))
    if (!any_dyn) {
      cli::cli_warn(
        "{.arg eval_time} is only used for dynamic survival metrics.",
        call = call
      )
      eval_time <- NULL
    }
    check_eval_time_in_tune_results(x, eval_time, call)
  }

  # But there could be NA eval times for non-dynamic metrics
  met <- estimate_tune_results(x) |> dplyr::filter(.metric %in% metric)

  if (any(names(met) == ".eval_time")) {
    if (any(is.na(met$.eval_time))) {
      eval_time <- unique(c(NA, eval_time))
    }
  }
  eval_time
}

check_singular_metric <- function(x, call) {
  if (all(vctrs::vec_count(x$.metric)$count == 1)) {
    cli::cli_abort(
      "Only one observation per metric was present. \\
      Unable to create meaningful plot.",
      call = call
    )
  }
  invisible(NULL)
}

library(yardstick)

test_that("selecting single eval time - non-survival case", {
  met_reg <- metric_set(rmse)

  times_1 <- 1 / 3
  times_2 <- as.numeric(5:4) / 7

  # eval time is not applicable outside of survival models; return null

  expect_null(first_eval_time(met_reg, eval_time = NULL))
  expect_null(first_eval_time(met_reg, eval_time = times_1))
  expect_null(first_eval_time(met_reg, eval_time = times_2))
})

test_that("selecting single eval time - pure metric sets", {
  met_int <- metric_set(brier_survival_integrated)
  met_dyn <- metric_set(brier_survival)
  met_stc <- metric_set(concordance_survival)

  times_1 <- 1 / 3
  times_2 <- as.numeric(5:4) / 7

  # all static; return NULL and add warning if times are given

  expect_null(first_eval_time(met_stc, eval_time = NULL))
  expect_null(first_eval_time(
    met_stc,
    metric = "concordance_survival",
    eval_time = NULL
  ))

  expect_silent(
    stc_one <- first_eval_time(met_stc, eval_time = times_1)
  )
  expect_null(stc_one)

  expect_silent(
    stc_multi <- first_eval_time(met_stc, eval_time = times_2)
  )
  expect_null(stc_multi)

  # all dynamic; return a single time and warn if there are more and error if
  # there are none

  expect_snapshot(
    first_eval_time(met_dyn, eval_time = NULL),
    error = TRUE
  )
  expect_snapshot(
    first_eval_time(met_dyn, metric = "brier_survival", eval_time = NULL),
    error = TRUE
  )

  expect_equal(
    first_eval_time(met_dyn, eval_time = times_1),
    times_1
  )

  expect_snapshot(
    dyn_multi <- first_eval_time(met_dyn, eval_time = times_2)
  )
  expect_equal(dyn_multi, times_2[1])

  # all integrated; return NULL and warn if there 1+ times

  expect_null(first_eval_time(met_int, eval_time = NULL))
  expect_null(
    first_eval_time(
      met_int,
      metric = "brier_survival_integrated",
      eval_time = NULL
    )
  )

  expect_silent(
    int_1 <- first_eval_time(met_int, eval_time = times_1)
  )
  expect_null(int_1)

  expect_silent(
    int_multi <- first_eval_time(met_int, eval_time = times_2)
  )
  expect_null(int_multi)
})

test_that("selecting single eval time - mixed metric sets - static first", {
  met_mix_stc <- metric_set(concordance_survival, brier_survival)
  met_mix_stc_all <- metric_set(
    concordance_survival,
    brier_survival,
    brier_survival_integrated
  )

  times_1 <- 1 / 3
  times_2 <- as.numeric(5:4) / 7

  # static is first but includes dynamic. Should return NULL and add warning
  # if times are given

  expect_null(
    first_eval_time(met_mix_stc, eval_time = NULL)
  )

  expect_silent(
    stc_1 <- first_eval_time(met_mix_stc, eval_time = times_1)
  )
  expect_null(stc_1)

  expect_silent(
    stc_multi <- first_eval_time(met_mix_stc, eval_time = times_2)
  )
  expect_null(stc_multi)

  # static is first but includes dynamic and integrated. Should return NULL and
  # add warning if times are given

  expect_null(
    first_eval_time(met_mix_stc_all, eval_time = NULL)
  )

  expect_silent(
    stc_1 <- first_eval_time(met_mix_stc_all, eval_time = times_1)
  )
  expect_null(stc_1)

  expect_silent(
    stc_multi <- first_eval_time(met_mix_stc_all, eval_time = times_2)
  )
  expect_null(stc_multi)
})

test_that("selecting single eval time - mixed metric sets - dynamic first", {
  met_mix_dyn <- metric_set(brier_survival, concordance_survival)
  met_mix_dyn_all <-
    metric_set(brier_survival, brier_survival_integrated, concordance_survival)

  times_1 <- 1 / 3
  times_2 <- as.numeric(5:4) / 7

  # dynamic is first but includes static. Should return single time and add warning
  # if 2+ times are given

  expect_snapshot(
    first_eval_time(met_mix_dyn, eval_time = NULL),
    error = TRUE
  )
  expect_equal(
    first_eval_time(met_mix_dyn, eval_time = times_1),
    times_1
  )
  expect_snapshot(
    dyn_multi <- first_eval_time(met_mix_dyn, eval_time = times_2)
  )
  expect_equal(dyn_multi, times_2[1])

  # dynamic is first but includes static and integrated. Should return single
  # time and add warning if 2+ times are given

  expect_snapshot(
    first_eval_time(met_mix_dyn_all, eval_time = NULL),
    error = TRUE
  )
  expect_equal(
    first_eval_time(met_mix_dyn_all, eval_time = times_1),
    times_1
  )
  expect_snapshot(
    dyn_multi <- first_eval_time(met_mix_dyn_all, eval_time = times_2)
  )
  expect_equal(dyn_multi, times_2[1])
})


test_that("selecting single eval time - mixed metric sets - integrated first", {
  met_mix_int <- metric_set(brier_survival_integrated, concordance_survival)
  met_mix_int_all <-
    metric_set(brier_survival_integrated, brier_survival, concordance_survival)

  times_1 <- 1 / 3
  times_2 <- as.numeric(5:4) / 7

  # integrated is first but includes static. Should return NULL and add warning
  # if 1+ times are given

  expect_null(first_eval_time(met_mix_int, eval_time = NULL))

  expect_silent(
    first_eval_time(met_mix_int, eval_time = times_1)
  )
  expect_silent(
    int_multi <- first_eval_time(met_mix_int, eval_time = times_2)
  )
  expect_null(int_multi)

  # integrated is first but includes static and dynamic. Should return NULL and
  # add warning if 1+ times are given

  expect_null(first_eval_time(met_mix_int_all, eval_time = NULL))

  expect_silent(
    first_eval_time(met_mix_int_all, eval_time = times_1)
  )
  expect_silent(
    int_multi <- first_eval_time(met_mix_int_all, eval_time = times_2)
  )
  expect_null(int_multi)
})


test_that("selecting an evaluation time", {
  # much of this is indirectly tested in show/select best

  surv_res <- readRDS(test_path("data", "surv_boost_tree_res.rds"))

  expect_snapshot(
    choose_eval_time(surv_res, "brier_survival")
  )
  expect_snapshot(
    choose_eval_time(surv_res, "concordance_survival")
  )
  expect_snapshot(
    choose_eval_time(surv_res, "concordance_survival", eval_time = 10)
  )

  data("example_ames_knn")
  expect_snapshot(choose_eval_time(ames_grid_search, "rmse", eval_time = 1))
})

test_that("selecting the first metric", {
  library(yardstick)

  met_1 <- metric_set(rmse)
  tbl_1 <- tibble::as_tibble(met_1)[1, ]
  met_2 <- metric_set(rmse, ccc)
  tbl_2 <- tibble::as_tibble(met_2)[1, ]

  expect_equal(first_metric(met_1), tbl_1)
  expect_equal(first_metric(met_2), tbl_2)
})

test_that("selecting a metric", {
  # much of this is indirectly tested in show/select best

  data("example_ames_knn")
  expect_snapshot(
    choose_metric(ames_grid_search, "rmse")
  )
  expect_snapshot(
    choose_metric(ames_grid_search, NULL)
  )
  expect_snapshot(
    choose_metric(ames_grid_search, "potato"),
    error = TRUE
  )
  expect_snapshot(
    choose_metric(ames_grid_search, c("rmse", "ccc"))
  )
})

test_that("identify survival metrics", {
  library(yardstick)

  expect_false(
    metric_set(rmse) |>
      tibble::as_tibble() |>
      tune:::contains_survival_metric()
  )

  expect_true(
    metric_set(brier_survival_integrated) |>
      tibble::as_tibble() |>
      tune:::contains_survival_metric()
  )

  expect_true(
    metric_set(brier_survival, concordance_survival) |>
      tibble::as_tibble() |>
      tune:::contains_survival_metric()
  )
})

CoordObsPred <-
  ggplot2::ggproto(
    "CoordObsPred",
    ggplot2::CoordFixed,
    setup_panel_params = function(self, scale_x, scale_y, params = list()) {
      # coord limits take precedence over scale limits
      rngs <- range(
        self$limits$x %||% scale_x$get_limits(),
        self$limits$y %||% scale_y$get_limits(),
        na.rm = TRUE
      )
      self$limits$y <- rngs
      self$limits$x <- rngs
      ggplot2::ggproto_parent(ggplot2::CoordFixed, self)$setup_panel_params(
        scale_x,
        scale_y,
        params
      )
    },
    aspect = function(self, ranges) {
      1 / self$ratio
    }
  )

#' Use same scale for plots of observed vs predicted values
#'
#' For regression models, `coord_obs_pred()` can be used in a ggplot to make the
#' x- and y-axes have the same exact scale along with an aspect ratio of one.
#' @param ratio	Aspect ratio, expressed as `y / x`. Defaults to 1.0.
#' @param xlim,ylim Limits for the x and y axes.
#' @param expand Not currently used.
#' @param clip Should drawing be clipped to the extent of the plot panel? A setting
#' of "on" (the default) means yes, and a setting of "off" means no. In most
#' cases, the default of "on" should not be changed, as setting `clip = "off"`
#' can cause unexpected results. It allows drawing of data points anywhere on
#' the plot, including in the plot margins. If limits are set via `xlim` and
#' `ylim` and some data points fall outside those limits, then those data points
#' may show up in places such as the axes, the legend, the plot title, or the
#' plot margins.
#' @return A `ggproto` object.
#' @examplesIf rlang::is_installed("modeldata")
#' # example code
#' data(solubility_test, package = "modeldata")
#'
#' library(ggplot2)
#' p <- ggplot(solubility_test, aes(x = solubility, y = prediction)) +
#'   geom_abline(lty = 2) +
#'   geom_point(alpha = 0.5)
#'
#' p
#'
#' p + coord_fixed()
#'
#' p + coord_obs_pred()
#' @export
coord_obs_pred <-
  function(ratio = 1, xlim = NULL, ylim = NULL, expand = TRUE, clip = "on") {
    ggplot2::ggproto(
      NULL,
      CoordObsPred,
      limits = list(x = xlim, y = ylim),
      ratio = ratio,
      expand = expand,
      clip = clip
    )
  }

# ------------------------------------------------------------------------------

#' @export
`[.resample_results` <- function(x, i, j, ...) {
  out <- NextMethod()
  resample_results_reconstruct(out, x)
}

#' @export
`names<-.resample_results` <- function(x, value) {
  out <- NextMethod()
  resample_results_reconstruct(out, x)
}

# ------------------------------------------------------------------------------

new_resample_results <-
  function(
    x,
    parameters,
    metrics,
    eval_time,
    eval_time_target = NULL,
    outcomes = character(0),
    rset_info,
    workflow = NULL
  ) {
    new_tune_results(
      x = x,
      parameters = parameters,
      metrics = metrics,
      eval_time = eval_time,
      eval_time_target = eval_time_target,
      outcomes = outcomes,
      rset_info = rset_info,
      workflow = workflow,
      class = "resample_results"
    )
  }

#' Fit a model to the numerically optimal configuration
#'
#' `fit_best()` takes the results from model tuning and fits it to the training
#' set using tuning parameters associated with the best performance.
#'
#' @param x The results of class `tune_results` (coming from functions such as
#' [tune_grid()], [tune_bayes()], etc). The control option
#' [`save_workflow = TRUE`][tune::control_grid] should have been used.
#' @param ... Not currently used, must be empty.
#' @param metric A character string (or `NULL`) for which metric to optimize. If
#' `NULL`, the first metric is used.
#' @param parameters An optional 1-row tibble of tuning parameter settings, with
#' a column for each tuning parameter. This tibble should have columns for each
#' tuning parameter identifier (e.g. `"my_param"` if `tune("my_param")` was used).
#' If `NULL`, this argument will be set to
#' [`select_best(metric, eval_time)`][tune::select_best.tune_results].
#' If not `NULL`, `parameters` overwrites the specification via `metric`, and
#' `eval_time`.
#' @param verbose A logical for printing logging.
#' @param add_validation_set When the resamples embedded in `x` are a split into
#' training set and validation set, should the validation set be included in the
#' data set used to train the model? If not, only the training set is used. If
#' `NULL`, the validation set is not used for resamples originating from
#' [rsample::validation_set()] while it is used for resamples originating
#' from [rsample::validation_split()].
#' @inheritParams select_best
#' @details
#' This function is a shortcut for the manual steps of:
#'
#' \preformatted{
#'   best_param <- select_best(tune_results, metric) # or other `select_*()`
#'   wflow <- finalize_workflow(wflow, best_param)  # or just `finalize_model()`
#'   wflow_fit <- fit(wflow, data_set)
#' }
#'
#' @template case-weights
#'
#' @inheritSection last_fit See also
#'
#' @examplesIf tune:::should_run_examples() && rlang::is_installed("modeldata") && !tune:::is_cran_check()
#' library(recipes)
#' library(rsample)
#' library(parsnip)
#' library(dplyr)
#'
#' data(meats, package = "modeldata")
#' meats <- meats |> select(-water, -fat)
#'
#' set.seed(1)
#' meat_split <- initial_split(meats)
#' meat_train <- training(meat_split)
#' meat_test  <- testing(meat_split)
#'
#' set.seed(2)
#' meat_rs <- vfold_cv(meat_train, v = 10)
#'
#' pca_rec <-
#'   recipe(protein ~ ., data = meat_train) |>
#'   step_normalize(all_numeric_predictors()) |>
#'   step_pca(all_numeric_predictors(), num_comp = tune())
#'
#' knn_mod <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")
#'
#' ctrl <- control_grid(save_workflow = TRUE)
#'
#' set.seed(128)
#' knn_pca_res <-
#'   tune_grid(knn_mod, pca_rec, resamples = meat_rs, grid = 10, control = ctrl)
#'
#' knn_fit <- fit_best(knn_pca_res, verbose = TRUE)
#' predict(knn_fit, meat_test)
#' @return A fitted workflow.
#' @export
fit_best <- function(x, ...) {
  UseMethod("fit_best")
}

#' @export
#' @rdname fit_best
fit_best.default <- function(x, ...) {
  cli::cli_abort(
    "No {.fn fit_best} exists for {.obj_type_friendly {x}}."
  )
}

#' @export
#' @rdname fit_best
fit_best.tune_results <- function(
  x,
  ...,
  metric = NULL,
  eval_time = NULL,
  parameters = NULL,
  verbose = FALSE,
  add_validation_set = NULL
) {
  rlang::check_dots_empty()
  wflow <- .get_tune_workflow(x)
  if (is.null(wflow)) {
    cli::cli_abort(c(
      "x" = "The control option `save_workflow = TRUE` should be used when tuning."
    ))
  }

  if (is.null(parameters)) {
    met_set <- .get_tune_metrics(x)

    if (is.null(metric)) {
      metric <- .get_tune_metric_names(x)[1]
    } else {
      check_metric_in_tune_results(tibble::as_tibble(met_set), metric)
    }

    if (is.null(eval_time) & is_dyn(met_set, metric)) {
      eval_time <- .get_tune_eval_times(x)[1]
    }

    parameters <- select_best(x, metric = metric, eval_time = eval_time)
    if (verbose) {
      format_final_param(parameters, metric)
    }
  } else {
    if (!is.null(metric)) {
      cli::cli_warn(
        "{.arg metric} is being ignored because {.arg parameters}
      has been specified."
      )
    }
    if (!is.null(eval_time)) {
      cli::cli_warn(
        "{.arg eval_time} is being ignored because {.arg parameters}
      has been specified."
      )
    }
  }

  # ----------------------------------------------------------------------------

  wflow <- finalize_workflow(wflow, parameters = parameters)

  # ----------------------------------------------------------------------------

  pset <- hardhat::extract_parameter_set_dials(wflow)
  p <- nrow(pset)
  if (p > 0) {
    cli::cli_abort(
      "The {cli::qty(pset$id)} parameter{?s} {.var {pset$id}} \\
       {?is/are} still marked for tuning."
    )
  }

  # ----------------------------------------------------------------------------

  if (inherits(x$splits[[1]], "val_split")) {
    if (is.null(add_validation_set)) {
      rset_info <- attr(x, "rset_info")
      originate_from_3way_split <- rset_info$att$origin_3way %||% FALSE
      if (originate_from_3way_split) {
        add_validation_set <- FALSE
      } else {
        add_validation_set <- TRUE
      }
    }
    if (add_validation_set) {
      dat <- x$splits[[1]]$data
    } else {
      dat <- rsample::training(x$splits[[1]])
    }
  } else {
    if (!is.null(add_validation_set)) {
      cli::cli_warn(
        "The option {.arg add_validation_set} is being ignored because the
         resampling object does not include a validation set."
      )
    }
    dat <- x$splits[[1]]$data
  }
  if (verbose) {
    cli::cli_inform(c("i" = "Fitting using {nrow(dat)} data points..."))
  }
  res <- parsnip::fit(wflow, dat)
  if (verbose) {
    cli::cli_inform(c("v" = "Done.", " "))
  }
  res
}

format_final_param <- function(x, metric) {
  x <- dplyr::select(x, -dplyr::any_of(".config"))
  if (ncol(x) == 0) {
    return(invisible(x))
  }
  lst <- as.list(x)
  chr_lst <- format(lst)
  nms_lst <- format(paste0(names(chr_lst), ": "))
  print_lst <- paste0("  ", nms_lst, chr_lst, collapse = "\n")
  prm <- ifelse(length(lst) == 1, "parameter was:\n", "parameters were:\n")
  cat("Using", metric, "as the metric, the optimal", prm)
  cat(print_lst, sep = "")
  cat("\n\n")
  invisible(x)
}

library(parsnip)
library(workflows)
library(tune)
library(rsample)
library(recipes)
library(yardstick)
library(dplyr)
library(dials)
library(rlang)
library(tailor)
library(tibble)

# ------------------------------------------------------------------------------

dt_spec <- parsnip::decision_tree(
  mode = "classification",
  min_n = tune(),
  engine = "C5.0"
)

knn_cls_spec <- parsnip::nearest_neighbor(
  mode = "classification",
  neighbors = tune()
)


if (rlang::is_installed("probably")) {
  cls_est_post <- tailor::tailor() |>
    tailor::adjust_probability_calibration(method = "logistic")

  cls_cal_tune_post <- tailor::tailor() |>
    tailor::adjust_probability_calibration(method = "logistic") |>
    tailor::adjust_probability_threshold(threshold = tune("cut"))

  cls_cal <- tailor::tailor() |>
    tailor::adjust_probability_calibration()

  cls_tenth <- tailor::tailor() |>
    tailor::adjust_probability_threshold(threshold = 1 / 10)

  cls_post <- tailor::tailor() |>
    tailor::adjust_probability_threshold(threshold = tune("cut"))
}

fac_2c <- structure(
  integer(0),
  levels = c("Class1", "Class2"),
  class = "factor"
)
cls_two_class_plist <-
  tibble::tibble(
    Class = fac_2c,
    .pred_class = fac_2c,
    .pred_Class1 = double(0),
    .pred_Class2 = double(0),
    .row = integer(0),
  )

sim_2c <- structure(
  integer(0),
  levels = c("class_1", "class_2"),
  class = "factor"
)
cls_sim_plist <-
  tibble::tibble(
    class = sim_2c,
    .pred_class = sim_2c,
    .pred_class_1 = double(0),
    .pred_class_2 = double(0),
    .row = integer(0),
  )

# ------------------------------------------------------------------------------

dt_grid <- tibble::tibble(min_n = c(2, 4))
knn_grid <- tibble::tibble(neighbors = 1:3)
svm_grid <- tibble::tibble(degree = 1:2)

make_post_data <- function(mode = "classification") {
  set.seed(1)
  if (mode == "classification") {
    dat <- modeldata::sim_classification(1000)
    nm <- "class"
  } else if (mode == "regression") {
    dat <- modeldata::sim_regression(1000)
    nm <- "outcome"
  } else if (mode == "censored") {
    require(survival)
    dat <- modeldata::deliveries |>
      dplyr::select(time_to_delivery, starts_with("item"))
    evt <- rep_len(c(rep(1, 9), 0), nrow(dat))
    dat$outcome <- survival::Surv(dat$time_to_delivery, evt)
    dat$time_to_delivery <- NULL
    nm <- "outcome"
  } else {
    cli::abort(
      "Only have modes for classification, regression, and censored regression so far"
    )
  }
  rs <- rsample::mc_cv(dat, times = 2)
  rs_split <- rs$splits[[1]]
  rs_args <- rsample::.get_split_args(rs)
  list(data = dat, rs = rs, split = rs_split, args = rs_args, y = nm)
}

# ------------------------------------------------------------------------------

puromycin <- tibble::as_tibble(Puromycin)
puromycin_rec <- recipes::recipe(rate ~ ., data = puromycin) |>
  recipes::step_dummy(state)

puromycin_tune_rec <- puromycin_rec |>
  recipes::step_poly(conc, degree = tune())

knn_reg_spec <- parsnip::nearest_neighbor(
  mode = "regression",
  neighbors = tune()
)
svm_spec <- parsnip::svm_poly(mode = "regression", cost = 1, degree = tune())

reg_post <- tailor::tailor() |>
  tailor::adjust_predictions_custom(.pred = .pred + 10000)


if (rlang::is_installed("probably")) {
  reg_cal_max <- tailor::tailor() |>
    tailor::adjust_numeric_calibration() |>
    tailor::adjust_numeric_range(upper_limit = tune())

  reg_cal <- tailor::tailor() |>
    tailor::adjust_numeric_calibration()

  reg_cal_tune <- tailor::tailor() |>
    tailor::adjust_numeric_calibration(method = tune())

  reg_max <- tailor::tailor() |>
    tailor::adjust_numeric_range(upper_limit = tune())
}

glmn_spec <- parsnip::linear_reg(penalty = tune(), mixture = tune()) |>
  parsnip::set_engine("glmnet")

reg_sim_plist <- tibble::tibble(
  outcome = double(0),
  .pred = double(0),
  .row = integer(0)
)

puromycin_plist <- tibble::tibble(
  rate = puromycin$rate[0],
  .pred = puromycin$rate[0],
  .row = integer(0)
)

# ------------------------------------------------------------------------------

surv_0 <- structure(
  numeric(0),
  type = "right",
  dim = c(0L, 2L),
  dimnames = list(NULL, c("time", "status")),
  class = "Surv"
)

pred_0 <- tibble::tibble(
  .eval_time = numeric(0),
  .pred_survival = numeric(0)
)

pred_dyn_0 <- tibble::tibble(
  .eval_time = numeric(0),
  .pred_survival = numeric(0),
  .weight_censored = numeric(0)
)

# wrapper for executions that catches warnings and errors if they occur. See
# https://adv-r.hadley.nz/conditions.html
catcher <- function(expr) {
  signals <- list()
  add_cond <- function(cnd) {
    signals <<- append(signals, list(rlang::cnd_entrace(cnd)))
    rlang::cnd_muffle(cnd)
  }

  res <-
    rlang::try_fetch(
      expr,
      warning = add_cond,
      error = function(e) {
        structure(
          catch_message(e),
          class = "try-error",
          # if a simple error, add a traceback.
          # otherwise, pass the condition right along.
          condition = rlang::`%||%`(rlang::cnd_entrace(e), e)
        )
      }
    )

  list(res = res, signals = signals)
}

# A simplified version of the error handler supplied in `try()` source
catch_message <- function(e) {
  paste0(
    "Error in ",
    deparse(conditionCall(e)),
    ": ",
    conditionMessage(e),
    "\n"
  )
}

#' Obtain and format results produced by tuning functions
#'
#' @param x The results of [tune_grid()], [tune_bayes()], [fit_resamples()],
#'  or [last_fit()]. For [collect_predictions()], the control option `save_pred
#'  = TRUE` should have been used.
#' @param ... Not currently used.
#' @param summarize A logical; should metrics be summarized over resamples
#' (`TRUE`) or return the values for each individual resample. Note that, if `x`
#' is created by [last_fit()], `summarize` has no effect. For the other object
#' types, the method of summarizing predictions is detailed below.
#' @param parameters An optional tibble of tuning parameter values that can be
#'  used to filter the predicted values before processing. This tibble should
#'  only have columns for each tuning parameter identifier (e.g. `"my_param"`
#'  if `tune("my_param")` was used).
#' @param type One of `"long"` (the default) or `"wide"`. When `type = "long"`,
#'  output has columns `.metric` and one of `.estimate` or `mean`.
#'  `.estimate`/`mean` gives the values for the `.metric`. When `type = "wide"`,
#'  each metric has its own column and the `n` and `std_err` columns are removed,
#'  if they exist.
#'
#' @return A tibble. The column names depend on the results and the mode of the
#' model.
#'
#' For [collect_metrics()] and [collect_predictions()], when unsummarized,
#' there are columns for each tuning parameter (using the `id` from [tune()],
#' if any).
#'
#' [collect_metrics()] also has columns `.metric`, and `.estimator` by default.
#' For [collect_metrics()] methods that have a `type` argument, supplying
#' `type = "wide"` will pivot the output such that each metric has its own
#' column. When the results are summarized, there are columns for `mean`, `n`,
#' and `std_err`. When not summarized, the additional columns for the resampling
#' identifier(s) and `.estimate`.
#'
#' For [collect_predictions()], there are additional columns for the resampling
#' identifier(s), columns for the predicted values (e.g., `.pred`,
#' `.pred_class`, etc.), and a column for the outcome(s) using the original
#' column name(s) in the data.
#'
#' [collect_predictions()] can summarize the various results over
#'  replicate out-of-sample predictions. For example, when using the bootstrap,
#'  each row in the original training set has multiple holdout predictions
#'  (across assessment sets). To convert these results to a format where every
#'  training set same has a single predicted value, the results are averaged
#'  over replicate predictions.
#'
#' For regression cases, the numeric predictions are simply averaged.
#'
#' For classification models, the problem is more complex. When class probabilities
#'  are used, these are averaged and then re-normalized to make sure that they
#'  add to one. If hard class predictions also exist in the data, then these are
#'  determined from the summarized probability estimates (so that they match).
#'  If only hard class predictions are in the results, then the mode is used to
#'  summarize.
#'
#' With censored outcome models, the predicted survival probabilities (if any)
#'  are averaged while the static predicted event times are summarized using the
#'  median.
#'
#' [collect_notes()] returns a tibble with columns for the resampling
#' indicators, the location (preprocessor, model, etc.), type (error or warning),
#' and the notes.
#'
#' [collect_extracts()] collects objects extracted from fitted workflows
#' via the `extract` argument to [control functions][control_grid()]. The
#' function returns a tibble with columns for the resampling
#' indicators, the location (preprocessor, model, etc.), and extracted objects.
#'
#' @section Hyperparameters and extracted objects:
#'
#' When making use of submodels, tune can generate predictions and calculate
#' metrics for multiple model `.config`urations using only one model fit.
#' However, this means that if a function was supplied to a
#' [control function's][control_grid()] `extract` argument, tune can only
#' execute that extraction on the one model that was fitted. As a result,
#' in the `collect_extracts()` output, tune opts to associate the
#' extracted objects with the hyperparameter combination used to
#' fit that one model workflow, rather than the hyperparameter
#' combination of a submodel. In the output, this appears like
#' a hyperparameter entry is recycled across many `.config`
#' entries---this is intentional.
#'
#' See \url{https://parsnip.tidymodels.org/articles/Submodels.html} to learn
#' more about submodels.
#'
#' @examplesIf tune:::should_run_examples(suggests = c("kknn", "splines2"))
#' data("example_ames_knn")
#' # The parameters for the model:
#' extract_parameter_set_dials(ames_wflow)
#'
#' # Summarized over resamples
#' collect_metrics(ames_grid_search)
#'
#' # Per-resample values
#' collect_metrics(ames_grid_search, summarize = FALSE)
#'
#'
#' # ---------------------------------------------------------------------------
#'
#' library(parsnip)
#' library(rsample)
#' library(dplyr)
#' library(recipes)
#' library(tibble)
#'
#' lm_mod <- linear_reg() |> set_engine("lm")
#' set.seed(93599150)
#' car_folds <- vfold_cv(mtcars, v = 2, repeats = 3)
#' ctrl <- control_resamples(save_pred = TRUE, extract = extract_fit_engine)
#'
#' spline_rec <-
#'   recipe(mpg ~ ., data = mtcars) |>
#'   step_spline_natural(disp, deg_free = tune("df"))
#'
#' grid <- tibble(df = 3:6)
#'
#' resampled <-
#'   lm_mod |>
#'   tune_grid(spline_rec, resamples = car_folds, control = ctrl, grid = grid)
#'
#' collect_predictions(resampled) |> arrange(.row)
#' collect_predictions(resampled, summarize = TRUE) |> arrange(.row)
#' collect_predictions(
#'   resampled,
#'   summarize = TRUE,
#'   parameters = grid[1, ]
#' ) |> arrange(.row)
#'
#' collect_extracts(resampled)
#'
#' @export
collect_predictions <- function(x, ...) {
  UseMethod("collect_predictions")
}

#' @export
#' @rdname collect_predictions
collect_predictions.default <- function(x, ...) {
  cli::cli_abort(
    "No {.fn collect_predictions} exists for {.obj_type_friendly {x}}."
  )
}

#' @export
#' @rdname collect_predictions
collect_predictions.tune_results <- function(
  x,
  ...,
  summarize = FALSE,
  parameters = NULL
) {
  rlang::check_dots_empty()

  names <- colnames(x)
  coll_col <- ".predictions"

  has_coll_col <- coll_col %in% names

  if (!has_coll_col) {
    cli::cli_abort(
      "The {.field .predictions} column does not exist. Please refit with the
      control argument {.code save_pred = TRUE} to save predictions."
    )
  }

  x <- filter_predictions(x, parameters)

  if (summarize) {
    x <- average_predictions(x, parameters)
  } else {
    x <- collector(x, coll_col = coll_col)
  }

  param_names <- .get_tune_parameter_names(x)
  y_name <- .get_tune_outcome_names(x)
  reorder_pred_cols(x, y_name, param_names)
}

filter_predictions <- function(x, parameters) {
  if (is.null(parameters)) {
    return(x)
  }
  params <- attr(x, "parameters")
  if (is.null(params)) {
    cli::cli_warn(
      "The object is missing some attributes; it is probably from an earlier
       version of {.pkg tune}. The predictions can't be filtered."
    )

    return(x)
  }

  param_names <- params$id
  parameters <- dplyr::select(parameters, dplyr::any_of(param_names))
  if (ncol(parameters) != length(param_names)) {
    cli::cli_abort(
      "{.arg parameters} should only have columns:
                    {.val {param_names}}."
    )
  }
  if (length(param_names) == 0) {
    x$.predictions <-
      purrr::map(x$.predictions, dplyr::cross_join, parameters)
  } else {
    x$.predictions <-
      purrr::map(
        x$.predictions,
        dplyr::inner_join,
        parameters,
        by = param_names
      )
  }
  x
}

# Functions for summarizing the various results over replicate out-of-sample
#  predictions. For regression cases, the numeric predictions are simply
#  averaged. For classification models, the problem is more complex. When class
#  probabilities are used, these are averaged and then re-normalized to make
#  sure that they add to one. If hard class predictions also exist in the data,
#  then these are determined from the summarized probability estimates (so that
#  they match). If only hard class predictions are in the results, then the
#  mode is used to summarize.

numeric_summarize <- function(x) {
  pred_cols <- grep("^\\.pred", names(x), value = TRUE)
  nms <- names(x)
  group_cols <- nms[!(nms %in% pred_cols)]
  x <-
    x |>
    dplyr::group_by(!!!rlang::syms(group_cols)) |>
    dplyr::summarise(
      dplyr::across(dplyr::starts_with(".pred"), mean_na_rm)
    )

  x
}

prob_summarize <- function(x, p) {
  pred_cols <- grep("^\\.pred", names(x), value = TRUE)

  if (any(names(x) == ".pred_class")) {
    compute_class <- TRUE
    x$.pred_class <- NULL
    pred_cols <- pred_cols[pred_cols != ".pred_class"]
  } else {
    compute_class <- FALSE
  }

  nms <- names(x)
  y_cols <- nms[
    !(nms %in% c(".row", ".iter", ".config", ".eval_time", pred_cols, p))
  ]
  group_cols <- nms[!(nms %in% pred_cols)]

  x <-
    x |>
    dplyr::group_by(!!!rlang::syms(group_cols)) |>
    dplyr::summarise(
      dplyr::across(dplyr::starts_with(".pred_"), mean_na_rm)
    ) |>
    ungroup()

  # In case the class probabilities do not add up to 1 after averaging
  group_cols <- group_cols[group_cols != y_cols]
  totals <-
    x |>
    dplyr::select(-dplyr::all_of(y_cols)) |>
    tidyr::pivot_longer(
      cols = c(dplyr::all_of(pred_cols)),
      names_to = ".column",
      values_to = ".value"
    ) |>
    dplyr::group_by(!!!rlang::syms(group_cols)) |>
    dplyr::summarize(.totals = sum(.value)) |>
    dplyr::ungroup()

  x <-
    x |>
    dplyr::full_join(totals, by = group_cols) |>
    dplyr::mutate(
      dplyr::across(dplyr::starts_with(".pred_"), ~ .x / .totals)
    ) |>
    dplyr::select(-.totals)

  # If we started with hard class predictions, recompute them based on the
  # newly averaged probability estimates.
  if (compute_class) {
    lvl <- levels(x[[y_cols]])
    ord <- is.ordered(x[[y_cols]])
    class_pred <-
      x |>
      dplyr::select(-dplyr::all_of(y_cols)) |>
      tidyr::pivot_longer(
        cols = c(dplyr::all_of(pred_cols)),
        names_to = ".column",
        values_to = ".value"
      ) |>
      dplyr::group_by(!!!rlang::syms(group_cols)) |>
      dplyr::arrange(dplyr::desc(.value), .by_group = TRUE) |>
      dplyr::slice(1) |>
      dplyr::mutate(
        .pred_class = gsub("\\.pred_", "", .column),
        .pred_class = factor(.pred_class, levels = lvl, ordered = ord)
      ) |>
      dplyr::ungroup() |>
      dplyr::select(-.value, -.column)
    x <- full_join(x, class_pred, by = group_cols)
  }
  x
}

# define a `mean()` wrapper to avoid slowdown in
# evaluation of anonymous function
mean_na_rm <- function(x) {
  mean(x, na.rm = TRUE)
}

class_summarize <- function(x, p) {
  pred_cols <- grep("^\\.pred", names(x), value = TRUE)
  nms <- names(x)
  group_cols <- nms[!(nms %in% pred_cols)]
  outcome_col <- group_cols[!(group_cols %in% c(p, ".row", ".iter", ".config"))]
  x <-
    x |>
    dplyr::group_by(!!!rlang::syms(group_cols)) |>
    dplyr::count(.pred_class) |>
    dplyr::arrange(dplyr::desc(n), .by_group = TRUE) |>
    dplyr::slice(1) |>
    dplyr::ungroup() |>
    dplyr::select(-n)
  x
}

surv_summarize <- function(x, param, y) {
  pred_cols <- grep("^\\.pred", names(x), value = TRUE)
  nms <- names(x)

  outcomes <- x[, c(".row", y)] |> dplyr::slice(1, .by = .row)

  res <- NULL

  # Use the median to summarize the static estimates
  if (any(pred_cols == ".pred_time")) {
    res <-
      dplyr::summarize(
        x,
        .pred_time = median(.pred_time),
        .by = c(.row, .config, dplyr::any_of(param), dplyr::any_of(".iter"))
      )
  }

  # Simple mean to summarize dynamic probability predictions
  if (any(pred_cols == ".pred")) {
    nest_cols <- c(".eval_time", ".pred_survival", ".weight_censored")
    tmp <-
      x |>
      dplyr::select(
        .pred,
        .config,
        .row,
        dplyr::any_of(param),
        dplyr::any_of(".iter")
      ) |>
      tidyr::unnest(.pred) |>
      dplyr::summarize(
        .pred_survival = mean(.pred_survival, na.rm = TRUE),
        .weight_censored = mean(.weight_censored, na.rm = TRUE),
        .by = c(
          .row,
          .eval_time,
          .config,
          dplyr::any_of(param),
          dplyr::any_of(".iter")
        )
      ) |>
      tidyr::nest(
        .pred = c(dplyr::all_of(nest_cols)),
        .by = c(.row, .config, dplyr::any_of(param), dplyr::any_of(".iter"))
      )

    if (!is.null(res)) {
      dot_iter <- grep("\\.iter", nms, value = TRUE)
      res <-
        dplyr::full_join(tmp, res, by = c(".row", ".config", param, dot_iter))
    } else {
      res <- tmp
    }
  }

  # Simple mean to summarize linear predictors
  if (any(pred_cols == ".pred_linear_pred")) {
    tmp <-
      dplyr::summarize(
        x,
        .pred_linear_pred = mean(.pred_linear_pred, na.rm = TRUE),
        .by = c(.row, .config, dplyr::any_of(param), dplyr::any_of(".iter"))
      )

    if (!is.null(res)) {
      dot_iter <- grep("\\.iter", nms, value = TRUE)
      res <-
        dplyr::full_join(tmp, res, by = c(".row", ".config", param, dot_iter))
    } else {
      res <- tmp
    }
  }

  res <- dplyr::full_join(outcomes, res, by = ".row")
  res[order(res$.row, res$.config), nms]
}

quantile_summarize <- function(x) {
  lvl_key <-
    tibble::tibble(
      .quantile_levels = hardhat::extract_quantile_levels(x$.pred_quantile)
    ) |>
    dplyr::arrange(.quantile_levels) |>
    dplyr::mutate(
      .index = gsub(" ", "0", format(dplyr::row_number())),
      .index = paste0(".qp_", .index)
    )
  nms <- names(x)
  group_cols <- nms[!(nms %in% ".pred_quantile")]
  group_cols <- group_cols[!grepl("^id", group_cols)]
  all_group_cols <- c(".quantile_levels", group_cols)
  tmp <-
    x |>
    dplyr::mutate(
      .pred_quantile = purrr::map(.pred_quantile, tibble::as_tibble)
    ) |>
    # There are .row columns in .pred_quantile and in the tibble. We want the
    # tibble column and, to avoid a name conflict, will temporarily rename
    # that, then remove the extra, the rename back to .row.
    dplyr::rename(.row_orig = .row) |>
    tidyr::unnest(cols = c(.pred_quantile)) |>
    dplyr::select(-.row) |>
    dplyr::rename(.row = .row_orig) |>
    dplyr::summarise(
      .pred_quantile = mean(.pred_quantile, na.rm = TRUE),
      .by = c(dplyr::all_of(all_group_cols))
    ) |>
    dplyr::full_join(lvl_key, by = ".quantile_levels") |>
    dplyr::select(-.quantile_levels) |>
    tidyr::pivot_wider(
      id_cols = c(dplyr::all_of(group_cols)),
      names_from = ".index",
      values_from = ".pred_quantile"
    )

  qp <-
    tmp |>
    dplyr::select(dplyr::starts_with(".qp_")) |>
    as.matrix() |>
    hardhat::quantile_pred(lvl_key$.quantile_levels)

  tmp$.pred_quantile <- qp

  tmp |>
    dplyr::select(-dplyr::starts_with(".qp_")) |>
    dplyr::relocate(.pred_quantile)
}

average_predictions <- function(x, grid = NULL) {
  metric_types <- metrics_info(attr(x, "metrics"))$type
  param_names <- attr(x, "parameters")$id
  y_nms <- .get_tune_outcome_names(x)

  if (!is.null(grid)) {
    grid <- dplyr::select(grid, dplyr::all_of(param_names))
    if (ncol(grid) != length(param_names)) {
      cli::cli_abort(
        "{.arg grid} should only have columns: {.val {param_names}}."
      )
    }
    if (length(param_names) == 0) {
      x$.predictions <-
        purrr::map(x$.predictions, dplyr::cross_join, grid)
    } else {
      x$.predictions <-
        purrr::map(
          x$.predictions,
          dplyr::inner_join,
          grid,
          by = param_names
        )
    }
  }

  x <-
    x |>
    collect_predictions() |>
    dplyr::select(-starts_with("id"))

  if (all(metric_types == "numeric")) {
    x <- numeric_summarize(x)
  } else if (any(metric_types %in% c("prob", "ordered_prob"))) {
    # Note that this will recompute the hard class predictions since the
    # probability estimates are changing. That's why there is a separate
    # branch below that summarizes the hard class predictions when those are
    # the only predictions.
    x <- prob_summarize(x, param_names)
  } else if (any(metric_types == "class")) {
    x <- class_summarize(x, param_names)
  } else if (any(metric_types %in% c("survival", "time", "linear_pred"))) {
    x <- surv_summarize(x, param_names, y_nms)
  } else if (any(metric_types == "quantile")) {
    x <- quantile_summarize(x)
  } else {
    cli::cli_abort(
      "We don't know about metrics of type: {.val {unique(metric_types)}}."
    )
  }

  if (dplyr::is_grouped_df(x)) {
    x <- dplyr::ungroup(x)
  }
  x
}


# ------------------------------------------------------------------------------

#' @export
#' @rdname collect_predictions
collect_metrics <- function(x, ...) {
  UseMethod("collect_metrics")
}

#' @export
collect_metrics.default <- function(x, ...) {
  cli::cli_abort(
    "No {.fn collect_metrics} exists for {.obj_type_friendly {x}}."
  )
}

#' @export
#' @rdname collect_predictions
collect_metrics.tune_results <- function(
  x,
  ...,
  summarize = TRUE,
  type = c("long", "wide")
) {
  rlang::check_dots_empty()
  rlang::arg_match0(type, values = c("long", "wide"))

  if (inherits(x, "last_fit")) {
    res <- x$.metrics[[1]]
  } else {
    if (summarize) {
      res <- estimate_tune_results(x)
    } else {
      res <- collector(x, coll_col = ".metrics")
    }
  }

  if (identical(type, "wide")) {
    res <- pivot_metrics(x, res)
  }

  res
}

pivot_metrics <- function(x, x_metrics) {
  params <- .get_tune_parameter_names(x)
  res <- paste_param_by(x_metrics)

  tidyr::pivot_wider(
    res,
    id_cols = c(
      dplyr::any_of(c(params, ".config", ".iter", ".eval_time")),
      starts_with("id")
    ),
    names_from = .metric,
    values_from = dplyr::any_of(c(".estimate", "mean"))
  )
}

collector <- function(x, coll_col = ".predictions") {
  is_iterative <- any(colnames(x) == ".iter")
  if (is_iterative) {
    keep_cols <- c(coll_col, ".iter")
  } else {
    keep_cols <- coll_col
  }

  id_cols <- colnames(x)[grepl("id", colnames(x))]
  keep_cols <- c(id_cols, keep_cols)
  x <- x[keep_cols]
  coll_col <- x[[coll_col]]
  sizes <- vctrs::list_sizes(coll_col)

  res <-
    vctrs::vec_cbind(
      vctrs::vec_rep_each(x[, id_cols], times = sizes),
      vctrs::list_unchop(coll_col)
    )

  if (is_iterative) {
    res <-
      vctrs::vec_cbind(
        res,
        vctrs::vec_rep_each(x[, ".iter"], times = sizes)
      )
  }

  arrange_cols <- c(".eval_time", ".iter", ".config")
  arrange_cols <- arrange_cols[rlang::has_name(res, arrange_cols)]

  vec_slice(res, vctrs::vec_order(res[arrange_cols]))
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
# Get relationship between the parameter values, .config, and (potentially) .iter
.config_key_from_metrics <- function(x) {
  param_names <- .get_tune_parameter_names(x)
  tibble_metrics <- purrr::map_lgl(x[[".metrics"]], tibble::is_tibble)
  x <- vctrs::vec_slice(x, tibble_metrics)
  x <- x[, colnames(x) %in% c(".iter", ".metrics")]

  metrics <- x[[".metrics"]]

  out <- vctrs::list_unchop(metrics)
  out <- out[c(param_names, ".config")]

  if (rlang::has_name(x, ".iter")) {
    iter <- x[[".iter"]]
    out[[".iter"]] <- vctrs::vec_rep_each(
      iter,
      times = vctrs::list_sizes(metrics)
    )
  }

  out <- vctrs::vec_unique(out)

  out
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
estimate_tune_results <- function(x, ..., col_name = ".metrics") {
  rlang::check_dots_empty()
  param_names <- .get_tune_parameter_names(x)
  id_names <- grep("^id", names(x), value = TRUE)
  group_cols <- .get_extra_col_names(x)

  all_bad <- .is_cataclysmic(x)
  if (all_bad) {
    cli::cli_abort(
      "All models failed. Run {.code show_notes(.Last.tune.result)} for more
       information."
    )
  }

  resample_weights <- .get_resample_weights(x)

  # The mapping of tuning parameters and .config.
  config_key <- .config_key_from_metrics(x)

  tibble_metrics <- purrr::map_lgl(x[[col_name]], tibble::is_tibble)
  x <- x[tibble_metrics, c(id_names, col_name)]
  x <- tidyr::unnest(x, cols = c(all_of(col_name)))

  if (col_name == ".extracts") {
    # For sub-model parameters, there are parameter columns in the current tibble
    # but the list column of tibbles may also have some of those parameter columns
    # too (and these are more accurate). We will see if there is overlap and
    # delete the top-level columns.
    outer_names <- names(x)
    inner_names <- names(x$.extracts[[1]])
    conflicts <- intersect(outer_names, inner_names)
    conflicts <- intersect(conflicts, param_names)
    if (length(conflicts) > 0) {
      x <- dplyr::select(x, -dplyr::all_of(conflicts))
    }
    # Now bring the extract values to the top
    x <- tidyr::unnest(x, cols = .extracts)
    # There may _still_ be list columns and, if there are, un-nest these.
    is_list_col <- purrr::map_lgl(x, is.list)
    if (any(is_list_col)) {
      list_cols <- names(is_list_col)[is_list_col]
      x <- tidyr::unnest(x, cols = c(dplyr::all_of(list_cols)))
    }
    # Again, for models with sub-model parameters, the current .config may not
    # be accurate so we remove it and merge back in later.
    x$.config <- NULL
    if (any(names(x) == ".iter")) {
      x$.iter <- NULL
    }

    x <- dplyr::distinct(x)
  }

  x <- tibble::as_tibble(x)
  x <- vctrs::vec_slice(x, x$id != "Apparent")

  # Join weights to the data if available
  if (!is.null(resample_weights)) {
    weight_data <- .create_weight_mapping(resample_weights, id_names, x)
    if (!is.null(weight_data)) {
      x <- dplyr::left_join(x, weight_data, by = id_names)
    } else {
      # If weight mapping failed, fall back to unweighted
      resample_weights <- NULL
    }
  }

  if (!is.null(resample_weights)) {
    # Use weighted aggregation
    x <- x |>
      dplyr::group_by(
        !!!rlang::syms(param_names),
        .metric,
        .estimator,
        !!!rlang::syms(group_cols)
      ) |>
      dplyr::summarize(
        mean = stats::weighted.mean(.estimate, .resample_weight),
        n = sum(!is.na(.estimate)),
        effective_n = .effective_sample_size(.resample_weight[
          !is.na(.estimate)
        ]),
        std_err = .weighted_sd(.estimate, .resample_weight) /
          sqrt(pmax(effective_n, 1)),
        .groups = "drop"
      ) |>
      dplyr::select(-effective_n)
  } else {
    x <- x |>
      dplyr::group_by(
        !!!rlang::syms(param_names),
        .metric,
        .estimator,
        !!!rlang::syms(group_cols)
      ) |>
      dplyr::summarize(
        mean = mean(.estimate, na.rm = TRUE),
        n = sum(!is.na(.estimate)),
        std_err = sd(.estimate, na.rm = TRUE) / sqrt(n),
        .groups = "drop"
      )
  }

  # only join when parameters are being tuned (#600)
  if (length(param_names) == 0) {
    x <- x |>
      dplyr::bind_cols(config_key)
  } else {
    x <- x |>
      dplyr::full_join(config_key, by = param_names)
  }

  arrange_names <- intersect(c(".iter", ".config", ".eval_time"), names(x))
  dplyr::arrange(x, !!!rlang::syms(arrange_names))
}

# ------------------------------------------------------------------------------

#' @export
#' @rdname collect_predictions
collect_notes <- function(x, ...) {
  UseMethod("collect_notes")
}

#' @export
collect_notes.default <- function(x, ...) {
  cli::cli_abort(
    "No {.fn collect_notes} exists for {.obj_type_friendly {x}}."
  )
}

#' @export
#' @rdname collect_predictions
collect_notes.tune_results <- function(x, ...) {
  if (inherits(x, "last_fit")) {
    return(x$.notes[[1]])
  }

  x |>
    dplyr::select(dplyr::starts_with("id"), dplyr::any_of(".iter"), .notes) |>
    tidyr::unnest(cols = .notes)
}

# ----------------------------------------------------------------------------

#' @export
#' @rdname collect_predictions
collect_extracts <- function(x, ...) {
  UseMethod("collect_extracts")
}

#' @export
collect_extracts.default <- function(x, ...) {
  cli::cli_abort(
    "No {.fn collect_extracts} exists for {.obj_type_friendly {x}}."
  )
}

#' @export
#' @rdname collect_predictions
collect_extracts.tune_results <- function(x, ...) {
  if (!".extracts" %in% colnames(x)) {
    cli::cli_abort(c(
      "!" = "The {.var .extracts} column does not exist.",
      "i" = "Please supply a {.help [control object](tune::control_grid)} with \\
             a non-{.var NULL} {.arg extract} argument during resample fitting."
    ))
  }

  x |>
    dplyr::select(
      dplyr::starts_with("id"),
      dplyr::any_of(".iter"),
      .extracts
    ) |>
    tidyr::unnest(cols = .extracts)
}

if (rlang::is_installed(c("modeldata", "splines2", "kernlab"))) {
  data(two_class_dat, package = "modeldata")

  # ------------------------------------------------------------------------------

  set.seed(6735)
  rep_folds <- rsample::vfold_cv(mtcars, v = 2, repeats = 2)

  spline_rec <- recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = 3)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  lm_splines <-
    fit_resamples(
      lin_mod,
      spline_rec,
      rep_folds,
      control = control_grid(save_pred = TRUE)
    )

  set.seed(93114)
  rep_folds_class <- rsample::vfold_cv(two_class_dat, v = 2, repeats = 3)

  svm_mod <-
    parsnip::svm_rbf(cost = tune("cost value")) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("classification")

  suppressMessages(
    svm_tune <-
      tune_bayes(
        svm_mod,
        Class ~ .,
        rep_folds_class,
        initial = 2,
        iter = 2,
        control = control_bayes(save_pred = TRUE)
      )
  )

  svm_tune_class <- svm_tune
  svm_tune_class$.predictions <-
    purrr::map(
      svm_tune_class$.predictions,
      \(.x) .x |> dplyr::select(-.pred_Class1, -.pred_Class2)
    )
  attr(svm_tune_class, "metrics") <- yardstick::metric_set(yardstick::kap)

  svm_grd <- show_best(svm_tune, metric = "roc_auc") |>
    dplyr::select(`cost value`)
}

# ------------------------------------------------------------------------------

test_that("`collect_predictions()` errors informatively if there is no `.predictions` column", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  expect_snapshot(error = TRUE, {
    collect_predictions(lm_splines |> dplyr::select(-.predictions))
  })
})

test_that("`collect_predictions()` errors informatively applied to unsupported class", {
  expect_snapshot(
    error = TRUE,
    collect_predictions(lm(mpg ~ disp, mtcars))
  )
})

# ------------------------------------------------------------------------------

test_that("`collect_predictions()`, un-averaged", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  res <- collect_predictions(lm_splines)
  exp_res <-
    unnest(
      lm_splines |> dplyr::select(.predictions, starts_with("id")),
      cols = c(.predictions)
    ) |>
    dplyr::select(all_of(names(res)))
  expect_equal(res, exp_res)

  res <- collect_predictions(svm_tune)
  exp_res <-
    unnest(
      svm_tune |> dplyr::select(.predictions, starts_with("id"), .iter),
      cols = c(.predictions)
    ) |>
    dplyr::select(all_of(names(res)))
  res_subset <- collect_predictions(svm_tune, parameters = svm_grd[1, ])
  exp_res_subset <- dplyr::filter(
    exp_res,
    `cost value` == svm_grd$`cost value`[[1]]
  )
  expect_equal(res_subset, exp_res_subset)
})

# ------------------------------------------------------------------------------

test_that("bad filter grid", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kernlab")

  expect_snapshot(
    error = TRUE,
    collect_predictions(svm_tune, parameters = tibble(wrong = "value"))
  )
  expect_true(
    nrow(collect_predictions(
      svm_tune,
      parameters = tibble(`cost value` = 1)
    )) ==
      0
  )
})

# ------------------------------------------------------------------------------

test_that("regression predictions, averaged", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  all_res <- collect_predictions(lm_splines)
  res <- collect_predictions(lm_splines, summarize = TRUE)
  expect_equal(nrow(res), nrow(mtcars))
  expect_false(dplyr::is_grouped_df(res))

  # pull out an example to test
  all_res_subset <- dplyr::filter(all_res, .row == 3)
  res_subset <- dplyr::filter(res, .row == 3)
  expect_equal(mean(all_res_subset$.pred), res_subset$.pred)
})

# ------------------------------------------------------------------------------

test_that("classification class predictions, averaged", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kernlab")

  all_res <- collect_predictions(svm_tune_class)
  res <- collect_predictions(svm_tune_class, summarize = TRUE)
  expect_equal(nrow(res), nrow(two_class_dat) * nrow(svm_grd))
  expect_false(dplyr::is_grouped_df(res))
  expect_named(
    collect_predictions(svm_tune, summarize = TRUE),
    c(
      ".pred_class",
      ".pred_Class1",
      ".pred_Class2",
      ".row",
      "cost value",
      "Class",
      ".config",
      ".iter"
    ),
    ignore.order = TRUE
  )

  # pull out an example to test
  all_res_subset <-
    dplyr::filter(all_res, .row == 5 & `cost value` == svm_grd$`cost value`[1])
  mode_val <- names(sort(table(all_res_subset$.pred_class)))[2]
  exp_val <- factor(mode_val, levels = levels(all_res_subset$Class))
  res_subset <-
    dplyr::filter(res, .row == 5 & `cost value` == svm_grd$`cost value`[1])
  expect_equal(exp_val, res_subset$.pred_class)
})

# ------------------------------------------------------------------------------

test_that("classification class and prob predictions, averaged", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("kernlab")

  all_res <- collect_predictions(svm_tune)
  res <- collect_predictions(svm_tune, summarize = TRUE)
  expect_equal(nrow(res), nrow(two_class_dat) * nrow(svm_grd))
  expect_false(dplyr::is_grouped_df(res))

  # pull out an example to test
  all_res_subset <-
    dplyr::filter(all_res, .row == 5 & `cost value` == svm_grd$`cost value`[1])
  .pred_Class1 <- mean(all_res_subset$.pred_Class1)
  .pred_Class2 <- 1 - .pred_Class1
  .pred_class <- ifelse(.pred_Class2 > .pred_Class1, "Class2", "Class1")
  .pred_class <- factor(.pred_class, levels = levels(all_res_subset$Class))
  res_subset <-
    dplyr::filter(res, .row == 5 & `cost value` == svm_grd$`cost value`[1])
  expect_equal(.pred_Class1, res_subset$.pred_Class1)
  expect_equal(.pred_Class2, res_subset$.pred_Class2)
  expect_equal(.pred_class, res_subset$.pred_class)
})

# ------------------------------------------------------------------------------

test_that("collecting notes - fit_resamples", {
  skip_if(new_rng_snapshots)
  skip_if(!rankdeficient_version)
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  mtcars2 <- mtcars |> mutate(wt2 = wt)
  set.seed(1)
  flds <- rsample::bootstraps(mtcars2, times = 2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect_snapshot(
    lm_splines <- fit_resamples(lin_mod, mpg ~ ., flds)
  )
  expect_snapshot(lm_splines)

  nts <- collect_notes(lm_splines)
  expect_true(all(nts$type == "warning"))
  expect_true(all(grepl("rank", nts$note)))
  expect_equal(names(nts), c("id", "location", "type", "note", "trace"))
})

test_that("collecting notes - last_fit", {
  skip_if(!rankdeficient_version)

  options(pillar.advice = FALSE, pillar.min_title_chars = Inf)

  mtcars2 <- mtcars |> mutate(wt2 = wt)
  set.seed(1)
  split <- rsample::initial_split(mtcars2)

  lin_mod <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  expect_snapshot(
    lst <- last_fit(lin_mod, mpg ~ ., split),
    transform = catalog_lines
  )
  expect_snapshot(lst)

  nts <- collect_notes(lst)
  expect_true(all(nts$type == "warning"))
  expect_true(all(grepl("rank", nts$note)))
  expect_equal(names(nts), c("location", "type", "note", "trace"))
})

test_that("`collect_notes()` errors informatively applied to unsupported class", {
  expect_snapshot(
    error = TRUE,
    collect_notes(lm(mpg ~ disp, mtcars))
  )
})

test_that("collecting extracted objects - fit_resamples", {
  # skip pre-R-4.0.0 so that snaps aren't affected by stringsAsFactors change
  skip_if(R.Version()$major < "4")

  spec <- parsnip::linear_reg()
  form <- mpg ~ .
  set.seed(1)
  boots <- rsample::bootstraps(mtcars, 5)

  ctrl_fit <- control_resamples(extract = extract_fit_engine)
  ctrl_err <- control_resamples(extract = function(x) {
    stop("eeeep! eep!")
  })

  res_fit <- fit_resamples(spec, form, boots, control = ctrl_fit)
  res_nothing <- fit_resamples(spec, form, boots)
  suppressMessages({
    res_error <- fit_resamples(spec, form, boots, control = ctrl_err)
  })

  expect_snapshot(collect_extracts(res_fit))
  expect_snapshot(collect_extracts(res_nothing), error = TRUE)
  expect_snapshot(collect_extracts(res_error))
})

test_that("`collect_extracts()` errors informatively applied to unsupported class", {
  expect_snapshot(
    error = TRUE,
    collect_extracts(lm(mpg ~ disp, mtcars))
  )
})

test_that("`collect_metrics()` errors informatively applied to unsupported class", {
  expect_snapshot(
    error = TRUE,
    collect_metrics(lm(mpg ~ disp, mtcars))
  )
})

test_that("`collect_metrics(type)` errors informatively with bad input", {
  skip_on_cran()

  expect_snapshot(
    error = TRUE,
    collect_metrics(ames_grid_search, type = "boop")
  )

  expect_snapshot(
    error = TRUE,
    collect_metrics(ames_grid_search, type = NULL)
  )
})

test_that("`pivot_metrics()`, grid search, typical metrics, summarized", {
  expect_equal(
    pivot_metrics(ames_grid_search, collect_metrics(ames_grid_search)) |>
      dplyr::slice(),
    tibble::tibble(
      K = integer(0),
      weight_func = character(0),
      dist_power = numeric(0),
      lon = integer(0),
      lat = integer(0),
      .config = character(0),
      rmse = numeric(0),
      rsq = numeric(0)
    )
  )
})

test_that("`pivot_metrics()`, grid search, typical metrics, unsummarized", {
  expect_equal(
    pivot_metrics(
      ames_grid_search,
      collect_metrics(ames_grid_search, summarize = FALSE)
    ) |>
      dplyr::slice(),
    tibble::tibble(
      K = integer(0),
      weight_func = character(0),
      dist_power = numeric(0),
      lon = integer(0),
      lat = integer(0),
      .config = character(0),
      id = character(0),
      rmse = numeric(0),
      rsq = numeric(0)
    )
  )
})

test_that("`pivot_metrics()`, iterative search, typical metrics, summarized", {
  expect_equal(
    pivot_metrics(ames_iter_search, collect_metrics(ames_iter_search)) |>
      dplyr::slice(),
    tibble::tibble(
      K = integer(0),
      weight_func = character(0),
      dist_power = numeric(0),
      lon = integer(0),
      lat = integer(0),
      .config = character(0),
      .iter = integer(0),
      rmse = numeric(0),
      rsq = numeric(0)
    )
  )
})

test_that("`pivot_metrics()`, resampled fits, fairness metrics, summarized", {
  skip_if_not_installed("kknn")

  mtcars_fair <- mtcars
  mtcars_fair$vs <- as.factor(mtcars_fair$vs)
  mtcars_fair$cyl <- as.factor(mtcars_fair$cyl)
  mtcars_fair$am <- as.factor(mtcars_fair$am)
  set.seed(4400)

  ms <-
    yardstick::metric_set(
      yardstick::demographic_parity(cyl),
      yardstick::demographic_parity(am)
    )

  res <-
    fit_resamples(
      nearest_neighbor("classification"),
      vs ~ mpg + hp + cyl,
      rsample::bootstraps(mtcars_fair, 3),
      metrics = ms
    )

  expect_equal(
    pivot_metrics(res, collect_metrics(res)) |> slice(),
    tibble::tibble(
      .config = character(0),
      `demographic_parity(am)` = integer(0),
      `demographic_parity(cyl)` = numeric(0),
    )
  )
})

test_that("resampling quantile regression models", {
  skip_if_not_installed("parsnip", minimum_version = "1.4.1.9002")
  skip_if_not_installed("xgboost", minimum_version = "3.1.3.1")
  skip_if_not_installed("modeldata")

  set.seed(1)
  sim_dat <- modeldata::sim_regression(200)
  sim_rs <- rsample::bootstraps(sim_dat, times = 3)

  bst_spec <-
    parsnip::boost_tree(min_n = 10, trees = 10) |>
    set_mode("quantile regression", quantile_levels = (0:10) / 10)

  set.seed(2)
  bst_res <-
    bst_spec |>
    fit_resamples(
      outcome ~ .,
      resamples = sim_rs,
      control = control_resamples(save_pred = TRUE)
    )

  expect_s3_class(bst_res, "tune_results")

  bst_mtr <- collect_metrics(bst_res)
  expect_true(bst_mtr$.metric == "weighted_interval_score")
  expect_true(nrow(bst_mtr) == 1)

  bst_pred_ind <- collect_predictions(bst_res, summarize = FALSE)
  expect_s3_class(
    bst_pred_ind$.pred_quantile,
    c("quantile_pred", "vctrs_rcrd", "vctrs_vctr")
  )

  bst_pred_mean <- collect_predictions(bst_res, summarize = TRUE)
  expect_s3_class(
    bst_pred_mean$.pred_quantile,
    c("quantile_pred", "vctrs_rcrd", "vctrs_vctr")
  )
  expect_true(nrow(bst_pred_mean) < nrow(bst_pred_ind))
  expect_true(all(table(bst_pred_mean$.row) == 1))
})

test_that("tuning quantile regression models", {
  skip_if_not_installed("parsnip", minimum_version = "1.4.1.9002")
  skip_if_not_installed("xgboost", minimum_version = "3.1.3.1")
  skip_if_not_installed("modeldata")

  set.seed(1)
  sim_dat <- modeldata::sim_regression(200)
  sim_rs <- rsample::bootstraps(sim_dat, times = 3)

  bst_spec <-
    parsnip::boost_tree(min_n = tune(), trees = 10) |>
    set_mode("quantile regression", quantile_levels = (0:10) / 10)

  set.seed(2)
  bst_res <-
    bst_spec |>
    tune_grid(
      outcome ~ .,
      resamples = sim_rs,
      control = control_grid(save_pred = TRUE, save_workflow = TRUE),
      grid = tibble::tibble(min_n = c(2, 5, 10))
    )

  expect_s3_class(bst_res, "tune_results")

  bst_mtr <- collect_metrics(bst_res)
  expect_true(all(bst_mtr$.metric == "weighted_interval_score"))
  expect_true(nrow(bst_mtr) == 3)

  bst_pred_ind <- collect_predictions(bst_res, summarize = FALSE)
  expect_s3_class(
    bst_pred_ind$.pred_quantile,
    c("quantile_pred", "vctrs_rcrd", "vctrs_vctr")
  )

  bst_pred_mean <- collect_predictions(bst_res, summarize = TRUE)
  expect_s3_class(
    bst_pred_mean$.pred_quantile,
    c("quantile_pred", "vctrs_rcrd", "vctrs_vctr")
  )
  expect_true(nrow(bst_pred_mean) < nrow(bst_pred_ind))
  expect_true(all(table(bst_pred_mean$.row) == 3))

  bst_best <- select_best(bst_res, metric = "weighted_interval_score")
  expect_equal(bst_best$min_n, 10L)

  set.seed(3)
  bst_fit_best <- fit_best(bst_res, metric = "weighted_interval_score")
  expect_s3_class(bst_fit_best, "workflow")
  expect_s3_class(bst_fit_best |> extract_fit_engine(), "xgb.Booster")

  bst_plot <- autoplot(bst_res)
  expect_s3_class(bst_plot, "ggplot")
})


test_that("bootstrap intervals for quantile regression models", {
  skip_if_not_installed("parsnip", minimum_version = "1.4.1.9002")
  skip_if_not_installed("xgboost", minimum_version = "3.1.3.1")
  skip_if_not_installed("modeldata")

  set.seed(1)
  sim_dat <- modeldata::sim_regression(200)
  sim_rs <- rsample::bootstraps(sim_dat, times = 3)

  bst_spec <-
    parsnip::boost_tree(min_n = tune(), trees = 10) |>
    set_mode("quantile regression", quantile_levels = (0:10) / 10)

  set.seed(2)
  bst_res <-
    bst_spec |>
    tune_grid(
      outcome ~ .,
      resamples = sim_rs,
      control = control_grid(save_pred = TRUE),
      grid = tibble::tibble(min_n = c(2, 5, 10))
    )

  set.seed(3)
  expect_warning(
    bst_int <- rsample::int_pctl(bst_res, alpha = 0.1, times = 100)
  )
  expect_true(tibble::is_tibble(bst_int))
  expect_equal(nrow(bst_int), 3L)
})

test_that("recipe only", {
  load(test_path("data", "test_objects.RData"))
  grid <- collect_metrics(mt_spln_lm_grid) |>
    dplyr::select(deg_free) |>
    dplyr::distinct()

  purrr::map2(
    mt_spln_lm_grid$splits,
    mt_spln_lm_grid$.predictions,
    check_predictions,
    grid,
    obj = mt_spln_lm_grid
  )

  # initial values for Bayes opt
  init <- mt_spln_lm_bo |> dplyr::filter(.iter == 0)
  init_grid <-
    collect_metrics(mt_spln_lm_bo) |>
    dplyr::filter(.iter == 0) |>
    dplyr::select(deg_free) |>
    dplyr::distinct()

  purrr::map2(
    init$splits,
    init$.predictions,
    check_predictions,
    init_grid,
    obj = mt_spln_lm_grid
  )

  # Now search iterations with a dummy grid
  bo <- mt_spln_lm_bo |> dplyr::filter(.iter > 0)
  bo_grid <- init_grid |> dplyr::slice(1)

  purrr::map2(
    bo$splits,
    bo$.predictions,
    check_predictions,
    bo_grid,
    obj = mt_spln_lm_bo
  )
})

# ------------------------------------------------------------------------------

test_that("model only", {
  load(test_path("data", "test_objects.RData"))
  grid <-
    collect_metrics(mt_knn_grid) |>
    dplyr::select(neighbors) |>
    dplyr::distinct()

  purrr::map2(
    mt_knn_grid$splits,
    mt_knn_grid$.predictions,
    check_predictions,
    tune_df = grid,
    obj = mt_knn_grid
  )

  # initial values for Bayes opt
  init <- mt_knn_bo |> dplyr::filter(.iter == 0)
  init_grid <-
    collect_metrics(mt_knn_bo) |>
    dplyr::filter(.iter == 0) |>
    dplyr::select(neighbors) |>
    distinct()

  purrr::map2(
    init$splits,
    init$.predictions,
    check_predictions,
    init_grid,
    obj = mt_knn_bo
  )

  # Now search iterations with a dummy grid
  bo <- mt_knn_bo |> dplyr::filter(.iter > 0)
  bo_grid <- init_grid |> dplyr::slice(1)

  purrr::map2(
    bo$splits,
    bo$.predictions,
    check_predictions,
    bo_grid,
    obj = mt_knn_bo
  )
})


# ------------------------------------------------------------------------------

test_that("model and recipe", {
  load(test_path("data", "test_objects.RData"))
  grid <-
    collect_metrics(mt_spln_knn_grid) |>
    dplyr::select(deg_free, neighbors) |>
    dplyr::distinct()

  purrr::map2(
    mt_spln_knn_grid$splits,
    mt_spln_knn_grid$.predictions,
    check_predictions,
    grid,
    obj = mt_spln_knn_grid
  )

  # initial values for Bayes opt
  init <- mt_spln_knn_bo |> dplyr::filter(.iter == 0)
  init_grid <-
    collect_metrics(mt_spln_knn_bo) |>
    dplyr::filter(.iter == 0) |>
    dplyr::select(deg_free, neighbors) |>
    dplyr::distinct()

  purrr::map2(
    init$splits,
    init$.predictions,
    check_predictions,
    init_grid,
    obj = mt_spln_knn_grid
  )

  # Now search iterations with a dummy grid
  bo <- mt_spln_knn_bo |> dplyr::filter(.iter > 0)
  bo_grid <- init_grid |> dplyr::slice(1)

  purrr::map2(
    bo$splits,
    bo$.predictions,
    check_predictions,
    bo_grid,
    obj = mt_spln_knn_bo
  )
})

# Determine any qualitative parameters and their ranges

find_qual_param <- function(pset) {
  is_qual <- purrr::map_lgl(pset$object, ~ inherits(.x, "qual_param"))
  if (!any(is_qual)) {
    return(list())
  }

  pset <- pset[is_qual, ]

  possible_lvl <- purrr::map(pset$object, ~ .x$values)
  names(possible_lvl) <- pset$id
  possible_lvl
}

# Leave as-is but scale the others; order by parameter id in pset

# Create the kernel for the continuous parameters
# Add a new kernel for each qualitative parameter

make_kernel <- function(pset, lvls) {
  qual_ind <- which(pset$id %in% names(lvls))
  quant_ind <- setdiff(seq_along(pset$id), qual_ind)
  quant_avail <- length(quant_ind) > 0

  if (length(qual_ind) == 0) {
    return(GauPro::k_Matern32(D = length(quant_ind)))
  }

  num_kernels <- length(qual_ind) + quant_avail

  kernels <- vector(mode = "list", length = num_kernels)
  kern_count <- 0

  if (quant_avail) {
    kernels[[1]] <- GauPro::k_IgnoreIndsKernel(
      k = GauPro::k_Matern32(D = length(quant_ind)),
      ignoreinds = qual_ind
    )
    kern_count <- 1
  }

  for (i in seq_along(qual_ind)) {
    kern_count <- kern_count + 1
    kernels[[kern_count]] <- GauPro::k_FactorKernel(
      D = 1,
      xindex = qual_ind[i],
      nlevels = length(lvls[[1]])
    )
  }

  for (i in 1:num_kernels) {
    if (i == 1) {
      res <- kernels[[1]]
    } else {
      res <- res * kernels[[i]]
    }
  }
  res
}

check_gp <- function(x, control) {
  model_fail <- inherits(x, "try-error")
  gp_threshold <- 0.1
  if (!model_fail) {
    loo_res <- summary(x)
    loo_bad <- loo_res$coverage95LOO < gp_threshold
    loo_rsq <- loo_res$r.squaredLOO
  } else {
    loo_bad <- TRUE
    loo_rsq <- 0.0
  }
  # convergence?
  if (model_fail) {
    message_wrap(
      cli::format_inline("GP failed: {as.character(x)}"),
      prefix = cli::symbol$checkbox_circle_on,
      color_text = get_tune_colors()$message$danger
    )
  } else if (loo_rsq < gp_threshold) {
    if (control$verbose_iter) {
      msg <- cli::format_inline(
        "GP has a LOO R\u00B2 of {round(loo_rsq * 100, 1)}% and is unreliable."
      )
      message_wrap(
        msg,
        prefix = cli::symbol$checkbox_circle_on,
        color_text = get_tune_colors()$message$danger
      )
    }
  } else if (loo_bad) {
    if (control$verbose_iter) {
      msg <- cli::format_inline(
        "GP has a coverage rate < {round(gp_threshold * 100, 1)}% and is unreliable."
      )
      message_wrap(
        msg,
        prefix = cli::symbol$checkbox_circle_on,
        color_text = get_tune_colors()$message$danger
      )
    }
  }
  list(use = !loo_bad && !model_fail && loo_rsq > gp_threshold, rsq = loo_rsq)
}

# encode_set() was created to work on all types of tuning parameters; usage of
# GauPro means that we should not encode qualitative tuning parameters so we
# need a wrapper
partial_encode <- function(dat, pset) {
  qual_info <- find_qual_param(pset)

  if (any(names(dat) == "mean")) {
    outcomes <- dat$mean
  }

  normalized <- encode_set(
    dat |> dplyr::select(dplyr::all_of(pset$id)),
    pset = pset,
    as_matrix = FALSE
  )

  # Replace with the original data when qualitative parameters
  # GauPro::gpkm can take factor encodings to work
  for (i in seq_along(qual_info)) {
    nm <- names(qual_info)[i]
    normalized[[nm]] <- factor(dat[[nm]], levels = qual_info[[i]])
  }
  if (any(names(dat) == "mean")) {
    normalized$.outcome <- dat$mean
  }
  normalized
}

# ------------------------------------------------------------------------------

fit_gp <- function(
  dat,
  pset,
  metric,
  eval_time = NULL,
  control,
  ...
) {
  tune::empty_ellipses(...)

  dat <- dat |> dplyr::filter(.metric == metric)

  if (!is.null(eval_time)) {
    dat <- dat |> dplyr::filter(.eval_time == eval_time)
  }

  dat <- dat |>
    check_gp_data() |>
    dplyr::select(dplyr::all_of(pset$id), mean)

  qual_info <- find_qual_param(pset)
  num_pred <- nrow(pset)
  num_cand <- nrow(dat)

  normalized <- partial_encode(dat, pset)
  # gaupro will look for * or : in names and fail :-(
  # change names and change back later
  og_names <- colnames(normalized)
  gp_names <- make.names(og_names)
  colnames(normalized) <- gp_names

  gp_kernel <- make_kernel(pset, qual_info)

  if (num_cand <= num_pred && num_cand > 0 & control$verbose_iter) {
    msg <- cli::format_inline(
      "The Gaussian process model is being fit using {num_pred} feature{?s} but
			only has {num_cand} data point{?s} to do so. This may cause errors or a
			poor model fit."
    )
    message_wrap(
      msg,
      prefix = "!",
      color_text = get_tune_colors()$message$warning
    )
  }

  gp_fit <- try(
    GauPro::gpkm(
      .outcome ~ .,
      data = normalized,
      kernel = gp_kernel,
      verbose = 0,
      restarts = 5,
      nug.est = FALSE,
      parallel = FALSE
    ),
    silent = TRUE
  )

  new_check <- check_gp(gp_fit, control)

  if (control$verbose_iter) {
    if (new_check$use) {
      msg <- cli::format_inline(
        "Gaussian process model (LOO R\u00B2: {round(new_check$rsq * 100, 1)}%)"
      )
      message_wrap(
        msg,
        prefix = cli::symbol$tick,
        color_text = get_tune_colors()$message$success
      )
    } else {
      message_wrap(
        "Gaussian process model failed",
        prefix = cli::symbol$tick,
        color_text = get_tune_colors()$message$danger
      )
    }
  }

  list(
    fit = gp_fit,
    use = new_check$use,
    rsq = new_check$rsq,
    tr = normalized
  )
}

# ------------------------------------------------------------------------------

quiet_pred_gp <- function(object, new_data, ...) {
  sssh_pred <- purrr::quietly(object$fit$pred)
  res <- sssh_pred(new_data, ...)
  wrn <- res$warnings

  if (length(wrn) > 0) {
    wrn <- wrn[!grepl("Too small", wrn)]
    for (i in seq_along(wrn)) {
      cli::cli_warn("{wrn[i]}")
    }
  }
  res$result
}

pred_gp <- function(object, pset, size = 5000, current = NULL, control) {
  candidates <- dials::grid_space_filling(
    pset,
    size = size,
    type = "latin_hypercube"
  ) |>
    dplyr::distinct()

  if (!object$use) {
    x <- partial_encode(candidates, pset)
    colnames(x) <- make.names(colnames(x))
    x_old <- object$tr
    x_old <- x_old[, names(x)]

    # Remove existing points
    x <- dplyr::anti_join(x, x_old, by = make.names(pset$id))

    keep_ind <- dissim_sample(x_old, x, pset, max_n = Inf)
    candidates <- candidates[keep_ind, ] |>
      dplyr::mutate(.mean = NA_real_, .sd = NA_real_)

    if (control$verbose_iter) {
      message_wrap(
        "Generating a candidate as far away from existing points as possible.",
        prefix = cli::symbol$info,
        color_text = get_tune_colors()$message$info
      )
    }

    return(candidates)
  } else {
    if (control$verbose_iter) {
      message_wrap(
        paste("Generating", nrow(candidates), "candidates."),
        prefix = cli::symbol$info,
        color_text = get_tune_colors()$message$info
      )
    }
  }

  if (!is.null(current)) {
    candidates <- candidates |>
      dplyr::anti_join(current, by = pset$id)
  }

  if (nrow(candidates) == 0) {
    message_wrap(
      "No remaining candidate models",
      prefix = cli::symbol$tick,
      color_text = get_tune_colors()$message$warning
    )
    return(candidates |> dplyr::mutate(.mean = NA_real_, .sd = NA_real_))
  }

  x <- partial_encode(candidates, pset)
  colnames(x) <- make.names(colnames(x))

  gp_pred <- quiet_pred_gp(object, x, se.fit = TRUE)

  gp_pred <- tibble::as_tibble(gp_pred) |>
    dplyr::select(.mean = mean, .sd = se)
  dplyr::bind_cols(candidates, gp_pred)
}


pick_candidate <- function(results, info, control) {
  bad_gp <- all(is.na(results$.mean))
  if (!bad_gp & info$uncertainty < control$uncertain) {
    results <- results |>
      dplyr::slice_max(objective, n = 1, with_ties = FALSE)
  } else {
    if (control$verbose_iter) {
      message_wrap(
        "Uncertainty sample",
        prefix = cli::symbol$info,
        color_text = get_tune_colors()$message$info
      )
    }
    n_top <- max(floor(.1 * nrow(results)), 1L)
    results <- results |>
      dplyr::arrange(dplyr::desc(.sd)) |>
      dplyr::slice(seq_len(n_top)) |>
      dplyr::sample_n(1)
  }
  results
}

dissim_sample <- function(ref_data, candidates, pset, max_n = Inf) {
  max_n <- min(max_n, nrow(candidates))
  candidates <- candidates[1:max_n, , drop = FALSE]
  all_data <- dplyr::bind_rows(ref_data, candidates)

  # Deal with any qualitative predictors by casting them to c(0,1)
  qual_info <- find_qual_param(pset)
  if (length(qual_info) > 0) {
    for (i in seq_along(qual_info)) {
      nm <- names(qual_info)[i]
      uniq <- sort(unique(all_data[[nm]]))
      all_data[[nm]] <- as.character(all_data[[nm]])
      all_data[[nm]] <- factor(all_data[[nm]], levels = uniq)
    }
  }
  all_data <- stats::model.matrix(~ . + 0, data = all_data)

  n_ref <- nrow(ref_data)
  n_all <- nrow(all_data)
  distances <- stats::dist(all_data)
  distances <- as.matrix(distances)
  distances <- distances[1:n_ref, (n_ref + 1):n_all]
  min_distances <- apply(distances, 2, function(x) min(x[x > 0]))
  max_ind <- which.max(min_distances)[1]
  max_ind
}

test_that("encoding before model", {
  skip_if_not_installed("dials", minimum_version = "1.4.0")
  knn_set <- readRDS(test_path("data", "knn_set.rds"))
  knn_grid <- readRDS(test_path("data", "knn_grid.rds"))

  knn_encoded <- tune:::encode_set(knn_grid, knn_set)

  expect_true(all(knn_encoded$K >= 0 & knn_encoded$K <= 1))
  expect_true(all(knn_encoded$exponent >= 0 & knn_encoded$exponent <= 1))
  expect_true(is.factor(knn_encoded$weight_func))
  expect_equal(levels(knn_encoded$weight_func), dials::weight_func()$values)
})

# ------------------------------------------------------------------------------

test_that("GP fit - svm - failure", {
  svm_results <- readRDS(test_path("data", "svm_results.rds"))
  svm_set <- attributes(svm_results)$parameters

  expect_snapshot({
    svm_gp <-
      tune:::fit_gp(
        collect_metrics(svm_results),
        pset = svm_set,
        metric = "accuracy",
        control = control_bayes(verbose = TRUE)
      )
  })

  expect_equal(class(svm_gp), "list")
  expect_named(
    svm_gp,
    c("fit", "use", "rsq", "tr")
  )
  expect_false(svm_gp$use)
  expect_named(
    svm_gp$tr,
    c("cost", "X....", "scale_factor", ".outcome")
  )

  curr <-
    collect_metrics(svm_results) |>
    dplyr::filter(.metric == "accuracy") |>
    mutate(.iter = 0)

  expect_snapshot({
    svm_scores <-
      tune:::pred_gp(
        svm_gp,
        pset = svm_set,
        size = 20,
        current = curr,
        control = control_bayes(verbose_iter = TRUE)
      )
  })
})

# ------------------------------------------------------------------------------

test_that("GP scoring with failed model", {
  svm_results <- readRDS(test_path("data", "svm_results.rds"))
  svm_set <- attributes(svm_results)$parameters

  ctrl <- control_bayes(verbose_iter = TRUE)
  curr <-
    collect_metrics(svm_results) |>
    dplyr::filter(.metric == "accuracy") |>
    mutate(.iter = 0)

  expect_snapshot({
    svm_gp <-
      tune:::fit_gp(
        collect_metrics(svm_results),
        pset = svm_set,
        metric = "accuracy",
        control = ctrl
      )
  })

  expect_snapshot({
    svm_scores <-
      tune:::pred_gp(
        svm_gp,
        pset = svm_set,
        size = 20,
        current = curr,
        control = ctrl
      )
  })
  expect_true(tibble::is_tibble(svm_scores))
  expect_named(
    svm_scores,
    c("cost", "%^*#", "scale_factor", ".mean", ".sd")
  )
  expect_equal(nrow(svm_scores), 1)
})


# ------------------------------------------------------------------------------

test_that("pick_candidate() selects best objective when GP succeeds", {
  results <- tibble::tibble(
    x = 1:10,
    .mean = seq(0.1, 1, by = 0.1),
    .sd = rev(seq(0.1, 1, by = 0.1)),
    objective = seq(0.1, 1, by = 0.1)
  )
  info <- list(uncertainty = 0)
  ctrl <- control_bayes(uncertain = 5)

  res <- tune:::pick_candidate(results, info, ctrl)
  expect_identical(nrow(res), 1L)
  expect_identical(res$objective, 1)
})

test_that("pick_candidate() falls back to uncertainty sample when all GP predictions are NA", {
  results <- tibble::tibble(
    x = 1:20,
    .mean = rep(NA_real_, 20),
    .sd = seq(0.05, 1, by = 0.05),
    objective = seq(0.05, 1, by = 0.05)
  )
  info <- list(uncertainty = 0)
  ctrl <- control_bayes(uncertain = 5, verbose_iter = FALSE)

  set.seed(1)
  res <- tune:::pick_candidate(results, info, ctrl)
  expect_identical(nrow(res), 1L)
  # Should pick from top 10% by .sd, not by objective
  expect_gte(res$.sd, 0.9)
})

test_that("pick_candidate() emits uncertainty sample message when verbose", {
  results <- tibble::tibble(
    x = 1:20,
    .mean = rep(NA_real_, 20),
    .sd = seq(0.05, 1, by = 0.05),
    objective = seq(0.05, 1, by = 0.05)
  )
  info <- list(uncertainty = 0)
  ctrl <- control_bayes(uncertain = 5, verbose_iter = TRUE)

  expect_snapshot({
    set.seed(1)
    res <- tune:::pick_candidate(results, info, ctrl)
  })
})

# ------------------------------------------------------------------------------

test_that("GP fit - knn", {
  knn_results <- readRDS(test_path("data", "knn_results.rds"))
  knn_set <- attributes(knn_results)$parameters

  knn_mtr <-
    collect_metrics(knn_results) |>
    dplyr::filter(.metric == "roc_auc")

  set.seed(1)
  knn_gp <-
    tune:::fit_gp(
      knn_mtr,
      pset = knn_set,
      metric = "roc_auc",
      control = control_bayes()
    )

  expect_equal(class(knn_gp), "list")
  expect_named(
    knn_gp,
    c("fit", "use", "rsq", "tr")
  )
  expect_true(knn_gp$use)
  expect_named(
    knn_gp$tr,
    c("K", "weight_func", "exponent", ".outcome")
  )

  expect_snapshot({
    set.seed(1)
    knn_scores <-
      tune:::pred_gp(
        knn_gp,
        pset = knn_set,
        size = 20,
        current = knn_mtr |> mutate(.iter = 0),
        control = control_bayes()
      )
  })

  expect_named(
    knn_scores,
    c("K", "weight_func", "exponent", ".mean", ".sd")
  )
  expect_equal(nrow(knn_scores), 20L)
})

#' Exponential decay function
#'
#' [expo_decay()] can be used to increase or decrease a function exponentially
#' over iterations. This can be used to dynamically set parameters for
#' acquisition functions as iterations of Bayesian optimization proceed.
#'
#' Note that, when used with the acquisition functions in [tune()], a wrapper
#' would be required since only the first argument would be evaluated during
#' tuning.
#'
#' @param iter An integer for the current iteration number.
#' @param start_val The number returned for the first iteration.
#' @param limit_val The number that the process converges to over iterations.
#' @param slope A coefficient for the exponent to control the rate of decay. The
#'  sign of the slope controls the direction of decay.
#' @return A single numeric value.
#' @examplesIf tune:::should_run_examples()
#' library(tibble)
#' library(purrr)
#' library(ggplot2)
#' library(dplyr)
#' tibble(
#'   iter = 1:40,
#'   value = map_dbl(
#'     1:40,
#'     expo_decay,
#'     start_val = .1,
#'     limit_val = 0,
#'     slope = 1 / 5
#'   )
#' ) |>
#'   ggplot(aes(x = iter, y = value)) +
#'   geom_path()
#' @export
expo_decay <- function(iter, start_val, limit_val, slope = 1 / 5) {
  # between 0:1
  unscaled <- 1 - exp(-slope * (iter - 1))
  (unscaled * (limit_val - start_val)) + start_val
}

#' Example Analysis of Ames Housing Data
#'
#' @details
#' These objects are the results of an analysis of the Ames
#'  housing data. A K-nearest neighbors model was used with a small
#'  predictor set that included natural spline transformations of
#'  the `Longitude` and `Latitude` predictors. The code used to
#'  generate these examples was:
#'
#' ```
#' library(tidymodels)
#' library(tune)
#' library(AmesHousing)
#'
#' # ------------------------------------------------------------------------------
#'
#' ames <- make_ames()
#'
#' set.seed(4595)
#' data_split <- initial_split(ames, strata = "Sale_Price")
#'
#' ames_train <- training(data_split)
#'
#' set.seed(2453)
#' rs_splits <- vfold_cv(ames_train, strata = "Sale_Price")
#'
#' # ------------------------------------------------------------------------------
#'
#' ames_rec <-
#'   recipe(Sale_Price ~ ., data = ames_train) |>
#'   step_log(Sale_Price, base = 10) |>
#'   step_YeoJohnson(Lot_Area, Gr_Liv_Area) |>
#'   step_other(Neighborhood, threshold = .1)  |>
#'   step_dummy(all_nominal()) |>
#'   step_zv(all_predictors()) |>
#'   step_spline_natural(Longitude, deg_free = tune("lon")) |>
#'   step_spline_natural(Latitude, deg_free = tune("lat"))
#'
#' knn_model <-
#'   nearest_neighbor(
#'     mode = "regression",
#'     neighbors = tune("K"),
#'     weight_func = tune(),
#'     dist_power = tune()
#'   ) |>
#'   set_engine("kknn")
#'
#' ames_wflow <-
#'   workflow() |>
#'   add_recipe(ames_rec) |>
#'   add_model(knn_model)
#'
#' ames_set <-
#'   extract_parameter_set_dials(ames_wflow) |>
#'   update(K = neighbors(c(1, 50)))
#'
#' set.seed(7014)
#' ames_grid <-
#'   ames_set |>
#'   grid_max_entropy(size = 10)
#'
#' ames_grid_search <-
#'   tune_grid(
#'     ames_wflow,
#'     resamples = rs_splits,
#'     grid = ames_grid
#'   )
#'
#' set.seed(2082)
#' ames_iter_search <-
#'   tune_bayes(
#'     ames_wflow,
#'     resamples = rs_splits,
#'     param_info = ames_set,
#'     initial = ames_grid_search,
#'     iter = 15
#'   )
#' ```
#'
#' __important note__: Since the `rsample` split columns contain a reference
#' to the same data, saving them to disk can results in large object sizes when
#' the object is later used. In essence, R replaces all of those references with
#' the actual data. For this reason, we saved zero-row tibbles in their place.
#' This doesn't affect how we use these objects in examples but be advised that
#' using some `rsample` functions on them will cause issues.
#'
#' @name example_ames_knn
#' @aliases ames_wflow ames_grid_search ames_iter_search
#' @docType data
#' @return \item{ames_wflow}{A workflow object}
#'  \item{ames_grid_search,ames_iter_search}{Results of model tuning. }
#'
#'
#' @keywords datasets
#' @examples
#' library(tune)
#'
#' ames_grid_search
#' ames_iter_search
NULL

# ---
# repo: tidymodels/parsnip
# file: standalone-survival.R
# last-updated: 2023-12-08
# license: https://unlicense.org
# ---

# This file provides a portable set of helper functions for survival analysis.
# The original is in the parsnip package.

# ## Changelog
# 2023-12-08
# * move .filter_eval_time to this file
#
# 2023-11-09
# * make sure survival vectors are unnamed.
#
# 2023-06-14
# * removed time to factor conversion
#
# 2023-05-18
# * added time to factor conversion
#
# 2023-02-28:
# * Initial version
#
# ------------------------------------------------------------------------------
#
# @param surv A [survival::Surv()] object
# @details
# `.is_censored_right()` always returns a logical while
# `.check_censored_right()` will fail if `FALSE`.
#
# `.extract_status()` will return the data as 0/1 even if the original object
# used the legacy encoding of 1/2. See [survival::Surv()].

# @return
# - `.extract_surv_status()` returns a vector.
# - `.extract_surv_time()` returns a vector when the type is `"right"` or `"left"`
#    and a tibble otherwise.
# - Functions starting with `.is_` or `.check_` return logicals although the
#   latter will fail when `FALSE`.

# nocov start
# These are tested in the extratests repo since it would require a dependency
# on the survival package. https://github.com/tidymodels/extratests/pull/78
.is_surv <- function(surv, fail = TRUE, call = rlang::caller_env()) {
  is_surv <- inherits(surv, "Surv")
  if (!is_surv && fail) {
    rlang::abort("The object does not have class `Surv`.", call = call)
  }
  is_surv
}

.extract_surv_type <- function(surv) {
  attr(surv, "type")
}

.check_cens_type <-
  function(surv, type = "right", fail = TRUE, call = rlang::caller_env()) {
    .is_surv(surv, call = call)
    obj_type <- .extract_surv_type(surv)
    good_type <- all(obj_type %in% type)
    if (!good_type && fail) {
      c_list <- paste0("'", type, "'")
      msg <- cli::format_inline(
        "For this usage, the allowed censoring type{?s} {?is/are}: {c_list}"
      )
      rlang::abort(msg, call = call)
    }
    good_type
  }

.is_censored_right <- function(surv) {
  .check_cens_type(surv, type = "right", fail = FALSE)
}

.check_censored_right <- function(surv, call = rlang::caller_env()) {
  .check_cens_type(surv, type = "right", fail = TRUE, call = call)
} # will add more as we need them

.extract_surv_time <- function(surv) {
  .is_surv(surv)
  keepers <- c("time", "start", "stop", "time1", "time2")
  cols <- colnames(surv)[colnames(surv) %in% keepers]
  res <- surv[, cols, drop = FALSE]
  if (length(cols) > 1) {
    res <- tibble::tibble(as.data.frame(res))
  } else {
    res <- as.numeric(res)
  }
  res
}

.extract_surv_status <- function(surv) {
  .is_surv(surv)
  res <- surv[, "status"]
  un_vals <- sort(unique(res))
  event_type_to_01 <-
    !(.extract_surv_type(surv) %in% c("interval", "interval2", "mstate"))
  if (
    event_type_to_01 &&
      (identical(un_vals, 1:2) | identical(un_vals, c(1.0, 2.0)))
  ) {
    res <- res - 1
  }
  unname(res)
}

# nocov end

# ------------------------------------------------------------------------------

# @param eval_time A vector of numeric time points
# @details
# `.filter_eval_time` checks the validity of the time points.
#
# @return A potentially modified vector of time points.
.filter_eval_time <- function(eval_time, fail = TRUE) {
  if (!is.null(eval_time)) {
    eval_time <- as.numeric(eval_time)
  }
  eval_time_0 <- eval_time
  # will still propagate nulls:
  eval_time <- eval_time[!is.na(eval_time)]
  eval_time <- eval_time[eval_time >= 0 & is.finite(eval_time)]
  eval_time <- unique(eval_time)
  if (fail && identical(eval_time, numeric(0))) {
    cli::cli_abort(
      "There were no usable evaluation times (finite, non-missing, and >= 0).",
      call = NULL
    )
  }
  if (!identical(eval_time, eval_time_0)) {
    diffs <- setdiff(eval_time_0, eval_time)
    cli::cli_warn(
      "There {?was/were} {length(diffs)} inappropriate evaluation
                  time point{?s} that {?was/were} removed.",
      call = NULL
    )
  }
  eval_time
}

#' Quietly load package namespace
#'
#' For one or more packages, load the namespace. This is used during parallel
#' processing since the different parallel backends handle the package
#' environments differently.
#' @param x A character vector of packages.
#' @param infra Should base tidymodels packages be loaded as well?
#' @return An invisible NULL.
#' @keywords internal
#' @export
load_pkgs <- function(x, ..., infra = TRUE) {
  UseMethod("load_pkgs")
}

#' @export
load_pkgs.character <- function(x, ...) {
  rlang::check_dots_empty()
  withr::with_preserve_seed(.load_namespace(x))
}

#' @export
load_pkgs.model_spec <- function(x, ..., infra = TRUE) {
  rlang::check_dots_empty()
  pkgs <- required_pkgs(x)
  if (infra) {
    pkgs <- c(infra_pkgs, pkgs)
  }
  .load_namespace(unique(pkgs))
}

#' @export
load_pkgs.workflow <- function(x, ..., infra = TRUE) {
  rlang::check_dots_empty()
  load_pkgs.model_spec(extract_spec_parsnip(x), infra = infra)
}

full_load <- c("kknn", "earth")

#' @export
#' @rdname tune-internal-functions
.load_namespace <- function(x) {
  if (length(x) == 0) {
    return(invisible(TRUE))
  }

  x_full <- x[x %in% full_load]
  x <- x[!(x %in% full_load)]

  loaded <- purrr::map_lgl(x, isNamespaceLoaded)
  x <- x[!loaded]

  if (length(x) > 0) {
    did_load <- purrr::map_lgl(x, requireNamespace, quietly = TRUE)
    if (any(!did_load)) {
      bad <- x[!did_load]
      cli::cli_abort("The package{?s} {.pkg {bad}} could not be loaded.")
    }
  }

  if (length(x_full) > 0) {
    purrr::map(
      x_full,
      \(.x) {
        try(suppressPackageStartupMessages(attachNamespace(.x)), silent = TRUE)
      }
    )
  }

  invisible(TRUE)
}

infra_pkgs <- c(
  "tune",
  "recipes",
  "parsnip",
  "yardstick",
  "purrr",
  "dplyr",
  "tibble",
  "dials",
  "rsample",
  "workflows",
  "tidyr",
  "rlang",
  "vctrs"
)

# ------------------------------------------------------------------------------
# tune_results

tune_results_reconstruct <- function(x, to) {
  if (tune_results_can_reconstruct(x, to)) {
    new_tune_results_from_template(x, to)
  } else {
    new_bare_tibble(x)
  }
}

new_tune_results_from_template <- function(x, to) {
  attrs <- attributes(to)

  new_tune_results(
    x = x,
    parameters = attrs$parameters,
    metrics = attrs$metrics,
    eval_time = attrs$eval_time,
    eval_time_target = attrs$eval_time_target,
    outcomes = attrs$outcomes,
    rset_info = attrs$rset_info
  )
}

# Invariants:
# - Rows cannot be added or removed
# - Rows can be reordered
# - Columns can be added
# - Columns can be reordered
# - The `splits`, `id`, `.metrics` and `.notes` columns all must exist
tune_results_can_reconstruct <- function(x, to) {
  results_can_reconstruct(
    x = x,
    to = to,
    detect_cols_fn = detect_cols_tune_results,
    detect_order_cols_fn = detect_order_cols_tune_results
  )
}

results_can_reconstruct <- function(
  x,
  to,
  detect_cols_fn,
  detect_order_cols_fn
) {
  x_names <- names(x)
  to_names <- names(to)

  x_indicator <- detect_cols_fn(x_names)
  to_indicator <- detect_cols_fn(to_names)

  x_tune_names <- x_names[x_indicator]
  to_tune_names <- to_names[to_indicator]

  # Ignore ordering
  x_tune_names <- sort(x_tune_names)
  to_tune_names <- sort(to_tune_names)

  # Early return if names aren't identical
  if (!identical(x_tune_names, to_tune_names)) {
    return(FALSE)
  }

  # Avoid all non-bare-data-frame S3 dispatch and
  # don't compare outer data frame attributes.
  # Only look at column names and actual column data.
  x <- new_data_frame(x)
  to <- new_data_frame(to)

  # Early return if number of rows doesn't match
  if (!identical(vec_size(x), vec_size(to))) {
    return(FALSE)
  }

  x_tune_cols <- x[x_tune_names]
  to_tune_cols <- to[x_tune_names]

  # Row order doesn't matter, so reorder both inputs in the same way
  order_cols_indicator <- detect_order_cols_fn(x_tune_names)

  x_order_cols <- x_tune_cols[order_cols_indicator]
  to_order_cols <- to_tune_cols[order_cols_indicator]

  x_order <- vec_order(x_order_cols)
  to_order <- vec_order(to_order_cols)

  if (!identical(x_order, to_order)) {
    x_tune_cols <- vec_slice(x_tune_cols, x_order)
    to_tune_cols <- vec_slice(to_tune_cols, to_order)
  }

  # Check identical structures of sorted tune specific columns
  identical(x_tune_cols, to_tune_cols)
}

# ------------------------------------------------------------------------------
# resample_results

resample_results_reconstruct <- function(x, to) {
  if (resample_results_can_reconstruct(x, to)) {
    new_resample_results_from_template(x, to)
  } else {
    new_bare_tibble(x)
  }
}

new_resample_results_from_template <- function(x, to) {
  attrs <- attributes(to)

  new_resample_results(
    x = x,
    parameters = attrs$parameters,
    metrics = attrs$metrics,
    eval_time = attrs$eval_time,
    eval_time_target = attrs$eval_time_target,
    outcomes = attrs$outcomes,
    rset_info = attrs$rset_info
  )
}

# Invariants:
# - Same as `tune_results`
resample_results_can_reconstruct <- function(x, to) {
  tune_results_can_reconstruct(x, to)
}

# ------------------------------------------------------------------------------
# iteration_results

iteration_results_reconstruct <- function(x, to) {
  if (iteration_results_can_reconstruct(x, to)) {
    new_iteration_results_from_template(x, to)
  } else {
    new_bare_tibble(x)
  }
}

new_iteration_results_from_template <- function(x, to) {
  attrs <- attributes(to)

  new_iteration_results(
    x = x,
    parameters = attrs$parameters,
    metrics = attrs$metrics,
    eval_time = attrs$eval_time,
    eval_time_target = attrs$eval_time_target,
    outcomes = attrs$outcomes,
    rset_info = attrs$rset_info,
    workflow = attrs$workflow
  )
}

# Invariants:
# - Same as `tune_results`, but also special cases the `.iter` column
iteration_results_can_reconstruct <- function(x, to) {
  results_can_reconstruct(
    x = x,
    to = to,
    detect_cols_fn = detect_cols_iteration_results,
    detect_order_cols_fn = detect_order_cols_iteration_results
  )
}

# ------------------------------------------------------------------------------

# - `detect_cols_*()` detects the "special" columns required by the subclass
# - `detect_order_cols_*()` detects the columns used to order the subclass rows.
#   The order columns are generally the special columns but without the
#   list-columns, since `vec_order()` doesn't really order those.

detect_cols_tune_results <- function(x) {
  col_equals_splits(x) |
    col_starts_with_id(x) |
    col_equals_dot_metrics(x) |
    col_equals_dot_notes(x)
}

detect_cols_iteration_results <- function(x) {
  detect_cols_tune_results(x) | col_equals_dot_iter(x)
}

detect_order_cols_tune_results <- function(x) {
  col_starts_with_id(x)
}

detect_order_cols_iteration_results <- function(x) {
  detect_order_cols_tune_results(x) | col_equals_dot_iter(x)
}

# ------------------------------------------------------------------------------

col_equals_splits <- function(x) {
  vec_equal(x, "splits")
}

col_equals_dot_notes <- function(x) {
  vec_equal(x, ".notes")
}

col_equals_dot_metrics <- function(x) {
  vec_equal(x, ".metrics")
}

col_equals_dot_iter <- function(x) {
  vec_equal(x, ".iter")
}

col_starts_with_id <- function(x) {
  grepl("(^id$)|(^id[1-9]$)", x)
}

# ------------------------------------------------------------------------------
# tune_results_can_reconstruct()

test_that("identical results are reconstructable", {
  x <- helper_tune_results$tune
  expect_true(tune_results_can_reconstruct(x, x))
})

test_that("column order doesn't matter", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)[rev(names(to))]
  expect_true(tune_results_can_reconstruct(x, to))
})

test_that("all columns must exist", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)["id"]
  expect_false(tune_results_can_reconstruct(x, to))
})

test_that("can have extra columns", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)
  x[["extra"]] <- 1
  expect_true(tune_results_can_reconstruct(x, to))
})

test_that("can't have different number of rows", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)

  a <- vec_slice(x, integer())
  b <- vec_slice(x, c(1, vec_seq_along(x)))

  expect_false(tune_results_can_reconstruct(a, to))
  expect_false(tune_results_can_reconstruct(b, to))
})

test_that("can't rename a column to unknown name", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)

  names <- names(x)
  names[[1]] <- "foo"
  names(x) <- names

  expect_false(tune_results_can_reconstruct(x, to))
})

test_that("swapping column names is caught", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)

  names <- names(x)
  one <- names[[1]]
  names[[1]] <- names[[3]]
  names[[3]] <- one
  names(x) <- names

  expect_false(tune_results_can_reconstruct(x, to))
})

test_that("column type can't change", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)

  x[["id"]] <- 1

  expect_false(tune_results_can_reconstruct(x, to))
})

test_that("row order doesn't matter", {
  to <- helper_tune_results$tune
  x <- as_tibble(to)

  idx <- rev(seq_len(nrow(x)))
  expect_true(tune_results_can_reconstruct(x[idx, ], to))
})

# ------------------------------------------------------------------------------
# resample_results_can_reconstruct()

# The same as `tune_results_can_reconstruct()`, so just make sure it works

test_that("identical results are reconstructable", {
  x <- helper_tune_results$resample
  expect_true(resample_results_can_reconstruct(x, x))
})

#' Determine if case weights should be passed on to yardstick
#'
#' @description
#' This S3 method defines the logic for deciding when a case weight vector
#' should be passed to yardstick metric functions and used to measure model
#' performance. The current logic is that frequency weights (i.e.
#' [hardhat::frequency_weights()]) are the only situation where this should
#' occur.
#'
#' @param x A vector
#' @return A single `TRUE` or `FALSE`.
#' @export
#' @examples
#' library(parsnip)
#' library(dplyr)
#'
#' frequency_weights(1:10) |>
#'   .use_case_weights_with_yardstick()
#'
#' importance_weights(seq(1, 10, by = .1))|>
#'   .use_case_weights_with_yardstick()
.use_case_weights_with_yardstick <- function(x) {
  UseMethod(".use_case_weights_with_yardstick")
}

#' @export
.use_case_weights_with_yardstick.default <- function(x) {
  message <- c(
    "An object with {.cls hardhat_case_weights} was expected, not
    {.obj_type_friendly {x}}.",
    i = paste0(
      "Define a {.fn .use_case_weights_with_yardstick} method for this type to ",
      "declare whether or not these case weights should be passed on to
       {.pkg yardstick}."
    ),
    i = "See {.help .use_case_weights_with_yardstick} for more information."
  )

  cli::cli_abort(message)
}

#' @rdname dot-use_case_weights_with_yardstick
#' @export
.use_case_weights_with_yardstick.hardhat_importance_weights <- function(x) {
  FALSE
}

#' @rdname dot-use_case_weights_with_yardstick
#' @export
.use_case_weights_with_yardstick.hardhat_frequency_weights <- function(x) {
  TRUE
}

# ------------------------------------------------------------------------------

extract_case_weights <- function(data, workflow) {
  col <- extract_case_weights_col(workflow)

  if (!rlang::is_quosure(col)) {
    cli::cli_abort(
      "{.arg col} must exist and be a quosure at this point.",
      .internal = TRUE
    )
  }

  loc <- eval_select_case_weights(col, data)
  if (length(loc) != 1) {
    # Can't happen now, make sure it doesn't happen in the future
    cli::cli_abort(
      "Only a single case weight column is allowed. {length(loc)}
                   were found: {.val {names(loc)}}."
    )
  }

  if (!tibble::is_tibble(data)) {
    data <- tibble::as_tibble(data)
  }
  case_weights <- data[, loc]
  case_weights <- stats::setNames(case_weights, case_weights_column_name())

  if (!hardhat::is_case_weights(data[[names(loc)]])) {
    cli::cli_abort(
      "Case weights must be a supported case weights type, as determined by
      {.fn hardhat::is_case_weights}."
    )
  }

  case_weights
}

extract_case_weights_col <- function(x) {
  # workflows:::extract_case_weights_col()
  x$pre$actions$case_weights$col
}

eval_select_case_weights <- function(
  col,
  data,
  ...,
  call = rlang::caller_env()
) {
  # workflows:::eval_select_case_weights()

  rlang::check_dots_empty()

  # `col` is saved as a quosure, so it carries along the evaluation environment
  env <- rlang::empty_env()

  loc <- tidyselect::eval_select(
    expr = col,
    data = data,
    env = env,
    error_call = call
  )

  if (length(loc) != 1L) {
    cli::cli_abort(
      "{.arg col} must specify exactly one column from {.arg data} to
       extract case weights from.",
      call = call
    )
  }

  loc
}

case_weights_column_name <- function() {
  ".case_weights"
}

# ------------------------------------------------------------------------------
# .use_case_weights_with_yardstick()

test_that("knows about importance weights", {
  x <- hardhat::importance_weights(1)
  expect_false(.use_case_weights_with_yardstick(x))
})

test_that("knows about frequency weights", {
  x <- hardhat::frequency_weights(1)
  expect_true(.use_case_weights_with_yardstick(x))
})

test_that("gives informative default error", {
  expect_snapshot(error = TRUE, {
    .use_case_weights_with_yardstick(1)
  })
})

# ------------------------------------------------------------------------------
# extract_case_weights()

test_that("`extract_case_weights()` errors if `col` doesn't exist", {
  skip_if(packageVersion("rlang") < "1.0.6.9000")
  wf <- workflows::workflow()

  expect_snapshot(error = TRUE, {
    extract_case_weights(mtcars, wf)
  })
})

test_that("`extract_case_weights()` errors if case weights column isn't the right class", {
  mtcars$weight <- 1L

  wf <- workflows::workflow()
  wf <- workflows::add_case_weights(wf, weight)

  expect_snapshot(error = TRUE, {
    extract_case_weights(mtcars, wf)
  })
})

test_that("`extract_case_weights()` works", {
  set.seed(1)

  mtcars$weight <- hardhat::frequency_weights(1:32)

  folds <- rsample::vfold_cv(mtcars, v = 2)

  spec <- parsnip::linear_reg()
  metrics <- yardstick::metric_set(yardstick::rmse)

  wf <- workflows::workflow()
  wf <- workflows::add_variables(wf, mpg, c(disp, cyl))
  wf <- workflows::add_model(wf, spec)
  wf <- workflows::add_case_weights(wf, weight)

  extracted <- tune:::extract_case_weights(mtcars, wf)
  expted <- mtcars |>
    tibble::as_tibble() |>
    dplyr::select(.case_weights = weight)
  expect_equal(extracted, expted)
})

# ------------------------------------------------------------------------------
# Passed on during tuning

test_that("weights are used during resampling", {
  set.seed(1)

  mtcars$weight <- hardhat::frequency_weights(1:32)

  folds <- rsample::vfold_cv(mtcars, v = 2)

  spec <- parsnip::linear_reg()
  metrics <- yardstick::metric_set(yardstick::rmse)

  wf <- workflows::workflow()
  wf <- workflows::add_variables(wf, mpg, c(disp, cyl))
  wf <- workflows::add_model(wf, spec)
  wf <- workflows::add_case_weights(wf, weight)

  res <- fit_resamples(
    object = wf,
    resamples = folds,
    metrics = metrics,
    control = control_resamples(save_pred = TRUE)
  )

  predictions <- res$.predictions[[1]]

  manual_with_wts <-
    predictions |>
    yardstick::rmse(mpg, .pred, case_weight = .case_weights)

  manual_with_no_wts <-
    predictions |>
    yardstick::rmse(mpg, .pred)

  res_metric <- res$.metrics[[1]] |> dplyr::select(-.config)

  expect_equal(res_metric, manual_with_wts)
  expect_true(!identical(res_metric, manual_with_no_wts))
})


test_that("weights are used during tuning", {
  skip_if_not_installed("splines2")
  set.seed(1)

  mtcars$weight <- hardhat::frequency_weights(1:32)

  folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = tune())

  spec <- parsnip::linear_reg()
  metrics <- yardstick::metric_set(yardstick::rmse)

  wf <- workflows::workflow()
  wf <- workflows::add_recipe(wf, rec)
  wf <- workflows::add_model(wf, spec)
  wf <- workflows::add_case_weights(wf, weight)

  df_grid <- tibble(deg_free = 3:4)

  res <- tune_grid(
    object = wf,
    resamples = folds,
    metrics = metrics,
    grid = df_grid,
    control = control_resamples(save_pred = TRUE)
  )

  predictions <- res$.predictions[[1]] |> group_by(.config)

  manual_with_wts <-
    predictions |>
    yardstick::rmse(mpg, .pred, case_weight = .case_weights)

  manual_with_no_wts <-
    predictions |>
    yardstick::rmse(mpg, .pred)

  res_metric <- res$.metrics[[1]] |>
    dplyr::select(all_of(names(manual_with_wts)))

  expect_equal(res_metric, manual_with_wts)
  expect_true(!identical(res_metric, manual_with_no_wts))
})


test_that("weights work with multi-predict", {
  # glmnet depends on >= 3.6.0 so we don't test on CRAN
  skip_if_not_installed("glmnet")

  set.seed(1)

  mtcars$weight <- hardhat::frequency_weights(1:32)

  folds <- rsample::vfold_cv(mtcars, v = 2)

  spec <- parsnip::linear_reg(penalty = tune(), mixture = tune())
  spec <- parsnip::set_engine(spec, "glmnet")

  grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)

  metrics <- yardstick::metric_set(yardstick::rmse)

  wf <- workflows::workflow()
  wf <- workflows::add_variables(wf, mpg, c(disp, cyl))
  wf <- workflows::add_model(wf, spec)
  wf <- workflows::add_case_weights(wf, weight)

  res <- tune_grid(
    object = wf,
    resamples = folds,
    grid = grid,
    control = control_grid(save_pred = TRUE),
    metrics = metrics
  )

  predictions <- res$.predictions[[1]] |> group_by(.config)

  manual_with_wts <-
    predictions |>
    yardstick::rmse(mpg, .pred, case_weight = .case_weights)

  manual_with_no_wts <-
    predictions |>
    yardstick::rmse(mpg, .pred)

  res_metric <- res$.metrics[[1]] |>
    dplyr::select(all_of(names(manual_with_wts)))

  expect_equal(res_metric, manual_with_wts)
  expect_true(!identical(res_metric, manual_with_no_wts))
})

test_that("importance weights are *not* used during prediction", {
  skip_if_not_installed("splines2")
  set.seed(1)

  mtcars$weight <- hardhat::importance_weights((1:32) / 32)

  folds <- rsample::vfold_cv(mtcars, v = 2)

  rec <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = tune())

  spec <- parsnip::linear_reg()
  metrics <- yardstick::metric_set(yardstick::rmse)

  wf <- workflows::workflow()
  wf <- workflows::add_recipe(wf, rec)
  wf <- workflows::add_model(wf, spec)
  wf <- workflows::add_case_weights(wf, weight)

  df_grid <- tibble(deg_free = 3:4)

  res <- tune_grid(
    object = wf,
    resamples = folds,
    metrics = metrics,
    grid = df_grid,
    control = control_resamples(save_pred = TRUE)
  )

  predictions <- res$.predictions[[1]] |> group_by(.config)

  expect_true(!any(names(predictions) == ".case_weights"))

  manual_with_no_wts <-
    predictions |>
    yardstick::rmse(mpg, .pred)

  res_metric <- res$.metrics[[1]] |>
    dplyr::select(all_of(names(manual_with_no_wts)))

  expect_equal(res_metric, manual_with_no_wts)
})

#' Splice final parameters into objects
#'
#' The `finalize_*` functions take a list or tibble of tuning parameter values and
#' update objects with those values.
#'
#' @param x A recipe, \pkg{parsnip} model specification, \pkg{tailor}
#'  postprocessor, or workflow.
#' @param parameters A list or 1-row tibble of parameter values. Note that the
#'  column names of the tibble should be the `id` fields attached to `tune()`.
#'  For example, in the `Examples` section below, the model has `tune("K")`. In
#'  this case, the parameter tibble should be "K" and not "neighbors".
#' @return An updated version of `x`.
#' @export
#' @examplesIf tune:::should_run_examples(suggests = "kknn")
#' data("example_ames_knn")
#'
#' library(parsnip)
#' knn_model <-
#'   nearest_neighbor(
#'     mode = "regression",
#'     neighbors = tune("K"),
#'     weight_func = tune(),
#'     dist_power = tune()
#'   ) |>
#'   set_engine("kknn")
#'
#' lowest_rmse <- select_best(ames_grid_search, metric = "rmse")
#' lowest_rmse
#'
#' knn_model
#' finalize_model(knn_model, lowest_rmse)
finalize_model <- function(x, parameters) {
  if (!inherits(x, "model_spec") && !inherits(x, "cluster_spec")) {
    cli::cli_abort(
      "{.arg x} should be a parsnip or tidymodels model specification, not
                    {.obj_type_friendly {x}}."
    )
  }
  check_final_param(parameters)
  pset <- hardhat::extract_parameter_set_dials(x)
  if (tibble::is_tibble(parameters)) {
    parameters <- as.list(parameters)
  }

  parameters <- parameters[names(parameters) %in% pset$id]

  discordant <- dplyr::filter(pset, id != name & id %in% names(parameters))
  if (nrow(discordant) > 0) {
    for (i in 1:nrow(discordant)) {
      names(parameters)[names(parameters) == discordant$id[i]] <-
        discordant$name[i]
    }
  }
  rlang::exec(update, object = x, !!!parameters)
}

#' @export
#' @rdname finalize_model
finalize_recipe <- function(x, parameters) {
  if (!inherits(x, "recipe")) {
    cli::cli_abort("{.arg x} should be a recipe, not {.obj_type_friendly {x}}.")
  }
  check_final_param(parameters)
  pset <-
    hardhat::extract_parameter_set_dials(x) |>
    dplyr::filter(id %in% names(parameters) & source == "recipe")

  if (tibble::is_tibble(parameters)) {
    parameters <- as.list(parameters)
  }

  parameters <- parameters[names(parameters) %in% pset$id]
  parameters <- parameters[pset$id]
  split <- vec_split(pset, pset$component_id)
  pset <- split[["val"]]
  for (i in seq_along(pset)) {
    pset_params <- parameters[names(parameters) %in% pset[[i]]$id]
    x <- complete_steps(pset_params, pset[[i]], x)
  }
  x
}

#' @export
#' @rdname finalize_model
finalize_workflow <- function(x, parameters) {
  if (!inherits(x, "workflow")) {
    cli::cli_abort(
      "{.arg x} should be a workflow, not {.obj_type_friendly {x}}."
    )
  }
  check_final_param(parameters)

  mod <- extract_spec_parsnip(x)
  mod <- finalize_model(mod, parameters)
  x <- .set_workflow_spec(x, mod)

  if (.has_preprocessor_recipe(x)) {
    rec <- extract_preprocessor(x)
    rec <- finalize_recipe(rec, parameters)
    x <- .set_workflow_recipe(x, rec)
  }

  if (has_postprocessor(x)) {
    tailor <- extract_postprocessor(x)
    tailor <- finalize_tailor(tailor, parameters)
    x <- set_workflow_tailor(x, tailor)
  }

  x
}

#' @export
#' @rdname finalize_model
finalize_tailor <- function(x, parameters) {
  if (!inherits(x, "tailor")) {
    cli::cli_abort("{.arg x} should be a tailor, not {.obj_type_friendly {x}}.")
  }
  check_final_param(parameters)
  pset <-
    hardhat::extract_parameter_set_dials(x) |>
    dplyr::filter(id %in% names(parameters) & source == "tailor") |>
    dplyr::as_tibble()

  if (tibble::is_tibble(parameters)) {
    parameters <- as.list(parameters)
  }

  parameters <- parameters[names(parameters) %in% pset$id]
  parameters <- parameters[pset$id]

  for (i in seq_along(x$adjustments)) {
    adj <- x$adjustments[[i]]
    adj_comps <- purrr::map_lgl(pset$component, \(.x) inherits(adj, .x))
    if (any(adj_comps)) {
      adj_ids <- pset$id[adj_comps]
      prm_nm <- pset$name[adj_comps]
      adj_prms <- parameters[names(parameters) %in% adj_ids]
      names(adj_prms) <- prm_nm
      adj$arguments <- purrr::list_modify(adj$arguments, !!!adj_prms)
      x$adjustments[[i]] <- adj
    }
  }
  x
}

# ------------------------------------------------------------------------------

check_final_param <- function(x) {
  if (!is.list(x) & !tibble::is_tibble(x)) {
    cli::cli_abort(
      "The parameter object should be a {.cls list} or
                   {.cls tibble}, not {.obj_type_friendly {x}}."
    )
  }
  if (tibble::is_tibble(x) && nrow(x) > 1) {
    cli::cli_abort("The parameter {.arg tibble} should have a single row.")
  }
  invisible(x)
}

complete_steps <- function(param, pset, object) {
  # find the corresponding step in the recipe
  step_ids <- purrr::map_chr(object$steps, "id")
  step_index <- which(unique(pset$component_id) == step_ids)
  step_to_update <- object$steps[[step_index]]

  names(param) <- pset$name

  step_to_update <- rlang::exec(update, object = step_to_update, !!!param)
  object$steps[[step_index]] <- step_to_update
  object
}

test_that("cannot finalize with recipe parameters", {
  skip_if_not_installed("randomForest")
  skip_if_not_installed("splines2")

  set.seed(21983)
  rs <- rsample::vfold_cv(mtcars)

  mod_1 <-
    parsnip::rand_forest(mtry = tune(), trees = 20, min_n = tune()) |>
    parsnip::set_engine("randomForest") |>
    parsnip::set_mode("regression")

  rec_1 <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = tune())

  rec_2 <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = 3)

  expect_snapshot(error = TRUE, {
    mod_1 |> tune_grid(rec_1, resamples = rs, grid = 3)
  })

  set.seed(987323)
  expect_no_error(
    suppressMessages(mod_1 |> tune_grid(rec_2, resamples = rs, grid = 3))
  )
})


test_that("skip error if grid is supplied", {
  skip_if_not_installed("randomForest")
  skip_if_not_installed("splines2")

  set.seed(21983)
  rs <- rsample::vfold_cv(mtcars)

  mod_1 <-
    parsnip::rand_forest(mtry = tune(), trees = 20, min_n = tune()) |>
    parsnip::set_engine("randomForest") |>
    parsnip::set_mode("regression")

  rec_1 <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = tune())

  grid <- tibble::tibble(mtry = 1:3, deg_free = c(3, 3, 4), min_n = c(5, 4, 6))

  set.seed(987323)
  expect_no_error(
    mod_1 |> tune_grid(rec_1, resamples = rs, grid = grid)
  )
})


test_that("finalize recipe step with multiple tune parameters", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  data(biomass, package = "modeldata")

  model_spec <- parsnip::linear_reg() |>
    parsnip::set_engine("lm")

  rec <- recipes::recipe(
    HHV ~ carbon + hydrogen + oxygen + nitrogen + sulfur,
    data = biomass
  ) |>
    recipes::step_spline_b(carbon, hydrogen, deg_free = tune(), degree = tune())

  best <- tibble(deg_free = 2, degree = 1, .config = "Preprocessor1_Model1")

  expect_s3_class(finalize_recipe(rec, best), "recipe")
  expect_equal(finalize_recipe(rec, best)$steps[[1]]$degree, 1)
  expect_equal(finalize_recipe(rec, best)$steps[[1]]$deg_free, 2)
})

# ------------------------------------------------------------------------------
# post-processing

test_that("finalize tailors", {
  skip_if_not_installed("probably")
  skip_if_not_installed("dials", "1.3.0.9000")
  library(tailor)

  adjust_rng <-
    tailor() |>
    adjust_numeric_range(lower_limit = tune(), upper_limit = tune())

  adj_1 <- finalize_tailor(adjust_rng, tibble(lower_limit = 2))
  expect_equal(adj_1$adjustments[[1]]$arguments$lower_limit, 2)
  expect_equal(adj_1$adjustments[[1]]$arguments$upper_limit, tune())

  adj_2 <- finalize_tailor(adjust_rng, tibble(lower_limit = 2, upper_limit = 3))
  expect_equal(adj_2$adjustments[[1]]$arguments$lower_limit, 2)
  expect_equal(adj_2$adjustments[[1]]$arguments$upper_limit, 3)

  adj_3 <- finalize_tailor(
    adjust_rng,
    tibble(lower_limit = 2, upper_limit = 3, a = 2)
  )
  expect_equal(adj_3$adjustments[[1]]$arguments$lower_limit, 2)
  expect_equal(adj_3$adjustments[[1]]$arguments$upper_limit, 3)

  adj_4 <- finalize_tailor(adjust_rng, tibble())
  expect_equal(adj_4, adjust_rng)

  expect_snapshot(
    finalize_tailor(linear_reg(), tibble()),
    error = TRUE
  )
})

test_that("finalize workflows with tailors", {
  skip_if_not_installed("probably")
  skip_if_not_installed("dials", "1.3.0.9000")
  library(tailor)
  library(purrr)

  adjust_rng <-
    tailor() |>
    adjust_numeric_range(lower_limit = tune(), upper_limit = tune())
  wflow <- workflow(y ~ ., linear_reg(), adjust_rng)

  wflow_1 <- finalize_workflow(wflow, tibble(lower_limit = 2))
  expect_equal(
    wflow_1 |>
      extract_postprocessor() |>
      pluck("adjustments") |>
      pluck(1) |>
      pluck("arguments") |>
      pluck("lower_limit"),
    2
  )
  expect_equal(
    wflow_1 |>
      extract_postprocessor() |>
      pluck("adjustments") |>
      pluck(1) |>
      pluck("arguments") |>
      pluck("upper_limit"),
    tune()
  )

  wflow_2 <- finalize_workflow(wflow, tibble(lower_limit = 2, upper_limit = 3))
  expect_equal(
    wflow_2 |>
      extract_postprocessor() |>
      pluck("adjustments") |>
      pluck(1) |>
      pluck("arguments") |>
      pluck("lower_limit"),
    2
  )
  expect_equal(
    wflow_2 |>
      extract_postprocessor() |>
      pluck("adjustments") |>
      pluck(1) |>
      pluck("arguments") |>
      pluck("upper_limit"),
    3
  )

  wflow_3 <- finalize_workflow(
    wflow,
    tibble(lower_limit = 2, upper_limit = 3, a = 2)
  )
  expect_equal(
    wflow_3 |>
      extract_postprocessor() |>
      pluck("adjustments") |>
      pluck(1) |>
      pluck("arguments") |>
      pluck("lower_limit"),
    2
  )
  expect_equal(
    wflow_3 |>
      extract_postprocessor() |>
      pluck("adjustments") |>
      pluck(1) |>
      pluck("arguments") |>
      pluck("upper_limit"),
    3
  )

  wflow_4 <- finalize_workflow(wflow, tibble())
  expect_equal(wflow_4, wflow)

  # https://github.com/tidymodels/tune/issues/998
  cls_post <- tailor() |>
    adjust_probability_threshold(threshold = tune("cut"))
  wflow_thrsh <- workflow(y ~ ., logistic_reg(), cls_post)
  thrsh_param <- tibble(min_n = 2, cut = 1 / 3)
  wflow_thrsh <- wflow_thrsh |> finalize_workflow(thrsh_param)

  expect_equal(
    wflow_thrsh |>
      extract_postprocessor() |>
      pluck("adjustments") |>
      pluck(1) |>
      pluck("arguments") |>
      pluck("threshold"),
    1 / 3
  )
})

#' Acquisition function for scoring parameter combinations
#'
#' These functions can be used to score candidate tuning parameter combinations
#' as a function of their predicted mean and variation.
#'
#' @details The acquisition functions often combine the mean and variance
#' predictions from the Gaussian process model into an objective to be
#' optimized.
#'
#' For this documentation, we assume that the metric in question is better when
#' _maximized_ (e.g. accuracy, the coefficient of determination, etc).
#'
#' The expected improvement of a point `x` is based on the predicted mean and
#' variation at that point as well as the current best value (denoted here as
#' `x_b`). The vignette linked below contains the formulas for this acquisition
#' function. When the `trade_off` parameter is greater than zero, the
#' acquisition function will down-play the effect of the _mean_ prediction and
#' give more weight to the variation. This has the effect of searching for new
#' parameter combinations that are in areas that have yet to be sampled.
#'
#' Note that for [exp_improve()] and [prob_improve()], the `trade_off` value is
#' in the units of the outcome. The functions are parameterized so that the
#' `trade_off` value should always be non-negative.
#'
#' The confidence bound function does not take into account the current best
#' results in the data.
#'
#' If a function is passed to  [exp_improve()] or [prob_improve()], the function
#' can have multiple arguments but only the first (the current iteration number)
#' is given to the function. In other words, the function argument should have
#' defaults for all but the first argument. See [expo_decay()] as an example of
#' a function.
#'
#' @param trade_off A number or function that describes the trade-off between
#'   exploitation and exploration. Smaller values favor exploitation.
#' @param eps A small constant to avoid division by zero.
#' @param kappa A positive number (or function) that corresponds to the
#'   multiplier of the standard deviation in a confidence bound (e.g. 1.96 in
#'   normal-theory 95 percent confidence intervals). Smaller values lean more
#'   towards exploitation.
#' @return An object of class `prob_improve`, `exp_improve`, or `conf_bounds`
#'   along with an extra class of `acquisition_function`.
#'
#' @seealso [tune_bayes()], [expo_decay()]
#' @examples
#' prob_improve()
#' @export
prob_improve <- function(trade_off = 0, eps = .Machine$double.eps) {
  if (!is.numeric(trade_off) & !is_function(trade_off)) {
    cli::cli_abort("{.arg trade_off} should be a number or a function.")
  }

  lab <- "the probability of improvement"

  if (rlang::is_function(trade_off)) {
    farg <- names(formals(trade_off))
    if (length(farg) == 0) {
      cli::cli_abort(
        "The {.fn trade_off} function should have at least one argument."
      )
    }
    lab <- paste(lab, "with variable trade-off values.")
  }

  res <- list(trade_off = trade_off, eps = eps, lable = lab)
  class(res) <- c("prob_improve", "acquisition_function")
  res
}

#' @export
print.prob_improve <- function(x, ...) {
  cat("Acquisition Function: probability of improvment\n")
  invisible(x)
}

#' @export
predict.prob_improve <-
  function(object, new_data, maximize, iter, best, ...) {
    check_bool(maximize)
    check_number_decimal(best, allow_infinite = FALSE)

    if (is.function(object$trade_off)) {
      trade_off <- object$trade_off(iter)
    } else {
      trade_off <- object$trade_off
    }

    new_data <-
      new_data |>
      mutate(.sd = ifelse(.sd <= object$eps, object$eps, .sd))

    if (maximize) {
      new_data <-
        new_data |>
        mutate(delta = ((.mean - best - trade_off) / .sd))
    } else {
      new_data <-
        new_data |>
        mutate(delta = ((trade_off + best - .mean) / .sd))
    }
    new_data |>
      dplyr::mutate(objective = pnorm(delta)) |>
      dplyr::select(objective)
  }

# ------------------------------------------------------------------------------

#' @export
#' @rdname prob_improve
exp_improve <- function(trade_off = 0, eps = .Machine$double.eps) {
  if (!is.numeric(trade_off) & !is_function(trade_off)) {
    cli::cli_abort("{.arg trade_off} should be a number or a function.")
  }

  lab <- "the expected improvement"

  if (rlang::is_function(trade_off)) {
    farg <- names(formals(trade_off))
    if (length(farg) == 0) {
      cli::cli_abort(
        "The {.fn trade_off} function should have at least one argument."
      )
    }
    lab <- paste(lab, "with variable trade-off values.")
  }
  res <- list(trade_off = trade_off, eps = eps, label = lab)
  class(res) <- c("exp_improve", "acquisition_function")
  res
}

#' @export
# NOTE `maximize` is the direction of the metric, not the acquisition function
predict.exp_improve <- function(object, new_data, maximize, iter, best, ...) {
  check_bool(maximize)
  check_number_decimal(best, allow_infinite = FALSE)

  if (is.function(object$trade_off)) {
    trade_off <- object$trade_off(iter)
  } else {
    trade_off <- object$trade_off
  }

  new_data <-
    new_data |>
    mutate(sd_trunc = ifelse(.sd <= object$eps, object$eps, .sd))

  if (maximize) {
    new_data <- new_data |> mutate(delta = .mean - best - trade_off)
  } else {
    new_data <- new_data |> mutate(delta = trade_off + best - .mean)
  }
  new_data <-
    new_data |>
    mutate(
      snr = delta / sd_trunc,
      z = ifelse(.sd <= object$eps, 0, snr),
      objective = (delta * pnorm(z)) + (sd_trunc * dnorm(z))
    )

  new_data |> dplyr::select(objective)
}


#' @export
#' @rdname prob_improve
conf_bound <- function(kappa = 0.1) {
  if (!is.numeric(kappa) & !is_function(kappa)) {
    cli::cli_abort("{.arg kappa} should be a number or a function.")
  }
  lab <- "the confidence bound"
  if (rlang::is_function(kappa)) {
    farg <- names(formals(kappa))
    if (length(farg) == 0) {
      cli::cli_abort(
        "The {.fn trade_off} function should have at least one argument."
      )
    }
    lab <- paste(lab, "with variable kappa values.")
  }
  res <- list(kappa = kappa, label = lab)
  class(res) <- c("conf_bound", "acquisition_function")
  res
}

#' @export
predict.conf_bound <- function(object, new_data, maximize, iter, ...) {
  check_bool(maximize)

  if (is.function(object$kappa)) {
    kappa <- object$kappa(iter)
  } else {
    kappa <- object$kappa
  }

  # `tune` is setup to always maximize the objective function
  if (maximize) {
    new_data <- new_data |> mutate(objective = .mean + kappa * .sd)
  } else {
    new_data <- new_data |> mutate(objective = -(.mean + kappa * .sd))
  }
  new_data |> dplyr::select(objective)
}

test_that("conf_bound interface", {
  expect_snapshot(error = TRUE, conf_bound("a"))
  expect_snapshot(error = TRUE, conf_bound(function() 1))

  test_res <- tibble::tibble(.mean = 1:10, .sd = c((1:9) / 10, NA_real_))
  expect_snapshot(
    error = TRUE,
    predict(conf_bound(), test_res, maximize = 2, iter = 1)
  )
})

test_that("conf_bound calculations", {
  test_res <- tibble::tibble(.mean = 1:10, .sd = c((1:9) / 10, NA_real_))
  dbled <- function(x) x^2

  expect_equal(
    predict(conf_bound(kappa = 1), test_res, maximize = TRUE, iter = 1),
    test_res |> mutate(objective = .mean + 1 * .sd) |> select(objective)
  )
  expect_equal(
    predict(conf_bound(2), test_res, maximize = TRUE, iter = 1),
    test_res |> mutate(objective = .mean + 2 * .sd) |> select(objective)
  )
  expect_equal(
    predict(conf_bound(dbled), test_res, maximize = TRUE, iter = 2),
    test_res |> mutate(objective = .mean + 4 * .sd) |> select(objective)
  )
  expect_equal(
    predict(conf_bound(kappa = 1), test_res, maximize = FALSE, iter = 1),
    test_res |> mutate(objective = -(.mean + 1 * .sd)) |> select(objective)
  )
})


# ------------------------------------------------------------------------------

test_that("prob_improve interface", {
  test_res <- tibble::tibble(.mean = 1:10, .sd = c((1:9) / 10, NA_real_))

  expect_snapshot(error = TRUE, prob_improve("a"))
  expect_snapshot(error = TRUE, prob_improve(function() 1))
  expect_snapshot(
    error = TRUE,
    predict(prob_improve(), test_res, maximize = 2, iter = 1)
  )
  expect_snapshot(
    error = TRUE,
    predict(prob_improve(), test_res, maximize = TRUE, iter = 1, best = NA)
  )
  expect_snapshot(
    error = TRUE,
    predict(prob_improve(), test_res, maximize = TRUE, iter = 1, best = "WAT")
  )
})

test_that("prob_improve calculations", {
  test_res <- tibble::tibble(.mean = 1:10, .sd = c((1:9) / 10, NA_real_))
  dbled <- function(x) x^2

  expect_equal(
    predict(prob_improve(), test_res, maximize = TRUE, iter = 1, best = 15),
    test_res |>
      mutate(objective = pnorm((.mean - 15) / .sd)) |>
      select(objective)
  )
  expect_equal(
    predict(prob_improve(), test_res, maximize = FALSE, iter = 1, best = -2),
    test_res |>
      mutate(objective = pnorm((-2 - .mean) / .sd)) |>
      select(objective)
  )

  expect_equal(
    predict(prob_improve(.1), test_res, maximize = TRUE, iter = 1, best = 15),
    test_res |>
      mutate(objective = pnorm((.mean - 15 - .1) / .sd)) |>
      select(objective)
  )
  expect_equal(
    predict(prob_improve(.1), test_res, maximize = FALSE, iter = 1, best = -2),
    test_res |>
      mutate(objective = pnorm((-2 + .1 - .mean) / .sd)) |>
      select(objective)
  )

  expect_equal(
    predict(
      prob_improve(dbled),
      test_res,
      maximize = TRUE,
      iter = 2,
      best = 15
    ),
    test_res |>
      mutate(objective = pnorm((.mean - 15 - 4) / .sd)) |>
      select(objective)
  )
  expect_equal(
    predict(
      prob_improve(dbled),
      test_res,
      maximize = FALSE,
      iter = 4,
      best = -2
    ),
    test_res |>
      mutate(objective = pnorm((-2 + 16 - .mean) / .sd)) |>
      select(objective)
  )
})


# ------------------------------------------------------------------------------

test_that("exp_improve interface", {
  test_res <- tibble::tibble(.mean = 1:10, .sd = c((1:9) / 10, NA_real_))

  expect_snapshot(error = TRUE, exp_improve("a"))
  expect_snapshot(error = TRUE, exp_improve(function() 2))
  expect_snapshot(
    error = TRUE,
    predict(exp_improve(), test_res, maximize = 2, iter = 1)
  )
  expect_snapshot(
    error = TRUE,
    predict(exp_improve(), test_res, maximize = TRUE, iter = 1, best = NA)
  )
  expect_snapshot(
    error = TRUE,
    predict(exp_improve(), test_res, maximize = TRUE, iter = 1, best = "WAT")
  )
})

test_that("exp_improve calculations", {
  test_res <- tibble::tibble(.mean = 1:10, .sd = c((1:9) / 10, NA_real_))
  dbled <- function(x) x^2

  expect_equal(
    predict(exp_improve(), test_res, maximize = TRUE, iter = 1, best = 15),
    test_res |>
      mutate(
        diff = .mean - 15,
        objective = (diff * pnorm(diff / .sd)) + (.sd * dnorm(diff / .sd))
      ) |>
      select(objective)
  )
  expect_equal(
    predict(exp_improve(), test_res, maximize = FALSE, iter = 1, best = 15),
    test_res |>
      mutate(
        diff = 15 - .mean,
        objective = (diff * pnorm(diff / .sd)) + (.sd * dnorm(diff / .sd))
      ) |>
      select(objective)
  )

  expect_equal(
    predict(exp_improve(1), test_res, maximize = TRUE, iter = 1, best = 15),
    test_res |>
      mutate(
        diff = .mean - 16,
        objective = (diff * pnorm(diff / .sd)) + (.sd * dnorm(diff / .sd))
      ) |>
      select(objective)
  )
  expect_equal(
    predict(exp_improve(1), test_res, maximize = FALSE, iter = 1, best = 15),
    test_res |>
      mutate(
        diff = 16 - .mean,
        objective = (diff * pnorm(diff / .sd)) + (.sd * dnorm(diff / .sd))
      ) |>
      select(objective)
  )

  expect_equal(
    predict(exp_improve(dbled), test_res, maximize = TRUE, iter = 2, best = 15),
    test_res |>
      mutate(
        diff = .mean - 19,
        objective = (diff * pnorm(diff / .sd)) + (.sd * dnorm(diff / .sd))
      ) |>
      select(objective)
  )
  expect_equal(
    predict(
      exp_improve(dbled),
      test_res,
      maximize = FALSE,
      iter = 2,
      best = 15
    ),
    test_res |>
      mutate(
        diff = 19 - .mean,
        objective = (diff * pnorm(diff / .sd)) + (.sd * dnorm(diff / .sd))
      ) |>
      select(objective)
  )
})

#' Compute average confusion matrix across resamples
#'
#' For classification problems, `conf_mat_resampled()` computes a separate
#' confusion matrix for each resample then averages the cell counts.
#'
#' @param x An object with class `tune_results` that was used with a
#' classification model that was run with `control_*(save_pred = TRUE)`.
#' @param ... Currently unused, must be empty.
#' @param parameters A tibble with a single tuning parameter combination. Only
#' one tuning parameter combination (if any were used) is allowed here.
#' @param tidy Should the results come back in a tibble (`TRUE`) or a `conf_mat`
#' object like `yardstick::conf_mat()` (`FALSE`)?
#' @return A tibble or `conf_mat` with the average cell count across resamples.
#' @examplesIf rlang::is_installed("modeldata")
#' # example code
#'
#' library(parsnip)
#' library(rsample)
#' library(dplyr)
#'
#' data(two_class_dat, package = "modeldata")
#'
#' set.seed(2393)
#' res <-
#'   logistic_reg() |>
#'   set_engine("glm") |>
#'   fit_resamples(
#'     Class ~ .,
#'     resamples = vfold_cv(two_class_dat, v = 3),
#'     control = control_resamples(save_pred = TRUE)
#'   )
#'
#' conf_mat_resampled(res)
#' conf_mat_resampled(res, tidy = FALSE)
#' @export
conf_mat_resampled <- function(x, ..., parameters = NULL, tidy = TRUE) {
  rlang::check_dots_empty()
  if (!inherits(x, "tune_results")) {
    cli::cli_abort(
      "The first argument needs to be {.cls tune_results} object,
                   not {.obj_type_friendly {mtcars}}."
    )
  }
  if (!any(names(x) == ".predictions")) {
    cli::cli_abort(
      "The function was not run with the {.code save_pred = TRUE} option.
       Please re-run with that option."
    )
  }
  preds <- collect_predictions(x, summarize = FALSE, parameters = parameters)

  if (!any(names(preds) == ".pred_class")) {
    cli::cli_abort(
      "Cannot find the predicted classes. Was this a classification model?"
    )
  }
  # check for multiple parameter combinations
  params <- .get_tune_parameter_names(x)
  if (length(params) > 0) {
    param_combos <-
      preds |>
      dplyr::select(!!!params) |>
      distinct()
    if (nrow(param_combos) > 1) {
      cli::cli_abort(
        "It looks like there are {nrow(param_combos)} tuning parameter
        combination in the data. Please use the {.arg parameters}
        argument to select one combination of parameters."
      )
    }
  }

  truth <- .get_tune_outcome_names(x)
  if (length(truth) != 1) {
    cli::cli_abort("Cannot determine the proper outcome name")
  }

  id_cols <- grep("(^id$)|($id[1-9]$)", names(preds), value = TRUE)
  preds <-
    preds |>
    dplyr::group_nest(!!!syms(id_cols)) |>
    dplyr::mutate(
      conf_mats = purrr::map(
        data,
        \(.x) {
          yardstick::conf_mat(.x, truth = {{ truth }}, estimate = .pred_class)
        }
      )
    )

  opt <- getOption("dplyr.summarise.inform", default = "FALSE")
  options(dplyr.summarise.inform = FALSE)

  res <-
    purrr::map(preds$conf_mats, \(.x) as.data.frame(.x$table)) |>
    purrr::list_rbind() |>
    dplyr::group_by(Prediction, Truth) |>
    dplyr::summarize(Freq = mean(Freq, na.rm = TRUE)) |>
    dplyr::ungroup()

  options(dplyr.summarise.inform = opt)

  if (!tidy) {
    lvls <- levels(res$Prediction)
    res <- matrix(res$Freq, ncol = length(lvls), byrow = TRUE)
    colnames(res) <- lvls
    rownames(res) <- lvls
    res <- as.table(res) |> yardstick::conf_mat()
  }
  res
}

test_that("appropriate return values", {
  svm_results <- readRDS(test_path("data", "svm_results.rds"))

  expect_no_error(
    cm_1 <-
      conf_mat_resampled(
        svm_results,
        parameters = select_best(svm_results, metric = "accuracy")
      )
  )
  expect_true(tibble::is_tibble(cm_1))

  expect_no_error(
    cm_2 <-
      conf_mat_resampled(
        svm_results,
        parameters = select_best(svm_results, metric = "accuracy"),
        tidy = FALSE
      )
  )
  expect_equal(class(cm_2), "conf_mat")

  expect_equal(
    cm_2$table["Class1", "Class2"],
    cm_1$Freq[cm_1$Prediction == "Class1" & cm_1$Truth == "Class2"]
  )
  expect_equal(
    cm_2$table["Class2", "Class1"],
    cm_1$Freq[cm_1$Prediction == "Class2" & cm_1$Truth == "Class1"]
  )
})

# ------------------------------------------------------------------------------

test_that("bad argss", {
  load(test_path("data", "test_objects.RData"))
  svm_results <- readRDS(test_path("data", "svm_results.rds"))

  expect_snapshot(error = TRUE, {
    conf_mat_resampled(svm_results)
  })
  expect_snapshot(error = TRUE, {
    conf_mat_resampled(mt_knn_bo)
  })
  broke_results <- svm_results
  broke_results$.predictions <- NULL

  expect_snapshot(error = TRUE, {
    conf_mat_resampled(broke_results)
  })
  expect_snapshot(error = TRUE, {
    conf_mat_resampled(tibble::as_tibble(svm_results))
  })

  broke_results <- svm_results
  broke_results$.predictions <-
    purrr::map(
      broke_results$.predictions,
      \(.x) .x |> dplyr::select(-.pred_class)
    )

  expect_snapshot(error = TRUE, {
    conf_mat_resampled(broke_results)
  })

  broke_results <- svm_results
  attr(broke_results, "outcomes") <- NULL

  expect_snapshot(error = TRUE, {
    conf_mat_resampled(
      broke_results,
      parameters = select_best(broke_results, metric = "accuracy")
    )
  })

  expect_snapshot(error = TRUE, {
    conf_mat_resampled(svm_results, argument_that_doesnt_exist = TRUE)
  })
})

#' Investigate best tuning parameters
#'
#' @description
#'
#' [show_best()] displays the top sub-models and their performance estimates.
#'
#' [select_best()] finds the tuning parameter combination with the best
#' performance values.
#'
#' [select_by_one_std_err()] uses the "one-standard error rule" (Breiman _el
#'  at, 1984) that selects the most simple model that is within one standard
#'  error of the numerically optimal results.
#'
#' [select_by_pct_loss()] selects the most simple model whose loss of
#'  performance is within some acceptable limit.
#'
#' @param x The results of [tune_grid()] or [tune_bayes()].
#' @param ... For [select_by_one_std_err()] and [select_by_pct_loss()], this
#' argument is passed directly to [dplyr::arrange()] so that the user can sort
#' the models from *most simple to most complex*. That is, for a parameter `p`,
#' pass the unquoted expression `p` if smaller values of `p` indicate a simpler
#' model, or `desc(p)` if larger values indicate a simpler model. At
#' least one term is required for these two functions. See the examples below.
#' @param metric A character value for the metric that will be used to sort
#'  the models. (See
#'  \url{https://yardstick.tidymodels.org/articles/metric-types.html} for
#'  more details). Not required if a single metric exists in `x`. If there are
#'  multiple metric and none are given, the first in the metric set is used (and
#'  a warning is issued).
#' @param n An integer for the number of top results/rows to return.
#' @param limit The limit of loss of performance that is acceptable (in percent
#' units). See details below.
#' @param eval_time A single numeric time point where dynamic event time
#' metrics should be chosen (e.g., the time-dependent ROC curve, etc). The
#' values should be consistent with the values used to create `x`. The `NULL`
#' default will automatically use the first evaluation time used by `x`.
#' @param call The call to be shown in errors and warnings.
#' @return A tibble with columns for the parameters. [show_best()] also
#'  includes columns for performance metrics.
#' @details
#' For percent loss, suppose the best model has an RMSE of 0.75 and a simpler
#' model has an RMSE of 1. The percent loss would be `(1.00 - 0.75)/1.00 * 100`,
#' or 25 percent. Note that loss will always be non-negative.
#' @references
#' Breiman, Leo; Friedman, J. H.; Olshen, R. A.; Stone, C. J. (1984).
#' _Classification and Regression Trees._ Monterey, CA: Wadsworth.
#' @examplesIf tune:::should_run_examples()
#' data("example_ames_knn")
#'
#' show_best(ames_iter_search, metric = "rmse")
#'
#' select_best(ames_iter_search, metric = "rsq")
#'
#' # To find the least complex model within one std error of the numerically
#' # optimal model, the number of nearest neighbors are sorted from the largest
#' # number of neighbors (the least complex class boundary) to the smallest
#' # (corresponding to the most complex model).
#'
#' select_by_one_std_err(ames_grid_search, metric = "rmse", desc(K))
#'
#' # Now find the least complex model that has no more than a 5% loss of RMSE:
#' select_by_pct_loss(
#'   ames_grid_search,
#'   metric = "rmse",
#'   limit = 5, desc(K)
#' )
#' @export
show_best <- function(x, ...) {
  UseMethod("show_best")
}

#' @export
#' @rdname show_best
show_best.default <- function(x, ...) {
  cli::cli_abort("No {.fn show_best} exists for this type of object.")
}

#' @export
#' @rdname show_best
show_best.tune_results <- function(
  x,
  ...,
  metric = NULL,
  eval_time = NULL,
  n = 5,
  call = rlang::current_env()
) {
  rlang::check_dots_empty()

  metric_info <- choose_metric(x, metric, call = call)
  metric <- metric_info$metric

  eval_time <- choose_eval_time(x, metric, eval_time = eval_time, call = call)

  summary_res <- .filter_perf_metrics(x, metric, eval_time)

  if (metric_info$direction == "maximize") {
    summary_res <- summary_res |> dplyr::arrange(dplyr::desc(mean))
  } else if (metric_info$direction == "minimize") {
    summary_res <- summary_res |> dplyr::arrange(mean)
  } else if (metric_info$direction == "zero") {
    summary_res <- summary_res |> dplyr::arrange(abs(mean))
  }
  show_ind <- 1:min(nrow(summary_res), n)
  summary_res |>
    dplyr::slice(show_ind)
}


#' @export
#' @rdname show_best
select_best <- function(x, ...) {
  UseMethod("select_best")
}

#' @export
#' @rdname show_best
select_best.default <- function(x, ...) {
  cli::cli_abort("No {.fn select_best} exists for this type of object.")
}

#' @export
#' @rdname show_best
select_best.tune_results <- function(x, ..., metric = NULL, eval_time = NULL) {
  rlang::check_dots_empty()

  metric_info <- choose_metric(x, metric)
  metric <- metric_info$metric

  param_names <- .get_tune_parameter_names(x)

  res <- show_best(
    x,
    metric = metric,
    n = 1,
    eval_time = eval_time,
    call = rlang::current_env()
  )
  res |> dplyr::select(dplyr::all_of(param_names), .config)
}

#' @export
#' @rdname show_best
select_by_pct_loss <- function(x, ...) {
  UseMethod("select_by_pct_loss")
}

#' @export
#' @rdname show_best
select_by_pct_loss.default <- function(x, ...) {
  cli::cli_abort("No {.fn select_by_pct_loss} exists for this type of object.")
}

#' @export
#' @rdname show_best
select_by_pct_loss.tune_results <- function(
  x,
  ...,
  metric = NULL,
  eval_time = NULL,
  limit = 2
) {
  metric_info <- choose_metric(x, metric)
  metric <- metric_info$metric

  param_names <- .get_tune_parameter_names(x)

  check_select_dots(...)

  eval_time <- choose_eval_time(x, metric, eval_time = eval_time)

  summary_res <- .filter_perf_metrics(x, metric, eval_time)

  if (metric_info$direction == "maximize") {
    best_metric <- max(summary_res$mean, na.rm = TRUE)
  } else if (metric_info$direction == "minimize") {
    best_metric <- min(summary_res$mean, na.rm = TRUE)
  } else if (metric_info$direction == "zero") {
    which_min <- which.min(abs(summary_res$mean))
    best_metric <- summary_res$mean[which_min]
  }

  summary_res <-
    summary_res |>
    dplyr::rowwise() |>
    dplyr::mutate(
      .best = best_metric,
      # can calculate loss without knowledge of direction since
      # we know that losses must be larger than that for the best
      .loss = abs((abs(mean) - abs(.best)) / .best) * 100
    ) |>
    dplyr::ungroup()

  dots <- rlang::enquos(...)
  summary_res <- try(dplyr::arrange(summary_res, !!!dots), silent = TRUE)
  if (inherits(summary_res, "try-error")) {
    var_nm <- rlang::eval_tidy(dots)
    var_nm <- purrr::map_chr(var_nm, rlang::quo_name)
    var_nm <- var_nm[!var_nm %in% colnames(collect_metrics(x))]
    cli::cli_abort("Could not sort results by {.var {var_nm}}.")
  }

  # discard models more complex than the best and
  # remove models with greater increase in loss than the limit
  best_index <- which(summary_res$.loss == 0)
  summary_res |>
    dplyr::slice(1:best_index) |>
    dplyr::filter(.loss < limit) |>
    dplyr::slice(1) |>
    dplyr::select(dplyr::all_of(param_names), .config)
}

#' @export
#' @rdname show_best
select_by_one_std_err <- function(x, ...) {
  UseMethod("select_by_one_std_err")
}

#' @export
#' @rdname show_best
select_by_one_std_err.default <- function(x, ...) {
  cli::cli_abort(
    "No {.fn select_by_one_std_err} exists for this type of object."
  )
}

#' @export
#' @rdname show_best
select_by_one_std_err.tune_results <- function(
  x,
  ...,
  metric = NULL,
  eval_time = NULL
) {
  metric_info <- choose_metric(x, metric)
  metric <- metric_info$metric

  param_names <- .get_tune_parameter_names(x)

  check_select_dots(...)

  eval_time <- choose_eval_time(x, metric, eval_time = eval_time)

  summary_res <- .filter_perf_metrics(x, metric, eval_time)

  if (metric_info$direction == "maximize") {
    best_index <- which.max(summary_res$mean)
    best <- summary_res$mean[best_index]
    bound <- best - summary_res$std_err[best_index]
    summary_res <-
      summary_res |>
      dplyr::mutate(
        .best = best,
        .bound = bound
      ) |>
      dplyr::filter(mean >= .bound)
  } else if (metric_info$direction == "minimize") {
    best_index <- which.min(summary_res$mean)
    best <- summary_res$mean[best_index]
    bound <- best + summary_res$std_err[best_index]
    summary_res <-
      summary_res |>
      dplyr::mutate(
        .best = best,
        .bound = bound
      ) |>
      dplyr::filter(mean <= .bound)
  } else if (metric_info$direction == "zero") {
    best_index <- which.min(abs(summary_res$mean))
    best <- summary_res$mean[best_index]
    bound_lower <- -abs(best) - summary_res$std_err[best_index]
    bound_upper <- abs(best) + summary_res$std_err[best_index]
    summary_res <-
      summary_res |>
      dplyr::rowwise() |>
      dplyr::mutate(
        .best = best,
        .bound = list(c(lower = bound_lower, upper = bound_upper))
      ) |>
      dplyr::filter(mean >= .bound[[1]] & mean <= .bound[[2]]) |>
      dplyr::ungroup()
  }

  dots <- rlang::enquos(...)
  summary_res <- try(dplyr::arrange(summary_res, !!!dots), silent = TRUE)
  if (inherits(summary_res, "try-error")) {
    var_nm <- rlang::eval_tidy(dots)
    var_nm <- purrr::map_chr(var_nm, rlang::quo_name)
    var_nm <- var_nm[!var_nm %in% colnames(collect_metrics(x))]
    cli::cli_abort("Could not sort results by {.var {var_nm}}.")
  }
  summary_res |>
    dplyr::slice(1) |>
    dplyr::select(dplyr::all_of(param_names), .config)
}

check_select_dots <- function(..., call = rlang::caller_env()) {
  dots <- rlang::enquos(...)
  if (length(dots) == 0) {
    cli::cli_abort(
      "Please choose at least one tuning parameter to sort in {.code ...}.",
      call = call
    )
  }
  invisible(NULL)
}

# ------------------------------------------------------------------------------
# objects for these tests are created using inst/test_objects.R

# ------------------------------------------------------------------------------

test_that("select_best()", {
  skip_if_not_installed("dials", minimum_version = "1.4.0")
  options(width = 200, pillar.advice = FALSE, pillar.min_title_chars = Inf)

  rcv_results <- readRDS(test_path("data", "rcv_results.rds"))

  expect_true(
    tibble::is_tibble(select_best(rcv_results, metric = "rmse"))
  )
  best_rmse <-
    tibble::tribble(
      ~deg_free    ,
      ~degree      ,
      ~`wt df`     ,
      ~`wt degree` ,
      6L           ,
      2L           ,
      2L           ,
      1L
    )
  best_rsq <-
    tibble::tribble(
      ~deg_free    ,
      ~degree      ,
      ~`wt df`     ,
      ~`wt degree` ,
      10L          ,
       2L          ,
       2L          ,
       2L
    )

  expect_equal(
    select_best(rcv_results, metric = "rmse") |> select(-.config),
    best_rmse
  )
  expect_equal(
    select_best(rcv_results, metric = "rsq") |> select(-.config),
    best_rsq
  )

  expect_snapshot(error = TRUE, {
    select_best(rcv_results, metric = "random")
  })
  expect_snapshot(
    select_best(rcv_results, metric = c("rmse", "rsq"))
  )
  expect_snapshot({
    best_default_metric <- select_best(rcv_results)
    best_rmse <- select_best(rcv_results, metric = "rmse")
  })
  expect_equal(best_default_metric, best_rmse)

  expect_snapshot(error = TRUE, {
    select_best(mtcars, metric = "disp")
  })
})


test_that("show_best()", {
  rcv_results <- readRDS(test_path("data", "rcv_results.rds"))

  rcv_rmse <-
    rcv_results |>
    collect_metrics() |>
    dplyr::filter(.metric == "rmse") |>
    dplyr::arrange(mean)

  expect_equal(
    show_best(rcv_results, metric = "rmse", n = 1),
    rcv_rmse |> slice(1)
  )
  expect_equal(
    show_best(rcv_results, metric = "rmse", n = nrow(rcv_rmse) + 1),
    rcv_rmse
  )
  expect_equal(
    show_best(rcv_results, metric = "rmse", n = 1) |> names(),
    rcv_rmse |> names()
  )
  expect_snapshot({
    best_default_metric <- show_best(rcv_results)
    best_rmse <- show_best(rcv_results, metric = "rmse")
  })
  expect_equal(best_default_metric, best_rmse)

  expect_snapshot(error = TRUE, {
    show_best(mtcars, metric = "disp")
  })
})

test_that("one-std error rule", {
  skip_if_not_installed("dials", minimum_version = "1.4.0")
  options(width = 200, pillar.advice = FALSE, pillar.min_title_chars = Inf)

  rcv_results <- readRDS(test_path("data", "rcv_results.rds"))
  knn_results <- readRDS(test_path("data", "knn_results.rds"))

  expect_true(
    tibble::is_tibble(select_by_one_std_err(
      knn_results,
      metric = "accuracy",
      K
    ))
  )

  expect_equal(
    select_by_one_std_err(
      rcv_results,
      metric = "rmse",
      deg_free,
      `wt degree`
    )$.config,
    "pre19_mod0_post0"
  )
  expect_equal(
    select_by_one_std_err(knn_results, metric = "accuracy", K)$K,
    25L
  )

  expect_snapshot(error = TRUE, {
    select_by_one_std_err(rcv_results, metric = "random", deg_free)
  })
  expect_snapshot(
    select_by_one_std_err(rcv_results, metric = c("rmse", "rsq"), deg_free)
  )
  expect_snapshot({
    select_via_default_metric <- select_by_one_std_err(knn_results, K)
    select_via_roc <- select_by_one_std_err(knn_results, K, metric = "roc_auc")
  })
  expect_equal(select_via_default_metric, select_via_roc)

  expect_snapshot(error = TRUE, {
    select_by_one_std_err(rcv_results, metric = "random")
  })

  expect_snapshot(error = TRUE, {
    select_by_one_std_err(mtcars, metric = "disp")
  })
  expect_snapshot(error = TRUE, {
    select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk)
  })
  expect_snapshot(error = TRUE, {
    select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk, K)
  })
  expect_snapshot(error = TRUE, {
    select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk, Kay)
  })
  expect_snapshot(error = TRUE, {
    select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk, desc(K))
  })
})


test_that("percent loss", {
  skip_if_not_installed("dials", minimum_version = "1.4.0")
  options(width = 200, pillar.advice = FALSE, pillar.min_title_chars = Inf)

  rcv_results <- readRDS(test_path("data", "rcv_results.rds"))
  knn_results <- readRDS(test_path("data", "knn_results.rds"))

  expect_true(
    tibble::is_tibble(select_by_pct_loss(knn_results, metric = "accuracy", K))
  )
  expect_equal(
    select_by_pct_loss(
      rcv_results,
      metric = "rmse",
      deg_free,
      `wt degree`
    )$.config,
    "pre19_mod0_post0"
  )
  expect_equal(
    select_by_pct_loss(knn_results, metric = "accuracy", K)$K,
    12L
  )

  expect_snapshot(error = TRUE, {
    select_by_pct_loss(rcv_results, metric = "random", deg_free)
  })
  expect_snapshot(
    select_by_pct_loss(rcv_results, metric = c("rmse", "rsq"), deg_free)
  )
  expect_snapshot({
    select_via_default_metric <- select_by_pct_loss(knn_results, K)
    select_via_roc <- select_by_pct_loss(knn_results, K, metric = "roc_auc")
  })
  expect_equal(select_via_default_metric, select_via_roc)

  expect_snapshot(error = TRUE, {
    select_by_pct_loss(rcv_results, metric = "random")
  })

  expect_snapshot(error = TRUE, {
    select_by_pct_loss(mtcars, metric = "disp")
  })
  expect_snapshot(error = TRUE, {
    select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk)
  })
  expect_snapshot(error = TRUE, {
    select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk, K)
  })
  expect_snapshot(error = TRUE, {
    select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk, Kay)
  })
  expect_snapshot(error = TRUE, {
    select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk, desc(K))
  })

  data("example_ames_knn")
  expect_equal(
    select_by_pct_loss(
      ames_grid_search,
      metric = "rmse",
      limit = 10,
      desc(K)
    )$K,
    40
  )
})

test_that("select_by_* can handle metrics with direction == 'zero'", {
  skip_on_cran()
  skip_if_not_installed("kknn")

  set.seed(1)
  resamples <- rsample::bootstraps(mtcars, times = 5)

  set.seed(1)
  tune_res <-
    tune::tune_grid(
      parsnip::nearest_neighbor(mode = "regression", neighbors = tune()),
      mpg ~ .,
      resamples,
      metrics = yardstick::metric_set(yardstick::mpe, yardstick::msd)
    )

  tune_res_metrics <- tune_res |> collect_metrics()

  expect_equal(
    select_best(tune_res, metric = "msd")$.config,
    tune_res_metrics |>
      filter(.metric == "msd") |>
      arrange(abs(mean)) |>
      slice(1) |>
      select(.config) |>
      pull()
  )

  expect_equal(
    select_best(tune_res, metric = "mpe")$.config,
    tune_res_metrics |>
      filter(.metric == "mpe") |>
      arrange(abs(mean)) |>
      slice(1) |>
      select(.config) |>
      pull()
  )

  expect_equal(
    show_best(tune_res, metric = "msd", n = 5)$.config,
    tune_res_metrics |>
      filter(.metric == "msd") |>
      arrange(abs(mean)) |>
      slice(1:5) |>
      select(.config) |>
      pull()
  )

  expect_equal(
    show_best(tune_res, metric = "mpe", n = 5)$.config,
    tune_res_metrics |>
      filter(.metric == "mpe") |>
      arrange(abs(mean)) |>
      slice(1:5) |>
      select(.config) |>
      pull()
  )

  # one std error, msd ----------
  best <-
    tune_res_metrics |>
    filter(.metric == "msd") |>
    arrange(min(abs(mean))) |>
    slice(1)

  bound_lower <- -abs(best$mean) - abs(best$std_err)
  bound_upper <- abs(best$mean) + abs(best$std_err)
  expect_equal(bound_lower, -bound_upper)

  simplest_within_bound <-
    tune_res_metrics |>
    filter(.metric == "msd") |>
    filter(abs(mean) < bound_upper) |>
    arrange(desc(neighbors)) |>
    slice(1)

  expect_equal(
    select_by_one_std_err(tune_res, metric = "msd", desc(neighbors))$.config,
    simplest_within_bound$.config
  )

  # one std error, mpe ----------
  best <-
    tune_res_metrics |>
    filter(.metric == "mpe") |>
    arrange(min(abs(mean))) |>
    slice(1)

  bound_lower <- -abs(best$mean) - abs(best$std_err)
  bound_upper <- abs(best$mean) + abs(best$std_err)
  expect_equal(bound_lower, -bound_upper)

  simplest_within_bound <-
    tune_res_metrics |>
    filter(.metric == "mpe") |>
    filter(abs(mean) < bound_upper) |>
    arrange(desc(neighbors)) |>
    slice(1)

  expect_equal(
    select_by_one_std_err(tune_res, metric = "mpe", desc(neighbors))$.config,
    simplest_within_bound$.config
  )

  # pct loss, msd ----------
  best <-
    tune_res_metrics |>
    filter(.metric == "msd") |>
    arrange(abs(mean)) |>
    slice(1)

  expect_equal(
    select_by_pct_loss(
      tune_res,
      metric = "msd",
      limit = 10,
      desc(neighbors)
    )$.config,
    tune_res_metrics |>
      filter(.metric == "msd") |>
      dplyr::rowwise() |>
      mutate(loss = abs((abs(mean) - abs(best$mean)) / best$mean) * 100) |>
      ungroup() |>
      arrange(desc(neighbors)) |>
      slice(1:which(.config == best$.config)) |>
      filter(loss < 10) |>
      slice(1) |>
      select(.config) |>
      pull()
  )

  # pct loss, mpe ----------
  best <-
    tune_res_metrics |>
    filter(.metric == "mpe") |>
    arrange(abs(mean)) |>
    slice(1)

  expect_equal(
    select_by_pct_loss(
      tune_res,
      metric = "mpe",
      limit = 10,
      desc(neighbors)
    )$.config,
    tune_res_metrics |>
      filter(.metric == "mpe") |>
      dplyr::rowwise() |>
      mutate(loss = abs((abs(mean) - abs(best$mean)) / best$mean) * 100) |>
      ungroup() |>
      arrange(desc(neighbors)) |>
      slice(1:which(.config == best$.config)) |>
      filter(loss < 10) |>
      slice(1) |>
      select(.config) |>
      pull()
  )
})

test_that("show_best with survival models", {
  surv_res <- readRDS(test_path("data", "surv_boost_tree_res.rds"))

  expect_snapshot(
    show_best(surv_res)
  )
  expect_snapshot(
    show_best(surv_res, metric = "concordance_survival")
  )
  expect_snapshot(
    show_best(surv_res, metric = "brier_survival_integrated")
  )
  expect_snapshot(
    show_best(surv_res, metric = "brier_survival")
  )
  expect_snapshot(
    show_best(surv_res, metric = c("brier_survival", "roc_auc_survival"))
  )
  expect_snapshot(
    show_best(surv_res, metric = "brier_survival", eval_time = 1)
  )
  expect_snapshot(
    show_best(surv_res, metric = "concordance_survival", eval_time = 1)
  )
  expect_snapshot(
    show_best(surv_res, metric = "concordance_survival", eval_time = 1.1)
  )
  expect_snapshot(
    show_best(surv_res, metric = "brier_survival", eval_time = 1.1),
    error = TRUE
  )
  expect_snapshot(
    show_best(surv_res, metric = "brier_survival", eval_time = 1:2)
  )
  expect_snapshot(
    show_best(surv_res, metric = "brier_survival", eval_time = 3:4),
    error = TRUE
  )
})

test_that("select_best with survival models", {
  surv_res <- readRDS(test_path("data", "surv_boost_tree_res.rds"))

  expect_snapshot(
    select_best(surv_res)
  )
  expect_snapshot(
    select_best(surv_res, metric = "concordance_survival")
  )
  expect_snapshot(
    select_best(surv_res, metric = "brier_survival_integrated")
  )
  expect_snapshot(
    select_best(surv_res, metric = "brier_survival")
  )
  expect_snapshot(
    select_best(surv_res, metric = c("brier_survival", "roc_auc_survival"))
  )
  expect_snapshot(
    select_best(surv_res, metric = "brier_survival", eval_time = 1)
  )
  expect_snapshot(
    select_best(surv_res, metric = "concordance_survival", eval_time = 1)
  )
  expect_snapshot(
    select_best(surv_res, metric = "concordance_survival", eval_time = 1.1)
  )
  expect_snapshot(
    select_best(surv_res, metric = "brier_survival", eval_time = 1.1),
    error = TRUE
  )
  expect_snapshot(
    select_best(surv_res, metric = "brier_survival", eval_time = 1:2)
  )
  expect_snapshot(
    select_best(surv_res, metric = "brier_survival", eval_time = 3:4),
    error = TRUE
  )
})

test_that("select_by_one_std_err with survival models", {
  surv_res <- readRDS(test_path("data", "surv_boost_tree_res.rds"))

  expect_snapshot(
    select_by_one_std_err(surv_res, trees)
  )
  expect_snapshot(
    select_by_one_std_err(surv_res, metric = "concordance_survival", trees)
  )
  expect_snapshot(
    select_by_one_std_err(surv_res, metric = "brier_survival_integrated", trees)
  )
  expect_snapshot(
    select_by_one_std_err(surv_res, metric = "brier_survival", trees)
  )
  expect_snapshot(
    select_by_one_std_err(
      surv_res,
      metric = c("brier_survival", "roc_auc_survival"),
      trees
    )
  )
  expect_snapshot(
    select_by_one_std_err(
      surv_res,
      metric = "brier_survival",
      eval_time = 1,
      trees
    )
  )
  expect_snapshot(
    select_by_one_std_err(
      surv_res,
      metric = "concordance_survival",
      eval_time = 1,
      trees
    )
  )
  expect_snapshot(
    select_by_one_std_err(
      surv_res,
      metric = "concordance_survival",
      eval_time = 1.1,
      trees
    )
  )
  expect_snapshot(
    select_by_one_std_err(
      surv_res,
      metric = "brier_survival",
      eval_time = 1.1,
      trees
    ),
    error = TRUE
  )
  expect_snapshot(
    select_by_one_std_err(
      surv_res,
      metric = "brier_survival",
      eval_time = 1:2,
      trees
    )
  )
  expect_snapshot(
    select_by_one_std_err(
      surv_res,
      metric = "brier_survival",
      eval_time = 3:4,
      trees
    ),
    error = TRUE
  )
})

test_that("select_by_pct_loss with survival models", {
  surv_res <- readRDS(test_path("data", "surv_boost_tree_res.rds"))

  expect_snapshot(
    select_by_pct_loss(surv_res, trees)
  )
  expect_snapshot(
    select_by_pct_loss(surv_res, metric = "concordance_survival", trees)
  )
  expect_snapshot(
    select_by_pct_loss(surv_res, metric = "brier_survival_integrated", trees)
  )
  expect_snapshot(
    select_by_pct_loss(surv_res, metric = "brier_survival", trees)
  )
  expect_snapshot(
    select_by_pct_loss(
      surv_res,
      metric = c("brier_survival", "roc_auc_survival"),
      trees
    )
  )
  expect_snapshot(
    select_by_pct_loss(
      surv_res,
      metric = "brier_survival",
      eval_time = 1,
      trees
    )
  )
  expect_snapshot(
    select_by_pct_loss(
      surv_res,
      metric = "concordance_survival",
      eval_time = 1,
      trees
    )
  )
  expect_snapshot(
    select_by_pct_loss(
      surv_res,
      metric = "concordance_survival",
      eval_time = 1.1,
      trees
    )
  )
  expect_snapshot(
    select_by_pct_loss(
      surv_res,
      metric = "brier_survival",
      eval_time = 1.1,
      trees
    ),
    error = TRUE
  )
  expect_snapshot(
    select_by_pct_loss(
      surv_res,
      metric = "brier_survival",
      eval_time = 1:2,
      trees
    )
  )
  expect_snapshot(
    select_by_pct_loss(
      surv_res,
      metric = "brier_survival",
      eval_time = 3:4,
      trees
    ),
    error = TRUE
  )
})

#' @importFrom dplyr dplyr_reconstruct
dplyr::dplyr_reconstruct

#' @noRd
#' @export
dplyr_reconstruct.tune_results <- function(data, template) {
  tune_results_reconstruct(data, template)
}

#' @noRd
#' @export
dplyr_reconstruct.resample_results <- function(data, template) {
  resample_results_reconstruct(data, template)
}

#' @noRd
#' @export
dplyr_reconstruct.iteration_results <- function(data, template) {
  iteration_results_reconstruct(data, template)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
check_rset <- function(x) {
  if (!inherits(x, "rset")) {
    cli::cli_abort(
      c(
        "The {.arg resamples} argument should be an {.cls rset} object,",
        "i" = "Such objects are produced by {.fn vfold_cv} or other
           {.pkg rsample} functions."
      )
    )
  }
  if (inherits(x, "loo_cv")) {
    cli::cli_abort(
      "Leave-one-out cross-validation is not currently supported with {.pkg tune}."
    )
  }
  if (inherits(x, "nested_cv")) {
    cli::cli_abort("Nested resampling is not currently supported with tune.")
  }
  if (inherits(x, "permutations")) {
    cli::cli_abort("Permutation samples are not suitable for tuning.")
  }

  # Check fold weights if present
  check_resample_weights(x)

  invisible(NULL)
}

#' Check fold weights in rset objects
#'
#' @param x An rset object.
#' @return `NULL` invisibly, or error if weights are invalid.
#' @keywords internal
check_resample_weights <- function(x) {
  weights <- attr(x, ".resample_weights")
  if (is.null(weights)) {
    return(invisible(NULL))
  }

  .validate_resample_weights(weights, nrow(x))

  invisible(NULL)
}

backend_options_msg <- "{.arg backend_options} should be created by {.fn tune::new_backend_options}."

check_backend_options <- function(backend_options) {
  if (
    !is.null(backend_options) &&
      !inherits(backend_options, "tune_backend_options")
  ) {
    cli::cli_abort(backend_options_msg)
  }

  invisible(NULL)
}


grid_msg <- "{.arg grid} should be a positive integer or a data frame."

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.check_grid <- function(grid, workflow, pset = NULL, call = caller_env()) {
  # `NULL` grid is the signal that we are using `fit_resamples()`
  if (is.null(grid)) {
    return(grid)
  }

  if (is.null(pset)) {
    pset <- hardhat::extract_parameter_set_dials(workflow)
  }

  if (nrow(pset) == 0L) {
    cli::cli_warn(c(
      "No tuning parameters have been detected, performance will be
       evaluated using the resamples with no tuning.",
      "Did you want to assign any parameters with a value of {.fn tune}?"
    ))

    # Return `NULL` as the new `grid`, like what is used in `fit_resamples()`
    return(NULL)
  }

  if (!is.numeric(grid)) {
    if (!is.data.frame(grid)) {
      cli::cli_abort(grid_msg)
    }

    grid_distinct <- distinct(grid)

    # remove attributes, particularly those tacked on by `expand.grid()`
    grid_distinct <- vctrs::new_data_frame(
      grid_distinct,
      n = nrow(grid_distinct)
    )

    if (!identical(nrow(grid_distinct), nrow(grid))) {
      cli::cli_warn(
        "Duplicate rows in grid of tuning combinations found and removed."
      )
    }
    grid <- grid_distinct

    tune_tbl <- tune_args(workflow)
    tune_params <- tune_tbl$id

    # when called from [tune_bayes()]
    tune_params <- tune_params[tune_params != ".iter"]

    grid_params <- names(grid)

    extra_grid_params <- setdiff(grid_params, tune_params)
    extra_tune_params <- setdiff(tune_params, grid_params)

    if (length(extra_grid_params) != 0L) {
      n_extra <- length(extra_grid_params)

      cli::cli_abort(
        "The provided grid has {n_extra} parameter column{?s}
         ({.var {extra_grid_params}}) that {?has/have} not been marked for tuning
         by {.fn tune}."
      )
    }

    if (length(extra_tune_params) != 0L) {
      n_extra <- length(extra_tune_params)

      cli::cli_abort(
        "The provided grid is missing the following {n_extra} parameter
         column{?s} that {?has/have} been marked for tuning by {.fn tune}:
         {.val {extra_tune_params}}."
      )
    }
  } else {
    grid <- as.integer(grid[1])
    if (grid < 1) {
      cli::cli_abort(grid_msg)
    }
    check_workflow(workflow, pset = pset, check_dials = TRUE, call = call)

    grid <- dials::grid_space_filling(pset, size = grid)
    grid <- dplyr::distinct(grid)
  }

  if (!tibble::is_tibble(grid)) {
    grid <- tibble::as_tibble(grid)
  }

  grid
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.needs_finalization <- function(x, nms = character(0)) {
  # If an unknown engine-specific parameter, the object column is missing and
  # no need for finalization
  x <- x[!is.na(x$object), ]
  # If the parameter is in a pre-defined grid, then no need to finalize
  x <- x[!(x$id %in% nms), ]
  if (length(x) == 0) {
    return(FALSE)
  }
  any(dials::has_unknowns(x$object))
}

#' @export
#' @param data The training data.
#' @param grid_names A character vector of column names from the grid.
#' @keywords internal
#' @rdname empty_ellipses
check_parameters <- function(
  wflow,
  pset = NULL,
  data,
  grid_names = character(0)
) {
  if (is.null(pset)) {
    pset <- hardhat::extract_parameter_set_dials(wflow)
  }
  unk <- purrr::map_lgl(pset$object, dials::has_unknowns)
  if (!any(unk)) {
    return(pset)
  }
  tune_param <- tune_args(wflow)
  tune_recipe <- tune_param$id[tune_param$source == "recipe"]
  tune_recipe <- length(tune_recipe) > 0

  if (.needs_finalization(pset, grid_names)) {
    if (tune_recipe) {
      cli::cli_abort(
        c(
          "Some model parameters require finalization but there are recipe
           parameters that require tuning.",
          "i" = "Please use {.fn extract_parameter_set_dials} to set parameter
           ranges manually and supply the output to the {.arg param_info}
           argument."
        )
      )
    }
    unk_names <- pset$id[unk]
    num_unk <- length(unk_names)
    msg <-
      cli::format_inline(
        "Creating pre-processing data to finalize {num_unk} unknown parameter{?s}: {.val {unk_names}}"
      )

    update_printer(
      list(verbose = TRUE),
      split_labels = NULL,
      task = msg,
      type = "info"
    )

    x <- workflows::.fit_pre(wflow, data)$pre$mold$predictors
    pset$object <- purrr::map(pset$object, dials::finalize, x = x)
  }
  pset
}

shhhh <- function(x) {
  suppressPackageStartupMessages(requireNamespace(x, quietly = TRUE))
}

is_installed <- function(pkg) {
  res <- try(shhhh(pkg), silent = TRUE)
  res
}

check_installs <- function(x, call = caller_env()) {
  if (x$engine == "unknown") {
    cli::cli_abort("Please declare an engine for the model")
  } else {
    m_type <- class(x)[1]
    deps <- parsnip::get_dependency(m_type)
    deps <- deps$pkg[deps$engine == x$engine]
    deps <- unlist(deps)
  }

  if (length(deps) > 0) {
    is_inst <- purrr::map_lgl(deps, is_installed)
    if (any(!is_inst)) {
      needs_installed <- unique(deps[!is_inst])
      cli::cli_abort(
        "{cli::qty(needs_installed)} Package install{?s} {?is/are}
         required for {.pkg {needs_installed}}.",
        call = call
      )
    }
  }
}

check_bayes_initial_size <- function(num_param, num_grid, race = FALSE) {
  msg <-
    cli::pluralize(
      "There {cli::qty(num_param)}{?is/are} {num_param} tuning parameter{?s}
      and {num_grid} grid point{?s} {?was/were} requested."
    )

  msg_list <- c(
    "{msg}",
    "i" = "The GP model requires 2+ initial points. For best performance,
               supply more initial points than there are tuning parameters."
  )
  bullet_msg <-
    c(
      `!` = "{msg}",
      `*` = cli::pluralize(
        "This is likely to cause numerical issues in the first few search iterations."
      )
    )

  if (race) {
    race_msg <- "With racing, only completely resampled parameters are used."
    msg_list <- c(msg_list, "i" = race_msg)
    bullet_msg <- c(bullet_msg, `*` = race_msg)
  }

  if (num_grid == 1) {
    cli::cli_abort(msg_list)
  }

  if (num_grid < num_param + 1) {
    diff <- num_param - num_grid + 1
    cli::cli_bullets(bullet_msg)
  }

  invisible(NULL)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.check_param_objects <- function(pset) {
  params <- purrr::map_lgl(pset$object, inherits, "param")

  if (!all(params)) {
    cli::cli_abort(
      "The workflow has arguments to be tuned that are missing some parameter
       objects: {.val {pset$id[!params]}}"
    )
  }
  invisible(pset)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
#' @param check_dials A logical for check for a NULL parameter object.
check_workflow <- function(
  x,
  ...,
  pset = NULL,
  check_dials = FALSE,
  call = caller_env()
) {
  if (!inherits(x, "workflow")) {
    cli::cli_abort(
      "The {.arg object} argument should be a {.cls workflow} object."
    )
  }

  if (!.has_preprocessor(x)) {
    cli::cli_abort("A formula, recipe, or variables preprocessor is required.")
  }

  if (!.has_spec(x)) {
    cli::cli_abort("A parsnip model is required.")
  }

  rlang::check_dots_empty(call = call)

  if (check_dials) {
    if (is.null(pset)) {
      pset <- hardhat::extract_parameter_set_dials(x)
    }

    .check_param_objects(pset)

    incompl <- dials::has_unknowns(pset$object)

    if (any(incompl)) {
      cli::cli_abort(
        "The workflow has arguments whose ranges are not finalized:
         {.val {pset$id[incompl]}}."
      )
    }
  }

  check_extra_tune_parameters(x)

  check_installs(hardhat::extract_spec_parsnip(x), call = call)

  invisible(NULL)
}

check_extra_tune_parameters <- function(x) {
  mod <- hardhat::extract_spec_parsnip(x)

  to_be_tuned <- hardhat::extract_parameter_set_dials(mod)
  marked_for_tuning <- generics::tune_args(mod)

  if (nrow(marked_for_tuning) > nrow(to_be_tuned)) {
    not_tunable <- marked_for_tuning$name[
      !marked_for_tuning$name %in% to_be_tuned$name
    ]
    msg <-
      c(
        "!" = "{cli::qty(not_tunable)}The parameter{?s} {.var {not_tunable}}
               {?was/were} marked with `tune()`, though will not be tuned.",
        "i" = "This usually means that the current modeling engine
               {.var {extract_spec_parsnip(x)$engine}}
               does not support tuning {.var {not_tunable}}."
      )

    cli::cli_abort(
      msg,
      call = rlang::caller_env(3),
      class = "not_tunable_error"
    )
  }
  invisible(NULL)
}


#' @export
#' @keywords internal
#' @rdname empty_ellipses
#' @param object
#'   `r lifecycle::badge("deprecated")`
#'
#'   A `workflow` object.
check_metrics <- function(x, object) {
  lifecycle::deprecate_warn("2.1.0", "check_metrics()", "check_metrics_arg()")
  mode <- extract_spec_parsnip(object)$mode

  if (is.null(x)) {
    switch(
      mode,
      regression = {
        x <- yardstick::metric_set(rmse, rsq)
      },
      classification = {
        x <- yardstick::metric_set(roc_auc, accuracy, brier_class)
      },
      'censored regression' = {
        x <- yardstick::metric_set(brier_survival)
      },
      'quantile regression' = {
        x <- yardstick::metric_set(weighted_interval_score)
      },
      unknown = {
        cli::cli_abort(
          "Internal error: {.fn check_installs} should have
                            caught an {.val unknown} mode."
        )
      },
      cli::cli_abort("Unknown {.val mode} for parsnip model.")
    )

    return(x)
  }

  is_numeric_metric_set <- inherits(x, "numeric_metric_set")
  is_class_prob_metric_set <- inherits(x, "class_prob_metric_set")
  is_surv_metric_set <- inherits(x, c("survival_metric_set"))

  if (
    !is_numeric_metric_set && !is_class_prob_metric_set && !is_surv_metric_set
  ) {
    cli::cli_abort(
      "The {.arg metrics} argument should be the results of
                   {.fn yardstick::metric_set}."
    )
  }

  if (mode == "regression" && !is_numeric_metric_set) {
    cli::cli_abort(
      c(
        "The parsnip model has {.code mode = 'regression'}, but {.arg metrics}
        is a metric set for a different model mode."
      )
    )
  }

  if (mode == "classification" && !is_class_prob_metric_set) {
    cli::cli_abort(
      c(
        "The parsnip model has {.code mode = 'classification'}, but {.arg metrics}
         is a metric set for a different model mode."
      )
    )
  }

  if (mode == "censored regression" && !is_surv_metric_set) {
    cli::cli_abort(c(
      "The parsnip model has {.code mode = 'censored regression'},
       but {.arg metrics} is a metric set for a different model mode."
    ))
  }
  x
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
#' @param wflow A `workflow` object.
#' @param resamples An `rset` object.
#' @param ctrl A `control_grid` object.
check_initial <- function(
  x,
  pset,
  wflow,
  resamples,
  metrics,
  eval_time,
  ctrl,
  checks = "grid"
) {
  if (is.null(x)) {
    cli::cli_abort(
      "{.arg initial} should be a positive integer or the results
                   of {.fn tune_grid}"
    )
  }
  if (is.numeric(x)) {
    x <- create_initial_set(pset, n = x, checks = checks)
    if (ctrl$verbose) {
      message()
      msg <- cli::format_inline(
        " Generating a set of {nrow(x)} initial parameter results"
      )
      update_printer(ctrl, split_labels = NULL, task = msg, type = "go")
    }

    grid_ctrl <- ctrl
    grid_ctrl$verbose <- FALSE
    x <- tune_grid(
      wflow,
      resamples = resamples,
      grid = x,
      metrics = metrics,
      eval_time = eval_time,
      param_info = pset,
      control = parsnip::condense_control(grid_ctrl, control_grid())
    )

    if (ctrl$verbose) {
      update_printer(
        ctrl,
        split_labels = NULL,
        task = "Initialization complete",
        type = "success"
      )
      message()
    }
  } else {
    if (!inherits(x, "tune_results")) {
      cli::cli_abort(
        "{.arg initial} should be a positive integer or the results
                      of {.fn tune_grid}"
      )
    }
    if (ctrl$save_pred & !any(names(x) == ".predictions")) {
      cli::cli_abort(
        "{.arg save_pred} can only be used if the initial results
                      saved predictions."
      )
    }
    if (!is.null(ctrl$extract) & !any(names(x) == ".extracts")) {
      cli::cli_abort(
        "The {.fn extract} function can only be used if the initial
                     results have extractions."
      )
    }
    param_nms <- .get_tune_parameter_names(x)
    if (inherits(x, "tune_race")) {
      num_resamples <-
        x |>
        collect_metrics(summarize = FALSE) |>
        dplyr::count(.config)
      max_resamples <- max(num_resamples$n)
      configs <- num_resamples$.config[num_resamples$n == max_resamples]
      x <- filter_parameters(x, .config %in% configs)
      num_grid <- length(configs)
      x$.order <- NULL
    } else {
      num_grid <-
        collect_metrics(x) |>
        dplyr::distinct(!!!rlang::syms(param_nms)) |>
        nrow()
    }
    if (any(checks == "bayes")) {
      check_bayes_initial_size(
        length(param_nms),
        num_grid,
        race = inherits(x, "tune_race")
      )
    }
  }
  if (!any(names(x) == ".iter")) {
    x <- x |> dplyr::mutate(.iter = 0L)
  }
  x
}

# ------------------------------------------------------------------------------

check_class_or_null <- function(x, cls = "numeric") {
  inherits(x, cls) | is.null(x)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
#' @param cls `r lifecycle::badge("deprecated")` A character vector of
#'   possible classes
#' @param where `r lifecycle::badge("deprecated")` A character string for
#'   the calling function.
val_class_or_null <- function(x, cls = "numeric", where = NULL) {
  lifecycle::deprecate_warn("2.1.0", "val_class_or_null()")

  cl <- match.call()
  fine <- check_class_or_null(x, cls)
  cls <- paste(cls, collapse = " or ")
  if (!fine) {
    msg <- glue::glue("Argument '{deparse(cl$x)}' should be a {cls} or NULL")
    if (!is.null(where)) {
      msg <- glue::glue(msg, " in `{where}`")
    }
    rlang::abort(msg)
  }
  invisible(NULL)
}

check_class_and_single <- function(x, cls = "numeric") {
  isTRUE(inherits(x, cls) & length(x) == 1)
}

#' @export
#' @keywords internal
#' @rdname empty_ellipses
val_class_and_single <- function(x, cls = "numeric", where = NULL) {
  lifecycle::deprecate_warn("2.1.0", "val_class_and_single()")

  cl <- match.call()
  fine <- check_class_and_single(x, cls)
  cls <- paste(cls, collapse = " or ")
  if (!fine) {
    msg <- glue::glue(
      "Argument '{deparse(cl$x)}' should be a single {cls} value"
    )
    if (!is.null(where)) {
      msg <- glue::glue(msg, " in `{where}`")
    }
    rlang::abort(msg)
  }
  invisible(NULL)
}

# Check the data going into the GP. If there are all missing values, fail. If some
# are missing, remove them and send a warning. If all metrics are the same, fail.
check_gp_data <- function(x) {
  met <- x$.metric[[1]]

  miss_y <- sum(is.nan(x$mean))
  if (miss_y > 0) {
    if (miss_y == nrow(x)) {
      msg <- cli::pluralize(
        "All of the {met} estimates were missing. The Gaussian process model cannot be fit to the data."
      )
      message_wrap(
        msg,
        prefix = "!",
        color_text = get_tune_colors()$message$danger
      )
    } else {
      msg <- cli::pluralize(
        "For the {met} estimates, {miss_y} missing {?value was/values were} found and removed before fitting the Gaussian process model."
      )
      message_wrap(
        msg,
        prefix = "!",
        color_text = get_tune_colors()$message$warning
      )
    }

    x <- x[!is.na(x$mean), ]
  }

  n_uni <- length(unique(x$mean))
  if (n_uni == 1) {
    msg <- cli::pluralize(
      "All of the {met} values were identical. The Gaussian process model cannot
       be fit to the data. Try expanding the range of the tuning parameters."
    )
    message_wrap(
      msg,
      prefix = "!",
      color_text = get_tune_colors()$message$danger
    )
  }

  x
}

check_no_tuning <- function(x) {
  tune_param <- tune_args(x)
  num_param <- nrow(tune_param)
  if (num_param == 0) {
    return(invisible(FALSE))
  }
  srcs <- unique(tune_param$source)
  cli::cli_abort(c(
    "{num_param} argument{?s} {?has/have} been tagged for tuning in
     {?this component/these components}: {srcs}.",
    "i" = "Please use one of the tuning functions (e.g. {.fn tune_grid})
           to optimize them."
  ))
}

check_time_limit_arg <- function(x, call = rlang::caller_env()) {
  if (!inherits(x, c("logical", "numeric")) || length(x) != 1L) {
    cli::cli_abort(
      "{.arg time_limit} should be either a single numeric or
                    logical value.",
      call = call
    )
  }
  invisible(NULL)
}

test_that("rsample objects", {
  obj_cv <- rsample::vfold_cv(mtcars)
  obj_loo <- rsample::loo_cv(mtcars)
  obj_nst <- rsample::nested_cv(mtcars, obj_cv, inside = rsample::bootstraps())
  obj_permut <- rsample::permutations(mtcars, hp)
  expect_no_error(tune:::check_rset(obj_cv))
  expect_snapshot(error = TRUE, tune:::check_rset(obj_loo))
  expect_snapshot(error = TRUE, tune:::check_rset(obj_nst))
  expect_snapshot(error = TRUE, tune:::check_rset(obj_permut))
})

# ------------------------------------------------------------------------------

test_that("grid objects", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")
  skip_if_not_installed("kernlab")
  data("Chicago", package = "modeldata")

  spline_rec <-
    recipes::recipe(ridership ~ ., data = head(Chicago)) |>
    recipes::step_date(date) |>
    recipes::step_holiday(date) |>
    recipes::step_rm(date, dplyr::ends_with("away")) |>
    recipes::step_impute_knn(
      recipes::all_predictors(),
      neighbors = tune("imputation")
    ) |>
    recipes::step_other(recipes::all_nominal(), threshold = tune()) |>
    recipes::step_dummy(recipes::all_nominal()) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::step_spline_b(
      recipes::all_predictors(),
      deg_free = tune(),
      degree = tune()
    )

  glmn <- parsnip::linear_reg(penalty = tune(), mixture = tune()) |>
    parsnip::set_engine("glmnet")

  chi_wflow <-
    workflows::workflow() |>
    workflows::add_recipe(spline_rec) |>
    workflows::add_model(glmn)

  grid_1 <- tibble::tibble(
    penalty = 1:10,
    mixture = 1:10,
    imputation = 1:10,
    threshold = 1:10,
    deg_free = 1:10,
    degree = 1:10
  )

  set_1 <- extract_parameter_set_dials(chi_wflow)
  set_2 <- set_1 |> update(deg_free = dials::deg_free(c(1, 3)))

  expect_equal(tune:::.check_grid(grid_1, chi_wflow), grid_1)

  expect_silent(tune:::.check_grid(grid_1, chi_wflow))

  expect_snapshot(
    tune:::.check_grid(rbind(grid_1, grid_1), chi_wflow)
  )

  expect_snapshot(error = TRUE, {
    tune:::.check_grid(chi_wflow, chi_wflow)
  })

  bare_rec <-
    recipes::recipe(ridership ~ ., data = head(Chicago))

  svm_mod <-
    parsnip::svm_rbf(cost = tune()) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("regression")

  wflow_1 <-
    workflow() |>
    add_model(svm_mod) |>
    add_recipe(bare_rec)

  expect_no_error(grid_2 <- tune:::.check_grid(6, wflow_1))
  expect_equal(nrow(grid_2), 6)
  expect_true(inherits(grid_2, "data.frame"))

  # For issue #56
  grid_3 <- as.data.frame(grid_1)
  expect_equal(tune:::.check_grid(grid_3, chi_wflow), grid_1)

  # For weird attributes
  grid_4 <- expand.grid(
    penalty = 1:10,
    mixture = 12,
    imputation = 1:2,
    threshold = 1:2,
    deg_free = 2:3,
    degree = 9:10
  )
  expect_equal(
    tune:::.check_grid(grid_4, chi_wflow),
    tibble::as_tibble(vctrs::data_frame(grid_4))
  )

  expect_silent(tune:::.check_grid(grid_4, chi_wflow))
})

test_that("Unknown `grid` columns are caught", {
  skip_if_not_installed("splines2")

  data <- data.frame(x = 1:2, y = 1:2)

  rec <- recipes::recipe(y ~ x, data = data)
  rec <- recipes::step_spline_b(rec, x, deg_free = tune())
  rec <- recipes::step_pca(rec, x, num_comp = tune())

  model <- parsnip::linear_reg()
  model <- parsnip::set_engine(model, "lm")

  workflow <- workflow()
  workflow <- add_recipe(workflow, rec)
  workflow <- add_model(workflow, model)

  grid <- tibble::tibble(deg_free = 2, num_comp = 0.01, other1 = 1, other2 = 1)

  expect_snapshot(error = TRUE, {
    tune:::.check_grid(grid, workflow)
  })
})

test_that("Missing required `grid` columns are caught", {
  skip_if_not_installed("splines2")

  data <- data.frame(x = 1:2, y = 1:2)

  rec <- recipes::recipe(y ~ x, data = data)
  rec <- recipes::step_spline_b(rec, x, deg_free = tune())
  rec <- recipes::step_pca(rec, x, num_comp = tune())

  model <- parsnip::linear_reg()
  model <- parsnip::set_engine(model, "lm")

  workflow <- workflow()
  workflow <- add_recipe(workflow, rec)
  workflow <- add_model(workflow, model)

  grid <- tibble::tibble(num_comp = 0.01)

  expect_snapshot(error = TRUE, {
    tune:::.check_grid(grid, workflow)
  })
})

# ------------------------------------------------------------------------------

test_that("workflow objects", {
  skip_if_not_installed("xgboost")

  bare_rec <-
    recipes::recipe(ridership ~ ., data = head(Chicago))

  svm_mod <-
    parsnip::svm_rbf(cost = tune()) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("regression")

  wflow_1 <-
    workflow() |>
    add_model(svm_mod) |>
    add_recipe(bare_rec)

  expect_null(tune:::check_workflow(x = wflow_1))

  wflow_2 <-
    workflow() |>
    add_model(
      parsnip::boost_tree(mtry = tune()) |>
        parsnip::set_engine("xgboost") |>
        parsnip::set_mode("regression")
    ) |>
    add_recipe(bare_rec)

  expect_null(tune:::check_workflow(x = wflow_2))
  expect_snapshot(error = TRUE, {
    tune:::check_workflow(x = wflow_2, check_dials = TRUE)
  })

  glmn <- parsnip::linear_reg(penalty = tune(), mixture = tune()) |>
    parsnip::set_engine("glmnet")

  wflow_3 <-
    workflow() |>
    add_model(glmn)
  expect_snapshot(error = TRUE, {
    tune:::check_workflow(wflow_3)
  })

  wflow_4 <-
    workflow() |>
    add_recipe(bare_rec)
  expect_snapshot(error = TRUE, {
    tune:::check_workflow(wflow_4)
  })
})

test_that("errors informatively when needed package isn't installed", {
  # rstanarm is not installed during CI runs
  # in contexts where it _is_ installed, skip the test.
  skip_if(rlang::is_installed("rstanarm"))
  stan_wflow <- workflow(mpg ~ ., parsnip::linear_reg(engine = "stan"))

  expect_snapshot(
    check_workflow(stan_wflow),
    error = TRUE
  )

  expect_snapshot(
    fit_resamples(stan_wflow, rsample::bootstraps(mtcars)),
    error = TRUE
  )
})

test_that("workflow objects (will not tune, tidymodels/tune#548)", {
  skip_if_not_installed("glmnet")
  skip_if_not_installed("splines2")

  # one recipe without tuning, one with:
  rec_bare <- recipes::recipe(ridership ~ ., data = head(Chicago, 30))
  rec_tune <- rec_bare |>
    recipes::step_spline_natural(temp_max, deg_free = tune())

  # well-defined:
  lr_lm_0 <- parsnip::linear_reg()

  # not well-defined:
  lr_lm_1 <- parsnip::linear_reg(penalty = tune())
  lr_lm_2 <- parsnip::linear_reg(penalty = tune(), mixture = tune())

  # well-defined:
  lr_glmnet_0 <- lr_lm_0 |> parsnip::set_engine("glmnet")
  lr_glmnet_1 <- lr_lm_1 |> parsnip::set_engine("glmnet")
  lr_glmnet_2 <- lr_lm_2 |> parsnip::set_engine("glmnet")

  # don't error when supplied tune args make sense given engine / steps
  expect_no_error(check_workflow(workflow(rec_bare, lr_lm_0)))
  expect_no_error(check_workflow(workflow(rec_bare, lr_glmnet_0)))
  expect_no_error(check_workflow(workflow(rec_bare, lr_glmnet_1)))
  expect_no_error(check_workflow(workflow(rec_bare, lr_glmnet_2)))

  expect_no_error(check_workflow(workflow(rec_tune, lr_lm_0)))
  expect_no_error(check_workflow(workflow(rec_tune, lr_glmnet_0)))
  expect_no_error(check_workflow(workflow(rec_tune, lr_glmnet_1)))
  expect_no_error(check_workflow(workflow(rec_tune, lr_glmnet_2)))

  # error when supplied tune args don't make sense given engine / steps
  expect_error_nt <- function(x) {
    testthat::expect_error(x, class = "not_tunable_error")
  }

  expect_error_nt(check_workflow(workflow(rec_bare, lr_lm_1)))
  expect_error_nt(check_workflow(workflow(rec_bare, lr_lm_2)))

  expect_error_nt(check_workflow(workflow(rec_tune, lr_lm_1)))
  expect_error_nt(check_workflow(workflow(rec_tune, lr_lm_2)))

  # ensure that error (incl `call`) is displayed as it ought to be:
  expect_snapshot(
    error = TRUE,
    tune_grid(
      lr_lm_1,
      rec_bare,
      rsample::bootstraps(Chicago, 2)
    )
  )
  expect_snapshot(
    error = TRUE,
    tune_bayes(
      lr_lm_2,
      rec_tune,
      rsample::bootstraps(Chicago, 2)
    )
  )
})

# ------------------------------------------------------------------------------

test_that("yardstick objects", {
  skip_if_not_installed("splines2")
  withr::local_options(lifecycle_verbosity = "quiet")

  spline_rec <-
    recipes::recipe(ridership ~ ., data = head(Chicago)) |>
    recipes::step_date(date) |>
    recipes::step_holiday(date) |>
    recipes::step_rm(date, dplyr::ends_with("away")) |>
    recipes::step_impute_knn(
      recipes::all_predictors(),
      neighbors = tune("imputation")
    ) |>
    recipes::step_other(recipes::all_nominal(), threshold = tune()) |>
    recipes::step_dummy(recipes::all_nominal()) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::step_spline_b(
      recipes::all_predictors(),
      deg_free = tune(),
      degree = tune()
    )

  glmn <- parsnip::linear_reg(penalty = tune(), mixture = tune()) |>
    parsnip::set_engine("glmnet")

  chi_wflow <-
    workflows::workflow() |>
    workflows::add_recipe(spline_rec) |>
    workflows::add_model(glmn)

  metrics_1 <- tune:::check_metrics(NULL, chi_wflow)
  metrics_2 <- yardstick::metric_set(yardstick:::rmse)
  expect_true(inherits(metrics_1, "numeric_metric_set"))
  expect_snapshot(error = TRUE, {
    tune:::check_metrics(yardstick::rmse, chi_wflow)
  })
  expect_true(inherits(
    tune:::check_metrics(metrics_2, chi_wflow),
    "numeric_metric_set"
  ))
})

test_that("metrics must match the parsnip engine", {
  withr::local_options(lifecycle_verbosity = "quiet")
  metric_set1 <- yardstick::metric_set(yardstick::accuracy)
  metric_set2 <- yardstick::metric_set(yardstick::rmse)

  mod1 <- parsnip::rand_forest(mode = "regression")
  mod2 <- parsnip::rand_forest(mode = "classification")

  workflow1 <- add_model(workflow(), mod1)
  workflow2 <- add_model(workflow(), mod2)

  expect_snapshot(error = TRUE, {
    tune:::check_metrics(metric_set1, workflow1)
  })

  expect_snapshot(error = TRUE, {
    tune:::check_metrics(metric_set2, workflow2)
  })
})

test_that("check_metrics() is deprecated", {
  mod <- parsnip::rand_forest(mode = "regression")
  wflow <- add_model(workflow(), mod)
  expect_snapshot(metric <- check_metrics(NULL, wflow))
})

# ------------------------------------------------------------------------------

test_that("grid control objects", {
  expect_no_error(control_grid())
  expect_snapshot(error = TRUE, control_grid(tomato = 1))
  expect_snapshot(error = TRUE, control_grid(verbose = 1))
  expect_snapshot(error = TRUE, control_grid(verbose = rep(TRUE, 2)))
  expect_snapshot(error = TRUE, control_grid(allow_par = 1))
  expect_snapshot(error = TRUE, control_grid(save_pred = "no"))
  expect_snapshot(error = TRUE, control_grid(extract = Inf))
  expect_snapshot(error = TRUE, control_grid(pkgs = Inf))

  expect_no_error(control_grid(verbose = TRUE))
  expect_no_error(control_grid(allow_par = FALSE))
  expect_no_error(control_grid(save_pred = TRUE))
  expect_no_error(control_grid(extract = NULL))
  expect_no_error(control_grid(extract = I))
  expect_no_error(control_grid(pkgs = NULL))
  expect_no_error(control_grid(pkgs = letters))
  expect_s3_class(control_grid(), c("control_grid", "control_resamples"))
})

test_that("Bayes control objects", {
  expect_no_error(control_bayes())
  expect_snapshot(error = TRUE, control_bayes(tomato = 1))
  expect_snapshot(error = TRUE, control_bayes(verbose = 1))
  expect_snapshot(error = TRUE, control_bayes(verbose = rep(TRUE, 2)))
  expect_snapshot(error = TRUE, control_bayes(no_improve = FALSE))
  expect_snapshot(error = TRUE, control_bayes(uncertain = FALSE))
  expect_snapshot(error = TRUE, control_bayes(seed = FALSE))
  expect_snapshot(error = TRUE, control_bayes(save_pred = "no"))
  expect_snapshot(error = TRUE, control_bayes(extract = Inf))
  expect_snapshot(error = TRUE, control_bayes(pkgs = Inf))
  expect_snapshot(error = TRUE, control_bayes(time_limit = "a"))

  expect_snapshot(
    tmp <- control_bayes(no_improve = 2, uncertain = 5)
  )

  expect_no_error(control_bayes(verbose = TRUE))
  expect_no_error(control_bayes(no_improve = 2))
  expect_no_error(control_bayes(uncertain = 2))
  expect_no_error(control_bayes(save_pred = TRUE))
  expect_no_error(control_bayes(extract = NULL))
  expect_no_error(control_bayes(extract = I))
  expect_no_error(control_bayes(pkgs = NULL))
  expect_no_error(control_bayes(pkgs = letters))
  expect_no_error(control_bayes(time_limit = 2))
  expect_s3_class(control_bayes(), "control_bayes")
})

# ------------------------------------------------------------------------------

test_that("initial values", {
  skip_if_not_installed("kernlab")

  svm_mod <-
    parsnip::svm_rbf(cost = tune()) |>
    parsnip::set_engine("kernlab") |>
    parsnip::set_mode("regression")
  wflow_1 <-
    workflow() |>
    add_model(svm_mod) |>
    add_recipe(recipes::recipe(mpg ~ ., data = mtcars))
  mtfolds <- rsample::vfold_cv(mtcars)

  grid_1 <- tune:::check_initial(
    2,
    pset = extract_parameter_set_dials(wflow_1),
    wflow = wflow_1,
    resamples = mtfolds,
    metrics = yardstick::metric_set(yardstick::rsq),
    eval_time = NULL,
    ctrl = control_bayes()
  )
  expect_true(is.data.frame(grid_1))
  expect_equal(nrow(grid_1), nrow(mtfolds))
  expect_true(all(purrr::map_lgl(grid_1$.metrics, \(.x) nrow(.x) == 2)))

  expect_snapshot(error = TRUE, {
    tune:::check_initial(
      data.frame(),
      pset = extract_parameter_set_dials(wflow_1),
      wflow = wflow_1,
      resamples = mtfolds,
      metrics = yardstick::metric_set(yardstick::rsq),
      ctrl = control_bayes()
    )
  })
})

# ------------------------------------------------------------------------------

test_that("val_class_or_null() is deprecated", {
  expect_snapshot(res <- val_class_or_null("a", "character"))
})

test_that("val_class_and_single() is deprecated", {
  expect_snapshot(res <- val_class_and_single("a", "character"))
})

test_that("validation helpers", {
  expect_true(tune:::check_class_or_null("a", "character"))
  expect_true(tune:::check_class_or_null(letters, "character"))
  expect_true(tune:::check_class_or_null(NULL, "character"))
  expect_false(tune:::check_class_or_null(NA, "character"))

  expect_true(tune:::check_class_and_single("a", "character"))
  expect_false(tune:::check_class_and_single(letters, "character"))
  expect_false(tune:::check_class_and_single(NULL, "character"))
  expect_false(tune:::check_class_and_single(NA, "character"))
})

# ------------------------------------------------------------------------------

test_that("check parameter finalization", {
  skip_if_not_installed("splines2")

  rec <-
    recipes::recipe(mpg ~ ., data = mtcars) |>
    recipes::step_spline_natural(disp, deg_free = 3)
  rec_tune <- rec |>
    recipes::step_pca(recipes::all_predictors(), num_comp = tune())
  f <- mpg ~ .
  rf1 <-
    parsnip::rand_forest(mtry = tune(), min_n = tune()) |>
    parsnip::set_engine("ranger") |>
    parsnip::set_mode("regression")
  lm1 <-
    parsnip::linear_reg(penalty = tune()) |>
    parsnip::set_engine("glmnet")

  w1 <-
    workflow() |>
    add_formula(f) |>
    add_model(rf1)

  expect_snapshot(
    expect_no_error(
      p1 <- tune:::check_parameters(
        w1,
        data = mtcars,
        grid_names = character(0)
      )
    )
  )
  expect_false(any(dials::has_unknowns(p1$object)))

  expect_no_error(
    p1 <- tune:::check_parameters(w1, data = mtcars, grid_names = "mtry")
  )

  w2 <-
    workflow() |>
    add_recipe(rec) |>
    add_model(rf1)

  expect_snapshot(
    expect_no_error(
      p2 <- tune:::check_parameters(w2, data = mtcars)
    )
  )
  expect_false(any(dials::has_unknowns(p2$object)))

  w3 <-
    workflow() |>
    add_recipe(rec) |>
    add_model(rf1)
  p3 <- extract_parameter_set_dials(w3)

  expect_snapshot(
    expect_no_error(
      p3_a <- tune:::check_parameters(w3, data = mtcars)
    )
  )
  expect_false(any(dials::has_unknowns(p3_a$object)))

  w4 <-
    workflow() |>
    add_recipe(rec_tune) |>
    add_model(rf1)

  expect_snapshot(error = TRUE, {
    tune:::check_parameters(w4, data = mtcars)
  })

  p4_a <-
    extract_parameter_set_dials(w4) |>
    update(mtry = dials::mtry(c(1, 10)))

  expect_no_error(
    p4_b <- tune:::check_parameters(w4, p4_a, data = mtcars)
  )
  expect_true(inherits(p4_b, "parameters"))

  w5 <-
    workflow() |>
    add_recipe(rec_tune) |>
    add_model(lm1)

  expect_no_error(
    p5 <- tune:::check_parameters(w5, data = mtcars)
  )
  expect_true(inherits(p5, "parameters"))
})

test_that("check fold weights", {
  folds <- rsample::vfold_cv(mtcars, v = 3)

  # No weights should pass silently
  expect_no_error(tune:::check_resample_weights(folds))

  # Valid weights should pass
  weights <- c(0.1, 0.5, 0.4)
  weighted_folds <- add_resample_weights(folds, weights)
  expect_no_error(tune:::check_resample_weights(weighted_folds))

  # Invalid weights should error
  expect_snapshot(
    add_resample_weights(folds, c("a", "b", "c")),
    error = TRUE
  )

  expect_snapshot(
    add_resample_weights(folds, c(0.5, 0.3)),
    error = TRUE
  )

  expect_snapshot(
    add_resample_weights(folds, c(-0.1, 0.5, 0.6)),
    error = TRUE
  )
})

test_that("fold weights integration test", {
  # Create simple data and resamples
  set.seed(1234)
  data_small <- mtcars[1:20, ]
  folds <- rsample::vfold_cv(data_small, v = 3)

  # Create simple model and recipe
  simple_rec <- recipes::recipe(mpg ~ wt + hp, data = data_small)
  simple_mod <- parsnip::linear_reg() |> parsnip::set_engine("lm")
  simple_wflow <- workflows::workflow() |>
    workflows::add_recipe(simple_rec) |>
    workflows::add_model(simple_mod)

  # Test with equal weights (should match unweighted results)
  equal_weights <- c(1, 1, 1)
  weighted_folds_equal <- add_resample_weights(folds, equal_weights)

  # Fit both weighted and unweighted
  unweighted_results <- fit_resamples(
    simple_wflow,
    folds,
    control = control_resamples(save_pred = FALSE)
  )
  weighted_results_equal <- fit_resamples(
    simple_wflow,
    weighted_folds_equal,
    control = control_resamples(save_pred = FALSE)
  )

  # Extract metrics
  unweighted_metrics <- collect_metrics(unweighted_results)
  weighted_metrics_equal <- collect_metrics(weighted_results_equal)

  # Should be nearly identical (allowing for small numerical differences)
  expect_equal(
    unweighted_metrics$mean,
    weighted_metrics_equal$mean,
    tolerance = 1e-10
  )

  # Test with unequal weights
  unequal_weights <- c(0.1, 0.3, 0.6) # Higher weight on last fold
  weighted_folds_unequal <- add_resample_weights(folds, unequal_weights)

  weighted_results_unequal <- fit_resamples(
    simple_wflow,
    weighted_folds_unequal,
    control = control_resamples(save_pred = FALSE)
  )
  weighted_metrics_unequal <- collect_metrics(weighted_results_unequal)

  # Should be different from unweighted results
  expect_false(all(
    abs(unweighted_metrics$mean - weighted_metrics_unequal$mean) < 1e-10
  ))

  # Verify that weights are properly stored and retrieved
  expect_equal(
    attr(weighted_folds_unequal, ".resample_weights"),
    unequal_weights
  )

  # Test fold size calculation
  calculated_weights <- calculate_resample_weights(folds)
  expect_length(calculated_weights, nrow(folds))
  expect_true(all(calculated_weights > 0))
  expect_equal(sum(calculated_weights), 1) # Should sum to 1 now
})

test_that("fold weights with tune_grid", {
  skip_if_not_installed("kernlab")

  # Create simple tuning scenario
  set.seed(5678)
  data_small <- mtcars[1:15, ]
  folds <- rsample::vfold_cv(data_small, v = 3)

  # Create tunable workflow
  tune_rec <- recipes::recipe(mpg ~ wt + hp, data = data_small) |>
    recipes::step_normalize(recipes::all_predictors())
  tune_mod <- parsnip::svm_rbf(
    cost = tune(),
    rbf_sigma = 0.001,
    mode = "regression"
  )

  tune_wflow <- workflows::workflow() |>
    workflows::add_recipe(tune_rec) |>
    workflows::add_model(tune_mod)

  # Create simple grid
  simple_grid <- tibble::tibble(cost = c(1, 10, 100))

  # Test with unequal weights
  weights <- c(0.2, 0.3, 0.5)
  weighted_folds <- add_resample_weights(folds, weights)

  # Tune with weights
  weighted_tune_results <- tune_grid(
    tune_wflow,
    weighted_folds,
    grid = simple_grid,
    control = control_grid(save_pred = FALSE)
  )

  # Verify results structure
  expect_s3_class(weighted_tune_results, "tune_results")

  # Extract metrics and verify they're computed
  weighted_metrics <- collect_metrics(weighted_tune_results)
  expect_true(nrow(weighted_metrics) > 0)
  expect_true(all(c("mean", "std_err") %in% names(weighted_metrics)))

  # Compare with unweighted results
  unweighted_tune_results <- tune_grid(
    tune_wflow,
    folds,
    grid = simple_grid,
    control = control_grid(save_pred = FALSE)
  )
  unweighted_metrics <- collect_metrics(unweighted_tune_results)

  # Results should differ due to weighting
  expect_false(all(
    abs(weighted_metrics$mean - unweighted_metrics$mean) < 1e-10
  ))
})

# ------------------------------------------------------------------------------

expect_s3_class_tune_results <- function(x) {
  expect_s3_class(x, "tune_results")
}

expect_s3_class_bare_tibble <- function(x) {
  expect_s3_class(x, c("tbl_df", "tbl", "data.frame"), exact = TRUE)
}

# ------------------------------------------------------------------------------
# Test object used when testing vctrs/dplyr compatibility.
# Delayed assign to only load it in once.

load_helper_tune_results <- function() {
  list(
    resample = readRDS(test_path("data", "lm_resamples.rds")),
    tune = readRDS(test_path("data", "knn_results.rds")),
    bayes = readRDS(test_path("data", "lm_bayes.rds"))
  )
}

delayedAssign("helper_tune_results", load_helper_tune_results())

# nocov start

# For use in setting the `tune_symbol` active binding in `.onLoad()`

tune_symbol_utf8 <- list(
  "success" = "\u2713"
)

tune_symbol_windows <- list(
  "success" = "\u221A"
)

tune_symbol_ascii <- list(
  "success" = "v"
)

## -----------------------------------------------------------------------------

# For use in setting the `tune_color` active binding in `.onLoad()`

tune_color_dark <- list(
  symbol = list(
    "warning" = cli::col_yellow,
    "go" = cli::col_white,
    "danger" = cli::col_red,
    "success" = cli::col_green,
    "info" = cli::col_blue
  ),
  message = list(
    "warning" = cli::col_yellow,
    "go" = cli::col_white,
    "danger" = cli::col_red,
    "success" = cli::col_white,
    "info" = cli::col_white
  )
)

tune_color_light <- list(
  symbol = list(
    "warning" = cli::col_yellow,
    "go" = cli::col_black,
    "danger" = cli::col_red,
    "success" = cli::col_green,
    "info" = cli::col_blue
  ),
  message = list(
    "warning" = cli::col_yellow,
    "go" = cli::col_black,
    "danger" = cli::col_red,
    "success" = cli::col_black,
    "info" = cli::col_black
  )
)

# ------------------------------------------------------------------------------

# cli:::is_latex_output()
is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) {
    return(FALSE)
  }

  get("is_latex_output", asNamespace("knitr"))()
}

# cli:::is_windows()
is_windows <- function() {
  .Platform$OS.type == "windows"
}

# nocov end

#' Get colors for tune text.
#'
#' @keywords internal
#' @export
#' @rdname empty_ellipses
get_tune_colors <- function() tune_color

#' @export
print.tune_results <- function(x, ...) {
  cl <- match.call()
  if (inherits(x, "resample_results")) {
    cat("# Resampling results\n")
  } else {
    cat("# Tuning results\n")
  }

  att <- attributes(x)
  rset_info <- att$rset_info

  if (is.null(rset_info)) {
    print_compat_tune_results_label(x)
  } else {
    cat("#", rset_info$label, "\n")
  }

  print(tibble::as_tibble(x), ...)

  summarize_notes(x)
}

# `tune_results` have been changed to no longer inherit from `rset`,
# and should instead use the `rset_info` attribute. This code
# ensures that printing still works on old versions of `tune_results`
# that might have been saved to disk and then loaded back up with
# a new version of tune.
print_compat_tune_results_label <- function(x) {
  # Somehow we don't have the `rset_info` attribute, but this isn't
  # an rset subclass. Just don't print a label to avoid erroring.
  if (!inherits(x, "rset")) {
    return()
  }

  label <- try(pretty(x), silent = TRUE)

  # Somehow the rset `pretty()` method failed.
  # Just don't print a label to avoid erroring.
  if (inherits(label, "try-error")) {
    return()
  }

  cat("#", label, "\n")
}


has_notes <- function(x) {
  if (is.null(x)) {
    return(0L)
  }
  nrow(x)
}

summarize_notes <- function(x) {
  num_notes <- sum(purrr::map_int(x$.notes, has_notes))
  if (num_notes == 0) {
    return(invisible(NULL))
  }
  notes <-
    x |>
    dplyr::select(dplyr::starts_with("id"), .notes) |>
    tidyr::unnest(cols = .notes)
  by_type <-
    notes |>
    dplyr::group_nest(type) |>
    dplyr::mutate(data = purrr::map(data, \(.x) dplyr::count(.x, note))) |>
    tidyr::unnest(data) |>
    dplyr::rowwise() |>
    dplyr::mutate(
      note = gsub("(Error:)", "", note),
      note = glue::glue_collapse(note, width = 0.85 * getOption("width")),
      note = gsub("\n", " ", note, fixed = TRUE),
      pre = ifelse(type == "error", "  - Error(s) x", "  - Warning(s) x"),
      note = paste0(pre, n, ": ", note)
    )
  cat("\nThere were issues with some computations:\n\n")
  cat(by_type$note, sep = "\n")
  cat("\nRun `show_notes(.Last.tune.result)` for more information.\n")
  invisible(NULL)
}


# ------------------------------------------------------------------------------

#' @export
`[.tune_results` <- function(x, i, j, ...) {
  out <- NextMethod()
  tune_results_reconstruct(out, x)
}

#' @export
`names<-.tune_results` <- function(x, value) {
  out <- NextMethod()
  tune_results_reconstruct(out, x)
}

# ------------------------------------------------------------------------------

new_tune_results <-
  function(
    x,
    parameters,
    metrics,
    eval_time,
    eval_time_target,
    outcomes = character(0),
    rset_info,
    ...,
    class = character()
  ) {
    if (any(names(x) == ".seeds")) {
      x$.seeds <- NULL
    }

    res <-
      new_bare_tibble(
        x = x,
        parameters = parameters,
        metrics = metrics,
        eval_time = eval_time,
        eval_time_target = eval_time_target,
        outcomes = outcomes,
        rset_info = rset_info,
        ...,
        class = c(class, "tune_results")
      )
    attr(res, "outcomes") <- outcomes
    res
  }

is_tune_results <- function(x) {
  inherits(x, "tune_results")
}

peek_tune_results_outcomes <- function(x) {
  if (!is_tune_results(x)) {
    cli::cli_abort(
      "Internal error: {.arg outcomes} can only be extracted from
                   {.cls tune_results}."
    )
  }

  out <- attr(x, "outcomes", exact = TRUE)

  if (is.null(out)) {
    cli::cli_abort(
      "The object of type {.cls tune_results} doesn't have an
                    {.code outcomes} attribute."
    )
  }

  out
}

# ------------------------------------------------------------------------------

#' Display distinct errors from tune objects
#' @param x An object of class `tune_results`.
#' @param n An integer for how many unique notes to show.
#' @return Invisibly, `x`. Function is called for side-effects and printing.
#' @export
show_notes <- function(x, n = 10) {
  res <-
    collect_notes(x) |>
    dplyr::distinct(type, note)

  if (nrow(res) == 0) {
    cat("Great job! No notes to show.\n")
    return(invisible(x))
  }

  n <- min(nrow(res), n)
  notes <- res$note[1:n]

  msg <- "unique notes:\n"
  if (n != nrow(res)) {
    msg <- paste0("first ", n, msg)
  }

  sub_notes <- strsplit(notes, split = "\n")[[1]]
  max_width <- max(purrr::map_int(sub_notes, nchar))
  max_width <- min(max_width, cli::console_width())

  notes <- paste(cli::rule(width = max_width), notes, sep = "\n")
  notes <- paste0(notes, "\n")
  cat(msg)
  cat(notes, sep = "")
  invisible(x)
}

library(workflows)
library(dplyr)
library(recipes)
library(parsnip)
library(rsample)
library(yardstick)

test_that("showing notes", {
  skip_if_not_installed("modeldata")
  skip_if(!rankdeficient_version)

  data(Chicago, package = "modeldata")

  base_wflow <-
    workflow() |>
    add_model(linear_reg())

  role_rec <-
    recipe(ridership ~ ., data = Chicago) |>
    step_date(date, id = "step_date") |>
    update_role(date, new_role = "date") |>
    update_role_requirements("date", bake = FALSE)

  role_bp_wflow <-
    base_wflow |>
    add_recipe(role_rec)

  set.seed(1)
  rs <- vfold_cv(Chicago)

  skip_if(packageVersion("dplyr") < "1.1.1")
  expect_snapshot(res_roles <- role_bp_wflow |> fit_resamples(rs))
  expect_snapshot(show_notes(res_roles))

  # example with warnings

  simple_rec <-
    recipe(ridership ~ ., data = Chicago) |>
    step_holiday(date) |>
    step_date(date, keep_original_cols = FALSE) |>
    step_dummy(all_nominal_predictors(), one_hot = TRUE)

  simple_wflow <-
    base_wflow |>
    add_recipe(simple_rec)

  expect_snapshot(res_simple <- simple_wflow |> fit_resamples(rs))
  expect_snapshot(show_notes(res_simple))

  # nothing to show

  clean_rec <-
    recipe(ridership ~ Austin + Belmont, data = Chicago)

  clean_wflow <-
    base_wflow |>
    add_recipe(clean_rec)

  res_clean <- clean_wflow |> fit_resamples(rs)
  expect_snapshot(show_notes(.Last.tune.result))

  # Get cli lines right
  set.seed(1)
  dat <- modeldata::sim_classification(150, intercept = 15)
  rs <- rsample::vfold_cv(dat)
  expect_snapshot(
    fit_lr <-
      parsnip::logistic_reg() |>
      fit_resamples(class ~ ., rs)
  )
  expect_snapshot(show_notes(fit_lr))
})

test_that("notes are sorted in the correct order", {
  # set `apparent = TRUE` so that resamples aren't in alphabetical order by id
  mt_boots <- bootstraps(mtcars, 3, apparent = TRUE)

  # induce the size zero yardstick error in Bootstrap1
  mt_boots$splits[[1]]$out_id <- integer(0)

  suppressMessages({
    mt_res <- fit_resamples(linear_reg(), mpg ~ ., mt_boots)
  })

  boots_1_loc <- which(mt_res$id == "Bootstrap1")
  boots_1_notes <- mt_res$.notes[[boots_1_loc]]
  expect_equal(nrow(boots_1_notes), 1)
})

#' Model tuning via grid search
#'
#' [tune_grid()] computes a set of performance metrics (e.g. accuracy or RMSE)
#'  for a pre-defined set of tuning parameters that correspond to a model or
#'  recipe across one or more resamples of the data.
#'
#' @inheritParams last_fit
#' @inheritParams fit_resamples
#' @param param_info A [dials::parameters()] object or `NULL`. If none is given,
#' a parameters set is derived from other arguments. Passing this argument can
#' be useful when parameter ranges need to be customized.
#' @param grid A data frame of tuning combinations or a positive integer. The
#'  data frame should have columns for each parameter being tuned and rows for
#'  tuning parameter candidates. An integer denotes the number of candidate
#'  parameter sets to be created automatically.
#' @param control An object used to modify the tuning process, likely created
#' by [control_grid()].
#' @param eval_time A numeric vector of time points where dynamic event time
#' metrics should be computed (e.g. the time-dependent ROC curve, etc). The
#' values must be non-negative and should probably be no greater than the
#' largest event time in the training set (See Details below).
#' @param ... Not currently used.
#' @return An updated version of `resamples` with extra list columns for `.metrics` and
#' `.notes` (optional columns are `.predictions` and `.extracts`). `.notes`
#' contains warnings and errors that occur during execution.
#' @seealso [control_grid()], [tune()], [fit_resamples()],
#' [autoplot.tune_results()], [show_best()], [select_best()],
#' [collect_predictions()], [collect_metrics()]
#' @details
#'
#' Suppose there are _m_ tuning parameter combinations. [tune_grid()] may not
#' require all _m_ model/recipe fits across each resample. For example:
#'
#' \itemize{
#'   \item In cases where a single model fit can be used to make predictions
#'         for different parameter values in the grid, only one fit is used.
#'         For example, for some boosted trees, if 100 iterations of boosting
#'         are requested, the model object for 100 iterations can be used to
#'         make predictions on iterations less than 100 (if all other
#'         parameters are equal).
#'   \item When the model is being tuned in conjunction with pre-processing
#'         and/or post-processing parameters, the minimum number of fits are
#'         used. For example, if the number of PCA components in a recipe step
#'         are being tuned over three values (along with model tuning
#'         parameters), only three recipes are trained. The alternative
#'         would be to re-train the same recipe multiple times for each model
#'         tuning parameter.
#' }
#'
#' tune supports parallel processing with the \pkg{future} package. To execute
#' the resampling iterations in parallel, specify a [plan][future::plan] with
#' future first. The `allow_par` argument can be used to avoid parallelism.
#'
#' For the most part, warnings generated during training are shown as they occur
#' and are associated with a specific resample when
#' `control_grid(verbose = TRUE)`. They are (usually) not aggregated until the
#' end of processing.
#'
#' @section Parameter Grids:
#'
#' If no tuning grid is provided, a grid (via [dials::grid_space_filling()]) is
#' created with 10 candidate parameter combinations.
#'
#' When provided, the grid should have column names for each parameter and
#'  these should be named by the parameter name or `id`. For example, if a
#'  parameter is marked for optimization using `penalty = tune()`, there should
#'  be a column named `penalty`. If the optional identifier is used, such as
#'  `penalty = tune(id = 'lambda')`, then the corresponding column name should
#'  be `lambda`.
#'
#' In some cases, the tuning parameter values depend on the dimensions of the
#'  data. For example, `mtry` in random forest models depends on the number of
#'  predictors. In this case, the default tuning parameter object requires an
#'  upper range. [dials::finalize()] can be used to derive the data-dependent
#'  parameters. Otherwise, a parameter set can be created (via
#'  [dials::parameters()]) and the `dials` `update()` function can be used to
#'  change the values. This updated parameter set can be passed to the function
#'  via the `param_info` argument.
#'
#'  The rows of the grid are called tuning parameter _candidates_. Each
#'  candidate has a unique `.config` value that, for grid search, has the
#'  pattern `pre{num}_mod{num}_post{num}`. The numbers include a zero when that
#'  element was static. For example, a value of `pre0_mod3_post4` means no
#'  preprocessors were tuned and the model and postprocessor(s) had at least
#'  three and four candidates, respectively. Also, the numbers are zero-padded
#'  to enable proper sorting.
#'
#' @section Performance Metrics:
#'
#' To use your own performance metrics, the [yardstick::metric_set()] function
#'  can be used to pick what should be measured for each model. If multiple
#'  metrics are desired, they can be bundled. For example, to estimate the area
#'  under the ROC curve as well as the sensitivity and specificity (under the
#'  typical probability cutoff of 0.50), the `metrics` argument could be given:
#'
#' \preformatted{
#'   metrics = metric_set(roc_auc, sens, spec)
#' }
#'
#' Each metric is calculated for each candidate model.
#'
#' If no metric set is provided, one is created:
#' \itemize{
#'   \item For regression models, the root mean squared error and coefficient
#'         of determination are computed.
#'   \item For classification, the area under the ROC curve and overall accuracy
#'         are computed.
#'   \item For censored regression, the dynamic Brier score
#'         ([yardstick::brier_survival()]) is used.
#'   \item For quantile regression, the weighted interval score
#'         ([yardstick::weighted_interval_score()]) is used.
#' }
#'
#' Note that the metrics also determine what type of predictions are estimated
#' during tuning. For example, in a classification problem, if metrics are used
#' that are all associated with hard class predictions, the classification
#' probabilities are not created.
#'
#' The out-of-sample estimates of these metrics are contained in a list column
#' called `.metrics`. This tibble contains a row for each metric and columns
#' for the value, the estimator type, and so on.
#'
#' [collect_metrics()] can be used for these objects to collapse the results
#' over the resampled (to obtain the final resampling estimates per tuning
#' parameter combination).
#'
#' @section Obtaining Predictions:
#'
#' When `control_grid(save_pred = TRUE)`, the output tibble contains a list
#' column called `.predictions` that has the out-of-sample predictions for each
#' parameter combination in the grid and each fold (which can be very large).
#'
#' The elements of the tibble are tibbles with columns for the tuning
#' parameters, the row number from the original data object (`.row`), the
#' outcome data (with the same name(s) of the original data), and any columns
#' created by the predictions. For example, for simple regression problems, this
#' function generates a column called `.pred` and so on. As noted above, the
#' prediction columns that are returned are determined by the type of metric(s)
#' requested.
#'
#' This list column can be `unnested` using [tidyr::unnest()] or using the
#'  convenience function [collect_predictions()].
#'
#' @section Extracting Information:
#'
#' The `extract` control option will result in an additional function to be
#'  returned called `.extracts`. This is a list column that has tibbles
#'  containing the results of the user's function for each tuning parameter
#'  combination. This can enable returning each model and/or recipe object that
#'  is created during resampling. Note that this could result in a large return
#'  object, depending on what is returned.
#'
#' The control function contains an option (`extract`) that can be used to
#'  retain any model or recipe that was created within the resamples. This
#'  argument should be a function with a single argument. The value of the
#'  argument that is given to the function in each resample is a workflow
#'  object (see [workflows::workflow()] for more information). Several
#'  helper functions can be used to easily pull out the preprocessing
#'  and/or model information from the workflow, such as
#'  [`extract_preprocessor()`][workflows::extract_preprocessor.workflow()] and
#'  [`extract_fit_parsnip()`][workflows::extract_fit_parsnip.workflow()].
#'
#' As an example, if there is interest in getting each parsnip model fit back,
#' one could use:
#' \preformatted{
#'   extract = function (x) extract_fit_parsnip(x)
#' }
#'
#' Note that the function given to the `extract` argument is evaluated on
#'  every model that is _fit_ (as opposed to every model that is _evaluated_).
#' As noted above, in some cases, model predictions can be derived for
#'  sub-models so that, in these cases, not every row in the tuning parameter
#'  grid has a separate R object associated with it.
#'
#' Finally, it is a good idea to include calls to [require()] for packages that
#' are used in the function. This helps prevent failures when using parallel
#' processing.
#'
#' @template case-weights
#' @template censored-regression
#'
#' @examplesIf tune:::should_run_examples(suggests = "kernlab") & rlang::is_installed("splines2") && !tune:::is_cran_check()
#' library(recipes)
#' library(rsample)
#' library(parsnip)
#' library(workflows)
#' library(ggplot2)
#'
#' # ---------------------------------------------------------------------------
#'
#' set.seed(6735)
#' folds <- vfold_cv(mtcars, v = 5)
#'
#' # ---------------------------------------------------------------------------
#'
#' # tuning recipe parameters:
#'
#' spline_rec <-
#'   recipe(mpg ~ ., data = mtcars) |>
#'   step_spline_natural(disp, deg_free = tune("disp")) |>
#'   step_spline_natural(wt, deg_free = tune("wt"))
#'
#' lin_mod <-
#'   linear_reg() |>
#'   set_engine("lm")
#'
#' # manually create a grid
#' spline_grid <- expand.grid(disp = 2:5, wt = 2:5)
#'
#' # Warnings will occur from making spline terms on the holdout data that are
#' # extrapolations.
#' spline_res <-
#'   tune_grid(lin_mod, spline_rec, resamples = folds, grid = spline_grid)
#' spline_res
#'
#'
#' show_best(spline_res, metric = "rmse")
#'
#' # ---------------------------------------------------------------------------
#'
#' # tune model parameters only (example requires the `kernlab` package)
#'
#' car_rec <-
#'   recipe(mpg ~ ., data = mtcars) |>
#'   step_normalize(all_predictors())
#'
#' svm_mod <-
#'   svm_rbf(cost = tune(), rbf_sigma = tune()) |>
#'   set_engine("kernlab") |>
#'   set_mode("regression")
#'
#' # Use a space-filling design with 7 points
#' set.seed(3254)
#' svm_res <- tune_grid(svm_mod, car_rec, resamples = folds, grid = 7)
#' svm_res
#'
#' show_best(svm_res, metric = "rmse")
#'
#' autoplot(svm_res, metric = "rmse") +
#'   scale_x_log10()
#'
#' # ---------------------------------------------------------------------------
#'
#' # Using a variables preprocessor with a workflow
#'
#' # Rather than supplying a preprocessor (like a recipe) and a model directly
#' # to `tune_grid()`, you can also wrap them up in a workflow and pass
#' # that along instead (note that this doesn't do any preprocessing to
#' # the variables, it passes them along as-is).
#' wf <- workflow() |>
#'   add_variables(outcomes = mpg, predictors = everything()) |>
#'   add_model(svm_mod)
#'
#' set.seed(3254)
#' svm_res_wf <- tune_grid(wf, resamples = folds, grid = 7)
#' @export
tune_grid <- function(object, ...) {
  UseMethod("tune_grid")
}

#' @export
tune_grid.default <- function(object, ...) {
  cli::cli_abort(
    "The first argument to {.fn tune_grid} should be either a model or workflow,
    not {.obj_type_friendly {object}}."
  )
}

#' @export
#' @rdname tune_grid
tune_grid.model_spec <- function(
  object,
  preprocessor,
  resamples,
  ...,
  param_info = NULL,
  grid = 10,
  metrics = NULL,
  eval_time = NULL,
  control = control_grid()
) {
  if (rlang::is_missing(preprocessor) || !is_preprocessor(preprocessor)) {
    cli::cli_abort(tune_pp_msg)
  }

  control <- parsnip::condense_control(control, control_grid())

  empty_ellipses(...)

  wflow <- add_model(workflow(), object)

  if (is_recipe(preprocessor)) {
    wflow <- add_recipe(wflow, preprocessor)
  } else if (rlang::is_formula(preprocessor)) {
    wflow <- add_formula(wflow, preprocessor)
  }

  tune_grid(
    wflow,
    resamples = resamples,
    param_info = param_info,
    grid = grid,
    metrics = metrics,
    eval_time = eval_time,
    control = control
  )
}

#' @export
#' @rdname tune_grid
tune_grid.workflow <- function(
  object,
  resamples,
  ...,
  param_info = NULL,
  grid = 10,
  metrics = NULL,
  eval_time = NULL,
  control = control_grid()
) {
  empty_ellipses(...)

  control <- parsnip::condense_control(control, control_grid())

  # Disallow `NULL` grids in `tune_grid()`, as this is the special signal
  # used when no tuning is required
  if (is.null(grid)) {
    cli::cli_abort(grid_msg)
  }

  res <- tune_grid_workflow(
    object,
    resamples = resamples,
    grid = grid,
    metrics = metrics,
    eval_time = eval_time,
    pset = param_info,
    control = control
  )
  .stash_last_result(res)
  res
}

# ------------------------------------------------------------------------------

tune_grid_workflow <- function(
  workflow,
  resamples,
  grid = 10,
  metrics = NULL,
  eval_time = NULL,
  pset = NULL,
  control = control_grid(),
  call = caller_env()
) {
  if (!catalog_is_active()) {
    initialize_catalog(control, workflow = workflow)
  }

  check_rset(resamples)

  metrics <- check_metrics_arg(metrics, workflow, call = call)
  eval_time <- check_eval_time_arg(eval_time, metrics, call = call)

  pset <- check_parameters(
    workflow,
    pset = pset,
    data = resamples$splits[[1]]$data,
    grid_names = names(grid)
  )

  check_workflow(workflow, pset = pset, call = call)
  check_backend_options(control$backend_options)

  grid <- .check_grid(
    grid = grid,
    workflow = workflow,
    pset = pset
  )

  # Save rset attributes, then fall back to a bare tibble
  rset_info <- pull_rset_attributes(resamples)
  split_args <- rsample::.get_split_args(resamples)

  resamples <- tune_grid_loop(
    resamples = resamples,
    grid = grid,
    workflow = workflow,
    param_info = pset,
    metrics = metrics,
    eval_time = eval_time,
    control = control
  )

  y_name <- outcome_names(resamples)

  if (.is_cataclysmic(resamples)) {
    cli::cli_warn(
      "All models failed. Run {.code show_notes(.Last.tune.result)} for more
       information."
    )
  }

  workflow <- .set_workflow(workflow, control)

  new_tune_results(
    x = resamples,
    parameters = pset,
    metrics = metrics,
    eval_time = eval_time,
    eval_time_target = NULL,
    outcomes = y_name,
    rset_info = rset_info,
    workflow = workflow
  )
}

# ------------------------------------------------------------------------------

#' @export
#' @keywords internal
#' @rdname empty_ellipses
pull_rset_attributes <- function(x) {
  excl_att <- c("names", "row.names")
  att <- attributes(x)
  att_nms <- names(att)
  att_nms <- setdiff(att_nms, excl_att)
  att$class <- setdiff(class(x), class(tibble::new_tibble(list())))
  att$class <- att$class[att$class != "rset"]

  if (!is.null(attr(x, ".resample_weights"))) {
    att[[".resample_weights"]] <- attr(x, ".resample_weights")
  }

  lab <- try(pretty(x), silent = TRUE)
  if (inherits(lab, "try-error")) {
    lab <- NA_character_
  }
  list(att = att[att_nms], label = lab)
}

# ------------------------------------------------------------------------------

#' @export
#' @keywords internal
#' @rdname empty_ellipses
.set_workflow <- function(workflow, control) {
  if (control$save_workflow) {
    if (!is.null(workflow)) {
      w_size <- utils::object.size(workflow)
      if (w_size / 1024^2 > control$workflow_size) {
        rounded <- round(w_size / 1024^2, 1)
        msg <-
          cli::format_inline(
            "The workflow being saved is large ({rounded} MB). If this
            was not intentional, please set the control setting
            {.arg save_workflow} to be {.code FALSE} or change the threshold
            for this warning (currently {control$workflow_size} MB) with the
            {.arg workflow_size} argument.",
            keep_whitespace = FALSE,
            collapse = FALSE
          )
        cols <- get_tune_colors()
        msg <- strwrap(
          msg,
          prefix = paste0(cols$symbol$info(cli::symbol$info), " ")
        )
        msg <- cols$message$info(paste0(msg, collapse = "\n"))
        cli::cli_bullets(msg)
      }
    }
    workflow
  } else {
    NULL
  }
}

#' Fit multiple models via resampling
#'
#' [fit_resamples()] computes a set of performance metrics across one or more
#' resamples. It does not perform any tuning (see [tune_grid()] and
#' [tune_bayes()] for that), and is instead used for fitting a single
#' model+recipe or model+formula combination across many resamples.
#'
#' @inheritParams last_fit
#' @inheritParams tune_grid
#'
#' @param resamples An `rset` resampling object created from an `rsample`
#' function, such as [rsample::vfold_cv()].
#'
#' @param control A [control_resamples()] object used to fine tune the resampling
#'   process.
#'
#' @inheritSection tune_grid Performance Metrics
#' @inheritSection tune_grid Obtaining Predictions
#' @inheritSection tune_grid Extracting Information
#' @template case-weights
#' @template censored-regression
#' @seealso [control_resamples()], [collect_predictions()], [collect_metrics()]
#' @examplesIf tune:::should_run_examples() & rlang::is_installed("splines2")
#' library(recipes)
#' library(rsample)
#' library(parsnip)
#' library(workflows)
#'
#' set.seed(6735)
#' folds <- vfold_cv(mtcars, v = 5)
#'
#' spline_rec <- recipe(mpg ~ ., data = mtcars) |>
#'   step_spline_natural(disp) |>
#'   step_spline_natural(wt)
#'
#' lin_mod <- linear_reg() |>
#'   set_engine("lm")
#'
#' control <- control_resamples(save_pred = TRUE)
#'
#' spline_res <- fit_resamples(lin_mod, spline_rec, folds, control = control)
#'
#' spline_res
#'
#' show_best(spline_res, metric = "rmse")
#'
#' # You can also wrap up a preprocessor and a model into a workflow, and
#' # supply that to `fit_resamples()` instead. Here, a workflows "variables"
#' # preprocessor is used, which lets you supply terms using dplyr selectors.
#' # The variables are used as-is, no preprocessing is done to them.
#' wf <- workflow() |>
#'   add_variables(outcomes = mpg, predictors = everything()) |>
#'   add_model(lin_mod)
#'
#' wf_res <- fit_resamples(wf, folds)
#' @export
fit_resamples <- function(object, ...) {
  UseMethod("fit_resamples")
}

#' @export
fit_resamples.default <- function(object, ...) {
  cli::cli_abort(
    "The first argument to {.fn fit_resamples} should be either a model or workflow,
    not {.obj_type_friendly {object}}."
  )
}

#' @export
#' @rdname fit_resamples
fit_resamples.model_spec <- function(
  object,
  preprocessor,
  resamples,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_resamples()
) {
  if (rlang::is_missing(preprocessor) || !is_preprocessor(preprocessor)) {
    cli::cli_abort(tune_pp_msg)
  }

  control <- parsnip::condense_control(control, control_resamples())

  empty_ellipses(...)

  wflow <- add_model(workflow(), object)

  if (is_recipe(preprocessor)) {
    wflow <- add_recipe(wflow, preprocessor)
  } else if (rlang::is_formula(preprocessor)) {
    wflow <- add_formula(wflow, preprocessor)
  }

  fit_resamples(
    wflow,
    resamples = resamples,
    metrics = metrics,
    eval_time = eval_time,
    control = control
  )
}


#' @rdname fit_resamples
#' @export
fit_resamples.workflow <- function(
  object,
  resamples,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_resamples()
) {
  empty_ellipses(...)

  control <- parsnip::condense_control(control, control_resamples())

  res <-
    resample_workflow(
      workflow = object,
      resamples = resamples,
      metrics = metrics,
      eval_time = eval_time,
      control = control
    )
  .stash_last_result(res)
  res
}

# ------------------------------------------------------------------------------

resample_workflow <- function(
  workflow,
  resamples,
  metrics,
  eval_time = NULL,
  control,
  call = caller_env()
) {
  check_no_tuning(workflow)

  # `NULL` is the signal that we have no grid to tune with
  grid <- NULL
  pset <- NULL

  out <- tune_grid_workflow(
    workflow = workflow,
    resamples = resamples,
    grid = grid,
    metrics = metrics,
    eval_time = eval_time,
    pset = pset,
    control = control,
    call = call
  )

  attributes <- attributes(out)

  new_resample_results(
    x = out,
    parameters = attributes$parameters,
    metrics = attributes$metrics,
    eval_time = attributes$eval_time,
    eval_time_target = NULL,
    outcomes = attributes$outcomes,
    rset_info = attributes$rset_info,
    workflow = attributes$workflow
  )
}

test_that("pretty tune objects", {
  expect_equal(
    readRDS(test_path("data", "knn_results.rds")) |> pretty(),
    "10-fold cross-validation repeated 5 times"
  )
  expect_equal(
    ames_grid_search |> pretty(),
    "10-fold cross-validation using stratification"
  )
  expect_equal(
    ames_iter_search |> pretty(),
    "10-fold cross-validation using stratification"
  )
})

check_merged_tibble <- function(x, type = "recipe", complete = TRUE) {
  expect_true(tibble::is_tibble(x))
  expect_equal(names(x), "x")
  expect_true(all(purrr::map_lgl(x$x, inherits, type)))
  if (complete) {
    any_args <- purrr::map_int(x$x, \(.x) tune_args(.x) |> nrow())
    expect_true(!any(any_args > 0))
  }
  invisible(TRUE)
}

test_that("Light mode / default `tune_color`s work", {
  old <- options(tidymodels.dark = NULL)
  on.exit(options(old))

  expect_equal(
    tune:::tune_color$symbol$go("hi"),
    cli::col_black("hi")
  )

  expect_equal(
    tune:::tune_color$message$info("hi"),
    cli::col_black("hi")
  )
})

test_that("Dark mode `tune_color`s work", {
  old <- options(tidymodels.dark = TRUE)
  on.exit(options(old))

  expect_equal(
    tune:::tune_color$symbol$go("hi"),
    cli::col_white("hi")
  )

  expect_equal(
    tune:::tune_color$message$info("hi"),
    cli::col_white("hi")
  )
})

test_that("`tune_color` falls back to light mode with back `tidymodels.dark` option", {
  old <- options(tidymodels.dark = "oh no")
  on.exit(options(old))

  expect_equal(
    tune:::tune_color$symbol$go("hi"),
    cli::col_black("hi")
  )

  expect_equal(
    tune:::tune_color$message$info("hi"),
    cli::col_black("hi")
  )
})

test_that("recipe merges", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  data("Chicago", package = "modeldata")
  spline_rec <-
    recipes::recipe(ridership ~ ., data = head(Chicago)) |>
    recipes::step_date(date) |>
    recipes::step_holiday(date) |>
    recipes::step_rm(date, dplyr::ends_with("away")) |>
    recipes::step_impute_knn(
      recipes::all_predictors(),
      neighbors = tune("imputation")
    ) |>
    recipes::step_other(recipes::all_nominal(), threshold = tune()) |>
    recipes::step_dummy(recipes::all_nominal()) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::step_spline_b(
      recipes::all_predictors(),
      deg_free = tune(),
      degree = tune()
    )
  spline_grid <-
    tibble::tribble(
      ~imputation ,
      ~threshold  ,
      ~deg_free   ,
      ~degree     ,
       3L         ,
       0.088      ,
      14L         ,
       1L         ,
       6L         ,
       0.058      ,
       8L         ,
       1L         ,
       8L         ,
       0.051      ,
      14L         ,
       1L         ,
       9L         ,
       0.007      ,
      10L         ,
       1L         ,
       1L         ,
       0.032      ,
      15L         ,
       2L         ,
       8L         ,
       0.018      ,
       9L         ,
       2L         ,
       1L         ,
       0.036      ,
       5L         ,
       1L         ,
      10L         ,
       0.1        ,
      10L         ,
       2L         ,
       9L         ,
       0.094      ,
      12L         ,
       1L         ,
       4L         ,
       0.025      ,
      12L         ,
       1L
    )

  expect_no_error(
    spline_updated <- merge(spline_rec, spline_grid)
  )
  check_merged_tibble(spline_updated)
  for (i in 1:nrow(spline_grid)) {
    expect_equal(
      spline_updated$x[[i]]$steps[[4]]$neighbors,
      spline_grid$imputation[[i]]
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[5]]$threshold,
      spline_grid$threshold[[i]]
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[8]]$deg_free,
      spline_grid$deg_free[[i]]
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[8]]$degree,
      spline_grid$degree[[i]]
    )
  }
})

test_that("partially recipe merge", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  data("Chicago", package = "modeldata")
  spline_rec <-
    recipes::recipe(ridership ~ ., data = head(Chicago)) |>
    recipes::step_date(date) |>
    recipes::step_holiday(date) |>
    recipes::step_rm(date, dplyr::ends_with("away")) |>
    recipes::step_impute_knn(
      recipes::all_predictors(),
      neighbors = tune("imputation")
    ) |>
    recipes::step_other(recipes::all_nominal(), threshold = tune()) |>
    recipes::step_dummy(recipes::all_nominal()) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::step_spline_b(
      recipes::all_predictors(),
      deg_free = tune(),
      degree = tune()
    )
  spline_grid <-
    tibble::tribble(
      ~imputation ,
      ~threshold  ,
      ~deg_free   ,
      ~degree     ,
       3L         ,
       0.088      ,
      14L         ,
       1L         ,
       6L         ,
       0.058      ,
       8L         ,
       1L         ,
       8L         ,
       0.051      ,
      14L         ,
       1L         ,
       9L         ,
       0.007      ,
      10L         ,
       1L         ,
       1L         ,
       0.032      ,
      15L         ,
       2L         ,
       8L         ,
       0.018      ,
       9L         ,
       2L         ,
       1L         ,
       0.036      ,
       5L         ,
       1L         ,
      10L         ,
       0.1        ,
      10L         ,
       2L         ,
       9L         ,
       0.094      ,
      12L         ,
       1L         ,
       4L         ,
       0.025      ,
      12L         ,
       1L
    )

  expect_no_error(
    spline_updated <- merge(spline_rec, spline_grid[, -1])
  )
  check_merged_tibble(spline_updated, complete = FALSE)
  for (i in 1:nrow(spline_grid)) {
    expect_equal(
      spline_updated$x[[i]]$steps[[4]]$neighbors,
      tune("imputation")
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[5]]$threshold,
      spline_grid$threshold[[i]]
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[8]]$deg_free,
      spline_grid$deg_free[[i]]
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[8]]$degree,
      spline_grid$degree[[i]]
    )
  }
})

test_that("umerged recipe merge", {
  skip_if_not_installed("modeldata")
  skip_if_not_installed("splines2")

  data("Chicago", package = "modeldata")
  spline_rec <-
    recipes::recipe(ridership ~ ., data = head(Chicago)) |>
    recipes::step_date(date) |>
    recipes::step_holiday(date) |>
    recipes::step_rm(date, dplyr::ends_with("away")) |>
    recipes::step_impute_knn(
      recipes::all_predictors(),
      neighbors = tune("imputation")
    ) |>
    recipes::step_other(recipes::all_nominal(), threshold = tune()) |>
    recipes::step_dummy(recipes::all_nominal()) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::step_spline_b(
      recipes::all_predictors(),
      deg_free = tune(),
      degree = tune()
    )
  bst_grid <- tibble::tibble(
    "funky name \n" = 1:4,
    rules = rep(c(TRUE, FALSE), each = 2)
  )

  expect_no_error(
    spline_updated <- merge(spline_rec, bst_grid)
  )
  check_merged_tibble(spline_updated, complete = FALSE)
  for (i in 1:nrow(bst_grid)) {
    expect_equal(
      spline_updated$x[[i]]$steps[[4]]$neighbors,
      tune("imputation")
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[5]]$threshold,
      tune()
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[8]]$deg_free,
      tune()
    )
    expect_equal(
      spline_updated$x[[i]]$steps[[8]]$degree,
      tune()
    )
  }
})

# ------------------------------------------------------------------------------

test_that("model spec merges", {
  library(parsnip)
  bst_model <-
    parsnip::boost_tree(
      mode = "classification",
      trees = tune("funky name \n")
    ) |>
    parsnip::set_engine("C5.0", rules = tune(), noGlobalPruning = TRUE)
  bst_grid <- tibble::tibble(
    "funky name \n" = 1:4,
    rules = rep(c(TRUE, FALSE), each = 2)
  )

  expect_no_error(
    bst_updated <- merge(bst_model, bst_grid)
  )
  check_merged_tibble(bst_updated, "model_spec")
  for (i in 1:nrow(bst_grid)) {
    expect_equal(
      bst_updated$x[[i]]$args$trees,
      rlang::as_quosure(bst_grid[["funky name \n"]][[i]], empty_env())
    )
    expect_equal(
      bst_updated$x[[i]]$eng_args$rules,
      rlang::as_quosure(bst_grid$rules[[i]], empty_env())
    )
  }

  # ensure that `grid` can handle list-columns
  bst_model_obj <-
    boost_tree(mode = "classification") |>
    set_args(objective = tune())

  bst_grid_obj <- tibble::tibble(objective = list("hey", "there"))

  bst_updated_obj <- merge(bst_model_obj, bst_grid_obj)

  expect_equal(
    rlang::quo_get_expr(bst_updated_obj$x[[1]]$eng_args$objective),
    "hey"
  )
  expect_equal(
    rlang::quo_get_expr(bst_updated_obj$x[[2]]$eng_args$objective),
    "there"
  )
})

test_that("partially model spec merge", {
  bst_model <-
    parsnip::boost_tree(
      mode = "classification",
      trees = tune("funky name \n")
    ) |>
    parsnip::set_engine("C5.0", rules = tune(), noGlobalPruning = TRUE)
  bst_grid <- tibble::tibble(
    "funky name \n" = 1:4,
    rules = rep(c(TRUE, FALSE), each = 2)
  )

  expect_no_error(
    bst_updated <- merge(bst_model, bst_grid[, -1])
  )
  check_merged_tibble(bst_updated, "model_spec", complete = FALSE)
  for (i in 1:nrow(bst_grid)) {
    expect_equal(
      rlang::get_expr(bst_updated$x[[i]]$args$trees),
      tune("funky name \n")
    )
    expect_equal(
      bst_updated$x[[i]]$eng_args$rules,
      rlang::as_quosure(bst_grid$rules[[i]], empty_env())
    )
  }
})

test_that("umerged model spec merge", {
  bst_model <-
    parsnip::boost_tree(
      mode = "classification",
      trees = tune("funky name \n")
    ) |>
    parsnip::set_engine("C5.0", rules = tune(), noGlobalPruning = TRUE)
  bst_grid <- tibble::tibble(
    "funky name \n" = 1:4,
    rules = rep(c(TRUE, FALSE), each = 2)
  )

  other_grid <- bst_grid
  names(bst_grid) <- letters[1:2]
  expect_no_error(
    bst_not_updated <- merge(bst_model, other_grid)
  )
  check_merged_tibble(bst_not_updated, "model_spec", complete = FALSE)
  # for (i in 1:nrow(other_grid)) {
  #   if (i == 1) {
  #     print(rlang::get_expr(bst_not_updated$x[[i]]$args$trees))
  #     print(rlang::get_expr(bst_not_updated$x[[i]]$eng_args$rules))
  #   }
  #   expect_equal(
  #     rlang::get_expr(bst_not_updated$x[[i]]$args$trees), tune("funky name \n")
  #   )
  #   expect_equal(
  #     rlang::get_expr(bst_not_updated$x[[i]]$eng_args$rules), tune()
  #   )
  # }
})

# ------------------------------------------------------------------------------

test_that("boosted tree grid reduction - xgboost", {
  skip_if_not_installed("xgboost")

  mod <- parsnip::boost_tree() |>
    parsnip::set_engine("xgboost")

  # A typical grid
  reg_grid <- expand.grid(trees = 1:3, min_n = 1:2)
  reg_grid_smol <- min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$trees, rep(3, 2))
  expect_equal(reg_grid_smol$min_n, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(trees = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(trees = 1:3, min_n = 1:2)[-3, ]
  reg_ish_grid_smol <- min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$trees, 2:3)
  expect_equal(reg_ish_grid_smol$min_n, 1:2)
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(trees = 1:2))
  }

  # Grid with a third parameter
  reg_grid_extra <- expand.grid(trees = 1:3, min_n = 1:2, tree_depth = 10:12)
  reg_grid_extra_smol <- min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$trees, rep(3, 6))
  expect_equal(reg_grid_extra_smol$min_n, rep(1:2, each = 3))
  expect_equal(reg_grid_extra_smol$tree_depth, rep(10:12, 2))
  for (i in 1:nrow(reg_grid_extra_smol)) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(trees = 1:2))
  }

  # Only trees
  only_trees <- expand.grid(trees = 1:3)
  only_trees_smol <- min_grid(mod, only_trees)

  expect_equal(only_trees_smol$trees, 3)
  expect_equal(only_trees_smol$.submodels, list(list(trees = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(trees = 1, min_n = 1:2)
  no_sub_smol <- min_grid(mod, no_sub)

  expect_equal(no_sub_smol$trees, rep(1, 2))
  expect_equal(no_sub_smol$min_n, 1:2)
  for (i in 1:nrow(no_sub_smol)) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- parsnip::boost_tree(trees = tune("Amos")) |>
    parsnip::set_engine("xgboost")
  reg_grid <- expand.grid(Amos = 1:3, min_n = 1:2)
  reg_grid_smol <- min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Amos, rep(3, 2))
  expect_equal(reg_grid_smol$min_n, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Amos = 1:2))
  }

  all_sub <- expand.grid(Amos = 1:3)
  all_sub_smol <- min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Amos, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Amos = 1:2))

  mod_2 <- parsnip::boost_tree(trees = tune("Ade Tukunbo")) |>
    parsnip::set_engine("xgboost")
  reg_grid <- expand.grid(`Ade Tukunbo` = 1:3, min_n = 1:2, ` \t123` = 10:11)
  reg_grid_smol <- min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Ade Tukunbo`, rep(3, 4))
  expect_equal(reg_grid_smol$min_n, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Ade Tukunbo` = 1:2))
  }
})

# ------------------------------------------------------------------------------

test_that("boosted tree grid reduction - C5.0", {
  mod <- parsnip::boost_tree() |> parsnip::set_engine("C5.0")

  # A typical grid
  reg_grid <- expand.grid(trees = 1:3, min_n = 1:2)
  reg_grid_smol <- min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$trees, rep(3, 2))
  expect_equal(reg_grid_smol$min_n, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(trees = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(trees = 1:3, min_n = 1:2)[-3, ]
  reg_ish_grid_smol <- min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$trees, 2:3)
  expect_equal(reg_ish_grid_smol$min_n, 1:2)
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(trees = 1:2))
  }

  # Grid with a third parameter
  reg_grid_extra <- expand.grid(trees = 1:3, min_n = 1:2, tree_depth = 10:12)
  reg_grid_extra_smol <- min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$trees, rep(3, 6))
  expect_equal(reg_grid_extra_smol$min_n, rep(1:2, each = 3))
  expect_equal(reg_grid_extra_smol$tree_depth, rep(10:12, 2))
  for (i in 1:nrow(reg_grid_extra_smol)) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(trees = 1:2))
  }

  # Only trees
  only_trees <- expand.grid(trees = 1:3)
  only_trees_smol <- min_grid(mod, only_trees)

  expect_equal(only_trees_smol$trees, 3)
  expect_equal(only_trees_smol$.submodels, list(list(trees = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(trees = 1, min_n = 1:2)
  no_sub_smol <- min_grid(mod, no_sub)

  expect_equal(no_sub_smol$trees, rep(1, 2))
  expect_equal(no_sub_smol$min_n, 1:2)
  for (i in 1:nrow(no_sub_smol)) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- parsnip::boost_tree(trees = tune("Marco")) |>
    parsnip::set_engine("C5.0")
  reg_grid <- expand.grid(Marco = 1:3, min_n = 1:2)
  reg_grid_smol <- min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Marco, rep(3, 2))
  expect_equal(reg_grid_smol$min_n, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Marco = 1:2))
  }

  all_sub <- expand.grid(Marco = 1:3)
  all_sub_smol <- min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Marco, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Marco = 1:2))

  mod_2 <- parsnip::boost_tree(trees = tune("Anderson Dawes")) |>
    parsnip::set_engine("C5.0")
  reg_grid <- expand.grid(`Anderson Dawes` = 1:3, min_n = 1:2, ` \t123` = 10:11)
  reg_grid_smol <- min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Anderson Dawes`, rep(3, 4))
  expect_equal(reg_grid_smol$min_n, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Anderson Dawes` = 1:2))
  }
})

# ------------------------------------------------------------------------------

test_that("linear regression grid reduction - glmnet", {
  # glmnet depends on >= 3.6.0 so we only test locally
  skip_if_not_installed("glmnet")

  mod <- parsnip::linear_reg() |> parsnip::set_engine("glmnet")

  # A typical grid
  reg_grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)
  reg_grid_smol <- min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$penalty, rep(3, 5))
  expect_equal(reg_grid_smol$mixture, (1:5) / 5)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)[-3, ]
  reg_ish_grid_smol <- min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$penalty, c(2, rep(3, 4)))
  expect_equal(reg_ish_grid_smol$mixture, (1:5) / 5)
  expect_equal(reg_ish_grid_smol$.submodels[[1]], list(penalty = 1))
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Grid with a third parameter
  reg_grid_extra <- expand.grid(
    penalty = 1:3,
    mixture = (1:5) / 5,
    blah = 10:12
  )
  reg_grid_extra_smol <- min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$penalty, rep(3, 15))
  expect_equal(reg_grid_extra_smol$mixture, rep((1:5) / 5, each = 3))
  expect_equal(reg_grid_extra_smol$blah, rep(10:12, 5))
  for (i in 1:nrow(reg_grid_extra_smol)) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Penalty not specified
  expect_snapshot(error = TRUE, {
    min_grid(mod, data.frame(mixture = 1:3))
  })

  # Only penalty
  only_penalty <- expand.grid(penalty = 1:3)
  only_penalty_smol <- min_grid(mod, only_penalty)

  expect_equal(only_penalty_smol$penalty, 3)
  expect_equal(only_penalty_smol$.submodels, list(list(penalty = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(penalty = 1:5, mixture = (1:5) / 5)
  no_sub_smol <- min_grid(mod, no_sub)

  expect_equal(no_sub_smol$penalty, 1:5)
  expect_equal(no_sub_smol$mixture, (1:5) / 5)
  for (i in 1:nrow(no_sub_smol)) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- parsnip::linear_reg(penalty = tune("Shaddid")) |>
    parsnip::set_engine("glmnet")
  reg_grid <- expand.grid(Shaddid = 1:3, mixture = 1:2)
  reg_grid_smol <- min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Shaddid, rep(3, 2))
  expect_equal(reg_grid_smol$mixture, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Shaddid = 1:2))
  }

  all_sub <- expand.grid(Shaddid = 1:3)
  all_sub_smol <- min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Shaddid, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Shaddid = 1:2))

  mod_2 <- parsnip::linear_reg(penalty = tune("Josephus Miller")) |>
    parsnip::set_engine("glmnet")
  reg_grid <- expand.grid(
    `Josephus Miller` = 1:3,
    mixture = 1:2,
    ` \t123` = 10:11
  )
  reg_grid_smol <- min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Josephus Miller`, rep(3, 4))
  expect_equal(reg_grid_smol$mixture, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Josephus Miller` = 1:2))
  }
})

# ------------------------------------------------------------------------------

test_that("logistic regression grid reduction - glmnet", {
  # glmnet depends on >= 3.6.0 so we only test locally
  skip_if_not_installed("glmnet")

  mod <- parsnip::logistic_reg() |> parsnip::set_engine("glmnet")

  # A typical grid
  reg_grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)
  reg_grid_smol <- min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$penalty, rep(3, 5))
  expect_equal(reg_grid_smol$mixture, (1:5) / 5)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)[-3, ]
  reg_ish_grid_smol <- min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$penalty, c(2, rep(3, 4)))
  expect_equal(reg_ish_grid_smol$mixture, (1:5) / 5)
  expect_equal(reg_ish_grid_smol$.submodels[[1]], list(penalty = 1))
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Grid with a third parameter
  reg_grid_extra <- expand.grid(
    penalty = 1:3,
    mixture = (1:5) / 5,
    blah = 10:12
  )
  reg_grid_extra_smol <- min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$penalty, rep(3, 15))
  expect_equal(reg_grid_extra_smol$mixture, rep((1:5) / 5, each = 3))
  expect_equal(reg_grid_extra_smol$blah, rep(10:12, 5))
  for (i in 1:nrow(reg_grid_extra_smol)) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Penalty not specified
  expect_snapshot(error = TRUE, {
    min_grid(mod, data.frame(mixture = 1:3))
  })

  # Only penalty
  only_penalty <- expand.grid(penalty = 1:3)
  only_penalty_smol <- min_grid(mod, only_penalty)

  expect_equal(only_penalty_smol$penalty, 3)
  expect_equal(only_penalty_smol$.submodels, list(list(penalty = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(penalty = 1:5, mixture = (1:5) / 5)
  no_sub_smol <- min_grid(mod, no_sub)

  expect_equal(no_sub_smol$penalty, 1:5)
  expect_equal(no_sub_smol$mixture, (1:5) / 5)
  for (i in 1:nrow(no_sub_smol)) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- parsnip::logistic_reg(penalty = tune("Prax")) |>
    parsnip::set_engine("glmnet")
  reg_grid <- expand.grid(Prax = 1:3, mixture = 1:2)
  reg_grid_smol <- min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Prax, rep(3, 2))
  expect_equal(reg_grid_smol$mixture, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Prax = 1:2))
  }

  all_sub <- expand.grid(Prax = 1:3)
  all_sub_smol <- min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Prax, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Prax = 1:2))

  mod_2 <- parsnip::logistic_reg(penalty = tune("Samara Rosenberg")) |>
    parsnip::set_engine("glmnet")
  reg_grid <- expand.grid(
    `Samara Rosenberg` = 1:3,
    mixture = 1:2,
    ` \t123` = 10:11
  )
  reg_grid_smol <- min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Samara Rosenberg`, rep(3, 4))
  expect_equal(reg_grid_smol$mixture, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Samara Rosenberg` = 1:2))
  }
})

# more of a negative control test
test_that("logistic regression grid reduction - spark", {
  reg_grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)
  reg_grid_smol <- min_grid(
    parsnip::logistic_reg() |>
      parsnip::set_engine("spark"),
    reg_grid
  )

  expect_equal(reg_grid_smol$penalty, reg_grid$penalty)
  expect_equal(reg_grid_smol$mixture, reg_grid$mixture)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list())
  }
})

# ------------------------------------------------------------------------------

test_that("MARS grid reduction - earth", {
  mod <- parsnip::mars() |> parsnip::set_engine("earth")

  # A typical grid
  reg_grid <- expand.grid(num_terms = 1:3, prod_degree = 1:2)
  reg_grid_smol <- min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$num_terms, rep(3, 2))
  expect_equal(reg_grid_smol$prod_degree, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(num_terms = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(num_terms = 1:3, prod_degree = 1:2)[-3, ]
  reg_ish_grid_smol <- min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$num_terms, 2:3)
  expect_equal(reg_ish_grid_smol$prod_degree, 1:2)
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(num_terms = 1:2))
  }

  # Grid with a third parameter
  reg_grid_extra <- expand.grid(
    num_terms = 1:3,
    prod_degree = 1:2,
    blah = 10:12
  )
  reg_grid_extra_smol <- min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$num_terms, rep(3, 6))
  expect_equal(reg_grid_extra_smol$prod_degree, rep(1:2, each = 3))
  expect_equal(reg_grid_extra_smol$blah, rep(10:12, 2))
  for (i in 1:nrow(reg_grid_extra_smol)) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(num_terms = 1:2))
  }

  # Only num_terms
  only_num_terms <- expand.grid(num_terms = 1:3)
  only_num_terms_smol <- min_grid(mod, only_num_terms)

  expect_equal(only_num_terms_smol$num_terms, 3)
  expect_equal(only_num_terms_smol$.submodels, list(list(num_terms = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(num_terms = 1, prod_degree = 1:2)
  no_sub_smol <- min_grid(mod, no_sub)

  expect_equal(no_sub_smol$num_terms, rep(1, 2))
  expect_equal(no_sub_smol$prod_degree, 1:2)
  for (i in 1:nrow(no_sub_smol)) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- parsnip::mars(num_terms = tune("Filip")) |>
    parsnip::set_engine("earth")
  reg_grid <- expand.grid(Filip = 1:3, prod_degree = 1:2)
  reg_grid_smol <- min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Filip, rep(3, 2))
  expect_equal(reg_grid_smol$prod_degree, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Filip = 1:2))
  }

  all_sub <- expand.grid(Filip = 1:3)
  all_sub_smol <- min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Filip, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Filip = 1:2))

  mod_2 <- parsnip::mars(num_terms = tune("Elvi Okoye")) |>
    parsnip::set_engine("earth")
  reg_grid <- expand.grid(
    `Elvi Okoye` = 1:3,
    prod_degree = 1:2,
    ` \t123` = 10:11
  )
  reg_grid_smol <- min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Elvi Okoye`, rep(3, 4))
  expect_equal(reg_grid_smol$prod_degree, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Elvi Okoye` = 1:2))
  }
})

# ------------------------------------------------------------------------------

test_that("multinomial regression grid reduction - glmnet", {
  # glmnet depends on >= 3.6.0 so we only test locally
  skip_if_not_installed("glmnet")

  mod <- parsnip::multinom_reg() |> parsnip::set_engine("glmnet")

  # A typical grid
  reg_grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)
  reg_grid_smol <- min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$penalty, rep(3, 5))
  expect_equal(reg_grid_smol$mixture, (1:5) / 5)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(penalty = 1:3, mixture = (1:5) / 5)[-3, ]
  reg_ish_grid_smol <- min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$penalty, c(2, rep(3, 4)))
  expect_equal(reg_ish_grid_smol$mixture, (1:5) / 5)
  expect_equal(reg_ish_grid_smol$.submodels[[1]], list(penalty = 1))
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Grid with a third parameter
  reg_grid_extra <- expand.grid(
    penalty = 1:3,
    mixture = (1:5) / 5,
    blah = 10:12
  )
  reg_grid_extra_smol <- min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$penalty, rep(3, 15))
  expect_equal(reg_grid_extra_smol$mixture, rep((1:5) / 5, each = 3))
  expect_equal(reg_grid_extra_smol$blah, rep(10:12, 5))
  for (i in 1:nrow(reg_grid_extra_smol)) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(penalty = 1:2))
  }

  # Penalty not specified
  expect_snapshot(error = TRUE, {
    min_grid(mod, data.frame(mixture = 1:3))
  })

  # Only penalty
  only_penalty <- expand.grid(penalty = 1:3)
  only_penalty_smol <- min_grid(mod, only_penalty)

  expect_equal(only_penalty_smol$penalty, 3)
  expect_equal(only_penalty_smol$.submodels, list(list(penalty = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(penalty = 1:5, mixture = (1:5) / 5)
  no_sub_smol <- min_grid(mod, no_sub)

  expect_equal(no_sub_smol$penalty, 1:5)
  expect_equal(no_sub_smol$mixture, (1:5) / 5)
  for (i in 1:nrow(no_sub_smol)) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- parsnip::multinom_reg(penalty = tune("Cortazar")) |>
    parsnip::set_engine("glmnet")
  reg_grid <- expand.grid(Cortazar = 1:3, mixture = 1:2)
  reg_grid_smol <- min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Cortazar, rep(3, 2))
  expect_equal(reg_grid_smol$mixture, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Cortazar = 1:2))
  }

  all_sub <- expand.grid(Cortazar = 1:3)
  all_sub_smol <- min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Cortazar, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Cortazar = 1:2))

  mod_2 <- parsnip::multinom_reg(penalty = tune("Shed Garvey")) |>
    parsnip::set_engine("glmnet")
  reg_grid <- expand.grid(`Shed Garvey` = 1:3, mixture = 1:2, ` \t123` = 10:11)
  reg_grid_smol <- min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Shed Garvey`, rep(3, 4))
  expect_equal(reg_grid_smol$mixture, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Shed Garvey` = 1:2))
  }
})

# ------------------------------------------------------------------------------

test_that("nearest neighbors grid reduction - kknn", {
  mod <- parsnip::nearest_neighbor() |> parsnip::set_engine("kknn")

  # A typical grid
  reg_grid <- expand.grid(neighbors = 1:3, dist_power = 1:2)
  reg_grid_smol <- min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$neighbors, rep(3, 2))
  expect_equal(reg_grid_smol$dist_power, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(neighbors = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(neighbors = 1:3, dist_power = 1:2)[-3, ]
  reg_ish_grid_smol <- min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$neighbors, 2:3)
  expect_equal(reg_ish_grid_smol$dist_power, 1:2)
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(neighbors = 1:2))
  }

  # Grid with a third parameter
  wts <- c("rectangular", "triangular", "epanechnikov")
  reg_grid_extra <- expand.grid(
    neighbors = 1:3,
    dist_power = 1:2,
    weight_func = wts
  )
  reg_grid_extra_smol <- min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$neighbors, rep(3, 6))
  expect_equal(reg_grid_extra_smol$dist_power, rep(1:2, each = 3))
  expect_equal(reg_grid_extra_smol$weight_func, rep(wts, 2))
  for (i in 1:nrow(reg_grid_extra_smol)) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(neighbors = 1:2))
  }

  # Only neighbors
  only_neighbors <- expand.grid(neighbors = 1:3)
  only_neighbors_smol <- min_grid(mod, only_neighbors)

  expect_equal(only_neighbors_smol$neighbors, 3)
  expect_equal(only_neighbors_smol$.submodels, list(list(neighbors = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(neighbors = 1, dist_power = 1:2)
  no_sub_smol <- min_grid(mod, no_sub)

  expect_equal(no_sub_smol$neighbors, rep(1, 2))
  expect_equal(no_sub_smol$dist_power, 1:2)
  for (i in 1:nrow(no_sub_smol)) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- parsnip::nearest_neighbor(neighbors = tune("Nami")) |>
    parsnip::set_engine("kknn")
  reg_grid <- expand.grid(Nami = 1:3, dist_power = 1:2)
  reg_grid_smol <- min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Nami, rep(3, 2))
  expect_equal(reg_grid_smol$dist_power, 1:2)
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Nami = 1:2))
  }

  all_sub <- expand.grid(Nami = 1:3)
  all_sub_smol <- min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Nami, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Nami = 1:2))

  mod_2 <- parsnip::nearest_neighbor(neighbors = tune("Michio Pa")) |>
    parsnip::set_engine("kknn")
  reg_grid <- expand.grid(`Michio Pa` = 1:3, dist_power = 1:2, ` \t123` = 10:11)
  reg_grid_smol <- min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Michio Pa`, rep(3, 4))
  expect_equal(reg_grid_smol$dist_power, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in 1:nrow(reg_grid_smol)) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Michio Pa` = 1:2))
  }
})

---
title: Loop over all stages aka loopy
---

We have a big ol loop at the heart of `tune_grid()` (and thus the rest of the tuning/resampling functions). That can be a lot to hold in your head at once, so here's a memory aid.

## Overview

We need to work our way through a whole lot of resamples and tuning parameter combinations. We could take the whole grid of tuning parameters, splice them into the workflow, and fit all those workflows on all resamples. This repeats potentially costly calculations, so we have carefully crafted the loop to avoid redundant computation.

The loop runs over a single resample. There are two paths depending on whether metrics are **prediction-based** (standard yardstick metrics) or **model-based** (metrics computed directly from the fitted workflow, e.g. clustering).

### Prediction-based metrics path (default)

- For each preproc parameter combination
	- Fit preprocessor
	- Apply it to analysis set
	- Apply it to the assessment set
	- Apply it to the calibration set
	- For each (non-sub) model parameter combination
		- Fit the model
		- Predict for all submodel parameters at once (via multi_predict)
		- Predict on calibration data for all submodels at once
		- For each submodel parameter value
			- Filter to this submodel's predictions
			- Filter to this submodel's calibration predictions
			- For each post parameter combination
				- Fit the postprocessor on the calibration predictions
				- Apply the postprocessor to the predictions (on the assessment set)
				- Combine (post-processed) predictions with the grid to `final_pred`
				- Save `final_pred` by appending it to `pred_reserve`
				- Do the extracts
- Compute the metrics from `pred_reserve`

### Model-based metrics path (e.g. clustering)

Used when `static$metric_type == "model"`. Metrics are computed directly from the fitted workflow and the assessment data, bypassing the predict/submodel/post stages entirely.

- For each preproc parameter combination
	- Fit preprocessor
	- Apply it to analysis set
	- For each model parameter combination
		- Fit the model
		- Compute metrics directly: `metrics(workflow, new_data = pred_data)`
		- Save metrics to `metrics_reserve`
		- Do the extracts
- Collect metrics from `metrics_reserve`

## Inputs

### The schedule

The schedule is a nested tibble created by `schedule_grid()` that organizes tuning parameters by stage. Each stage contains the next stage in a list-column:

```
sched
├── [preproc params]
└── model_stage (list-col)
    └── tibble
        ├── [model params]
        └── predict_stage (list-col)
            └── tibble
                ├── [submodel params]
                └── post_stage (list-col)
                    └── tibble
                        └── [post params]
```

The four stages are:

- **pre**: preprocessing via recipes
- **model**: the model fit via parsnip (submodel parameters are collapsed via `min_grid()`)
- **predict**: prediction (submodel parameters are expanded)
- **post**: postprocessing via tailor

To access the next stage, extract with `[[1]]`:

- `current_sched_pre$model_stage[[1]]` → tibble of model param combinations
- `current_sched_model$predict_stage[[1]]` → tibble of submodel param values
- `current_sched_pred$post_stage[[1]]` → tibble of post param combinations

For model-based metrics (e.g. clustering), the predict and post stages are never entered.

### The `static` object

The `static` list contains everything that stays constant throughout the loop:

- `wflow` - the original workflow (template for finalization)
- `param_info` - parameter set info from `tune_args()`
- `configs` - tibble mapping parameter values to `.config` labels
- `metrics` - the metric set
- `metric_type` - `"prediction"` for standard yardstick metrics, `"model"` for metrics computed directly from the fitted workflow (e.g. clustering)
- `pred_types` - prediction types needed (e.g., "class", "prob", "numeric"); empty for model-based metrics
- `eval_time` - evaluation times for survival models
- `control` - control options
- `data` - list with `fit`, `pred`, and `cal` data partitions (added after setup)
- `y_name` - outcome column name(s); not set for model-based metrics

### Data partitions

The `static$data` list contains three partitions (set up once per resample):

- `fit` - training data for preprocessor and model
- `pred` - assessment data for predictions (used to compute metrics)
- `cal` - calibration data for postprocessors that need fitting (e.g., probability calibration)

When there's no postprocessor requiring calibration, `cal` is NULL and `fit` uses the full analysis set. When calibration is needed, the analysis set is further split into `fit` and `cal`.

## Reading the code

### Naming conventions

**Loop variables:**

- `iter_{stage}` - iteration counter (e.g., `iter_pre`, `iter_model`, `iter_pred`, `iter_post`)
- `num_iterations_{stage}` - total iterations for that stage

**Schedule objects:**

- `sched` - the full nested schedule tibble
- `current_sched_{stage}` - current row of the schedule at each stage

**Grid objects:**

- `current_grid` - progressively accumulates tuning params as we descend into loops
- `grid_with_pre` - snapshot with pre params (before model loop)
- `grid_with_pre_model` - snapshot with pre + model params, without submodel col (before pred loop)
- `grid_with_pre_model_pred` - snapshot with pre + model + pred params (before post loop)

**Workflow snapshots** (saved to allow re-finalization in inner loops):

- `current_wflow` - the workflow being modified
- `wflow_with_fitted_pre` - snapshot after fitting preprocessor
- `wflow_with_fitted_pre_and_model` - snapshot after fitting model

**Prediction objects (prediction-based metrics path):**

- `pred_data` - processed prediction data (features + outcomes)
- `pred_all_submodels` - batched predictions for all submodel values (source for filtering)
- `current_pred` - predictions for current pred iteration (filtered from `pred_all_submodels`)
- `final_pred` - predictions after postprocessing (ready to save)
- `pred_reserve` - accumulator for all final predictions

**Metric objects (model-based metrics path):**

- `metrics_reserve` - accumulator for metrics computed directly per model fit

**Calibration objects:**

- `cal_pred_data` - processed calibration data (features + outcomes), analogous to `pred_data`
- `cal_pred_all_submodels` - batched calibration predictions for all submodel values
- `current_cal_pred` - calibration predictions for current pred iteration (filtered from `cal_pred_all_submodels`)

**General conventions:**

- `static` - things that don't change during the loop
- `current_*` - value for the current iteration
- `*_all_submodels` - batched values for all submodel params (e.g., `pred_all_submodels`)
- Stage suffixes: `_pre`, `_model`, `_pred`, `_post`

### Key helper functions

**Finalization** (splice tuning params into workflow and fit):

- `finalize_fit_pre()` - finalize recipe params, fit preprocessor
- `finalize_fit_model()` - finalize model params, fit model
- `finalize_fit_post()` - finalize tailor params, fit postprocessor on pre-computed predictions
- `fit_post_from_predictions()` - fit the postprocessor from predictions (instead of raw data)

**Grid helpers:**

- `extend_grid()` - extend a grid with params from a schedule row (strips `*_stage` columns)
- `remove_stage()` - remove nested stage columns from a schedule row

**Prediction:**

- `process_prediction_data()` - apply fitted preprocessor to data (`source = "pred"` for assessment, `source = "cal"` for calibration)
- `predict_all_types()` - generate all needed prediction types (`source = "pred"` for assessment, `source = "cal"` for calibration)

### Error handling

The loop uses a consistent error handling pattern:

```r
result <- .catch_and_log(
  some_operation(),
  control = static$control,
  split_labels = split_labs,
  location = location,
  notes = notes
)

if (is_failure(result)) {
  next
}
result <- remove_log_notes(result)
```

- `.catch_and_log()` wraps operations to capture errors/warnings without stopping
- `is_failure()` checks if the operation failed
- `next` skips to the next iteration (the failed config won't have results)
- `remove_log_notes()` strips logging metadata from successful results

## Efficiency

The nested structure avoids redundant computation:

### Prediction-based metrics

| What | Computed | Reused for |
|------|----------|------------|
| Preprocessor fit | Once per preproc param combo | All model params below it |
| Processed prediction data | Once per preproc param combo | All predictions below it |
| Model fit | Once per model param combo | All submodel predictions |
| Submodel predictions | Once per model (batched) | All submodel × post combos |
| Calibration predictions | Once per model (batched) | All submodel × post combos |
| Postprocessor fit | Once per post param combo | That specific config |

Submodel parameters (like `penalty` in glmnet) are predicted all at once using `multi_predict()`, which is much faster than predicting one at a time.

### Model-based metrics

| What | Computed | Reused for |
|------|----------|------------|
| Preprocessor fit | Once per preproc param combo | All model params below it |
| Model fit | Once per model param combo | Metric computation for that config |

There are no predictions, submodels, calibration, or postprocessing steps. Metrics are computed immediately after each model fit from the assessment data (`static$data$pred$data`).

## Debugging

Set `control = control_grid(allow_par = FALSE)` to run sequentially with `lapply()` so you can see output and use `browser()`.

For parallel debugging, `options(future.debug = TRUE)` can help.

---
title: "Event time prediction with tidymodels"
output:
  pdf_document: default
  html_document: default
date: "2023-04-13"
---

```{r}
#| label: get-engines
#| include: false

library(tidymodels)
library(censored)
library(joineR)
library(future)

# ------------------------------------------------------------------------------

tidymodels_prefer()
theme_set(theme_bw())
options(pillar.advice = FALSE, pillar.min_title_chars = Inf)
plan("multisession")

# ------------------------------------------------------------------------------

engines <- 
  get_from_env("models") %>% 
  map(~ get_from_env(.x) %>% mutate(model = .x)) %>% 
  list_rbind() %>% 
  filter(mode == "censored regression")

num_models <- engines %>% distinct(model) %>% nrow()
```

The censored package was released in June 2022, enabling users to fit event time/survival time models using the tidymodels framework. As of this writing, there are now a total of `r nrow(engines)` different engines that can be used with `r num_models` different models. 

This document is intended as a tutorial for using the broader tidymodels framework for event time analysis, including model tuning, evaluation, and selection. 

To reproduce these results, you might need to update some package versions: 

```{r}
#| label: installs
#| eval: false

# Get CRAN versions of
pak::pak(c("parsnip", "censored"), ask = FALSE)

# Get GitHub versions of: 
pak::pak(c("tidymodels/tune@ipcw"), ask = FALSE)
pak::pak(c("tidymodels/yardstick"), ask = FALSE)
```


## An Example

We'll use the heart valve data set in the joineR package (also described in [this publication](https://doi.org/10.1186/s12874-021-01375-x)). There are `r length(unique(heart.valve$num))` patients in the study that experienced aortic valve replacement surgery. The data has time-dependent covariates, but we will skip those to simplify the analysis here. The outcome is the time to death after surgery:

```{r}
#| label: print-data-head
library(joineR)

data(heart.valve, package = "joineR")

str(heart.valve)
```

Loading needed tidymodels packages:

```{r}
#| label: load-tm
library(tidymodels)
library(censored)

# ------------------------------------------------------------------------------

tidymodels_prefer()
theme_set(theme_bw())
options(pillar.advice = FALSE, pillar.min_title_chars = Inf)
```

We'll retrieve the appropriate event times for the outcome (since there are multiple time points where patients were measured). Then, we'll identify the predictors that have the same values across the multiple time points and merge them. Functions in the joineR package will help us out here: 

```{r}
outcome_data <- 
  UniqueVariables(heart.valve, var.col = c("fuyrs", "status"), id.col = "num")
covar_data <- 
  UniqueVariables(heart.valve, 
                  var.col = c("age", "hs", "sex", "lv", "emergenc", "hc", "sten.reg.mix"), 
                  id.col = "num")

heart_data <- 
  full_join(outcome_data, covar_data, by = "num") %>% 
  select(-num) %>%
  as_tibble()

heart_data
```

We'll reformat some of the categorical predictors since they are currently encoded as integers. 

Also, tidymodels expects that the event times and corresponding status data are pre-formatted using the `Surv` function in the survival package. We'll do that, then remove the original `fuyrs` and `status` columns.

```{r}
#| label: reformat-data
heart_data <- 
  heart_data %>% 
  mutate(
    event_time = Surv(fuyrs, status),
    lv =
      case_when(
        lv == 1 ~ "good",
        lv == 2 ~ "moderate",
        lv == 3 ~ "poor"
      ),
    emergenc =
      case_when(
        emergenc == 0 ~ "elective",
        emergenc == 1 ~ "urgent",
        emergenc == 2 ~ "emergency"
      ),
    hc =
      case_when(
        hc == 0 ~ "absent",
        hc == 1 ~ "present_treated",
        hc == 2 ~ "present_untreated"
      ),
    sten.reg.mix =
      case_when(
        sten.reg.mix == 1 ~ "stenosis",
        sten.reg.mix == 2 ~ "regurgitation",
        sten.reg.mix == 3 ~ "mixed"
      ),
    hs =
      case_when(
        hs == "Homograft" ~ "homograft",
        TRUE ~ "stentless_porcine_tissue"
      ),
    across(where(is.character), factor)
  ) %>% 
  select(-fuyrs, -status)
```

Since our focus is on prediction, the standard tidymodels methods for data splitting are used to create training and test sets. We'll also make cross-validation folds: 

```{r}
#| label: data-splitting
set.seed(6941)
valve_split <- initial_split(heart_data)
valve_tr <- training(valve_split)
valve_te <- testing(valve_split)
```

In the training set, the observed time values range from `r format(min(valve_tr$event_time[,1]), digits = 2)` years to `r format(max(valve_tr$event_time[,1]), digits = 2)` years and `r round(mean(valve_tr$event_time[,2]) * 100, 2)`% of the patients died (i.e. were events). 

## New Prediction Types

There are different types of predictions for event time analysis. _Dynamic_ predictions require a specific time to make the prediction at. That time is sometimes called a "landmark time", we call it "evaluation time" since our focus is prediction. For example, we might want to know the probability of survival up to some evaluation time $t$. A _static_ prediction is one that is not dependent on an evaluation time point. For example, we might predict the event time from a model. 

To demonstrate, let's fit a bagged tree to the training data: 

```{r}
#| label: initial-example

bag_spec <- 
  bag_tree() %>%
  set_mode("censored regression") %>% 
  set_engine("rpart", nbagg = 50)

set.seed(29872)
bag_fit <- 
  bag_spec %>% 
  fit(event_time ~ ., data = valve_tr)
```

Instead of using the training or testing sets, let's make two fake patients by randomly selecting rows from the training set: 

```{r}
#| fake-patients
set.seed(4853)
fake_examples <- 
  slice_sample(valve_tr, n = 2)

fake_examples
```

The standard `predict()` machinery can be used to get static (e.g., `type = "time"`) or dynamic predictions (e.g., `type = "survival"`). We'll create a grid of `r length(seq(0, 10, by = .1))` evaluation time points for the latter:


```{r}
#| fake-predictions
time_points <- seq(0, 10, by = .1)
bag_pred <- 
  predict(bag_fit, fake_examples, type = "survival", eval_time = time_points) %>% 
  bind_cols(
    predict(bag_fit, fake_examples),
    fake_examples %>% select(event_time)
  ) %>% 
  add_rowindex()
bag_pred
```

As usual, the prediction columns are prefixed with `.pred_`. What is unusual is that `.pred` is a list column, and each list element is a tibble with `r ncol(bag_pred$.pred[[1]])` columns and `r nrow(bag_pred$.pred[[1]])` rows. They contain the survival estimates for each patient: 

```{r}
#| dot-pred
bag_pred$.pred[[1]] %>% slice(1:5)
```

We can unnest these and plot the per-patient survival curves: 

```{r}
#| label: example-survival-probs
#| out-width: "60%"
#| fig-align: center

bag_pred %>% 
  unnest(.pred) %>% 
  mutate(sample = format(.row)) %>% 
  ggplot(aes(.eval_time, .pred_survival, group = sample, col = sample)) + 
  geom_step() + 
  lims(y = 0:1) +
  labs(x = "Time", y = "Probability of Survival") +
  scale_color_brewer(palette = "Paired")
```

The static/dynamic prediction types make these models' tuning and evaluations a little more complex. In many tidymodels functions, there is a new argument called `eval_time` that is used to specify the time points for dynamic predictions (as we'll see in a minute). 

## Measures of Performance

Metrics to measure how well our model performs can also be split into dynamic and static metrics. 

For static, a common choice is the [concordance statistic](https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6041741/), accessible via the `concordance_survival()` function. If we were looking at the test set results for the bagged tree model: 

```{r}
#| label: test-concordance

test_pred <- 
  predict(bag_fit, valve_te, type = "survival", eval_time = time_points) %>% 
  bind_cols(
    predict(bag_fit, valve_te),
    valve_te %>% select(event_time)
  )

test_pred %>% 
  concordance_survival(truth = event_time, estimate = .pred_time)
```

Dynamic metrics usually are classification metrics re-purposed for survival analysis. For example, if we wanted to evaluate the model at $t = 5$, we could use the predicted survival probabilities and try to classify each data point as dead or alive. This ends up being a two class situation, and metrics like the Brier Score or the area under the ROC curve can be used to quantify how well the model works at evaluation time $t$. 

The main difficulty is that, due to censoring, some data can't be cleanly classified. If we have a censored event at time 6, we definitely know that it should not be classified as an event. However, if the observed time were 2 and censored, we don't know if it is an event at $t = 5$ or not. 

There are a lot of ways to deal with this issue. We've done an exhaustive reading of the literature and have come to a somewhat opinionated conclusion. Most of the survival metrics in the literature are developed to univariately score a collection of predictors, typically biomarkers, regarding how well they are associated with the event times. That's not what we are doing; we have model predictions. 

Our choice for dynamic metrics is to use the inverse probability of censoring weights (IPCW), specifically the scheme used by [Graf _et al._ (1999)](https://scholar.google.com/scholar?hl=en&as_sdt=0%2C7&q=%22Assessment+and+comparison+of+prognostic+classification+schemes+for+survival+data). They compute the probability that every data point might have been censored and uses the inverse of this value as a case weight. If the observed time is a censoring time that occurs before the evaluation time, the data point should make no contribution to the performance metric. 

If you were to compute model performance manually (as above), these weights are computed using:

```{r}
#| ipcw-comp
ipcw_data <- 
  test_pred %>% 
  .censoring_weights_graf(bag_fit, .) %>% 
  select(-.pred_time)
```

This adds a column called `.weight_censored` to the tibble of predicted survival probabilities which is used as a case weight in calculating the performance metric.

```{r}
ipcw_data

# The adjusted data:
ipcw_data$.pred[[1]] %>% slice(1:5)
```

With the data in this format, we can use a yardstick function for dynamic metrics like `brier_survival()`: 

```{r}
#| label: brier-surv

brier_scores <-
  ipcw_data %>% 
  # No argument name is used for .pred
  brier_survival(truth = event_time, .pred)
brier_scores %>% slice(1:5)
```

We compute a score for each evaluation time:

```{r}
#| label: brier-surv-plot
#| out-width: "60%"
#| fig-align: center

brier_scores %>% 
  ggplot(aes(.eval_time, .estimate)) +
  geom_hline(yintercept = 0.25, col = "red", alpha = 1 / 2, lty = 2) +
  geom_line() +
  labs(x = "years", y = "Brier Score")
```

The vertical line is the level of performance that you would get with a non-informative model.

The other dynamic metrics that are currently implemented are `brier_survival_integrated()` (for an AUC of the curve above) and `roc_auc_survival().` 

Multiple static and dynamic metrics can be combined via a metric set. 

## Resampling the Model

tidymodels strongly focuses on empirical validation via resampling, which is also true of event time models. 

We can use the `fit_resamples()` function with an rsample object to compute performance without using the test set. We need to tell the function what times to use for the dynamic metrics:

```{r}
#| label: resample

# Create resamples
set.seed(12)
valve_rs <- vfold_cv(valve_tr, repeats = 5)

bag_tree_res <- 
  bag_spec %>% 
  fit_resamples(event_time ~ ., resamples = valve_rs, eval_time = time_points)
```

By default, the Brier score is used:

```{r}
#| label: resampled-brier
#| out-width: "60%"
#| fig-align: center

collect_metrics(bag_tree_res) %>% slice(1:5)

bag_tree_res  %>%
  collect_metrics() %>%
  mutate(
    lower = mean - 1.96 * std_err,
    upper = mean + 1.96 * std_err
  ) %>%
  ggplot(aes(.eval_time)) +
  geom_hline(yintercept = 0.25, col = "red", alpha = 1 / 2, lty = 2) +
  geom_line(aes(y = mean)) +
  geom_ribbon(aes(ymin = lower, ymax = upper),
              col = NA,
              alpha = 1 / 10) +
  labs(x = "years", y = "Brier Score") 
```

## Model Tuning

Suppose we try a regularized Cox model for these data. We'll add a recipe to the analysis and tune a lasso model. The code is pretty standard tidymodels syntax, with the added `eval_time` argument. We'll also use a metric set to include the integrated Brier score, which computes the AUC of the Brier/time curve: 

```{r}
#| label: lasso-model
#| out-width: "60%"
#| fig-align: center

lasso_spec <- 
  proportional_hazards(penalty = tune(), mixture = 0) %>%
  set_engine("glmnet") %>%
  set_mode("censored regression")

lasso_rec <- 
  recipe(event_time ~ ., data = valve_tr) %>%
  step_dummy(all_nominal_predictors()) %>%
  step_zv(all_predictors()) %>%
  step_normalize(all_numeric_predictors())

lasso_wflow <- workflow(lasso_rec, lasso_spec)

surv_metrics <- metric_set(brier_survival_integrated, brier_survival)

lasso_tune_res <-
  lasso_wflow %>%
  tune_grid(
    resamples = valve_rs,
    eval_time = time_points,
    grid = tibble(penalty = 10^seq(-3, -1, length.out = 20)),
    metrics = surv_metrics
  )
```

We can plot the results for that specific metric: 

```{r}
#| label: integrated-tune
#| out-width: "60%"
#| fig-align: center
autoplot(lasso_tune_res, metric = "brier_survival", eval_time = 5)
```

For these plot methods, `eval_time` can be passed in as shown. If a dynamic metric is used and `eval_time` is not set, the function will pick a time near the middle of the range. 

We can also choose the best penalty. If we use an integrated method, no `eval_time` is needed: 

```{r}
best_penalty <- select_best(lasso_tune_res, metric = "brier_survival_integrated")
```

Now we can update the workflow and, assuming that this is the model that we want to keep, evaluate it on the test set: 

```{r}
#| label: finalize
lasso_final_wflow <- 
  lasso_wflow %>% 
  finalize_workflow(best_penalty)

lasso_final_wflow
```

For performance assessment on the test set, you can manually predict it and calculate performance or use `last_fit()` with the original split object to do these steps for you: 

```{r}
#| label: last-fit
test_res <- 
  last_fit(lasso_final_wflow, valve_split, eval_time = time_points)
```

As usual, you can get the test set metrics via: 

```{r}
#| label: test-metrics
collect_metrics(test_res)
```

How do the Brier Score estimates compare between the test set and resampling? 

```{r}
#| label: lasso-brier-curves
#| out-width: "60%"
#| fig-align: center
#| warning: false

collect_metrics(test_res) %>% 
  mutate(estimator = "testing") %>% 
  select(.eval_time, estimator, Brier = .estimate) %>% 
  bind_rows(
    lasso_tune_res %>% 
    collect_metrics() %>% 
      mutate(estimator = "resampling") %>% 
      select(.eval_time, estimator, Brier = mean, penalty) %>% 
      inner_join(best_penalty, by = "penalty")
  ) %>% 
  ggplot(aes(.eval_time)) +
  geom_hline(yintercept = 0.25, col = "red", alpha = 1 / 2, lty = 2) +
  geom_line(aes(y = Brier, col = estimator)) +
  labs(x = "years", y = "Brier Score") +
  scale_color_brewer(palette = "Set2")
```

Good!

## Things still to do

* Update finetune to use `eval_time`
* Update Bayesian analysis methods in tidyposterior
* Update parsnip `augment()` to produce IPCW values. 

## Session Info

```{r}
#| label: session

sessioninfo::session_info()
```

# not marginal plot with grid search

    Code
      autoplot(knn_results, type = "performance")
    Condition
      Error in `autoplot()`:
      ! `type = "performance"` is only used with iterative search results.

---

    Code
      autoplot(knn_results, type = "parameters")
    Condition
      Error in `autoplot()`:
      ! `type = "parameters"` is only used with iterative search results.

# coord_obs_pred

    Removed 1 row containing missing values or values outside the scale range (`geom_point()`).

# regular grid plot

    Code
      autoplot(res)
    Condition
      Error in `autoplot()`:
      ! Only one observation per metric was present. Unable to create meaningful plot.

# evaluation time warning for non-survival model

    Code
      foo <- autoplot(res, metric = "rmse", eval_time = 10)
    Condition
      Warning in `autoplot()`:
      `eval_time` is only used for dynamic survival metrics.

# `metrics` argument works (differing class metric types)

    Code
      compute_metrics(res_auc, m_set_acc)
    Condition
      Error in `compute_metrics()`:
      ! The supplied `metrics` argument has metrics of type "class", while the metrics used to generate predictions only used "prob" metrics.
      i To save predictions for class metrics, generate `x` with metrics of that type.

---

    Code
      compute_metrics(res_acc, m_set_auc)
    Condition
      Error in `compute_metrics()`:
      ! The supplied `metrics` argument has metrics of type "prob", while the metrics used to generate predictions only used "class" metrics.
      i To save predictions for prob metrics, generate `x` with metrics of that type.

# `metrics` argument works (iterative tuning)

    Code
      set.seed(1)
      res_rmse <- tune_bayes(nearest_neighbor("regression", neighbors = tune()), mpg ~
        ., vfold_cv(mtcars, v = 3), metrics = m_set_rmse, control = tune::control_bayes(
        save_pred = TRUE), iter = 2, initial = 3)

---

    Code
      set.seed(1)
      res_both <- tune_bayes(nearest_neighbor("regression", neighbors = tune()), mpg ~
        ., vfold_cv(mtcars, v = 3), metrics = m_set_both, control = tune::control_bayes(
        save_pred = TRUE), iter = 2, initial = 3)

# errors informatively with bad input

    Code
      compute_metrics(res_rmse, metric_set(rsq))
    Condition
      Error in `compute_metrics()`:
      ! `x` must have been generated with the control argument `save_pred = TRUE`.

---

    Code
      compute_metrics("boop", metric_set(rsq))
    Condition
      Error in `compute_metrics()`:
      ! No `compute_metrics()` method exists for a string.

---

    Code
      compute_metrics(res_rmse_save_pred, "wheee")
    Condition
      Error in `compute_metrics()`:
      ! `metrics` must be a metric set.

# evaluation time

    Code
      tune_grid(spec, Surv(time, status) ~ ., resamples = rs, metrics = mtr)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(spec, Surv(time, status) ~ ., resamples = rs, metrics = reg_mtr)
    Condition
      Error in `tune_grid()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_grid(linear_reg(), age ~ ., resamples = rs, metrics = reg_mtr, eval_time = 1)
    Condition
      Warning in `tune_grid()`:
      `eval_time` is only used for models with mode "censored regression".
      Warning:
      No tuning parameters have been detected, performance will be evaluated using the resamples with no tuning.
      Did you want to assign any parameters with a value of `tune()`?
    Output
      # Tuning results
      # 10-fold cross-validation using stratification 
      # A tibble: 10 x 4
         splits           id     .metrics         .notes          
         <list>           <chr>  <list>           <list>          
       1 <split [164/20]> Fold01 <tibble [1 x 4]> <tibble [0 x 4]>
       2 <split [165/19]> Fold02 <tibble [1 x 4]> <tibble [0 x 4]>
       3 <split [165/19]> Fold03 <tibble [1 x 4]> <tibble [0 x 4]>
       4 <split [166/18]> Fold04 <tibble [1 x 4]> <tibble [0 x 4]>
       5 <split [166/18]> Fold05 <tibble [1 x 4]> <tibble [0 x 4]>
       6 <split [166/18]> Fold06 <tibble [1 x 4]> <tibble [0 x 4]>
       7 <split [166/18]> Fold07 <tibble [1 x 4]> <tibble [0 x 4]>
       8 <split [166/18]> Fold08 <tibble [1 x 4]> <tibble [0 x 4]>
       9 <split [166/18]> Fold09 <tibble [1 x 4]> <tibble [0 x 4]>
      10 <split [166/18]> Fold10 <tibble [1 x 4]> <tibble [0 x 4]>

---

    Code
      no_usable_times <- tune_grid(spec, Surv(time, status) ~ ., resamples = rs,
      metrics = mtr, eval_time = c(-1, Inf))
    Condition
      Error:
      ! There were no usable evaluation times (finite, non-missing, and >= 0).

# process_prediction_data errors when recipe drops rows

    Code
      tune:::process_prediction_data(wflow_fit, static)
    Condition
      Error in `tune:::process_prediction_data()`:
      ! Some assessment set rows are not available at prediction time.
      i Consider using `skip = TRUE` on any recipe steps that remove rows to avoid calling them on the assessment set.

# process_prediction_data errors when recipe drops calibration rows

    Code
      tune:::process_prediction_data(wflow_fit, static, source = "cal")
    Condition
      Error in `tune:::process_prediction_data()`:
      ! Some calibration set rows are not available at prediction time.
      i Consider using `skip = TRUE` on any recipe steps that remove rows to avoid calling them on the calibration set.

# process_prediction_data errors when non-recipe preprocessor drops rows

    Code
      tune:::process_prediction_data(wflow_fit, static)
    Condition
      Error in `tune:::process_prediction_data()`:
      ! Some assessment set rows are not available at prediction time.
      i Did your preprocessing steps filter or remove rows?

# workflow size warning

    Code
      set.seed(1)
      warns <- fit_resamples(lm_wflow, resamples = vfold_cv(MTCARS), control = control_resamples(save_workflow = TRUE, workflow_size = 2))
    Message
      i The workflow being saved is large (2.7 MB). If this was not intentional,
      please set the control setting `save_workflow` to be `FALSE` or change the
      threshold for this warning (currently 2 MB) with the `workflow_size` argument.

# control object print methods

    Code
      control_grid()
    Message
      Grid/resamples control object
        `verbose`: FALSE
        `allow_par`: TRUE
        `extract`: NULL
        `save_pred`: FALSE
        `pkgs`: NULL
        `save_workflow`: FALSE
        `event_level`: "first"
        `parallel_over`: NULL
        `backend_options`: NULL
        `workflow_size`: 100

---

    Code
      control_grid(verbose = TRUE, save_pred = TRUE)
    Message
      Grid/resamples control object
        `verbose`: TRUE
        `allow_par`: TRUE
        `extract`: NULL
        `save_pred`: TRUE
        `pkgs`: NULL
        `save_workflow`: FALSE
        `event_level`: "first"
        `parallel_over`: NULL
        `backend_options`: NULL
        `workflow_size`: 100

---

    Code
      control_grid(pkgs = c("pkg1", "pkg2"), extract = I)
    Message
      Grid/resamples control object
        `verbose`: FALSE
        `allow_par`: TRUE
        `extract`: <function>
        `save_pred`: FALSE
        `pkgs`: "pkg1" and "pkg2"
        `save_workflow`: FALSE
        `event_level`: "first"
        `parallel_over`: NULL
        `backend_options`: NULL
        `workflow_size`: 100

---

    Code
      control_bayes()
    Message
      Bayes control object
        `verbose`: FALSE
        `verbose_iter`: FALSE
        `allow_par`: TRUE
        `no_improve`: 10
        `uncertain`: Inf
        `seed`: 51663
        `extract`: NULL
        `save_pred`: FALSE
        `time_limit`: NA
        `pkgs`: NULL
        `save_workflow`: FALSE
        `save_gp_scoring`: FALSE
        `event_level`: "first"
        `parallel_over`: NULL
        `backend_options`: NULL
        `workflow_size`: 100

---

    Code
      control_bayes(verbose_iter = TRUE, no_improve = 5, save_gp_scoring = TRUE)
    Message
      Bayes control object
        `verbose`: FALSE
        `verbose_iter`: TRUE
        `allow_par`: TRUE
        `no_improve`: 5
        `uncertain`: Inf
        `seed`: 2986
        `extract`: NULL
        `save_pred`: FALSE
        `time_limit`: NA
        `pkgs`: NULL
        `save_workflow`: FALSE
        `save_gp_scoring`: TRUE
        `event_level`: "first"
        `parallel_over`: NULL
        `backend_options`: NULL
        `workflow_size`: 100

---

    Code
      control_last_fit()
    Message
      Last fit control object
        `verbose`: FALSE
        `allow_par`: FALSE
        `extract`: <function>
        `save_pred`: TRUE
        `pkgs`: NULL
        `save_workflow`: FALSE
        `event_level`: "first"
        `parallel_over`: NULL
        `backend_options`: NULL
        `workflow_size`: 100

---

    Code
      control_last_fit(verbose = TRUE)
    Message
      Last fit control object
        `verbose`: TRUE
        `allow_par`: FALSE
        `extract`: <function>
        `save_pred`: TRUE
        `pkgs`: NULL
        `save_workflow`: FALSE
        `event_level`: "first"
        `parallel_over`: NULL
        `backend_options`: NULL
        `workflow_size`: 100

# control object print methods with default = TRUE

    Code
      print(control_grid(verbose = TRUE, pkgs = c("pkg1", "pkg2")), default = TRUE)
    Message
      Grid/resamples control object
        `verbose`: TRUE
        `pkgs`: "pkg1" and "pkg2"

---

    Code
      print(control_bayes(verbose_iter = TRUE, no_improve = 5), default = TRUE)
    Message
      Bayes control object
        `verbose_iter`: TRUE
        `no_improve`: 5
        `seed`: 13797

---

    Code
      print(control_last_fit(verbose = TRUE), default = TRUE)
    Message
      Last fit control object
        `verbose`: TRUE
        `extract`: <function>

# preprocessor error doesn't stop grid

    Code
      res_fit <- tune_grid(parsnip::nearest_neighbor("regression", "kknn",
        dist_power = tune()), Sale_Price ~ ., folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | error:   invalid type (list) for variable 'First_Flr_SF'
    Condition

# model error doesn't stop grid

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | error:   invalid type (list) for variable 'First_Flr_SF'
    Condition

# prediction error doesn't stop grid

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | error:   Assigned data `.ind` must be compatible with existing data.
    Condition

# capturing error correctly in notes

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | error:   Error in `step_logging_helper()`:
    Condition

# capturing warning correctly in notes

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | warning: testing warning

# doesn't capturing message in notes

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message

# captures extract errors

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE, extract = extract_error))
    Message
      > A | error:   extract error

# captures kknn R errors

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | error:   NA/NaN/Inf in foreign function call (arg 1)
    Condition

# captures xgboost C errors

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
    Condition

# captures cli styled errors

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | error:   Error in `step_logging_helper()`:
    Condition

# emitter works with errors

    Code
      res_fit <- tune_grid(wf_spec, folds, grid = 2, control = control_grid(
        allow_par = FALSE))
    Message
      > A | error:   Error in `step_logging_helper()`:
    Condition

# mis-specified extract function

    Code
      res_extract_warning <- fit_resamples(wf, boots, control = control_resamples(
        extract = raise_warning))
    Message
      > A | warning: AHHH

---

    Code
      res_extract_error <- fit_resamples(wf, boots, control = control_resamples(
        extract = raise_error))
    Message
      > A | error:   AHHH

---

    Code
      res_extract_both <- fit_resamples(wf, boots, control = control_resamples(
        extract = raise_both))
    Message
      > A | warning: AH
      > B | error:   AHHH

---

    Code
      res_extract_error_once <- fit_resamples(wf, boots, control = control_resamples(
        extract = raise_error_once))
    Message
      > A | error:   oh no

---

    Code
      res_extract_warning
    Output
      # Resampling results
      # Bootstrap sampling 
      # A tibble: 3 x 5
        splits          id         .metrics         .notes           .extracts       
        <list>          <chr>      <list>           <list>           <list>          
      1 <split [32/13]> Bootstrap1 <tibble [2 x 4]> <tibble [1 x 4]> <tibble [1 x 2]>
      2 <split [32/17]> Bootstrap2 <tibble [2 x 4]> <tibble [1 x 4]> <tibble [1 x 2]>
      3 <split [32/11]> Bootstrap3 <tibble [2 x 4]> <tibble [1 x 4]> <tibble [1 x 2]>
      
      There were issues with some computations:
      
        - Warning(s) x3: AHHH
      
      Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      res_extract_error
    Output
      # Resampling results
      # Bootstrap sampling 
      # A tibble: 3 x 5
        splits          id         .metrics         .notes           .extracts       
        <list>          <chr>      <list>           <list>           <list>          
      1 <split [32/13]> Bootstrap1 <tibble [2 x 4]> <tibble [1 x 4]> <tibble [1 x 2]>
      2 <split [32/17]> Bootstrap2 <tibble [2 x 4]> <tibble [1 x 4]> <tibble [1 x 2]>
      3 <split [32/11]> Bootstrap3 <tibble [2 x 4]> <tibble [1 x 4]> <tibble [1 x 2]>
      
      There were issues with some computations:
      
        - Error(s) x3: AHHH
      
      Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      res_extract_both
    Output
      # Resampling results
      # Bootstrap sampling 
      # A tibble: 3 x 5
        splits          id         .metrics         .notes           .extracts       
        <list>          <chr>      <list>           <list>           <list>          
      1 <split [32/13]> Bootstrap1 <tibble [2 x 4]> <tibble [2 x 4]> <tibble [1 x 2]>
      2 <split [32/17]> Bootstrap2 <tibble [2 x 4]> <tibble [2 x 4]> <tibble [1 x 2]>
      3 <split [32/11]> Bootstrap3 <tibble [2 x 4]> <tibble [2 x 4]> <tibble [1 x 2]>
      
      There were issues with some computations:
      
        - Error(s) x3: AHHH
        - Warning(s) x3: AH
      
      Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      res_extract_error_once
    Output
      # Resampling results
      # Bootstrap sampling 
      # A tibble: 3 x 5
        splits          id         .metrics         .notes           .extracts       
        <list>          <chr>      <list>           <list>           <list>          
      1 <split [32/13]> Bootstrap1 <tibble [2 x 4]> <tibble [1 x 4]> <tibble [1 x 2]>
      2 <split [32/17]> Bootstrap2 <tibble [2 x 4]> <tibble [0 x 4]> <tibble [1 x 2]>
      3 <split [32/11]> Bootstrap3 <tibble [2 x 4]> <tibble [0 x 4]> <tibble [1 x 2]>
      
      There were issues with some computations:
      
        - Error(s) x1: oh no
      
      Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      res_extract_warning$.notes[[1]]
    Output
      # A tibble: 1 x 4
        location                               type    note  trace              
        <chr>                                  <chr>   <chr> <list>             
      1 preprocessor 1/1, model 1/1 (extracts) warning AHHH  <rlng_trc [24 x 5]>

---

    Code
      res_extract_error$.extracts[[1]]$.extracts[[1]]
    Output
      [1] "Error in extractor(object): AHHH\n"
      attr(,"class")
      [1] "try-error"
      attr(,"condition")
      <simpleError in extractor(object): AHHH>

---

    Code
      res_extract_error$.notes[[1]]
    Output
      # A tibble: 1 x 4
        location                               type  note  trace              
        <chr>                                  <chr> <chr> <list>             
      1 preprocessor 1/1, model 1/1 (extracts) error AHHH  <rlng_trc [21 x 5]>

---

    Code
      res_extract_both$.extracts[[1]]$.extracts[[1]]
    Output
      [1] "Error in extractor(object): AHHH\n"
      attr(,"class")
      [1] "try-error"
      attr(,"condition")
      <simpleError in extractor(object): AHHH>

---

    Code
      res_extract_error_once$.extracts[[1]]$.extracts[[1]]
    Output
      [1] "Error in extractor(object): oh no\n"
      attr(,"class")
      [1] "try-error"
      attr(,"condition")
      <simpleError in extractor(object): oh no>

# showing notes

    Code
      res_roles <- fit_resamples(role_bp_wflow, rs)
    Message
      > A | error:   The following required column is missing from `new_data`: date.
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      show_notes(res_roles)
    Output
      unique notes:
      ---------------------------------------------------------------
      The following required column is missing from `new_data`: date.

---

    Code
      res_simple <- fit_resamples(simple_wflow, rs)
    Message
      > A | warning: prediction from rank-deficient fit; consider predict(., rankdeficient="NA")

---

    Code
      show_notes(res_simple)
    Output
      unique notes:
      ---------------------------------------------------------------------------
      prediction from rank-deficient fit; consider predict(., rankdeficient="NA")

---

    Code
      show_notes(.Last.tune.result)
    Output
      Great job! No notes to show.

---

    Code
      fit_lr <- fit_resamples(parsnip::logistic_reg(), class ~ ., rs)
    Message
      > A | warning: glm.fit: algorithm did not converge
      > B | warning: glm.fit: fitted probabilities numerically 0 or 1 occurred
      > C | warning: No control observations were detected in `truth` with control level 'class_2'.

---

    Code
      show_notes(fit_lr)
    Output
      unique notes:
      -----------------------------------
      glm.fit: algorithm did not converge
      -----------------------------------
      glm.fit: fitted probabilities numerically 0 or 1 occurred
      -----------------------------------
      No control observations were detected in `truth` with control level 'class_2'.

# GP fit - svm - failure

    Code
      svm_gp <- tune:::fit_gp(collect_metrics(svm_results), pset = svm_set, metric = "accuracy",
      control = control_bayes(verbose = TRUE))

---

    Code
      svm_scores <- tune:::pred_gp(svm_gp, pset = svm_set, size = 20, current = curr,
        control = control_bayes(verbose_iter = TRUE))
    Message
      i Generating a candidate as far away from existing points as possible.

# GP scoring with failed model

    Code
      svm_gp <- tune:::fit_gp(collect_metrics(svm_results), pset = svm_set, metric = "accuracy",
      control = ctrl)
    Message
      (x) GP has a LOO R² of -6.1% and is unreliable.
      v Gaussian process model failed

---

    Code
      svm_scores <- tune:::pred_gp(svm_gp, pset = svm_set, size = 20, current = curr,
        control = ctrl)
    Message
      i Generating a candidate as far away from existing points as possible.

# pick_candidate() emits uncertainty sample message when verbose

    Code
      set.seed(1)
      res <- tune:::pick_candidate(results, info, ctrl)
    Message
      i Uncertainty sample

# GP fit - knn

    Code
      set.seed(1)
      knn_scores <- tune:::pred_gp(knn_gp, pset = knn_set, size = 20, current = mutate(
        knn_mtr, .iter = 0), control = control_bayes())

# bad inputs

    Code
      filter_parameters(collect_metrics(svm_reg_results), parameters = tibble::tibble(
        `%^*#` = 1))
    Condition
      Error in `filter_parameters()`:
      ! `collect_metrics(svm_reg_results)` should have class <tune_results>; a tibble was passed.

---

    Code
      filter_parameters(svm_reg_results, parameters = tibble::tibble(soup = 1))
    Condition
      Error in `filter_by_join()`:
      ! There are no columns in `parameters` that match with "svm_reg_results".

---

    Code
      filter_parameters(svm_reg_results, tibble::tibble(soup = 1))
    Condition
      Error in `filter_parameters()`:
      ! An element passed to `...` is a data frame rather than a filter expression.
      i Did you forget to name the `parameters` argument?

---

    Code
      filter_parameters(svm_reg_results, cost < 1, tibble::tibble(soup = 1))
    Condition
      Error in `filter_parameters()`:
      ! An element passed to `...` is a data frame rather than a filter expression.
      i Did you forget to name the `parameters` argument?

---

    Code
      filter_parameters(svm_reg_results, parameters = tibble::tibble(`%^*#` = 1 / 3))
    Condition
      Error in `filter_by_join()`:
      ! No parameter combinations were selected using your subset.

---

    Code
      filter_parameters(svm_reg_results, parameters = tibble::tibble(`%^*#` = 1,
        soup = 2))
    Condition
      Warning:
      The column `soup` passed in `parameters` is not needed and will be ignored.
    Output
      # Tuning results
      # 10-fold cross-validation repeated 5 times 
      # A tibble: 50 x 6
         splits           id      id2    .metrics          .notes   .predictions      
         <list>           <chr>   <chr>  <list>            <list>   <list>            
       1 <split [711/80]> Repeat1 Fold01 <tibble [30 x 7]> <tibble> <tibble [800 x 7]>
       2 <split [712/79]> Repeat1 Fold02 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
       3 <split [712/79]> Repeat1 Fold03 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
       4 <split [712/79]> Repeat1 Fold04 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
       5 <split [712/79]> Repeat1 Fold05 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
       6 <split [712/79]> Repeat1 Fold06 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
       7 <split [712/79]> Repeat1 Fold07 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
       8 <split [712/79]> Repeat1 Fold08 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
       9 <split [712/79]> Repeat1 Fold09 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
      10 <split [712/79]> Repeat1 Fold10 <tibble [30 x 7]> <tibble> <tibble [790 x 7]>
      # i 40 more rows

---

    Code
      res <- filter_parameters(svm_reg_results, parameters = tibble::tibble(`%^*#` = 1,
        soup = 2, boop = 3))
    Condition
      Warning:
      The columns `soup` and `boop` passed in `parameters` are not needed and will be ignored.

---

    Code
      res <- filter_parameters(svm_reg_results, parameters = tibble::tibble(`%^*#` = 1,
        soup = 2, boop = 3, loop = 4))
    Condition
      Warning:
      The columns `soup`, `boop`, and `loop` passed in `parameters` are not needed and will be ignored.

# conf_bound interface

    Code
      conf_bound("a")
    Condition
      Error in `conf_bound()`:
      ! `kappa` should be a number or a function.

---

    Code
      conf_bound(function() 1)
    Condition
      Error in `conf_bound()`:
      ! The `trade_off()` function should have at least one argument.

---

    Code
      predict(conf_bound(), test_res, maximize = 2, iter = 1)
    Condition
      Error in `predict()`:
      ! `maximize` must be `TRUE` or `FALSE`, not the number 2.

# prob_improve interface

    Code
      prob_improve("a")
    Condition
      Error in `prob_improve()`:
      ! `trade_off` should be a number or a function.

---

    Code
      prob_improve(function() 1)
    Condition
      Error in `prob_improve()`:
      ! The `trade_off()` function should have at least one argument.

---

    Code
      predict(prob_improve(), test_res, maximize = 2, iter = 1)
    Condition
      Error in `predict()`:
      ! `maximize` must be `TRUE` or `FALSE`, not the number 2.

---

    Code
      predict(prob_improve(), test_res, maximize = TRUE, iter = 1, best = NA)
    Condition
      Error in `predict()`:
      ! `best` must be a number, not `NA`.

---

    Code
      predict(prob_improve(), test_res, maximize = TRUE, iter = 1, best = "WAT")
    Condition
      Error in `predict()`:
      ! `best` must be a number, not the string "WAT".

# exp_improve interface

    Code
      exp_improve("a")
    Condition
      Error in `exp_improve()`:
      ! `trade_off` should be a number or a function.

---

    Code
      exp_improve(function() 2)
    Condition
      Error in `exp_improve()`:
      ! The `trade_off()` function should have at least one argument.

---

    Code
      predict(exp_improve(), test_res, maximize = 2, iter = 1)
    Condition
      Error in `predict()`:
      ! `maximize` must be `TRUE` or `FALSE`, not the number 2.

---

    Code
      predict(exp_improve(), test_res, maximize = TRUE, iter = 1, best = NA)
    Condition
      Error in `predict()`:
      ! `best` must be a number, not `NA`.

---

    Code
      predict(exp_improve(), test_res, maximize = TRUE, iter = 1, best = "WAT")
    Condition
      Error in `predict()`:
      ! `best` must be a number, not the string "WAT".

# tune model only - failure in recipe is caught elegantly

    Code
      cars_res <- tune_grid(helper_objects$svm_mod, preprocessor = rec, resamples = data_folds,
      grid = cars_grid, control = control_grid(extract = function(x) {
        1
      }, save_pred = TRUE))
    Message
      > A | error:   Error in `step_spline_b()`:
                     Caused by error in `prep()`:
                     ! `deg_free` must be a whole number, not a numeric `NA`.
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.

# tune model only - failure in formula is caught elegantly

    Code
      cars_res <- tune_grid(helper_objects$svm_mod, y ~ z, resamples = data_folds,
      grid = cars_grid, control = control_grid(extract = function(x) {
        1
      }, save_pred = TRUE))
    Message
      > A | error:   The following predictor was not found in `data`: "z".
    Condition

# argument order gives errors for recipes

    Code
      tune_grid(helper_objects$rec_tune_1, helper_objects$lm_mod, rsample::vfold_cv(
        mtcars, v = 2))
    Condition
      Error in `tune_grid()`:
      ! The first argument to `tune_grid()` should be either a model or workflow, not a <recipe> object.

# argument order gives errors for formula

    Code
      tune_grid(mpg ~ ., helper_objects$lm_mod, rsample::vfold_cv(mtcars, v = 2))
    Condition
      Error in `tune_grid()`:
      ! The first argument to `tune_grid()` should be either a model or workflow, not a <formula> object.

# ellipses with tune_grid

    Code
      tune_grid(wflow, resamples = folds, grid = 3, something = "wrong")
    Condition
      Warning:
      The `...` are not used in this function but 1 object was passed: "something"
    Output
      # Tuning results
      # 10-fold cross-validation 
      # A tibble: 10 x 4
         splits         id     .metrics         .notes          
         <list>         <chr>  <list>           <list>          
       1 <split [28/4]> Fold01 <tibble [6 x 5]> <tibble [0 x 4]>
       2 <split [28/4]> Fold02 <tibble [6 x 5]> <tibble [0 x 4]>
       3 <split [29/3]> Fold03 <tibble [6 x 5]> <tibble [0 x 4]>
       4 <split [29/3]> Fold04 <tibble [6 x 5]> <tibble [0 x 4]>
       5 <split [29/3]> Fold05 <tibble [6 x 5]> <tibble [0 x 4]>
       6 <split [29/3]> Fold06 <tibble [6 x 5]> <tibble [0 x 4]>
       7 <split [29/3]> Fold07 <tibble [6 x 5]> <tibble [0 x 4]>
       8 <split [29/3]> Fold08 <tibble [6 x 5]> <tibble [0 x 4]>
       9 <split [29/3]> Fold09 <tibble [6 x 5]> <tibble [0 x 4]>
      10 <split [29/3]> Fold10 <tibble [6 x 5]> <tibble [0 x 4]>

# finalize_workflow_preprocessor() is deprecated

    Code
      wflow <- finalize_workflow_preprocessor(wflow, grid)
    Condition
      Warning:
      `finalize_workflow_preprocessor()` was deprecated in tune 2.1.0.

# low-level messages

    Code
      siren("a", "werewolf")
    Condition
      Error in `match.arg()`:
      ! 'arg' should be one of "warning", "go", "danger", "success", "info"

---

    Code
      siren("bat", "info")
    Message
      i bat

---

    Code
      siren("bat", "go")
    Message
      > bat

---

    Code
      siren("bat", "danger")
    Message
      x bat

---

    Code
      siren("bat", "warning")
    Message
      ! bat

---

    Code
      siren("bat", "success")
    Message
      v bat

# update_printer

    Code
      update_printer(ctrl_t, rs, task = "cube", type = "go")
    Message
      > Fold01: cube
    Output
      NULL

---

    Code
      update_printer(ctrl_t, NULL, task = "cube", type = "go")
    Message
      > cube
    Output
      NULL

---

    Code
      update_printer(ctrl_t, rs, task = "cube", type = "success")
    Message
      v Fold01: cube
    Output
      NULL

# logging iterations

    Code
      log_best(ctrl_t, 10, sc_1)
    Message
      
      -- Iteration 10 ----------------------------------------------------------------
      
      i Current best:		0.8=7 (@iter 2)
    Output
      NULL

# logging search info

    Code
      check_and_log_flow(ctrl_t, mutate(tb_1, .mean = .mean * NA))
    Message
      x Skipping to next iteration
    Condition
      Error:
      ! no loop for break/next, jumping to top level

---

    Code
      check_and_log_flow(ctrl_t, slice(mutate(tb_1, .mean = .mean * NA), 1))
    Message
      x Halting search
    Condition
      Error:
      ! no loop for break/next, jumping to top level

# current results

    Code
      log_progress(ctrl_t, tb_2, maximize = FALSE, objective = "a")
    Message
      (x) Newest results:	a=4 (+/-0.4)

---

    Code
      log_progress(ctrl_t, tb_2, maximize = TRUE, objective = "b")
    Message
      <3 Newest results:	b=8 (+/-0.8)

---

    Code
      log_progress(ctrl_t, tb_2, maximize = TRUE, objective = "a")
    Message
      <3 Newest results:	a=4 (+/-0.4)

# show parameters

    Code
      param_msg(ctrl_t, iris[1, 4:5])
    Message
      i Petal.Width=0.2, Species=setosa

# acquisition functions

    Code
      acq_summarizer(ctrl_t, 1, conf_bound(I))
    Message
      i Kappa value: 1

---

    Code
      acq_summarizer(ctrl_t, 1, exp_improve(I))
    Message
      i Trade-off value: 1

---

    Code
      acq_summarizer(ctrl_t, 1, prob_improve(I))
    Message
      i Trade-off value: 1

# interactive logger works (fit_resamples, warning + error)

    Code
      res_fit <- fit_resamples(parsnip::nearest_neighbor("regression", "kknn"),
      Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5), control = control_resamples(
        extract = function(x) {
          raise_warning()
          raise_error()
        }))
    Message
      > A | warning: ope! yikes.
      > B | error:   AHHhH

---

    Code
      catalog_summary_test
    Output
      A: x5   B: x5

# interactive logger works (fit_resamples, rlang warning + error)

    Code
      res_fit <- fit_resamples(parsnip::nearest_neighbor("regression", "kknn"),
      Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5), control = control_resamples(
        extract = function(x) {
          raise_warning_rl()
          raise_error_rl()
        }))
    Message
      > A | warning: ope! yikes. (but rlang)
      > B | error:   AHHhH (but rlang)

---

    Code
      catalog_summary_test
    Output
      A: x5   B: x5

# interactive logger works (fit_resamples, multiline)

    Code
      res_fit <- fit_resamples(parsnip::nearest_neighbor("regression", "kknn"),
      Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5), control = control_resamples(
        extract = raise_multiline_conditions))
    Message
      > A | warning: hmmm what's happening
      > B | error:   aHHHksdjvndiuf

---

    Code
      catalog_summary_test
    Output
      A: x5   B: x5

# interactive logger works (fit_resamples, occasional error)

    Code
      res_fit <- fit_resamples(parsnip::nearest_neighbor("regression", "kknn"),
      Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5), control = control_resamples(
        extract = later))
    Message
      > A | error:   this errors now! ha!

---

    Code
      catalog_summary_test
    Output
      A: x2

# interactive logger works (fit_resamples, occasional errors)

    Code
      res_fit <- fit_resamples(parsnip::nearest_neighbor("regression", "kknn"),
      Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 10),
      control = control_resamples(extract = function(x) {
        once()
        later()
      }))
    Message
      > A | error:   oh no
      > B | error:   this errors now! ha!

---

    Code
      catalog_summary_test
    Output
      A: x1   B: x6

# interactive logger works (fit_resamples, many distinct errors)

    Code
      res_fit <- fit_resamples(parsnip::nearest_neighbor("regression", "kknn"),
      Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(72, 40:45)], 5), control = control_resamples(
        extract = numbered))
    Message
      > A | error:   error number 1
      > B | error:   error number 2
      > C | error:   error number 3
      > D | error:   error number 4
      > E | error:   error number 5

---

    Code
      catalog_summary_test
    Output
      A: x1   B: x1   C: x1   D: x1   E: x1

# interactive logger works (tune grid, error)

    Code
      res_fit <- tune_grid(parsnip::nearest_neighbor("regression", "kknn",
        dist_power = tune()), Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(
        72, 40:45)], 5), grid = 5, control = control_grid(extract = raise_error))
    Message
      > A | error:   AHHhH

---

    Code
      catalog_summary_test
    Output
      A: x75

# interactive logger works (bayesian, error)

    Code
      res_grid <- tune_bayes(parsnip::nearest_neighbor("regression", "kknn",
        dist_power = tune()), Sale_Price ~ ., rsample::vfold_cv(modeldata::ames[, c(
        72, 40:45)], 5), initial = 5, iter = 5, control = control_bayes(extract = raise_error))
    Message
      > A | error:   AHHhH

---

    Code
      catalog_summary_test
    Output
      A: x100

# select_best()

    Code
      select_best(rcv_results, metric = "random")
    Condition
      Error in `select_best()`:
      ! "random" was not in the metric set. Please choose from: "rmse" and "rsq".

---

    Code
      select_best(rcv_results, metric = c("rmse", "rsq"))
    Condition
      Warning in `select_best()`:
      2 metrics were given; "rmse" will be used.
    Output
      # A tibble: 1 x 5
        deg_free degree `wt df` `wt degree` .config         
           <int>  <int>   <int>       <int> <chr>           
      1        6      2       2           1 pre19_mod0_post0

---

    Code
      best_default_metric <- select_best(rcv_results)
    Condition
      Warning in `select_best()`:
      No value of `metric` was given; "rmse" will be used.
    Code
      best_rmse <- select_best(rcv_results, metric = "rmse")

---

    Code
      select_best(mtcars, metric = "disp")
    Condition
      Error in `select_best()`:
      ! No `select_best()` exists for this type of object.

# show_best()

    Code
      best_default_metric <- show_best(rcv_results)
    Condition
      Warning in `show_best()`:
      No value of `metric` was given; "rmse" will be used.
    Code
      best_rmse <- show_best(rcv_results, metric = "rmse")

---

    Code
      show_best(mtcars, metric = "disp")
    Condition
      Error in `show_best()`:
      ! No `show_best()` exists for this type of object.

# one-std error rule

    Code
      select_by_one_std_err(rcv_results, metric = "random", deg_free)
    Condition
      Error in `select_by_one_std_err()`:
      ! "random" was not in the metric set. Please choose from: "rmse" and "rsq".

---

    Code
      select_by_one_std_err(rcv_results, metric = c("rmse", "rsq"), deg_free)
    Condition
      Warning in `select_by_one_std_err()`:
      2 metrics were given; "rmse" will be used.
    Output
      # A tibble: 1 x 5
        deg_free degree `wt df` `wt degree` .config         
           <int>  <int>   <int>       <int> <chr>           
      1        6      2       2           1 pre19_mod0_post0

---

    Code
      select_via_default_metric <- select_by_one_std_err(knn_results, K)
    Condition
      Warning in `select_by_one_std_err()`:
      No value of `metric` was given; "roc_auc" will be used.
    Code
      select_via_roc <- select_by_one_std_err(knn_results, K, metric = "roc_auc")

---

    Code
      select_by_one_std_err(rcv_results, metric = "random")
    Condition
      Error in `select_by_one_std_err()`:
      ! "random" was not in the metric set. Please choose from: "rmse" and "rsq".

---

    Code
      select_by_one_std_err(mtcars, metric = "disp")
    Condition
      Error in `select_by_one_std_err()`:
      ! No `select_by_one_std_err()` exists for this type of object.

---

    Code
      select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk)
    Condition
      Error in `select_by_one_std_err()`:
      ! Could not sort results by `weight_funk`.

---

    Code
      select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk, K)
    Condition
      Error in `select_by_one_std_err()`:
      ! Could not sort results by `weight_funk`.

---

    Code
      select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk, Kay)
    Condition
      Error in `select_by_one_std_err()`:
      ! Could not sort results by `weight_funk` and `Kay`.

---

    Code
      select_by_one_std_err(knn_results, metric = "roc_auc", weight_funk, desc(K))
    Condition
      Error in `select_by_one_std_err()`:
      ! Could not sort results by `weight_funk` and `desc(K)`.

# percent loss

    Code
      select_by_pct_loss(rcv_results, metric = "random", deg_free)
    Condition
      Error in `select_by_pct_loss()`:
      ! "random" was not in the metric set. Please choose from: "rmse" and "rsq".

---

    Code
      select_by_pct_loss(rcv_results, metric = c("rmse", "rsq"), deg_free)
    Condition
      Warning in `select_by_pct_loss()`:
      2 metrics were given; "rmse" will be used.
    Output
      # A tibble: 1 x 5
        deg_free degree `wt df` `wt degree` .config         
           <int>  <int>   <int>       <int> <chr>           
      1        6      2       2           1 pre19_mod0_post0

---

    Code
      select_via_default_metric <- select_by_pct_loss(knn_results, K)
    Condition
      Warning in `select_by_pct_loss()`:
      No value of `metric` was given; "roc_auc" will be used.
    Code
      select_via_roc <- select_by_pct_loss(knn_results, K, metric = "roc_auc")

---

    Code
      select_by_pct_loss(rcv_results, metric = "random")
    Condition
      Error in `select_by_pct_loss()`:
      ! "random" was not in the metric set. Please choose from: "rmse" and "rsq".

---

    Code
      select_by_pct_loss(mtcars, metric = "disp")
    Condition
      Error in `select_by_pct_loss()`:
      ! No `select_by_pct_loss()` exists for this type of object.

---

    Code
      select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk)
    Condition
      Error in `select_by_pct_loss()`:
      ! Could not sort results by `weight_funk`.

---

    Code
      select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk, K)
    Condition
      Error in `select_by_pct_loss()`:
      ! Could not sort results by `weight_funk`.

---

    Code
      select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk, Kay)
    Condition
      Error in `select_by_pct_loss()`:
      ! Could not sort results by `weight_funk` and `Kay`.

---

    Code
      select_by_pct_loss(knn_results, metric = "roc_auc", weight_funk, desc(K))
    Condition
      Error in `select_by_pct_loss()`:
      ! Could not sort results by `weight_funk` and `desc(K)`.

# show_best with survival models

    Code
      show_best(surv_res)
    Condition
      Warning in `show_best()`:
      No value of `metric` was given; "brier_survival" will be used.
      Warning in `show_best()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 5 x 8
        trees .metric        .estimator .eval_time  mean     n std_err .config        
        <dbl> <chr>          <chr>           <dbl> <dbl> <int>   <dbl> <chr>          
      1   100 brier_survival standard           10 0.118    10  0.0177 pre0_mod5_post0
      2    20 brier_survival standard           10 0.136    10  0.0153 pre0_mod4_post0
      3    10 brier_survival standard           10 0.155    10  0.0175 pre0_mod3_post0
      4     5 brier_survival standard           10 0.172    10  0.0198 pre0_mod2_post0
      5     1 brier_survival standard           10 0.194    10  0.0221 pre0_mod1_post0

---

    Code
      show_best(surv_res, metric = "concordance_survival")
    Output
      # A tibble: 5 x 8
        trees .metric              .estimator .eval_time  mean     n std_err .config  
        <dbl> <chr>                <chr>           <dbl> <dbl> <int>   <dbl> <chr>    
      1    20 concordance_survival standard           NA 0.677    10  0.0354 pre0_mod~
      2   100 concordance_survival standard           NA 0.670    10  0.0329 pre0_mod~
      3    10 concordance_survival standard           NA 0.668    10  0.0376 pre0_mod~
      4     5 concordance_survival standard           NA 0.663    10  0.0363 pre0_mod~
      5     1 concordance_survival standard           NA 0.626    10  0.0326 pre0_mod~

---

    Code
      show_best(surv_res, metric = "brier_survival_integrated")
    Output
      # A tibble: 5 x 8
        trees .metric                .estimator .eval_time  mean     n std_err .config
        <dbl> <chr>                  <chr>           <dbl> <dbl> <int>   <dbl> <chr>  
      1   100 brier_survival_integr~ standard           NA 0.117    10 0.00699 pre0_m~
      2    20 brier_survival_integr~ standard           NA 0.127    10 0.00981 pre0_m~
      3    10 brier_survival_integr~ standard           NA 0.141    10 0.0134  pre0_m~
      4     5 brier_survival_integr~ standard           NA 0.151    10 0.0159  pre0_m~
      5     1 brier_survival_integr~ standard           NA 0.164    10 0.0182  pre0_m~

---

    Code
      show_best(surv_res, metric = "brier_survival")
    Condition
      Warning in `show_best()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 5 x 8
        trees .metric        .estimator .eval_time  mean     n std_err .config        
        <dbl> <chr>          <chr>           <dbl> <dbl> <int>   <dbl> <chr>          
      1   100 brier_survival standard           10 0.118    10  0.0177 pre0_mod5_post0
      2    20 brier_survival standard           10 0.136    10  0.0153 pre0_mod4_post0
      3    10 brier_survival standard           10 0.155    10  0.0175 pre0_mod3_post0
      4     5 brier_survival standard           10 0.172    10  0.0198 pre0_mod2_post0
      5     1 brier_survival standard           10 0.194    10  0.0221 pre0_mod1_post0

---

    Code
      show_best(surv_res, metric = c("brier_survival", "roc_auc_survival"))
    Condition
      Warning in `show_best()`:
      2 metrics were given; "brier_survival" will be used.
      Warning in `show_best()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 5 x 8
        trees .metric        .estimator .eval_time  mean     n std_err .config        
        <dbl> <chr>          <chr>           <dbl> <dbl> <int>   <dbl> <chr>          
      1   100 brier_survival standard           10 0.118    10  0.0177 pre0_mod5_post0
      2    20 brier_survival standard           10 0.136    10  0.0153 pre0_mod4_post0
      3    10 brier_survival standard           10 0.155    10  0.0175 pre0_mod3_post0
      4     5 brier_survival standard           10 0.172    10  0.0198 pre0_mod2_post0
      5     1 brier_survival standard           10 0.194    10  0.0221 pre0_mod1_post0

---

    Code
      show_best(surv_res, metric = "brier_survival", eval_time = 1)
    Output
      # A tibble: 5 x 8
        trees .metric        .estimator .eval_time   mean     n std_err .config       
        <dbl> <chr>          <chr>           <dbl>  <dbl> <int>   <dbl> <chr>         
      1    20 brier_survival standard            1 0.0381    10  0.0149 pre0_mod4_pos~
      2    10 brier_survival standard            1 0.0386    10  0.0151 pre0_mod3_pos~
      3   100 brier_survival standard            1 0.0386    10  0.0147 pre0_mod5_pos~
      4     5 brier_survival standard            1 0.0389    10  0.0152 pre0_mod2_pos~
      5     1 brier_survival standard            1 0.0392    10  0.0153 pre0_mod1_pos~

---

    Code
      show_best(surv_res, metric = "concordance_survival", eval_time = 1)
    Condition
      Warning in `show_best()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 5 x 8
        trees .metric              .estimator .eval_time  mean     n std_err .config  
        <dbl> <chr>                <chr>           <dbl> <dbl> <int>   <dbl> <chr>    
      1    20 concordance_survival standard           NA 0.677    10  0.0354 pre0_mod~
      2   100 concordance_survival standard           NA 0.670    10  0.0329 pre0_mod~
      3    10 concordance_survival standard           NA 0.668    10  0.0376 pre0_mod~
      4     5 concordance_survival standard           NA 0.663    10  0.0363 pre0_mod~
      5     1 concordance_survival standard           NA 0.626    10  0.0326 pre0_mod~

---

    Code
      show_best(surv_res, metric = "concordance_survival", eval_time = 1.1)
    Condition
      Warning in `show_best()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 5 x 8
        trees .metric              .estimator .eval_time  mean     n std_err .config  
        <dbl> <chr>                <chr>           <dbl> <dbl> <int>   <dbl> <chr>    
      1    20 concordance_survival standard           NA 0.677    10  0.0354 pre0_mod~
      2   100 concordance_survival standard           NA 0.670    10  0.0329 pre0_mod~
      3    10 concordance_survival standard           NA 0.668    10  0.0376 pre0_mod~
      4     5 concordance_survival standard           NA 0.663    10  0.0363 pre0_mod~
      5     1 concordance_survival standard           NA 0.626    10  0.0326 pre0_mod~

---

    Code
      show_best(surv_res, metric = "brier_survival", eval_time = 1.1)
    Condition
      Error in `show_best()`:
      ! Evaluation time 1.1 is not in the results.

---

    Code
      show_best(surv_res, metric = "brier_survival", eval_time = 1:2)
    Condition
      Warning in `show_best()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 1`).
    Output
      # A tibble: 5 x 8
        trees .metric        .estimator .eval_time   mean     n std_err .config       
        <dbl> <chr>          <chr>           <dbl>  <dbl> <int>   <dbl> <chr>         
      1    20 brier_survival standard            1 0.0381    10  0.0149 pre0_mod4_pos~
      2    10 brier_survival standard            1 0.0386    10  0.0151 pre0_mod3_pos~
      3   100 brier_survival standard            1 0.0386    10  0.0147 pre0_mod5_pos~
      4     5 brier_survival standard            1 0.0389    10  0.0152 pre0_mod2_pos~
      5     1 brier_survival standard            1 0.0392    10  0.0153 pre0_mod1_pos~

---

    Code
      show_best(surv_res, metric = "brier_survival", eval_time = 3:4)
    Condition
      Warning in `show_best()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 3`).
      Error in `show_best()`:
      ! Evaluation time 3 is not in the results.

# select_best with survival models

    Code
      select_best(surv_res)
    Condition
      Warning in `select_best()`:
      No value of `metric` was given; "brier_survival" will be used.
      Warning in `select_best()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_best(surv_res, metric = "concordance_survival")
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_best(surv_res, metric = "brier_survival_integrated")
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_best(surv_res, metric = "brier_survival")
    Condition
      Warning in `select_best()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_best(surv_res, metric = c("brier_survival", "roc_auc_survival"))
    Condition
      Warning in `select_best()`:
      2 metrics were given; "brier_survival" will be used.
      Warning in `select_best()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_best(surv_res, metric = "brier_survival", eval_time = 1)
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_best(surv_res, metric = "concordance_survival", eval_time = 1)
    Condition
      Warning in `select_best()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_best(surv_res, metric = "concordance_survival", eval_time = 1.1)
    Condition
      Warning in `select_best()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_best(surv_res, metric = "brier_survival", eval_time = 1.1)
    Condition
      Error in `select_best()`:
      ! Evaluation time 1.1 is not in the results.

---

    Code
      select_best(surv_res, metric = "brier_survival", eval_time = 1:2)
    Condition
      Warning in `select_best()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 1`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_best(surv_res, metric = "brier_survival", eval_time = 3:4)
    Condition
      Warning in `select_best()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 3`).
      Error in `select_best()`:
      ! Evaluation time 3 is not in the results.

# select_by_one_std_err with survival models

    Code
      select_by_one_std_err(surv_res, trees)
    Condition
      Warning in `select_by_one_std_err()`:
      No value of `metric` was given; "brier_survival" will be used.
      Warning in `select_by_one_std_err()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "concordance_survival", trees)
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1     5 pre0_mod2_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "brier_survival_integrated", trees)
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "brier_survival", trees)
    Condition
      Warning in `select_by_one_std_err()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_by_one_std_err(surv_res, metric = c("brier_survival", "roc_auc_survival"),
      trees)
    Condition
      Warning in `select_by_one_std_err()`:
      2 metrics were given; "brier_survival" will be used.
      Warning in `select_by_one_std_err()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    20 pre0_mod4_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "brier_survival", eval_time = 1, trees)
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1     1 pre0_mod1_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "concordance_survival", eval_time = 1,
        trees)
    Condition
      Warning in `select_by_one_std_err()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1     5 pre0_mod2_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "concordance_survival", eval_time = 1.1,
        trees)
    Condition
      Warning in `select_by_one_std_err()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1     5 pre0_mod2_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "brier_survival", eval_time = 1.1,
        trees)
    Condition
      Error in `select_by_one_std_err()`:
      ! Evaluation time 1.1 is not in the results.

---

    Code
      select_by_one_std_err(surv_res, metric = "brier_survival", eval_time = 1:2,
      trees)
    Condition
      Warning in `select_by_one_std_err()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 1`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1     1 pre0_mod1_post0

---

    Code
      select_by_one_std_err(surv_res, metric = "brier_survival", eval_time = 3:4,
      trees)
    Condition
      Warning in `select_by_one_std_err()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 3`).
      Error in `select_by_one_std_err()`:
      ! Evaluation time 3 is not in the results.

# select_by_pct_loss with survival models

    Code
      select_by_pct_loss(surv_res, trees)
    Condition
      Warning in `select_by_pct_loss()`:
      No value of `metric` was given; "brier_survival" will be used.
      Warning in `select_by_pct_loss()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "concordance_survival", trees)
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    10 pre0_mod3_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "brier_survival_integrated", trees)
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "brier_survival", trees)
    Condition
      Warning in `select_by_pct_loss()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_by_pct_loss(surv_res, metric = c("brier_survival", "roc_auc_survival"),
      trees)
    Condition
      Warning in `select_by_pct_loss()`:
      2 metrics were given; "brier_survival" will be used.
      Warning in `select_by_pct_loss()`:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1   100 pre0_mod5_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "brier_survival", eval_time = 1, trees)
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1     5 pre0_mod2_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "concordance_survival", eval_time = 1,
        trees)
    Condition
      Warning in `select_by_pct_loss()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    10 pre0_mod3_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "concordance_survival", eval_time = 1.1,
        trees)
    Condition
      Warning in `select_by_pct_loss()`:
      `eval_time` is only used for dynamic survival metrics.
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1    10 pre0_mod3_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "brier_survival", eval_time = 1.1, trees)
    Condition
      Error in `select_by_pct_loss()`:
      ! Evaluation time 1.1 is not in the results.

---

    Code
      select_by_pct_loss(surv_res, metric = "brier_survival", eval_time = 1:2, trees)
    Condition
      Warning in `select_by_pct_loss()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 1`).
    Output
      # A tibble: 1 x 2
        trees .config        
        <dbl> <chr>          
      1     5 pre0_mod2_post0

---

    Code
      select_by_pct_loss(surv_res, metric = "brier_survival", eval_time = 3:4, trees)
    Condition
      Warning in `select_by_pct_loss()`:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 3`).
      Error in `select_by_pct_loss()`:
      ! Evaluation time 3 is not in the results.

# fit_best

    Code
      fit_best(knn_pca_res, verbose = TRUE)
    Output
      Using rmse as the metric, the optimal parameters were:
        neighbors: 1
        num_comp:  4
      
    Message
      i Fitting using 161 data points...
      v Done.
    Output
      == Workflow [trained] ==========================================================
      Preprocessor: Recipe
      Model: nearest_neighbor()
      
      -- Preprocessor ----------------------------------------------------------------
      1 Recipe Step
      
      * step_pca()
      
      -- Model -----------------------------------------------------------------------
      
      Call:
      kknn::train.kknn(formula = ..y ~ ., data = data, ks = min_rows(1L,     data, 5))
      
      Type of response variable: continuous
      minimal mean absolute error: 1.015528
      Minimal mean squared error: 2.448261
      Best kernel: optimal
      Best k: 1

---

    Code
      tmp <- fit_best(knn_pca_res, verbose = TRUE, parameters = tibble(neighbors = 1,
        num_comp = 1))
    Message
      i Fitting using 161 data points...
      v Done.

---

    No `fit_best()` exists for an integer.

---

    No `fit_best()` exists for a tibble.

---

    "WAT" was not in the metric set. Please choose from: "rmse" and "rsq".

---

    The parameters `neighbors` and `num_comp` are still marked for tuning.

---

    The parameter `num_comp` is still marked for tuning.

---

    `...` must be empty.
    x Problematic argument:
    * chickens = 2

---

    x The control option `save_workflow = TRUE` should be used when tuning.

# fit_best() warns when metric or eval_time are specified in addition to parameters

    Code
      manual_wf <- fit_best(res, metric = "rmse", parameters = tune_params)
    Condition
      Warning:
      `metric` is being ignored because `parameters` has been specified.

---

    Code
      manual_wf <- fit_best(res, metric = "rmse", eval_time = 10, parameters = tune_params)
    Condition
      Warning:
      `metric` is being ignored because `parameters` has been specified.
      Warning:
      `eval_time` is being ignored because `parameters` has been specified.

# check for no arguments pass to ...

    Code
      empty_ellipses(a = 5)
    Condition
      Warning:
      The `...` are not used in this function but 1 object was passed: "a"

---

    Code
      empty_ellipses(a = 5, b = 1)
    Condition
      Warning:
      The `...` are not used in this function but 2 objects were passed: "a" and "b"

---

    Code
      empty_ellipses(5)
    Condition
      Warning:
      The `...` are not used in this function but 1 unnamed object was passed.

---

    Code
      empty_ellipses(1, 5)
    Condition
      Warning:
      The `...` are not used in this function but 2 unnamed objects were passed.

---

    Code
      empty_ellipses(1, second = 2)
    Condition
      Warning:
      The `...` are not used in this function but 2 objects were passed.

# percentile intervals - resamples only

    Code
      int_res_1 <- int_pctl(lm_res, times = 500)
    Condition
      Warning in `rsample::int_pctl()`:
      Recommend at least 1000 non-missing bootstrap resamples for terms `rmse` and `rsq`.

---

    Code
      int_pctl(lm_res, times = 2000, metrics = "rmse")
    Condition
      Error in `int_pctl()`:
      ! `metrics` should be a metric set as generated by `yardstick::metric_set()`.

---

    Code
      int_res_2 <- int_pctl(lm_res, times = 500, alpha = 0.25)
    Condition
      Warning in `rsample::int_pctl()`:
      Recommend at least 1000 non-missing bootstrap resamples for terms `rmse` and `rsq`.

# percentile intervals - last fit

    Code
      int_res_1 <- int_pctl(lm_res, times = 200)
    Condition
      Warning in `rsample::int_pctl()`:
      Recommend at least 1000 non-missing bootstrap resamples for term `mae`.

---

    Code
      int_res_2 <- int_pctl(lm_res, times = 200)
    Condition
      Warning in `rsample::int_pctl()`:
      Recommend at least 1000 non-missing bootstrap resamples for term `mae`.

# percentile intervals - grid + bayes tuning

    Code
      int_res_1 <- int_pctl(c5_res)

---

    Code
      set.seed(92)
      c5_bo_res <- tune_bayes(set_mode(set_engine(decision_tree(min_n = tune()),
      "C5.0"), "classification"), Class ~ ., resamples = cls_rs, initial = c5_res,
      iter = 1, metrics = metric_set(sens), control = control_bayes(save_pred = TRUE))

# percentile intervals - grid tuning with validation set

    Code
      int_res_1 <- int_pctl(c5_res)

# tune recipe only

    Code
      set.seed(2)
      res <- tune_bayes(wflow, resamples = folds, param_info = pset, initial = iter1,
        iter = iter2, control = control)

---

    Code
      tune_bayes(wflow, resamples = folds, param_info = pset, initial = iter1, iter = iter2,
        control = control_bayes(verbose = TRUE))
    Message
      
      >  Generating a set of 2 initial parameter results
      v Initialization complete
      
      i Estimating performance
      i Fold01: preprocessor 1/1
      i Fold01: preprocessor 1/1 (prediction data)
      i Fold01: preprocessor 1/1, model 1/1
      i Fold01: preprocessor 1/1, model 1/1 (predictions)
      i Fold02: preprocessor 1/1
      i Fold02: preprocessor 1/1 (prediction data)
      i Fold02: preprocessor 1/1, model 1/1
      i Fold02: preprocessor 1/1, model 1/1 (predictions)
      i Fold03: preprocessor 1/1
      i Fold03: preprocessor 1/1 (prediction data)
      i Fold03: preprocessor 1/1, model 1/1
      i Fold03: preprocessor 1/1, model 1/1 (predictions)
      i Fold04: preprocessor 1/1
      i Fold04: preprocessor 1/1 (prediction data)
      i Fold04: preprocessor 1/1, model 1/1
      i Fold04: preprocessor 1/1, model 1/1 (predictions)
      i Fold05: preprocessor 1/1
      i Fold05: preprocessor 1/1 (prediction data)
      i Fold05: preprocessor 1/1, model 1/1
      i Fold05: preprocessor 1/1, model 1/1 (predictions)
      i Fold06: preprocessor 1/1
      i Fold06: preprocessor 1/1 (prediction data)
      i Fold06: preprocessor 1/1, model 1/1
      i Fold06: preprocessor 1/1, model 1/1 (predictions)
      i Fold07: preprocessor 1/1
      i Fold07: preprocessor 1/1 (prediction data)
      i Fold07: preprocessor 1/1, model 1/1
      i Fold07: preprocessor 1/1, model 1/1 (predictions)
      i Fold08: preprocessor 1/1
      i Fold08: preprocessor 1/1 (prediction data)
      i Fold08: preprocessor 1/1, model 1/1
      i Fold08: preprocessor 1/1, model 1/1 (predictions)
      i Fold09: preprocessor 1/1
      i Fold09: preprocessor 1/1 (prediction data)
      i Fold09: preprocessor 1/1, model 1/1
      i Fold09: preprocessor 1/1, model 1/1 (predictions)
      i Fold10: preprocessor 1/1
      i Fold10: preprocessor 1/1 (prediction data)
      i Fold10: preprocessor 1/1, model 1/1
      i Fold10: preprocessor 1/1, model 1/1 (predictions)
      v Estimating performance
      i Estimating performance
      i Fold01: preprocessor 1/1
      i Fold01: preprocessor 1/1 (prediction data)
      i Fold01: preprocessor 1/1, model 1/1
      i Fold01: preprocessor 1/1, model 1/1 (predictions)
      i Fold02: preprocessor 1/1
      i Fold02: preprocessor 1/1 (prediction data)
      i Fold02: preprocessor 1/1, model 1/1
      i Fold02: preprocessor 1/1, model 1/1 (predictions)
      i Fold03: preprocessor 1/1
      i Fold03: preprocessor 1/1 (prediction data)
      i Fold03: preprocessor 1/1, model 1/1
      i Fold03: preprocessor 1/1, model 1/1 (predictions)
      i Fold04: preprocessor 1/1
      i Fold04: preprocessor 1/1 (prediction data)
      i Fold04: preprocessor 1/1, model 1/1
      i Fold04: preprocessor 1/1, model 1/1 (predictions)
      i Fold05: preprocessor 1/1
      i Fold05: preprocessor 1/1 (prediction data)
      i Fold05: preprocessor 1/1, model 1/1
      i Fold05: preprocessor 1/1, model 1/1 (predictions)
      i Fold06: preprocessor 1/1
      i Fold06: preprocessor 1/1 (prediction data)
      i Fold06: preprocessor 1/1, model 1/1
      i Fold06: preprocessor 1/1, model 1/1 (predictions)
      i Fold07: preprocessor 1/1
      i Fold07: preprocessor 1/1 (prediction data)
      i Fold07: preprocessor 1/1, model 1/1
      i Fold07: preprocessor 1/1, model 1/1 (predictions)
      i Fold08: preprocessor 1/1
      i Fold08: preprocessor 1/1 (prediction data)
      i Fold08: preprocessor 1/1, model 1/1
      i Fold08: preprocessor 1/1, model 1/1 (predictions)
      i Fold09: preprocessor 1/1
      i Fold09: preprocessor 1/1 (prediction data)
      i Fold09: preprocessor 1/1, model 1/1
      i Fold09: preprocessor 1/1, model 1/1 (predictions)
      i Fold10: preprocessor 1/1
      i Fold10: preprocessor 1/1 (prediction data)
      i Fold10: preprocessor 1/1, model 1/1
      i Fold10: preprocessor 1/1, model 1/1 (predictions)
      v Estimating performance
    Output
      # Tuning results
      # 10-fold cross-validation 
      # A tibble: 30 x 5
         splits         id     .metrics         .notes           .iter
         <list>         <chr>  <list>           <list>           <int>
       1 <split [28/4]> Fold01 <tibble [4 x 5]> <tibble [0 x 4]>     0
       2 <split [28/4]> Fold02 <tibble [4 x 5]> <tibble [0 x 4]>     0
       3 <split [29/3]> Fold03 <tibble [4 x 5]> <tibble [0 x 4]>     0
       4 <split [29/3]> Fold04 <tibble [4 x 5]> <tibble [0 x 4]>     0
       5 <split [29/3]> Fold05 <tibble [4 x 5]> <tibble [0 x 4]>     0
       6 <split [29/3]> Fold06 <tibble [4 x 5]> <tibble [0 x 4]>     0
       7 <split [29/3]> Fold07 <tibble [4 x 5]> <tibble [0 x 4]>     0
       8 <split [29/3]> Fold08 <tibble [4 x 5]> <tibble [0 x 4]>     0
       9 <split [29/3]> Fold09 <tibble [4 x 5]> <tibble [0 x 4]>     0
      10 <split [29/3]> Fold10 <tibble [4 x 5]> <tibble [0 x 4]>     0
      # i 20 more rows

---

    Code
      tune_bayes(wflow, resamples = folds, param_info = pset, initial = iter1, iter = iter2,
        control = control_bayes(verbose_iter = TRUE))
    Message
      Optimizing rmse using the expected improvement
      
      -- Iteration 1 -----------------------------------------------------------------
      
      i Current best:		rmse=2.505 (@iter 0)
      (x) GP has a LOO R² of 0% and is unreliable.
      v Gaussian process model failed
      i Generating a candidate as far away from existing points as possible.
      i Uncertainty sample
      i num_comp=11
      i Estimating performance
      v Estimating performance
      (x) Newest results:	rmse=3.589 (+/-0.499)
      
      -- Iteration 2 -----------------------------------------------------------------
      
      i Current best:		rmse=2.505 (@iter 0)
      v Gaussian process model (LOO R²: 30.5%)
      i Generating 15 candidates.
      i num_comp=4
      i Estimating performance
      v Estimating performance
      <3 Newest results:	rmse=2.461 (+/-0.37)
    Output
      # Tuning results
      # 10-fold cross-validation 
      # A tibble: 30 x 5
         splits         id     .metrics         .notes           .iter
         <list>         <chr>  <list>           <list>           <int>
       1 <split [28/4]> Fold01 <tibble [4 x 5]> <tibble [0 x 4]>     0
       2 <split [28/4]> Fold02 <tibble [4 x 5]> <tibble [0 x 4]>     0
       3 <split [29/3]> Fold03 <tibble [4 x 5]> <tibble [0 x 4]>     0
       4 <split [29/3]> Fold04 <tibble [4 x 5]> <tibble [0 x 4]>     0
       5 <split [29/3]> Fold05 <tibble [4 x 5]> <tibble [0 x 4]>     0
       6 <split [29/3]> Fold06 <tibble [4 x 5]> <tibble [0 x 4]>     0
       7 <split [29/3]> Fold07 <tibble [4 x 5]> <tibble [0 x 4]>     0
       8 <split [29/3]> Fold08 <tibble [4 x 5]> <tibble [0 x 4]>     0
       9 <split [29/3]> Fold09 <tibble [4 x 5]> <tibble [0 x 4]>     0
      10 <split [29/3]> Fold10 <tibble [4 x 5]> <tibble [0 x 4]>     0
      # i 20 more rows

---

    Code
      tune_bayes(wflow, resamples = folds, param_info = pset, initial = iter1, iter = iter2,
        control = control_bayes(verbose_iter = TRUE, verbose = TRUE))
    Message
      
      >  Generating a set of 2 initial parameter results
      v Initialization complete
      
      Optimizing rmse using the expected improvement
      
      -- Iteration 1 -----------------------------------------------------------------
      
      i Current best:		rmse=2.505 (@iter 0)
      (x) GP has a LOO R² of 0% and is unreliable.
      v Gaussian process model failed
      i Generating a candidate as far away from existing points as possible.
      i Uncertainty sample
      i num_comp=11
      i Estimating performance
      i Fold01: preprocessor 1/1
      i Fold01: preprocessor 1/1 (prediction data)
      i Fold01: preprocessor 1/1, model 1/1
      i Fold01: preprocessor 1/1, model 1/1 (predictions)
      i Fold02: preprocessor 1/1
      i Fold02: preprocessor 1/1 (prediction data)
      i Fold02: preprocessor 1/1, model 1/1
      i Fold02: preprocessor 1/1, model 1/1 (predictions)
      i Fold03: preprocessor 1/1
      i Fold03: preprocessor 1/1 (prediction data)
      i Fold03: preprocessor 1/1, model 1/1
      i Fold03: preprocessor 1/1, model 1/1 (predictions)
      i Fold04: preprocessor 1/1
      i Fold04: preprocessor 1/1 (prediction data)
      i Fold04: preprocessor 1/1, model 1/1
      i Fold04: preprocessor 1/1, model 1/1 (predictions)
      i Fold05: preprocessor 1/1
      i Fold05: preprocessor 1/1 (prediction data)
      i Fold05: preprocessor 1/1, model 1/1
      i Fold05: preprocessor 1/1, model 1/1 (predictions)
      i Fold06: preprocessor 1/1
      i Fold06: preprocessor 1/1 (prediction data)
      i Fold06: preprocessor 1/1, model 1/1
      i Fold06: preprocessor 1/1, model 1/1 (predictions)
      i Fold07: preprocessor 1/1
      i Fold07: preprocessor 1/1 (prediction data)
      i Fold07: preprocessor 1/1, model 1/1
      i Fold07: preprocessor 1/1, model 1/1 (predictions)
      i Fold08: preprocessor 1/1
      i Fold08: preprocessor 1/1 (prediction data)
      i Fold08: preprocessor 1/1, model 1/1
      i Fold08: preprocessor 1/1, model 1/1 (predictions)
      i Fold09: preprocessor 1/1
      i Fold09: preprocessor 1/1 (prediction data)
      i Fold09: preprocessor 1/1, model 1/1
      i Fold09: preprocessor 1/1, model 1/1 (predictions)
      i Fold10: preprocessor 1/1
      i Fold10: preprocessor 1/1 (prediction data)
      i Fold10: preprocessor 1/1, model 1/1
      i Fold10: preprocessor 1/1, model 1/1 (predictions)
      v Estimating performance
      (x) Newest results:	rmse=3.589 (+/-0.499)
      
      -- Iteration 2 -----------------------------------------------------------------
      
      i Current best:		rmse=2.505 (@iter 0)
      v Gaussian process model (LOO R²: 30.5%)
      i Generating 15 candidates.
      i num_comp=4
      i Estimating performance
      i Fold01: preprocessor 1/1
      i Fold01: preprocessor 1/1 (prediction data)
      i Fold01: preprocessor 1/1, model 1/1
      i Fold01: preprocessor 1/1, model 1/1 (predictions)
      i Fold02: preprocessor 1/1
      i Fold02: preprocessor 1/1 (prediction data)
      i Fold02: preprocessor 1/1, model 1/1
      i Fold02: preprocessor 1/1, model 1/1 (predictions)
      i Fold03: preprocessor 1/1
      i Fold03: preprocessor 1/1 (prediction data)
      i Fold03: preprocessor 1/1, model 1/1
      i Fold03: preprocessor 1/1, model 1/1 (predictions)
      i Fold04: preprocessor 1/1
      i Fold04: preprocessor 1/1 (prediction data)
      i Fold04: preprocessor 1/1, model 1/1
      i Fold04: preprocessor 1/1, model 1/1 (predictions)
      i Fold05: preprocessor 1/1
      i Fold05: preprocessor 1/1 (prediction data)
      i Fold05: preprocessor 1/1, model 1/1
      i Fold05: preprocessor 1/1, model 1/1 (predictions)
      i Fold06: preprocessor 1/1
      i Fold06: preprocessor 1/1 (prediction data)
      i Fold06: preprocessor 1/1, model 1/1
      i Fold06: preprocessor 1/1, model 1/1 (predictions)
      i Fold07: preprocessor 1/1
      i Fold07: preprocessor 1/1 (prediction data)
      i Fold07: preprocessor 1/1, model 1/1
      i Fold07: preprocessor 1/1, model 1/1 (predictions)
      i Fold08: preprocessor 1/1
      i Fold08: preprocessor 1/1 (prediction data)
      i Fold08: preprocessor 1/1, model 1/1
      i Fold08: preprocessor 1/1, model 1/1 (predictions)
      i Fold09: preprocessor 1/1
      i Fold09: preprocessor 1/1 (prediction data)
      i Fold09: preprocessor 1/1, model 1/1
      i Fold09: preprocessor 1/1, model 1/1 (predictions)
      i Fold10: preprocessor 1/1
      i Fold10: preprocessor 1/1 (prediction data)
      i Fold10: preprocessor 1/1, model 1/1
      i Fold10: preprocessor 1/1, model 1/1 (predictions)
      v Estimating performance
      <3 Newest results:	rmse=2.461 (+/-0.37)
    Output
      # Tuning results
      # 10-fold cross-validation 
      # A tibble: 30 x 5
         splits         id     .metrics         .notes           .iter
         <list>         <chr>  <list>           <list>           <int>
       1 <split [28/4]> Fold01 <tibble [4 x 5]> <tibble [0 x 4]>     0
       2 <split [28/4]> Fold02 <tibble [4 x 5]> <tibble [0 x 4]>     0
       3 <split [29/3]> Fold03 <tibble [4 x 5]> <tibble [0 x 4]>     0
       4 <split [29/3]> Fold04 <tibble [4 x 5]> <tibble [0 x 4]>     0
       5 <split [29/3]> Fold05 <tibble [4 x 5]> <tibble [0 x 4]>     0
       6 <split [29/3]> Fold06 <tibble [4 x 5]> <tibble [0 x 4]>     0
       7 <split [29/3]> Fold07 <tibble [4 x 5]> <tibble [0 x 4]>     0
       8 <split [29/3]> Fold08 <tibble [4 x 5]> <tibble [0 x 4]>     0
       9 <split [29/3]> Fold09 <tibble [4 x 5]> <tibble [0 x 4]>     0
      10 <split [29/3]> Fold10 <tibble [4 x 5]> <tibble [0 x 4]>     0
      # i 20 more rows

# tune recipe only - failure in recipe is caught elegantly

    Code
      cars_init_res <- tune_grid(model, preprocessor = rec, resamples = data_folds,
        grid = cars_grid)
    Message
      > A | error:   Error in `step_spline_b()`:
                     Caused by error in `prep()`:
                     ! `deg_free` must be a whole number, not a numeric `NA`.
      > B | warning: Some 'x' values beyond boundary knots may cause ill-conditioned basis
                     functions.

---

    Code
      set.seed(283)
      cars_bayes_res <- tune_bayes(model, preprocessor = rec, resamples = data_folds,
        initial = cars_init_res, iter = 2)
    Message
      > A | error:   Error in `step_spline_b()`:
                     Caused by error in `prep()`:
                     ! `degree` (3) must be less than or equal to `deg_free` (2) when `complete_set = TRUE`.
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.
    Message
      > A | warning: Some 'x' values beyond boundary knots may cause ill-conditioned basis
                     functions.

# tune model only - failure in recipe is caught elegantly

    Code
      cars_res <- tune_bayes(svm_mod, preprocessor = rec, resamples = data_folds)
    Message
      > A | error:   Error in `step_spline_b()`:
                     Caused by error in `prep()`:
                     ! `deg_free` must be a whole number, not a numeric `NA`.
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.
    Message
      x Optimization stopped prematurely; returning current results.

# tune model only - failure in formula is caught elegantly

    Code
      cars_res <- tune_bayes(wflow, resamples = data_folds, control = control_bayes(
        extract = function(x) {
          1
        }, save_pred = TRUE))
    Message
      > A | error:   The following predictor was not found in `data`: "z".
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.
    Message
      x Optimization stopped prematurely; returning current results.

# argument order gives an error for recipes

    Code
      tune_bayes(rec_tune_1, model = lm_mod, resamples = rsample::vfold_cv(mtcars, v = 2),
      param_info = extract_parameter_set_dials(rec_tune_1), iter = iter1, initial = iter2)
    Condition
      Error in `tune_bayes()`:
      ! The first argument to `tune_bayes()` should be either a model or workflow, not a <recipe> object.

# argument order gives an error for formula

    Code
      tune_bayes(mpg ~ ., svm_mod, resamples = rsample::vfold_cv(mtcars, v = 2),
      param_info = extract_parameter_set_dials(svm_mod), initial = iter1, iter = iter2)
    Condition
      Error in `tune_bayes()`:
      ! The first argument to `tune_bayes()` should be either a model or workflow, not a <formula> object.

# retain extra attributes and saved GP candidates

    Code
      setdiff(new_obj, current_objs)
    Output
      [1] "candidates" "gp_fit"     "i"          "score_card"

---

    Code
      res2 <- tune_bayes(wflow, resamples = folds, param_info = pset, initial = iter1,
        iter = iter2, control = control_bayes(save_workflow = TRUE))

# too few starting values

    Code
      tune:::check_bayes_initial_size(5, 3, FALSE)
    Message
      ! There are 5 tuning parameters and 3 grid points were requested.
      * This is likely to cause numerical issues in the first few search iterations.

---

    Code
      tune:::check_bayes_initial_size(5, 3, TRUE)
    Message
      ! There are 5 tuning parameters and 3 grid points were requested.
      * This is likely to cause numerical issues in the first few search iterations.
      * With racing, only completely resampled parameters are used.

---

    Code
      tune:::check_bayes_initial_size(2, 2, FALSE)
    Message
      ! There are 2 tuning parameters and 2 grid points were requested.
      * This is likely to cause numerical issues in the first few search iterations.

---

    Code
      tune:::check_bayes_initial_size(5, 1, FALSE)
    Condition
      Error in `tune:::check_bayes_initial_size()`:
      ! There are 5 tuning parameters and 1 grid point was requested.
      i The GP model requires 2+ initial points. For best performance, supply more initial points than there are tuning parameters.

---

    Code
      tune:::check_bayes_initial_size(5, 1, TRUE)
    Condition
      Error in `tune:::check_bayes_initial_size()`:
      ! There are 5 tuning parameters and 1 grid point was requested.
      i The GP model requires 2+ initial points. For best performance, supply more initial points than there are tuning parameters.
      i With racing, only completely resampled parameters are used.

---

    Code
      tune:::check_bayes_initial_size(1, 1, FALSE)
    Condition
      Error in `tune:::check_bayes_initial_size()`:
      ! There is 1 tuning parameter and 1 grid point was requested.
      i The GP model requires 2+ initial points. For best performance, supply more initial points than there are tuning parameters.

# missing performance values

    Code
      set.seed(1)
      res <- tune_bayes(mod, Sale_Price ~ Neighborhood + Gr_Liv_Area + Year_Built +
        Bldg_Type + Latitude + Longitude, resamples = folds, initial = 3, metrics = yardstick::metric_set(
        rsq), param_info = parameters(dials::cost_complexity(c(-2, 0))))
    Message
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 1 missing value was found and removed before fitting
        the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 2 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 3 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 4 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 5 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 6 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 7 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 8 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 9 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! For the rsq estimates, 10 missing values were found and removed before
        fitting the Gaussian process model.
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! No improvement for 10 iterations; returning current results.

---

    Code
      set.seed(2)
      res_fail <- tune_bayes(mod, Sale_Price ~ Neighborhood + Gr_Liv_Area +
      Year_Built + Bldg_Type + Latitude + Longitude, resamples = folds, initial = 5,
      metrics = yardstick::metric_set(rsq), param_info = parameters(dials::cost_complexity(
        c(0, 0.5))))
    Message
      > A | warning: A correlation computation is required, but `estimate` is constant and has 0 standard deviation, resulting in a divide by 0 error. `NA` will be returned.
      ! All of the rsq estimates were missing. The Gaussian process model cannot be
        fit to the data.
      (x) GP failed: Error in initialize(...) : is.numeric(X) is not TRUE
    Condition
      Error in `apply()`:
      ! dim(X) must have a positive length
    Message
      x Optimization stopped prematurely; returning current results.

# tune_bayes() output for `iter` edge cases (#721)

    Code
      tune_bayes(wf, boots, iter = -1)
    Condition
      Error in `tune_bayes()`:
      ! The `iter` argument must be a non-negative integer.

---

    Code
      tune_bayes(wf, boots, iter = c(-1, 0, 1))
    Condition
      Error in `tune_bayes()`:
      ! The `iter` argument must be a non-negative integer.

---

    Code
      tune_bayes(wf, boots, iter = c(0, 1, 2))
    Condition
      Error in `tune_bayes()`:
      ! The `iter` argument must be a non-negative integer.

---

    Code
      tune_bayes(wf, boots, iter = NA)
    Condition
      Error in `tune_bayes()`:
      ! The `iter` argument must be a non-negative integer.

---

    Code
      tune_bayes(wf, boots, iter = NULL)
    Condition
      Error in `tune_bayes()`:
      ! The `iter` argument must be a non-negative integer.

# rsample objects

    Code
      tune:::check_rset(obj_loo)
    Condition
      Error in `tune:::check_rset()`:
      ! Leave-one-out cross-validation is not currently supported with tune.

---

    Code
      tune:::check_rset(obj_nst)
    Condition
      Error in `tune:::check_rset()`:
      ! Nested resampling is not currently supported with tune.

---

    Code
      tune:::check_rset(obj_permut)
    Condition
      Error in `tune:::check_rset()`:
      ! Permutation samples are not suitable for tuning.

# grid objects

    Code
      tune:::.check_grid(rbind(grid_1, grid_1), chi_wflow)
    Condition
      Warning:
      Duplicate rows in grid of tuning combinations found and removed.
    Output
      # A tibble: 10 x 6
         penalty mixture imputation threshold deg_free degree
           <int>   <int>      <int>     <int>    <int>  <int>
       1       1       1          1         1        1      1
       2       2       2          2         2        2      2
       3       3       3          3         3        3      3
       4       4       4          4         4        4      4
       5       5       5          5         5        5      5
       6       6       6          6         6        6      6
       7       7       7          7         7        7      7
       8       8       8          8         8        8      8
       9       9       9          9         9        9      9
      10      10      10         10        10       10     10

---

    Code
      tune:::.check_grid(chi_wflow, chi_wflow)
    Condition
      Error in `tune:::.check_grid()`:
      ! `grid` should be a positive integer or a data frame.

# Unknown `grid` columns are caught

    Code
      tune:::.check_grid(grid, workflow)
    Condition
      Error in `tune:::.check_grid()`:
      ! The provided grid has 2 parameter columns (`other1` and `other2`) that have not been marked for tuning by `tune()`.

# Missing required `grid` columns are caught

    Code
      tune:::.check_grid(grid, workflow)
    Condition
      Error in `tune:::.check_grid()`:
      ! The provided grid is missing the following 1 parameter column that has been marked for tuning by `tune()`: "deg_free".

# workflow objects

    Code
      tune:::check_workflow(x = wflow_2, check_dials = TRUE)
    Condition
      Error in `tune:::check_workflow()`:
      ! The workflow has arguments whose ranges are not finalized: "mtry".

---

    Code
      tune:::check_workflow(wflow_3)
    Condition
      Error in `tune:::check_workflow()`:
      ! A formula, recipe, or variables preprocessor is required.

---

    Code
      tune:::check_workflow(wflow_4)
    Condition
      Error in `tune:::check_workflow()`:
      ! A parsnip model is required.

# errors informatively when needed package isn't installed

    Code
      check_workflow(stan_wflow)
    Condition
      Error:
      ! Package install is required for rstanarm.

---

    Code
      fit_resamples(stan_wflow, rsample::bootstraps(mtcars))
    Condition
      Error in `fit_resamples()`:
      ! Package install is required for rstanarm.

# workflow objects (will not tune, tidymodels/tune#548)

    Code
      tune_grid(lr_lm_1, rec_bare, rsample::bootstraps(Chicago, 2))
    Condition
      Error in `tune_grid()`:
      ! The parameter `penalty` was marked with `tune()`, though will not be tuned.
      i This usually means that the current modeling engine `lm` does not support tuning `penalty`.

---

    Code
      tune_bayes(lr_lm_2, rec_tune, rsample::bootstraps(Chicago, 2))
    Condition
      Error in `tune_bayes()`:
      ! The parameters `penalty` and `mixture` were marked with `tune()`, though will not be tuned.
      i This usually means that the current modeling engine `lm` does not support tuning `penalty` and `mixture`.

# yardstick objects

    Code
      tune:::check_metrics(yardstick::rmse, chi_wflow)
    Condition
      Error in `tune:::check_metrics()`:
      ! The `metrics` argument should be the results of `yardstick::metric_set()`.

# metrics must match the parsnip engine

    Code
      tune:::check_metrics(metric_set1, workflow1)
    Condition
      Error in `tune:::check_metrics()`:
      ! The parsnip model has `mode = 'regression'`, but `metrics` is a metric set for a different model mode.

---

    Code
      tune:::check_metrics(metric_set2, workflow2)
    Condition
      Error in `tune:::check_metrics()`:
      ! The parsnip model has `mode = 'classification'`, but `metrics` is a metric set for a different model mode.

# check_metrics() is deprecated

    Code
      metric <- check_metrics(NULL, wflow)
    Condition
      Warning:
      `check_metrics()` was deprecated in tune 2.1.0.
      i Please use `check_metrics_arg()` instead.

# grid control objects

    Code
      control_grid(tomato = 1)
    Condition
      Error in `control_grid()`:
      ! unused argument (tomato = 1)

---

    Code
      control_grid(verbose = 1)
    Condition
      Error in `control_grid()`:
      ! `verbose` must be `TRUE` or `FALSE`, not the number 1.

---

    Code
      control_grid(verbose = rep(TRUE, 2))
    Condition
      Error in `control_grid()`:
      ! `verbose` must be `TRUE` or `FALSE`, not a logical vector.

---

    Code
      control_grid(allow_par = 1)
    Condition
      Error in `control_grid()`:
      ! `allow_par` must be `TRUE` or `FALSE`, not the number 1.

---

    Code
      control_grid(save_pred = "no")
    Condition
      Error in `control_grid()`:
      ! `save_pred` must be `TRUE` or `FALSE`, not the string "no".

---

    Code
      control_grid(extract = Inf)
    Condition
      Error in `control_grid()`:
      ! `extract` must be a function or `NULL`, not `Inf`.

---

    Code
      control_grid(pkgs = Inf)
    Condition
      Error in `control_grid()`:
      ! `pkgs` must be a character vector or `NULL`, not `Inf`.

# Bayes control objects

    Code
      control_bayes(tomato = 1)
    Condition
      Error in `control_bayes()`:
      ! unused argument (tomato = 1)

---

    Code
      control_bayes(verbose = 1)
    Condition
      Error in `control_bayes()`:
      ! `verbose` must be `TRUE` or `FALSE`, not the number 1.

---

    Code
      control_bayes(verbose = rep(TRUE, 2))
    Condition
      Error in `control_bayes()`:
      ! `verbose` must be `TRUE` or `FALSE`, not a logical vector.

---

    Code
      control_bayes(no_improve = FALSE)
    Condition
      Error in `control_bayes()`:
      ! `no_improve` must be a whole number, not `FALSE`.

---

    Code
      control_bayes(uncertain = FALSE)
    Condition
      Error in `control_bayes()`:
      ! `uncertain` must be a whole number, not `FALSE`.

---

    Code
      control_bayes(seed = FALSE)
    Condition
      Error in `control_bayes()`:
      ! `seed` must be a whole number, not `FALSE`.

---

    Code
      control_bayes(save_pred = "no")
    Condition
      Error in `control_bayes()`:
      ! `save_pred` must be `TRUE` or `FALSE`, not the string "no".

---

    Code
      control_bayes(extract = Inf)
    Condition
      Error in `control_bayes()`:
      ! `extract` must be a function or `NULL`, not `Inf`.

---

    Code
      control_bayes(pkgs = Inf)
    Condition
      Error in `control_bayes()`:
      ! `pkgs` must be a character vector or `NULL`, not `Inf`.

---

    Code
      control_bayes(time_limit = "a")
    Condition
      Error in `control_bayes()`:
      ! `time_limit` should be either a single numeric or logical value.

---

    Code
      tmp <- control_bayes(no_improve = 2, uncertain = 5)
    Condition
      Warning:
      Uncertainty sample scheduled after 5 poor iterations but the search will stop after 2.

# initial values

    Code
      tune:::check_initial(data.frame(), pset = extract_parameter_set_dials(wflow_1),
      wflow = wflow_1, resamples = mtfolds, metrics = yardstick::metric_set(
        yardstick::rsq), ctrl = control_bayes())
    Condition
      Error in `tune:::check_initial()`:
      ! `initial` should be a positive integer or the results of `tune_grid()`

# val_class_or_null() is deprecated

    Code
      res <- val_class_or_null("a", "character")
    Condition
      Warning:
      `val_class_or_null()` was deprecated in tune 2.1.0.

# val_class_and_single() is deprecated

    Code
      res <- val_class_and_single("a", "character")
    Condition
      Warning:
      `val_class_and_single()` was deprecated in tune 2.1.0.

# check parameter finalization

    Code
      expect_no_error(p1 <- tune:::check_parameters(w1, data = mtcars, grid_names = character(
        0)))
    Message
      i Creating pre-processing data to finalize 1 unknown parameter: "mtry"

---

    Code
      expect_no_error(p2 <- tune:::check_parameters(w2, data = mtcars))
    Message
      i Creating pre-processing data to finalize 1 unknown parameter: "mtry"

---

    Code
      expect_no_error(p3_a <- tune:::check_parameters(w3, data = mtcars))
    Message
      i Creating pre-processing data to finalize 1 unknown parameter: "mtry"

---

    Code
      tune:::check_parameters(w4, data = mtcars)
    Condition
      Error in `tune:::check_parameters()`:
      ! Some model parameters require finalization but there are recipe parameters that require tuning.
      i Please use `extract_parameter_set_dials()` to set parameter ranges manually and supply the output to the `param_info` argument.

# check fold weights

    Code
      add_resample_weights(folds, c("a", "b", "c"))
    Condition
      Error in `.validate_resample_weights()`:
      ! `weights` must be numeric.

---

    Code
      add_resample_weights(folds, c(0.5, 0.3))
    Condition
      Error in `.validate_resample_weights()`:
      ! Length of `weights` (2) must equal number of resamples (3).

---

    Code
      add_resample_weights(folds, c(-0.1, 0.5, 0.6))
    Condition
      Error in `.validate_resample_weights()`:
      ! `weights` must be non-negative.

# maker static object

    Code
      tune:::.make_static(1, param_info = extract_parameter_set_dials(wflow), grid = dt_grid,
      metrics = metric_set(accuracy), eval_time = NULL, split_args = mc_cv_args,
      control = control_resamples())
    Condition
      Error in `tune:::.make_static()`:
      ! `workflow` should be a <workflow> object.

---

    Code
      tune:::.make_static(wflow, param_info = 2, grid = dt_grid, metrics = metric_set(
        accuracy), eval_time = NULL, split_args = mc_cv_args, control = control_resamples())
    Condition
      Error in `tune:::.make_static()`:
      ! `param_info` should be a <parameters> object.

---

    Code
      tune:::.make_static(wflow, param_info = extract_parameter_set_dials(wflow),
      grid = dt_grid, metrics = 3, eval_time = NULL, split_args = mc_cv_args,
      control = control_resamples())
    Condition
      Error in `tune:::.make_static()`:
      ! `metrics` should be a <metric_set> object.

---

    Code
      tune:::.make_static(wflow, param_info = extract_parameter_set_dials(wflow),
      grid = dt_grid, metrics = metric_set(accuracy), eval_time = "four", split_args = mc_cv_args,
      control = control_resamples())
    Condition
      Error in `tune:::.make_static()`:
      ! `eval_time` should be a numeric vector.

---

    Code
      tune:::update_static(res, bad_part)
    Condition
      Error in `check_static_data()`:
      ! Element `data` should be a tibble in the fit slot.

---

    Code
      tune:::update_static(res, bad_part)
    Condition
      Error in `check_static_data()`:
      ! Element `ind` should be an integer in the fit slot.

# no parallelism

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is not active.
      future is active with 1 worker.
      Too few workers for parallel processing.
    Output
      [1] "sequential"

# enable future parallelism

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is not active.
      future is active with 1 worker.
      Too few workers for parallel processing.
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is not active.
      future is active with 2 workers.
      future will be used for parallel processing}.
    Output
      [1] "future"

---

    Code
      tune:::choose_framework(control = ctrl_no, verbose = TRUE)
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(control = ctrl_java, verbose = TRUE)
    Message
      These packages cannot be used with explicit parallel processing: rJava.
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is not active.
      future is active with 1 worker.
      Too few workers for parallel processing.
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is not active.
      future is active with 1 worker.
      Too few workers for parallel processing.
    Output
      [1] "sequential"

# enable mirai parallelism

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is not active.
      future is active with 1 worker.
      Too few workers for parallel processing.
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is active with 2 workers.
      future is active with 1 worker.
      mirai will be used for parallel processing}.
    Output
      [1] "mirai"

---

    Code
      tune:::choose_framework(control = ctrl_no, verbose = TRUE)
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(control = ctrl_java, verbose = TRUE)
    Message
      These packages cannot be used with explicit parallel processing: rJava.
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is active with 1 worker.
      future is active with 1 worker.
      Too few workers for parallel processing.
    Output
      [1] "sequential"

---

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is not active.
      future is active with 1 worker.
      Too few workers for parallel processing.
    Output
      [1] "sequential"

# break parallelism tie

    Code
      tune:::choose_framework(verbose = TRUE)
    Message
      mirai is active with 2 workers.
      future is active with 2 workers.
      Multiple workers exist for both mirai and future; falling back to the default of mirai.
    Output
      [1] "mirai"

# loop execution code

    Code
      tune:::loop_call("resamples", "sequential", list())
    Output
      lapply(resamples, tune::.loop_over_all_stages, grid = grid, static = static)

---

    Code
      tune:::loop_call("resamples", "sequential", list(a = quote(a)))
    Output
      lapply(resamples, tune::.loop_over_all_stages, grid = grid, static = static, 
          a = a)

---

    Code
      tune:::loop_call("everything", "sequential", list())
    Output
      lapply(inds, tune::.loop_over_all_stages2, resamples = resamples, 
          grid = candidates, static = static)

---

    Code
      tune:::loop_call("everything", "sequential", list(a = quote(a)))
    Output
      lapply(inds, tune::.loop_over_all_stages2, resamples = resamples, 
          grid = candidates, static = static, a = a)

---

    Code
      tune:::loop_call("resamples", "future", list())
    Output
      future.apply::future_lapply(resamples, tune::.loop_over_all_stages, 
          grid = grid, static = static, future.label = "tune-grid-%d", 
          future.stdout = TRUE, future.seed = TRUE, future.globals = c("grid", 
          "static"))

---

    Code
      tune:::loop_call("resamples", "future", list(a = quote(a)))
    Output
      future.apply::future_lapply(resamples, tune::.loop_over_all_stages, 
          grid = grid, static = static, a = a, future.label = "tune-grid-%d", 
          future.stdout = TRUE, future.seed = TRUE, future.globals = c("grid", 
          "static", "a"))

---

    Code
      tune:::loop_call("everything", "future", list())
    Output
      future.apply::future_lapply(inds, tune::.loop_over_all_stages2, 
          resamples = resamples, grid = candidates, static = static, 
          future.label = "tune-grid-%d", future.stdout = TRUE, future.seed = TRUE, 
          future.globals = c("resamples", "grid", "static"))

---

    Code
      tune:::loop_call("everything", "future", list(a = quote(a)))
    Output
      future.apply::future_lapply(inds, tune::.loop_over_all_stages2, 
          resamples = resamples, grid = candidates, static = static, 
          a = a, future.label = "tune-grid-%d", future.stdout = TRUE, 
          future.seed = TRUE, future.globals = c("resamples", "grid", 
          "static", "a"))

---

    Code
      tune:::loop_call("resamples", "mirai", list())
    Output
      eval_mirai(resamples, tune::.loop_over_all_stages, .args = list(
          grid = grid, static = static))

---

    Code
      tune:::loop_call("resamples", "mirai", list(a = quote(a)))
    Output
      eval_mirai(resamples, tune::.loop_over_all_stages, .args = list(
          grid = grid, static = static, a = a))

---

    Code
      tune:::loop_call("everything", "mirai", list())
    Output
      eval_mirai(inds, tune::.loop_over_all_stages2, .args = list(resamples = resamples, 
          grid = candidates, static = static))

---

    Code
      tune:::loop_call("everything", "mirai", list(a = quote(a)))
    Output
      eval_mirai(inds, tune::.loop_over_all_stages2, .args = list(resamples = resamples, 
          grid = candidates, static = static, a = a))

# failure in recipe is caught elegantly

    Code
      result <- fit_resamples(lin_mod, rec, folds, control = control)
    Message
      > A | error:   Error in `step_spline_natural()`:
                     Caused by error in `prep()`:
                     ! `deg_free` must be a whole number, not a numeric `NA`.
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      note
    Output
      [1] "Error in `step_spline_natural()`:\nCaused by error in `prep()`:\n! `deg_free` must be a whole number, not a numeric `NA`."

# failure in variables tidyselect specification is caught elegantly

    Code
      result <- fit_resamples(workflow, folds, control = control)
    Message
      > A | error:   Can't select columns that don't exist.
                     x Column `foobar` doesn't exist.
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      note
    Output
      [1] "Can't select columns that don't exist.\nx Column `foobar` doesn't exist."

# classification models generate correct error message

    Code
      result <- fit_resamples(log_mod, rec, folds, control = control)
    Message
      > A | error:   For a classification model, the outcome should be a <factor>, not a double vector.
    Condition
      Warning:
      All models failed. Run `show_notes(.Last.tune.result)` for more information.

---

    Code
      note
    Output
      [1] "For a classification model, the outcome should be a <factor>, not a double vector."

# `tune_grid()` falls back to `fit_resamples()` - formula

    Code
      result <- tune_grid(lin_mod, mpg ~ ., folds)
    Condition
      Warning:
      No tuning parameters have been detected, performance will be evaluated using the resamples with no tuning.
      Did you want to assign any parameters with a value of `tune()`?

# `tune_grid()` falls back to `fit_resamples()` - workflow variables

    Code
      result <- tune_grid(wf, folds)
    Condition
      Warning:
      No tuning parameters have been detected, performance will be evaluated using the resamples with no tuning.
      Did you want to assign any parameters with a value of `tune()`?

# `tune_grid()` ignores `grid` if there are no tuning parameters

    Code
      result <- tune_grid(lin_mod, mpg ~ ., grid = data.frame(x = 1), folds)
    Condition
      Warning:
      No tuning parameters have been detected, performance will be evaluated using the resamples with no tuning.
      Did you want to assign any parameters with a value of `tune()`?

# cannot autoplot `fit_resamples()` results

    Code
      autoplot(result)
    Condition
      Error in `autoplot()`:
      ! There is no `autoplot()` implementation for <resample_results>.

# ellipses with fit_resamples

    Code
      fit_resamples(lin_mod, mpg ~ ., folds, something = "wrong")
    Condition
      Warning:
      The `...` are not used in this function but 1 object was passed: "something"
    Output
      # Resampling results
      # 2-fold cross-validation 
      # A tibble: 2 x 4
        splits          id    .metrics         .notes          
        <list>          <chr> <list>           <list>          
      1 <split [16/16]> Fold1 <tibble [2 x 4]> <tibble [0 x 4]>
      2 <split [16/16]> Fold2 <tibble [2 x 4]> <tibble [0 x 4]>

# argument order gives errors for recipe/formula

    Code
      fit_resamples(rec, lin_mod, folds)
    Condition
      Error in `fit_resamples()`:
      ! The first argument to `fit_resamples()` should be either a model or workflow, not a <recipe> object.

---

    Code
      fit_resamples(mpg ~ ., lin_mod, folds)
    Condition
      Error in `fit_resamples()`:
      ! The first argument to `fit_resamples()` should be either a model or workflow, not a <formula> object.

# `fit_resamples()` when objects need tuning

    2 arguments have been tagged for tuning in these components: model_spec and recipe.
    i Please use one of the tuning functions (e.g. `tune_grid()`) to optimize them.

---

    1 argument has been tagged for tuning in this component: model_spec.
    i Please use one of the tuning functions (e.g. `tune_grid()`) to optimize them.

---

    1 argument has been tagged for tuning in this component: recipe.
    i Please use one of the tuning functions (e.g. `tune_grid()`) to optimize them.

# gives informative default error

    Code
      .use_case_weights_with_yardstick(1)
    Condition
      Error in `.use_case_weights_with_yardstick()`:
      ! An object with <hardhat_case_weights> was expected, not a number.
      i Define a `.use_case_weights_with_yardstick()` method for this type to declare whether or not these case weights should be passed on to yardstick.
      i See `?.use_case_weights_with_yardstick()` for more information.

# `extract_case_weights()` errors if `col` doesn't exist

    Code
      extract_case_weights(mtcars, wf)
    Condition
      Error in `extract_case_weights()`:
      ! `col` must exist and be a quosure at this point.
      i This is an internal error that was detected in the tune package.
        Please report it at <https://github.com/tidymodels/tune/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

# `extract_case_weights()` errors if case weights column isn't the right class

    Code
      extract_case_weights(mtcars, wf)
    Condition
      Error in `extract_case_weights()`:
      ! Case weights must be a supported case weights type, as determined by `hardhat::is_case_weights()`.

# linear regression grid reduction - glmnet

    Code
      min_grid(mod, data.frame(mixture = 1:3))
    Condition
      Error in `no_penalty()`:
      ! At least one penalty value is required for glmnet.

# logistic regression grid reduction - glmnet

    Code
      min_grid(mod, data.frame(mixture = 1:3))
    Condition
      Error in `no_penalty()`:
      ! At least one penalty value is required for glmnet.

# multinomial regression grid reduction - glmnet

    Code
      min_grid(mod, data.frame(mixture = 1:3))
    Condition
      Error in `no_penalty()`:
      ! At least one penalty value is required for glmnet.

# eval time inputs are checked for regression models

    Code
      check_eval_time_arg(NULL, met_reg)
    Output
      NULL

---

    Code
      check_eval_time_arg(times, met_reg)
    Condition
      Warning:
      `eval_time` is only used for models with mode "censored regression".
    Output
      NULL

---

    Code
      res <- fit_resamples(wflow, rs, eval_time = times)
    Condition
      Warning in `fit_resamples()`:
      `eval_time` is only used for models with mode "censored regression".

---

    Code
      set.seed(1)
      res <- tune_grid(wflow_tune, rs, eval_time = times)
    Condition
      Warning in `tune_grid()`:
      `eval_time` is only used for models with mode "censored regression".

---

    Code
      set.seed(1)
      res <- tune_bayes(wflow_tune, rs, iter = 1, eval_time = times)
    Condition
      Warning in `tune_bayes()`:
      `eval_time` is only used for models with mode "censored regression".

---

    Code
      res <- last_fit(wflow, split, eval_time = times)
    Condition
      Warning in `last_fit()`:
      `eval_time` is only used for models with mode "censored regression".

# eval time are checked for classification models

    Code
      check_eval_time_arg(NULL, met_cls)
    Output
      NULL

---

    Code
      check_eval_time_arg(times, met_cls)
    Condition
      Warning:
      `eval_time` is only used for models with mode "censored regression".
    Output
      NULL

---

    Code
      res <- fit_resamples(wflow, rs, eval_time = times)
    Condition
      Warning in `fit_resamples()`:
      `eval_time` is only used for models with mode "censored regression".

---

    Code
      set.seed(1)
      res <- tune_grid(wflow_tune, rs, eval_time = times)
    Condition
      Warning in `tune_grid()`:
      `eval_time` is only used for models with mode "censored regression".

---

    Code
      set.seed(1)
      res <- tune_bayes(wflow_tune, rs, iter = 1, eval_time = times)
    Condition
      Warning in `tune_bayes()`:
      `eval_time` is only used for models with mode "censored regression".

---

    Code
      res <- last_fit(wflow, split, eval_time = times)
    Condition
      Warning in `last_fit()`:
      `eval_time` is only used for models with mode "censored regression".

# eval time inputs are checked for censored regression models

    Code
      check_eval_time_arg(NULL, met_stc)
    Output
      NULL

---

    Code
      check_eval_time_arg(NULL, met_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(NULL, met_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(NULL, met_stc_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(NULL, met_stc_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(NULL, met_dyn_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(NULL, met_dyn_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(NULL, met_int_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(NULL, met_int_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      check_eval_time_arg(2, met_stc)
    Condition
      Warning:
      `eval_time` is only used for dynamic or integrated survival metrics.
    Output
      NULL

---

    Code
      check_eval_time_arg(2, met_dyn)
    Output
      [1] 2

---

    Code
      check_eval_time_arg(2, met_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      check_eval_time_arg(2, met_stc_dyn)
    Output
      [1] 2

---

    Code
      check_eval_time_arg(2, met_stc_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      check_eval_time_arg(2, met_dyn_stc)
    Output
      [1] 2

---

    Code
      check_eval_time_arg(2, met_dyn_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      check_eval_time_arg(2, met_int_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      check_eval_time_arg(2, met_int_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      check_eval_time_arg(1:3, met_stc)
    Condition
      Warning:
      `eval_time` is only used for dynamic or integrated survival metrics.
    Output
      NULL

---

    Code
      check_eval_time_arg(1:3, met_dyn)
    Output
      [1] 1 2 3

---

    Code
      check_eval_time_arg(1:3, met_int)
    Output
      [1] 1 2 3

---

    Code
      check_eval_time_arg(1:3, met_stc_dyn)
    Output
      [1] 1 2 3

---

    Code
      check_eval_time_arg(1:3, met_stc_int)
    Output
      [1] 1 2 3

---

    Code
      check_eval_time_arg(1:3, met_dyn_stc)
    Output
      [1] 1 2 3

---

    Code
      check_eval_time_arg(1:3, met_dyn_int)
    Output
      [1] 1 2 3

---

    Code
      check_eval_time_arg(1:3, met_int_stc)
    Output
      [1] 1 2 3

---

    Code
      check_eval_time_arg(1:3, met_int_dyn)
    Output
      [1] 1 2 3

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_stc)

---

    Code
      fit_resamples(wflow, rs, metrics = met_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric". Only 0 unique times were given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_stc_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_stc_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_dyn_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_dyn_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_int_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_int_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_stc, eval_time = 2)
    Condition
      Warning in `fit_resamples()`:
      `eval_time` is only used for dynamic or integrated survival metrics.

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_dyn, eval_time = 2)

---

    Code
      fit_resamples(wflow, rs, metrics = met_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_stc_dyn, eval_time = 2)

---

    Code
      fit_resamples(wflow, rs, metrics = met_stc_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_dyn_stc, eval_time = 2)

---

    Code
      fit_resamples(wflow, rs, metrics = met_dyn_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_int_stc, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      fit_resamples(wflow, rs, metrics = met_int_dyn, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_stc, eval_time = 1:3)
    Condition
      Warning in `fit_resamples()`:
      `eval_time` is only used for dynamic or integrated survival metrics.

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_dyn, eval_time = 1:3)

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_int, eval_time = 1:3)

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_stc_dyn, eval_time = 1:3)

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_stc_int, eval_time = 1:3)

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_dyn_stc, eval_time = 1:3)

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_dyn_int, eval_time = 1:3)

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_int_stc, eval_time = 1:3)

---

    Code
      res <- fit_resamples(wflow, rs, metrics = met_int_dyn, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_stc)

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_stc_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_stc_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_dyn_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_dyn_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_int_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_int_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_stc, eval_time = 2)
    Condition
      Warning in `tune_grid()`:
      `eval_time` is only used for dynamic or integrated survival metrics.

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_dyn, eval_time = 2)

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_stc_dyn, eval_time = 2)

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_stc_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_dyn_stc, eval_time = 2)

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_dyn_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_int_stc, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_int_dyn, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_stc, eval_time = 1:3)
    Condition
      Warning in `tune_grid()`:
      `eval_time` is only used for dynamic or integrated survival metrics.

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_dyn, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_int, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_stc_dyn, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_stc_int, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_dyn_stc, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_dyn_int, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_int_stc, eval_time = 1:3)

---

    Code
      res <- tune_grid(wflow_tune, rs, metrics = met_int_dyn, eval_time = 1:3)

---

    Code
      last_fit(wflow, split, metrics = met_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric". Only 0 unique times were given.

---

    Code
      last_fit(wflow, split, metrics = met_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      last_fit(wflow, split, metrics = met_stc_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      last_fit(wflow, split, metrics = met_stc_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      last_fit(wflow, split, metrics = met_dyn_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 1 evaluation time is required for the metric type(s) requested: "dynamic_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      last_fit(wflow, split, metrics = met_dyn_int)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      last_fit(wflow, split, metrics = met_int_stc)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 0 unique times were given.

---

    Code
      last_fit(wflow, split, metrics = met_int_dyn)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 0 unique times were given.

---

    Code
      res <- last_fit(wflow, split, metrics = met_stc, eval_time = 2)
    Condition
      Warning in `last_fit()`:
      `eval_time` is only used for dynamic or integrated survival metrics.

---

    Code
      res <- last_fit(wflow, split, metrics = met_dyn, eval_time = 2)

---

    Code
      last_fit(wflow, split, metrics = met_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      res <- last_fit(wflow, split, metrics = met_stc_dyn, eval_time = 2)

---

    Code
      last_fit(wflow, split, metrics = met_stc_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      res <- last_fit(wflow, split, metrics = met_dyn_stc, eval_time = 2)

---

    Code
      last_fit(wflow, split, metrics = met_dyn_int, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      last_fit(wflow, split, metrics = met_int_stc, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "integrated_survival_metric" and "static_survival_metric". Only 1 unique time was given.

---

    Code
      last_fit(wflow, split, metrics = met_int_dyn, eval_time = 2)
    Condition
      Error in `check_enough_eval_times()`:
      ! At least 2 evaluation times are required for the metric type(s) requested: "dynamic_survival_metric" and "integrated_survival_metric". Only 1 unique time was given.

---

    Code
      res <- last_fit(wflow, split, metrics = met_stc, eval_time = 1:3)
    Condition
      Warning in `last_fit()`:
      `eval_time` is only used for dynamic or integrated survival metrics.

---

    Code
      res <- last_fit(wflow, split, metrics = met_dyn, eval_time = 1:3)

---

    Code
      res <- last_fit(wflow, split, metrics = met_int, eval_time = 1:3)

---

    Code
      res <- last_fit(wflow, split, metrics = met_stc_dyn, eval_time = 1:3)

---

    Code
      res <- last_fit(wflow, split, metrics = met_stc_int, eval_time = 1:3)

---

    Code
      res <- last_fit(wflow, split, metrics = met_dyn_stc, eval_time = 1:3)

---

    Code
      res <- last_fit(wflow, split, metrics = met_dyn_int, eval_time = 1:3)

---

    Code
      res <- last_fit(wflow, split, metrics = met_int_stc, eval_time = 1:3)

---

    Code
      res <- last_fit(wflow, split, metrics = met_int_dyn, eval_time = 1:3)

# selecting a metric

    Code
      choose_metric(ames_grid_search, "rmse")
    Output
      # A tibble: 1 x 3
        metric class          direction
        <chr>  <chr>          <chr>    
      1 rmse   numeric_metric minimize 

---

    Code
      choose_metric(ames_grid_search, NULL)
    Condition
      Warning:
      No value of `metric` was given; "rmse" will be used.
    Output
      # A tibble: 1 x 3
        metric class          direction
        <chr>  <chr>          <chr>    
      1 rmse   numeric_metric minimize 

---

    Code
      choose_metric(ames_grid_search, "potato")
    Condition
      Error:
      ! "potato" was not in the metric set. Please choose from: "rmse" and "rsq".

---

    Code
      choose_metric(ames_grid_search, c("rmse", "ccc"))
    Condition
      Warning:
      2 metrics were given; "rmse" will be used.
    Output
      # A tibble: 1 x 3
        metric class          direction
        <chr>  <chr>          <chr>    
      1 rmse   numeric_metric minimize 

# add_resample_weights() validates inputs correctly

    Code
      add_resample_weights("not_an_rset", c(0.5, 0.3, 0.2))
    Condition
      Error in `add_resample_weights()`:
      ! `rset` must be an rset object.

---

    Code
      add_resample_weights(folds, c("a", "b", "c"))
    Condition
      Error in `.validate_resample_weights()`:
      ! `weights` must be numeric.

---

    Code
      add_resample_weights(folds, c(0.5, 0.3))
    Condition
      Error in `.validate_resample_weights()`:
      ! Length of `weights` (2) must equal number of resamples (3).

---

    Code
      add_resample_weights(folds, c(-0.1, 0.5, 0.6))
    Condition
      Error in `.validate_resample_weights()`:
      ! `weights` must be non-negative.

---

    Code
      add_resample_weights(folds, c(0, 0, 0))
    Condition
      Error in `.validate_resample_weights()`:
      ! At least one weight must be positive.

# extract_resample_weights() validates input types

    Code
      extract_resample_weights("not_valid_input")
    Condition
      Error in `extract_resample_weights()`:
      ! `x` must be an rset or tune_results object.

---

    Code
      extract_resample_weights(data.frame(x = 1:3))
    Condition
      Error in `extract_resample_weights()`:
      ! `x` must be an rset or tune_results object.

# vec_cast() is working

    Code
      vec_cast(x, x)
    Condition
      Error in `stop_incompatible_cast_resample_results()`:
      ! Can't convert `x` <resample_results> to <resample_results>.
      Can't cast to a <resample_results> because attributes are likely incompatible.

---

    Code
      vec_cast(tbl, x)
    Condition
      Error in `stop_incompatible_cast_resample_results()`:
      ! Can't convert `tbl` <tibble> to <resample_results>.
      Can't cast to a <resample_results> because attributes are likely incompatible.

---

    Code
      vec_cast(df, x)
    Condition
      Error in `stop_incompatible_cast_resample_results()`:
      ! Can't convert `df` <data.frame> to <resample_results>.
      Can't cast to a <resample_results> because attributes are likely incompatible.

---

    Code
      vec_cast(x, x)
    Condition
      Error in `stop_incompatible_cast_tune_results()`:
      ! Can't convert `x` <tune_results> to <tune_results>.
      Can't cast to a <tune_results> because attributes are likely incompatible.

---

    Code
      vec_cast(tbl, x)
    Condition
      Error in `stop_incompatible_cast_tune_results()`:
      ! Can't convert `tbl` <tibble> to <tune_results>.
      Can't cast to a <tune_results> because attributes are likely incompatible.

---

    Code
      vec_cast(df, x)
    Condition
      Error in `stop_incompatible_cast_tune_results()`:
      ! Can't convert `df` <data.frame> to <tune_results>.
      Can't cast to a <tune_results> because attributes are likely incompatible.

---

    Code
      vec_cast(x, x)
    Condition
      Error in `stop_incompatible_cast_iteration_results()`:
      ! Can't convert `x` <iteration_results> to <iteration_results>.
      Can't cast to a <iteration_results> because attributes are likely incompatible.

---

    Code
      vec_cast(tbl, x)
    Condition
      Error in `stop_incompatible_cast_iteration_results()`:
      ! Can't convert `tbl` <tibble> to <iteration_results>.
      Can't cast to a <iteration_results> because attributes are likely incompatible.

---

    Code
      vec_cast(df, x)
    Condition
      Error in `stop_incompatible_cast_iteration_results()`:
      ! Can't convert `df` <data.frame> to <iteration_results>.
      Can't cast to a <iteration_results> because attributes are likely incompatible.

# predict errors informatively when supplied tuning results

    Code
      predict(ames_grid_search)
    Condition
      Error in `predict()`:
      ! `predict()` is not well-defined for tuning results.
      i To predict with the optimal model configuration from tuning results, ensure that the tuning result was generated with the control option (`?tune::control_grid()`) `save_workflow = TRUE`, run `fit_best()` (`?tune::fit_best()`), and then predict using `predict()` (`?workflows::predict.workflow()`) on its output.
      i To collect predictions from tuning results, ensure that the tuning result was generated with the control option (`?tune::control_grid()`) `save_pred = TRUE` and run `collect_predictions()` (`?tune::collect_predictions()`).

# predict errors informatively when supplied last_fit() results

    Code
      predict(spline_res)
    Condition
      Error in `predict()`:
      ! `predict()` is not well-defined for `last_fit()` results.
      i To predict with the model fitted in `last_fit()`, first run `extract_workflow()` (`?tune::extract_workflow.last_fit()`) and then predict using `predict()` (`?workflows::predict.workflow()`) on its output.

# metric inputs are checked for regression models

    Code
      check_metrics_arg(NULL, wflow)
    Output
      A metric set, consisting of:
      - `rmse()`, a numeric metric | direction: minimize
      - `rsq()`, a numeric metric  | direction: maximize

---

    Code
      check_metrics_arg(met_reg, wflow)
    Output
      A metric set, consisting of:
      - `rmse()`, a numeric metric | direction: minimize

---

    Code
      check_metrics_arg(met_cls, wflow)
    Condition
      Error:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      check_metrics_arg(met_mix_int, wflow)
    Condition
      Error:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      fit_resamples(wflow, rs, metrics = met_cls)
    Condition
      Error in `fit_resamples()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      fit_resamples(wflow, rs, metrics = met_mix_int)
    Condition
      Error in `fit_resamples()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_cls)
    Condition
      Error in `tune_grid()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_mix_int)
    Condition
      Error in `tune_grid()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_bayes(wflow_tune, rs, metrics = met_cls)
    Condition
      Error in `tune_bayes()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_bayes(wflow_tune, rs, metrics = met_mix_int)
    Condition
      Error in `tune_bayes()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      last_fit(wflow, split, metrics = met_cls)
    Condition
      Error in `last_fit()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      last_fit(wflow, split, metrics = met_mix_int)
    Condition
      Error in `last_fit()`:
      ! The parsnip model has `mode` value of "regression", but the `metrics` is a metric set for a different model mode.

# metric inputs are checked for classification models

    Code
      check_metrics_arg(NULL, wflow)
    Output
      A metric set, consisting of:
      - `roc_auc()`, a probability metric     | direction: maximize
      - `accuracy()`, a class metric          | direction: maximize
      - `brier_class()`, a probability metric | direction: minimize

---

    Code
      check_metrics_arg(met_reg, wflow)
    Condition
      Error:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      check_metrics_arg(met_cls, wflow)
    Output
      A metric set, consisting of:
      - `brier_class()`, a probability metric | direction: minimize

---

    Code
      check_metrics_arg(met_mix_int, wflow)
    Condition
      Error:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      fit_resamples(wflow, rs, metrics = met_reg)
    Condition
      Error in `fit_resamples()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      fit_resamples(wflow, rs, metrics = met_mix_int)
    Condition
      Error in `fit_resamples()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_reg)
    Condition
      Error in `tune_grid()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_mix_int)
    Condition
      Error in `tune_grid()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_bayes(wflow_tune, rs, metrics = met_reg)
    Condition
      Error in `tune_bayes()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_bayes(wflow_tune, rs, metrics = met_mix_int)
    Condition
      Error in `tune_bayes()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      last_fit(wflow, split, metrics = met_reg)
    Condition
      Error in `last_fit()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

---

    Code
      last_fit(wflow, split, metrics = met_mix_int)
    Condition
      Error in `last_fit()`:
      ! The parsnip model has `mode` value of "classification", but the `metrics` is a metric set for a different model mode.

# metric inputs are checked for censored regression models

    Code
      check_metrics_arg(NULL, wflow)
    Output
      A metric set, consisting of:
      - `brier_survival()`, a dynamic survival metric | direction: minimize

---

    Code
      check_metrics_arg(met_reg, wflow)
    Condition
      Error:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      check_metrics_arg(met_cls, wflow)
    Condition
      Error:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      check_metrics_arg(met_srv, wflow)
    Output
      A metric set, consisting of:
      - `concordance_survival()`, a static survival metric | direction: maximize

---

    Code
      fit_resamples(wflow, rs, metrics = met_cls)
    Condition
      Error in `fit_resamples()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      fit_resamples(wflow, rs, metrics = met_reg)
    Condition
      Error in `fit_resamples()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_cls)
    Condition
      Error in `tune_grid()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_grid(wflow_tune, rs, metrics = met_reg)
    Condition
      Error in `tune_grid()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_bayes(wflow_tune, rs, metrics = met_cls)
    Condition
      Error in `tune_bayes()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      tune_bayes(wflow_tune, rs, metrics = met_reg)
    Condition
      Error in `tune_bayes()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      last_fit(wflow, split, metrics = met_cls)
    Condition
      Error in `last_fit()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

---

    Code
      last_fit(wflow, split, metrics = met_reg)
    Condition
      Error in `last_fit()`:
      ! The parsnip model has `mode` value of "censored regression", but the `metrics` is a metric set for a different model mode.

# selecting single eval time - pure metric sets

    Code
      first_eval_time(met_dyn, eval_time = NULL)
    Condition
      Error:
      ! A single evaluation time is required to use this metric.

---

    Code
      first_eval_time(met_dyn, metric = "brier_survival", eval_time = NULL)
    Condition
      Error:
      ! A single evaluation time is required to use this metric.

---

    Code
      dyn_multi <- first_eval_time(met_dyn, eval_time = times_2)
    Condition
      Warning:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 0.71429`).

# selecting single eval time - mixed metric sets - dynamic first

    Code
      first_eval_time(met_mix_dyn, eval_time = NULL)
    Condition
      Error:
      ! A single evaluation time is required to use this metric.

---

    Code
      dyn_multi <- first_eval_time(met_mix_dyn, eval_time = times_2)
    Condition
      Warning:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 0.71429`).

---

    Code
      first_eval_time(met_mix_dyn_all, eval_time = NULL)
    Condition
      Error:
      ! A single evaluation time is required to use this metric.

---

    Code
      dyn_multi <- first_eval_time(met_mix_dyn_all, eval_time = times_2)
    Condition
      Warning:
      2 evaluation times are available; the first will be used (i.e. `eval_time = 0.71429`).

# selecting an evaluation time

    Code
      choose_eval_time(surv_res, "brier_survival")
    Condition
      Warning:
      4 evaluation times are available; the first will be used (i.e. `eval_time = 10`).
    Output
      [1] 10

---

    Code
      choose_eval_time(surv_res, "concordance_survival")
    Output
      NULL

---

    Code
      choose_eval_time(surv_res, "concordance_survival", eval_time = 10)
    Condition
      Warning:
      `eval_time` is only used for dynamic survival metrics.
    Output
      NULL

---

    Code
      choose_eval_time(ames_grid_search, "rmse", eval_time = 1)
    Condition
      Warning:
      `eval_time` is only used for models with mode "censored regression".
    Output
      NULL

# model_fit method

    Code
      last_fit(lm_fit)
    Condition
      Error in `last_fit()`:
      ! `last_fit()` (`?tune::last_fit()`) is not well-defined for fitted model objects.
      i `last_fit()` (`?tune::last_fit()`) takes a model specification (`?parsnip::model_spec()`) or unfitted workflow (`?workflows::workflow()`) as its first argument.

# workflow method

    Code
      last_fit(lm_fit)
    Condition
      Error in `last_fit()`:
      ! `last_fit()` is not well-defined for a fitted workflow.

# ellipses with last_fit

    Code
      last_fit(set_engine(linear_reg(), "lm"), f, split, something = "wrong")
    Condition
      Warning:
      The `...` are not used in this function but 1 object was passed: "something"
    Output
      # Resampling results
      # Manual resampling 
      # A tibble: 1 x 6
        splits         id               .metrics         .notes           .predictions     .workflow 
        <list>         <chr>            <list>           <list>           <list>           <list>    
      1 <split [24/8]> train/test split <tibble [2 x 4]> <tibble [0 x 4]> <tibble [8 x 4]> <workflow>

# argument order gives errors for recipe/formula

    Code
      last_fit(rec, lin_mod, split)
    Condition
      Error in `last_fit()`:
      ! The first argument to `last_fit()` should be either a model or workflow, not a <recipe> object.

---

    Code
      last_fit(f, lin_mod, split)
    Condition
      Error in `last_fit()`:
      ! The first argument to `last_fit()` should be either a model or workflow, not a <formula> object.

# `last_fit()` when objects need tuning

    2 arguments have been tagged for tuning in these components: model_spec and recipe.
    i Please use one of the tuning functions (e.g. `tune_grid()`) to optimize them.

---

    1 argument has been tagged for tuning in this component: model_spec.
    i Please use one of the tuning functions (e.g. `tune_grid()`) to optimize them.

---

    1 argument has been tagged for tuning in this component: recipe.
    i Please use one of the tuning functions (e.g. `tune_grid()`) to optimize them.

# bad argss

    Code
      conf_mat_resampled(svm_results)
    Condition
      Error in `conf_mat_resampled()`:
      ! It looks like there are 5 tuning parameter combination in the data. Please use the `parameters` argument to select one combination of parameters.

---

    Code
      conf_mat_resampled(mt_knn_bo)
    Condition
      Error in `conf_mat_resampled()`:
      ! Cannot find the predicted classes. Was this a classification model?

---

    Code
      conf_mat_resampled(broke_results)
    Condition
      Error in `conf_mat_resampled()`:
      ! The function was not run with the `save_pred = TRUE` option. Please re-run with that option.

---

    Code
      conf_mat_resampled(tibble::as_tibble(svm_results))
    Condition
      Error in `conf_mat_resampled()`:
      ! The first argument needs to be <tune_results> object, not a data frame.

---

    Code
      conf_mat_resampled(broke_results)
    Condition
      Error in `conf_mat_resampled()`:
      ! Cannot find the predicted classes. Was this a classification model?

---

    Code
      conf_mat_resampled(broke_results, parameters = select_best(broke_results,
        metric = "accuracy"))
    Condition
      Error in `conf_mat_resampled()`:
      ! Cannot determine the proper outcome name

---

    Code
      conf_mat_resampled(svm_results, argument_that_doesnt_exist = TRUE)
    Condition
      Error in `conf_mat_resampled()`:
      ! `...` must be empty.
      x Problematic argument:
      * argument_that_doesnt_exist = TRUE

# set configurations - single element

    Code
      tune:::.get_config_key(tibble::tibble(polynimial_degree = 1:2), wflow_pre)
    Condition
      Error:
      ! Some parameters are tagged for tuning but are not in the grid: `degree`

---

    Code
      tune:::.get_config_key(tibble::tibble(degree = 1:3), wflow_pre)
    Output
      # A tibble: 3 x 2
        degree .config        
         <int> <chr>          
      1      1 pre1_mod0_post0
      2      2 pre2_mod0_post0
      3      3 pre3_mod0_post0

---

    Code
      tune:::.get_config_key(tibble::tibble(degree = 1), wflow_pre)
    Output
      # A tibble: 1 x 2
        degree .config        
         <dbl> <chr>          
      1      1 pre1_mod0_post0

---

    Code
      tune:::.get_config_key(tibble::tibble(min_n = 10:12), wflow_mod)
    Output
      # A tibble: 3 x 2
        min_n .config        
        <int> <chr>          
      1    10 pre0_mod1_post0
      2    11 pre0_mod2_post0
      3    12 pre0_mod3_post0

---

    Code
      tune:::.get_config_key(tibble::tibble(cut = seq(0, 1, length.out = 5)),
      wflow_post)
    Output
      # A tibble: 5 x 2
          cut .config        
        <dbl> <chr>          
      1  0    pre0_mod0_post1
      2  0.25 pre0_mod0_post2
      3  0.5  pre0_mod0_post3
      4  0.75 pre0_mod0_post4
      5  1    pre0_mod0_post5

# set configurations - two elements

    Code
      tune:::.get_config_key(tidyr::crossing(degree = 1:3, neighbors = 1:2), wflow_1)
    Output
      # A tibble: 6 x 3
        neighbors degree .config        
            <int>  <int> <chr>          
      1         1      1 pre1_mod1_post0
      2         2      1 pre1_mod2_post0
      3         1      2 pre2_mod1_post0
      4         2      2 pre2_mod2_post0
      5         1      3 pre3_mod1_post0
      6         2      3 pre3_mod2_post0

---

    Code
      tune:::.get_config_key(slice(tidyr::crossing(min_n = 1:3, cut = (1:5) / 5), -1),
      wflow_2)
    Output
      # A tibble: 14 x 3
         min_n   cut .config        
         <int> <dbl> <chr>          
       1     1   0.4 pre0_mod1_post2
       2     1   0.6 pre0_mod1_post3
       3     1   0.8 pre0_mod1_post4
       4     1   1   pre0_mod1_post5
       5     2   0.2 pre0_mod2_post1
       6     2   0.4 pre0_mod2_post2
       7     2   0.6 pre0_mod2_post3
       8     2   0.8 pre0_mod2_post4
       9     2   1   pre0_mod2_post5
      10     3   0.2 pre0_mod3_post1
      11     3   0.4 pre0_mod3_post2
      12     3   0.6 pre0_mod3_post3
      13     3   0.8 pre0_mod3_post4
      14     3   1   pre0_mod3_post5

---

    Code
      tune:::.get_config_key(grid, wflow_3)
    Output
      # A tibble: 10 x 3
         degree   cut .config         
          <int> <dbl> <chr>           
       1      1 0     pre1_mod0_post01
       2      1 0.333 pre1_mod0_post04
       3      1 0.444 pre1_mod0_post05
       4      1 0.778 pre1_mod0_post08
       5      2 0.111 pre2_mod0_post02
       6      2 0.556 pre2_mod0_post06
       7      2 0.889 pre2_mod0_post09
       8      3 0.222 pre3_mod0_post03
       9      3 0.667 pre3_mod0_post07
      10      3 1     pre3_mod0_post10

# set configurations - all or none

    Code
      tune:::.get_config_key(tibble(), wflow_none)
    Output
      # A tibble: 1 x 1
        .config        
        <chr>          
      1 pre0_mod0_post0

---

    Code
      tune:::.get_config_key(grid, wflow_all)
    Output
      # A tibble: 5 x 4
        min_n degree   cut .config        
        <int>  <int> <dbl> <chr>          
      1    21      1  1    pre1_mod3_post5
      2    30      1  0.25 pre1_mod4_post2
      3     2      2  0    pre2_mod1_post1
      4    40      2  0.5  pre2_mod5_post3
      5    11      3  0.75 pre3_mod2_post4

# tuning with engine parameters without dials objects

    Code
      rf_tune <- tune_grid(rf_mod, mpg ~ ., resamples = rs, grid = 3)
    Condition
      Error in `.check_param_objects()`:
      ! The workflow has arguments to be tuned that are missing some parameter objects: "corr.bias"

---

    Code
      p <- autoplot(rf_tune)
    Condition
      Error in `autoplot()`:
      ! Some parameters do not have corresponding parameter objects and cannot be used with `autoplot()`: `corr.bias`.

---

    Code
      rf_search <- tune_bayes(rf_mod, mpg ~ ., resamples = rs)
    Condition
      Error in `.check_param_objects()`:
      ! The workflow has arguments to be tuned that are missing some parameter objects: "corr.bias"

# cannot finalize with recipe parameters

    Code
      tune_grid(mod_1, rec_1, resamples = rs, grid = 3)
    Condition
      Error in `check_parameters()`:
      ! Some model parameters require finalization but there are recipe parameters that require tuning.
      i Please use `extract_parameter_set_dials()` to set parameter ranges manually and supply the output to the `param_info` argument.

# finalize tailors

    Code
      finalize_tailor(linear_reg(), tibble())
    Condition
      Error in `finalize_tailor()`:
      ! `x` should be a tailor, not a <linear_reg> object.

# `collect_predictions()` errors informatively if there is no `.predictions` column

    Code
      collect_predictions(dplyr::select(lm_splines, -.predictions))
    Condition
      Error in `collect_predictions()`:
      ! The .predictions column does not exist. Please refit with the control argument `save_pred = TRUE` to save predictions.

# `collect_predictions()` errors informatively applied to unsupported class

    Code
      collect_predictions(lm(mpg ~ disp, mtcars))
    Condition
      Error in `collect_predictions()`:
      ! No `collect_predictions()` exists for a <lm> object.

# bad filter grid

    Code
      collect_predictions(svm_tune, parameters = tibble(wrong = "value"))
    Condition
      Error in `filter_predictions()`:
      ! `parameters` should only have columns: "cost value".

# collecting notes - fit_resamples

    Code
      lm_splines <- fit_resamples(lin_mod, mpg ~ ., flds)
    Message
      > A | warning: prediction from rank-deficient fit; consider predict(., rankdeficient="NA")

---

    Code
      lm_splines
    Output
      # Resampling results
      # Bootstrap sampling 
      # A tibble: 2 x 4
        splits          id         .metrics         .notes          
        <list>          <chr>      <list>           <list>          
      1 <split [32/13]> Bootstrap1 <tibble [2 x 4]> <tibble [1 x 4]>
      2 <split [32/17]> Bootstrap2 <tibble [2 x 4]> <tibble [1 x 4]>
      
      There were issues with some computations:
      
        - Warning(s) x2: prediction from rank-deficient fit; consider predict(., rankdefic...
      
      Run `show_notes(.Last.tune.result)` for more information.

# collecting notes - last_fit

    Code
      lst <- last_fit(lin_mod, mpg ~ ., split)
    Message
      > A | warning: prediction from rank-deficient fit; consider predict(., rankdeficient="NA")

---

    Code
      lst
    Output
      # Resampling results
      # Manual resampling 
      # A tibble: 1 x 6
        splits         id               .metrics .notes   .predictions     .workflow 
        <list>         <chr>            <list>   <list>   <list>           <list>    
      1 <split [24/8]> train/test split <tibble> <tibble> <tibble [8 x 4]> <workflow>
      
      There were issues with some computations:
      
        - Warning(s) x1: prediction from rank-deficient fit; consider predict(., rankdefic...
      
      Run `show_notes(.Last.tune.result)` for more information.

# `collect_notes()` errors informatively applied to unsupported class

    Code
      collect_notes(lm(mpg ~ disp, mtcars))
    Condition
      Error in `collect_notes()`:
      ! No `collect_notes()` exists for a <lm> object.

# collecting extracted objects - fit_resamples

    Code
      collect_extracts(res_fit)
    Output
      # A tibble: 5 x 3
        id         .extracts .config        
        <chr>      <list>    <chr>          
      1 Bootstrap1 <lm>      pre0_mod0_post0
      2 Bootstrap2 <lm>      pre0_mod0_post0
      3 Bootstrap3 <lm>      pre0_mod0_post0
      4 Bootstrap4 <lm>      pre0_mod0_post0
      5 Bootstrap5 <lm>      pre0_mod0_post0

---

    Code
      collect_extracts(res_nothing)
    Condition
      Error in `collect_extracts()`:
      ! The `.extracts` column does not exist.
      i Please supply a control object (`?tune::control_grid()`) with a non-`NULL` `extract` argument during resample fitting.

---

    Code
      collect_extracts(res_error)
    Output
      # A tibble: 5 x 3
        id         .extracts      .config        
        <chr>      <list>         <chr>          
      1 Bootstrap1 <try-errr [1]> pre0_mod0_post0
      2 Bootstrap2 <try-errr [1]> pre0_mod0_post0
      3 Bootstrap3 <try-errr [1]> pre0_mod0_post0
      4 Bootstrap4 <try-errr [1]> pre0_mod0_post0
      5 Bootstrap5 <try-errr [1]> pre0_mod0_post0

# `collect_extracts()` errors informatively applied to unsupported class

    Code
      collect_extracts(lm(mpg ~ disp, mtcars))
    Condition
      Error in `collect_extracts()`:
      ! No `collect_extracts()` exists for a <lm> object.

# `collect_metrics()` errors informatively applied to unsupported class

    Code
      collect_metrics(lm(mpg ~ disp, mtcars))
    Condition
      Error in `collect_metrics()`:
      ! No `collect_metrics()` exists for a <lm> object.

# `collect_metrics(type)` errors informatively with bad input

    Code
      collect_metrics(ames_grid_search, type = "boop")
    Condition
      Error in `collect_metrics()`:
      ! `type` must be one of "long" or "wide", not "boop".

---

    Code
      collect_metrics(ames_grid_search, type = NULL)
    Condition
      Error in `collect_metrics()`:
      ! `type` must be a string or character vector.

# augment fit_resamples

    Code
      names(aug_1)
    Output
      [1] "Class"        ".pred_class"  ".pred_Class1" ".pred_Class2" "A"           
      [6] "B"           

---

    Code
      augment(fit_1, hey = "you")
    Condition
      Error in `augment()`:
      ! `...` must be empty.
      x Problematic argument:
      * hey = "you"

---

    Code
      aug_2 <- augment(fit_2)
    Condition
      Warning:
      The original data had 791 rows but there were 593 hold-out predictions.

---

    Code
      names(aug_2)
    Output
      [1] "Class"        ".pred_class"  ".pred_Class1" ".pred_Class2" "A"           
      [6] "B"           

# augment tune_grid

    Code
      names(aug_1)
    Output
       [1] "mpg"    ".pred"  ".resid" "cyl"    "disp"   "hp"     "drat"   "wt"    
       [9] "qsec"   "vs"     "am"     "gear"   "carb"  

---

    Code
      augment(fit_1, parameters = list(cost = 3))
    Condition
      Error in `augment()`:
      ! `parameters` should be a single row data frame.

---

    Code
      augment(fit_1, parameters = data.frame(cost = 3:4))
    Condition
      Error in `augment()`:
      ! `parameters` should be a single row data frame.

---

    Code
      augment(fit_1, cost = 4)
    Condition
      Error in `augment()`:
      ! `...` must be empty.
      x Problematic argument:
      * cost = 4

---

    Code
      names(aug_3)
    Output
       [1] "mpg"    ".pred"  ".resid" "cyl"    "disp"   "hp"     "drat"   "wt"    
       [9] "qsec"   "vs"     "am"     "gear"   "carb"  

# augment last_fit

    Code
      names(aug_1)
    Output
      [1] "Class"        ".pred_class"  ".pred_Class1" ".pred_Class2" "A"           
      [6] "B"           

---

    Code
      augment(fit_1, potato = TRUE)
    Condition
      Error in `augment()`:
      ! `...` must be empty.
      x Problematic argument:
      * potato = TRUE

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expo_decay.R
\name{expo_decay}
\alias{expo_decay}
\title{Exponential decay function}
\usage{
expo_decay(iter, start_val, limit_val, slope = 1/5)
}
\arguments{
\item{iter}{An integer for the current iteration number.}

\item{start_val}{The number returned for the first iteration.}

\item{limit_val}{The number that the process converges to over iterations.}

\item{slope}{A coefficient for the exponent to control the rate of decay. The
sign of the slope controls the direction of decay.}
}
\value{
A single numeric value.
}
\description{
\code{\link[=expo_decay]{expo_decay()}} can be used to increase or decrease a function exponentially
over iterations. This can be used to dynamically set parameters for
acquisition functions as iterations of Bayesian optimization proceed.
}
\details{
Note that, when used with the acquisition functions in \code{\link[=tune]{tune()}}, a wrapper
would be required since only the first argument would be evaluated during
tuning.
}
\examples{
\dontshow{if (tune:::should_run_examples()) withAutoprint(\{ # examplesIf}
library(tibble)
library(purrr)
library(ggplot2)
library(dplyr)
tibble(
  iter = 1:40,
  value = map_dbl(
    1:40,
    expo_decay,
    start_val = .1,
    limit_val = 0,
    slope = 1 / 5
  )
) |>
  ggplot(aes(x = iter, y = value)) +
  geom_path()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resample.R
\name{fit_resamples}
\alias{fit_resamples}
\alias{fit_resamples.model_spec}
\alias{fit_resamples.workflow}
\title{Fit multiple models via resampling}
\usage{
fit_resamples(object, ...)

\method{fit_resamples}{model_spec}(
  object,
  preprocessor,
  resamples,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_resamples()
)

\method{fit_resamples}{workflow}(
  object,
  resamples,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_resamples()
)
}
\arguments{
\item{object}{A \code{parsnip} model specification or an unfitted
\link[workflows:workflow]{workflow()}. No tuning parameters are allowed; if arguments
have been marked with \link[hardhat:tune]{tune()}, their values must be
\link[=finalize_model]{finalized}.}

\item{...}{Currently unused.}

\item{preprocessor}{A traditional model formula or a recipe created using
\code{\link[recipes:recipe]{recipes::recipe()}}.}

\item{resamples}{An \code{rset} resampling object created from an \code{rsample}
function, such as \code{\link[rsample:vfold_cv]{rsample::vfold_cv()}}.}

\item{metrics}{A \code{\link[yardstick:metric_set]{yardstick::metric_set()}}, or \code{NULL} to compute a standard
set of metrics.}

\item{eval_time}{A numeric vector of time points where dynamic event time
metrics should be computed (e.g. the time-dependent ROC curve, etc). The
values must be non-negative and should probably be no greater than the
largest event time in the training set (See Details below).}

\item{control}{A \code{\link[=control_resamples]{control_resamples()}} object used to fine tune the resampling
process.}
}
\description{
\code{\link[=fit_resamples]{fit_resamples()}} computes a set of performance metrics across one or more
resamples. It does not perform any tuning (see \code{\link[=tune_grid]{tune_grid()}} and
\code{\link[=tune_bayes]{tune_bayes()}} for that), and is instead used for fitting a single
model+recipe or model+formula combination across many resamples.
}
\section{Case Weights}{

Some models can utilize case weights during training. tidymodels currently
supports two types of case weights: importance weights (doubles) and
frequency weights (integers). Frequency weights are used during model
fitting and evaluation, whereas importance weights are only used during
fitting.

To know if your model is capable of using case weights, create a model spec
and test it using \code{\link[parsnip:case_weights_allowed]{parsnip::case_weights_allowed()}}.

To use them, you will need a numeric column in your data set that has been
passed through either \code{\link[hardhat:importance_weights]{hardhat:: importance_weights()}} or
\code{\link[hardhat:frequency_weights]{hardhat::frequency_weights()}}.

For functions such as \code{\link[=fit_resamples]{fit_resamples()}} and the \verb{tune_*()} functions, the
model must be contained inside of a \code{\link[workflows:workflow]{workflows::workflow()}}. To declare that
case weights are used, invoke \code{\link[workflows:add_case_weights]{workflows::add_case_weights()}} with the
corresponding (unquoted) column name.

From there, the packages will appropriately handle the weights during model
fitting and (if appropriate) performance estimation.
}

\section{Censored Regression Models}{


Three types of metrics can be used to assess the quality of censored
regression models:
\itemize{
\item static: the prediction is independent of time.
\item dynamic: the prediction is a time-specific probability (e.g., survival
probability) and is measured at one or more particular times.
\item integrated: same as the dynamic metric but returns the integral of the
different metrics from each time point.
}

Which metrics are chosen by the user affects how many evaluation times
should be specified. For example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# Needs no `eval_time` value
metric_set(concordance_survival)

# Needs at least one `eval_time`
metric_set(brier_survival)
metric_set(brier_survival, concordance_survival)

# Needs at least two eval_time` values
metric_set(brier_survival_integrated, concordance_survival)
metric_set(brier_survival_integrated, concordance_survival)
metric_set(brier_survival_integrated, concordance_survival, brier_survival)
}\if{html}{\out{</div>}}

Values of \code{eval_time} should be less than the largest observed event
time in the training data. For many non-parametric models, the results beyond
the largest time corresponding to an event are constant (or \code{NA}).
}

\section{Performance Metrics}{


To use your own performance metrics, the \code{\link[yardstick:metric_set]{yardstick::metric_set()}} function
can be used to pick what should be measured for each model. If multiple
metrics are desired, they can be bundled. For example, to estimate the area
under the ROC curve as well as the sensitivity and specificity (under the
typical probability cutoff of 0.50), the \code{metrics} argument could be given:

\preformatted{
  metrics = metric_set(roc_auc, sens, spec)
}

Each metric is calculated for each candidate model.

If no metric set is provided, one is created:
\itemize{
\item For regression models, the root mean squared error and coefficient
of determination are computed.
\item For classification, the area under the ROC curve and overall accuracy
are computed.
\item For censored regression, the dynamic Brier score
(\code{\link[yardstick:brier_survival]{yardstick::brier_survival()}}) is used.
\item For quantile regression, the weighted interval score
(\code{\link[yardstick:weighted_interval_score]{yardstick::weighted_interval_score()}}) is used.
}

Note that the metrics also determine what type of predictions are estimated
during tuning. For example, in a classification problem, if metrics are used
that are all associated with hard class predictions, the classification
probabilities are not created.

The out-of-sample estimates of these metrics are contained in a list column
called \code{.metrics}. This tibble contains a row for each metric and columns
for the value, the estimator type, and so on.

\code{\link[=collect_metrics]{collect_metrics()}} can be used for these objects to collapse the results
over the resampled (to obtain the final resampling estimates per tuning
parameter combination).
}

\section{Obtaining Predictions}{


When \code{control_grid(save_pred = TRUE)}, the output tibble contains a list
column called \code{.predictions} that has the out-of-sample predictions for each
parameter combination in the grid and each fold (which can be very large).

The elements of the tibble are tibbles with columns for the tuning
parameters, the row number from the original data object (\code{.row}), the
outcome data (with the same name(s) of the original data), and any columns
created by the predictions. For example, for simple regression problems, this
function generates a column called \code{.pred} and so on. As noted above, the
prediction columns that are returned are determined by the type of metric(s)
requested.

This list column can be \code{unnested} using \code{\link[tidyr:unnest]{tidyr::unnest()}} or using the
convenience function \code{\link[=collect_predictions]{collect_predictions()}}.
}

\section{Extracting Information}{


The \code{extract} control option will result in an additional function to be
returned called \code{.extracts}. This is a list column that has tibbles
containing the results of the user's function for each tuning parameter
combination. This can enable returning each model and/or recipe object that
is created during resampling. Note that this could result in a large return
object, depending on what is returned.

The control function contains an option (\code{extract}) that can be used to
retain any model or recipe that was created within the resamples. This
argument should be a function with a single argument. The value of the
argument that is given to the function in each resample is a workflow
object (see \code{\link[workflows:workflow]{workflows::workflow()}} for more information). Several
helper functions can be used to easily pull out the preprocessing
and/or model information from the workflow, such as
\code{\link[workflows:extract-workflow]{extract_preprocessor()}} and
\code{\link[workflows:extract-workflow]{extract_fit_parsnip()}}.

As an example, if there is interest in getting each parsnip model fit back,
one could use:
\preformatted{
  extract = function (x) extract_fit_parsnip(x)
}

Note that the function given to the \code{extract} argument is evaluated on
every model that is \emph{fit} (as opposed to every model that is \emph{evaluated}).
As noted above, in some cases, model predictions can be derived for
sub-models so that, in these cases, not every row in the tuning parameter
grid has a separate R object associated with it.

Finally, it is a good idea to include calls to \code{\link[=require]{require()}} for packages that
are used in the function. This helps prevent failures when using parallel
processing.
}

\examples{
\dontshow{if (tune:::should_run_examples() & rlang::is_installed("splines2")) withAutoprint(\{ # examplesIf}
library(recipes)
library(rsample)
library(parsnip)
library(workflows)

set.seed(6735)
folds <- vfold_cv(mtcars, v = 5)

spline_rec <- recipe(mpg ~ ., data = mtcars) |>
  step_spline_natural(disp) |>
  step_spline_natural(wt)

lin_mod <- linear_reg() |>
  set_engine("lm")

control <- control_resamples(save_pred = TRUE)

spline_res <- fit_resamples(lin_mod, spline_rec, folds, control = control)

spline_res

show_best(spline_res, metric = "rmse")

# You can also wrap up a preprocessor and a model into a workflow, and
# supply that to `fit_resamples()` instead. Here, a workflows "variables"
# preprocessor is used, which lets you supply terms using dplyr selectors.
# The variables are used as-is, no preprocessing is done to them.
wf <- workflow() |>
  add_variables(outcomes = mpg, predictors = everything()) |>
  add_model(lin_mod)

wf_res <- fit_resamples(wf, folds)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=control_resamples]{control_resamples()}}, \code{\link[=collect_predictions]{collect_predictions()}}, \code{\link[=collect_metrics]{collect_metrics()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_parameters.R
\name{filter_parameters}
\alias{filter_parameters}
\title{Remove some tuning parameter results}
\usage{
filter_parameters(x, ..., parameters = NULL)
}
\arguments{
\item{x}{An object of class \code{tune_results} that has multiple tuning parameters.}

\item{...}{Expressions that return a logical value, and are defined in terms
of the tuning parameter values. If multiple expressions are included, they
are combined with the \code{&} operator. Only rows for which all conditions
evaluate to \code{TRUE} are kept.}

\item{parameters}{A tibble of tuning parameter values that can be used to
filter the predicted values before processing. This tibble should only have
columns for tuning parameter identifiers (e.g. \code{"my_param"} if
\code{tune("my_param")} was used). There can be multiple rows and one or more
columns. \strong{If used, this parameter must be named.}}
}
\value{
A version of \code{x} where the lists columns only retain the parameter
combinations in \code{parameters} or satisfied by the filtering logic.
}
\description{
For objects produced by the \verb{tune_*()} functions, there may only be a subset
of tuning parameter combinations of interest. For large data sets, it might be
helpful to be able to remove some results. This function trims the \code{.metrics}
column of unwanted results as well as columns \code{.predictions} and \code{.extracts}
(if they were requested).
}
\details{
Removing some parameter combinations might affect the results of \code{autoplot()}
for the object.
}
\examples{
library(dplyr)
library(tibble)

# For grid search:
data("example_ames_knn")

## -----------------------------------------------------------------------------
# select all combinations using the 'rank' weighting scheme

ames_grid_search |>
  collect_metrics()

filter_parameters(ames_grid_search, weight_func == "rank") |>
  collect_metrics()

rank_only <- tibble::tibble(weight_func = "rank")
filter_parameters(ames_grid_search, parameters = rank_only) |>
  collect_metrics()

## -----------------------------------------------------------------------------
# Keep only the results from the numerically best combination

ames_iter_search |>
  collect_metrics()

best_param <- select_best(ames_iter_search, metric = "rmse")
ames_iter_search |>
  filter_parameters(parameters = best_param) |>
  collect_metrics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R, R/collect.R,
%   R/grid_helpers.R, R/grid_performance.R, R/iteration_results.R,
%   R/loop_over_all_stages-helpers.R, R/loop_over_all_stages.R, R/merge.R,
%   R/parallel.R, R/symbol.R, R/tune_bayes.R, R/tune_grid.R, R/utils.R
\name{check_rset}
\alias{check_rset}
\alias{.check_grid}
\alias{.needs_finalization}
\alias{check_parameters}
\alias{.check_param_objects}
\alias{check_workflow}
\alias{check_metrics}
\alias{check_initial}
\alias{val_class_or_null}
\alias{val_class_and_single}
\alias{.config_key_from_metrics}
\alias{estimate_tune_results}
\alias{.has_preprocessor}
\alias{.has_preprocessor_recipe}
\alias{.has_preprocessor_formula}
\alias{.has_preprocessor_variables}
\alias{.has_spec}
\alias{.set_workflow_spec}
\alias{.set_workflow_recipe}
\alias{metrics_info}
\alias{new_iteration_results}
\alias{.make_static}
\alias{.get_data_subsets}
\alias{.get_config_key}
\alias{.determine_pred_types}
\alias{.loop_over_all_stages}
\alias{.loop_over_all_stages2}
\alias{.update_model}
\alias{.update_recipe}
\alias{.update_parallel_over}
\alias{loop_call}
\alias{get_tune_colors}
\alias{encode_set}
\alias{.is_cataclysmic}
\alias{check_time}
\alias{pull_rset_attributes}
\alias{.set_workflow}
\alias{empty_ellipses}
\alias{is_recipe}
\alias{is_preprocessor}
\alias{is_workflow}
\alias{new_bare_tibble}
\title{Get colors for tune text.}
\usage{
check_rset(x)

.check_grid(grid, workflow, pset = NULL, call = caller_env())

.needs_finalization(x, nms = character(0))

check_parameters(wflow, pset = NULL, data, grid_names = character(0))

.check_param_objects(pset)

check_workflow(x, ..., pset = NULL, check_dials = FALSE, call = caller_env())

check_metrics(x, object)

check_initial(
  x,
  pset,
  wflow,
  resamples,
  metrics,
  eval_time,
  ctrl,
  checks = "grid"
)

val_class_or_null(x, cls = "numeric", where = NULL)

val_class_and_single(x, cls = "numeric", where = NULL)

.config_key_from_metrics(x)

estimate_tune_results(x, ..., col_name = ".metrics")

.has_preprocessor(workflow)

.has_preprocessor_recipe(workflow)

.has_preprocessor_formula(workflow)

.has_preprocessor_variables(workflow)

.has_spec(workflow)

.set_workflow_spec(workflow, spec)

.set_workflow_recipe(workflow, recipe)

metrics_info(x)

new_iteration_results(
  x,
  parameters,
  metrics,
  eval_time,
  eval_time_target,
  outcomes = character(0),
  rset_info,
  workflow
)

.make_static(
  workflow,
  param_info,
  grid,
  metrics,
  eval_time,
  split_args,
  control,
  pkgs = "tune",
  strategy = "sequential",
  data = list(fit = NULL, pred = NULL, cal = NULL)
)

.get_data_subsets(wflow, split, split_args = NULL)

.get_config_key(grid, wflow)

.determine_pred_types(wflow, metrics)

.loop_over_all_stages(resamples, grid, static)

.loop_over_all_stages2(index, resamples, grid, static)

.update_model(grid, object, pset, step_id, nms, ..., source = "model_spec")

.update_recipe(grid, object, pset, step_id, nms, ...)

.update_parallel_over(control, resamples, grid)

loop_call(strategy, framework, opts)

get_tune_colors()

encode_set(x, pset, ..., as_matrix = FALSE)

.is_cataclysmic(x)

check_time(origin, limit)

pull_rset_attributes(x)

.set_workflow(workflow, control)

empty_ellipses(...)

is_recipe(x)

is_preprocessor(x)

is_workflow(x)

new_bare_tibble(x, ..., class = character())
}
\arguments{
\item{x}{An object.}

\item{workflow}{The workflow used to fit the iteration results.}

\item{pset}{A \code{parameters} object.}

\item{wflow}{A \code{workflow} object.}

\item{data}{The training data.}

\item{grid_names}{A character vector of column names from the grid.}

\item{...}{Other options}

\item{check_dials}{A logical for check for a NULL parameter object.}

\item{object}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

A \code{workflow} object.}

\item{resamples}{An \code{rset} object.}

\item{metrics}{A metric set.}

\item{eval_time}{A numeric vector of time points where dynamic event time
metrics should be computed (e.g. the time-dependent ROC curve, etc).}

\item{ctrl}{A \code{control_grid} object.}

\item{cls}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} A character vector of
possible classes}

\item{where}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} A character string for
the calling function.}

\item{parameters}{A \code{parameters} object.}

\item{outcomes}{A character vector of outcome names.}

\item{rset_info}{Attributes from an \code{rset} object.}

\item{as_matrix}{A logical for the return type.}

\item{origin}{The calculation start time.}

\item{limit}{The allowable time (in minutes).}
}
\description{
These are not intended for use by the general public.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{calculate_resample_weights}
\alias{calculate_resample_weights}
\title{Calculate resample weights from resample sizes}
\usage{
calculate_resample_weights(rset)
}
\arguments{
\item{rset}{An rset object from \pkg{rsample}.}
}
\value{
A numeric vector of weights proportional to resample sizes, normalized
to sum to 1.
}
\description{
This convenience function calculates weights proportional to the number of
observations in each resample's analysis set. Larger resamples get higher weights.
This ensures that resamples with more data have proportionally more influence
on the final aggregated metrics.
}
\details{
This is particularly useful for time-based resamples (e.g., expanding window CV)
or stratified sampling  where resamples might have slightly different sizes, in
which resamples are imbalanced.
}
\examples{
library(rsample)
folds <- vfold_cv(mtcars, v = 3)
weights <- calculate_resample_weights(folds)
weighted_folds <- add_resample_weights(folds, weights)
}
\seealso{
\code{\link[=add_resample_weights]{add_resample_weights()}}, \code{\link[=extract_resample_weights]{extract_resample_weights()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_best.R
\name{show_best}
\alias{show_best}
\alias{show_best.default}
\alias{show_best.tune_results}
\alias{select_best}
\alias{select_best.default}
\alias{select_best.tune_results}
\alias{select_by_pct_loss}
\alias{select_by_pct_loss.default}
\alias{select_by_pct_loss.tune_results}
\alias{select_by_one_std_err}
\alias{select_by_one_std_err.default}
\alias{select_by_one_std_err.tune_results}
\title{Investigate best tuning parameters}
\usage{
show_best(x, ...)

\method{show_best}{default}(x, ...)

\method{show_best}{tune_results}(
  x,
  ...,
  metric = NULL,
  eval_time = NULL,
  n = 5,
  call = rlang::current_env()
)

select_best(x, ...)

\method{select_best}{default}(x, ...)

\method{select_best}{tune_results}(x, ..., metric = NULL, eval_time = NULL)

select_by_pct_loss(x, ...)

\method{select_by_pct_loss}{default}(x, ...)

\method{select_by_pct_loss}{tune_results}(x, ..., metric = NULL, eval_time = NULL, limit = 2)

select_by_one_std_err(x, ...)

\method{select_by_one_std_err}{default}(x, ...)

\method{select_by_one_std_err}{tune_results}(x, ..., metric = NULL, eval_time = NULL)
}
\arguments{
\item{x}{The results of \code{\link[=tune_grid]{tune_grid()}} or \code{\link[=tune_bayes]{tune_bayes()}}.}

\item{...}{For \code{\link[=select_by_one_std_err]{select_by_one_std_err()}} and \code{\link[=select_by_pct_loss]{select_by_pct_loss()}}, this
argument is passed directly to \code{\link[dplyr:arrange]{dplyr::arrange()}} so that the user can sort
the models from \emph{most simple to most complex}. That is, for a parameter \code{p},
pass the unquoted expression \code{p} if smaller values of \code{p} indicate a simpler
model, or \code{desc(p)} if larger values indicate a simpler model. At
least one term is required for these two functions. See the examples below.}

\item{metric}{A character value for the metric that will be used to sort
the models. (See
\url{https://yardstick.tidymodels.org/articles/metric-types.html} for
more details). Not required if a single metric exists in \code{x}. If there are
multiple metric and none are given, the first in the metric set is used (and
a warning is issued).}

\item{eval_time}{A single numeric time point where dynamic event time
metrics should be chosen (e.g., the time-dependent ROC curve, etc). The
values should be consistent with the values used to create \code{x}. The \code{NULL}
default will automatically use the first evaluation time used by \code{x}.}

\item{n}{An integer for the number of top results/rows to return.}

\item{call}{The call to be shown in errors and warnings.}

\item{limit}{The limit of loss of performance that is acceptable (in percent
units). See details below.}
}
\value{
A tibble with columns for the parameters. \code{\link[=show_best]{show_best()}} also
includes columns for performance metrics.
}
\description{
\code{\link[=show_best]{show_best()}} displays the top sub-models and their performance estimates.

\code{\link[=select_best]{select_best()}} finds the tuning parameter combination with the best
performance values.

\code{\link[=select_by_one_std_err]{select_by_one_std_err()}} uses the "one-standard error rule" (Breiman _el
at, 1984) that selects the most simple model that is within one standard
error of the numerically optimal results.

\code{\link[=select_by_pct_loss]{select_by_pct_loss()}} selects the most simple model whose loss of
performance is within some acceptable limit.
}
\details{
For percent loss, suppose the best model has an RMSE of 0.75 and a simpler
model has an RMSE of 1. The percent loss would be \code{(1.00 - 0.75)/1.00 * 100},
or 25 percent. Note that loss will always be non-negative.
}
\examples{
\dontshow{if (tune:::should_run_examples()) withAutoprint(\{ # examplesIf}
data("example_ames_knn")

show_best(ames_iter_search, metric = "rmse")

select_best(ames_iter_search, metric = "rsq")

# To find the least complex model within one std error of the numerically
# optimal model, the number of nearest neighbors are sorted from the largest
# number of neighbors (the least complex class boundary) to the smallest
# (corresponding to the most complex model).

select_by_one_std_err(ames_grid_search, metric = "rmse", desc(K))

# Now find the least complex model that has no more than a 5\% loss of RMSE:
select_by_pct_loss(
  ames_grid_search,
  metric = "rmse",
  limit = 5, desc(K)
)
\dontshow{\}) # examplesIf}
}
\references{
Breiman, Leo; Friedman, J. H.; Olshen, R. A.; Stone, C. J. (1984).
\emph{Classification and Regression Trees.} Monterey, CA: Wadsworth.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{extract_resample_weights}
\alias{extract_resample_weights}
\title{Extract resample weights from rset or tuning objects}
\usage{
extract_resample_weights(x)
}
\arguments{
\item{x}{An rset object with resample weights, or a \code{tune_results} object.}
}
\value{
A numeric vector of resample weights, or NULL if no weights are present.
}
\description{
This function provides a consistent interface to access resample weights
regardless of whether they were added to an rset object or are stored
in \code{tune_results} after tuning.
}
\examples{
\dontrun{
library(rsample)
folds <- vfold_cv(mtcars, v = 3)
weighted_folds <- add_resample_weights(folds, c(0.2, 0.3, 0.5))
extract_resample_weights(weighted_folds)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control_grid}
\alias{control_grid}
\alias{control_resamples}
\alias{new_backend_options}
\title{Control aspects of the grid search process}
\usage{
control_grid(
  verbose = FALSE,
  allow_par = TRUE,
  extract = NULL,
  save_pred = FALSE,
  pkgs = NULL,
  save_workflow = FALSE,
  event_level = "first",
  parallel_over = NULL,
  backend_options = NULL,
  workflow_size = 100
)

control_resamples(
  verbose = FALSE,
  allow_par = TRUE,
  extract = NULL,
  save_pred = FALSE,
  pkgs = NULL,
  save_workflow = FALSE,
  event_level = "first",
  parallel_over = NULL,
  backend_options = NULL,
  workflow_size = 100
)

new_backend_options(..., class = character())
}
\arguments{
\item{verbose}{A logical for logging results (other than warnings and errors,
which are always shown) as they are generated during training in a single
R process. When using most parallel backends, this argument typically will
not result in any logging. If using a dark IDE theme, some logging messages
might be hard to see; try setting the \code{tidymodels.dark} option with
\code{options(tidymodels.dark = TRUE)} to print lighter colors.}

\item{allow_par}{A logical to allow parallel processing (if a parallel
backend is registered).}

\item{extract}{An optional function with at least one argument (or \code{NULL})
that can be used to retain arbitrary objects from the model fit object,
recipe, or other elements of the workflow.}

\item{save_pred}{A logical for whether the out-of-sample predictions should
be saved for each model \emph{evaluated}.}

\item{pkgs}{An optional character string of R package names that should be
loaded (by namespace) during parallel processing.}

\item{save_workflow}{A logical for whether the workflow should be appended
to the output as an attribute.}

\item{event_level}{A single string containing either \code{"first"} or \code{"second"}.
This argument is passed on to yardstick metric functions when any type
of class prediction is made, and specifies which level of the outcome
is considered the "event".}

\item{parallel_over}{A single string containing either \code{"resamples"} or
\code{"everything"} describing how to use parallel processing. Alternatively,
\code{NULL} is allowed, which chooses between \code{"resamples"} and \code{"everything"}
automatically.

If \code{"resamples"}, then tuning will be performed in parallel over resamples
alone. Within each resample, the preprocessor (i.e. recipe or formula) is
processed once, and is then reused across all models that need to be fit.

If \code{"everything"}, then tuning will be performed in parallel at two levels.
An outer parallel loop will iterate over resamples. Additionally, an
inner parallel loop will iterate over all unique combinations of
preprocessor and model tuning parameters for that specific resample. This
will result in the preprocessor being re-processed multiple times, but
can be faster if that processing is extremely fast.

If \code{NULL}, chooses \code{"resamples"} if there are more than one resample,
otherwise chooses \code{"everything"} to attempt to maximize core utilization.

Note that switching between \code{parallel_over} strategies is not guaranteed
to use the same random number generation schemes. However, re-tuning a
model using the same \code{parallel_over} strategy is guaranteed to be
reproducible between runs.}

\item{backend_options}{An object of class \code{"tune_backend_options"} as created
by \code{tune::new_backend_options()}, used to pass arguments to specific tuning
backend. Defaults to \code{NULL} for default backend options.}

\item{workflow_size}{A non-negative number (in MB) that is used as a
threshold for a warning regarding the size of the workflow. Only used when
\code{save_workflow = TRUE}.}
}
\description{
Control aspects of the grid search process
}
\details{
For \code{extract}, this function can be used to output the model object, the
recipe (if used), or some components of either or both. When evaluated, the
function's sole argument has a fitted workflow If the formula method is used,
the recipe element will be \code{NULL}.

The results of the \code{extract} function are added to a list column in the
output called \code{.extracts}. Each element of this list is a tibble with tuning
parameter column and a list column (also called \code{.extracts}) that contains
the results of the function. If no extraction function is used, there is no
\code{.extracts} column in the resulting object. See \code{\link[=tune_bayes]{tune_bayes()}} for more
specific details.

Note that for \code{\link[=collect_predictions]{collect_predictions()}}, it is possible that each row of the
original data point might be represented multiple times per tuning
parameter. For example, if the bootstrap or repeated cross-validation are
used, there will be multiple rows since the sample data point has been
evaluated multiple times. This may cause issues when merging the predictions
with the original data.

\code{\link[=control_resamples]{control_resamples()}} is an alias for \code{\link[=control_grid]{control_grid()}} and is meant to be
used with \code{\link[=fit_resamples]{fit_resamples()}}.
}
\section{Hyperparameters and extracted objects}{


When making use of submodels, tune can generate predictions and calculate
metrics for multiple model \code{.config}urations using only one model fit.
However, this means that if a function was supplied to a
\link[=control_grid]{control function's} \code{extract} argument, tune can only
execute that extraction on the one model that was fitted. As a result,
in the \code{collect_extracts()} output, tune opts to associate the
extracted objects with the hyperparameter combination used to
fit that one model workflow, rather than the hyperparameter
combination of a submodel. In the output, this appears like
a hyperparameter entry is recycled across many \code{.config}
entries---this is intentional.

See \url{https://parsnip.tidymodels.org/articles/Submodels.html} to learn
more about submodels.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schedule.R
\name{schedule_grid}
\alias{schedule_grid}
\title{Schedule a grid}
\usage{
schedule_grid(grid, wflow)
}
\arguments{
\item{grid}{A tibble containing the parameter grid.}

\item{wflow}{The workflow object for which we schedule the grid.}
}
\value{
A schedule object, inheriting from either 'single_schedule',
'grid_schedule', or 'resample_schedule'.
}
\description{
Schedule a grid
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune-package.R
\docType{package}
\name{tune-package}
\alias{tune-package}
\title{tune: Tidy Tuning Tools}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

The ability to tune models is important. 'tune' contains functions and classes to be used in conjunction with other 'tidymodels' packages for finding reasonable values of hyper-parameters in models, preprocessing methods, and post-processing steps.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://tune.tidymodels.org/}
  \item \url{https://github.com/tidymodels/tune}
  \item Report bugs at \url{https://github.com/tidymodels/tune/issues}
}

}
\author{
\strong{Maintainer}: Max Kuhn \email{max@posit.co} (\href{https://orcid.org/0000-0003-2402-136X}{ORCID})

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel.R
\name{parallelism}
\alias{parallelism}
\title{Support for parallel processing in tune}
\description{
\pkg{tune} can enable simultaneous parallel computations. Tierney (2008)
defined different classes of parallel processing techniques:
\itemize{
\item \emph{Implicit} is when a function uses low-level tools to perform a
calculation that is small in scope in parallel. Examples are using
multithreaded linear algebra libraries (e.g., BLAS) or basic R vectorization
functions.
\item \emph{Explicit} parallelization occurs when the user requests that some
calculations should be run by generating multiple new R (sub)processes. These
calculations can be more complex than those for implicit parallel
processing.
}

For example, some decision tree libraries can implicitly parallelize their
search for the optimal splitting routine using multiple threads.

Alternatively, if you are resampling a model \emph{B} times, you can explicitly
create \emph{B} new R jobs to train \emph{B} boosted trees in parallel and return their
resampling results to the main R process (e.g., \code{\link[=fit_resamples]{fit_resamples()}}).

There are two frameworks that can be used to explicitly parallel process
your work in \pkg{tune}: the \link[future:future]{future} package and the
\link[mirai:mirai]{mirai} package. Previously, you could use the
\link[foreach:foreach]{foreach} package, but this has been deprecated as of
version 1.2.1 of tune.

By default, no parallelism is used to process models in \pkg{tune}; you have
to opt-in.
\subsection{Using future}{

You should install the package and choose your flavor of parallelism using
the \link[future:plan]{plan} function. This allows you to specify the number of
worker processes and the specific technology to use.

For example, you can use:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{   library(future)
   plan(multisession, workers = 4)
}\if{html}{\out{</div>}}

and work will be conducted simultaneously (unless there is an exception; see
the section below).

If you had previously used \pkg{foreach}, this would replace your existing
code that probably looked like:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{   library(doBackend)
   registerDoBackend(cores = 4)
}\if{html}{\out{</div>}}

See \code{\link[future:plan]{future::plan()}} for possible options other than \code{multisession}.

Note that \pkg{tune} resets the \emph{maximum} limit of memory of global variables
(e.g., attached packages) to be greater than the default when the package is
loaded. This value can be altered using \code{options(future.globals.maxSize)}.

If you want \pkg{future} to use \pkg{mirai} parallel workers, you can
install and load the \pkg{future.mirai} package.
}

\subsection{Using mirai}{

To set the specific for parallel processing with \pkg{mirai}, use the
\code{\link[mirai:daemons]{mirai::daemons()}} function. The first argument, \code{n}, determines the number
of parallel workers. Using \code{daemons(0)} reverts to sequential processing.

The arguments \code{url} and \code{remote} are used to set up and launch parallel
processes over the network for distributed computing. See \code{\link[mirai:daemons]{mirai::daemons()}}
documentation for more details.
}

\subsection{Reverting to sequential processing}{

There are a few times when you might specify that you wish to use parallel
processing, but it will revert to sequential execution:
\itemize{
\item Many of the control functions (e.g. \code{\link[=control_grid]{control_grid()}}) have an argument
called \code{allow_par}. If this is set to \code{FALSE}, parallel backends will
always be ignored.
\item Some packages, such as \pkg{rJava} and \pkg{keras} are not compatible with
explicit parallelization. If any of these packages are used, sequential
processing occurs.
\item If you specify fewer than two workers, or if there is only a single task,
the computations will occur sequentially.
}
}

\subsection{Expectations for reproducibility}{

We advise that you \emph{always} run \code{\link[=set.seed]{set.seed()}} with a seed value just prior to
using a function that uses (or might use) random numbers. Given this:
\itemize{
\item You should expect to get the same results if you run that section of code
repeatedly, conditional on using version 1.4.0 of tune.
\item You should expect differences in results between version 1.4.0 of tune and
previous versions.
\item When using \code{\link[=last_fit]{last_fit()}}, you should be able to get the same results as
manually using \code{\link[generics:fit]{generics::fit()}} and \code{\link[=predict]{predict()}} to do the same work.
\item When running with or without parallel processing (using any backend
package), you should be able to achieve the same results from
\code{\link[=fit_resamples]{fit_resamples()}} and the various tuning functions.
}

Specific exceptions:
\itemize{
\item For SVM classification models using the \pkg{kernlab} package, the random
number generator is independent of R, and there is no argument to control
it. Unfortunately, it is likely to give you different results from
run-to-run.
\item For some deep learning packages (e.g., \pkg{tensorflow}, \pkg{keras}, and
\pkg{torch}), it is very difficult to achieve reproducible results. This
is especially true when using GPUs for computations. Additionally, we have
seen differences in computations (stochastic or non-random) between
platforms due to the packages' use of different numerical tolerance
constants across operating systems.
}
}

\subsection{Handling package dependencies}{

\pkg{tune} knows what packages are required to fit a workflow object.

When computations are run sequentially, an initial check is made to see if
they are installed. This triggers the packages to be loaded but not visible
in the search path.

In parallel, the required packages are fully loaded (i.e., loaded and seen
in the search path), as they were previously with \pkg{foreach}, in the
worker processes (but not the main R session).
}
}
\references{
\url{https://www.tmwr.org/grid-search#parallel-processing}

Tierney, Luke. "Implicit and explicit parallel computing in R." COMPSTAT
2008: Proceedings in Computational Statistics. Physica-Verlag HD, 2008.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_pctl.R
\name{int_pctl.tune_results}
\alias{int_pctl.tune_results}
\title{Bootstrap confidence intervals for performance metrics}
\usage{
\method{int_pctl}{tune_results}(
  .data,
  metrics = NULL,
  eval_time = NULL,
  times = 1001,
  parameters = NULL,
  alpha = 0.05,
  allow_par = TRUE,
  event_level = "first",
  keep_replicates = FALSE,
  ...
)
}
\arguments{
\item{.data}{A object with class \code{tune_results} where the \code{save_pred = TRUE}
option was used in the control function.}

\item{metrics}{A \code{\link[yardstick:metric_set]{yardstick::metric_set()}}. By default, it uses the same
metrics as the original object.}

\item{eval_time}{A vector of evaluation times for censored regression models.
\code{NULL} is appropriate otherwise. If \code{NULL} is used with censored models, a
evaluation time is selected, and a warning is issued.}

\item{times}{The number of bootstrap samples.}

\item{parameters}{An optional tibble of tuning parameter values that can be
used to filter the predicted values before processing. This tibble should
only have columns for each tuning parameter identifier (e.g. \code{"my_param"}
if \code{tune("my_param")} was used).}

\item{alpha}{Level of significance.}

\item{allow_par}{A logical to allow parallel processing (if a parallel
backend is registered).}

\item{event_level}{A single string. Either \code{"first"} or \code{"second"} to specify
which level of truth to consider as the "event".}

\item{keep_replicates}{A logic for saving the individual estimates from each
bootstrap sample (as a list column called \code{.values}).}

\item{...}{Not currently used.}
}
\value{
A tibble of metrics with additional columns for \code{.lower} and
\code{.upper} (and potentially, \code{.values}).
}
\description{
Using out-of-sample predictions, the bootstrap is used to create percentile
confidence intervals.
}
\details{
For each model configuration (if any), this function takes bootstrap samples
of the out-of-sample predicted values. For each bootstrap sample, the metrics
are computed and these are used to compute confidence intervals.
See \code{\link[rsample:int_pctl]{rsample::int_pctl()}} and the references therein for more details.

Note that the \code{.estimate} column is likely to be different from the results
given by \code{\link[=collect_metrics]{collect_metrics()}} since a different estimator is used. Since
random numbers are used in sampling, set the random number seed prior to
running this function.

The number of bootstrap samples should be large to have reliable intervals.
The defaults reflect the fewest samples that should be used.

The computations for each configuration can be extensive. To increase
computational efficiency parallel processing can be used. The \pkg{future}
package is used here. To execute the resampling iterations in parallel,
specify a \link[future:plan]{plan} with future first. The \code{allow_par} argument
can be used to avoid parallelism.

Also, if a censored regression model used numerous evaluation times, the
computations can take a long time unless the times are filtered with the
\code{eval_time} argument.
}
\examples{
\dontshow{if (!tune:::is_cran_check()) withAutoprint(\{ # examplesIf}
if (rlang::is_installed("modeldata")) {
  data(Sacramento, package = "modeldata")
  library(rsample)
  library(parsnip)

  set.seed(13)
  sac_rs <- vfold_cv(Sacramento)

  lm_res <-
    linear_reg() |>
    fit_resamples(
      log10(price) ~ beds + baths + sqft + type + latitude + longitude,
      resamples = sac_rs,
      control = control_resamples(save_pred = TRUE)
    )

  set.seed(31)
  int_pctl(lm_res)
}
\dontshow{\}) # examplesIf}
}
\references{
Davison, A., & Hinkley, D. (1997). \emph{Bootstrap Methods and their
Application}. Cambridge: Cambridge University Press.
doi:10.1017/CBO9780511802843
}
\seealso{
\code{\link[rsample:int_pctl]{rsample::int_pctl()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_metrics.R
\name{compute_metrics}
\alias{compute_metrics}
\alias{compute_metrics.default}
\alias{compute_metrics.tune_results}
\title{Calculate and format metrics from tuning functions}
\usage{
compute_metrics(x, metrics, summarize, event_level, ...)

\method{compute_metrics}{default}(x, metrics, summarize = TRUE, event_level = "first", ...)

\method{compute_metrics}{tune_results}(x, metrics, ..., summarize = TRUE, event_level = "first")
}
\arguments{
\item{x}{The results of a tuning function like \code{\link[=tune_grid]{tune_grid()}} or
\code{\link[=fit_resamples]{fit_resamples()}}, generated with the control option \code{save_pred = TRUE}.}

\item{metrics}{A \link[yardstick:metric_set]{metric set} of new metrics
to compute. See the "Details" section below for more information.}

\item{summarize}{A single logical value indicating whether metrics should
be summarized over resamples (\code{TRUE}) or return the values for each
individual resample. See \code{\link[=collect_metrics]{collect_metrics()}} for more details on how
metrics are summarized.}

\item{event_level}{A single string containing either \code{"first"} or \code{"second"}.
This argument is passed on to yardstick metric functions when any type
of class prediction is made, and specifies which level of the outcome
is considered the "event".}

\item{...}{Not currently used.}
}
\value{
A tibble. See \code{\link[=collect_metrics]{collect_metrics()}} for more details on the return value.
}
\description{
This function computes metrics from tuning results. The arguments and
output formats are closely related to those from \code{\link[=collect_metrics]{collect_metrics()}}, but
this function additionally takes a \code{metrics} argument with a
\link[yardstick:metric_set]{metric set} for new metrics to compute. This
allows for computing new performance metrics without requiring users to
re-evaluate models against resamples.

Note that the \link[=control_grid]{control option} \code{save_pred = TRUE} must
have been supplied when generating \code{x}.
}
\details{
Each metric in the set supplied to the \code{metrics} argument must have a metric
type (usually \code{"numeric"}, \code{"class"}, or \code{"prob"}) that matches some metric
evaluated when generating \code{x}. e.g. For example, if \code{x} was generated with
only hard \code{"class"} metrics, this function can't compute metrics that take in
class probabilities (\code{"prob"}.) By default, the tuning functions used to
generate \code{x} compute metrics of all needed types.
}
\examples{
\dontshow{if (tune:::should_run_examples()) withAutoprint(\{ # examplesIf}
# load needed packages:
library(parsnip)
library(rsample)
library(yardstick)

# evaluate a linear regression against resamples.
# note that we pass `save_pred = TRUE`:
res <-
  fit_resamples(
    linear_reg(),
    mpg ~ cyl + hp,
    bootstraps(mtcars, 5),
    control = control_grid(save_pred = TRUE)
  )

# to return the metrics supplied to `fit_resamples()`:
collect_metrics(res)

# to compute new metrics:
compute_metrics(res, metric_set(mae))

# if `metrics` is the same as that passed to `fit_resamples()`,
# then `collect_metrics()` and `compute_metrics()` give the same
# output, though `compute_metrics()` is quite a bit slower:
all.equal(
  collect_metrics(res),
  compute_metrics(res, metric_set(rmse, rsq))
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logging.R
\name{message_wrap}
\alias{message_wrap}
\title{Write a message that respects the line width}
\usage{
message_wrap(
  x,
  width = options()$width - 2,
  prefix = "",
  color_text = NULL,
  color_prefix = color_text
)
}
\arguments{
\item{x}{A character string of the message text.}

\item{width}{An integer for the width.}

\item{prefix}{An optional string to go on the first line of the message.}

\item{color_text, color_prefix}{A function (or \code{NULL}) that is used to color
the text and/or prefix.}
}
\value{
The processed text is returned (invisibly) but a message is written.
}
\description{
Write a message that respects the line width
}
\examples{
library(cli)
Gaiman <-
  paste(
    '"Good point." Bod was pleased with himself, and glad he had thought of',
    "asking the poet for advice. Really, he thought, if you couldn't trust a",
    "poet to offer sensible advice, who could you trust?",
    collapse = ""
  )
message_wrap(Gaiman)
message_wrap(Gaiman, width = 20, prefix = "-")
message_wrap(Gaiman,
  width = 30, prefix = "-",
  color_text = cli::col_silver
)
message_wrap(Gaiman,
  width = 30, prefix = "-",
  color_text = cli::style_underline,
  color_prefix = cli::col_green
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{extract-tune}
\alias{extract-tune}
\alias{extract_workflow.last_fit}
\alias{extract_workflow.tune_results}
\alias{extract_spec_parsnip.tune_results}
\alias{extract_recipe.tune_results}
\alias{extract_fit_parsnip.tune_results}
\alias{extract_fit_engine.tune_results}
\alias{extract_mold.tune_results}
\alias{extract_preprocessor.tune_results}
\title{Extract elements of \code{tune} objects}
\usage{
\method{extract_workflow}{last_fit}(x, ...)

\method{extract_workflow}{tune_results}(x, ...)

\method{extract_spec_parsnip}{tune_results}(x, ...)

\method{extract_recipe}{tune_results}(x, ..., estimated = TRUE)

\method{extract_fit_parsnip}{tune_results}(x, ...)

\method{extract_fit_engine}{tune_results}(x, ...)

\method{extract_mold}{tune_results}(x, ...)

\method{extract_preprocessor}{tune_results}(x, ...)
}
\arguments{
\item{x}{A \code{tune_results} object.}

\item{...}{Not currently used.}

\item{estimated}{A logical for whether the original (unfit) recipe or the
fitted recipe should be returned.}
}
\value{
The extracted value from the \code{tune} tune_results, \code{x}, as described in the
description section.
}
\description{
These functions extract various elements from a tune object. If they do
not exist yet, an error is thrown.
\itemize{
\item \code{\link[=extract_preprocessor.tune_results]{extract_preprocessor()}} returns
the formula, recipe, or variable
expressions used for preprocessing.
\item \code{\link[=extract_spec_parsnip.tune_results]{extract_spec_parsnip()}} returns
the parsnip model specification.
\item \code{\link[=extract_fit_parsnip.tune_results]{extract_fit_parsnip()}} returns the
parsnip model fit object.
\item \code{\link[=extract_fit_engine.tune_results]{extract_fit_engine()}} returns the
engine specific fit embedded within
a parsnip model fit. For example, when using \code{\link[parsnip:linear_reg]{parsnip::linear_reg()}}
with the \code{"lm"} engine, this returns the underlying \code{lm} object.
\item \code{\link[=extract_mold.tune_results]{extract_mold()}} returns the preprocessed
"mold" object returned
from \code{\link[hardhat:mold]{hardhat::mold()}}. It contains information about the preprocessing,
including either the prepped recipe, the formula terms object, or
variable selectors.
\item \code{\link[=extract_recipe.tune_results]{extract_recipe()}} returns the recipe.
The \code{estimated} argument specifies
whether the fitted or original recipe is returned.
\item \code{\link[=extract_workflow.tune_results]{extract_workflow()}} returns the
workflow object if the control option
\code{save_workflow = TRUE} was used. The workflow will only have been
estimated for objects produced by \code{\link[=last_fit]{last_fit()}}.
}
}
\details{
These functions supersede \code{extract_model()}.
}
\examples{
\dontshow{if (rlang::is_installed("splines2")) withAutoprint(\{ # examplesIf}
# example code

library(recipes)
library(rsample)
library(parsnip)

set.seed(6735)
tr_te_split <- initial_split(mtcars)

spline_rec <- recipe(mpg ~ ., data = mtcars) |>
  step_spline_natural(disp)

lin_mod <- linear_reg() |>
  set_engine("lm")

spline_res <- last_fit(lin_mod, spline_rec, split = tr_te_split)

extract_preprocessor(spline_res)

# The `spec` is the parsnip spec before it has been fit.
# The `fit` is the fitted parsnip model.
extract_spec_parsnip(spline_res)
extract_fit_parsnip(spline_res)
extract_fit_engine(spline_res)

# The mold is returned from `hardhat::mold()`, and contains the
# predictors, outcomes, and information about the preprocessing
# for use on new data at `predict()` time.
extract_mold(spline_res)

# A useful shortcut is to extract the fitted recipe from the workflow
extract_recipe(spline_res)

# That is identical to
identical(
  extract_mold(spline_res)$blueprint$recipe,
  extract_recipe(spline_res)
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel.R
\name{has_non_par_pkgs}
\alias{has_non_par_pkgs}
\alias{future_installed}
\alias{mirai_installed}
\alias{get_future_workers}
\alias{get_mirai_workers}
\alias{choose_framework}
\alias{get_parallel_seeds}
\alias{eval_mirai}
\alias{.par_fns}
\title{Internal functions to help use parallel processing}
\usage{
has_non_par_pkgs(object, control, verbose = FALSE)

future_installed()

mirai_installed()

get_future_workers(verbose)

get_mirai_workers(verbose)

choose_framework(
  object = NULL,
  control = NULL,
  verbose = FALSE,
  default = "mirai"
)

get_parallel_seeds(workers)

eval_mirai(.x, .f, ..., .args)

.par_fns(framework)
}
\arguments{
\item{object}{A workflow.}

\item{control}{A control object}

\item{verbose}{A logical for printing}

\item{default}{The default parallel processor.}

\item{workers}{The number of existing workers}

\item{.x}{A list.}

\item{.f}{A function}

\item{..., .args}{Options to pass to other functions.}
}
\description{
Internal functions to help use parallel processing
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_ns.R
\name{load_pkgs}
\alias{load_pkgs}
\title{Quietly load package namespace}
\usage{
load_pkgs(x, ..., infra = TRUE)
}
\arguments{
\item{x}{A character vector of packages.}

\item{infra}{Should base tidymodels packages be loaded as well?}
}
\value{
An invisible NULL.
}
\description{
For one or more packages, load the namespace. This is used during parallel
processing since the different parallel backends handle the package
environments differently.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R, R/grid_performance.R,
%   R/load_ns.R, R/logging.R
\name{forge_from_workflow}
\alias{forge_from_workflow}
\alias{finalize_workflow_preprocessor}
\alias{tune-internal-functions}
\alias{.estimate_metrics}
\alias{.load_namespace}
\alias{initialize_catalog}
\alias{.catch_and_log}
\title{Internal functions used by other tidymodels packages}
\usage{
forge_from_workflow(new_data, workflow)

finalize_workflow_preprocessor(workflow, grid_preprocessor)

.estimate_metrics(
  dat,
  metric,
  param_names,
  outcome_name,
  event_level,
  metrics_info = metrics_info(metrics)
)

.load_namespace(x)

initialize_catalog(control, env = rlang::caller_env(), workflow = NULL)

.catch_and_log(.expr, ..., bad_only = FALSE, notes, catalog = TRUE)
}
\arguments{
\item{new_data}{A data frame or matrix of predictors to process.}

\item{workflow}{A workflow.}

\item{grid_preprocessor}{A tibble with parameter information.}

\item{dat}{A data set.}

\item{metric}{A metric set.}

\item{param_names}{A character vector of tuning parameter names.}

\item{outcome_name}{A character string for the column of \code{dat} that is the
outcome.}

\item{event_level}{A logical passed from the control function.}

\item{metrics_info}{The output of \code{tune:::metrics_info(metrics)}---only
included as an argument to allow for pre-computing.}

\item{x}{A character vector of package names.}

\item{.expr}{Code to execute.}

\item{...}{Object to pass to the internal \code{update_printer()} function.}

\item{bad_only}{A logical for whether warnings and errors should be caught.}

\item{notes}{Character data to add to the logging.}

\item{catalog}{A logical passed to \code{update_printer()} giving whether the message
is compatible with the issue cataloger. Defaults to \code{TRUE}. Updates that are
always unique and do not represent a tuning "issue" can bypass the cataloger
by setting \code{catalog = FALSE}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{finalize_workflow_preprocessor()} is deprecated.

These are not to be meant to be invoked directly by users.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.get_tune_parameters}
\alias{.get_tune_parameters}
\alias{.get_tune_parameter_names}
\alias{.get_extra_col_names}
\alias{.get_tune_metrics}
\alias{.get_tune_metric_names}
\alias{.get_tune_eval_times}
\alias{.get_tune_eval_time_target}
\alias{.get_tune_outcome_names}
\alias{.get_tune_workflow}
\alias{.get_fingerprint.tune_results}
\title{Various accessor functions}
\usage{
.get_tune_parameters(x)

.get_tune_parameter_names(x)

.get_extra_col_names(x)

.get_tune_metrics(x)

.get_tune_metric_names(x)

.get_tune_eval_times(x)

.get_tune_eval_time_target(x)

.get_tune_outcome_names(x)

.get_tune_workflow(x)

\method{.get_fingerprint}{tune_results}(x, ...)
}
\arguments{
\item{x}{An object of class \code{tune_result}.}
}
\value{
\itemize{
\item \code{.get_tune_parameters()} returns a \code{dials} \code{parameter} object or a tibble.
\item \code{.get_tune_parameter_names()}, \code{.get_tune_metric_names()}, and
\code{.get_tune_outcome_names()} return a character string.
\item \code{.get_tune_metrics()} returns a metric set or NULL.
\item \code{.get_tune_workflow()} returns the workflow used to fit the
resamples (if \code{save_workflow} was set to \code{TRUE} during fitting) or NULL.
}
}
\description{
These functions return different attributes from objects with class
\code{tune_result}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outcome-names.R
\name{outcome_names}
\alias{outcome_names}
\alias{outcome_names.terms}
\alias{outcome_names.formula}
\alias{outcome_names.recipe}
\alias{outcome_names.workflow}
\alias{outcome_names.tune_results}
\alias{outcome_names.workflow_variables}
\title{Determine names of the outcome data in a workflow}
\usage{
outcome_names(x, ...)

\method{outcome_names}{terms}(x, ...)

\method{outcome_names}{formula}(x, ...)

\method{outcome_names}{recipe}(x, ...)

\method{outcome_names}{workflow}(x, ...)

\method{outcome_names}{tune_results}(x, ...)

\method{outcome_names}{workflow_variables}(x, data = NULL, ...)
}
\arguments{
\item{x}{An object.}

\item{...}{Not used.}
}
\value{
A character string of variable names
}
\description{
Determine names of the outcome data in a workflow
}
\examples{
library(dplyr)
lm(cbind(mpg, wt) ~ ., data = mtcars) |>
  purrr::pluck(terms) |>
  outcome_names()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{autoplot.tune_results}
\alias{autoplot.tune_results}
\title{Plot tuning search results}
\usage{
\method{autoplot}{tune_results}(
  object,
  type = c("marginals", "parameters", "performance"),
  metric = NULL,
  eval_time = NULL,
  width = NULL,
  call = rlang::current_env(),
  ...
)
}
\arguments{
\item{object}{A tibble of results from \code{\link[=tune_grid]{tune_grid()}} or \code{\link[=tune_bayes]{tune_bayes()}}.}

\item{type}{A single character value. Choices are \code{"marginals"} (for a plot
of each predictor versus performance; see Details below), \code{"parameters"}
(each parameter versus search iteration), or \code{"performance"} (performance
versus iteration). The latter two choices are only used for \code{\link[=tune_bayes]{tune_bayes()}}.}

\item{metric}{A character vector or \code{NULL} for which metric to plot. By
default, all metrics will be shown via facets. Possible options are
the entries in \code{.metric} column of \code{collect_metrics(object)}.}

\item{eval_time}{A numeric vector of time points where dynamic event time
metrics should be chosen (e.g. the time-dependent ROC curve, etc). The
values should be consistent with the values used to create \code{object}.}

\item{width}{A number for the width of the confidence interval bars when
\code{type = "performance"}. A value of zero prevents them from being shown.}

\item{call}{The call to be displayed in warnings or errors.}

\item{...}{For plots with a regular grid, this is passed to \code{format()} and is
applied to a parameter used to color points. Otherwise, it is not used.}
}
\value{
A \code{ggplot2} object.
}
\description{
Plot tuning search results
}
\details{
When the results of \code{tune_grid()} are used with \code{autoplot()}, it tries to
determine whether a \emph{regular grid} was used.
\subsection{Regular grids}{

For regular grids with one or more numeric tuning parameters, the parameter
with the most unique values is used on the x-axis. If there are categorical
parameters, the first is used to color the geometries. All other parameters
are used in column faceting.

The plot has the performance metric(s) on the y-axis. If there are multiple
metrics, these are row-faceted.

If there are more than five tuning parameters, the "marginal effects" plots
are used instead.
}

\subsection{Irregular grids}{

For space-filling or random grids, a \emph{marginal} effect plot is created. A
panel is made for each numeric parameter so that each parameter is on the
x-axis and performance is on the y-xis. If there are multiple metrics, these
are row-faceted.

A single categorical parameter is shown as colors. If there are two or more
non-numeric parameters, an error is given. A similar result occurs is only
non-numeric parameters are in the grid. In these cases, we suggest using
\code{collect_metrics()} and \code{ggplot()} to create a plot that is appropriate for
the data.

If a parameter has an associated transformation associated with it (as
determined by the parameter object used to create it), the plot shows the
values in the transformed units (and is labeled with the transformation type).

Parameters are labeled using the labels found in the parameter object
\emph{except} when an identifier was used (e.g. \code{neighbors = tune("K")}).
}
}
\examples{
\dontshow{if (tune:::should_run_examples()) withAutoprint(\{ # examplesIf}
# For grid search:
data("example_ames_knn")

# Plot the tuning parameter values versus performance
autoplot(ames_grid_search, metric = "rmse")


# For iterative search:
# Plot the tuning parameter values versus performance
autoplot(ames_iter_search, metric = "rmse", type = "marginals")

# Plot tuning parameters versus iterations
autoplot(ames_iter_search, metric = "rmse", type = "parameters")

# Plot performance over iterations
autoplot(ames_iter_search, metric = "rmse", type = "performance")
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=tune_grid]{tune_grid()}}, \code{\link[=tune_bayes]{tune_bayes()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/last_fit.R
\name{last_fit}
\alias{last_fit}
\alias{last_fit.model_spec}
\alias{last_fit.workflow}
\title{Fit the final best model to the training set and evaluate the test set}
\usage{
last_fit(object, ...)

\method{last_fit}{model_spec}(
  object,
  preprocessor,
  split,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_last_fit(),
  add_validation_set = FALSE
)

\method{last_fit}{workflow}(
  object,
  split,
  ...,
  metrics = NULL,
  eval_time = NULL,
  control = control_last_fit(),
  add_validation_set = FALSE
)
}
\arguments{
\item{object}{A \code{parsnip} model specification or an unfitted
\link[workflows:workflow]{workflow()}. No tuning parameters are allowed; if arguments
have been marked with \link[hardhat:tune]{tune()}, their values must be
\link[=finalize_model]{finalized}.}

\item{...}{Currently unused.}

\item{preprocessor}{A traditional model formula or a recipe created using
\code{\link[recipes:recipe]{recipes::recipe()}}.}

\item{split}{An \code{rsplit} object created from \code{\link[rsample:initial_split]{rsample::initial_split()}} or
\code{\link[rsample:initial_validation_split]{rsample::initial_validation_split()}}.}

\item{metrics}{A \code{\link[yardstick:metric_set]{yardstick::metric_set()}}, or \code{NULL} to compute a standard
set of metrics.}

\item{eval_time}{A numeric vector of time points where dynamic event time
metrics should be computed (e.g. the time-dependent ROC curve, etc). The
values must be non-negative and should probably be no greater than the
largest event time in the training set (See Details below).}

\item{control}{A \code{\link[=control_last_fit]{control_last_fit()}} object used to fine tune the last fit
process.}

\item{add_validation_set}{For 3-way splits into training, validation, and test
set via \code{\link[rsample:initial_validation_split]{rsample::initial_validation_split()}}, should the validation set be
included in the data set used to train the model. If not, only the training
set is used.}
}
\value{
A single row tibble that emulates the structure of \code{fit_resamples()}.
However, a list column called \code{.workflow} is also attached with the fitted
model (and recipe, if any) that used the training set. Helper functions
for formatting tuning results like \code{\link[=collect_metrics]{collect_metrics()}} and
\code{\link[=collect_predictions]{collect_predictions()}} can be used with \code{last_fit()} output.
}
\description{
\code{\link[=last_fit]{last_fit()}} emulates the process where, after determining the best model,
the final fit on the entire training set is needed and is then evaluated on
the test set.
}
\details{
This function is intended to be used after fitting a \emph{variety of models}
and the final tuning parameters (if any) have been finalized. The next step
would be to fit using the entire training set and verify performance using
the test data.
}
\section{Case Weights}{

Some models can utilize case weights during training. tidymodels currently
supports two types of case weights: importance weights (doubles) and
frequency weights (integers). Frequency weights are used during model
fitting and evaluation, whereas importance weights are only used during
fitting.

To know if your model is capable of using case weights, create a model spec
and test it using \code{\link[parsnip:case_weights_allowed]{parsnip::case_weights_allowed()}}.

To use them, you will need a numeric column in your data set that has been
passed through either \code{\link[hardhat:importance_weights]{hardhat:: importance_weights()}} or
\code{\link[hardhat:frequency_weights]{hardhat::frequency_weights()}}.

For functions such as \code{\link[=fit_resamples]{fit_resamples()}} and the \verb{tune_*()} functions, the
model must be contained inside of a \code{\link[workflows:workflow]{workflows::workflow()}}. To declare that
case weights are used, invoke \code{\link[workflows:add_case_weights]{workflows::add_case_weights()}} with the
corresponding (unquoted) column name.

From there, the packages will appropriately handle the weights during model
fitting and (if appropriate) performance estimation.
}

\section{Censored Regression Models}{


Three types of metrics can be used to assess the quality of censored
regression models:
\itemize{
\item static: the prediction is independent of time.
\item dynamic: the prediction is a time-specific probability (e.g., survival
probability) and is measured at one or more particular times.
\item integrated: same as the dynamic metric but returns the integral of the
different metrics from each time point.
}

Which metrics are chosen by the user affects how many evaluation times
should be specified. For example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# Needs no `eval_time` value
metric_set(concordance_survival)

# Needs at least one `eval_time`
metric_set(brier_survival)
metric_set(brier_survival, concordance_survival)

# Needs at least two eval_time` values
metric_set(brier_survival_integrated, concordance_survival)
metric_set(brier_survival_integrated, concordance_survival)
metric_set(brier_survival_integrated, concordance_survival, brier_survival)
}\if{html}{\out{</div>}}

Values of \code{eval_time} should be less than the largest observed event
time in the training data. For many non-parametric models, the results beyond
the largest time corresponding to an event are constant (or \code{NA}).
}

\section{See also}{


\code{\link[=last_fit]{last_fit()}} is closely related to \code{\link[=fit_best]{fit_best()}}. They both
give you access to a workflow fitted on the training data but are situated
somewhat differently in the modeling workflow. \code{\link[=fit_best]{fit_best()}} picks up
after a tuning function like \code{\link[=tune_grid]{tune_grid()}} to take you from tuning results
to fitted workflow, ready for you to predict and assess further. \code{\link[=last_fit]{last_fit()}}
assumes you have made your choice of hyperparameters and finalized your
workflow to then take you from finalized workflow to fitted workflow and
further to performance assessment on the test data. While \code{\link[=fit_best]{fit_best()}} gives
a fitted workflow, \code{\link[=last_fit]{last_fit()}} gives you the performance results. If you
want the fitted workflow, you can extract it from the result of \code{\link[=last_fit]{last_fit()}}
via \link[=extract_workflow.tune_results]{extract_workflow()}.
}

\examples{
\dontshow{if (tune:::should_run_examples() & rlang::is_installed("splines2")) withAutoprint(\{ # examplesIf}
library(recipes)
library(rsample)
library(parsnip)

set.seed(6735)
tr_te_split <- initial_split(mtcars)

spline_rec <- recipe(mpg ~ ., data = mtcars) |>
  step_spline_natural(disp)

lin_mod <- linear_reg() |>
  set_engine("lm")

spline_res <- last_fit(lin_mod, spline_rec, split = tr_te_split)
spline_res

# test set metrics
collect_metrics(spline_res)

# test set predictions
collect_predictions(spline_res)

# or use a workflow

library(workflows)
spline_wfl <-
  workflow() |>
  add_recipe(spline_rec) |>
  add_model(lin_mod)

last_fit(spline_wfl, split = tr_te_split)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord_obs_pred.R
\name{coord_obs_pred}
\alias{coord_obs_pred}
\title{Use same scale for plots of observed vs predicted values}
\usage{
coord_obs_pred(ratio = 1, xlim = NULL, ylim = NULL, expand = TRUE, clip = "on")
}
\arguments{
\item{ratio}{Aspect ratio, expressed as \code{y / x}. Defaults to 1.0.}

\item{xlim, ylim}{Limits for the x and y axes.}

\item{expand}{Not currently used.}

\item{clip}{Should drawing be clipped to the extent of the plot panel? A setting
of "on" (the default) means yes, and a setting of "off" means no. In most
cases, the default of "on" should not be changed, as setting \code{clip = "off"}
can cause unexpected results. It allows drawing of data points anywhere on
the plot, including in the plot margins. If limits are set via \code{xlim} and
\code{ylim} and some data points fall outside those limits, then those data points
may show up in places such as the axes, the legend, the plot title, or the
plot margins.}
}
\value{
A \code{ggproto} object.
}
\description{
For regression models, \code{coord_obs_pred()} can be used in a ggplot to make the
x- and y-axes have the same exact scale along with an aspect ratio of one.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
# example code
data(solubility_test, package = "modeldata")

library(ggplot2)
p <- ggplot(solubility_test, aes(x = solubility, y = prediction)) +
  geom_abline(lty = 2) +
  geom_point(alpha = 0.5)

p

p + coord_fixed()

p + coord_obs_pred()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{get_metric_time}
\alias{get_metric_time}
\title{Get time for analysis of dynamic survival metrics}
\usage{
get_metric_time(metrics, eval_time)
}
\arguments{
\item{metrics}{A metric set.}

\item{eval_time}{A vector of evaluation times.}
}
\description{
Get time for analysis of dynamic survival metrics
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/finalize.R
\name{finalize_model}
\alias{finalize_model}
\alias{finalize_recipe}
\alias{finalize_workflow}
\alias{finalize_tailor}
\title{Splice final parameters into objects}
\usage{
finalize_model(x, parameters)

finalize_recipe(x, parameters)

finalize_workflow(x, parameters)

finalize_tailor(x, parameters)
}
\arguments{
\item{x}{A recipe, \pkg{parsnip} model specification, \pkg{tailor}
postprocessor, or workflow.}

\item{parameters}{A list or 1-row tibble of parameter values. Note that the
column names of the tibble should be the \code{id} fields attached to \code{tune()}.
For example, in the \code{Examples} section below, the model has \code{tune("K")}. In
this case, the parameter tibble should be "K" and not "neighbors".}
}
\value{
An updated version of \code{x}.
}
\description{
The \verb{finalize_*} functions take a list or tibble of tuning parameter values and
update objects with those values.
}
\examples{
\dontshow{if (tune:::should_run_examples(suggests = "kknn")) withAutoprint(\{ # examplesIf}
data("example_ames_knn")

library(parsnip)
knn_model <-
  nearest_neighbor(
    mode = "regression",
    neighbors = tune("K"),
    weight_func = tune(),
    dist_power = tune()
  ) |>
  set_engine("kknn")

lowest_rmse <- select_best(ames_grid_search, metric = "rmse")
lowest_rmse

knn_model
finalize_model(knn_model, lowest_rmse)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/case_weights.R
\name{.use_case_weights_with_yardstick}
\alias{.use_case_weights_with_yardstick}
\alias{.use_case_weights_with_yardstick.hardhat_importance_weights}
\alias{.use_case_weights_with_yardstick.hardhat_frequency_weights}
\title{Determine if case weights should be passed on to yardstick}
\usage{
.use_case_weights_with_yardstick(x)

\method{.use_case_weights_with_yardstick}{hardhat_importance_weights}(x)

\method{.use_case_weights_with_yardstick}{hardhat_frequency_weights}(x)
}
\arguments{
\item{x}{A vector}
}
\value{
A single \code{TRUE} or \code{FALSE}.
}
\description{
This S3 method defines the logic for deciding when a case weight vector
should be passed to yardstick metric functions and used to measure model
performance. The current logic is that frequency weights (i.e.
\code{\link[hardhat:frequency_weights]{hardhat::frequency_weights()}}) are the only situation where this should
occur.
}
\examples{
library(parsnip)
library(dplyr)

frequency_weights(1:10) |>
  .use_case_weights_with_yardstick()

importance_weights(seq(1, 10, by = .1))|>
  .use_case_weights_with_yardstick()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collect.R
\name{collect_predictions}
\alias{collect_predictions}
\alias{collect_predictions.default}
\alias{collect_predictions.tune_results}
\alias{collect_metrics}
\alias{collect_metrics.tune_results}
\alias{collect_notes}
\alias{collect_notes.tune_results}
\alias{collect_extracts}
\alias{collect_extracts.tune_results}
\title{Obtain and format results produced by tuning functions}
\usage{
collect_predictions(x, ...)

\method{collect_predictions}{default}(x, ...)

\method{collect_predictions}{tune_results}(x, ..., summarize = FALSE, parameters = NULL)

collect_metrics(x, ...)

\method{collect_metrics}{tune_results}(x, ..., summarize = TRUE, type = c("long", "wide"))

collect_notes(x, ...)

\method{collect_notes}{tune_results}(x, ...)

collect_extracts(x, ...)

\method{collect_extracts}{tune_results}(x, ...)
}
\arguments{
\item{x}{The results of \code{\link[=tune_grid]{tune_grid()}}, \code{\link[=tune_bayes]{tune_bayes()}}, \code{\link[=fit_resamples]{fit_resamples()}},
or \code{\link[=last_fit]{last_fit()}}. For \code{\link[=collect_predictions]{collect_predictions()}}, the control option \code{save_pred = TRUE} should have been used.}

\item{...}{Not currently used.}

\item{summarize}{A logical; should metrics be summarized over resamples
(\code{TRUE}) or return the values for each individual resample. Note that, if \code{x}
is created by \code{\link[=last_fit]{last_fit()}}, \code{summarize} has no effect. For the other object
types, the method of summarizing predictions is detailed below.}

\item{parameters}{An optional tibble of tuning parameter values that can be
used to filter the predicted values before processing. This tibble should
only have columns for each tuning parameter identifier (e.g. \code{"my_param"}
if \code{tune("my_param")} was used).}

\item{type}{One of \code{"long"} (the default) or \code{"wide"}. When \code{type = "long"},
output has columns \code{.metric} and one of \code{.estimate} or \code{mean}.
\code{.estimate}/\code{mean} gives the values for the \code{.metric}. When \code{type = "wide"},
each metric has its own column and the \code{n} and \code{std_err} columns are removed,
if they exist.}
}
\value{
A tibble. The column names depend on the results and the mode of the
model.

For \code{\link[=collect_metrics]{collect_metrics()}} and \code{\link[=collect_predictions]{collect_predictions()}}, when unsummarized,
there are columns for each tuning parameter (using the \code{id} from \code{\link[=tune]{tune()}},
if any).

\code{\link[=collect_metrics]{collect_metrics()}} also has columns \code{.metric}, and \code{.estimator} by default.
For \code{\link[=collect_metrics]{collect_metrics()}} methods that have a \code{type} argument, supplying
\code{type = "wide"} will pivot the output such that each metric has its own
column. When the results are summarized, there are columns for \code{mean}, \code{n},
and \code{std_err}. When not summarized, the additional columns for the resampling
identifier(s) and \code{.estimate}.

For \code{\link[=collect_predictions]{collect_predictions()}}, there are additional columns for the resampling
identifier(s), columns for the predicted values (e.g., \code{.pred},
\code{.pred_class}, etc.), and a column for the outcome(s) using the original
column name(s) in the data.

\code{\link[=collect_predictions]{collect_predictions()}} can summarize the various results over
replicate out-of-sample predictions. For example, when using the bootstrap,
each row in the original training set has multiple holdout predictions
(across assessment sets). To convert these results to a format where every
training set same has a single predicted value, the results are averaged
over replicate predictions.

For regression cases, the numeric predictions are simply averaged.

For classification models, the problem is more complex. When class probabilities
are used, these are averaged and then re-normalized to make sure that they
add to one. If hard class predictions also exist in the data, then these are
determined from the summarized probability estimates (so that they match).
If only hard class predictions are in the results, then the mode is used to
summarize.

With censored outcome models, the predicted survival probabilities (if any)
are averaged while the static predicted event times are summarized using the
median.

\code{\link[=collect_notes]{collect_notes()}} returns a tibble with columns for the resampling
indicators, the location (preprocessor, model, etc.), type (error or warning),
and the notes.

\code{\link[=collect_extracts]{collect_extracts()}} collects objects extracted from fitted workflows
via the \code{extract} argument to \link[=control_grid]{control functions}. The
function returns a tibble with columns for the resampling
indicators, the location (preprocessor, model, etc.), and extracted objects.
}
\description{
Obtain and format results produced by tuning functions
}
\section{Hyperparameters and extracted objects}{


When making use of submodels, tune can generate predictions and calculate
metrics for multiple model \code{.config}urations using only one model fit.
However, this means that if a function was supplied to a
\link[=control_grid]{control function's} \code{extract} argument, tune can only
execute that extraction on the one model that was fitted. As a result,
in the \code{collect_extracts()} output, tune opts to associate the
extracted objects with the hyperparameter combination used to
fit that one model workflow, rather than the hyperparameter
combination of a submodel. In the output, this appears like
a hyperparameter entry is recycled across many \code{.config}
entries---this is intentional.

See \url{https://parsnip.tidymodels.org/articles/Submodels.html} to learn
more about submodels.
}

\examples{
\dontshow{if (tune:::should_run_examples(suggests = c("kknn", "splines2"))) withAutoprint(\{ # examplesIf}
data("example_ames_knn")
# The parameters for the model:
extract_parameter_set_dials(ames_wflow)

# Summarized over resamples
collect_metrics(ames_grid_search)

# Per-resample values
collect_metrics(ames_grid_search, summarize = FALSE)


# ---------------------------------------------------------------------------

library(parsnip)
library(rsample)
library(dplyr)
library(recipes)
library(tibble)

lm_mod <- linear_reg() |> set_engine("lm")
set.seed(93599150)
car_folds <- vfold_cv(mtcars, v = 2, repeats = 3)
ctrl <- control_resamples(save_pred = TRUE, extract = extract_fit_engine)

spline_rec <-
  recipe(mpg ~ ., data = mtcars) |>
  step_spline_natural(disp, deg_free = tune("df"))

grid <- tibble(df = 3:6)

resampled <-
  lm_mod |>
  tune_grid(spline_rec, resamples = car_folds, control = ctrl, grid = grid)

collect_predictions(resampled) |> arrange(.row)
collect_predictions(resampled, summarize = TRUE) |> arrange(.row)
collect_predictions(
  resampled,
  summarize = TRUE,
  parameters = grid[1, ]
) |> arrange(.row)

collect_extracts(resampled)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acquisition.R
\name{prob_improve}
\alias{prob_improve}
\alias{exp_improve}
\alias{conf_bound}
\title{Acquisition function for scoring parameter combinations}
\usage{
prob_improve(trade_off = 0, eps = .Machine$double.eps)

exp_improve(trade_off = 0, eps = .Machine$double.eps)

conf_bound(kappa = 0.1)
}
\arguments{
\item{trade_off}{A number or function that describes the trade-off between
exploitation and exploration. Smaller values favor exploitation.}

\item{eps}{A small constant to avoid division by zero.}

\item{kappa}{A positive number (or function) that corresponds to the
multiplier of the standard deviation in a confidence bound (e.g. 1.96 in
normal-theory 95 percent confidence intervals). Smaller values lean more
towards exploitation.}
}
\value{
An object of class \code{prob_improve}, \code{exp_improve}, or \code{conf_bounds}
along with an extra class of \code{acquisition_function}.
}
\description{
These functions can be used to score candidate tuning parameter combinations
as a function of their predicted mean and variation.
}
\details{
The acquisition functions often combine the mean and variance
predictions from the Gaussian process model into an objective to be
optimized.

For this documentation, we assume that the metric in question is better when
\emph{maximized} (e.g. accuracy, the coefficient of determination, etc).

The expected improvement of a point \code{x} is based on the predicted mean and
variation at that point as well as the current best value (denoted here as
\code{x_b}). The vignette linked below contains the formulas for this acquisition
function. When the \code{trade_off} parameter is greater than zero, the
acquisition function will down-play the effect of the \emph{mean} prediction and
give more weight to the variation. This has the effect of searching for new
parameter combinations that are in areas that have yet to be sampled.

Note that for \code{\link[=exp_improve]{exp_improve()}} and \code{\link[=prob_improve]{prob_improve()}}, the \code{trade_off} value is
in the units of the outcome. The functions are parameterized so that the
\code{trade_off} value should always be non-negative.

The confidence bound function does not take into account the current best
results in the data.

If a function is passed to  \code{\link[=exp_improve]{exp_improve()}} or \code{\link[=prob_improve]{prob_improve()}}, the function
can have multiple arguments but only the first (the current iteration number)
is given to the function. In other words, the function argument should have
defaults for all but the first argument. See \code{\link[=expo_decay]{expo_decay()}} as an example of
a function.
}
\examples{
prob_improve()
}
\seealso{
\code{\link[=tune_bayes]{tune_bayes()}}, \code{\link[=expo_decay]{expo_decay()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_resample_weights}
\alias{check_resample_weights}
\title{Check fold weights in rset objects}
\usage{
check_resample_weights(x)
}
\arguments{
\item{x}{An rset object.}
}
\value{
\code{NULL} invisibly, or error if weights are invalid.
}
\description{
Check fold weights in rset objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_grid.R
\name{tune_grid}
\alias{tune_grid}
\alias{tune_grid.model_spec}
\alias{tune_grid.workflow}
\title{Model tuning via grid search}
\usage{
tune_grid(object, ...)

\method{tune_grid}{model_spec}(
  object,
  preprocessor,
  resamples,
  ...,
  param_info = NULL,
  grid = 10,
  metrics = NULL,
  eval_time = NULL,
  control = control_grid()
)

\method{tune_grid}{workflow}(
  object,
  resamples,
  ...,
  param_info = NULL,
  grid = 10,
  metrics = NULL,
  eval_time = NULL,
  control = control_grid()
)
}
\arguments{
\item{object}{A \code{parsnip} model specification or an unfitted
\link[workflows:workflow]{workflow()}. No tuning parameters are allowed; if arguments
have been marked with \link[hardhat:tune]{tune()}, their values must be
\link[=finalize_model]{finalized}.}

\item{...}{Not currently used.}

\item{preprocessor}{A traditional model formula or a recipe created using
\code{\link[recipes:recipe]{recipes::recipe()}}.}

\item{resamples}{An \code{rset} resampling object created from an \code{rsample}
function, such as \code{\link[rsample:vfold_cv]{rsample::vfold_cv()}}.}

\item{param_info}{A \code{\link[dials:parameters]{dials::parameters()}} object or \code{NULL}. If none is given,
a parameters set is derived from other arguments. Passing this argument can
be useful when parameter ranges need to be customized.}

\item{grid}{A data frame of tuning combinations or a positive integer. The
data frame should have columns for each parameter being tuned and rows for
tuning parameter candidates. An integer denotes the number of candidate
parameter sets to be created automatically.}

\item{metrics}{A \code{\link[yardstick:metric_set]{yardstick::metric_set()}}, or \code{NULL} to compute a standard
set of metrics.}

\item{eval_time}{A numeric vector of time points where dynamic event time
metrics should be computed (e.g. the time-dependent ROC curve, etc). The
values must be non-negative and should probably be no greater than the
largest event time in the training set (See Details below).}

\item{control}{An object used to modify the tuning process, likely created
by \code{\link[=control_grid]{control_grid()}}.}
}
\value{
An updated version of \code{resamples} with extra list columns for \code{.metrics} and
\code{.notes} (optional columns are \code{.predictions} and \code{.extracts}). \code{.notes}
contains warnings and errors that occur during execution.
}
\description{
\code{\link[=tune_grid]{tune_grid()}} computes a set of performance metrics (e.g. accuracy or RMSE)
for a pre-defined set of tuning parameters that correspond to a model or
recipe across one or more resamples of the data.
}
\details{
Suppose there are \emph{m} tuning parameter combinations. \code{\link[=tune_grid]{tune_grid()}} may not
require all \emph{m} model/recipe fits across each resample. For example:

\itemize{
\item In cases where a single model fit can be used to make predictions
for different parameter values in the grid, only one fit is used.
For example, for some boosted trees, if 100 iterations of boosting
are requested, the model object for 100 iterations can be used to
make predictions on iterations less than 100 (if all other
parameters are equal).
\item When the model is being tuned in conjunction with pre-processing
and/or post-processing parameters, the minimum number of fits are
used. For example, if the number of PCA components in a recipe step
are being tuned over three values (along with model tuning
parameters), only three recipes are trained. The alternative
would be to re-train the same recipe multiple times for each model
tuning parameter.
}

tune supports parallel processing with the \pkg{future} package. To execute
the resampling iterations in parallel, specify a \link[future:plan]{plan} with
future first. The \code{allow_par} argument can be used to avoid parallelism.

For the most part, warnings generated during training are shown as they occur
and are associated with a specific resample when
\code{control_grid(verbose = TRUE)}. They are (usually) not aggregated until the
end of processing.
}
\section{Parameter Grids}{


If no tuning grid is provided, a grid (via \code{\link[dials:grid_space_filling]{dials::grid_space_filling()}}) is
created with 10 candidate parameter combinations.

When provided, the grid should have column names for each parameter and
these should be named by the parameter name or \code{id}. For example, if a
parameter is marked for optimization using \code{penalty = tune()}, there should
be a column named \code{penalty}. If the optional identifier is used, such as
\code{penalty = tune(id = 'lambda')}, then the corresponding column name should
be \code{lambda}.

In some cases, the tuning parameter values depend on the dimensions of the
data. For example, \code{mtry} in random forest models depends on the number of
predictors. In this case, the default tuning parameter object requires an
upper range. \code{\link[dials:finalize]{dials::finalize()}} can be used to derive the data-dependent
parameters. Otherwise, a parameter set can be created (via
\code{\link[dials:parameters]{dials::parameters()}}) and the \code{dials} \code{update()} function can be used to
change the values. This updated parameter set can be passed to the function
via the \code{param_info} argument.

The rows of the grid are called tuning parameter \emph{candidates}. Each
candidate has a unique \code{.config} value that, for grid search, has the
pattern \verb{pre\{num\}_mod\{num\}_post\{num\}}. The numbers include a zero when that
element was static. For example, a value of \code{pre0_mod3_post4} means no
preprocessors were tuned and the model and postprocessor(s) had at least
three and four candidates, respectively. Also, the numbers are zero-padded
to enable proper sorting.
}

\section{Performance Metrics}{


To use your own performance metrics, the \code{\link[yardstick:metric_set]{yardstick::metric_set()}} function
can be used to pick what should be measured for each model. If multiple
metrics are desired, they can be bundled. For example, to estimate the area
under the ROC curve as well as the sensitivity and specificity (under the
typical probability cutoff of 0.50), the \code{metrics} argument could be given:

\preformatted{
  metrics = metric_set(roc_auc, sens, spec)
}

Each metric is calculated for each candidate model.

If no metric set is provided, one is created:
\itemize{
\item For regression models, the root mean squared error and coefficient
of determination are computed.
\item For classification, the area under the ROC curve and overall accuracy
are computed.
\item For censored regression, the dynamic Brier score
(\code{\link[yardstick:brier_survival]{yardstick::brier_survival()}}) is used.
\item For quantile regression, the weighted interval score
(\code{\link[yardstick:weighted_interval_score]{yardstick::weighted_interval_score()}}) is used.
}

Note that the metrics also determine what type of predictions are estimated
during tuning. For example, in a classification problem, if metrics are used
that are all associated with hard class predictions, the classification
probabilities are not created.

The out-of-sample estimates of these metrics are contained in a list column
called \code{.metrics}. This tibble contains a row for each metric and columns
for the value, the estimator type, and so on.

\code{\link[=collect_metrics]{collect_metrics()}} can be used for these objects to collapse the results
over the resampled (to obtain the final resampling estimates per tuning
parameter combination).
}

\section{Obtaining Predictions}{


When \code{control_grid(save_pred = TRUE)}, the output tibble contains a list
column called \code{.predictions} that has the out-of-sample predictions for each
parameter combination in the grid and each fold (which can be very large).

The elements of the tibble are tibbles with columns for the tuning
parameters, the row number from the original data object (\code{.row}), the
outcome data (with the same name(s) of the original data), and any columns
created by the predictions. For example, for simple regression problems, this
function generates a column called \code{.pred} and so on. As noted above, the
prediction columns that are returned are determined by the type of metric(s)
requested.

This list column can be \code{unnested} using \code{\link[tidyr:unnest]{tidyr::unnest()}} or using the
convenience function \code{\link[=collect_predictions]{collect_predictions()}}.
}

\section{Extracting Information}{


The \code{extract} control option will result in an additional function to be
returned called \code{.extracts}. This is a list column that has tibbles
containing the results of the user's function for each tuning parameter
combination. This can enable returning each model and/or recipe object that
is created during resampling. Note that this could result in a large return
object, depending on what is returned.

The control function contains an option (\code{extract}) that can be used to
retain any model or recipe that was created within the resamples. This
argument should be a function with a single argument. The value of the
argument that is given to the function in each resample is a workflow
object (see \code{\link[workflows:workflow]{workflows::workflow()}} for more information). Several
helper functions can be used to easily pull out the preprocessing
and/or model information from the workflow, such as
\code{\link[workflows:extract-workflow]{extract_preprocessor()}} and
\code{\link[workflows:extract-workflow]{extract_fit_parsnip()}}.

As an example, if there is interest in getting each parsnip model fit back,
one could use:
\preformatted{
  extract = function (x) extract_fit_parsnip(x)
}

Note that the function given to the \code{extract} argument is evaluated on
every model that is \emph{fit} (as opposed to every model that is \emph{evaluated}).
As noted above, in some cases, model predictions can be derived for
sub-models so that, in these cases, not every row in the tuning parameter
grid has a separate R object associated with it.

Finally, it is a good idea to include calls to \code{\link[=require]{require()}} for packages that
are used in the function. This helps prevent failures when using parallel
processing.
}

\section{Case Weights}{

Some models can utilize case weights during training. tidymodels currently
supports two types of case weights: importance weights (doubles) and
frequency weights (integers). Frequency weights are used during model
fitting and evaluation, whereas importance weights are only used during
fitting.

To know if your model is capable of using case weights, create a model spec
and test it using \code{\link[parsnip:case_weights_allowed]{parsnip::case_weights_allowed()}}.

To use them, you will need a numeric column in your data set that has been
passed through either \code{\link[hardhat:importance_weights]{hardhat:: importance_weights()}} or
\code{\link[hardhat:frequency_weights]{hardhat::frequency_weights()}}.

For functions such as \code{\link[=fit_resamples]{fit_resamples()}} and the \verb{tune_*()} functions, the
model must be contained inside of a \code{\link[workflows:workflow]{workflows::workflow()}}. To declare that
case weights are used, invoke \code{\link[workflows:add_case_weights]{workflows::add_case_weights()}} with the
corresponding (unquoted) column name.

From there, the packages will appropriately handle the weights during model
fitting and (if appropriate) performance estimation.
}

\section{Censored Regression Models}{


Three types of metrics can be used to assess the quality of censored
regression models:
\itemize{
\item static: the prediction is independent of time.
\item dynamic: the prediction is a time-specific probability (e.g., survival
probability) and is measured at one or more particular times.
\item integrated: same as the dynamic metric but returns the integral of the
different metrics from each time point.
}

Which metrics are chosen by the user affects how many evaluation times
should be specified. For example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# Needs no `eval_time` value
metric_set(concordance_survival)

# Needs at least one `eval_time`
metric_set(brier_survival)
metric_set(brier_survival, concordance_survival)

# Needs at least two eval_time` values
metric_set(brier_survival_integrated, concordance_survival)
metric_set(brier_survival_integrated, concordance_survival)
metric_set(brier_survival_integrated, concordance_survival, brier_survival)
}\if{html}{\out{</div>}}

Values of \code{eval_time} should be less than the largest observed event
time in the training data. For many non-parametric models, the results beyond
the largest time corresponding to an event are constant (or \code{NA}).
}

\examples{
\dontshow{if (tune:::should_run_examples(suggests = "kernlab") & rlang::is_installed("splines2") && !tune:::is_cran_check()) withAutoprint(\{ # examplesIf}
library(recipes)
library(rsample)
library(parsnip)
library(workflows)
library(ggplot2)

# ---------------------------------------------------------------------------

set.seed(6735)
folds <- vfold_cv(mtcars, v = 5)

# ---------------------------------------------------------------------------

# tuning recipe parameters:

spline_rec <-
  recipe(mpg ~ ., data = mtcars) |>
  step_spline_natural(disp, deg_free = tune("disp")) |>
  step_spline_natural(wt, deg_free = tune("wt"))

lin_mod <-
  linear_reg() |>
  set_engine("lm")

# manually create a grid
spline_grid <- expand.grid(disp = 2:5, wt = 2:5)

# Warnings will occur from making spline terms on the holdout data that are
# extrapolations.
spline_res <-
  tune_grid(lin_mod, spline_rec, resamples = folds, grid = spline_grid)
spline_res


show_best(spline_res, metric = "rmse")

# ---------------------------------------------------------------------------

# tune model parameters only (example requires the `kernlab` package)

car_rec <-
  recipe(mpg ~ ., data = mtcars) |>
  step_normalize(all_predictors())

svm_mod <-
  svm_rbf(cost = tune(), rbf_sigma = tune()) |>
  set_engine("kernlab") |>
  set_mode("regression")

# Use a space-filling design with 7 points
set.seed(3254)
svm_res <- tune_grid(svm_mod, car_rec, resamples = folds, grid = 7)
svm_res

show_best(svm_res, metric = "rmse")

autoplot(svm_res, metric = "rmse") +
  scale_x_log10()

# ---------------------------------------------------------------------------

# Using a variables preprocessor with a workflow

# Rather than supplying a preprocessor (like a recipe) and a model directly
# to `tune_grid()`, you can also wrap them up in a workflow and pass
# that along instead (note that this doesn't do any preprocessing to
# the variables, it passes them along as-is).
wf <- workflow() |>
  add_variables(outcomes = mpg, predictors = everything()) |>
  add_model(svm_mod)

set.seed(3254)
svm_res_wf <- tune_grid(wf, resamples = folds, grid = 7)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=control_grid]{control_grid()}}, \code{\link[=tune]{tune()}}, \code{\link[=fit_resamples]{fit_resamples()}},
\code{\link[=autoplot.tune_results]{autoplot.tune_results()}}, \code{\link[=show_best]{show_best()}}, \code{\link[=select_best]{select_best()}},
\code{\link[=collect_predictions]{collect_predictions()}}, \code{\link[=collect_metrics]{collect_metrics()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{resample_weights_utils}
\alias{resample_weights_utils}
\alias{.get_resample_weights}
\alias{.create_weight_mapping}
\alias{.weighted_sd}
\alias{.effective_sample_size}
\alias{.validate_resample_weights}
\title{Resampling weights utility functions}
\usage{
.get_resample_weights(x)

.create_weight_mapping(weights, id_names, metrics_data)

.weighted_sd(x, w)

.effective_sample_size(w)

.validate_resample_weights(weights, num_resamples)
}
\arguments{
\item{x}{A tune_results object.}

\item{weights}{Numeric vector of weights.}

\item{id_names}{Character vector of ID column names.}

\item{metrics_data}{The metrics data frame.}

\item{w}{Numeric vector of weights.}

\item{num_resamples}{Integer number of resamples.}
}
\value{
Various return values depending on the function.
}
\description{
These are internal functions for handling variable resampling weights in
hyperparameter tuning.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metric-selection.R
\name{choose_metric}
\alias{choose_metric}
\alias{check_metric_in_tune_results}
\alias{choose_eval_time}
\alias{maybe_choose_eval_time}
\alias{first_metric}
\alias{first_eval_time}
\alias{.filter_perf_metrics}
\alias{check_metrics_arg}
\alias{check_eval_time_arg}
\title{Tools for selecting metrics and evaluation times}
\usage{
choose_metric(x, metric, ..., call = rlang::caller_env())

check_metric_in_tune_results(mtr_info, metric, ..., call = rlang::caller_env())

choose_eval_time(
  x,
  metric,
  ...,
  eval_time = NULL,
  quietly = FALSE,
  call = rlang::caller_env()
)

maybe_choose_eval_time(x, mtr_set, eval_time)

first_metric(mtr_set)

first_eval_time(
  mtr_set,
  ...,
  metric = NULL,
  eval_time = NULL,
  quietly = FALSE,
  call = rlang::caller_env()
)

.filter_perf_metrics(x, metric, eval_time)

check_metrics_arg(mtr_set, wflow, ..., call = rlang::caller_env())

check_eval_time_arg(eval_time, mtr_set, ..., call = rlang::caller_env())
}
\arguments{
\item{x}{An object with class \code{tune_results}.}

\item{metric}{A character value for which metric is being used.}

\item{...}{These dots are for future extensions and must be empty.}

\item{call}{The call to be displayed in warnings or errors.}

\item{eval_time}{An optional vector of times to compute dynamic and/or
integrated metrics.}

\item{quietly}{Logical. Should warnings be muffled?}

\item{mtr_set}{A \code{\link[yardstick:metric_set]{yardstick::metric_set()}}.}

\item{wflow}{A \code{\link[workflows:workflow]{workflows::workflow()}}.}
}
\description{
Tools for selecting metrics and evaluation times
}
\details{
These are developer-facing functions used to compute and validate choices
for performance metrics. For survival analysis models, there are similar
functions for the evaluation time(s) required for dynamic and/or integrated
metrics.

\code{choose_metric()} is used with functions such as \code{\link[=show_best]{show_best()}} or
\code{\link[=select_best]{select_best()}} where a single valid metric is required to rank models. If
no value is given by the user, the first metric value is used (with a
warning).

For evaluation times, one is only required when the metric type is dynamic
(e.g. \code{\link[yardstick:brier_survival]{yardstick::brier_survival()}} or \code{\link[yardstick:roc_auc_survival]{yardstick::roc_auc_survival()}}). For
these metrics, we require a single numeric value that was originally given
to the function used to produce \code{x} (such as \code{\link[=tune_grid]{tune_grid()}}).

If a time is required and none is given, the first value in the vector
originally given in the \code{eval_time} argument is used (with a warning).

\code{maybe_choose_eval_time()} is for cases where multiple evaluation times are
acceptable but you need to choose a good default. The "maybe" is because
the function that would use \code{maybe_choose_eval_time()} can accept multiple
metrics (like \code{\link[=autoplot]{autoplot()}}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control_last_fit}
\alias{control_last_fit}
\title{Control aspects of the last fit process}
\usage{
control_last_fit(verbose = FALSE, event_level = "first", allow_par = FALSE)
}
\arguments{
\item{verbose}{A logical for logging results (other than warnings and errors,
which are always shown) as they are generated during training in a single
R process. When using most parallel backends, this argument typically will
not result in any logging. If using a dark IDE theme, some logging messages
might be hard to see; try setting the \code{tidymodels.dark} option with
\code{options(tidymodels.dark = TRUE)} to print lighter colors.}

\item{event_level}{A single string containing either \code{"first"} or \code{"second"}.
This argument is passed on to yardstick metric functions when any type
of class prediction is made, and specifies which level of the outcome
is considered the "event".}

\item{allow_par}{A logical to allow parallel processing (if a parallel
backend is registered).}
}
\description{
Control aspects of the last fit process
}
\details{
\code{\link[=control_last_fit]{control_last_fit()}} is a wrapper around \code{\link[=control_resamples]{control_resamples()}} and is meant
to be used with \code{\link[=last_fit]{last_fit()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_results.R
\name{show_notes}
\alias{show_notes}
\title{Display distinct errors from tune objects}
\usage{
show_notes(x, n = 10)
}
\arguments{
\item{x}{An object of class \code{tune_results}.}

\item{n}{An integer for how many unique notes to show.}
}
\value{
Invisibly, \code{x}. Function is called for side-effects and printing.
}
\description{
Display distinct errors from tune objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conf_mat_resampled.R
\name{conf_mat_resampled}
\alias{conf_mat_resampled}
\title{Compute average confusion matrix across resamples}
\usage{
conf_mat_resampled(x, ..., parameters = NULL, tidy = TRUE)
}
\arguments{
\item{x}{An object with class \code{tune_results} that was used with a
classification model that was run with \code{control_*(save_pred = TRUE)}.}

\item{...}{Currently unused, must be empty.}

\item{parameters}{A tibble with a single tuning parameter combination. Only
one tuning parameter combination (if any were used) is allowed here.}

\item{tidy}{Should the results come back in a tibble (\code{TRUE}) or a \code{conf_mat}
object like \code{yardstick::conf_mat()} (\code{FALSE})?}
}
\value{
A tibble or \code{conf_mat} with the average cell count across resamples.
}
\description{
For classification problems, \code{conf_mat_resampled()} computes a separate
confusion matrix for each resample then averages the cell counts.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
# example code

library(parsnip)
library(rsample)
library(dplyr)

data(two_class_dat, package = "modeldata")

set.seed(2393)
res <-
  logistic_reg() |>
  set_engine("glm") |>
  fit_resamples(
    Class ~ .,
    resamples = vfold_cv(two_class_dat, v = 3),
    control = control_resamples(save_pred = TRUE)
  )

conf_mat_resampled(res)
conf_mat_resampled(res, tidy = FALSE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{add_resample_weights}
\alias{add_resample_weights}
\title{Add resample weights to an rset object}
\usage{
add_resample_weights(rset, weights)
}
\arguments{
\item{rset}{An rset object from \pkg{rsample}.}

\item{weights}{A numeric vector of weights, one per resample. Weights will be
normalized.}
}
\value{
The rset object with weights added as an attribute.
}
\description{
This function allows you to specify custom weights for resamples. Weights
are automatically normalized to sum to 1.
}
\details{
Resampling weights are useful when assessment sets (i.e., held out data) have
different sizes or when you want to upweight certain resamples in the evaluation.
The weights are stored as an attribute and used automatically during
metric aggregation.
}
\examples{
library(rsample)
folds <- vfold_cv(mtcars, v = 3)
# Give equal weight to all folds
weighted_folds <- add_resample_weights(folds, c(1, 1, 1))
# Emphasize the first fold
weighted_folds <- add_resample_weights(folds, c(0.5, 0.25, 0.25))
}
\seealso{
\code{\link[=calculate_resample_weights]{calculate_resample_weights()}}, \code{\link[=extract_resample_weights]{extract_resample_weights()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compat-dplyr.R, R/reexports.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{dplyr_reconstruct}
\alias{parameters}
\alias{autoplot}
\alias{required_pkgs}
\alias{tune}
\alias{tunable}
\alias{tune_args}
\alias{min_grid}
\alias{augment}
\alias{.get_fingerprint}
\alias{int_pctl}
\alias{extract_spec_parsnip}
\alias{extract_recipe}
\alias{extract_fit_parsnip}
\alias{extract_fit_engine}
\alias{extract_mold}
\alias{extract_preprocessor}
\alias{extract_workflow}
\alias{extract_parameter_set_dials}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{dials}{\code{\link[dials]{parameters}}}

  \item{dplyr}{\code{\link[dplyr:dplyr_extending]{dplyr_reconstruct}}}

  \item{generics}{\code{\link[generics]{augment}}, \code{\link[generics]{min_grid}}, \code{\link[generics]{required_pkgs}}, \code{\link[generics]{tunable}}, \code{\link[generics]{tune_args}}}

  \item{ggplot2}{\code{\link[ggplot2]{autoplot}}}

  \item{hardhat}{\code{\link[hardhat:hardhat-extract]{extract_fit_engine}}, \code{\link[hardhat:hardhat-extract]{extract_fit_parsnip}}, \code{\link[hardhat:hardhat-extract]{extract_mold}}, \code{\link[hardhat:hardhat-extract]{extract_parameter_set_dials}}, \code{\link[hardhat:hardhat-extract]{extract_preprocessor}}, \code{\link[hardhat:hardhat-extract]{extract_recipe}}, \code{\link[hardhat:hardhat-extract]{extract_spec_parsnip}}, \code{\link[hardhat:hardhat-extract]{extract_workflow}}, \code{\link[hardhat]{tune}}}

  \item{rsample}{\code{\link[rsample:get_fingerprint]{.get_fingerprint}}, \code{\link[rsample]{int_pctl}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/augment.R
\name{augment.tune_results}
\alias{augment.tune_results}
\alias{augment.resample_results}
\alias{augment.last_fit}
\title{Augment data with holdout predictions}
\usage{
\method{augment}{tune_results}(x, ..., parameters = NULL)

\method{augment}{resample_results}(x, ...)

\method{augment}{last_fit}(x, ...)
}
\arguments{
\item{x}{An object resulting from one of the \verb{tune_*()} functions,
\code{fit_resamples()}, or \code{last_fit()}. The control specifications for these
objects should have used the option \code{save_pred = TRUE}.}

\item{...}{Not currently used.}

\item{parameters}{A data frame with a single row that indicates what
tuning parameters should be used to generate the predictions (for \verb{tune_*()}
objects only). If \code{NULL}, \code{select_best(x)} will be used with the first
metric and, if applicable, the first evaluation time point, used to
create \code{x}.}
}
\value{
A data frame with one or more additional columns for model
predictions.
}
\description{
For \code{tune} objects that use resampling, these \code{augment()} methods will add
one or more columns for the hold-out predictions (i.e. from the assessment
set(s)).
}
\details{
For some resampling methods where rows may be replicated in multiple
assessment sets, the prediction columns will be averages of the holdout
results. Also, for these methods, it is possible that all rows of the
original data do not have holdout predictions (like a single bootstrap
resample). In this case, all rows are return and a warning is issued.

For objects created by \code{last_fit()}, the test set data and predictions are
returned.

Unlike other \code{augment()} methods, the predicted values for regression models
are in a column called \code{.pred} instead of \code{.fitted} (to be consistent with
other tidymodels conventions).

For regression problems, an additional \code{.resid} column is added to the
results.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_best.R
\name{fit_best}
\alias{fit_best}
\alias{fit_best.default}
\alias{fit_best.tune_results}
\title{Fit a model to the numerically optimal configuration}
\usage{
fit_best(x, ...)

\method{fit_best}{default}(x, ...)

\method{fit_best}{tune_results}(
  x,
  ...,
  metric = NULL,
  eval_time = NULL,
  parameters = NULL,
  verbose = FALSE,
  add_validation_set = NULL
)
}
\arguments{
\item{x}{The results of class \code{tune_results} (coming from functions such as
\code{\link[=tune_grid]{tune_grid()}}, \code{\link[=tune_bayes]{tune_bayes()}}, etc). The control option
\code{\link[=control_grid]{save_workflow = TRUE}} should have been used.}

\item{...}{Not currently used, must be empty.}

\item{metric}{A character string (or \code{NULL}) for which metric to optimize. If
\code{NULL}, the first metric is used.}

\item{eval_time}{A single numeric time point where dynamic event time
metrics should be chosen (e.g., the time-dependent ROC curve, etc). The
values should be consistent with the values used to create \code{x}. The \code{NULL}
default will automatically use the first evaluation time used by \code{x}.}

\item{parameters}{An optional 1-row tibble of tuning parameter settings, with
a column for each tuning parameter. This tibble should have columns for each
tuning parameter identifier (e.g. \code{"my_param"} if \code{tune("my_param")} was used).
If \code{NULL}, this argument will be set to
\code{\link[=select_best.tune_results]{select_best(metric, eval_time)}}.
If not \code{NULL}, \code{parameters} overwrites the specification via \code{metric}, and
\code{eval_time}.}

\item{verbose}{A logical for printing logging.}

\item{add_validation_set}{When the resamples embedded in \code{x} are a split into
training set and validation set, should the validation set be included in the
data set used to train the model? If not, only the training set is used. If
\code{NULL}, the validation set is not used for resamples originating from
\code{\link[rsample:validation_set]{rsample::validation_set()}} while it is used for resamples originating
from \code{\link[rsample:validation_split]{rsample::validation_split()}}.}
}
\value{
A fitted workflow.
}
\description{
\code{fit_best()} takes the results from model tuning and fits it to the training
set using tuning parameters associated with the best performance.
}
\details{
This function is a shortcut for the manual steps of:

\preformatted{
  best_param <- select_best(tune_results, metric) # or other `select_*()`
  wflow <- finalize_workflow(wflow, best_param)  # or just `finalize_model()`
  wflow_fit <- fit(wflow, data_set)
}
}
\section{Case Weights}{

Some models can utilize case weights during training. tidymodels currently
supports two types of case weights: importance weights (doubles) and
frequency weights (integers). Frequency weights are used during model
fitting and evaluation, whereas importance weights are only used during
fitting.

To know if your model is capable of using case weights, create a model spec
and test it using \code{\link[parsnip:case_weights_allowed]{parsnip::case_weights_allowed()}}.

To use them, you will need a numeric column in your data set that has been
passed through either \code{\link[hardhat:importance_weights]{hardhat:: importance_weights()}} or
\code{\link[hardhat:frequency_weights]{hardhat::frequency_weights()}}.

For functions such as \code{\link[=fit_resamples]{fit_resamples()}} and the \verb{tune_*()} functions, the
model must be contained inside of a \code{\link[workflows:workflow]{workflows::workflow()}}. To declare that
case weights are used, invoke \code{\link[workflows:add_case_weights]{workflows::add_case_weights()}} with the
corresponding (unquoted) column name.

From there, the packages will appropriately handle the weights during model
fitting and (if appropriate) performance estimation.
}

\section{See also}{


\code{\link[=last_fit]{last_fit()}} is closely related to \code{\link[=fit_best]{fit_best()}}. They both
give you access to a workflow fitted on the training data but are situated
somewhat differently in the modeling workflow. \code{\link[=fit_best]{fit_best()}} picks up
after a tuning function like \code{\link[=tune_grid]{tune_grid()}} to take you from tuning results
to fitted workflow, ready for you to predict and assess further. \code{\link[=last_fit]{last_fit()}}
assumes you have made your choice of hyperparameters and finalized your
workflow to then take you from finalized workflow to fitted workflow and
further to performance assessment on the test data. While \code{\link[=fit_best]{fit_best()}} gives
a fitted workflow, \code{\link[=last_fit]{last_fit()}} gives you the performance results. If you
want the fitted workflow, you can extract it from the result of \code{\link[=last_fit]{last_fit()}}
via \link[=extract_workflow.tune_results]{extract_workflow()}.
}

\examples{
\dontshow{if (tune:::should_run_examples() && rlang::is_installed("modeldata") && !tune:::is_cran_check()) withAutoprint(\{ # examplesIf}
library(recipes)
library(rsample)
library(parsnip)
library(dplyr)

data(meats, package = "modeldata")
meats <- meats |> select(-water, -fat)

set.seed(1)
meat_split <- initial_split(meats)
meat_train <- training(meat_split)
meat_test  <- testing(meat_split)

set.seed(2)
meat_rs <- vfold_cv(meat_train, v = 10)

pca_rec <-
  recipe(protein ~ ., data = meat_train) |>
  step_normalize(all_numeric_predictors()) |>
  step_pca(all_numeric_predictors(), num_comp = tune())

knn_mod <- nearest_neighbor(neighbors = tune()) |> set_mode("regression")

ctrl <- control_grid(save_workflow = TRUE)

set.seed(128)
knn_pca_res <-
  tune_grid(knn_mod, pca_rec, resamples = meat_rs, grid = 10, control = ctrl)

knn_fit <- fit_best(knn_pca_res, verbose = TRUE)
predict(knn_fit, meat_test)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.stash_last_result}
\alias{.stash_last_result}
\title{Save most recent results to search path}
\usage{
.stash_last_result(x)
}
\arguments{
\item{x}{An object.}
}
\value{
NULL, invisibly.
}
\description{
Save most recent results to search path
}
\details{
The function will assign \code{x} to \code{.Last.tune.result} and put it in
the search path.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.R
\name{merge.recipe}
\alias{merge.recipe}
\alias{merge.model_spec}
\title{Merge parameter grid values into objects}
\usage{
\method{merge}{recipe}(x, y, ...)

\method{merge}{model_spec}(x, y, ...)
}
\arguments{
\item{x}{A recipe or model specification object.}

\item{y}{A data frame or a parameter grid resulting from one of the
\verb{grid_*} functions. The column names should correspond to the parameter
names (or their annotations) in the object.}

\item{...}{Not used but required for S3 completeness.}
}
\value{
A tibble with a column \code{x} that has as many rows as were in \code{y}.
}
\description{
\code{merge()} can be used to easily update any of the arguments in a
\pkg{parsnip} model or recipe.
}
\examples{
\dontshow{if (tune:::should_run_examples(suggests = c("xgboost", "modeldata", "splines2"))) withAutoprint(\{ # examplesIf}
library(tibble)
library(recipes)
library(parsnip)
library(dials)

pca_rec <-
  recipe(mpg ~ ., data = mtcars) |>
  step_impute_knn(all_predictors(), neighbors = tune()) |>
  step_pca(all_predictors(), num_comp = tune())

pca_grid <-
  tribble(
    ~neighbors, ~num_comp,
             1,         1,
             5,         1,
             1,         2,
             5,         2
  )

merge(pca_rec, pca_grid)

spline_rec <-
  recipe(mpg ~ ., data = mtcars) |>
  step_spline_natural(disp, deg_free = tune("disp df")) |>
  step_spline_natural(wt, deg_free = tune("wt df"))

spline_grid <-
  tribble(
    ~"disp df", ~ "wt df",
    3,         3,
    5,         3,
    3,         5,
    5,         5
  )

merge(pca_rec, pca_grid)

data(hpc_data, package = "modeldata")

xgb_mod <-
  boost_tree(trees = tune(), min_n = tune()) |>
  set_engine("xgboost")

set.seed(254)
xgb_grid <-
  extract_parameter_set_dials(xgb_mod) |>
  finalize(hpc_data) |>
  grid_max_entropy(size = 3)

merge(xgb_mod, xgb_grid)
\dontshow{\}) # examplesIf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control_bayes}
\alias{control_bayes}
\title{Control aspects of the Bayesian search process}
\usage{
control_bayes(
  verbose = FALSE,
  verbose_iter = FALSE,
  no_improve = 10L,
  uncertain = Inf,
  seed = sample.int(10^5, 1),
  extract = NULL,
  save_pred = FALSE,
  time_limit = NA,
  pkgs = NULL,
  save_workflow = FALSE,
  save_gp_scoring = FALSE,
  event_level = "first",
  parallel_over = NULL,
  backend_options = NULL,
  allow_par = TRUE,
  workflow_size = 100
)
}
\arguments{
\item{verbose}{A logical for logging results (other than warnings and errors,
which are always shown) as they are generated during training in a single
R process. When using most parallel backends, this argument typically will
not result in any logging. If using a dark IDE theme, some logging messages
might be hard to see; try setting the \code{tidymodels.dark} option with
\code{options(tidymodels.dark = TRUE)} to print lighter colors.}

\item{verbose_iter}{A logical for logging results of the Bayesian search
process. Defaults to FALSE. If using a dark IDE theme, some logging
messages might be hard to see; try setting the \code{tidymodels.dark} option
with \code{options(tidymodels.dark = TRUE)} to print lighter colors.}

\item{no_improve}{The integer cutoff for the number of iterations without
better results.}

\item{uncertain}{The number of iterations with no improvement before an
uncertainty sample is created where a sample with high predicted variance is
chosen (i.e., in a region that has not yet been explored). The iteration
counter is reset after each uncertainty sample. For example, if \code{uncertain = 10}, this condition is triggered every 10 samples with no improvement.}

\item{seed}{An integer for controlling the random number stream. Tuning
functions are sensitive to both the state of RNG set outside of tuning
functions with \code{set.seed()} as well as the value set here. The value of the
former determines RNG for the higher-level tuning process, like grid
generation and setting the value of this argument if left as default. The
value of this argument determines RNG state in workers for each iteration
of model fitting, determined by the value of \code{parallel_over}.}

\item{extract}{An optional function with at least one argument (or \code{NULL})
that can be used to retain arbitrary objects from the model fit object,
recipe, or other elements of the workflow.}

\item{save_pred}{A logical for whether the out-of-sample predictions should
be saved for each model \emph{evaluated}.}

\item{time_limit}{A number for the minimum number of \emph{minutes} (elapsed) that
the function should execute. The elapsed time is evaluated at internal
checkpoints and, if over time, the results at that time are returned (with
a warning). This means that the \code{time_limit} is not an exact limit, but a
minimum time limit.

Note that timing begins immediately on execution. Thus, if the
\code{initial} argument to \code{\link[=tune_bayes]{tune_bayes()}} is supplied as a number, the elapsed
time will include the time needed to generate initialization results.}

\item{pkgs}{An optional character string of R package names that should be
loaded (by namespace) during parallel processing.}

\item{save_workflow}{A logical for whether the workflow should be appended
to the output as an attribute.}

\item{save_gp_scoring}{A logical to save the intermediate Gaussian process
models for each iteration of the search. These are saved to
\code{tempdir()} with names \verb{gp_candidates_\{i\}.RData} where \code{i} is the iteration.
These results are deleted when the R session ends. This option is only
useful for teaching purposes.}

\item{event_level}{A single string containing either \code{"first"} or \code{"second"}.
This argument is passed on to yardstick metric functions when any type
of class prediction is made, and specifies which level of the outcome
is considered the "event".}

\item{parallel_over}{A single string containing either \code{"resamples"} or
\code{"everything"} describing how to use parallel processing. Alternatively,
\code{NULL} is allowed, which chooses between \code{"resamples"} and \code{"everything"}
automatically.

If \code{"resamples"}, then tuning will be performed in parallel over resamples
alone. Within each resample, the preprocessor (i.e. recipe or formula) is
processed once, and is then reused across all models that need to be fit.

If \code{"everything"}, then tuning will be performed in parallel at two levels.
An outer parallel loop will iterate over resamples. Additionally, an
inner parallel loop will iterate over all unique combinations of
preprocessor and model tuning parameters for that specific resample. This
will result in the preprocessor being re-processed multiple times, but
can be faster if that processing is extremely fast.

If \code{NULL}, chooses \code{"resamples"} if there are more than one resample,
otherwise chooses \code{"everything"} to attempt to maximize core utilization.

Note that switching between \code{parallel_over} strategies is not guaranteed
to use the same random number generation schemes. However, re-tuning a
model using the same \code{parallel_over} strategy is guaranteed to be
reproducible between runs.}

\item{backend_options}{An object of class \code{"tune_backend_options"} as created
by \code{tune::new_backend_options()}, used to pass arguments to specific tuning
backend. Defaults to \code{NULL} for default backend options.}

\item{allow_par}{A logical to allow parallel processing (if a parallel
backend is registered).}

\item{workflow_size}{A non-negative number (in MB) that is used as a
threshold for a warning regarding the size of the workflow. Only used when
\code{save_workflow = TRUE}.}
}
\description{
Control aspects of the Bayesian search process
}
\details{
For \code{extract}, this function can be used to output the model object, the
recipe (if used), or some components of either or both. When evaluated, the
function's sole argument has a fitted workflow If the formula method is used,
the recipe element will be \code{NULL}.

The results of the \code{extract} function are added to a list column in the
output called \code{.extracts}. Each element of this list is a tibble with tuning
parameter column and a list column (also called \code{.extracts}) that contains
the results of the function. If no extraction function is used, there is no
\code{.extracts} column in the resulting object. See \code{\link[=tune_bayes]{tune_bayes()}} for more
specific details.

Note that for \code{\link[=collect_predictions]{collect_predictions()}}, it is possible that each row of the
original data point might be represented multiple times per tuning
parameter. For example, if the bootstrap or repeated cross-validation are
used, there will be multiple rows since the sample data point has been
evaluated multiple times. This may cause issues when merging the predictions
with the original data.
}
\section{Hyperparameters and extracted objects}{


When making use of submodels, tune can generate predictions and calculate
metrics for multiple model \code{.config}urations using only one model fit.
However, this means that if a function was supplied to a
\link[=control_grid]{control function's} \code{extract} argument, tune can only
execute that extraction on the one model that was fitted. As a result,
in the \code{collect_extracts()} output, tune opts to associate the
extracted objects with the hyperparameter combination used to
fit that one model workflow, rather than the hyperparameter
combination of a submodel. In the output, this appears like
a hyperparameter entry is recycled across many \code{.config}
entries---this is intentional.

See \url{https://parsnip.tidymodels.org/articles/Submodels.html} to learn
more about submodels.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{example_ames_knn}
\alias{example_ames_knn}
\alias{ames_wflow}
\alias{ames_grid_search}
\alias{ames_iter_search}
\title{Example Analysis of Ames Housing Data}
\value{
\item{ames_wflow}{A workflow object}
\item{ames_grid_search,ames_iter_search}{Results of model tuning. }
}
\description{
Example Analysis of Ames Housing Data
}
\details{
These objects are the results of an analysis of the Ames
housing data. A K-nearest neighbors model was used with a small
predictor set that included natural spline transformations of
the \code{Longitude} and \code{Latitude} predictors. The code used to
generate these examples was:

\if{html}{\out{<div class="sourceCode">}}\preformatted{library(tidymodels)
library(tune)
library(AmesHousing)

# ------------------------------------------------------------------------------

ames <- make_ames()

set.seed(4595)
data_split <- initial_split(ames, strata = "Sale_Price")

ames_train <- training(data_split)

set.seed(2453)
rs_splits <- vfold_cv(ames_train, strata = "Sale_Price")

# ------------------------------------------------------------------------------

ames_rec <-
  recipe(Sale_Price ~ ., data = ames_train) |>
  step_log(Sale_Price, base = 10) |>
  step_YeoJohnson(Lot_Area, Gr_Liv_Area) |>
  step_other(Neighborhood, threshold = .1)  |>
  step_dummy(all_nominal()) |>
  step_zv(all_predictors()) |>
  step_spline_natural(Longitude, deg_free = tune("lon")) |>
  step_spline_natural(Latitude, deg_free = tune("lat"))

knn_model <-
  nearest_neighbor(
    mode = "regression",
    neighbors = tune("K"),
    weight_func = tune(),
    dist_power = tune()
  ) |>
  set_engine("kknn")

ames_wflow <-
  workflow() |>
  add_recipe(ames_rec) |>
  add_model(knn_model)

ames_set <-
  extract_parameter_set_dials(ames_wflow) |>
  update(K = neighbors(c(1, 50)))

set.seed(7014)
ames_grid <-
  ames_set |>
  grid_max_entropy(size = 10)

ames_grid_search <-
  tune_grid(
    ames_wflow,
    resamples = rs_splits,
    grid = ames_grid
  )

set.seed(2082)
ames_iter_search <-
  tune_bayes(
    ames_wflow,
    resamples = rs_splits,
    param_info = ames_set,
    initial = ames_grid_search,
    iter = 15
  )
}\if{html}{\out{</div>}}

\strong{important note}: Since the \code{rsample} split columns contain a reference
to the same data, saving them to disk can results in large object sizes when
the object is later used. In essence, R replaces all of those references with
the actual data. For this reason, we saved zero-row tibbles in their place.
This doesn't affect how we use these objects in examples but be advised that
using some \code{rsample} functions on them will cause issues.
}
\examples{
library(tune)

ames_grid_search
ames_iter_search
}
\keyword{datasets}
